/* 
* Copyright 1996 Regents of the University of California.
* All rights reserved.
*/

/* hip~ - simplest hipass filter. */

#include "m_extern.h"
#include "d_graph.h"
#include "d_ugen.h"
#include "math.h"

typedef struct hipctl
{
    float c_x;
    float c_coef;
} t_hipctl;

typedef struct sighip
{
    t_head x_h;
    t_sig *x_io[IN1+OUT1];
    float x_sr;
    float x_hz;
    t_hipctl x_cspace;
    t_hipctl *x_ctl;
} t_sighip;

void cu_sighip(float *in, float *out, t_hipctl *c, long n)
{
    int i;
    float last = c->c_x;
    for (i = 0; i < n; i++)
    {
	float new = *in++ + c->c_coef * last;
	*out++ = new - last;
	last = new;
    }
    c->c_x = last;
}

void sighip_ft1(t_sighip *x, double f)
{
    if (f < 0.001) f = 10;
    x->x_hz = f;
    x->x_ctl->c_coef = 1 - f * (2 * 3.14159) / x->x_sr;
}

void sighip_put(t_sighip *x, long whether)
{
    if (whether)
    {
	u_stdout(x);
	x->x_ctl = dspchain_addstate(&x->x_cspace,
	    sizeof(t_hipctl), sizeof(float));
	x->x_sr = x->x_io[0]->s_sr;
	sighip_ft1(x,  x->x_hz);
	dspchain_addc(cu_sighip, 4,
	    x->x_io[0]->s_shit, x->x_io[1]->s_shit, 
		x->x_ctl, x->x_io[0]->s_n);
    }
    else
    {
	x->x_cspace = *(x->x_ctl);
	x->x_ctl = &x->x_cspace;
    }
}

t_externclass *sighip_class;

void *sighip_new(double f)
{
    t_sighip *x = (t_sighip *)obj_new(&sighip_class, 0);
    inlet_new(x, x, gensym("float"), gensym("ft1"));
    u_setup(x, IN1, OUT1);
    x->x_sr = 44100;
    x->x_ctl = &x->x_cspace;
    x->x_cspace.c_x = 0;
    sighip_ft1(x, f);
    return (x);
}

void sighip_setup(void)
{
    c_extern(&sighip_class, sighip_new, u_clean,
	gensym("hip~"), sizeof(t_sighip), 0, A_FLOAT, 0);
    c_addmess(sighip_put, gensym("put"), A_CANT, 0);
    c_addmess(sighip_ft1, gensym("ft1"), A_FLOAT, 0);
    u_inletmethod(0);
}

