/*
 *      Copyright (c) 1987-1994 IRCAM
 *                      Copyright (c) 1995-1996 UC REGENTS
 *                          All rights reserved.
 */

/*
 * "graphics routines" for Max X/Motif interface
 * by Paul Foley and Miller Puckette
 */

#define DEF_TEXT

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/resource.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Protocols.h>
#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/Text.h>
#include <Xm/ToggleBG.h>

#include "m_mess.h"
#include "cp_mess.h"
#include "h_cp.h"
#include "t_mac.h"
#include "t_wind.h"

#include "p_pic.h"
#include "icon.xbm"
#include "pic0.xbm"
#include "pic1.xbm"

struct _text {
    Widget t_widget;
    char *t_contents;
    char *t_textwhenactivated;
    Rect t_rect;
    int t_fontsize;
    int t_active;
    int t_totalheight;
};

extern int wind_xoffset;
extern int wind_yoffset;

extern Display *display;
extern GC draw_gc;
extern GC undraw_gc;
extern GC invert_gc;

Widget t_getview();

#define MAX_ARGS 20
#define ARGS  args,narg

#ifdef __sgi
#define __STRICT_ANSI__
#endif

#if 1
#define SetArg(n,v) XtSetArg(args[narg],XmN##n,v),narg++
#else
#define SetArg(n,v) XtSetArg(args[narg],XmN/**/n,v),narg++
#endif
#define SetValues(w) XtSetValues(w,ARGS), narg=0
#define GetValues(w) XtGetValues(w,ARGS), narg=0
#define CreateText(p,n) XmCreateText(p,n,ARGS), narg=0

static int narg = 0;
static Arg args[MAX_ARGS];
#define Manage(w) XtManageChild(w)
#define Unmanage(w) XtUnmanageChild(w)

/* current window */
extern t_wind *current_window;

/* Fontsizes are numbers in the range 0..7, and forced somehow to produce fonts 
   in a pointsize corresponding to their entry in pointsizes[] */
#define FONT_COUNT 8
int pointsizes[] = { 8, 10, 12, 14, 16, 18, 20, 24 };

/* The font list contains the fonts as specified by user */
XmFontList fontlist;

/* The application FontList should specify fonts like this: */
#define FONTLIST_NAME "maxfont-%d"      /* e.g., "maxfont-0" */


char *index();

int t_getfontindex(int pointsize)       /* pointsize -> fontsize */
{
    /* Same as NeXT */
    int i = 0;
    while (i < FONT_COUNT - 1 && pointsizes[i] < pointsize)
        i++;
    return i;
}

int t_realfontsize(int fontsize)        /* fontsize -> pointsize */
{
    /* Same as NeXT */
    if (fontsize >= FONT_COUNT)
        fontsize = FONT_COUNT - 1;
    else if (fontsize < 0)
        fontsize = 0;
    return (pointsizes[fontsize]);
}

#define maximum(a,b) ((a > b) ? a : b)

t_setfont(char *font)
{
}                               /* fill in later if you care. */

static int t_currentfontsize;   /* current font index, 0-7 */

void t_textsize(int fontsize)
{
    t_currentfontsize = fontsize;
}

static int t_defaultfontsize;   /* default font index, 0-7 */

int t_setdeffont(int fontsize)
{
    t_textsize(fontsize);
    t_defaultfontsize = fontsize;
}

int t_setfontsize(char *size)   /* call if X resource is found for font size */
{
    /* same as NeXT */
    int i;
    if (sscanf(size, "%d", &i) == 1)
        t_setdeffont(t_getfontindex(i));
}

    /* this is called first.  Font size is set to "1", i.e., 10 pts. */
int t_initdeffont()
{
    char buf[12];
    int size = 1;

    t_setdeffont(size);         /* set default and current fontsize */
}

int t_getdeffont()
{
    /* Same as NeXT */
    return t_defaultfontsize;
}

int t_getcurrentfont()
{
    /* Same as NeXT */
    return t_currentfontsize;
}

int t_deffontheightfor(fontno)  /* don't have a good definition of what this means */
{
    char charset[12];
    XmString string;
    int height;

    sprintf(charset, FONTLIST_NAME, fontno);    /* identify current charset */
    string = XmStringCreate("A", charset);      /* create a string in it */
    height = XmStringHeight(fontlist, string);  /* get its height */
    XmStringFree(string);       /* free compound string */
    return height;
}

int t_charwidth(int c)
{
    char buf[2];
    char charset[12];
    XmString string;
    int width;

    sprintf(buf, "%c", c);
    sprintf(charset, FONTLIST_NAME, t_currentfontsize); /* identify current charset */
    string = XmStringCreate(buf, charset);      /* create a string in it */
    width = XmStringWidth(fontlist, string);    /* get its height */
    XmStringFree(string);       /* free compound string */
    return width;
}

int t_stringwidth(char *s)
{
    char charset[12];
    XmString string;
    int width;

    sprintf(charset, FONTLIST_NAME, t_currentfontsize); /* identify current charset */
    string = XmStringCreate(s, charset);        /* create a string in it */
    width = XmStringWidth(fontlist, string);    /* get its width */
    XmStringFree(string);       /* free compound string */
    return width;
}

char **te_text(struct _text *x)
{
    if (x->t_widget)
    {
        if (x->t_contents)
            XtFree(x->t_contents);
        x->t_contents = XmTextGetString(x->t_widget);
    }
    if (!x->t_contents)
        x->t_contents = XtNewString("");
    return (&x->t_contents);
}

int te_length(t_text *x)
{
    char *contents = *te_text(x);
    int length = strlen(contents);
    return (length);
}

void te_set(t_text *x, char **h, long int n)
{
    char buf[257];
    char *foo = (n > 256 ? malloc(n + 1) : buf);
    char *src = *h;
    int i;
    if (foo)
    {
        for (i = 0; i < n; i++)
            foo[i] = src[i];
        foo[n] = 0;
        if (x->t_widget)
            XmTextSetString(x->t_widget, foo);
        else
        {
            if (x->t_contents)
                XtFree(x->t_contents);
            x->t_contents = XtNewString(foo);
        }
        if (n > 256)
            free(foo);
    }
}

void te_setview(t_text *x, Rect *r)
{
    x->t_rect = *r;
    if (x->t_widget)
    {
        SetArg(x, r->left + wind_xoffset);
        SetArg(y, r->top + wind_yoffset);
        SetArg(width, r->right - r->left);
        /* SetArg(height, r->bottom - r->top); */
        SetValues(x->t_widget);
    }
}

#define MAXW 80
int te_wantwidth(t_text *x, int doit)
{
    char buf[MAXW + 5];
    int maxw = 15;
    char *nextchar = *te_text(x);
    int fontwas = t_currentfontsize;
    t_currentfontsize = x->t_fontsize;
    while (*nextchar)
    {
        int position = 0, width;
        char *bp = buf;
        while (*nextchar && isspace(*nextchar))
            nextchar++;
        while (position <= 80 && *nextchar && *nextchar != '\n')
        {
            *bp++ = *nextchar++;
            *bp = 0;
            position++;
        }
        width = t_stringwidth(buf);
        if (width > maxw)
            maxw = width;
    }
    t_currentfontsize = fontwas;
    return (maxw);
}

/* this routine figures out line breaks and (optionally) draws the text
 * 
 */
void te_doupdate(t_text *x, int doit)
{
    char buf[4096], *bp = buf;
    char *space;
    char *intext;
    int width = x->t_rect.right - x->t_rect.left, line = 1;
    int fontwas = t_currentfontsize;
    int lineheight = t_lineheight(x->t_fontsize);
    t_currentfontsize = x->t_fontsize;
    intext = *te_text(x);
#ifdef CHARWRAP                 /* old code which wrapped chars; not used any more. */
    while (1)
    {
        space = 0;
        bp[0] = *intext;
        bp[1] = 0;
        if (bp[0] == '\n' || bp[0] == '\0')
        {
            bp[0] = 0;
            if (*buf)
            {
                if (doit)
                    t_drawstring(x->t_rect.left,
                                 x->t_rect.top + line * lineheight, buf);
                line++;
            }
            bp = buf;
            if (*intext == 0)
                break;
        }
        else if (buf[1] && t_stringwidth(buf) > width)
        {
            bp[0] = 0;
            intext--;
            if (doit)
                t_drawstring(x->t_rect.left,
                             x->t_rect.top + line * lineheight, buf);
            line++;
            bp = buf;
        }
        else
            bp++;
        intext++;
    }
#else
    while (1)
    {
        space = 0;
        bp[0] = *intext;
        bp[1] = 0;
        intext++;
        /* on carriage return or null, write out the line */
        if (bp[0] == '\n' || bp[0] == '\0')
        {
            bp[0] = 0;
            if (*buf)
            {
                if (doit)
                    t_drawstring(x->t_rect.left,
                                 x->t_rect.top + line * lineheight, buf);
                line++;
            }
            bp = buf;
            if (intext[-1] == 0)
                break;
        }
        /* if we've overgrown the width of the text box, we must print
         * something out and start a new line out of the rest.
         */
        else if (buf[1] && t_stringwidth(buf) > width)
        {
            int lastspace = 0, nextspace = isspace(bp[0]);
            char *bp2;
            /* put back the last character */
            intext--;
            bp[0] = 0;
            bp--;
            /* find last space */
            for (bp2 = bp; bp2 >= buf; bp2--)
            {
                char c = *bp2;
                if (isspace(c))
                {
                    lastspace = 1;
                    break;
                }
            }
            /* if there was a space, put everything after it back in buf */
            if (lastspace)
            {
                intext -= (bp - bp2);
                bp2[1] = 0;
            }
            if (doit)
                t_drawstring(x->t_rect.left,
                             x->t_rect.top + line * lineheight, buf);
            line++;
            bp = buf;
            /* eat leading spaces from next line */
            while (*intext && isspace(*intext))
                intext++;
        }
        else
            bp++;
    }
#endif
    t_currentfontsize = fontwas;
    x->t_totalheight = (line == 1 ? 1 : (line - 1)) * lineheight;
}

int te_height(t_text *x)
{
    if (x->t_widget)
    {
        Dimension h;
        SetArg(height, &h);
        GetValues(x->t_widget);
        return (h);
    }
    else
    {
        te_doupdate(x, 0);
        return (x->t_totalheight);
    }
}

int t_lineheight(fontno)
{
    char charset[12];
    XmString string;
    int height;

    sprintf(charset, FONTLIST_NAME, fontno);    /* identify current charset */
    string = XmStringCreate("fg", charset);     /* create a string in it */
    height = XmStringHeight(fontlist, string);  /* get its height */
    XmStringFree(string);       /* free compound string */
    return height;
}

void te_update(t_text *x)
{
    te_doupdate(x, 1);
}

static int textchanged;

int te_changed()
{
    /* wrong */
    return (textchanged);
}

void te_activate(t_text *x)
{
    int foo = 0;
    /* post("activate %lx",  x); */
    if (x->t_active)
        return;
    SetArg(editable, True);
    SetArg(editMode, XmMULTI_LINE_EDIT);
    SetArg(wordWrap, True);
    SetArg(resizeWidth, True);
    SetArg(resizeHeight, True);
    SetArg(shadowThickness, 0);
    SetArg(maxLength, /*MAXTEXT*/ 4096);
    SetArg(marginHeight, 0);
    SetArg(marginWidth, 0);
    SetArg(x, x->t_rect.left + wind_xoffset);
    SetArg(y, x->t_rect.top + wind_yoffset);
    SetArg(width, x->t_rect.right - x->t_rect.left);
    /* create text widget; not managed yet */
    x->t_widget = CreateText(t_getview(), "textBox");
    if (x->t_contents)
        XmTextSetString(x->t_widget, x->t_contents);
    XtManageChild(x->t_widget);
    x->t_active = 1;
    textchanged = 0;
    if (x->t_textwhenactivated)
        XtFree(x->t_textwhenactivated);
    x->t_textwhenactivated = XmTextGetString(x->t_widget);
    /* give text widget immediate focus */
#ifdef MOTIF_1_0
    undocumented function ... _XmGrabTheFocus(x->t_widget);
#else
    /* There may be problems with this as noted in the Motif FAQ ... */
    if (!XmProcessTraversal(x->t_widget, XmTRAVERSE_CURRENT))
        bug("XmProcessTraversal 1");
#endif
    XmTextSetSelection(x->t_widget, 0, strlen(x->t_contents), 0);
    editcursor();
}

void te_deactivate(t_text *x)
{
    if (!x->t_active)
        return;
    /* post("deactivate %lx",  x); */
    /* if (!XmProcessTraversal(XtParent(x->t_widget), XmTRAVERSE_CURRENT))
       bug("XmProcessTraversal 2"); */
    XtUnmanageChild(x->t_widget);
    /* XtUnrealizeWidget(x->t_widget); */
    x->t_active = 0;
    (void) te_text(x);
    textchanged = (!x->t_textwhenactivated) ||
        strcmp(x->t_textwhenactivated, x->t_contents);
    /* post("text %s -> %s",  x->t_textwhenactivated, x->t_contents); */
    XtDestroyWidget(x->t_widget);
    x->t_widget = 0;
    wind_setpollon();
}

t_text *te_new(Rect *destrect, Rect *viewrect)
{
    t_text *x = (t_text *) getbytes(sizeof(t_text));
    x->t_rect = *viewrect;
    x->t_active = 0;
    x->t_fontsize = t_currentfontsize;

    x->t_totalheight = 0;
    x->t_contents = 0;
    x->t_textwhenactivated = 0;
    x->t_widget = 0;
    return x;
}

void te_free(t_text *x)
{
    if (x->t_contents)
        XtFree(x->t_contents);
    if (x->t_textwhenactivated)
        XtFree(x->t_textwhenactivated);
    if (x->t_widget)
        XtDestroyWidget(x->t_widget);
    freebytes(x, sizeof(t_text));
}

void t_drawstring(int a, int b, char *s)
{
    char charset[12];
    XmString string;
    int width, height;

    sprintf(charset, FONTLIST_NAME, t_currentfontsize); /* identify current charset */
    string = XmStringCreate(s, charset);        /* create a string in it */
    width = XmStringWidth(fontlist, string);    /* get its width */
    height = XmStringHeight(fontlist, string);  /* get its height */
    XmStringDraw(display, XtWindow(t_getview()), fontlist, string, draw_gc,
                 a + wind_xoffset, b + wind_yoffset - height,
                 width, XmALIGNMENT_BEGINNING,
                 XmSTRING_DIRECTION_L_TO_R, NULL);
    XmStringFree(string);       /* free compound string */
}

void t_drawlong(int a, int b, long int n)
{
    /* Same as NeXT */
    char buf[20];
    t_printi(buf, n);
    t_drawstring(a, b, buf);
}

void te_key(t_text *x, int k)
{
    /* Same as NeXT */
}

void te_click(t_text *x, Point pt, int shift)
{
    /* Same as NeXT */
}

void te_delete(t_text *x)
{
    /* Same as NeXT */
}

void te_copy(t_text *x)
{
    /* Same as NeXT */
}

void te_paste(t_text *x)
{
    /* Same as NeXT */
}

void te_idle(t_text *x)
{
    /* Same as NeXT */
}

long te_selstart(t_text *x)
{
    /* Same as NeXT */
    return 0;
}

long te_selend(t_text *x)
{
    /* Same as NeXT */
    return 0;
}

void te_setselect(t_text *x, long int b, long int e)
{
    /* Same as NeXT */
}

Rect *te_destrect(t_text *x)
{
    /* same as NeXT */
    return &(x->t_rect);
}

Rect *te_viewrect(t_text *x)
{
    /* same as NeXT */
    return &(x->t_rect);
}

void te_setdest(t_text *x, Rect *r)
{
    /* same as NeXT */
}

/* other graphics code */

void t_setrect(Rect *r, int a, int b, int c, int d)
{
    /* same as NeXT */
    r->left = a;
    r->top = b;
    r->right = c;
    r->bottom = d;
}

void t_offsetrect(Rect *r, int a, int b)
{
    /* same as NeXT */
    r->left += a;
    r->right += a;
    r->top += b;
    r->bottom += b;
}

int t_ptinrect(Point pt, Rect *r)
{
    /* same as NeXT */
    return (pt.h >= r->left && pt.h < r->right &&
            pt.v >= r->top && pt.v < r->bottom);
}

#define maximum(a,b) ((a > b) ? a : b)

void t_framerect(a, b, c, d)
{
    XDrawRectangle(display, XtWindow(t_getview()), draw_gc,
                   a + wind_xoffset, b + wind_yoffset, maximum(c - a - 1,
                                                               1),
                   maximum(d - b - 1, 1));
}

void t_unframerect(a, b, c, d)
{
    XDrawRectangle(display, XtWindow(t_getview()), undraw_gc,
                   a + wind_xoffset, b + wind_yoffset, maximum(c - a - 1,
                                                               1),
                   maximum(d - b - 1, 1));
}

void t_paintrect(a, b, c, d)
{
    XFillRectangle(display, XtWindow(t_getview()), draw_gc,
                   a + wind_xoffset, b + wind_yoffset, maximum(c - a, 1),
                   maximum(d - b, 1));
}

void t_eraserect(a, b, c, d)
{
    XFillRectangle(display, XtWindow(t_getview()), undraw_gc,
                   a + wind_xoffset, b + wind_yoffset, maximum(c - a, 1),
                   maximum(d - b, 1));
}

void t_invertrect(a, b, c, d)
{
    XFillRectangle(display, XtWindow(t_getview()), invert_gc,
                   a + wind_xoffset, b + wind_yoffset, maximum(c - a, 1),
                   maximum(d - b, 1));
}

void t_hiliterect(a, b, c, d)
{
    console_msg("t_hiliterect() on %s\n", XtName(t_getview()));
}

void t_frameoval(a, b, c, d)
{
    XDrawArc(display, XtWindow(t_getview()), draw_gc,
             a + wind_xoffset, b + wind_yoffset, c - a, d - b, 0,
             -(180 * 64));
    XDrawArc(display, XtWindow(t_getview()), draw_gc, a + wind_xoffset,
             b + wind_yoffset, c - a, d - b, 0, (180 * 64));
}

void t_line(a, b, c, d)
{
    XDrawLine(display, XtWindow(t_getview()), draw_gc,
              a + wind_xoffset, b + wind_yoffset,
              c + wind_xoffset, d + wind_yoffset);
}

void t_eraseline(a, b, c, d)
{
    XDrawLine(display, XtWindow(t_getview()), undraw_gc,
              a + wind_xoffset, b + wind_yoffset,
              c + wind_xoffset, d + wind_yoffset);
}

void t_penmode(int n)
{
    /* same as NeXT */
}

static int lwidth = 1;

void t_linewidth(int n)
{
    lwidth = n;
    /* Set line width.  Use the default Line, Cap and Join styles */
    XSetLineAttributes(display, draw_gc, lwidth, LineSolid, CapButt,
                       JoinMiter);
    XSetLineAttributes(display, undraw_gc, lwidth, LineSolid, CapButt,
                       JoinMiter);
}

void t_textfont(int n)
{
    /* same as NeXT */
}

/* end of t_gfx.c */
/* ============== */


/* from nx_bitmap.m */
/* ================ */

Pixmap icon_pixmap;             /* Application icon */
static Pixmap pic0_pixmap;      /* Palette */
static Pixmap pic1_pixmap;      /* CPU boxes */
static Pixmap pic10_pixmap, pic11_pixmap, pic12_pixmap,
    pic14_pixmap, pic15_pixmap, pic17_pixmap, pic18_pixmap, pic19_pixmap;
extern char icon_bitmap10[], icon_bitmap11[], icon_bitmap12[],
    icon_bitmap14[], icon_bitmap15[], icon_bitmap17[], icon_bitmap18[],
    icon_bitmap19[];
extern int icon_width10, icon_width11, icon_width12,
    icon_width14, icon_width15, icon_width17, icon_width18, icon_width19;
extern int icon_height10, icon_height11, icon_height12,
    icon_height14, icon_height15, icon_height17, icon_height18,
    icon_height19;

t_initbits()
{
    /* make Bitmaps */
    icon_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bits, icon_width, icon_height);
    pic0_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              pic0_bits, pic0_width, pic0_height);
    pic1_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              pic1_bits, pic1_width, pic1_height);
    pic10_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap10, icon_width10, icon_height10);
    pic11_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap11, icon_width11, icon_height11);
    pic12_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap12, icon_width12, icon_height12);
    pic14_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap14, icon_width14, icon_height14);
    pic15_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap15, icon_width15, icon_height15);
    pic17_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap17, icon_width17, icon_height17);
    pic18_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap18, icon_width18, icon_height18);
    pic19_pixmap =
        XCreateBitmapFromData(display, DefaultRootWindow(display),
                              icon_bitmap19, icon_width19, icon_height19);
}

t_freebits()
{
    XFreePixmap(display, icon_pixmap);
    XFreePixmap(display, pic0_pixmap);
    XFreePixmap(display, pic1_pixmap);
    XFreePixmap(display, pic10_pixmap);
    XFreePixmap(display, pic11_pixmap);
    XFreePixmap(display, pic12_pixmap);
    XFreePixmap(display, pic14_pixmap);
    XFreePixmap(display, pic15_pixmap);
    XFreePixmap(display, pic17_pixmap);
    XFreePixmap(display, pic18_pixmap);
    XFreePixmap(display, pic19_pixmap);
}

t_drawbits(int mapno, int a, int b, int x1, int y1, int x2, int y2)
{
    /* 
       Using (a, b) as origin of bitmap even though it seems to be (1,0)
       for the palette.  Also using x2 - x1 rather than picX_width and
       y2 - y1 rather than picX_height.  
     */
    Pixmap whichmap = 0;
    switch (mapno)
    {
    case PICPATCH:
        whichmap = pic0_pixmap;
        break;                  /* palette */
    case PICSEND:
        whichmap = pic1_pixmap;
        break;                  /* boxes */
    case 10:
        whichmap = pic10_pixmap;
        break;
    case 11:
        whichmap = pic11_pixmap;
        break;
    case 12:
        whichmap = pic12_pixmap;
        break;
    case 14:
        whichmap = pic14_pixmap;
        break;
    case 15:
        whichmap = pic15_pixmap;
        break;
    case 17:
        whichmap = pic17_pixmap;
        break;
    case 18:
        whichmap = pic18_pixmap;
        break;
    case 19:
        whichmap = pic19_pixmap;
        break;
    }
    if (whichmap)
        XCopyPlane(display, whichmap, XtWindow(t_getview()),
                   draw_gc, a, b,
                   x2 - x1, y2 - y1, x1 + wind_xoffset, y1 + wind_yoffset,
                   1);
    else
        t_framerect(x1, y1, x2, y2);
}

/* end of nx_bitmap.m */
/* ================== */
