#ifndef lint
static char rcsid[] =
    "$Revision: 1.21 $ IRCAM $Date: 1994/02/07 11:00:42 $";
#endif

/*
 *      Copyright (c) 1987-1994 IRCAM
 *                      Copyright (c) 1995-1996 UC REGENTS
 *                          All rights reserved.
 */

/*
 * main() for Max X/Motif interface
 * by Paul Foley and Miller Puckette
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Protocols.h>
#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/Text.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>

#include "m_mess.h"
#include "cp_mess.h"
#include "h_cp.h"
#include "t_mac.h"
#include "t_wind.h"

#include "p_pic.h"

/*
 * Widget resources
 */

#define MAX_ARGS 20
#define ARGS  args,narg

#ifdef __sgi
#define __STRICT_ANSI__
#endif

#if 1
#define SetArg(n,v) XtSetArg(args[narg],XmN##n,v),narg++
#else
#define SetArg(n,v) XtSetArg(args[narg],XmN/**/n,v),narg++
#endif
#define SetValues(w) XtSetValues(w,ARGS), narg=0
#define GetValues(w) XtGetValues(w,ARGS), narg=0

static int narg = 0;
static Arg args[MAX_ARGS];

XtAppContext app_context;       /* application context */
Widget toplevel;                /* Top level shell */


/*
 * Application Resources
 */

typedef struct {
    String throttle;            /* Host transfer throttle values */
    Boolean help;               /* Print help and exit */
    Boolean version;            /* Print version and exit */
    String configuration;       /* FTS configuration */
    int sample_rate;            /* Sample rate */
    String search_path;         /* Search path for patches */
    String help_dir;            /* Help patches */
    String font_name;           /* Default font */
    int font_size;              /* Default fontsize */
    String advances;            /* List of scheduler advances */
    String boot_dir;            /* Boot command directory */
    Boolean logfile;            /* Make a logfile */
    String fts_dir;             /* Directory to find FTS */
    Boolean killer;             /* Debugging */
    Boolean nort;               /* non-real-time */
    Boolean acs;                /* use ACS real-time hack */
    String openfile;            /* file to open on startup */
    Boolean copy_to_stderr;     /* Copy Max window to stderr */
    Boolean verbose;            /* Verbose startup */
} MaxResources;

#define Offset(field) (XtOffset(MaxResources*, field))

MaxResources app_resources;     /* application resources */

/*
 * Function declarations 
 */

void pdisplay(const char *s);

/*
 * Globals
 */

char *progname = "max";

extern int infile;
extern int outfile;
extern ftsType XFts;
extern char max_version[];      /* version info */

extern t_symbol s_;

#ifdef USE_TIMEFN
void max_timefn(void *x, char *buf, int size)
{
    long foo = *(long *) buf;
    restart_settime(foo);
}
#endif

char *getappname()
{
    /* The leading space foils attempts to set Max resources -- FIXME */
    return " Max";
}

void max_startfts()
{
    h_start();
    h_vmess(NIL, gensym("setfts"), "s", gensym(h_getftsdir()));
#ifdef USE_TIMEFN
    h_vmess(NIL, gensym("#_settimefn_"), "l", max_timefn);
#endif
}

void max_openfile(s)
char *s;
{
    char buf[4096], *b2 = buf;
    if (*s == '/')
        strcpy(buf, s);
    else
    {
        FILE *fd = popen("pwd", "r");
        int i;
        if (fd)
        {
            while ((i = getc(fd)) != -1)
                *b2++ = i;
            pclose(fd);
        }
        if (b2[-1] == '\n')
            b2--;
        *b2++ = '/';
        *b2 = 0;
        strcat(buf, s);
    }
    openfrompath(buf);
}


/* 
 * System call error messages
 *
 * Used in perror() and hate_perror() calls.
 * Only used in max_forkfts() --- should change rest of code too.
 */

#define MAXERR_CREATE_PIPE  "Cannot create pipe"
#define MAXERR_NONBLOCK   "Cannot set stream non-blocking"
#define MAXERR_DUP    "Cannot duplicate file descriptor"
#define MAXERR_CATCH_SIGPIPE  "Cannot catch SIGPIPE"
#define MAXERR_CATCH_SIGCHLD  "Cannot catch SIGCHLD"
#define MAXERR_FORK     "Cannot fork"
#define MAXERR_EXEC     "Cannot exec"

static void hate_perror(char *message)
{
    /* perror() -like message on dialog */
    hate("%s: %s", message, strerror(errno));
}

/*
 * max_forkfts()
 *
 * Routine to fork & exec FTS.
 * IPC is set up as follows:
 *
 * |          MAX            |
 * |                         |
 * | inpipe[0]    outpipe[1] |
 * \-------------------------/
 *     ^^            ||
 *     ||            vv
 * /-------------------------\
 * | inpipe[1]    outpipe[0] |
 * | (stdout)     (stdin)    |
 * |                         |
 * |          FTS            |
 *
 *
 * We provide aliases for the pipe names.
 * MAX: inpipe[0] == infile; outpipe[1] == outfile
 * FTS: inpipe[1] == outfile; outpipe[0] == infile
 *
 * To make the connection with the exec'd process we have
 * to duplicate the file descriptors onto stdin and stdout.
 */

void max_forkfts()
{
    int fts_pid;                /* FTS process ID */
    int inpipe[2], outpipe[2];
    char sbuf[4096];

    /* Set up pipes */
    if (pipe(inpipe) == -1)
    {
        perror(MAXERR_CREATE_PIPE);
        exit(1);
    }
    if (pipe(outpipe) == -1)
    {
        perror(MAXERR_CREATE_PIPE);
        exit(1);
    }

    /* Max's input and output streams must be non-blocking for fast feedback */
    if (fcntl(inpipe[0], F_SETFL, O_NONBLOCK) == -1)
    {
        perror(MAXERR_NONBLOCK);
        exit(1);
    }
    if (fcntl(outpipe[1], F_SETFL, O_NONBLOCK) == -1)
    {
        perror(MAXERR_NONBLOCK);
        exit(1);
    }

    /* Ignore pipe errors */
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
    {
        perror(MAXERR_CATCH_SIGPIPE);
        exit(1);
    }

    /* fork */
#define FTS_IMAGE "fts"
    if ((fts_pid = fork()) == -1)
    {
        perror(MAXERR_FORK);
        exit(1);
    }
    else if (fts_pid == 0)
    {

        /* Child */
        /* ===== */

        /* Close pipe ends not used by child */
        close(outpipe[1]);
        close(inpipe[0]);

        /* Child stdin reads from outpipe[0] -- see diagram */
        if (outpipe[0] != STDIN_FILENO)
        {
            if (dup2(outpipe[0], STDIN_FILENO) == -1)
            {
                perror(MAXERR_DUP);
                exit(1);
            }
            close(outpipe[0]);  /* not needed after dup2() */
        }

        /* Childs stdout writes to inpipe[1] -- see diagram */
        if (inpipe[1] != STDOUT_FILENO)
        {
            if (dup2(inpipe[1], STDOUT_FILENO) == -1)
            {
                perror(MAXERR_DUP);
                exit(1);
            }
            close(inpipe[1]);   /* not needed after dup2() */
        }

        if (app_resources.acs)
        {
            strcpy(sbuf, "/usr/local/bin/ndfts");
            app_resources.nort = 1;
        }
        else
        {
            strcpy(sbuf, h_getftsdir());
            strcat(sbuf, "/fts");
        }

        /* exec FTS */
#if 1
        execl(sbuf, sbuf,
              (app_resources.advances ? app_resources.advances : "0"),
              (app_resources.nort ? "1" : "0"), 0);
#endif
#if 0
        execl("/bin/sh", "/bin/sh", "-c", sbuf,
              (app_resources.advances ? app_resources.advances : "0"),
              (app_resources.nort ? "1" : "0"), 0);
#endif
#if 0
        execl("/usr/bin/dbx", "dbx", sbuf, 0);
#endif

        /* shouldn't get here */
        perror(MAXERR_EXEC);
        exit(1);                /* useless child */
    }

    /* Close pipe ends not used by parent */
    close(outpipe[0]);
    close(inpipe[1]);

    /* Alias pipe ends in parent
     * -- only needed because lots of code refers to `infile' and `outfile'
     */
    outfile = outpipe[1];
    infile = inpipe[0];
}

/* 
 * Xt resources
 *
 * TBW
 */

static XtResource resources[] = {
    { "throttle", "Throttle", XtRString, sizeof(char *),
     Offset(throttle), XtRImmediate, "1:2:3" },
    { "help", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(help), XtRImmediate, False },
    { "version", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(version), XtRImmediate, False },
    { "configuration", "Maxoptions", XtRString, sizeof(char *),
     Offset(configuration), XtRImmediate, NULL },
    { "sampleRate", "Maxoptions", XtRInt, sizeof(int),
     Offset(sample_rate), XtRImmediate, 44100 },
    { "patchPath", "Maxoptions", XtRString, sizeof(char *),
     Offset(search_path), XtRImmediate, NULL },
    { "helpDir", "Maxoptions", XtRString, sizeof(char *),
     Offset(help_dir), XtRImmediate, NULL },
    { "fontName", "Maxoptions", XtRString, sizeof(char *),
     Offset(font_name), XtRImmediate, NULL },
    { "fontSize", "Maxoptions", XtRInt, sizeof(int),
     Offset(font_size), XtRImmediate, NULL },
    { "advances", "Maxoptions", XtRString, sizeof(char *),
     Offset(advances), XtRImmediate, NULL },
    { "bootDir", "Maxoptions", XtRString, sizeof(char *),
     Offset(boot_dir), XtRImmediate, NULL },
    { "logfile", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(logfile), XtRImmediate, False },
    { "ftsDir", "Maxoptions", XtRString, sizeof(char *),
     Offset(fts_dir), XtRImmediate, NULL },
    { "killer", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(killer), XtRImmediate, False },
    { "nort", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(nort), XtRImmediate, False },
    { "acs", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(acs), XtRImmediate, False },
    { "open", "Maxoptions", XtRString, sizeof(char *),
     Offset(openfile), XtRImmediate, NULL },
    { "stderr", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(copy_to_stderr), XtRImmediate, False },
    { "verbose", "Maxoptions", XtRBoolean, sizeof(Boolean),
     Offset(verbose), XtRImmediate, False },
};

/*
 * Command line options
 *
 * I have tried to make these the same as on NeXT
 * but allow longer names too.  Ambiguity is unavoidable
 * between say, `help' and `helpdir' so both have long names.
 */

static XrmOptionDescRec options[] = {
    { "-help", "help", XrmoptionNoArg, "True" },
    { "-version", "version", XrmoptionNoArg, "True" },
    { "-configuration", "configuration", XrmoptionSepArg, NULL },
    { "-rate", "sampleRate", XrmoptionSepArg, NULL },
    { "-patchpath", "patchPath", XrmoptionSepArg, NULL },
    { "-helpdir", "helpDir", XrmoptionSepArg, NULL },
    { "-fontname", "fontName", XrmoptionSepArg, NULL }, /* maybe should replace with standard X */
    { "-size", "fontSize", XrmoptionSepArg, NULL },     /* ditto */
    { "-advances", "advances", XrmoptionSepArg, NULL },
    { "-bootdir", "bootDir", XrmoptionSepArg, NULL },
    { "-logfile", "logfile", XrmoptionNoArg, "True" },
    { "-d", "ftsDir", XrmoptionSepArg, NULL },
    { "-killer", "killer", XrmoptionNoArg, "True" },
    { "-acs", "acs", XrmoptionNoArg, "True" },
    { "-nort", "nort", XrmoptionNoArg, "True" },
    { "-open", "open", XrmoptionSepArg, NULL },
    { "-verbose", "verbose", XrmoptionNoArg, "True" },
    { "-stderr", "stderr", XrmoptionNoArg, "True" },
};

static void Usage()
{
    fprintf(stderr, "Usage: %s [<option> ...]\n\n", progname);
    fprintf(stderr, "Where the options are:\n\n");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-help", "Show this help");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-version", "Show version and exit");
    fprintf(stderr, "\t%s\t\t%s\n", "-configuration <config>",
            "Set FTS configuration");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-rate <number>",
            "Set DAC sample rate");
    fprintf(stderr, "\t%s\t\t%s\n", "-patchpath <path>",
            "Search path for patches and externs");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-helpdir <path>",
            "Help file directory");
    fprintf(stderr, "\t%s\t\t%s\n", "-fontname <name>", "Text font name");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-size <number>",
            "Text font size (in points??)");
    fprintf(stderr, "\t%s\t\t%s\n", "-advances <values,...>",
            "Comma-separated list of scheduler advances");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-bootdir <path>",
            "Boot command directory (for standalone only)");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-logfile", "Make a logfile");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-d <path>",
            "Directory to find FTS and whatnot");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-killer", "Debugging");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-nort", "non-real-time");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-acs",
            "ACS-style real-time operation");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-open <path>",
            "open a patch on startup");
    fprintf(stderr, "\t%s\t\t\t%s\n", "-verbose", "Verbose startup");
    fprintf(stderr, "\t%s\t\t\t\t%s\n", "-stderr",
            "Copy Max window to stderr");
    fprintf(stderr, "\n");
    exit(1);
}


/*
 * Fallback Resources
 *
 * In the absence of an application defaults file the program
 * falls back on settings found in the following list.
 */

static String fallbacks[] = {

/* Fontlist */
    "*fontList:-*-helvetica-*-r-*--14-*-*-*-*-*-*-*=gui,\\\n\
-*-helvetica-*-r-*--8-*-*-*-*-*-*-*=maxfont-0,\\\n\
-*-helvetica-*-r-*--10-*-*-*-*-*-*-*=maxfont-1,\\\n\
-*-helvetica-*-r-*--12-*-*-*-*-*-*-*=maxfont-2,\\\n\
-*-helvetica-*-r-*--14-*-*-*-*-*-*-*=maxfont-3,\\\n\
-*-helvetica-*-r-*--14-*-*-*-*-*-*-*=maxfont-4,\\\n\
-*-helvetica-*-r-*--18-*-*-*-*-*-*-*=maxfont-5,\\\n\
-*-helvetica-*-r-*--18-*-*-*-*-*-*-*=maxfont-6,\\\n\
-*-helvetica-*-r-*--24-*-*-*-*-*-*-*=maxfont-7",

/* Menus - in English */
    "*fileMenu.labelString: File",
    "*editMenu.labelString: Edit",
    "*fontMenu.labelString: Font",
    "*miscMenu.labelString: Misc",
    "*helpMenu.labelString: Help",

    "*newOption.labelString: New Patcher",
    "*openOption.labelString: Open...",
    "*saveOption.labelString: Save",
    "*saveAsOption.labelString: Save As...",
    "*printOption.labelString: Print",
    "*closeOption.labelString: Close",
    "*exitOption.labelString: Exit",

    "*undoOption.labelString: Undo",
    "*cutOption.labelString: Cut",
    "*copyOption.labelString: Copy",
    "*clearOption.labelString: Clear",
    "*pasteOption.labelString: Paste",
    "*duplicateOption.labelString: Duplicate",
    "*selectAllOption.labelString: Select All",

    "*bombOption.labelString: FONT BOMB",

    "*editOption.labelString: Edit",
    "*stretchOption.labelString: Stretch",
    "*contractOption.labelString: Contract",
    "*settingsOption.labelString: Settings...",
    "*fixwidthOption.labelString: Fix width",
    "*lineupOption.labelString: Line up",
    "*restartOption.labelString: Restart...",
    "*findOption.labelString: Find...",
    "*findAgainOption.labelString: Find Again",

    "*infoOption.labelString: About Max...",
    "*introOption.labelString: Overview",
    "*contextOption.labelString: Help...",

/* accelerators */
    "*newOption.accelerator: Alt<Key>N",
    "*newOption.acceleratorText: Alt-N",
    "*openOption.accelerator: Alt<Key>O",
    "*openOption.acceleratorText: Alt-O",
    "*saveOption.accelerator: Alt<Key>S",
    "*saveOption.acceleratorText: Alt-S",
    "*closeOption.accelerator: Alt<Key>W",
    "*closeOption.acceleratorText: Alt-W",
    "*exitOption.accelerator: Alt<Key>Q",
    "*exitOption.acceleratorText: Alt-Q",
    "*undoOption.accelerator: Alt<Key>Z",
    "*undoOption.acceleratorText: Alt-Z",
    "*cutOption.accelerator: Alt<Key>X",
    "*cutOption.acceleratorText: Alt-X",
    "*copyOption.accelerator: Alt<Key>C",
    "*copyOption.acceleratorText: Alt-C",
    "*pasteOption.accelerator: Alt<Key>V",
    "*pasteOption.acceleratorText: Alt-V",
    "*duplicateOption.accelerator: Alt<Key>D",
    "*duplicateOption.acceleratorText: Alt-D",
    "*selectAllOption.accelerator: Alt<Key>A",
    "*selectAllOption.acceleratorText: Alt-A",
    "*infoOption.accelerator: Alt<Key>I",
    "*infoOption.acceleratorText: Alt-I",
    "*introOption.accelerator: Alt<Key>H",
    "*introOption.acceleratorText: Alt-H",
    "*editOption.accelerator: Alt<Key>E",
    "*editOption.acceleratorText: Alt-E",

/* Dialogs */
    "*infoDialog.dialogTitle: About Max",
    "*infoDialog.okLabelString: Dismiss",
    "*errorDialog.dialogTitle: Max Error",
    "*errorDialog.okLabelString: OK",
    "*closeDialog.dialogTitle: Close Patcher",
    "*closeDialog.messageString: Patcher has unsaved changes --- Close anyway?",
    "*closeDialog.okLabelString: OK",
    "*closeDialog.cancelLabelString: Cancel",
    "*closeDialog.helpLabelString: Save",
    "*exitDialog.dialogTitle: Exit Max",
    "*exitDialog.messageString: Exit Max?",
    "*exitDialog.okLabelString: OK",
    "*exitDialog.cancelLabelString: Cancel",
    "*loadDialog.dialogTitle: Load Patcher",
    "*loadDialog.messageString: Patcher is loading",
    "*loadDialog.cancelLabelString: Cancel",

/* Console */
    "*consoleTitle.labelString: Console",
    "*console.rows: 30",
    "*console.columns: 60",
    "*console.scrollHorizontal: False",
    "*console.scrollTopSide: False",
    "*console.scrollLeftSide: False",
    "*consoleTitle.fontList: -*-helvetica-*-r-*--14-*-*-*-*-*-*-*",
    "*console.fontList: -*-helvetica-*-r-*--14-*-*-*-*-*-*-*",

#if PALETTE_WINDOW
/* Palette */
    "*palette.foreground: black",
    "*palette.background: white",
#endif

/* Canvas */
/*
"*canvas.foreground: black",
"*canvas.background: white",
*/

    "*canvas.translations: #override\\n\\\n\
 Alt<Key>n: file-new()\\n\\\n\
 Alt<Key>o: file-open()\\n\\\n\
 Alt<Key>s: file-save()\\n\\\n\
 Alt<Key>w: file-close()\\n\\\n\
 Alt<Key>q: file-exit()\\n\\\n\
 Alt<Key>x: edit-cut()\\n\\\n\
 Alt<Key>c: edit-copy()\\n\\\n\
 Alt<Key>v: edit-paste()\\n\\\n\
 Alt<Key>d: edit-duplicate()\\n\\\n\
 Alt<Key>e: misc-edit()\\n\\\n\
 Alt<Key>a: edit-select-all()\\n\\\n\
 Alt<Key>i: help-info()\\n\\\n\
 Alt<Key>h: help-intro()\\n\\\n\
 <Btn1Down>(2): double-click()\n\
 Alt<BtnDown>: alt-click()\\n\\",

    "*foreground: black",
    "*background: white",
#if 0
    "*textBox.foreground: black",
    "*textBox.background: blue",
    "*textSelectedForeground: red",
    "*textSelectedBackgrouns: blue",
    "*TextSelectedForeground: red",
    "*TextSelectedBackgrouns: blue",
#endif

    "*textBox.shadowThickness: 0",
    "*textBox.highlightThickness: 0",

/*
"*SgNuseEnhancedFSB: TRUE", 
"*useSchemes: all",
"*scheme: Base",
"*sgiMode: TRUE",
 -- this make selected text white instead of black;
        otherwise I'd use it. 
*/

    NULL
};



/*
 * main()
 *
 * Based as closely as possible on NeXT main()
 */

int main(int argc, char **argv)
{
    /* Command line options set these ... */
    char *config;
    char *ftsdir;
    char *advances;
    char *font;
    char *fontsize;
    char *helpdir;
    char *bootdir;
    char *searchpath;
    int rate;
    int deffontsize;

    /* Toplevel shell */
    SetArg(allowShellResize, True);
    toplevel = XtAppInitialize(&app_context, getappname(), options,
                               XtNumber(options),
                               &argc, argv, fallbacks, ARGS), narg = 0;

    /* Read resources database */
    XtGetApplicationResources(toplevel, &app_resources, resources,
                              XtNumber(resources), NULL, 0);

/* #define DEBUG_CMDLINE        /**/
#ifdef DEBUG_CMDLINE
    printf("copy_to_stderr is %s\n",
           app_resources.copy_to_stderr ? "on" : "off");
    printf("verbose is %s\n", app_resources.verbose ? "on" : "off");
    printf("help is %s\n", app_resources.help ? "on" : "off");
    printf("killer is %s\n", app_resources.killer ? "on" : "off");
    printf("nort is %s\n", app_resources.nort ? "on" : "off");
    printf("acs is %s\n", app_resources.acs ? "on" : "off");
    printf("openfile is %s\n", app_resources.openfile);
    printf("logfile is %s\n", app_resources.logfile ? "on" : "off");
    printf("configuration is %s\n", app_resources.configuration);
    printf("fts_dir is %s\n", app_resources.fts_dir);
    printf("advances is %s\n", app_resources.advances);
    printf("boot_dir is %s\n", app_resources.boot_dir);
    printf("help_dir is %s\n", app_resources.help_dir);
    printf("search_path is %s\n", app_resources.search_path);
    printf("font_name is %s\n", app_resources.font_name);
    printf("font_size is %d\n", app_resources.font_size);
    printf("sample_rate is %d\n", app_resources.sample_rate);
#endif

    if (argc > 1 || app_resources.help)
        Usage();

#ifndef DEBUG
    {
        struct rlimit rlim;
        if (getrlimit(RLIMIT_CORE, &rlim))
            perror("getrlimit");
        else
        {
            rlim.rlim_cur = 0;
            if (setrlimit(RLIMIT_CORE, &rlim))
                perror("setrlimit");
        }
    }
#endif
    if (app_resources.version)
    {
        printf("Max version %s\n(c) IRCAM 1990, 1993\n", max_version);
        exit(0);
    }

    /* init GUI */
    setupmainwindow();

    if (app_resources.copy_to_stderr)
        post_copytostd();
    if (app_resources.verbose)
        h_setverbose(1);

    config = app_resources.configuration;
    /* if (!config) config = t_config(); */
    if (!config)
        config = "3/5,2/3";

    h_setmax();

    ftsdir = app_resources.fts_dir;
    if (!ftsdir)
    {
        char *appname = *argv;
        if (appname)
        {
            char *slashp = 0, *sp, *ftsspace;
            if (!(ftsspace = malloc(strlen(appname) + 1)))
                hate("out of space");
            strcpy(ftsspace, appname);
            for (sp = ftsspace; *sp; sp++)
                if (*sp == '/')
                    slashp = sp;
            if (slashp)
            {
                *slashp = 0;
                ftsdir = ftsspace;
            }
            else
                ftsdir = ".";
        }
        else
            ftsdir = ".";
    }

    h_setconfig(config);

    if (advances = app_resources.advances)
        h_setadvancesstring(advances);
    if (rate = app_resources.sample_rate)
        h_setrate((float) rate);
    if (font = app_resources.font_name)
        t_setfont(font);
    if (app_resources.killer)
        h_setkiller(1);
    if (app_resources.logfile)
        h_setlogfile(1);
    if (deffontsize = app_resources.font_size)
        t_setfontsize(deffontsize);

    h_setfts(XFts);
    if (ftsdir)
        h_setftsdir(ftsdir);
    if (bootdir = app_resources.boot_dir)
        h_setbootdir(bootdir);

    h_init();

    /* start FTS */
    if (!app_resources.killer)
        max_forkfts();
    else
        infile = outfile = -1;

    mess_init();
    {
        char dir[MAXPATHLEN + 1];
        if (getcwd(dir, MAXPATHLEN))
            defvolume = gensym(dir);
    }
    h_start();
    max_startpolling();

    searchpath = app_resources.search_path;
    if (!searchpath)
        searchpath = "";
    t_setdefpathandflag(searchpath, 1);
    t_setpath(t_getdefpath());

    helpdir = app_resources.help_dir;
    if (helpdir)
        t_sethelppath(gensym(helpdir));
    else
    {
        char helphelp[512];
        sprintf(helphelp, "%s/help", ftsdir);
        t_sethelppath(gensym(helphelp));
    }

#if OLD
    max_startfts();
#else                           /* already did h_start */
    h_vmess(NIL, gensym("setfts"), "s", gensym(h_getftsdir()));
#endif

    /* Bring GUI up */
    XtRealizeWidget(toplevel);

    if (app_resources.openfile)
        max_openfile(app_resources.openfile);

    XtAppMainLoop(app_context);

    /* never gets here */
}
