#ifndef lint
static char rcsid[] =
    "$Revision: 1.21 $ IRCAM $Date: 1994/02/07 11:00:42 $";
#endif

/*
 *      Copyright (c) 1987-1994 IRCAM
 *                      Copyright (c) 1995-1996 UC REGENTS
 *                          All rights reserved.
 */

/*
 * X/Motif implementation of the "wind" object for Max
 * By Paul Foley and Miller Puckette
 *
 */

typedef struct windpub *GrafPort;
#define DEFINEDGRAFPORT

/*
 * Includes
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/resource.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Protocols.h>
#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/Text.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>
#ifdef __sgi
#include <Xm/FileSB.h>
#endif

#include "m_mess.h"
#include "cp_mess.h"
#include "h_cp.h"
#include "t_mac.h"
#include "t_wind.h"

#define MAX_ARGS 20
#define ARGS  args,narg

#ifdef __sgi
#define __STRICT_ANSI__
#endif

#if 1
#define SetArg(n,v) XtSetArg(args[narg],XmN##n,v),narg++
#else
#define SetArg(n,v) XtSetArg(args[narg],XmN/**/n,v),narg++
#endif
#define SetValues(w) XtSetValues(w,ARGS), narg=0
#define GetValues(w) XtGetValues(w,ARGS), narg=0

typedef off_t caddr_t;

static int narg = 0;
static Arg args[MAX_ARGS];

typedef void (*t_callbackfunction)(Widget w, void *x1, void *x2);
typedef void (*t_timercallbackfunction)(void *x1, XtIntervalId * x2);
typedef Boolean(*t_workcallbackfunction)
    ( /* XtAppContext x1, XtWorkProc x2, */ XtPointer x3);

/* 
 * Widget Creation
 */

#define CreateCascadeButton(p,n) XmCreateCascadeButton(p,n,ARGS), narg=0
#define CreateDrawingArea(p,n) XmCreateDrawingArea(p,n,ARGS), narg=0
#define CreateFileSelectionDialog(p,n) \
    XmCreateFileSelectionDialog(p,n,ARGS), narg=0
#define CreateForm(p,n) XmCreateForm(p,n,ARGS), narg=0
#define CreateFrame(p,n) XmCreateFrame(p,n,ARGS), narg=0
#define CreateLabel(p,n) XmCreateLabel(p,n,ARGS), narg=0
#define CreateMenuBar(p,n) XmCreateMenuBar(p,n,ARGS), narg=0
#define CreateOptionMenu(p,n) XmCreateOptionMenu(p,n,ARGS), narg=0
#define CreatePulldownMenu(p,n) XmCreatePulldownMenu(p,n,ARGS), narg=0
#define CreateScrolledText(p,n) XmCreateScrolledText(p,n,ARGS), narg=0
#define CreateSeparator(p,n) Manage(XmCreateSeparator(p,n,ARGS)), narg=0
#define CreateText(p,n) XmCreateText(p,n,ARGS), narg=0

#define CreateMessageDialog(p,n) XmCreateMessageDialog(p,n,ARGS), narg=0
#define CreateInformationDialog(p,n) XmCreateInformationDialog(p,n,ARGS), narg=0
#define CreateErrorDialog(p,n) XmCreateErrorDialog(p,n,ARGS), narg=0
#define CreateQuestionDialog(p,n) XmCreateQuestionDialog(p,n,ARGS), narg=0
#define CreateWarningDialog(p,n) XmCreateWarningDialog(p,n,ARGS), narg=0
#define CreateWorkingDialog(p,n) XmCreateWorkingDialog(p,n,ARGS), narg=0

#define CreateMenuButton(button, parent, label, cb, arg) \
button = XmCreatePushButtonGadget(parent, label, ARGS), narg=0; \
Manage(button); \
XtAddCallback(button, XmNactivateCallback, cb, arg)

#define CreateRadioButton(button, parent, label, cb, arg) \
button = XmCreateToggleButtonGadget(parent, label, ARGS), narg=0; \
Manage(button); \
XtAddCallback(button, XmNvalueChangedCallback, cb, arg)

#define CreateOptionButton(parent, label) \
XmCreatePushButtonGadget(parent, label, ARGS), narg=0

#define Manage(w) XtManageChild(w)
#define Unmanage(w) XtUnmanageChild(w)

/* 
 * Motif Strings
 */

#define STRset(a,b) a = XmStringCreateLtoR(b,XmSTRING_DEFAULT_CHARSET)
#define STRfree(a) if (a) XmStringFree(a)

/* Widget sensitivity */
#define Dim(w) XtSetSensitive(w, False)
#define Undim(w) XtSetSensitive(w, True)


/* 
 * Menu options
 */

#define FILE_NEW  0
#define FILE_OPEN 1
#define FILE_SAVE 2
#define FILE_SAVE_AS  3
#define FILE_PRINT  4
#define FILE_CLOSE  5
#define EXIT_MAX  6

#define EDIT_CUT  10
#define EDIT_COPY 11
#define EDIT_PASTE  12
#define EDIT_CLEAR  13
#define EDIT_DUPLICATE  14
#define EDIT_SELECT_ALL 15
#define EDIT_UNDO 16

#define MISC_EDIT   20
#define MISC_STRETCH  21
#define MISC_CONTRACT   22
#define MISC_FIXWIDTH   23
#define MISC_LINEUP   24
#define MISC_SETTINGS   25
#define MISC_RESTART  26
#define MISC_FIND   27
#define MISC_FIND_AGAIN 28

#define HELP_INFO 30
#define HELP_INTRO  31
#define HELP_CONTEXT  32

#define CANCEL_SELECT 100

/*
 * other defines 
 */

/*#define USE_TIMEFN        /**/

char *getcreateerror();

/* panel_ouch() messages */
#define OUCH_ACCESS "Error accessing file: %s"
#define OUCH_ORDINARY "Not an ordinary file"
#define OUCH_MAX "Not a Max file"

/*
 * Function declarations 
 */


Widget wind_menubar(Widget form, int really);
void setopenpaneldir(char *s);
void *openfrompath(char *filename);
void panel_ouch(char *fmt, ...);
void post_copytostd();

void mouseDown(t_wind * wind, int x, int y, int flags, int doubleclick);
void mouseUp(t_wind * wind, int x, int y);
void keyDown(t_wind * wind, char charCode, int flags);
void keyUp(t_wind * wind, char charCode, int flags);
void chknoport();
void wind_clear(t_wind * wind);
void undo(t_wind * wind);
void cut(t_wind * wind);
void copy(t_wind * wind);
void paste(t_wind * wind);
void duplicate(t_wind * wind);
void selectall(t_wind * wind);
void edit(t_wind * wind);
void wind_generic(t_wind * x, char *s, int n);
void font(t_wind * wind, int fontsize);
void stretch(t_wind * wind);
void contract(t_wind * wind);
void fixwidth(t_wind * wind);
void lineup(t_wind * wind);

    /*************** variables ***************/
int wind_xoffset;
int wind_yoffset;

Display *display;               /* connection to X server */
static XtIntervalId timer;      /* Xt timeout handle */
Atom wm_delete_window;          /* WM_DELETE_WINDOW protocol */

/* Graphic contexts for drawing onto canvas */
static int gcs_created = 0;
XGCValues gc_values;
GC draw_gc;
GC undraw_gc;
GC invert_gc;

extern t_symbol s_;
char *getappname();
extern char *progname;
extern char max_version[];      /* version info */
extern XtAppContext app_context;        /* application context */
extern Widget toplevel;         /* Top level shell */
Widget main_window;             /* Main Window */
Widget menu_bar;                /* Menu Bar */
Widget console_title;           /* Console Title */
Widget frame;
Widget console;                 /* Console for error messages etc. */


/* File Selector and dialog boxes */
Widget file_selector = 0;
Widget error_dialog = 0;
Widget info_dialog = 0;
Widget exit_dialog = 0;
Widget load_dialog = 0;

extern XmFontList fontlist;     /* filled in here */
extern int pointsizes[];

/* To make menu labels based on the pointsizes ... */
#define FONT_LABEL_TEMPLATE "%dpt"      /* e.g., "10pt" */
char font_option_label[5];      /* enough for FONT_LABEL_TEMPLATE + 1 */

/* current window */
t_wind *current_window = 0;


/*
 * Polling functions
 */

#define POLLINTERVAL 0.1
static float pollinterval = POLLINTERVAL;
void wind_poll();
static int wind_pollon;

void h_domaxpoll()
{
    h_dopoll();
    wind_poll();
    /* cast to long is mandatory here */
    timer = XtAppAddTimeOut(app_context, (long) (pollinterval * 1000),
                            (t_timercallbackfunction) h_domaxpoll, NULL);
}

#ifdef USE_TIMEFN
void max_timefn(void *x, char *buf, int size)
{
    long foo = *(long *) buf;
    restart_settime(foo);
}
#endif

void max_startpolling()
{
    h_domaxpoll();
}

void max_stoppolling()
{
    if (timer)
        XtRemoveTimeOut(timer);
    h_setpollfunction(0);
}

h_setpollinterval(float f)
{
    pollinterval = .05;
}

void console_msg(char *fmt, ...)
{
    va_list arglist;
    char buf[BUFSIZ];
    static XmTextPosition position;

    va_start(arglist, fmt);
    vsprintf(buf, fmt, arglist);
    va_end(arglist);

    XmTextInsert(console, position, buf);       /* append text */
    position += strlen(buf);
    SetArg(cursorPosition, position);   /* advance EOT marker */
    SetValues(console);
    XmTextShowPosition(console, position);      /* ensure text is visible */
}

/*
 * X/Xt error handlers
 *
 * Log errors on console
 */

static int X_error_handler(Display *display, XErrorEvent *err_event)
{
    char buf[BUFSIZ];

    XGetErrorText(display, err_event->error_code, buf, sizeof buf);
    console_msg("X Error: %s\n", buf);
    return (0);
}

static void Xt_error_handler(char *message)
{
    console_msg("Xt Error: %s\n", message);
}

/*
 * cleanexit()
 *
 * Release all resources and exit Max.  
 * Called on final exit from the program.
 */

void cleanexit()
{
    max_stoppolling();
    h_eof();

    /* close all windows */
    while (current_window)
    {
        current_window->w_dirty = 0;    /* force close: FIXME */
        close_patcher_window(current_window);
        /* focus event will update `current_window' */
    }

    if (gcs_created)
    {
        XFreeGC(display, draw_gc);
        XFreeGC(display, undraw_gc);
        XFreeGC(display, invert_gc);
    }

    t_freebits();

    if (file_selector)
        XtDestroyWidget(file_selector);
    if (error_dialog)
        XtDestroyWidget(error_dialog);
    if (info_dialog)
        XtDestroyWidget(info_dialog);
    if (exit_dialog)
        XtDestroyWidget(exit_dialog);
    if (load_dialog)
        XtDestroyWidget(load_dialog);

    XtDestroyWidget(toplevel);
    XCloseDisplay(display);
    exit(0);
}

/*
 * app_killed(),  Called on f.kill from WM menu.
 * query and exit.
 */

static void app_killed()
{
    do_exit_dialog();
}

/*
 * app_closed(),  Called on f.kill on a t_wind window.
 * Act as if "close" menu item were called.
 */

static void app_closed(Widget wd, t_wind *w)
{
    close_patcher_window(w);
}

#ifdef DEBUG
/* 
 * Debugging
 *
 * Print all descendants of the given widget in outline form on stdout.
 * The level is used to indicated indenting level. You should
 * always specify level 0 when you call this function.
 */

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>

static void widget_tree(Widget w, int level)
{
    char name_column[128] = { '\0' };
    int ndots;
    int nlines;

    /* Print continuation lines for previous levels. */
    for (nlines = level - 1; nlines > 0; nlines--)
        sprintf(name_column + strlen(name_column), "|  ");

    /* Print start line for this widget. */
    if (level > 0)
        sprintf(name_column + strlen(name_column), "|- ");

    /* Print widget name. */
    sprintf(name_column + strlen(name_column), "%s", XtName(w));
    printf("%s", name_column);

    /* Print leader to class name. */
    for (ndots = strlen(name_column) < 50 ? 50 - strlen(name_column) : 2;
         ndots > 0; ndots--)
        printf(".");
    printf("%s\n", XtClass(w)->core_class.class_name);

    if (XtIsComposite(w))
    {
        Cardinal nc;
        WidgetList c;
        SetArg(numChildren, &nc);
        SetArg(children, &c);
        GetValues(w);
        while (nc-- > 0)
            widget_tree(*(c++), level + 1);
    }
}
#endif                          /* DEBUG */

/*
 * Dialogs
 *
 * All must get focus immediately.
 */

static void popup_dialog(Widget w)
{
    Manage(w);
    /* ensure comes to top */
    XtPopup(XtParent(w), XtGrabNone);
    XMapRaised(XtDisplay(w), XtWindow(XtParent(w)));
}

/*
 * Working Dialogs
 *
 * Working dialogs are handled as follows:
 * 1) start_load() is called to set the `work_done' flag to False.
 *    It also attachs a work procedure to the working dialog and pops
 *    the dialog up.  Special processing is needed to force the dialog
 *    up immediately.
 * 2) The work procedure workingCallback() checks the `work_done' flag.
 *    If it is set it removes the work procedure and pops down the dialog.
 *    Otherwise it returns 0 to indicate the work is not finished.
 * 3) end_load() sets `work_done' to True so the next time workingCallback()
 *    gets called the dialog will disappear.
 */

do_exit_dialog()
{
    if (!exit_dialog)
    {
        SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE);
        exit_dialog = CreateQuestionDialog(toplevel, "exitDialog");
        Unmanage(XmMessageBoxGetChild(exit_dialog, XmDIALOG_HELP_BUTTON));
        XtAddCallback(exit_dialog, XmNokCallback,
                      (t_callbackfunction) cleanexit, NULL);
    }
    /* If user confirms control is passed to cleanexit() */
    popup_dialog(exit_dialog);
}

XtWorkProcId work_id;
static int work_done = 0;       /* True => remove dialog */

static int workingCallback(Widget w)
{
    if (work_done)
    {
        SetArg(userData, &work_id);
        GetValues(load_dialog);
        XtRemoveWorkProc(work_id);      /* remove work procedure */
        XtPopdown(XtParent(load_dialog));       /* remove dialog */
        return 1;               /* Done */
    }
    else
    {
        return 0;               /* not Done */
    }
}

static void start_load()
{
    Widget diashell;
    Window diawindow, topwindow;
    XWindowAttributes xwa;
    XEvent event;

    /*  Load dialog
     */
    if (!load_dialog)
    {
        SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE);
        SetArg(dialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
        load_dialog = CreateWorkingDialog(toplevel, "loadDialog");
        Unmanage(XmMessageBoxGetChild(load_dialog, XmDIALOG_OK_BUTTON));
        Unmanage(XmMessageBoxGetChild(load_dialog, XmDIALOG_HELP_BUTTON));
        /* Cancelling the load is not supported */
        Dim(XmMessageBoxGetChild(load_dialog, XmDIALOG_CANCEL_BUTTON));
    }

    work_done = 0;
    work_id =
        XtAppAddWorkProc(app_context,
                         (t_workcallbackfunction) workingCallback,
                         load_dialog);
    SetArg(userData, work_id);
    SetValues(load_dialog);
    popup_dialog(load_dialog);

    /*
     * This procedure will ensure that, if a dialog window is being mapped,
     * its contents become visible before returning.  It is intended to be
     * used just before a bout of computing that doesn't service the display.
     * You should still call XmUpdateDisplay() at intervals during this
     * computing if possible.
     *
     * The monitoring of window states is necessary because attempts to map
     * the dialog are redirected to the window manager (if there is one) and
     * this introduces a significant delay before the window is actually mapped
     * and exposed.  This code works under mwm, twm, uwm, and no-wm.  It
     * doesn't work (but doesn't hang) with olwm if the mainwindow is iconified.
     *
     * The argument to ForceDialog is any widget in the dialog (often it
     * will be the BulletinBoard child of a DialogShell).
     */

    diashell = XtParent(load_dialog);
    if (XtIsRealized(diashell) && XtIsRealized(toplevel))
    {
        diawindow = XtWindow(diashell);
        topwindow = XtWindow(toplevel);

        /* Wait for the dialog to be mapped.  It's guaranteed to become so unless... */
        while (XGetWindowAttributes(display, diawindow, &xwa),
               xwa.map_state != IsViewable)
        {

            /* ...if the primary is (or becomes) unviewable or unmapped, it's
               probably iconified, and nothing will happen. */
            if (XGetWindowAttributes(display, topwindow, &xwa),
                xwa.map_state != IsViewable)
                break;

            /* At this stage, we are guaranteed there will be an event of some
               kind. Beware; we are in a callback, so this can recurse. */
            XtAppNextEvent(app_context, &event);
            XtDispatchEvent(&event);
        }
    }
    /* The next XSync() will get an expose event if the dialog was unmapped. */
    XmUpdateDisplay(toplevel);
}

static void end_load()
{
    work_done = 1;
}

/*
 * Mouse coords
 *
 * The mouse coordinates are in the globals `mouse_x' and `mouse_y'
 * which are kept up to date by handling a ButtonMotion event.
 */

static int mouse_x, mouse_y;

/*
 * getMouseCoords()
 *
 * Called by mouseUp() mouseDown() and view_poll().
 * Just copy last recorded coords into the Point.
 * No transformation is required.
 */

void getMouseCoords(Point *point)
{
    point->h = mouse_x;
    point->v = mouse_y;
}

/*
 * Patcher events
 *
 * We respond to Focus and MotionNotify events as follows:
 * FocusIn;  We change the current patcher.
 * FocusOut;  Ignore.
 * MotionNotify;  Store mouse coordinates in `mouse_x' and `mouse_y'.
 *
 * The Focus events are attached to the Form.
 * The MotionNotify is a ButtonMotion event (attached to the canvas) 
 * since DrawingAreas are too dumb to recognise `dragging'.
 */

int trace_events = 0;

static void patcher_event(Widget w, XtPointer ptr, XEvent *event,
                          Boolean *x4)
{
    t_wind *wind = (t_wind *) ptr;
    switch (event->xany.type)
    {

    case FocusIn:
        if (trace_events)
            fprintf(stderr, "FocusIn\n");
        if (current_window != wind)
        {                       /* avoid unnecessary drawing */
            undim_menus(wind->w_titled);
            current_window = wind;
            /* wind_update(wind); */
        }
        break;

    case FocusOut:
        if (trace_events)
            fprintf(stderr, "FocusOut\n");
        /* ignored */ ;
        break;

    case EnterNotify:
        if (trace_events)
            fprintf(stderr, "EnterNotify\n");
        wind_pollon = 1;
        break;

    case LeaveNotify:
        if (trace_events)
            fprintf(stderr, "LeaveNotify\n");
        wind_pollon = 0;
        break;

    case MotionNotify:
        mouse_x = event->xmotion.x;
        mouse_y = event->xmotion.y;
        break;
    }
}

#define KBUFSIZE 10

static int eventtoascii(XKeyEvent *e)
{
    KeySym keysym;
    char keybuf[KBUFSIZE];
    int mofo = XLookupString(e, keybuf, KBUFSIZE, &keysym, 0);
    if (!mofo)
        return (0);
    else
        return (keybuf[0] & 0xff);
}

/*
 * Canvas events
 *
 * We are interested in all (DrawingArea) callbacks: Expose, Resize and Input.
 */

extern void *key_keyout, *keyup_keyout; /* Keychord info ... */
static struct timeval mouseup_time;
static int mouseup_x, mouseup_y;
#define DBLCLICKPIX 2

static void canvasCallback(Widget w, void *voidwind, void *voidcb)
{
    int asciikeycode;
    int doubleclick, shift;
    t_wind *wind = voidwind;
    XmDrawingAreaCallbackStruct *cb = voidcb;
    Position x, y, width, height;

    /* no matter what the event might be, update the window coords
     * so that "save" will remember them.
     */
    SetArg(x, &x);
    SetArg(y, &y);
    GetValues(XtParent(w));
    SetArg(width, &width);
    SetArg(height, &height);
    GetValues(w);
    wind->w_x1 = x;
    wind->w_y1 = y;
    /* wind->w_x2 = x + width;
       wind->w_y2 = y + height; */
    wind->w_x2 = x + width - wind->w_leftmargin;
    wind->w_y2 = y + height - wind->w_topmargin;

    switch (cb->reason)
    {

    case XmCR_EXPOSE:
    case XmCR_RESIZE:
        if (XtIsRealized(XtParent(wind->w_maincanvas)))
            wind_update(wind);
        break;

    case XmCR_INPUT:
        switch (cb->event->xany.type)
        {

        case ButtonPress:
            if (trace_events)
                fprintf(stderr, "ButtonPress\n");
            mouse_x = cb->event->xbutton.x;
            mouse_y = cb->event->xbutton.y;
            doubleclick = (msecsince(&mouseup_time) < 500 &&
                           mouse_x >= mouseup_x - DBLCLICKPIX &&
                           mouse_x <= mouseup_x + DBLCLICKPIX &&
                           mouse_y >= mouseup_y - DBLCLICKPIX &&
                           mouse_y <= mouseup_y + DBLCLICKPIX);
            shift = ((cb->event->xbutton.state & ShiftMask) ?
                     shiftKey : 0);
#if PALETTE_WINDOW
            mouseDown(wind, mouse_x, mouse_y, shift, doubleclick);
#else
            if (mouse_y <= wind->w_topmargin)
            {
                mouseDown(wind, mouse_x, mouse_y,
                          inMargin | shift, doubleclick);
            }
            else
            {
                mouseDown(wind, mouse_x - wind->w_leftmargin,
                          mouse_y - wind->w_topmargin, shift, doubleclick);
            }
#endif
            break;

        case ButtonRelease:
            if (trace_events)
                fprintf(stderr, "ButtonRelease\n");
            mouse_x = mouseup_x = cb->event->xbutton.x;
            mouse_y = mouseup_x = cb->event->xbutton.y;
            getrealtime(&mouseup_time);
            mouseUp(wind, mouse_x - wind_xoffset, mouse_y - wind_yoffset);
            break;

        case KeyPress:
            asciikeycode = eventtoascii(&cb->event->xkey);
            if (asciikeycode == 8)
                wind_clear(wind);
            else if (asciikeycode == 27)
                p_mute(), post("stopping DSP");
            if (trace_events)
                fprintf(stderr, "KeyPress\n");
            keyDown(wind, eventtoascii(&cb->event->xkey), 0);
            if (key_keyout)
                outlet_int(key_keyout, eventtoascii(&cb->event->xkey));

            break;

        case KeyRelease:
            if (trace_events)
                fprintf(stderr, "KeyRelease\n");
            keyUp(wind, eventtoascii(&cb->event->xkey), 0);
            if (keyup_keyout)
                outlet_int(keyup_keyout, eventtoascii(&cb->event->xkey));
            break;
        }
    }
}

#if PALETTE_WINDOW_NOT_YET
/*
 * Palette events
 */

static void paletteCallback(Widget w, t_wind *wind,
                            XmDrawingAreaCallbackStruct *cb)
{
    switch (cb->reason)
    {

    case XmCR_INPUT:
        switch (cb->event->xany.type)
        {

        case ButtonPress:
            bug("pallettecallback");
            break;

        case ButtonRelease:
            mouse_x = cb->event->xbutton.x;
            mouse_y = cb->event->xbutton.y;
            mouseUp(wind, mouse_x - wind_xoffset, mouse_y - wind_yoffset);
            break;

            /* currently ignoring keypresses in palette */
        }
    }
}
#endif

/*
 * Window Lists
 *
 * The windows created by Max are kept in a circular doubly-linked list.
 * The list starts as nil, of course, and there is a special case of one member.
 * Having a list allows us to have handles on the windows which cannot be got from
 * the window system in any reliable way that I can find.
 *
 * The list is manipulated by the functions:
 * - add_to_window_list()
 * - remove_from_window_list()
 */

void add_to_window_list(t_wind *new_window)
{
    if (current_window)
    {
        new_window->w_next = current_window->w_next;
        current_window->w_next = new_window;
        new_window->w_previous = current_window;
        new_window->w_next->w_previous = new_window;
    }
    else
    {
        current_window = new_window;
        current_window->w_next = current_window;
        current_window->w_previous = current_window;
    }
}

void remove_from_window_list(t_wind *window)
{
    if (window->w_previous != window)
    {
        window->w_previous->w_next = window->w_next;
        window->w_next->w_previous = window->w_previous;
        if (window == current_window)
            current_window = window->w_next;
    }
    else
        current_window = 0;
}

t_wind *find_window_for_widget(Widget foo)
{
    t_wind *w = current_window;
    if (!w)
        bug("find_window_for_widget");
    do
    {
        if (w->w_palette == foo || w->w_maincanvas == foo ||
            w->w_margincanvas == foo || w->w_canvasshell == foo)
            return (w);
        w = w->w_next;
    }
    while (w != current_window);
    bug("find_window_for_widget 2");
    return (0);
}

/*
 * create_patcher()
 * 
 * TBW
 */

static void create_patcher(int mode, char *name)
{
    void *new_patcher;

    switch (mode)
    {

    case FILE_NEW:
        new_patcher = (void *) patcher();
        break;

    case FILE_OPEN:
        start_load();
        new_patcher = (void *) openfrompath(name);
        end_load();
        break;

    case HELP_INTRO:
        start_load();
        patcher__help();        /* This opens 3 windows, not one. */
        new_patcher = (void *) 1;       /* bad bad bad */
        end_load();
        break;

    default:
        new_patcher = 0;
        break;
    }

    if (!new_patcher)
    {
        console_msg("Unable to create new patcher \"%s\"\n", name);
        return;
    }
}

/*void*/ close_patcher_window(t_wind *w)
{

    /* The action of wind_willclose() in the case of a patcher abstraction leaves
     * the window in an intermediate and invalid state 
     * -- go directly to wind_close() in this case 
     */
    if (!wind_willclose(w))
        wind_close(w);
}

/* 
 * File Selector
 */

static void fileselectCallback(Widget w, int mode,
                               XmFileSelectionBoxCallbackStruct *cb)
{
    char *filepath;
    char *filename;
    char *filedir;
    struct stat statbuf;

    /* extract dir and file name */
    XmStringGetLtoR(cb->value, XmSTRING_DEFAULT_CHARSET, &filepath);
    if (!filepath)
        filepath = "";
    if (filename = (char *) (strrchr(filepath, '/')))
        filename++;
    else
        filename = "";
    filedir = strdup(filepath);
    filedir[strlen(filepath) - strlen(filename)] = 0;

    Unmanage(file_selector);
    XtRemoveAllCallbacks(file_selector, XmNokCallback);

    if (filename)
    {
        switch (mode)
        {

        case CANCEL_SELECT:
            break;              /* nop */

        case FILE_OPEN:
            if (stat(filepath, &statbuf) == -1)
            {
                panel_ouch(OUCH_ACCESS, strerror(errno));       /* cannot stat */
                /* check file is ordinary */
            }
            else if (S_ISREG(statbuf.st_mode))
            {
                create_patcher(mode, filepath);
            }
            else
            {
                panel_ouch(OUCH_ORDINARY);      /* not ordinary */
            }
            break;

        case FILE_SAVE_AS:
            if (current_window)
            {
                if (wind_saveas(current_window, filepath) == 1)
                    panel_ouch("%s", getcreateerror());
            }
            else
                console_msg("SAVE AS: no current patcher\n");
            break;
        }
    }
}

static void popup_file_selector(char *title, int mode)
{
    XmString title_string = NULL;

    /*  Create file selector 
     */
    if (!file_selector)
    {
        XmString pattern_string = NULL;
        XmString dir_string = NULL;
        STRset(pattern_string, "*.pat");
        STRset(dir_string, "");
        SetArg(dirMask, pattern_string);        /* Pattern */
        SetArg(directory, dir_string);  /* Directory */
        SetArg(mwmDecorations,
               MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_RESIZEH);
        file_selector =
            CreateFileSelectionDialog(toplevel, "fileSelector");
        STRfree(pattern_string);
        STRfree(dir_string);
        Unmanage(XmFileSelectionBoxGetChild
                 (file_selector, XmDIALOG_HELP_BUTTON));
        XtAddCallback(file_selector, XmNcancelCallback,
                      (t_callbackfunction) fileselectCallback,
                      (XtPointer) CANCEL_SELECT);
    }

    /* Change the title to give a visual cue as to what you are selecting */
    STRset(title_string, title);
    SetArg(dialogTitle, title_string);
    SetValues(file_selector);
    STRfree(title_string);

    /* 
     * The okCallback can have different effects, depending on whether we
     * are opening or saving a patch, noted in `mode'.  
     * For this reason we set the callback dynamically.
     */
    XtAddCallback(file_selector, XmNokCallback,
                  (t_callbackfunction) fileselectCallback,
                  (XtPointer) mode);

    /* force scan to get up-to-the-minute listing */
    XmFileSelectionDoSearch(file_selector, NULL);

    popup_dialog(file_selector);
}

/*
 * verifyCloseCallback()
 *
 * The Close dialog is popped up in wind_willclose() if the
 * patcher is dirty.  We can save the patcher before closing it,
 * or just close it.
 * The flow of control in this case is:
 * 1) wind_willclose() pops up dialog and returns with `not closed' code
 * 2) close dialog (optionally) saves patcher, marks not dirty and calls
 *    wind_willclose() again, which contines as it would have the first
 *    time if the patcher had been clean to begin with.
 */

static void verifyCloseCallback(Widget w, void *mode, void *call_data)
{
    if (current_window)         /* FIXME -- send as arg */
    {
        current_window->w_dirty = 0;    /* force close */
        close_patcher_window(current_window);
    }

    XtPopdown(XtParent(w));     /* popdown dialog */
}

static void oopsCallback(Widget w, void *mode, void *call_data)
{
    bug("wind_save");
#if 0
    if (wind_save(current_window) == 1)
    {
        panel_ouch(OUCH_SAVE, strerror(errno));
        return;                 /* i.e. don't close */
    }
    else
    {
        current_window->w_dirty = 0;    /* force close */
        close_patcher_window(current_window);
    }
#endif
    XtPopdown(XtParent(w));     /* popdown dialog */
}


/*
 * Menu callbacks
 *
 * Each main window menu has one callback routine which takes action based
 * on the value of `mode'.
 * These routines can be invoked via a menu item callback or a translation
 * to an Xt "action" (exception: the Font menu.  There are no actions to 
 * change fontsize).
 */

static void menuCallback(Widget w, void *mode, void *call_data)
{
    switch ((int) mode)
    {

    case FILE_NEW:
        create_patcher(FILE_NEW, "NEW");
        break;

    case FILE_OPEN:
        /* pass control to fileselectCallback */
        popup_file_selector("Open Patch", FILE_OPEN);
        break;

    case FILE_SAVE:
        if (current_window)
        {
            if (wind_save(current_window) == 1)
                panel_ouch("%s", getcreateerror());
        }
        else
            console_msg("SAVE: no current patcher\n");
        break;

    case FILE_SAVE_AS:
        /* pass control to fileselectCallback */
        popup_file_selector("Save Patch", FILE_SAVE_AS);
        break;

    case FILE_PRINT:
        console_msg("\"Print\" not supported\n");
        break;

    case FILE_CLOSE:
        if (current_window)
            close_patcher_window(current_window);
        else
            console_msg("CLOSE: no current patcher\n");
        break;

    case EXIT_MAX:
        /*  Exit dialog --  confirm exit
         */
        do_exit_dialog();
        break;

    case EDIT_UNDO:
        if (current_window)
            undo(current_window);
        break;

    case EDIT_CUT:
        if (!current_window)
            console_msg("EDIT: no current patcher\n");
        else
            cut(current_window);
        break;

    case EDIT_CLEAR:
        if (current_window)
            wind_clear(current_window);
        break;

    case EDIT_COPY:
        if (!current_window)
            console_msg("EDIT: no current patcher\n");
        else
            copy(current_window);
        break;

    case EDIT_PASTE:
        if (!current_window)
            console_msg("EDIT: no current patcher\n");
        else
            paste(current_window);
        break;

    case EDIT_DUPLICATE:
        if (!current_window)
            console_msg("EDIT: no current patcher\n");
        else
            duplicate(current_window);
        break;

    case EDIT_SELECT_ALL:
        if (!current_window)
            console_msg("EDIT: no current patcher\n");
        else
            selectall(current_window);
        break;

    case MISC_EDIT:
        if (current_window)
            edit(current_window);
        else
            console_msg("MISC EDIT: no current patcher\n");
        break;

    case MISC_STRETCH:
        if (current_window)
            stretch(current_window);
        else
            console_msg("MISC STRETCH: no current patcher\n");
        break;

    case MISC_CONTRACT:
        if (current_window)
            contract(current_window);
        else
            console_msg("MISC CONTRACT: no current patcher\n");
        break;

    case MISC_FIXWIDTH:
        if (current_window)
            fixwidth(current_window);
        break;
    case MISC_LINEUP:
        if (current_window)
            lineup(current_window);
        break;

    case MISC_SETTINGS:
        console_msg("\"Settings\" not supported\n");
        break;

    case MISC_RESTART:
        max_forkfts();
        break;

    case MISC_FIND:
        console_msg("\"Find\" not supported\n");
        break;

    case MISC_FIND_AGAIN:
        console_msg("\"Find Again\" not supported\n");
        break;

    case HELP_INFO:
        /*  Info dialog --- message text from nx_charlie.m 
         */
        if (!info_dialog)
        {
            static char message[BUFSIZ];
            XmString message_string = NULL;
            sprintf(message, "Max version %s\nBy Miller Puckette\n\
with contributions from Maurizio de Cecco, Joseph Francis, Cort Lippe,\n\
Zack Settel, Eric Viara, David Yadegari, and David Zicarelli.\n\
(c) IRCAM 1990-1994\n\
(C) Regents of the University of California 1995-1996\n", max_version);
            STRset(message_string, message);
            SetArg(messageString, message_string);
            SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE);
            info_dialog = CreateInformationDialog(toplevel, "infoDialog");
            STRfree(message_string);
            Unmanage(XmMessageBoxGetChild
                     (info_dialog, XmDIALOG_CANCEL_BUTTON));
            Unmanage(XmMessageBoxGetChild
                     (info_dialog, XmDIALOG_HELP_BUTTON));
        }
        popup_dialog(info_dialog);
        break;

    case HELP_INTRO:
        create_patcher(HELP_INTRO, "HELP");
        break;

    case HELP_CONTEXT:
        {
            static Cursor context_cursor = 0;
            Widget context_widget;

            if (context_cursor == 0)
                context_cursor =
                    XCreateFontCursor(display, XC_question_arrow);
            /* Change cursor shape and allow user to click on a patcher object */
            context_widget =
                XmTrackingLocate(toplevel, context_cursor, FALSE);
            /* Ignore clicks not on patcher windows */
            if (context_widget != NULL
                && !strcmp(XtName(context_widget), "canvas"))
            {
                /* Get the (x, y) coords of the pointer */
                Window root, child;
                int rootx, rooty, wx, wy;
                unsigned int mask;
                Point pt;
                t_wind *w = 0;
                XQueryPointer(display, XtWindow(context_widget), &root,
                              &child, &rootx, &rooty, &wx, &wy, &mask);
                /* Try and get help by jumping into patcher_click() with the modifiers set
                 * to trigger help 
                 */
                SetArg(userData, &w);
                GetValues(context_widget);
                pt.h = wx;
                pt.v = wy;
                if (w && ispatcher(w->w_assoc))
                    patcher_click(w->w_assoc, pt, 0, optionKey);
            }
        }
        break;
    }
}

/*
 * fontCallback()
 *
 * Is a valueChangedCallback so handled differently from other menu callbacks
 */

static void fontCallback(Widget w, void *fontsize, void *cb)
{
    if (((XmToggleButtonCallbackStruct *) cb)->set)
    {                           /* ignore unsetting */
        if (current_window)
        {
            font(current_window, (int) fontsize);
            wind_update(current_window);
        }
        else if ((int) fontsize != -1)
            t_setdeffont((int) fontsize);       /* no current patcher: set default and current fontsize */
    }
}

/* FIXME replace canvas actions by key-and mouse-down event parsing ala MAC  */


/*
 * Double Clicks
 *
 * Have to handle double clicks via actions, which seems totally broken.
 */

static void doubleclickAction(Widget w, XEvent *event, String *x3,
                              Cardinal *x4)
{
    /* mouse_x and mouse_y are set by the normal event handling */
    /* last parameter `2' is the double-click */
    mouseDown(find_window_for_widget(w), mouse_x - wind_xoffset,
              mouse_y - wind_yoffset,
              mouse_y <= current_window->w_topmargin ? inMargin : 0, 2);
}

/*
 * Alt Clicks
 *
 * Have to handle alt clicks via actions, which seems totally broken.
 * This routine will only work when the action triggering it is bound to a canvas.
 */

static void altclickAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    Point pt;
    t_wind *wind = NULL;

    /* mouse_x and mouse_y are set by the normal event handling */
    pt.h = event->xbutton.x;
    pt.v = event->xbutton.y;
    SetArg(userData, &wind);    /* i.e. must be a canvas */
    GetValues(w);
    if (wind && ispatcher(wind->w_assoc))
        patcher_click(wind->w_assoc, pt, 0, optionKey); /* optionKey -> get help */
}

/*
 * Action procedures
 *
 * There is an action procedure for every menu option except the font menu items.
 */

static void filenewAction(Widget w, XEvent *event, String *x3,
                          Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_NEW, NULL);
}

static void fileopenAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_OPEN, NULL);
}

static void filesaveAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_SAVE, NULL);
}

static void filesaveasAction(Widget w, XEvent *event, String *x3,
                             Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_SAVE_AS, NULL);
}

static void fileprintAction(Widget w, XEvent *event, String *x3,
                            Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_PRINT, NULL);
}

static void filecloseAction(Widget w, XEvent *event, String *x3,
                            Cardinal *x4)
{
    menuCallback(w, (caddr_t) FILE_CLOSE, NULL);
}

static void exitAction(Widget w, XEvent *event, String *x3, Cardinal *x4)
{
    menuCallback(w, (caddr_t) EXIT_MAX, NULL);
}

static void editcutAction(Widget w, XEvent *event, String *x3,
                          Cardinal *x4)
{
    menuCallback(w, (caddr_t) EDIT_CUT, NULL);
}

static void editcopyAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) EDIT_COPY, NULL);
}

static void editpasteAction(Widget w, XEvent *event, String *x3,
                            Cardinal *x4)
{
    menuCallback(w, (caddr_t) EDIT_PASTE, NULL);
}

static void editduplicateAction(Widget w, XEvent *event, String *x3,
                                Cardinal *x4)
{
    menuCallback(w, (caddr_t) EDIT_DUPLICATE, NULL);
}

static void editselectallAction(Widget w, XEvent *event, String *x3,
                                Cardinal *x4)
{
    menuCallback(w, (caddr_t) EDIT_SELECT_ALL, NULL);
}

static void misceditAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_EDIT, NULL);
}

static void miscstretchAction(Widget w, XEvent *event, String *x3,
                              Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_STRETCH, NULL);
}

static void misccontractAction(Widget w, XEvent *event, String *x3,
                               Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_CONTRACT, NULL);
}

static void miscfixwidthAction(Widget w, XEvent *event, String *x3,
                               Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_FIXWIDTH, NULL);
}

static void misclineupAction(Widget w, XEvent *event, String *x3,
                             Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_LINEUP, NULL);
}

static void miscsettingsAction(Widget w, XEvent *event, String *x3,
                               Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_SETTINGS, NULL);
}

static void miscrestartAction(Widget w, XEvent *event, String *x3,
                              Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_RESTART, NULL);
}

static void miscfindAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_FIND, NULL);
}

static void miscfindagainAction(Widget w, XEvent *event, String *x3,
                                Cardinal *x4)
{
    menuCallback(w, (caddr_t) MISC_FIND_AGAIN, NULL);
}

static void helpinfoAction(Widget w, XEvent *event, String *x3,
                           Cardinal *x4)
{
    menuCallback(w, (caddr_t) HELP_INFO, NULL);
}

static void helpintroAction(Widget w, XEvent *event, String *x3,
                            Cardinal *x4)
{
    menuCallback(w, (caddr_t) HELP_INTRO, NULL);
}

static void helpcontextAction(Widget w, XEvent *event, String *x3,
                              Cardinal *x4)
{
    menuCallback(w, (caddr_t) HELP_CONTEXT, NULL);
}

/*
 * Actions
 *
 * All menu options have an action name so that they can be
 * described in translation tables.
 */

static XtActionsRec actions_table[] = {
    { "file-new", filenewAction },
    { "file-open", fileopenAction },
    { "file-save", filesaveAction },
    { "file-save-as", filesaveasAction },
    { "file-print", fileprintAction },
    { "file-close", filecloseAction },
    { "file-exit", exitAction },

    { "edit-cut", editcutAction },
    { "edit-copy", editcopyAction },
    { "edit-paste", editpasteAction },
    { "edit-duplicate", editduplicateAction },
    { "edit-select-all", editselectallAction },

    /* nothing for fonts */

    { "misc-edit", misceditAction },
    { "misc-stretch", miscstretchAction },
    { "misc-contract", misccontractAction },
    { "misc-settings", miscsettingsAction },
    { "misc-restart", miscrestartAction },
    { "misc-find", miscfindAction },
    { "misc-find-again", miscfindagainAction },

    { "help-info", helpinfoAction },
    { "help-intro", helpintroAction },
    { "help-context", helpcontextAction },

    { "double-click", doubleclickAction },
    { "alt-click", altclickAction },
};

/*
 * Menu sensitivity
 */

static void dim_menus()
{
    /* bullshit */
}

/*void*/ undim_menus(int titled)
{
    /* bullshit */
}

/*
 * setupmainwindow()
 *
 * Create the main window which includes the menubar and all menus, and the
 * console.
 *
 * Also creates some dialogs so they can be popped up quickly.  Most dialogs are 
 * created when they are first needed to improve startup time.
 *
 * We also set the global fontlist (by asking a widget for it).  This assumes that
 * the fontlist resource is set so we better have a fallback for it.
 */
static Widget anybutton;
extern Pixmap icon_pixmap;

void setupmainwindow()
{
    /* Set default font size */
    t_initdeffont();

    display = XtDisplay(toplevel);

    /* Main window */
    main_window = CreateForm(toplevel, "main_window");
    Manage(main_window);

    /* Menu bar */
    SetArg(leftAttachment, XmATTACH_FORM);
    SetArg(rightAttachment, XmATTACH_FORM);
    menu_bar = wind_menubar(main_window, 0);

    /* The console will display FTS and error output */

    /* Console title label */
    SetArg(topWidget, menu_bar);
    SetArg(topAttachment, XmATTACH_WIDGET);
    SetArg(rightAttachment, XmATTACH_FORM);
    SetArg(leftAttachment, XmATTACH_FORM);
    console_title = CreateLabel(main_window, "consoleTitle");
    Manage(console_title);

    /* Frame console */
    SetArg(topWidget, console_title);
    SetArg(topAttachment, XmATTACH_WIDGET);
    SetArg(rightAttachment, XmATTACH_FORM);
    SetArg(leftAttachment, XmATTACH_FORM);
    SetArg(bottomAttachment, XmATTACH_FORM);
    SetArg(shadowType, XmSHADOW_OUT);
    frame = CreateFrame(main_window, "frame");
    Manage(frame);

    /* Console */
    SetArg(editable, False);
    SetArg(editMode, XmMULTI_LINE_EDIT);
    SetArg(wordWrap, True);
    SetArg(scrollVertical, True);
    SetArg(blinkRate, 0);
    SetArg(autoShowCursorPosition, True);
    SetArg(cursorPositionVisible, False);
    SetArg(scrollingPolicy, XmAUTOMATIC);
    SetArg(scrollBarDisplayPolicy, XmAS_NEEDED);
    SetArg(shadowThickness, 2);

    console = CreateScrolledText(frame, "console");
    console_msg("Max version %s\n(c) IRCAM 1990, 1993\n", max_version);
    Manage(console);

    /* get the Max FontList */
    SetArg(fontList, &fontlist);
    GetValues(anybutton);       /* any menu button or option will have it */

    t_initbits();

    /* Set icon pixmap, icon name, and toplevel window title */
    SetArg(iconPixmap, icon_pixmap);
    SetArg(iconName, getappname());
    SetArg(title, getappname());
    SetArg(deleteResponse, XmDO_NOTHING);
    SetValues(toplevel);

    /* 
     * From Motif FAQ --- to trap MWM f.kill (i.e. Close)
     *
     * This will still kill the application.  To turn this behaviour off 
     * so that the application is not killed, set the shell resource 
     * XmNdeleteResponse to XmDO_NOTHING.
     * This means that users cannot kill your application via the system menu, 
     * and may be a bad thing.
     */
    wm_delete_window = XmInternAtom(display, "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(toplevel, wm_delete_window,
                            (XtCallbackProc) app_killed, NULL);

    /* Set Xt error handler */
    XtAppSetErrorHandler(app_context, Xt_error_handler);
    XtAppSetWarningHandler(app_context, Xt_error_handler);

    /* Set X error handler */
    XSetErrorHandler(X_error_handler);

    /* Actions */
    XtAppAddActions(app_context, actions_table, XtNumber(actions_table));
}


#ifdef NO_STRDUP                /* not everyone has this, incredibly */
char *strdup(char *s1)
{
    char *copy;
    if (s1)
    {
        copy = malloc(strlen(s1) + 1);
        strcpy(copy, s1);
        return copy;
    }
    else
    {
        return NULL;
    }
}
#endif

/* from nx_view.m */
/* ============== */

typedef void *(*t_foo)();
#define hit0(x, y)    (*((t_foo)egetfn(x, y)))(x)
#define hit1(x, y, z1)    (*((t_foo)egetfn(x, y)))(x, z1)
#define hit2(x, y, z1, z2)  (*((t_foo)egetfn(x, y)))(x, z1,z2)
#define hit3(x, y, z1, z2, z3)  (*((t_foo)egetfn(x, y)))(x, z1,z2,z3)

long max_serialno;              /* incremented each event */

void wind_setpollon()
{
    wind_pollon = 1;
}

static int lastx, lasty;

void wind_poll()
{
    if (wind_pollon && current_window && current_window->w_vis &&
        (lastx != mouse_x || lasty != mouse_y))
    {
        GrafPort *gp = wind_setport(current_window);
        Point pt;
        pt.h = lastx = mouse_x - wind_xoffset;
        pt.v = lasty = mouse_y - wind_yoffset;
        hit1(current_window->w_assoc, &s_idle, pt);
        t_resetport(gp);
    }
}

/* void wind_forcepoll() { lastx = lasty = -12345; } */

static t_method drag_fun;
static void *drag_arg;
static Point drag_pt;
static t_wind *drag_wind;

void dragfnpt(Point pt, int x)
{
    /* same as NeXT */
    GrafPort *gunk;

    if (!drag_wind)
        return;
    gunk = wind_setport(drag_wind);
    if (!gunk)
        bug("gunk");
    if (pt.h != drag_pt.h || pt.v != drag_pt.v || !x)
    {
        drag_pt = pt;
        (*drag_fun) (drag_arg, pt, x);
    }
    if (!x)
        wind_nodrag();
    t_resetport(gunk);
}

void mouseDown(t_wind *wind, int x, int y, int flags, int doubleclick)
{
    Point p2;
    GrafPort *gunk;

    /* post("mouse down %d %d %d",  x,  y,  flags); */
    max_serialno++;
    if (flags & inMargin)
        gunk = wind_marginport(wind);
    else
        gunk = wind_setport(wind);
    getMouseCoords(&p2);
    if (!(flags & inMargin))
        p2.v -= wind->w_topmargin, p2.h -= wind->w_leftmargin;
    hit3(wind->w_assoc, &s_click, p2, doubleclick, flags);
    t_resetport(gunk);
    chknoport();
}

void mouseUp(t_wind *wind, int x, int y)
{
    GrafPort *gunk;
    Point p2;

    max_serialno++;
    gunk = wind_setport(wind);
    getMouseCoords(&p2);
    p2.v -= wind->w_topmargin, p2.h -= wind->w_leftmargin, dragfnpt(p2, 0);
    t_resetport(gunk);
    chknoport();
}

void keyDown(t_wind *wind, char charCode, int flags)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit2(wind->w_assoc, &s_key, charCode, flags);
    t_resetport(gunk);
    chknoport();
}

void keyUp(t_wind *wind, char charCode, int flags)
{
    max_serialno++;
}

void wind_clear(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    hit0(wind->w_assoc, &s_clr);
    t_resetport(gunk);
    chknoport();
}

void undo(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    hit0(wind->w_assoc, &s_undo);
    t_resetport(gunk);
    chknoport();
}

void cut(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    hit0(wind->w_assoc, &s_cut);
    t_resetport(gunk);
    chknoport();
}

void copy(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit0(wind->w_assoc, &s_copy);
    t_resetport(gunk);
    chknoport();
}

void paste(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit0(wind->w_assoc, &s_paste);
    t_resetport(gunk);
    chknoport();
}

void duplicate(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit0(wind->w_assoc, &s_dup);
    t_resetport(gunk);
    chknoport();
}

void selectall(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit0(wind->w_assoc, gensym("selectall"));
    t_resetport(gunk);
    chknoport();
}

void edit(t_wind *wind)
{
    GrafPort *gunk = wind_setport(wind);
    max_serialno++;
    hit0(wind->w_assoc, gensym("edit"));
    t_resetport(gunk);
    chknoport();
}

void wind_generic(t_wind *x, char *s, int n)
{
    GrafPort *gunk = wind_setport(x);
    max_serialno++;
    hit1(x->w_assoc, gensym(s), n);
    t_resetport(gunk);
    chknoport();
}

void font(t_wind *wind, int fontsize)
{
    wind_generic(wind, "font", fontsize);
}

void stretch(t_wind *wind)
{
    wind_generic(wind, "stretch", 0);
}

void contract(t_wind *wind)
{
    wind_generic(wind, "stretch", 1);
}

void fixwidth(t_wind *wind)
{
    wind_generic(wind, "fixwidth", 0);
}

void lineup(t_wind *wind)
{
    wind_generic(wind, "lineup", 0);
}

void view_poll()
{
    Point pt;
    GrafPort *gp;
    if (!drag_wind)
        return;
    gp = wind_setport(drag_wind);
    getMouseCoords(&pt);
    pt.v -= wind_yoffset;
    pt.h -= wind_xoffset;
    dragfnpt(pt, 1);
    t_resetport(gp);
}

void setwindowtitle(t_wind *x, char *s, char *t)
{
    char namebuf[256];
    char dirt = (x->w_dirty ? '*' : ' ');
    if (*t)
    {
        if (x->w_titled)
            sprintf(namebuf, "%.50s%c - %.70s", s, dirt, t);
        else
            sprintf(namebuf, "[%.50s%c - %.70s]", s, dirt, t);
    }
    else
    {
        sprintf(namebuf, "%.50s%c", s, dirt);
    }
    if (x->w_maincanvas)
    {
        SetArg(title, namebuf);
        SetArg(iconName, s);
        SetValues(x->w_canvasshell);
    }
}

void wind_filename(t_wind *x, char *fn, t_symbol *vol, int bin)
{
    /* same as NeXT */
    strcpy(x->w_name, fn);
    x->w_vol = vol;
    x->w_bin = bin;
    x->w_titled = 1;
    if (x->w_vis)
        setwindowtitle(x, x->w_name, x->w_vol->s_name);
}

void wind_setuntitled(t_wind *x)
{
    /* same as NeXT */
    wind_filename(x, "Untitled", defvolume, 0);
    x->w_titled = 0;
}

void wind_unfilename(t_wind *x)
{
    /* same as NeXT */
    x->w_titled = 0;
    if (x->w_vis)
        setwindowtitle(x, x->w_name, x->w_vol->s_name);
}

void wind_settitle(t_wind *x, char *s)
{
    /* same as NeXT */
    char buf[80];
    sprintf(buf, "[%s]", s);
    strcpy(x->w_name, buf);
    x->w_vol = &s_;
    x->w_bin = 0;
    x->w_titled = 0;
    if (x->w_vis)
        setwindowtitle(x, buf, "");
}

void wind_nodrag()
{
    /* same as NeXT */
    drag_wind = 0;
}

void wind_drag(t_wind *x, t_method fun, void *arg)
{
    /* same as NeXT */
    if (!fun)
        wind_nodrag();
    else
    {
        drag_fun = fun;
        drag_arg = arg;
        drag_wind = x;
    }
}

t_method wind_getdrag()
{
    if (drag_wind)
        return (drag_fun);
    else
        return (0);
}

void wind_dirty(t_wind *x, int d)
{
    x->w_dirty = d;
    setwindowtitle(x, x->w_name, x->w_vol->s_name);
}

int wind_save(t_wind *x)
{
    noundo(x);
    if (x->w_titled)
    {
        if (!hit2(x->w_assoc, gensym("saveto"), x->w_name, x->w_vol))
        {
            wind_dirty(x, 0);
            return 0;           /* Save OK */
        }
        else
            return 1;           /* error */
    }
    else
    {
        /* no title - try to get one */
        popup_file_selector("Save Patch", FILE_SAVE_AS);
        return 0;               /* OK */
    }
}

int wind_saveas(t_wind *x, char *name)
{
    char basename[512], dir[512], wtitle[100], *savedir;
    int doit;

    noundo(x);
    t_decomposename(name, 512, dir, basename);
    setopenpaneldir(name);
    if (!hit2(x->w_assoc, gensym("saveto"), basename, gensym(dir)))
    {
        wind_dirty(x, 0);
        if (!x->w_keeptitle)
        {
            strcpy(x->w_name, basename);
            x->w_vol = gensym(dir);
            x->w_titled = 1;
            setwindowtitle(x, x->w_name, dir);
        }
        return 0;               /* OK */
    }
    else
        return 1;               /* error */
}

void wind_issubpatch(t_wind *x)
{
    /* same as NeXT */
    x->w_keeptitle = 1;
}

void wind_vis(t_wind *x)
{
    if (!x->w_vis)
    {
        wind_dovis(x);
        wind_update(x);
    }
    else
        XRaiseWindow(display, XtWindow(x->w_canvasshell));
}

void wind_visnoupdate(t_wind *x)
{
    if (!x->w_vis)
    {
        wind_dovis(x);
    }
}

wind_close(t_wind *x)
{
    x->w_vis = 0;
    if (x == drag_wind)
        wind_nodrag();

    if (x->w_maincanvas)
    {

        XtDestroyWidget(x->w_canvasshell);
        x->w_palette = 0;
        x->w_maincanvas = 0;
        x->w_margincanvas = 0;
        x->w_canvasshell = 0;

        /* want to switch to another window here */
        if (x == current_window)
        {
            current_window = x->w_next;
            if (current_window == x)
                current_window = 0;
        }
        remove_from_window_list(x);

        if (current_window && current_window->w_maincanvas)
        {
            XSetInputFocus(display,
                           XtWindow(current_window->w_canvasshell),
                           RevertToPointerRoot, CurrentTime);
        }
        else
        {
            dim_menus();
        }
    }
}

void wind_invis(t_wind *x)
{
    /* same as NeXT */
    wind_close(x);
}

void wind_ff(t_wind *x)
{
    /* same as NeXT */
    wind_close(x);
}

int wind_willclose(t_wind *x)
{
    noundo(x);
    if (x->w_refcount || !zgetfn(x->w_assoc, &s_close))
    {
        hit0(x->w_assoc, &s_invis);
        x->w_vis = 0;
    }
    else
    {
        if (x->w_dirty)
        {
            /* verify closure ... */
            if (!x->w_closedialog)
            {
                SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE);
                SetArg(dialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
                x->w_closedialog =
                    CreateWarningDialog(x->w_canvasshell, "closeDialog");
                XtAddCallback(x->w_closedialog, XmNokCallback,
                              verifyCloseCallback, (XtPointer) FILE_CLOSE);
                XtAddCallback(x->w_closedialog, XmNhelpCallback,
                              oopsCallback, (XtPointer) FILE_SAVE);
            }
            popup_dialog(x->w_closedialog);
            /* postpone close until close dialog confirms it */
            return 1;
        }
        hit0(x->w_assoc, &s_close);
    }
    return 0;                   /* close ok */
}

void *wind_front()
{
    /* NOTE: I am not 100% sure of this ... foley@iannis */
    Window win;
    int state;

    XGetInputFocus(display, &win, &state);
    return XtWindowToWidget(display, win);
}

wind_reframe(t_wind *x)
{
    /* for scrolling --- not currently implemented */
}

void wind_copytorect(t_wind *x, int left, int top, int bottom, int right)
{
    /* copy rectangle into paste/clip-board whatever --- not currently implemented */
}

Widget wind_menubar(Widget form, int really)
{
    Widget file_menu, edit_menu, font_menu, misc_menu, help_menu;
    Widget file_button, edit_button, font_button, misc_button, help_button;


    /* Menu options  */
    Widget new_option, open_option, open_text_option, print_option,
        exit_option;
    Widget font0_option, font1_option, font2_option, font3_option,
        font4_option;
    Widget font5_option, font6_option, font7_option, fontbomb_option;
#ifdef NOTYET
    Widget settings_option;
    Widget find_option, find_again_option;
#endif
    Widget info_option, intro_option, context_option;
    Widget save_option, save_as_option, close_option;
    Widget cut_option, copy_option, paste_option, dup_option,
        select_all_option;
    Widget edit_option, stretch_option, contract_option;
    Widget undo_option, lineup_option, fixwidth_option, clear_option;
    SetArg(leftAttachment, XmATTACH_FORM);
    SetArg(rightAttachment, XmATTACH_FORM);
    menu_bar = CreateMenuBar(form, "menuBar");

    /* File menu */
    file_menu = CreatePulldownMenu(menu_bar, "file_menu");
    SetArg(subMenuId, file_menu);
    anybutton = file_button = CreateCascadeButton(menu_bar, "fileMenu");
    Manage(file_button);

    CreateMenuButton(new_option, file_menu, "newOption",
                     menuCallback, (caddr_t) FILE_NEW);
    CreateMenuButton(open_option, file_menu, "openOption",
                     menuCallback, (caddr_t) FILE_OPEN);
    if (really)
    {
        CreateMenuButton(close_option, file_menu, "closeOption",
                         menuCallback, (caddr_t) FILE_CLOSE);
        CreateSeparator(file_menu, "separator");
        CreateMenuButton(save_option, file_menu, "saveOption",
                         menuCallback, (caddr_t) FILE_SAVE);
        CreateMenuButton(save_as_option, file_menu, "saveAsOption",
                         menuCallback, (caddr_t) FILE_SAVE_AS);
#if 0
        CreateSeparator(file_menu, "separator");
        CreateMenuButton(print_option, file_menu, "printOption",
                         menuCallback, (caddr_t) FILE_PRINT);
#endif
    }
    CreateSeparator(file_menu, "separator");
    CreateMenuButton(exit_option, file_menu, "exitOption",
                     menuCallback, (caddr_t) EXIT_MAX);

    if (really)
    {
        /* Edit menu */
        edit_menu = CreatePulldownMenu(menu_bar, "edit_menu");
        SetArg(subMenuId, edit_menu);
        edit_button = CreateCascadeButton(menu_bar, "editMenu");
        Manage(edit_button);

        CreateMenuButton(undo_option, edit_menu, "undoOption",
                         menuCallback, (caddr_t) EDIT_UNDO);
        CreateSeparator(edit_menu, "separator");
        CreateMenuButton(cut_option, edit_menu, "cutOption",
                         menuCallback, (caddr_t) EDIT_CUT);
        CreateMenuButton(copy_option, edit_menu, "copyOption",
                         menuCallback, (caddr_t) EDIT_COPY);
        CreateMenuButton(paste_option, edit_menu, "pasteOption",
                         menuCallback, (caddr_t) EDIT_PASTE);
        CreateMenuButton(clear_option, edit_menu, "clearOption",
                         menuCallback, (caddr_t) EDIT_CLEAR);
        CreateSeparator(edit_menu, "separator");
        CreateMenuButton(dup_option, edit_menu, "duplicateOption",
                         menuCallback, (caddr_t) EDIT_DUPLICATE);
        CreateMenuButton(select_all_option, edit_menu,
                         "selectAllOption", menuCallback,
                         (caddr_t) EDIT_SELECT_ALL);

        /* Font menu */
        SetArg(radioBehavior, True);
        SetArg(radioAlwaysOne, True);
        font_menu = CreatePulldownMenu(menu_bar, "optionMenu");
        SetArg(subMenuId, font_menu);
        font_button = CreateCascadeButton(menu_bar, "fontMenu");
        Manage(font_button);

        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[0]);
        CreateRadioButton(font0_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 0);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[1]);
        CreateRadioButton(font1_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 1);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[2]);
        CreateRadioButton(font2_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 2);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[3]);
        CreateRadioButton(font3_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 3);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[4]);
        CreateRadioButton(font4_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 4);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[5]);
        CreateRadioButton(font5_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 5);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[6]);
        CreateRadioButton(font6_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 6);
        sprintf(font_option_label, FONT_LABEL_TEMPLATE, pointsizes[7]);
        CreateRadioButton(font7_option, font_menu, font_option_label,
                          fontCallback, (caddr_t) 7);
        CreateSeparator(font_menu, "separator");
        CreateRadioButton(fontbomb_option, font_menu, "bombOption",
                          fontCallback, (caddr_t) - 1);

        /* Turn on the box of the default font size */
        sprintf(font_option_label, FONT_LABEL_TEMPLATE,
                pointsizes[t_getdeffont()]);
        SetArg(set, True);
        SetValues(XtNameToWidget(font_menu, font_option_label));

        /* Misc menu */
        misc_menu = CreatePulldownMenu(menu_bar, "misc_menu");
        SetArg(subMenuId, misc_menu);
        misc_button = CreateCascadeButton(menu_bar, "miscMenu");
        Manage(misc_button);

        CreateMenuButton(edit_option, misc_menu, "editOption",
                         menuCallback, (caddr_t) MISC_EDIT);
        CreateMenuButton(stretch_option, misc_menu, "stretchOption",
                         menuCallback, (caddr_t) MISC_STRETCH);
        CreateMenuButton(contract_option, misc_menu, "contractOption",
                         menuCallback, (caddr_t) MISC_CONTRACT);
        CreateMenuButton(fixwidth_option, misc_menu, "fixwidthOption",
                         menuCallback, (caddr_t) MISC_FIXWIDTH);
        CreateMenuButton(lineup_option, misc_menu, "lineupOption",
                         menuCallback, (caddr_t) MISC_LINEUP);
#ifdef NOTYET
        CreateMenuButton(settings_option, misc_menu, "settingsOption",
                         menuCallback, (caddr_t) MISC_SETTINGS);
        CreateMenuButton(find_option, misc_menu, "findOption",
                         menuCallback, (caddr_t) MISC_FIND);
        CreateMenuButton(find_again_option, misc_menu, "findAgainOption",
                         menuCallback, (caddr_t) MISC_FIND_AGAIN);
#endif
    }

    /* Help */
    help_menu = CreatePulldownMenu(menu_bar, "help_menu");
    SetArg(subMenuId, help_menu);
    help_button = CreateCascadeButton(menu_bar, "helpMenu");
    SetArg(menuHelpWidget, help_button);
    SetValues(menu_bar);
    Manage(help_button);

    CreateMenuButton(info_option, help_menu, "infoOption", menuCallback,
                     (caddr_t) HELP_INFO);
    CreateMenuButton(intro_option, help_menu, "introOption", menuCallback,
                     (caddr_t) HELP_INTRO);
    CreateMenuButton(context_option, help_menu, "contextOption",
                     menuCallback, (caddr_t) HELP_CONTEXT);
    Manage(menu_bar);

    dim_menus();
    return (menu_bar);
}

static void new_patcher_window(t_wind *new_window)
{
    Widget patcher_shell;
    Widget patcher_window, menubar, text_editor, palette, canvas;
    /* fprintf(stderr,  "new %x\n",  new_window); */
    /* Shell */
    /* ===== */
    new_window->w_canvasshell = patcher_shell =
        XtAppCreateShell(progname, getappname(),
                         applicationShellWidgetClass, display, NULL, 0);
    SetArg(iconPixmap, icon_pixmap);
    /* Get x,y coords from patcher */
    SetArg(x, new_window->w_x1);
    SetArg(y, new_window->w_y1);
    /* pretty arbitrary limits ... */
#define PATCHER_MIN_WIDTH 200
#define PATCHER_MIN_HEIGHT 200
    SetArg(minWidth, PATCHER_MIN_WIDTH);
    SetArg(minHeight, PATCHER_MIN_HEIGHT);
    SetArg(deleteResponse, XmDO_NOTHING);
    SetValues(patcher_shell);
    /* Catch MWM f.kill */

    XmAddWMProtocolCallback(patcher_shell, wm_delete_window,
                            (XtCallbackProc) app_closed, new_window);
    /* XtAddEventHandler(patcher_shell, FocusChangeMask, FALSE, patcher_event,
       new_window); */



    /* Form */
    /* ==== */
    SetArg(width, new_window->w_x2 - new_window->w_x1); /* from patcher */
    patcher_window = CreateForm(patcher_shell, "patcher");
    Manage(patcher_window);
    XtAddEventHandler(patcher_window, EnterWindowMask, FALSE,
                      patcher_event, new_window);
    XtAddEventHandler(patcher_window, LeaveWindowMask, FALSE,
                      patcher_event, new_window);
    XtAddEventHandler(patcher_window, FocusChangeMask, FALSE,
                      patcher_event, new_window);

    /* MENU BAR */
    /* ==== === */

    menubar = wind_menubar(patcher_window, 1);

#if PALETTE_WINDOW
    /* Palette */
    /* ======= */
    SetArg(topAttachment, XmATTACH_FORM);
    SetArg(leftAttachment, XmATTACH_FORM);
    SetArg(rightAttachment, XmATTACH_FORM);
    SetArg(height, pic0_height);        /* just big enough for palette pixmap */
    palette = CreateDrawingArea(patcher_window, "palette");     /* width same as Form's */
    Manage(palette);
    new_window->w_palette = palette;
    /* Palette event handling */
    XtAddCallback(palette, XmNinputCallback, paletteCallback, new_window);
#else
    new_window->w_palette = 0;
#endif

    /* Canvas */
    /* ====== */
#if PALETTE_WINDOW
    SetArg(topWidget, palette);
    SetArg(topAttachment, XmATTACH_WIDGET);
#else
    SetArg(topWidget, menubar);
    SetArg(topAttachment, XmATTACH_WIDGET);
#endif
    SetArg(leftAttachment, XmATTACH_FORM);
    SetArg(rightAttachment, XmATTACH_FORM);
    SetArg(bottomAttachment, XmATTACH_FORM);
    SetArg(height, new_window->w_topmargin + new_window->w_y2 - new_window->w_y1);      /* from patcher */
    SetArg(userData, new_window);       /* only used by context-sensitive help */
    canvas = CreateDrawingArea(patcher_window, "canvas");       /* width same as Form's */
    Manage(canvas);
    new_window->w_maincanvas = canvas;
    /* Canvas event handling */
    XtAddCallback(canvas, XmNresizeCallback, canvasCallback, new_window);
    XtAddCallback(canvas, XmNexposeCallback, canvasCallback, new_window);
    XtAddCallback(canvas, XmNinputCallback, canvasCallback, new_window);
    /* DrawingArea inputCallback does not handle Motion */
    XtAddEventHandler(canvas, PointerMotionMask, FALSE, patcher_event,
                      NULL);



    /* GCs */
    /* === */
    if (!gcs_created)
    {                           /* created with first patcher */
        Pixel fg, bg;
        SetArg(foreground, &fg);
        SetArg(background, &bg);
        GetValues(canvas);
        gc_values.foreground = fg;
        gc_values.background = bg;
        gc_values.function = GXcopy;
        draw_gc = XCreateGC(display, DefaultRootWindow(display),
                            GCForeground | GCBackground | GCFunction,
                            &gc_values);
        gc_values.foreground = bg;
        gc_values.background = fg;
        gc_values.function = GXxor;
        invert_gc = XCreateGC(display, DefaultRootWindow(display),
                              GCForeground | GCBackground | GCFunction,
                              &gc_values);
        gc_values.foreground = bg;
        gc_values.background = fg;
        gc_values.function = GXcopy;
        undraw_gc = XCreateGC(display, DefaultRootWindow(display),
                              GCForeground | GCBackground | GCFunction,
                              &gc_values);
/*
 * Before drawing with XmStringDraw, you must set the font id of the GC.
 * Otherwise you get a Bad Font error.
 *
 * You could call this a bug in Motif. You pass a GC to XmStringDraw, however, 
 * Motif wants to use the fonts from the font list to draw the string.  
 * Therefore it replaces the font of the GC temporarily with some fonts of its own 
 * as specified in the font list. In the end it tries to restore the old font of 
 * the GC. If a GC uses the default font, the client side GC structure does not have a
 * valid font id (that is the 0xffffffff you may see in the error message). Motif
 * tries to restore this invalid id at the end.
 *
 * The workaround is set the font id of the GC to any valid font ID
 */
        XSetFont(display, draw_gc, XLoadFont(display, "fixed"));

        gcs_created = 1;
    }

    add_to_window_list(new_window);
    new_window->w_closedialog = 0;

    /* up she comes */
    XtRealizeWidget(patcher_shell);
}


wind_dovis(t_wind *x)
{
    GrafPort *gp;

    /* Scrolling stuff not implemented */
    /* ... */

    x->w_vis = 1;
    if (!x->w_maincanvas)
        new_patcher_window(x);

    /* Scrolling stuff not implemented */
    /* ... */

    x->w_willclose = 0;

    gp = wind_setport(x);
    t_eraserect(0, 0, x->w_x2 - x->w_x1, x->w_y2 - x->w_y1);
    t_resetport(gp);

    setwindowtitle(x, x->w_name, x->w_vol->s_name);
}

static t_wind *noport;
/* #define NOVIEW ((Widget)1)
Widget thisview = NOVIEW; */
static GrafPort *thisport = &noport;

Widget t_getview()
{
    if (thisport == &noport)
        return (0);
    else
        return ((*thisport)->w_maincanvas);
}

void chknoport()
{
    /* same as NeXT */
    if (thisport != &noport)
        bug("port not reset");
    thisport = &noport;
}

void t_resetport(GrafPort *gp)
{
    if (thisport != gp)
    {
        if (thisport == &noport)
            fprintf(stderr, "ouch: t_resetport without setport\n");
        else
        {
            if (gp != &noport)
            {
                if (gp == &(*gp)->w_viacontent)
                    wind_xoffset =
                        (*gp)->w_leftmargin /*- (*gp)->w_xoffset*/ ,
                        wind_yoffset =
                        (*gp)->w_topmargin /*- (*gp)->w_yoffset*/ ;
                else if (gp == &(*gp)->w_viamargin)
                    wind_xoffset = 0, wind_yoffset = 0;
                else
                    bug("t_resetport");
            }
            thisport = gp;
        }
    }
}

GrafPort *wind_setportto(t_wind *x, GrafPort *gp)
{
    if (x->w_vis)
    {
        GrafPort *port = thisport;
        thisport = gp;
        return (port);
    }
    else
        return 0;
}

GrafPort *wind_setport(t_wind *x)
{
    wind_xoffset = x->w_leftmargin /*- x->w_xoffset*/ ;
    wind_yoffset = x->w_topmargin /*- x->w_yoffset*/ ;
    return wind_setportto(x, &x->w_viacontent);
}

GrafPort *wind_marginport(t_wind *x)
{
#if PALETTE_WINDOW
    return wind_setportto(x, x->w_margincanvas);
#else
    wind_xoffset = 0;
    wind_yoffset = 0;
    return wind_setportto(x, &x->w_viamargin);
#endif
}

/*void*/ wind_update(t_wind *x)
{
    GrafPort *gp;
    /* t_eraserect(0, 0, x->w_x2 - x->w_x1, x->w_y2 - x->w_y1); */
    gp = wind_setport(x);
    hit1(x->w_assoc, &s_update, 0);     /* draw patch */
    t_resetport(gp);
    gp = wind_marginport(x);
    hit1(x->w_assoc, &s_update, 1);     /* draw palette */
    t_resetport(gp);
}

void wind_getupdaterect(t_wind *x, int *a, int *b, int *c, int *d)
{
    Dimension x1 = 15, y1 = 25, w = 35, h = 45;

    SetArg(x, &x1);
    SetArg(y, &y1);
    SetArg(width, &w);
    SetArg(height, &h);
    GetValues(x->w_canvasshell);
    *a = 0;
    *b = 0;
    *c = x1 + w;
    *d = y1 + h;
}

void c_wind()
{
    /* same as NeXT */
    c_new(&s_int, wind_ff, sizeof(t_wind), 0, 0);
}

void *wind_messlist;            /* documentation would be nice ... */

int phase = 0;                  /* should be static to wind_new() */

t_wind *wind_new(void *assoc, int a, int b, int c, int d, int flags)
{
    t_wind *x =
        (t_wind *) obj_new((t_messlist **) & wind_messlist, c_wind);
    /* post("new: %lx",  x); */
    /* Skew each window a little more */
    phase = (phase + 1) % 6;
    if (a >= c || b >= d)
    {
        a = 100 + 50 * phase;
        b = 400 - 50 * phase;
        c = a + 400;
        d = b + 400;
    }
    x->w_x1 = a;
    x->w_y1 = b;
    x->w_x2 = c;
    x->w_y2 = d;
    x->w_xoffset = x->w_yoffset = 0;
    x->w_topmargin = x->w_bottommargin =
        x->w_leftmargin = x->w_rightmargin = 0;
    x->w_scrollgrain = 1;
    x->w_scroll = 0;
    x->w_scrollx = ((flags & WSCROLLX) != 0);
    x->w_scrolly = ((flags & WSCROLLY) != 0);
    x->w_refcount = x->w_titled = 0;
    x->w_assoc = assoc;
    x->w_vis = 0;
    x->w_palette = 0;
    x->w_maincanvas = 0;
    x->w_margincanvas = 0;
    x->w_canvasshell = 0;
    x->w_dirty = 0;
    x->w_keeptitle = 0;
    strcpy(x->w_name, "Untitled");
    x->w_vol = defvolume;
    x->w_printing = 0;
    x->w_previous = x->w_next = 0;
    x->w_closedialog = 0;
    x->w_viacontent = x->w_viamargin = x;
    if (flags & WVIS)
        wind_vis(x);
    return x;
}

/* Scrolling stuff which isn't implemented */

void wind_setsval(t_wind *x, int a, int b)      /* obsolete */
{
    /* same as NeXT */
}

void wind_setsmax(t_wind *x, int a, int b)      /* obsolete */
{
    /* same as NeXT */
}

void wind_setscrollrect(t_wind *x, int left, int top, int right,
                        int bottom)
{
    /* same as NeXT */
    x->w_xmin = left;
    x->w_xmax = right;
    x->w_ymin = top;
    x->w_ymax = bottom;
    wind_reframe(x);
}

void wind_setscrollparams(t_wind *x, int left, int top, int right,
                          int bottom)
{
    /* same as NeXT */
    x->w_scroll = 1;
    x->w_topmargin = top;
    x->w_leftmargin = left;
    x->w_bottommargin = bottom;
    x->w_rightmargin = right;
}

void wind_getscroll(t_wind *x, int *h, int *v)
{
#if 0
    int x1, y1;

    if (x->w_maincanvas)
    {
        SetArg(x, &x1);
        SetArg(y, &y1);
        GetValues(x->w_canvasshell);
        *h = x1;
        *v = y1;
    }
    else
    {
        *h = x->w_xoffset;
        *v = x->w_yoffset;
    }
#else
    *h = *v = 0;                /* X origin */
#endif
}

int wind_defaultscroll(t_wind *x, Point pt)
{
    /* same as NeXT */
    return 0;
}

/* end of nx_view.m */
/* ================ */


void panel_ouch(char *fmt, ...)
{
    va_list arglist;
    char buf[BUFSIZ];
    XmString message_string = NULL;

    va_start(arglist, fmt);
    vsprintf(buf, fmt, arglist);
    va_end(arglist);

    /*  Error dialog
     */
    if (!error_dialog)
    {
        SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE);
        error_dialog = CreateErrorDialog(toplevel, "errorDialog");
        Unmanage(XmMessageBoxGetChild
                 (error_dialog, XmDIALOG_CANCEL_BUTTON));
        Unmanage(XmMessageBoxGetChild(error_dialog, XmDIALOG_HELP_BUTTON));
    }
    STRset(message_string, buf);
    SetArg(messageString, message_string);
    SetValues(error_dialog);
    STRfree(message_string);

    popup_dialog(error_dialog);
}

/***************** cursor management ****************/

#if 0                           /* this isn't used */
static void arrowwindow(Window w)
{
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_top_left_arrow);
    XDefineCursor(display, XtWindow(t_getview()), c);
}
#endif

arrowcursor()
{
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_top_left_arrow);
    if (thisport != &noport)
        XDefineCursor(display, XtWindow(t_getview()), c);
}


waitcursor()
{
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_clock);
    if (thisport != &noport)
        XDefineCursor(display, XtWindow(t_getview()), c);
}

editcursor()                    /* should have been called "textcursor" */
{
    /* on the other hand I can't find a text selection cursor.
     * This does something inappropriate instead.
     */
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_top_left_arrow);
    if (thisport != &noport)
        XDefineCursor(display, XtWindow(t_getview()), c);
}

growcursor()
{
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_right_side);
    if (thisport != &noport)
        XDefineCursor(display, XtWindow(t_getview()), c);
}

connectcursor()
{
    static Cursor c = 0;
    if (!c)
        c = XCreateFontCursor(display, XC_circle);
    if (thisport != &noport)
        XDefineCursor(display, XtWindow(t_getview()), c);
}

/* from nx_charlie.m */
/* ================= */

extern void *lowload(char *name, t_symbol * volume, int ac,
                     struct atom *av, int couldedit);
extern t_symbol *t_getpath();
extern t_symbol *defvolume;
static t_symbol *openpaneldir = &s_;

void setopenpaneldir(char *s)
{
    /* same as NeXT */
    char dir[512], basename[512];
    t_decomposename(s, 512, dir, basename);
    openpaneldir = gensym(dir);
}

void *openfrompath(char *filename)
{
    char *name;
    int setit;                  /* sets file dialog window directory */
    void *x;
    struct atom at[10];
    char dir[512], basename[512];

    t_symbol *pathwas = t_getpath();
    name = strdup(filename);
    setit = 1;
    t_decomposename(name, 512, dir, basename);
    crosshatch(at);
    defvolume = gensym(dir);
    if (setit)
        setopenpaneldir(name);
    loadbang__doit(0);
    if ((x = lowload(basename, defvolume, 9, at, 1)) != (void *) 0)
        (*(egetfn(x, gensym("front")))) (x);
    else
        binbuf_ouch("%s: can't open", name);
    loadbang__doit(1);
    t_setpath(pathwas);
    free(name);
    return x;
}

/*
 * support for setting up open and save panels.  The panels are
 * non-modal; you supply a callback function and argument.
 */

static Widget t_panel = 0;
static void *t_panelclient;
static t_method t_panelfunction;

static void panelcallback(Widget w, int mode,
                          XmFileSelectionBoxCallbackStruct *cb)
{
    char *filepath;
    struct stat statbuf;
    if (!t_panelclient)
        return;
    /* extract dir and file name */
    XmStringGetLtoR(cb->value, XmSTRING_DEFAULT_CHARSET, &filepath);

    Unmanage(t_panel);
    XtRemoveAllCallbacks(t_panel, XmNokCallback);

    if (filepath)
    {
        switch (mode)
        {
        case CANCEL_SELECT:
            break;              /* nop */
        case FILE_OPEN:
            /* check file is ordinary */
            if (stat(filepath, &statbuf) == -1)
                panel_ouch(OUCH_ACCESS, strerror(errno));       /* cannot stat */
            else if (S_ISREG(statbuf.st_mode))
                (*t_panelfunction) (t_panelclient, filepath);
            else
                panel_ouch(OUCH_ORDINARY);      /* not ordinary */
            break;
        case FILE_SAVE_AS:
            (*t_panelfunction) (t_panelclient, filepath);
            break;
        }
    }
}

static void panel_popup(int mode, t_method f, void *arg)
{
    t_panelclient = arg;
    t_panelfunction = f;

    /* Create file selector */
    if (!t_panel)
    {
        XmString pattern_string = NULL;
        XmString dir_string = NULL;
        STRset(pattern_string, "*");
        STRset(dir_string, "");
        SetArg(dirMask, pattern_string);        /* Pattern */
        SetArg(directory, dir_string);  /* Directory */
        SetArg(mwmDecorations, MWM_DECOR_BORDER | MWM_DECOR_TITLE |
               MWM_DECOR_RESIZEH);
        t_panel = CreateFileSelectionDialog(toplevel, "fileSelector");
        STRfree(pattern_string);
        STRfree(dir_string);
        Unmanage(XmFileSelectionBoxGetChild
                 (t_panel, XmDIALOG_HELP_BUTTON));
        XtAddCallback(t_panel, XmNcancelCallback,
                      (t_callbackfunction) panelcallback,
                      (XtPointer) CANCEL_SELECT);
    }
    XtAddCallback(t_panel, XmNokCallback,
                  (t_callbackfunction) panelcallback, (XtPointer) mode);

    /* force scan to get up-to-the-minute listing */
    XmFileSelectionDoSearch(t_panel, NULL);
    popup_dialog(t_panel);
}

/* here are the functions you call: */

void panel_open(t_method f, void *arg)
{
    panel_popup(FILE_OPEN, f, arg);
}

void panel_save(t_method f, void *arg)
{
    panel_popup(FILE_SAVE_AS, f, arg);
}

void panel_cancel(void)
{
    t_panelclient = 0;
    t_panelfunction = 0;
}
