/*
 *      Copyright (c) 1994 Regents of the Univ. of California
 *                All rights reserved.
 *
 */

#define MIDI
#define MIDI5_3
#include <sys/lock.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/schedctl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/resource.h>
#include <errno.h>

#include "m_mess.h"
#include "cp_mess.h"
#include "d_ugen.h"


#include <fcntl.h>
#include "audio.h"

#include <sys/time.h>
#include <sys/times.h>
#include <sys/param.h>

/*
 * Globals
 */

char *progname = "fts";
int infile;                     /* input to Max */
int outfile;                    /* output from Max */

int cp_quit = 0;                /* quit flag; can be set from sched_tick */
int cp_this = 0;                /* ID number of this CP */
int cp_nproc = 1;               /* number of CPs */
int cp_usedacs;
int cp_slotno;
int cp_sr = 44100;              /* integer sampling rate use in dac_timefwd */
int cp_nchans;
int cp_use_audiofile;
static int cp_audio_open;       /* true if audio is open */
static int cp_advance;          /* scheduler advance */
static int sgi_realtime = 0;
static int cp_up2 = 0;

extern float dsp_sr;
#define NSPINCOUNT 16
int cp_spincount[NSPINCOUNT], cp_spinphase, cp_nextspin;
static long dactime = 1;
extern int dsp_ticks;

static long ticktime = 0;
extern long logicaltime;

extern float soundout[], soundin[];

#define DACVS 64                /* vector size for a/d/a and inter-CP signal paths */
#define NCHAN 8                 /* maximum number of DACs and ADCs */
#define XFERSIZE 64             /* number of samples we pass to dacs */
#define DACALIGN 16

int cp_xferframes;

static ALport iport;
static ALport oport;
static int sgi_opened;
static ALconfig sgi_config;
#define BUFTIME 100             /* .1 sec buffering -- should depend on sched advance */
#define MAXCH 4
#define DEFAULTCH 2


/* timout facility -- can be set to cause syscalls to return silently or to
 * dump core.
 */

static int nalarm = 0;
static int dump = 0;

void alarmhandler(int n)
{
    nalarm++;
    if (dump)
    {
        fprintf(stderr, "Abort 1\n");
        abort();
    }
    dump = 0;
}

void setalarm(int microsec, int dflag)
{
    struct itimerval gonzo, bonzo;
    gonzo.it_interval.tv_sec = 0;
    gonzo.it_interval.tv_usec = 0;
    gonzo.it_value.tv_sec = 0;
    gonzo.it_value.tv_usec = microsec;
    dump = 0;
    signal(SIGALRM, alarmhandler);
    setitimer(ITIMER_REAL, &gonzo, &bonzo);
    signal(SIGALRM, alarmhandler);
    dump = dflag;
}

#if 0
micropause(int microsec)
{
    setalarm(microsec, 0);
    pause();
}
#endif


void micropause(int microsec)
{
    fd_set readset, writeset, exceptset;
    struct timeval timout;
    timout.tv_sec = 0;
    timout.tv_usec = microsec;
    FD_ZERO(&writeset);
    FD_ZERO(&readset);
    FD_ZERO(&exceptset);
    select(0, &readset, &writeset, &exceptset, &timout);
}

void caught_exit()
{
    static int reentering;
    volatile int foo = reentering;
    reentering = 1;
#ifndef MIDI5_3
    if (!foo)
        system("stopmidi");
#endif
    fprintf(stderr, "Abort 2\n");
    abort();
}

cp_exit(n)
{
    system("stopmidi");
    exit(n);
}

void virtalarmhandler(int n)
{
    fprintf(stderr, "FTS died due to timeout\n");
    post("FTS died due to timeout");
    fflush(stderr);
    sys_trytoflush();
    sys_trytoflush();
#ifndef MIDI5_3
    system("stopmidi");
#endif
#ifdef DEBUG
    fprintf(stderr, "Abort 3\n");
    abort();
#endif
    exit(0);
}

void setvirtualalarm(int sec)
{
    struct itimerval gonzo, bonzo;
    gonzo.it_interval.tv_sec = 0;
    gonzo.it_interval.tv_usec = 0;
    gonzo.it_value.tv_sec = sec;
    gonzo.it_value.tv_usec = 0;
    signal(SIGVTALRM, virtalarmhandler);
    setitimer(ITIMER_VIRTUAL, &gonzo, &bonzo);
    signal(SIGVTALRM, virtalarmhandler);
}

#if 0
static struct timeval watchtimer;

startwatch()
{
    getrealtime(&watchtimer);
}

void bstopwatch(int n, int usec)
{
    if (sgi_realtime && logicaltime > 500 && usecsince(&watchtimer) > usec)
    {
        fprintf(stderr, "stopwatch %d %d->%d\n", n,
                usec, usecsince(&watchtimer));
        startwatch();
    }

}

void stopwatch(int n)
{
    bstopwatch(n, 5000);
}
#else
#define startwatch()
#define stopwatch(a)
#define bstopwatch(a,  b)

#endif

#include <sys/fpu.h>

  /*
     set the special "flush zero" but (FS, bit 24) in the
     Control Status Register of the FPU of R4k and beyond
     so that the result of any underflowing operation will
     be clamped to zero, and no exception of any kind will
     be generated on the CPU.  This has no effect on 
     an R3000.

     thanks to cpirazzi@cp.esd.sgi.com (Chris Pirazzi).

   */

void sgi_flush_all_underflows_to_zero()
{
    union fpc_csr f;
    f.fc_word = get_fpc_csr();
    f.fc_struct.flush = 1;
    set_fpc_csr(f.fc_word);
}

int sgi_open_audio()
{
    long pvbuf[4];
    long pvbuflen;
    int samplerate;
    /*
     *  get current sample rate -- should initialize to this...
     */
    pvbuf[0] = AL_INPUT_RATE;
    pvbuf[2] = AL_OUTPUT_RATE;
    pvbuflen = 4;

    ALgetparams(AL_DEFAULT_DEVICE, pvbuf, pvbuflen);

    if (pvbuf[1] < 1)
        pvbuf[1] = 44100;
    post("sample rate = %ld", pvbuf[1]);

    sgi_setchans(DEFAULTCH);
    sgi_setsamplerate((int) (samplerate = pvbuf[1]));

    pvbuf[3] = pvbuf[1];
    ALsetparams(AL_DEFAULT_DEVICE, pvbuf, pvbuflen);

    sgi_config = ALnewconfig();

    /* this is not good.  the queue size should reflect scheduler
       addvance; we should use 24 bit samples; and number of channels
       should be configurable. */
    if (sgi_realtime)
        ALsetqueuesize(sgi_config, BUFTIME * 48);
    else
        ALsetqueuesize(sgi_config, 1024L);
    /* ALsetqueuesize(sgi_config, 1024L); */
    ALsetwidth(sgi_config, AL_SAMPLE_16);
    ALsetsampfmt(sgi_config, AL_SAMPFMT_FLOAT);
    ALsetchannels(sgi_config, DEFAULTCH);

    oport = ALopenport("the ouput port", "w", sgi_config);
    if (!oport)
    {
        fprintf(stderr, "FTS: failed to open audio write port\n");
        return (1);
    }

    iport = ALopenport("the input port", "r", sgi_config);
    if (!iport)
    {
        fprintf(stderr, "FTS: failed to open audio read port\n");
        return (1);
    }
    sgi_opened = 1;
    return (0);
}

void sgi_set_dacs(int chans, int rate)
{
    if (rate < 1)
        rate = 44100;
    if (chans != 2 && chans != 4)
        chans = 2;
    if (sgi_config)
    {
        long pvbuf[4];
        long pvbuflen;

        pvbuf[0] = AL_INPUT_RATE;
        pvbuf[2] = AL_OUTPUT_RATE;
        pvbuf[3] = pvbuf[1] = (long) rate;
        pvbuflen = 4;
        ALsetparams(AL_DEFAULT_DEVICE, pvbuf, pvbuflen);

        ALgetparams(AL_DEFAULT_DEVICE, pvbuf, pvbuflen);

        rate = pvbuf[1];
        if (rate < 1)
            rate = 44100;
        post("sample rate = %d", rate);
        ALsetchannels(sgi_config, chans);

        ALcloseport(oport);
        ALcloseport(iport);
        oport = ALopenport("the ouput port", "w", sgi_config);
        if (!oport)
        {
            fprintf(stderr, "FTS: failed to open audio write port\n");
        }
        iport = ALopenport("the input port", "r", sgi_config);
        if (!iport)
        {
            fprintf(stderr, "FTS: failed to open audio read port\n");
        }

        sgi_setsamplerate(rate);
        sgi_setchans(chans);
    }
}

void sgi_close_audio()
{
    if (iport)
        ALcloseport(iport);
    if (oport)
        ALcloseport(oport);
}

void cp_setup2(void *x, long n)
{
    cp_up2 = n;
}

void sgi_senddacs(float *inbuffer, float *outbuffer, int n)
{
    float buf[2 * MAXCH * MAXVS], *fp1, *fp2;
    long fill, infill, fsamps = cp_advance * MAXVS;
    int i, nwait = 0;
    int sampstomove =
        (cp_up2 ? cp_nchans * (2 * MAXVS) : cp_nchans * MAXVS);
    fill = ALgetfilled(oport);
    infill = ALgetfilled(iport);
    if (fill <= sampstomove)
    {
        static int timeslipped;
        if (logicaltime > timeslipped)
        {
            timeslipped = logicaltime + 1000;
#if 0
            post("napping...");
#endif
            sginap(3);
            startwatch();
        }
        while ((infill =
                ALgetfilled(iport)) > sampstomove * (cp_advance + 1))
        {
            ALreadsamps(iport, buf, sampstomove);
        }
    }
    else if (sgi_realtime)
    {
        while (fill > sampstomove * cp_advance)
        {
            if (nwait++ > 1000)
            {
                post("FTS: DACS appear to have died");
                fprintf(stderr, "FTS: DACS appear to have died\n");
                sys_trytoflush();
                sys_trytoflush();
                hate("giving up on DACs");
            }
            startwatch();
            micropause((16000000 * cp_advance) / cp_sr);
            bstopwatch(1, 5000 + (16000000 * cp_advance) / cp_sr);
            startwatch();
            fill = ALgetfilled(oport);
        }
    }
    if (cp_up2)
    {
        if (cp_nchans == 2)
        {
            for (i = MAXVS, fp1 = outbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 4)
            {
                fp2[0] = fp2[2] = fp1[0];
                fp2[1] = fp2[3] = fp1[MAXVS];
            }
            ALwritesamps(oport, buf, 4 * MAXVS);
            if (infill > 4 * MAXVS)
                ALreadsamps(iport, buf, 4 * MAXVS);
            else
                bzero(buf, 4 * MAXVS * sizeof(float));
            for (i = MAXVS, fp1 = inbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 4)
            {
                fp1[0] = fp2[0];
                fp1[MAXVS] = fp2[1];
            }
        }
        else
        {
            for (i = MAXVS, fp1 = outbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 8)
            {
                fp2[0] = fp2[4] = fp1[0];
                fp2[1] = fp2[5] = fp1[MAXVS];
                fp2[2] = fp2[6] = fp1[2 * MAXVS];
                fp2[3] = fp2[7] = fp1[3 * MAXVS];
            }
            ALwritesamps(oport, buf, 8 * MAXVS);
            if (infill > 8 * MAXVS)
                ALreadsamps(iport, buf, 8 * MAXVS);
            else
                bzero(buf, 8 * MAXVS * sizeof(float));
            for (i = MAXVS, fp1 = inbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 8)
            {
                fp1[0] = fp2[0];
                fp1[MAXVS] = fp2[1];
                fp1[2 * MAXVS] = fp2[2];
                fp1[3 * MAXVS] = fp2[3];
            }
        }

    }
    else
    {
        if (cp_nchans == 2)
        {
            for (i = MAXVS, fp1 = outbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 2)
            {
                fp2[0] = fp1[0];
                fp2[1] = fp1[MAXVS];
            }
            ALwritesamps(oport, buf, 2 * MAXVS);
            if (infill > 2 * MAXVS)
                ALreadsamps(iport, buf, 2 * MAXVS);
            else
                bzero(buf, 2 * MAXVS * sizeof(float));
            for (i = MAXVS, fp1 = inbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 2)
            {
                fp1[0] = fp2[0];
                fp1[MAXVS] = fp2[1];
            }
        }
        else
        {
            for (i = MAXVS, fp1 = outbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 4)
            {
                fp2[0] = fp1[0];
                fp2[1] = fp1[MAXVS];
                fp2[2] = fp1[2 * MAXVS];
                fp2[3] = fp1[3 * MAXVS];
            }
            ALwritesamps(oport, buf, 4 * MAXVS);
            if (infill > 4 * MAXVS)
                ALreadsamps(iport, buf, 4 * MAXVS);
            else
                bzero(buf, 4 * MAXVS * sizeof(float));
            for (i = MAXVS, fp1 = inbuffer, fp2 = buf; i--;
                 fp1++, fp2 += 4)
            {
                fp1[0] = fp2[0];
                fp1[MAXVS] = fp2[1];
                fp1[2 * MAXVS] = fp2[2];
                fp1[3 * MAXVS] = fp2[3];
            }
        }
    }
    /* setalarm(10000,  1); */
}

static struct timeval fake_starttime;
static long fake_logicaltime;

void fake_openconnection()
{
    struct timezone tz;
    gettimeofday(&fake_starttime, &tz);
    cp_use_audiofile = 0;
    fake_logicaltime = 0;
}

int fake_dacshungry()
{
    struct timeval timenow;
    struct timezone tz;
    long elapsedtime;
    gettimeofday(&timenow, &tz);
    elapsedtime = 1000 * (timenow.tv_sec - fake_starttime.tv_sec) +
        (timenow.tv_usec - fake_starttime.tv_usec) / 1000;
    return (elapsedtime > fake_logicaltime);
}

void fake_senddacs()
{
    fake_logicaltime = logicaltime;
}

printlimit(int n)
{
    struct rlimit rlim;
    if (getrlimit(RLIMIT_CORE, &rlim))
        perror("getrlimit");
    fprintf(stderr, "lim %d %lx", n, rlim.rlim_cur);
}

int sgi_gotrealtime()
{
    return (sgi_realtime);
}

static void fixpri(void)
{
    char *shitfo = 0;
    char buf[100000];
    if (geteuid())
    {
        fprintf(stderr, "fts: warning: can't get real-time status\n");
        return;
    }
    sgi_realtime = 1;
    buf[0] = 1;
    /* bash priority to a high value.  */
#if 0
    if (!(shitfo = malloc(0x200000)))
        fprintf(stderr, "fts: warning: couldn't preallocate memory\n");
    else if (plock(PROCLOCK) < 0)
        fprintf(stderr, "fts: warning: couldn't lock into memory\n"),
            sgi_realtime = 0;
#endif
    if (schedctl(NDPRI, 0, NDPHIMAX) < 0)
        fprintf(stderr, "fts: warning: couldn't set priority\n"),
            sgi_realtime = 0;
    seteuid(getuid());
    if (shitfo)
        free(shitfo);
    if (!buf[0])
        post("foo");
}

main(int argc, char **argv)
{
    int realtimeplease;
    /* Max (the parent process) has duplicated stdin and stdout for a connection.
     * Alias the names since much code uses the names `infile' and `outfile'.
     */
    infile = STDIN_FILENO;
    outfile = STDOUT_FILENO;

    /* Trap Unix signals */
    signal(SIGINT, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGILL, caught_exit);
    signal(SIGEMT, caught_exit);
    signal(SIGBUS, caught_exit);
    signal(SIGSEGV, caught_exit);
    sgi_flush_all_underflows_to_zero();

    h_init(1);
    h_canpost();

    if (fcntl(outfile, F_SETFL, O_NONBLOCK) == -1)
    {
        perror("fcntl");
        hate("can't fix output port");
    }

    {
        struct rlimit rlim;
        if (getrlimit(RLIMIT_CORE, &rlim))
            perror("getrlimit");
        else
        {
#ifdef DEBUG
            rlim.rlim_cur = rlim.rlim_max;
#else
            rlim.rlim_cur = 0;
#endif
            if (setrlimit(RLIMIT_CORE, &rlim))
                perror("setrlimit");
        }
    }
#ifdef DEBUG
    printlimit(1);
#endif
    if (argc < 2 || (sscanf(argv[1], "%d", &cp_advance) < 1))
        cp_advance = 8;
    if (cp_advance < 3 || cp_advance > 50)
    {
        if (cp_advance)
            post("advance out of range (3-50); reset to 8");
        cp_advance = 8;
    }
    realtimeplease = (argc < 3 || *(argv[2]) == '0');

    if (realtimeplease)
        fixpri();
    else
        post("cancelling real-time priority setup...");


    if (!sgi_open_audio())
    {
        cp_audio_open = 1;
    }
    else
    {
        fake_openconnection();
        sgi_setchans(1);
        sgi_setsamplerate(8000);
    }

    cp_print(1);                /* start profiling */

    while (!cp_quit)
    {
        int foo = 0;
        cp_spinphase = (cp_spinphase + 1) & (NSPINCOUNT - 1);
        cp_spincount[cp_spinphase] = cp_nextspin;
        cp_nextspin = 0;
        sched_poll();           /* increment logicaltime */
        stopwatch(5);
#ifdef MIDI
        midi_poll();            /* keep up with MIDI in */
#endif
        stopwatch(6);
        dac_poll();
        host_poll();
        stopwatch(3);
        if (!foo--)
        {
            static int nexttime;
            hq_poll();
            if (sys_trytoflush())
                cp_quit = 1;
            if (logicaltime >= nexttime)
            {
                int foo = nexttime & 4;
                if (!foo)
                    max_sendtimeup();
                nexttime += 500;
                setvirtualalarm(10);
            }
            foo = 7;
        }
        stopwatch(4);
    }
    fprintf(stderr, "fts: stopping\n");
    h_done();
    sgi_close_audio();
#ifndef MIDI5_3
    system("stopmidi");
#endif
    exit(0);
}

long realtime()
{
    return dactime;
}

host_poll()
{
    char cmd[BUFSIZ];
    int ret, i, xxx;
    fd_set check;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = !cp_audio_open;

    FD_ZERO(&check);
    FD_SET(infile, &check);
    select(64, &check, (fd_set *) 0, (fd_set *) 0, &timeout);
    if (FD_ISSET(infile, &check))
    {
        if ((ret = read(infile, cmd, BUFSIZ)) < 0)
        {
            perror("read");
            hate("host connection broken");
        }
        else if (ret == 0)
        {
            cp_quit = 1;
            return;             /* EOF */
        }
        /* read ok */
        for (i = 0; i < ret; i++)
            cp_dochar(0, cmd[i]);
    }
}

  /* tell DSP network etc about new sample rate */
sgi_setsamplerate(int n)
{

    if (n <= 0)
        n = 1;
    dsp_sr = cp_sr = n;
}

    /* tell everyone what channel count we configured */
sgi_setchans(int n)
{
    if (n <= 0)
    {
        n = 0;
        cp_usedacs = 0;
    }
    else
    {
        if (n > NCHAN)
            n = NCHAN;
        cp_usedacs = 1;
    }
    cp_xferframes = XFERSIZE / n;
    cp_nchans = n;
    cu_zero(soundout, cp_nchans * MAXVS);
    cu_zero(soundin, cp_nchans * MAXVS);
}

dac_timefwd()
{
    static int dactime_frac;

    dactime_frac += 64000;
    while (dactime_frac >= cp_sr)
    {
        dactime++;
        dactime_frac -= cp_sr;
    }
}

static int cp_insomnia;         /* number of ticks since last sleep */

  /* NB: the use of `short' here is based on their being 16 bits */
dac_poll()
{
    sgi_senddacs(soundin, soundout, cp_nchans * DACVS);
    cu_zero(soundout, cp_nchans * MAXVS);
    dac_timefwd();
    stopwatch(2);
}

/* routine to stop performance, which sched_poll calls if rtfifo_tickin fails */
cp_canit()
{
    cp_quit = 1;
}

#ifdef MIDI5_3

#include <dmedia/midi.h>
#include "sys/select.h"

#define NSYSEX 1024
typedef struct mparse {
    MDport m_inport;
    MDport m_outport;
    int m_need;                 /* this is all for output... */
    int m_got;
    int m_buf[3];
    int m_nsysex;
    char m_sysexbuf[NSYSEX];
} t_mparse;

static int wasopened;
static t_mparse parser1, parser2;

openmidi()
{
    int nports, x;
    wasopened = 1;
    parser1.m_inport = parser1.m_outport = NULL;
    parser2.m_inport = parser2.m_outport = NULL;
    nports = mdInit();
    if (nports <= 0)
    {
        post("no serial ports are configured for MIDI...");
        post("if you want to use MIDI, try exiting Max, typing");
        post("'startmidi -d /dev/ttyd2' to a shell, and restarting Max");
    }
    else if (nports == 1)
        post("Found one MIDI port on %s", mdGetName(0));
    else
        post("Found MIDI ports on %s and %s", mdGetName(0), mdGetName(1));
    if (nports >= 1)
    {
        parser1.m_inport = mdOpenInPort(0);
        if (parser1.m_inport == NULL)
            error("couldn't open MIDI input port %s", mdGetName(0));
        else
            mdSetStampMode(parser1.m_inport, MD_NOSTAMP);
        parser1.m_outport = mdOpenOutPort(0);
        if (parser1.m_outport == NULL)
            error("couldn't open MIDI output port %s", mdGetName(0));
        else
            mdSetStampMode(parser1.m_outport, MD_NOSTAMP);
    }
    if (nports >= 2)
    {
        parser2.m_inport = mdOpenInPort(mdGetName(1));
        if (parser2.m_inport == NULL)
            error("couldn't open MIDI input port %s", mdGetName(1));
        else
            mdSetStampMode(parser2.m_inport, MD_NOSTAMP);
        parser2.m_outport = mdOpenOutPort(mdGetName(1));
        if (parser2.m_outport == NULL)
            error("couldn't open MIDI output port %s", mdGetName(1));
        else
            mdSetStampMode(parser2.m_outport, MD_NOSTAMP);
    }
}

putmidimess(MDport midiport, int a, int b, int c)
{
    MDevent mdv;
    mdv.msg[0] = a;
    mdv.msg[1] = b;
    mdv.msg[2] = c;
    mdv.msg[3] = 0;
    mdv.sysexmsg = 0;
    mdv.stamp = 0;
    mdv.msglen = 0;
    if (mdSend(midiport, &mdv, 1) < 0)
        hate("failure sending message\n");
}

putsysex(MDport midiport, char *s, int n)
{
    MDevent mdv;
    mdv.msg[0] = 0;
    mdv.msg[1] = 0;
    mdv.msg[2] = 0;
    mdv.msg[3] = 0;
    mdv.sysexmsg = s;
    mdv.stamp = 0;
    mdv.msglen = n;
    if (mdSend(midiport, &mdv, 1) < 0)
        hate("failure sending message\n");
}

void io_putmidi(t_mparse *m, int n)
{
    if (!m->m_outport)
        return;
    if (n == 0xf0)
    {
        m->m_nsysex = 1;
        m->m_sysexbuf[0] = 0xf0;
        return;
    }
    else if (m->m_nsysex)
    {
        if (m->m_nsysex == NSYSEX)
            m->m_nsysex = NSYSEX - 1;
        m->m_sysexbuf[m->m_nsysex] = n;
        if (n == 0xf7)
        {
            putsysex(m->m_outport, m->m_sysexbuf, m->m_nsysex);
            m->m_nsysex = 0;
        }
        return;
    }
    if (n & 0x80)
    {
        m->m_got = 0;
        switch (n & 0xf0)
        {
        case MD_NOTEOFF:
        case MD_NOTEON:
        case MD_POLYKEYPRESSURE:
        case MD_CONTROLCHANGE:
        case MD_PITCHBENDCHANGE:
            m->m_need = 2;
            break;
        case MD_PROGRAMCHANGE:
        case MD_CHANNELPRESSURE:
            m->m_need = 1;
            break;
        default:
            m->m_need = 0;
            break;
        }
        m->m_buf[0] = n;
        return;
    }
    if (!m->m_need)
        return;
    else if (m->m_need == 1)
    {
        putmidimess(m->m_outport, m->m_buf[0], n, 0);
        m->m_need = m->m_got = 0;
    }
    else if (m->m_need == 2)
    {
        if (m->m_got == 0)
        {
            m->m_buf[1] = n;
            m->m_got = 1;
        }
        else if (m->m_got == 1)
        {
            putmidimess(m->m_outport, m->m_buf[0], m->m_buf[1], n);
            m->m_need = m->m_got = 0;
        }
    }
    else
        bug("putmidi");
}

#if 0
steal code from here: {
    int nports, x;
    MDevent mdv;
    fd_set inports, outports;
    int nfds, highfd;

     FD_SET(mdGetFd(inport), &inports);
     FD_SET(mdGetFd(outport), &outports);
     highfd = mdGetFd(outport) + 1;

    while (1) {
        nfds = select(highfd, &inports, 0, 0, 0);
        for (x = 0; x < nports; x++) {
            if (FD_ISSET(mdGetFd(inport), &inports))
            {
                if (mdReceive(inport, &mdv, 1) < 0)
                {
                    hate("failure receiving message\n");
                } if (mdv.msg[0] == MD_SYSEX)
                    mdFree(mdv.sysexmsg);
            }
            else
            {
                FD_SET(mdGetFd(inport), &inports);
            }
        }
    }
}
#endif                          /* 0 */

int io_peekmidi(t_mparse *m)
{
    int ret, status, b1, b2, nfds;
    MDevent mdv;
    fd_set inports;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    if (!m->m_inport)
        return (0);

    FD_ZERO(&inports);
    FD_SET(mdGetFd(parser1.m_inport), &inports);

    if (select(mdGetFd(m->m_inport) + 1, &inports, 0, 0, &timeout) < 0)
        perror("midi select");
    if (FD_ISSET(mdGetFd(m->m_inport), &inports))
    {
        if (mdReceive(m->m_inport, &mdv, 1) < 0)
        {
            printf("failure receiving message\n");
            m->m_inport = 0;
            return (0);
        }
        status = 0xff & mdv.msg[0];
        if (status == MD_SYSEX)
        {
            int i;
            for (i = 0; i < mdv.msglen; i++)
                sched_domidi(0xff & mdv.sysexmsg[i]);
            mdFree(mdv.sysexmsg);
            return (1);
        }
        switch (status & 0xf0)
        {
        case MD_NOTEOFF:
        case MD_NOTEON:
        case MD_POLYKEYPRESSURE:
        case MD_CONTROLCHANGE:
        case MD_PITCHBENDCHANGE:
            sched_domidi(status);
            sched_domidi(0xff & mdv.msg[1]);
            sched_domidi(0xff & mdv.msg[2]);
            break;
        case MD_PROGRAMCHANGE:
        case MD_CHANNELPRESSURE:
            sched_domidi(status);
            sched_domidi(0xff & mdv.msg[1]);
            break;
        }
        return (1);
    }
    return (0);
}

#endif                          /* MIDI5_3 */

midi_poll()
{
    if (!wasopened)
        openmidi();
    while (io_peekmidi(&parser1));
}

Xmit(long port, long val)
{
    io_putmidi((port ? &parser2 : &parser1), (int) val);
}

void cp_doprint(long louder)
{
    int i, sum;

    for (i = sum = 0; i < NSPINCOUNT; i++)
        sum += cp_spincount[i];
    sum -= cp_spincount[cp_spinphase];
    sum = sum * .99;
    post((u_chain ? "idle %d.%d%%; %d signals used" :
          "idle %d.%d%%"), sum / 10, sum % 10, sig_getcount());
    if (louder)
    {
        post("ticks:");
        for (i = 0; i < NSPINCOUNT; i++)
            postlong(cp_spincount[(cp_spinphase - i) & (NSPINCOUNT - 1)]);
    }
}

#define HBUFSIZE 4096
char hostbuf[HBUFSIZE];
int hosthead, hosttail;

int sys_roomtowrite()
{
    return ((HBUFSIZE - 1) - ((hosthead - hosttail) & (HBUFSIZE - 1)));
}

int sys_asyncwrite(char *buf, int len)
{
    int room = sys_roomtowrite();
    if (room >= len)
    {
        while (len--)
            hostbuf[hosthead] = *buf++, hosthead =
                (hosthead + 1) & (HBUFSIZE - 1);
        return (1);
    }
    else
        return (0);
}

int sys_trytoflush()
{
    int nwritten, length;
    if (hosthead == hosttail)
        return (0);
    length = (hosthead > hosttail ?
              hosthead - hosttail : HBUFSIZE - hosttail);
    nwritten = write(outfile, hostbuf + hosttail, length);
    if (nwritten < 0 && errno != EAGAIN)
    {
        perror("write");
        return (-1);
    }
    if (nwritten > 0)
        hosttail = (hosttail + nwritten) & (HBUFSIZE - 1);
    return (0);
}

int sys_roomtowrite_animal()
{
    return 0x7fffffff;
}

cp_tick()
{
}

extern void *u_chain;

cp_print(x, start)
void *x;
long start;
{
    static struct timeval startrealtime;
    static long startlogicaltime;
    static struct tms startcputime;
    struct timezone tz;
    if (start)
    {
        gettimeofday(&startrealtime, &tz);
        times(&startcputime);
        startlogicaltime = logicaltime;
        nalarm = 0;
    }
    else
    {
        struct timeval newrealtime;
        long newlogicaltime;
        struct tms newcputime;
        long elapsedlogical, elapsedcpu;
        gettimeofday(&newrealtime, &tz);
        times(&newcputime);
        elapsedlogical = logicaltime - startlogicaltime;
        elapsedcpu = 1000 * (newcputime.tms_utime + newcputime.tms_stime -
                             startcputime.tms_utime -
                             startcputime.tms_stime) / HZ;
        post(u_chain ? "------ DSP ON ------" : "------ DSP OFF ------");
        post("elapsed logical time = %ld", elapsedlogical);
        post("elapsed CPU time =     %ld", elapsedcpu);
        post("elapsed real time =    %ld",
             1000 * (newrealtime.tv_sec - startrealtime.tv_sec) +
             (newrealtime.tv_usec - startrealtime.tv_usec) / 1000);
        post("percentage %6.1f", 100. * elapsedcpu / elapsedlogical);
    }
}
