#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1994/01/04 17:54:10 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/04 17:54:10 $
 *
 * FTS by Miller Puckette
 *
 */


/* LATER need to fix animal to do host updates from other than the
   master CP (exthq_doit is wired to send to MAX) */

#include "m_mess.h"
#include "cp_mess.h"

extern int cpmess_debug;
extern int outfile;

/* the number of hq messages the master keeps for each other CP: */
#define HQ_NBUF 5

/* element of a host queue (to send updates to host) */
typedef struct hq {
    long h_fn;
    long h_freefn;
    long h_ptr;
    void (*h_howto)();
    void *h_client;
    int h_size;
    struct hq *h_next;
    char h_zombie;
    char h_set;
#ifdef CPOS
    ClientConnection h_port;
#else
    char h_animal;
#endif
} t_hq;

typedef struct exthq            /* host queue element maintained by master CP to */
{                               /* double for a remote one */
    t_hq *e_hq;                 /* the host queue element */
    struct exthq *e_next;       /* next in free list */
    char e_buf[HQ_MAX];         /* buffer */
} t_exthq;

t_hq *hq_head, *hq_tail;        /* first and last hq in "set" list */
int hq_nremote;                 /* how many HQs the master is buffering for us */
t_exthq *exthq_freelist;        /* freelist for external hqs -- master only */

#ifdef I860
cvt_long(void *vp)
{
    char *lp = (char *) vp;
    char temp;
    temp = lp[0];
    lp[0] = lp[3];
    lp[3] = temp;
    temp = lp[1];
    lp[1] = lp[2];
    lp[2] = temp;
}
#else
cvt_long()
{
}
#endif

#ifdef I860
cvt_float(void *vp)
{
    char temp;
    char *lp = (char *) vp;
    temp = lp[0];
    lp[0] = lp[3];
    lp[3] = temp;
    temp = lp[1];
    lp[1] = lp[2];
    lp[2] = temp;
}
#else
cvt_float(void *lp)
{
}
#endif

/* set up an hq on behalf of another CP (done on master CP only.) */

/* @@@
   NOTE: THIS Function assume that there is only one ClientConnection
   and use it as h_port; this waiting for the remote port handling,
   coming soon :->
   */

void exthq_setup(t_qmbuf *b, int cpno)
{
    t_exthq *e = exthq_freelist;
    t_hq *hq;
    if (!exthq_freelist)
    {
        bug("out of extern HQs");
        return;
    }
    exthq_freelist = e->e_next;
    hq = e->e_hq;
    hq->h_size = b->b_m.m_size - sizeof(struct qmess);
    hq->h_fn = (long) b->b_q.qm_fn;
    hq->h_ptr = (long) (b->b_q.qm_ptr);

#ifdef CPOS
    /* HACK BEGIN !! */
    hq->h_port = (ClientConnection) 0;
    /* HACK END!! */
#endif

    /* hide the requesting CP in the hp's (unused) freefn slot */
    hq->h_freefn = cpno;
    bcopy(b->b_s, e->e_buf, hq->h_size);
    hq_set(hq);
}

/* this is called when the exthq's "hq" is ready to be sent up.  A message
   is also sent back to the originating CP (which was stuffed into the h_freefun
   slot when the exthq was set up with the message). */

void exthq_doit(t_exthq *x, t_hq *hq)
{
#ifdef CPOS
    sendtohost(hq->h_fn, hq->h_ptr, x->e_buf, hq->h_size, hq->h_port);
#else
    sendtohost(hq->h_fn, hq->h_ptr, x->e_buf, hq->h_size, 0);
#endif

    rtfifo_thanks((int) (hq->h_freefn));
    x->e_next = exthq_freelist;
    exthq_freelist = x;
}

void sendtohost(long int fn, long int ptr, char *buf, int size, int port)
{
    t_qmbuf b;
    char *s, *t;
    int n;
    int len = size + (sizeof(struct qmess) + sizeof(struct messhdr));
    b.b_m.m_command = H_QUEUE;
    b.b_m.m_size = size + sizeof(struct qmess);
    b.b_q.qm_fn = (t_method) fn;
    b.b_q.qm_ptr = (void *) ptr;
    for (s = b.b_s, t = buf, n = size; n--;)
        *s++ = *t++;
    if (cp_this)
    {
        rtfifo_sendtohost(&b, len);     /* should take port as argument ? */
        hq_nremote++;
    }
    else
    {
        cvt_long(&b.b_q.qm_fn);
        cvt_long(&b.b_q.qm_ptr);
        cvt_long(&b.b_m.m_command);
        cvt_long(&b.b_m.m_size);
#ifdef I860
#ifdef CPOS
        while (!sys_asyncwrite(&b, len, port));
#else
        if (port)
            while (!sys_asyncwrite_animal(&b, len));
        else
            while (!sys_asyncwrite(&b, len));
#endif
#else
#ifdef __unix
        while (!sys_asyncwrite(&b, len));
#else
        if (!cpmess_debug)
        {
            if (write(outfile, &b, len) < len)
            {
                perror("write");
                hate("host connection broken");
            }
        }
#endif                          /* __unix */
#endif                          /* I860 */
    }
}

/* if we're the master, we can send anything there's room for; if another
   CP, can only have HQ_NBUF host updates outstanding. */

static int hq_cansend(t_hq *x)
{
#ifdef CPOS
    if (cp_this)
        return (hq_nremote < HQ_NBUF);
    else
        return (sizeof(t_qmbuf) + x->h_size - HQ_MAX <=
                sys_roomtowrite(x->h_port));
#else
    if (cp_this)
        return (hq_nremote < HQ_NBUF);
    else if (x->h_animal)
        return (sizeof(t_qmbuf) + x->h_size - HQ_MAX <=
                sys_roomtowrite_animal());
    else
        return (sizeof(t_qmbuf) + x->h_size - HQ_MAX <= sys_roomtowrite());
#endif
}

hq_thanks()
{
    hq_nremote--;
}

void hq_putlong(long int *ip, long int *bp)
{
    *bp = *ip;
    cvt_long(bp);
}

void hq_putfloat(float *fp, float *bp)
{
    *bp = *fp;
    cvt_float(bp);
}

void hq_set(t_hq *x)
{
    if (!x->h_ptr)
        return;
    if (x->h_set)
        return;
    if (hq_tail)
        hq_tail->h_next = x, hq_tail = x;
    else
        hq_head = hq_tail = x;
    x->h_next = 0;
    x->h_set = 1;
}

void hq_resize(t_hq *x, int size)
{
    if (size > HQ_MAX)
        bug("maximum host queue element size");
    x->h_size = size;
}

void hq_connect(t_hq *x, long int fn, long int freefn, long int ptr)
{
    if (x->h_ptr)
        error("hq_connect: already connected");
    x->h_fn = fn;
    x->h_freefn = freefn;
    x->h_ptr = ptr;
    /* hq_set(x); */
}

/* an object can call this to generate a standard host update; this is
   called from a "h_howto" funtion, which the hq calls when it's ready to
   be sent.  The exthq's "howto" function does something special. */

hq_doit(t_hq *x, char *buf, int size)
{
#ifdef CPOS
    sendtohost(x->h_fn, x->h_ptr, buf, size, x->h_port);
#else
    sendtohost(x->h_fn, x->h_ptr, buf, size, (int) (x->h_animal));
#endif
}

/* try to send an hq upstairs; return 1 if anything was sent.
   first check if the hq can be sent and if so, take it off the queue
   and call the client's h_howto
   function to send it.  If it's a zombie (the owner has been deleted)
   we tell the host it's gone and then delete it. */

hq_poll()
{
    t_hq *x = hq_head;
    if (!x)
        return (0);
    if (!x->h_ptr || hq_cansend(x))
    {
        hq_head = x->h_next;
        if (hq_tail == x)
            hq_tail = 0;
        if (x->h_zombie)
        {
            if (x->h_freefn)
#ifdef CPOS
                sendtohost(x->h_freefn, x->h_ptr, NIL, 0, x->h_port);
#else
                sendtohost(x->h_freefn, x->h_ptr, NIL, 0,
                           (int) (x->h_animal));
#endif
            freebytes(x, sizeof(t_hq));
        }
        else
        {
            x->h_set = 0;
            if (x->h_ptr)
                (*x->h_howto) (x->h_client, x);
        }
        return (1);
    }
    else
        return (0);
}

/* call this first, with cp_this and cp_nproc defined */
hq_init()
{
    if (!cp_this)
    {
        int i, n = (cp_nproc - 1) * HQ_NBUF;

        for (i = n; i--;)
        {
            t_exthq *e = (t_exthq *) getbytes(sizeof(t_exthq));

            e->e_hq = hq_new(0L, 0L, 0L, e, exthq_doit, 0,
                             (ClientConnection) 0);
            e->e_next = exthq_freelist;
            exthq_freelist = e;
        }
    }
}

void hq_free(t_hq *x)
{
    x->h_zombie = 1;
    hq_set(x);
}

t_hq *hq_new(long int fn, long int freefn, long int ptr, void *client,
             void (*howto)(), int size, ClientConnection port)
{
    t_hq *x =(t_hq *) getbytes(sizeof(t_hq));
    if (size > HQ_MAX)
        bug("maximum host queue element size");
    x->h_fn = fn;
    x->h_freefn = freefn;
    x->h_ptr = ptr;
    x->h_client = client;
    x->h_howto = howto;
    x->h_size = size;
    x->h_zombie = x->h_set = 0;
#ifdef CPOS
    x->h_port = port;
#else
    x->h_animal = (port != 0);
#endif
    return (x);
}

cp_ping(void *x, long int n)
{
    post("call %x on host", n);
#ifdef CPOS
    hq_free(hq_new(0, n, 1, 0, 0, 0, currentPort));
#else
    hq_free(hq_new(0, n, 1, 0, 0, 0, (ClientConnection) 1));
#endif
}
