/*
 *      Copyright (c) 2000 IRCAM
 *      All rights reserved.
 *
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>

#include "m_mess.h"
#include "cp_mess.h"
#include "d_ugen.h"
#include "s_linux.h"

#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/times.h>
#include <sys/param.h>

/*
 * Globals
 */

char *progname = "fts";
int infile;                     /* input to Max */
int outfile;                    /* output from Max */

int cp_quit = 0;                /* quit flag; can be set from sched_tick */
int cp_this = 0;                /* ID number of this CP */
int cp_nproc = 1;               /* number of CPs */
int cp_usedacs;
int cp_slotno;
int cp_sr = 44100;              /* integer sampling rate use in dac_timefwd */
int cp_inchans;
int cp_outchans;
int cp_use_audiofile;
static int cp_audio_open;       /* true if audio is open */
static int cp_advance;          /* scheduler advance */
static int sgi_realtime = 0;
static int cp_up2 = 0;
#define MAXCH 8

extern float dsp_sr;
#define NSPINCOUNT 16
int cp_spincount[NSPINCOUNT], cp_spinphase, cp_nextspin;
static long dactime = 1;
extern int dsp_ticks;

static long ticktime = 0;
extern long logicaltime;

extern float soundout[], soundin[];

#define DACVS 64                /* vector size for a/d/a and inter-CP signal paths */
#define NCHAN 8                 /* maximum number of DACs and ADCs */
#define XFERSIZE 64             /* number of samples we pass to dacs */
#define DACALIGN 16

int cp_inxferframes;
int cp_outxferframes;
int sys_verbose;

/* ----------------------I/o and timing routines -------------- */

void sys_setalarm()
{
}

double sys_getrealtime(void)    /* get "real time" in seconds */
{
    static struct timeval then;
    struct timeval now;
    gettimeofday(&now, 0);
    if (then.tv_sec == 0)
        then = now;
    return ((now.tv_sec - then.tv_sec) +
            (1. / 1000000.) * (now.tv_usec - then.tv_usec));
}

void sys_log_error(int type)
{
}

void sys_microsleep(int microsec)
{
    fd_set readset, writeset, exceptset;
    struct timeval timout;
    timout.tv_sec = 0;
    timout.tv_usec = microsec;
    FD_ZERO(&writeset);
    FD_ZERO(&readset);
    FD_ZERO(&exceptset);
    select(0, &readset, &writeset, &exceptset, &timout);
}

cp_exit(n)
{
    exit(n);
}

int sgi_open_audio()
{
#if 1
    sgi_setlogicalsamplerate(44100);
    sgi_setlogicalchans(2, 2);
    linux_frags(8);
    fprintf(stderr, "open in chans %d out chans %d; using fragments\n",
            cp_inchans, cp_outchans);
    sys_open_audio_and_midi(1, 1, cp_inchans, cp_outchans, (int) dsp_sr);
    fprintf(stderr, "open 2\n");
    return (0);
#else
    return (-1);
#endif
}

void sgi_set_chans(int in, int out)
{
    sgi_setlogicalchans(in, out);
    cu_zero(soundin, 8 * MAXVS);
}

void sgi_close_audio()
{
    sys_close_audio_and_midi();
}

void sgi_setup2(void *x, long n)
{
    cp_up2 = n;
}

void sgi_senddacs(float *inbuffer, float *outbuffer)
{
#if 0
    struct timeval tv1, tv2;
    int i, nwait = 0, usec;
    int samps = cp_outchans * DACVS;
    for (i = 0; i < samps; i++)
        sys_soundout[i] = outbuffer[i];
    gettimeofday(&tv1, 0);
    while (sys_send_dacs() == SENDDACS_NO)
        sys_microsleep(1500);
#ifdef FIDDLE
    gettimeofday(&tv2, 0);
    usec = tv2.tv_usec - tv1.tv_usec + 1000000 * (tv2.tv_sec - tv1.tv_sec);
    if (usec < 0 || usec > 2000)
        fprintf(stderr, "xx usec %d\n", usec);
#endif
    samps = cp_inchans * DACVS;
    for (i = 0; i < samps; i++)
        inbuffer[i] = sys_soundin[i];
#else
    sys_microsleep(1450);
#endif
}

static struct timeval fake_starttime;
static long fake_logicaltime;

void fake_openconnection()
{
    gettimeofday(&fake_starttime, 0);
    cp_use_audiofile = 0;
    fake_logicaltime = 0;
}

int fake_dacshungry()
{
    struct timeval timenow;
    long elapsedtime;
    gettimeofday(&timenow, 0);
    elapsedtime = 1000 * (timenow.tv_sec - fake_starttime.tv_sec) +
        (timenow.tv_usec - fake_starttime.tv_usec) / 1000;
    return (elapsedtime > fake_logicaltime);
}

void fake_senddacs()
{
    fake_logicaltime = logicaltime;
}

int sgi_gotrealtime()
{
    return (sgi_realtime);
}

static void fixpri(void)
{
    if (geteuid())
    {
        fprintf(stderr, "fts: warning: can't get real-time status\n");
        return;
    }
    sgi_realtime = 1;
    /* bash priority to a high value.  */
    if (1)
        fprintf(stderr, "fts: warning: couldn't set priority\n"),
            sgi_realtime = 0;
    seteuid(getuid());
}

void floatfuck(int x)
{
    fprintf(stderr, "float exception\n");
}

main(int argc, char **argv)
{
    int realtimeplease;
    /* Max (the parent process) has duplicated stdin and stdout for a connection.
     * Alias the names since much code uses the names `infile' and `outfile'.
     */

    infile = STDIN_FILENO;
    outfile = STDOUT_FILENO;

    /* Trap Unix signals */
    signal(SIGINT, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGFPE, floatfuck);

    h_init(1);

    h_canpost();
    if (fcntl(outfile, F_SETFL, O_NONBLOCK) == -1)
    {
        perror("fcntl");
        hate("can't fix output port");
    }

    if (argc < 2 || (sscanf(argv[1], "%d", &cp_advance) < 1))
        cp_advance = 8;
    if (cp_advance < 3 || cp_advance > 50)
    {
        if (cp_advance)
            post("advance out of range (3-50); reset to 8");
        cp_advance = 8;
    }
    realtimeplease = (argc < 3 || *(argv[2]) == '0');

    if (realtimeplease)
        fixpri();
    else
        post("cancelling real-time priority setup...");


    if (!sgi_open_audio())
    {
        cp_audio_open = 1;
    }
    else
    {
        fake_openconnection();
        sgi_setlogicalchans(1, 1);
        sgi_setlogicalsamplerate(8000);
    }

    cp_print(1);                /* start profiling */

    while (!cp_quit)
    {
        int foo = 0;
        cp_spinphase = (cp_spinphase + 1) & (NSPINCOUNT - 1);
        cp_spincount[cp_spinphase] = cp_nextspin;
        cp_nextspin = 0;
        sched_poll();           /* increment logicaltime */
#ifdef MIDI
        midi_poll();            /* keep up with MIDI in */
#endif
        dac_poll();
        host_poll();
        if (!foo--)
        {
            static int nexttime;
            hq_poll();
            if (sys_trytoflush())
                cp_quit = 1;
            if (logicaltime >= nexttime)
            {
                int foo = nexttime & 4;
                if (!foo)
                    max_sendtimeup();
                nexttime += 500;
            }
            foo = 7;
        }
    }
    fprintf(stderr, "fts: stopping\n");
    h_done();
    sgi_close_audio();
    exit(0);
}

long realtime()
{
    return dactime;
}

host_poll()
{
    char cmd[BUFSIZ];
    int ret, i, xxx;
    fd_set check;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = !cp_audio_open;

    FD_ZERO(&check);
    FD_SET(infile, &check);
    select(64, &check, (fd_set *) 0, (fd_set *) 0, &timeout);
    if (FD_ISSET(infile, &check))
    {
        if ((ret = read(infile, cmd, BUFSIZ)) < 0)
        {
            perror("read");
            hate("host connection broken");
        }
        else if (ret == 0)
        {
            cp_quit = 1;
            return;             /* EOF */
        }
        /* fprintf(stderr, "poll %d\n", ret); */
        /* read ok */
        for (i = 0; i < ret; i++)
            cp_dochar(0, cmd[i]);
    }
}

  /* tell DSP network etc about new sample rate */
sgi_setlogicalsamplerate(int n)
{
    if (n <= 0)
        n = 1;
    dsp_sr = cp_sr = n;
    post("setting logical sample rate to %ld", n);
}

    /* tell everyone what channel count we configured */
sgi_setlogicalchans(int in, int out)
{
    if (in != 2 && in != 4 && in != 8)
        in = 2;
    if (out != 2 && out != 4 && out != 8)
        out = 2;
    cp_usedacs = 1;

    cp_inxferframes = XFERSIZE / in;
    cp_inchans = in;
    cu_zero(soundin, cp_inchans * MAXVS);
    cp_outxferframes = XFERSIZE / out;
    cp_outchans = out;
    cu_zero(soundout, cp_outchans * MAXVS);
}

dac_timefwd()
{
    static int dactime_frac;

    dactime_frac += 64000;
    while (dactime_frac >= cp_sr)
    {
        dactime++;
        dactime_frac -= cp_sr;
    }
}

static int cp_insomnia;         /* number of ticks since last sleep */

  /* NB: the use of `short' here is based on their being 16 bits */
dac_poll()
{
    sgi_senddacs(soundin, soundout);
    cu_zero(soundout, cp_outchans * MAXVS);
    dac_timefwd();
}

/* routine to stop performance, which sched_poll calls if rtfifo_tickin fails */
cp_canit()
{
    cp_quit = 1;
}

/* ---------------------- no MIDI ----------------------- */

openmidi()
{
}

Xmit(long port, long val)
{
    static int warned;
    if (!warned)
    {
        fprintf(stderr, "WARNING: midi output isn't implemented\n");
        warned = 1;
    }
}

void inmidi_noteon(int portno, int channel, int pitch, int velo)
{
}

void inmidi_controlchange(int portno, int channel, int ctlnumber,
                          int value)
{
}

void inmidi_programchange(int portno, int channel, int value)
{
}

void inmidi_pitchbend(int portno, int channel, int value)
{
}

void inmidi_aftertouch(int portno, int channel, int value)
{
}

void inmidi_polyaftertouch(int portno, int channel, int pitch, int value)
{
}

/* --------------------- printout ---------------------------- */

void cp_doprint(long louder)
{
    int i, sum;

    for (i = sum = 0; i < NSPINCOUNT; i++)
        sum += cp_spincount[i];
    sum -= cp_spincount[cp_spinphase];
    sum = sum * .99;
    post((u_chain ? "idle %d.%d%%; %d signals used" :
          "idle %d.%d%%"), sum / 10, sum % 10, sig_getcount());
    if (louder)
    {
        post("ticks:");
        for (i = 0; i < NSPINCOUNT; i++)
            postlong(cp_spincount[(cp_spinphase - i) & (NSPINCOUNT - 1)]);
    }
}

#define HBUFSIZE 4096
char hostbuf[HBUFSIZE];
int hosthead, hosttail;

int sys_roomtowrite()
{
    return ((HBUFSIZE - 1) - ((hosthead - hosttail) & (HBUFSIZE - 1)));
}

int sys_asyncwrite(char *buf, int len)
{
    int room = sys_roomtowrite();
    if (room >= len)
    {
        while (len--)
            hostbuf[hosthead] = *buf++, hosthead =
                (hosthead + 1) & (HBUFSIZE - 1);
        return (1);
    }
    else
        return (0);
}

int sys_trytoflush()
{
    int nwritten, length;
    if (hosthead == hosttail)
        return (0);
    length = (hosthead > hosttail ?
              hosthead - hosttail : HBUFSIZE - hosttail);
    nwritten = write(outfile, hostbuf + hosttail, length);
    if (nwritten < 0 && errno != EAGAIN)
    {
        perror("write");
        return (-1);
    }
    if (nwritten > 0)
        hosttail = (hosttail + nwritten) & (HBUFSIZE - 1);
    return (0);
}

int sys_roomtowrite_animal()
{
    return 0x7fffffff;
}

cp_tick()
{
}

extern void *u_chain;

cp_print(x, start)
void *x;
long start;
{
    static struct timeval startrealtime;
    static long startlogicaltime;
    static struct tms startcputime;
    if (start)
    {
        gettimeofday(&startrealtime, 0);
        times(&startcputime);
        startlogicaltime = logicaltime;
    }
    else
    {
        struct timeval newrealtime;
        long newlogicaltime;
        struct tms newcputime;
        long elapsedlogical, elapsedcpu;
        gettimeofday(&newrealtime, 0);
        times(&newcputime);
        elapsedlogical = logicaltime - startlogicaltime;
        elapsedcpu = 1000 * (newcputime.tms_utime + newcputime.tms_stime -
                             startcputime.tms_utime -
                             startcputime.tms_stime) / HZ;
        post(u_chain ? "------ DSP ON ------" : "------ DSP OFF ------");
        post("elapsed logical time = %ld", elapsedlogical);
        post("elapsed CPU time =     %ld", elapsedcpu);
        post("elapsed real time =    %ld",
             1000 * (newrealtime.tv_sec - startrealtime.tv_sec) +
             (newrealtime.tv_usec - startrealtime.tv_usec) / 1000);
        post("percentage %6.1f", 100. * elapsedcpu / elapsedlogical);
    }
}
