
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/04 17:54:21 $
 *
 * FTS by Miller Puckette
 *
 */

#ifndef _CP_MESS_H
#define _CP_MESS_H

/* commands from host to CP */
#define CP_SPECIAL  0           /* special message between CPs, such as "tick" */
#define CP_MESS   1             /* send a message to a CP object */
#define CP_CONNECT  2           /* make a MAX-style connection between two objects */
#define CP_DISCONNECT   3       /* cut one */
#define CP_FREE   4             /* free a CP object */
#define CP_TICK   5             /* advance logical time (for simulation only) */
#define CP_NEW    6             /* make a new CP object */
/* (note: CP_NEW must be last in list because CP number is added to it.) */

/* inter-CP messages or messages back to host (PRINT, QUEUE only). */
#define H_ENDOFTICK   0         /* end of message transmissions for a given tick */
#define H_PRINT   1             /* printout */
#define H_QUEUE   2             /* a "graphic update" (host queue transmission) */
#define H_THANKS  3             /* confirm that an H_QUEUE has been sent to host */
#define H_FCALL   4             /* call a function remotely */

/* parameters that might fix the size of some structure */
#define MSIZE 4096              /* maximum size of a host->CP message */
#define MARG 256                /* maximum number of args in a host->CP message */
#define HQ_MAX 256              /* max size of a host-queue element */
#define PR_MAX 256              /* max size of a printout */
#define MAXCP 24                /* maximum imaginable number of CPs */

typedef struct messhdr {        /* header for messages between CP and host */
    long m_command;             /* from CP_ or H_ def above depending on direction */
    long m_size;                /* ("size" is used only in CP->host direction!) */
} t_messhdr;

typedef struct messbuf {        /* structure of a message */
    struct messhdr m_h;
    char m_shit[MSIZE];
} t_messbuf;

typedef struct qmess {          /* first part of m_shit for H_QUEUE messages */
    void (*qm_fn)();            /* function to call on host */
    void *qm_ptr;               /* first argument to call it with */
} t_qmess;                      /* ... the rest is pointed to by another argument. */

typedef struct qmbuf {          /* structure of an H_QUEUE message */
    struct messhdr b_m;
    struct qmess b_q;
    char b_s[HQ_MAX];
} t_qmbuf;

/* this could be hidden if we drop CPOS compatibility: */
#define MAX_ARGLEN 512
typedef struct _inmess {
    t_messhdr i_msg;
    char i_buf[MAX_ARGLEN];
    char *i_pt;
    int i_gothdr;
    int i_ac;
    t_atom i_av[MARG];
    t_atom *i_ap;
} t_inmess;

extern int cp_this;             /* ID number of this CP */
extern int cp_nproc;            /* number of CPs */

/* a macro to check validity of an object's address.  The master ALIST
uses the reserved range of addresses, 0 <= x < cp_nproc, for special
purposes. */

#define CP_LOCAL(x) (((long)(x)) < 0 || ((long)(x)) >= cp_nproc)

#define MAXADVANCE 50           /* why? really should just try to alloc a bigger buf */
#define MINADVANCE 2            /* should really be 1 but doesn't work for 1 */

struct alist;
void *alist_get(struct alist *x, long int a);
void alist_put(struct alist *x, long int a, void *b);
void alist_bag(struct alist *x, long int a);
void alist_free(struct alist *x);
struct alist *alist_new(int bits);

typedef void *ClientConnection;
#define currentPort ((ClientConnection)0)

struct hq;
struct hq *hq_new(long int fn, long int freefn, long int ptr,
                  void *client, void (*howto)(), int size,
                  ClientConnection port);
void hq_set(struct hq *x);
void sendtohost(long int fn, long int ptr, char *buf, int size, int port);

#endif
