#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:00 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:00 $
 *
 * The code that builds call lists for MAX unit generator networks.  Great
 * care is taken to keep data compacted and local to get good cache statistics
 * on the I860.
 *
 * Tricky detail: there's an implicit guarantee that you can read MARGIN bytes past 
 * the end of the pages allocated here without causing a trap; most unit
 * generators read past the end of their signal inputs, and the 68000 version of
 * the DSPCHAIN traversor can read off the end of a page if the last funcall on
 * the page had less than the 5 arguments that are always picked up.
 */

#define MARGIN 256              /* number of extra bytes guaranteed */

#include <stdlib.h>
#include <stdarg.h>
/* #include <stdio.h> */
#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__


#ifdef I860
char *malloc();
#endif                          /* I860 */

/* we'll allocate several contiguous pages at a time so that when we
    read off the end of one we hit the next one */
#define PAGE_GET_AT_ONCE  8

extern t_symbol s_sig, s_dspchain, s__sig, s_tick;
long *u_free_pages;             /* pages we have allocated but are not now using */

/* the ucalls are terminated by the value U_EXIT.
    The FIRST word of a ucall block points to the next one or zero
    if none.  This is so that you can traverse the block list without
    traversing all the ucalls.
    */

#define WORDS_TO_TERMINATE 1
#ifdef I860
extern void dspchain_tick();    /* it's in assembler */
extern int dspchain_callc(), dspchain_exit();
#else
long *dspchain_callc(long *);
long *dspchain_exit(long *);
#endif
#define U_CALLC (long)dspchain_callc
#define U_EXIT (long)dspchain_exit

t_dspchain *dsp_current;
float dsp_sr = 44100;

extern int max_debug;
static int lasttick = 0;
int dsp_ticks;

#ifndef I860
long *dspchain_callc(long *where)
{
    t_method f = (t_method) (where[1]);
    long n = where[2];
    (*f) (where[3], where[4], where[5], where[6], where[7]);
    return (where + (n + 3));
}

long *dspchain_exit(long *where)
{
    long *nextpage = *(long **) (((long) where) & (-UCALL_PAGE_SIZE));
    if (!nextpage)
        return (0);
    else
        return (nextpage + 1);
}
#endif                          /* I860 */

#ifdef i860
#else
void dspchain_tick(t_dspchain *x)
{
    int count = 0;
    long *f = x->d_firstpage + 1;
    /* fprintf(stderr,  "tick\n"); */
    while (f)
    {
        /* fprintf(stderr,  "%x: %x %x %d %x %x\n",
           f,  f[0],  f[1], f[2],  f[3],  f[4]); */
        f = (*(t_ucall) (f[0])) (f);
    }
    dsp_ticks++;
}
#endif

/* put a c function call on the list. */
void dspchain_putc(t_dspchain *x, t_method f, int n, ...)
{
    va_list ap;
    long *nextucall = x->d_nextucall;
    va_start(ap, n);
    if (n > 5)
        error("dspchain_putc: sorry, 5 args max"), n = 5;
    if (nextucall + (3 + WORDS_TO_TERMINATE + n) > x->d_ceiling)
    {
        if (!dspchain_newpage(x))
            hate("out of memory building ugen chain");
        nextucall = x->d_nextucall;
    }
    *nextucall++ = U_CALLC;
    *nextucall++ = (long) f;
    *nextucall++ = n;
    if (n >= 1)
        *nextucall++ = va_arg(ap, int);
    if (n >= 2)
        *nextucall++ = va_arg(ap, int);
    if (n >= 3)
        *nextucall++ = va_arg(ap, int);
    if (n >= 4)
        *nextucall++ = va_arg(ap, int);
    if (n >= 5)
        *nextucall++ = va_arg(ap, int);
    x->d_nextucall = nextucall;
    va_end(ap);
}

void dspchain_addc(t_method f, int n, ...)
{
    va_list ap;
    va_start(ap, n);
    dspchain_putc(u_chain, f, n, ap);
    va_end(ap);
}

long dspchain_next(t_dspchain *x)
{
    return ((long) (x->d_nextucall));
}

/* put an assembly-language function on the list, with the following call
    convention: r15 points to a structure which contains
    (function) (arg) (arg) ...;
    the function should not return but should advance r15 to the next call
    (at the end of this one) and jump to (*r15).
    */
#ifdef I860
void dspchain_put860(x, f, n, a1, a2, a3, a4, a5)
t_dspchain *x;
t_method f;
long a1, a2, a3, a4, a5;
{
    long *nextucall = x->d_nextucall;
    if (n > 5)
        error("dspchain_put860: sorry, 5 args max"), n = 5;
    if (nextucall + (1 + WORDS_TO_TERMINATE + n) > x->d_ceiling)
    {
        if (!dspchain_newpage(x))
            hate("out of memory building ugen chain");
        nextucall = x->d_nextucall;
        /* post("switch pages"); */
    }
    *nextucall++ = (long) f;
    if (n >= 1)
        *nextucall++ = a1;
    if (n >= 2)
        *nextucall++ = a2;
    if (n >= 3)
        *nextucall++ = a3;
    if (n >= 4)
        *nextucall++ = a4;
    if (n >= 5)
        *nextucall++ = a5;
    x->d_nextucall = nextucall;
}

void dspchain_add860(f, n, a1, a2, a3, a4, a5)
t_method f;
long a1, a2, a3, a4, a5;
{
    dspchain_put860(u_chain, f, n, a1, a2, a3, a4, a5);
}
#else                           /* otherwise, it's the C version: */
void dspchain_addugen(t_ucall f, int n, ...)
{
    va_list ap;    
    t_dspchain *x = (t_dspchain *) u_chain;
    long *nextucall = x->d_nextucall;
    va_start(ap, n);
    if (n > 5)
        error("dspchain_put860: sorry, 5 args max"), n = 5;
    if (nextucall + (1 + WORDS_TO_TERMINATE + n) > x->d_ceiling)
    {
        if (!dspchain_newpage(x))
            hate("out of memory building ugen chain");
        nextucall = x->d_nextucall;
        /* post("switch pages"); */
    }
    *nextucall++ = (long) f;
    if (n >= 1)
        *nextucall++ = va_arg(ap, int);
    if (n >= 2)
        *nextucall++ = va_arg(ap, int);
    if (n >= 3)
        *nextucall++ = va_arg(ap, int);
    if (n >= 4)
        *nextucall++ = va_arg(ap, int);
    if (n >= 5)
        *nextucall++ = va_arg(ap, int);
    x->d_nextucall = nextucall;
    va_end(ap);
}
#endif
/* put a ugen's "state space" on the ucall list.  Alignment must be a power
    of two.
    */

void *dspchain_putstate(t_dspchain *x, void *shit, int size, int alignment)
{
    char *wouldbe = ((char *) x->d_ceiling) - size;
    if (size > UCALL_PAGE_SIZE - (2 + WORDS_TO_TERMINATE) * sizeof(long))
    {
        error("dspchain_putstate: buffer too large");
        return (shit);
    }
    wouldbe -= (((long) wouldbe) & (alignment - 1));
    if ((char *) (x->d_nextucall + WORDS_TO_TERMINATE) > wouldbe)
    {
        if (!dspchain_newpage(x))
            hate("out of memory building ugen chain");
        wouldbe = ((char *) x->d_ceiling) - size;
        wouldbe -= (((long) wouldbe) & (alignment - 1));
        /* post("switch pages"); */
    }
    x->d_ceiling = (long *) wouldbe;
    bcopy(shit, wouldbe, size);
    return (wouldbe);
}

void *dspchain_addstate(void *shit, int size, int alignment)
{
    return (dspchain_putstate(u_chain, shit, size, alignment));
}

/* allocate a page of memory; guarantee that you can read at least MARGIN
    bytes past the end of the page without trapping. */
void *dspchain_getpage()
{
    long *lp, *elp;
    if (!u_free_pages)          /* if free list is empty renew it */
    {
        lp = (long *) malloc(PAGE_GET_AT_ONCE * UCALL_PAGE_SIZE);
        if (!lp)
            return (0);
        elp = lp + PAGE_GET_AT_ONCE * UCALL_PAGE_SIZE / sizeof(long);
        /* round lp up to page boundary */
        lp += ((-(long) lp) & (UCALL_PAGE_SIZE - 1)) / sizeof(long);
        elp -= MARGIN / sizeof(long);
        while (lp + UCALL_PAGE_SIZE / sizeof(long) <= elp)
        {
            *lp = (long) u_free_pages;
            u_free_pages = lp;
            lp += UCALL_PAGE_SIZE / sizeof(long);
        }
    }
    /* pull a page off free list */
    lp = u_free_pages;
    u_free_pages = (long *) (*lp);
    return ((void *) lp);
}

/* deallocate same */
dspchain_ungetpage(long int *lp)
{
    *lp = (long) u_free_pages;
    u_free_pages = lp;
}

dspchain_newpage(t_dspchain *x)
{
    long *newpage = (long *) (dspchain_getpage());
    if (!newpage)
        return (0);
    /* if there's already a page there, make it point to the new one */
    if (x->d_firstpage)
    {
        *(x->d_nextucall) = U_EXIT;
        *(long **) (((long) (x->d_nextucall)) & (-UCALL_PAGE_SIZE)) =
            newpage;
    }
    /* otherwise it's the first one */
    else
        x->d_firstpage = newpage;
    x->d_nextucall = newpage + 1;
    x->d_ceiling = (long *) ((char *) newpage + UCALL_PAGE_SIZE);
    return (1);
}

/* free a dspchain's call list */
dspchain_freecalls(t_dspchain *x)
{
    long *page = x->d_firstpage, *nextpage;
    while (page)
    {
        nextpage = (long *) (*page);
        dspchain_ungetpage(page);
        page = nextpage;
    }
    x->d_firstpage = 0;
}

/* blow away the call list and set up a new, empty one. */
dspchain_begin(t_dspchain *x)
{
    dspchain_freecalls(x);      /* clear current call list */
    if (!dspchain_newpage(x))   /* allocate first page */
        hate("out of memory building ugen chain");
}

/* terminate new call list */
dspchain_end(t_dspchain *x)
{
    *(x->d_nextucall) = U_EXIT;
    *(long **) (((long) (x->d_nextucall)) & (-UCALL_PAGE_SIZE)) = 0;
}

void dspchain_freefun(t_dspchain *x)
{
    dspchain_freecalls(x);      /* clear current call list */
}

void c_dspchain()
{
    c_new(&s_dspchain, dspchain_freefun, sizeof(t_dspchain), 1, 0);
    c_addmess(dspchain_tick, &s_tick, A_DEFLONG, 0);
}

t_maxclass *dspchain_messlist;
void *dspchain_new()
{
    t_dspchain *x =
        (t_dspchain *) obj_new((t_messlist **) & dspchain_messlist,
                               c_dspchain);
    dsp_current = x;
    x->d_firstpage = 0;
    return (x);
}

dspchain_free(t_dspchain *x)
{
    freeobject((t_object *) x);
}
