#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:09 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:09 $
 *
 * FTS by Miller Puckette
 * delay write, delay read, interpolating delay read
 *
 */


#include <stdio.h>
#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#ifdef I860
#include "../6a/6a.h"
#include "../6a/proto56.h"
#endif
#define DELAY_ALIGN 16          /* alignment needed, in bytes, for delay line */
#include <math.h>
#include <stdio.h>              /* get rid of this */


char *uncachedMalloc(unsigned int n);

extern struct symbol s_put, s_set,
    s__sigdelread, s__sigdelwrite, s__sigvd, s_ft1;

#ifdef I860
void su_delwrite(), su_delread(), su_vd();
#endif

typedef struct delbuf {
    float *b_shit;
    float b_conv;
    long b_phase;
    long b_nsamps;
} t_delbuf;

#define DELCHUNK 4

void cu_delwrite(float *in, t_delbuf * b, long int n);
void cu_delread(float *out, t_delbuf * b, long int n, float *deltime);

/******************** DELWRITE ***********************/
typedef struct delwrite {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_symbol *x_sym;
    struct delwrite *x_next;    /* list of delwrites hanging from symbol */
    t_delbuf x_buf;
    char *x_malloc;
} t_delwrite;
extern void *delwrite_class;

/*
  Consistency requirements for delwrite, delread to work: the logical length
  of the buffer (b->b_nsamps) and the phase (b->b_phase) must be multiples of the
  vector size.  The actual delay buffer has vecsize elements at the end which are
  copies of the first vecsize elements.  Thus you can read vecsize consecutive
  samples starting anywhere as long as they don't cross the write pointer.  The
  phase can't be zero but can equal the logical length. */

#ifndef I860
void cu_delwrite(float *in, t_delbuf *b, long int n)
{
    float *bp = b->b_shit + b->b_phase;
    int i;
    for (i = 0; i < n; i++, bp++, in++)
    {
        float f = *in;
        if (BADFLOAT(f))
            *bp = 0;
        else
            *bp = f;
    }
    if (b->b_phase == b->b_nsamps)
    {
        bcopy(b->b_shit + b->b_phase, b->b_shit, n * sizeof(float));
        b->b_phase = n;
    }
    else
        b->b_phase += n;
}
#endif                          /* I860 */

void delwrite_put(t_delwrite *x, long int whether)
{
    if (whether)
    {
        if (!x->x_malloc)
        {
            error("delwrite: no memory -- aborting DSP");
            u_abort();
            return;
        }
        u_stdout((t_ugen *) x);
        if (!x->x_sym->s_thing)
            warn("delwrite: %s: something else had the same name",
                 x->x_sym->s_name);
        else if (x->x_sym->s_thing != (t_object *) x)
            warn("delwrite: %s: multiply defined", x->x_sym->s_name);
        x->x_buf.b_conv = dsp_sr * .001;
#ifdef I860
        dspchain_put860(u_chain, su_delwrite, 3,
                        x->x_io[0]->s_shit, &x->x_buf, x->x_io[0]->s_n);
#else
        dspchain_putc(u_chain, cu_delwrite, 3,
                      x->x_io[0]->s_shit, &x->x_buf, x->x_io[0]->s_n);
#endif
        if (u_debug)
            post("delwrite %lx: %lx->", x, x->x_io[0]->s_shit);
    }
}

void delwrite_clean(t_delwrite *x)
{
    u_clean((t_ugen *) x);
    if (x->x_sym->s_thing
        && ob_messlist(x->x_sym->s_thing) == delwrite_class)
    {
        t_delwrite stub, *y, *z;
        stub.x_next = (t_delwrite *) (x->x_sym->s_thing);
        for (y = &stub; z = y->x_next; y = z)
        {
            if (z == x)
            {
                y->x_next = z->x_next;
                break;
            }
        }
        x->x_sym->s_thing = (t_object *) (stub.x_next);
    }
    if (x->x_malloc)
        uncachedFree((void *) (x->x_malloc));
}

void *delwrite_class;
void delwrite_setup()
{
    c_new(&s__sigdelwrite, delwrite_clean, sizeof(t_delwrite), 0, 0);
    c_addmess(delwrite_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
}

/* problem -- we don't know the sampling rate yet!  How do you know
   how big to make the buffer??    */

extern int h_ishost;
void *sigdelwrite_new(t_symbol *s, double maxdel, long int phys)
{
    t_delwrite *x =
        (t_delwrite *) obj_new((t_messlist **) & delwrite_class,
                               delwrite_setup);
    long samps, length, i;
    float *fp;
    if (h_ishost)
        maxdel = 0;             /* cheat: if on host, don't need buffer */
    if (maxdel < 0)
        maxdel = 0;
    samps = MAXVS + 44.1 * maxdel;
    if (samps < MAXVS)
        samps = MAXVS;
    samps += (-samps) & (MAXVS - 1);
    length = samps + MAXVS;

    /* just ignore the phys argument, and allocate the memory */
    x->x_malloc =
        (char *) uncachedMalloc(length * sizeof(float) + DELAY_ALIGN);
    if (x->x_malloc)
    {
#ifndef __alpha
        /* fix the alignment */
        x->x_buf.b_shit = (float *) ((long) (x->x_malloc) +
                                     ((-(long) (x->x_malloc)) &
                                      (DELAY_ALIGN - 1)));
#else
        x->x_buf.b_shit = (float *) x->x_malloc;
#endif
    }
    else
    {
        error("delwrite: no space for delay line");
        x->x_buf.b_shit = 0;
    }
    if (x->x_buf.b_shit)
    {
        for (i = length, fp = x->x_buf.b_shit; i--;)
            *(fp++) = 0;
    }

    x->x_buf.b_nsamps = samps;
    x->x_buf.b_phase = 0;
    u_setup((t_ugen *) x, IN1, OUT1);
    if (u_debug)
        post("delwrite %lx", x);
    x->x_sym = s;
    if (s->s_thing)
    {
        if (ob_messlist(s->s_thing) != delwrite_class)
        {
            error("%s: already exists and is not a delay line", s->s_name);
        }
        else
        {
            x->x_next = ((t_delwrite *) (s->s_thing));
            s->s_thing = (t_object *) x;
        }
    }
    else
    {
        s->s_thing = (t_object *) x;
        x->x_next = 0;
    }
    x->x_buf.b_conv = dsp_sr * .001;
    return (x);
#if BUGGY_VERSION
    t_delwrite *x =
        (t_delwrite *) obj_new((t_messlist **) & delwrite_class,
                               delwrite_setup);
    long samps, length, i;
    float *fp;
    if (h_ishost)
        maxdel = 0;             /* cheat: if on host, don't need buffer */
    if (maxdel < 0)
        maxdel = 0;
    samps = MAXVS + .001 * dsp_sr * maxdel;
    if (samps < MAXVS)
        samps = MAXVS;
    samps += (-samps) & (MAXVS - 1);
    length = samps + MAXVS;

    /* just ignore the phys argument, and allocate the memory */
    x->x_malloc = uncachedMalloc(length * sizeof(float) + DELAY_ALIGN);
    if (x->x_malloc)
    {
        /* fix the alignment */
        x->x_buf.b_shit = (float *) ((long) (x->x_malloc) +
                                     ((-(long) (x->x_malloc)) &
                                      (DELAY_ALIGN - 1)));
    }
    else
    {
        error("delwrite: no space for delay line");
        x->x_buf.b_shit = 0;
    }
    if (x->x_buf.b_shit)
    {
        for (i = length, fp = x->x_buf.b_shit; i--;)
            *(fp++) = 0;
    }

    x->x_buf.b_nsamps = samps;
    x->x_buf.b_phase = 0;
    u_setup((t_ugen *) x, IN1, OUT1);
    if (u_debug)
        post("delwrite %lx", x);
    x->x_sym = s;
    if (s->s_thing)
    {
        if (ob_messlist(s->s_thing) != delwrite_class)
        {
            error("%s: already exists and is not a delay line", s->s_name);
        }
        else
        {
            x->x_next = ((t_delwrite *) (s->s_thing));
            s->s_thing = (t_object *) x;
        }
    }
    else
    {
        s->s_thing = (t_object *) x;
        x->x_next = 0;
    }
    x->x_buf.b_conv = dsp_sr * .001;
    return (x);
#endif
}

/******************** DELREAD ***********************/

typedef struct delread {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_symbol *x_sym;
    float x_rawdeltime;         /* nominal delay time */
    float x_delonset;           /* 0 if read before write, vecsize/SR otherwise */
    float x_deltime;            /* difference of two above */
    float *x_ctl;
    int x_hack;                 /* hack flag for I860 microcode */
} t_delread;

#ifndef I860
void cu_delread(float *out, t_delbuf *b, long int n, float *deltime)
{
    int offset = *deltime * b->b_conv + n;
    float *bp;
    if (offset > b->b_nsamps)
        offset = b->b_nsamps;
    else if (offset < n)
        offset = n;
    bp = b->b_shit + b->b_phase - offset;
    if (bp < b->b_shit)
        bp += b->b_nsamps;
    bcopy(bp, out, n * sizeof(float));
}
#endif

void delread_float(t_delread *x, double f)
{
    *x->x_ctl = (x->x_rawdeltime = f) - x->x_delonset;
}

void delread_int(t_delread *x, long int n)
{
    delread_float(x, (double) n);
}

void delread_put(t_delread *x, long int whether)
{
    if (whether)
    {
        t_delwrite *wt;
        u_stdout((t_ugen *) x);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_deltime,
                                     sizeof(float), sizeof(float));
        if ((wt = (t_delwrite *) (x->x_sym->s_thing)) &&
            ob_messlist(wt) == delwrite_class)
        {
            x->x_delonset = (wt->x_h.h_bazzed ?
                             0 : x->x_io[0]->s_n / (dsp_sr * .001));
            if (u_debug)
                post("onset %f", x->x_delonset);
            *x->x_ctl = x->x_rawdeltime - x->x_delonset;
#ifdef I860
            dspchain_put860(u_chain, su_delread, 5,
                            x->x_io[1]->s_shit, &wt->x_buf,
                            x->x_io[1]->s_n, x->x_ctl, x->x_hack);
#else
            dspchain_putc(u_chain, cu_delread, 4,
                          x->x_io[1]->s_shit, &wt->x_buf,
                          x->x_io[1]->s_n, x->x_ctl);
#endif
        }
        else
            error("delread: no corresponding write");
        if (u_debug)
            post("delread %lx: -> %lx", x, x->x_io[1]->s_shit);
    }
    else
    {
        x->x_deltime = *x->x_ctl;
        x->x_ctl = &x->x_deltime;
    }
}

void delread_setup()
{
    c_new(&s__sigdelread, u_clean, sizeof(t_delread), 0, 0);
    c_addmess(delread_put, &s_put, A_CANT, 0);
    c_addfloat(delread_float);
    c_addint(delread_int);
    u_inletmethod(0);
}

void *delread_class;
void *sigdelread_new(t_symbol *s, double f, long int hack)
{
    t_delread *x =
        (t_delread *) obj_new((t_messlist **) & delread_class,
                              delread_setup);
    u_setup((t_ugen *) x, IN1, OUT1);
    u_setforcer(x);
    x->x_sym = s;
    x->x_rawdeltime = x->x_deltime = f;
    x->x_delonset = 0;
    x->x_ctl = &x->x_deltime;
    x->x_hack = hack;
    if (u_debug)
        post("delread %lx", x);
    return (x);
}

/******************** VD (variable delay) ***********************/
/* a four-point interpolating delay read with sample-by-sample variation in
   delay time.  The interpolation algorithm is by cubic fit.  */

typedef struct vdctl {
    float c_fudge;              /* 1/(n-1) */
    long c_maxspan;             /* maximum travel allowed per call */
    float c_delonset;           /* correction to subtract from deltimes */
    float c_writevecsize;       /* number of samps in delwrite per vd call */
} t_vdctl;

typedef struct vd {
    t_head x_h;
    t_sig *x_io[IN2 + OUT1];
    t_symbol *x_sym;
    float x_maxinc;             /* maximum inc allowed per samp */
    float x_fixdel;             /* (sr of write) / (sr of vd) */
    t_vdctl *x_ctl;
    t_vdctl x_cspace;
} t_vd;

#ifndef I860
/* unused in the mockup but used in the i860 version */
void cu_vd(float *in, float *out, t_delbuf *b, t_vdctl *x, long int n)
{
    /* bp is incremented each sample and corresponds to zero delay;
       at the outset it points into the buffer proper but at the end it
       could point n-1 samps past the logical end.  Since the offset is
       bounded below by n, bp - offset does not point past the end;
       in fact you can read four consecutive samples starting from there
       (but bp-offset might point before the beginning; this is guarded
       against in the loop.) */
    float *bp = b->b_shit + b->b_phase;
    long minoffset = n * SAMP_NPOINTS;
    long maxoffset = (b->b_nsamps - 3) * SAMP_NPOINTS;
    float nforward = 0, finc = SAMP_NPOINTS * x->c_writevecsize;
    float conv = b->b_conv, delonset = x->c_delonset;
    float *shit = b->b_shit;
    int nsamps = b->b_nsamps;
    while (n--)
    {
        float *fp;
        float foffset = *(in++) * conv - delonset;
        long bigoffset, offset, interpindex;
        t_sampfilt *t;
        bigoffset = foffset * SAMP_NPOINTS;
        if (bigoffset < minoffset)
            bigoffset = minoffset;
        else if (bigoffset > maxoffset)
            bigoffset = maxoffset;
        bigoffset -= nforward;

        nforward += finc;
        offset = bigoffset >> SAMP_NBITS;
        interpindex = bigoffset & (SAMP_NPOINTS - 1);
        t = sampfilt_tab + interpindex;
        /* printf("(in %f) %f, %d\n", in[-1], foffset, offset); */
        fp = bp - offset;
        if (fp < shit + 3)
            fp += nsamps;
        *out++ = fp[0] * t->f1 + fp[-1] * t->f2
            + fp[-2] * t->f3 + fp[-3] * t->f4;
    }
}
#endif

void fuck();

char *getpage(int size, int cacheable);
int make_samp_tab()
{
    int i;
    if (SAMP_NPOINTS * sizeof(t_sampfilt) != 0x1000)
        bug("make_samp_tab");
    if (!(sampfilt_tab =
          (t_sampfilt *) getpage(SAMP_NPOINTS * sizeof(t_sampfilt), 1)))
        return (0);
    for (i = 0; i < SAMP_NPOINTS; i++)
    {
        double f = i * 0.00390625;
        double fp1 = 1. + f;
        double p1mf = 1. - f;
        double p2mf = 2. - f;
        sampfilt_tab[i].f1 = -.1666667 * f * p1mf * p2mf;
        fuck(&f);
        sampfilt_tab[i].f2 = .5 * fp1 * p1mf * p2mf;
        fuck(&f);
        sampfilt_tab[i].f3 = .5 * fp1 * f * p2mf;
        fuck(&f);
        sampfilt_tab[i].f4 = -.1666667 * fp1 * f * p1mf;
    }
    return (1);
}

void vd_put(t_vd *x, long int whether)
{
    if (whether)
    {
        t_delwrite *wt;
        u_stdout((t_ugen *) x);
        x->x_cspace.c_maxspan = x->x_maxinc * x->x_io[0]->s_n;
        x->x_cspace.c_maxspan += ((-x->x_cspace.c_maxspan) & (MINVS - 1));
        x->x_cspace.c_fudge =
            (x->x_io[0]->s_n > 1 ? 1. / (x->x_io[0]->s_n - 1) : 0);
        x->x_cspace.c_writevecsize = x->x_fixdel;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_vdctl), sizeof(float));
        if ((wt = (t_delwrite *) (x->x_sym->s_thing)) &&
            ob_messlist(wt) == delwrite_class)
        {
            if (u_debug)
                post("onset %f", x->x_ctl->c_delonset);
#ifdef I860
            x->x_ctl->c_delonset = (wt->x_h.h_bazzed ?
                                    -x->x_io[0]->s_n / (dsp_sr *
                                                        .001) : 0);
            dspchain_put860(u_chain, su_vd, 5, x->x_io[0]->s_shit,
                            x->x_io[2]->s_shit, &wt->x_buf, x->x_ctl,
                            x->x_io[0]->s_n);
#else
            x->x_ctl->c_delonset = (wt->x_h.h_bazzed ?
                                    1 - x->x_io[0]->s_n : 1);
            dspchain_putc(u_chain, cu_vd, 5,
                          x->x_io[0]->s_shit, x->x_io[2]->s_shit,
                          &wt->x_buf, x->x_ctl, x->x_io[0]->s_n);
#endif
            if (u_debug)
                post("vd %lx: (%lx) -> %lx", x,
                     x->x_io[0]->s_shit, x->x_io[2]->s_shit);
        }
        else
            error("vd: no corresponding write");
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_vdctl));
        x->x_ctl = &x->x_cspace;
    }
}

void vd_setup()
{
    c_new(&s__sigvd, u_clean, sizeof(t_vd), 0, 0);
    c_addmess(vd_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    u_inletmethod(1);
}

void *vd_class;
void *sigvd_new(t_symbol *s, double maxinc, long int down)
{
    t_vd *x;
    if (!sampfilt_tab && !make_samp_tab())
    {
        error("vd: out of memory");
        return (0);
    }
    x = (t_vd *) obj_new((t_messlist **) & vd_class, vd_setup);
    u_setup((t_ugen *) x, IN2, OUT1);
    x->x_sym = s;
    if (maxinc < 1.)
        x->x_maxinc = 4.;
    else
        x->x_maxinc = maxinc;
    x->x_cspace.c_maxspan = x->x_maxinc * DEFAULTVS;
    x->x_cspace.c_fudge = 1. / (DEFAULTVS - 1);
    x->x_ctl = &x->x_cspace;
    if (down < 0)
        down = 0;
    else if (down > 15)
        down = 15;
    x->x_fixdel = 1. / (1 << down);
    if (u_debug)
        post("vd %lx", x);
    return (x);
}
