#ifndef lint
static char rcsid[] =
    "$Revision: 1.13 $ IRCAM $Date: 1994/01/29 07:47:19 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.13 $ IRCAM $Date: 1994/01/29 07:47:19 $
 *
 * FTS by Miller Puckette
 * writesf~, readsf~: disk recording and playback
 *
 */

extern struct symbol
 s__sigwritesf, s__sigreadsf, s__sigvreadsf, s_put, s_open, s_close,
    s_tick, s_play, s_start, s_stop, s_debug, s_, s_print;

#include <stdio.h>
#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"

#ifdef ISPW
#include "../6a/6a.h"
#include "../6a/proto56.h"
#else
#include "d_disk.h"
#endif


#include "h_cp.h"
#ifdef GRAPHICS
#include "getNameValue.h"
#endif

#ifdef NEXTSTEP
#include <stdio.h>
#include <sound/soundstruct.h>
SNDSoundStruct io_soundheader =
    { SND_MAGIC, sizeof(SNDSoundStruct), 0x10000000,
    SND_FORMAT_LINEAR_16, 44100., 1, 0x62617a00
};

extern int errno;
#else
#include <errno.h>
#endif

#if !defined(GRAPHICS) && (defined(I860) || defined(__unix))
#define FTSSOUNDDISK
#endif

#if defined(NEXTSTEP) && !defined(GRAPHICS)
#define NEXTSOUNDHACK
#endif

#if defined(NEXTSTEP) && defined(GRAPHICS)
#define HOSTSTARTSERVER
#endif

extern t_symbol *defvolume;
extern int h_ishost;

#ifdef HOSTSTARTSERVER
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>

extern int io_slots[MAXSLOTNO];
#define SOUNDDISK "max-sounddisk-v2"
char *strerror();

io_openserver(char *directory, int serverno, int throttle1, int throttle2,
              int throttle3, char **errorstring)
{
    char namebuf[1024], servname[10], throttle1name[10],
        throttle2name[10], throttle3name[10], slotstring[100];
    int spoon, pipe1[2], slot, qfd;
    struct stat statbuf;

    sprintf(namebuf, "%s/%s", directory, SOUNDDISK);
    sprintf(servname, "%d", serverno);
    sprintf(throttle1name, "%d", throttle1);
    sprintf(throttle2name, "%d", throttle2);
    sprintf(throttle3name, "%d", throttle3);
    slotstring[0] = '\0';
    /* check that the file exists, belongs to root, and is setuid */
    if (stat(namebuf, &statbuf) < 0)
    {
        *errorstring = strerror(errno);
        return (-1);
    }
    else if (statbuf.st_uid || !(statbuf.st_mode & S_ISUID))
    {
        *errorstring = "must be owned by root and setuid";
        return (-1);
    }
    for (slot = 0; slot < MAXSLOTNO; slot++)
        if (io_slots[slot])
            sprintf(slotstring + strlen(slotstring), "%d,", slot);
    slotstring[strlen(slotstring) - 1] = '\0';

    if (pipe(pipe1) < 0)
    {
        *errorstring = strerror(errno);
        return -1;
    }
    if ((spoon = fork()) < 0)
    {
        *errorstring = strerror(errno);
        return -1;
    }

    /* Child */
    if (!spoon)
    {
        close(pipe1[1]);
        if (pipe1[0] != 0)
        {
            if (dup2(pipe1[0], 0) < 0)
            {
                *errorstring = strerror(errno);
                return -1;
            }
            close(pipe1[0]);
        }

        execlp(namebuf, "-SOUNDDISK", servname, throttle1name,
               throttle2name, throttle3name, slotstring, 0);
        fprintf(stderr, "%s: exec failed\n", namebuf);
        exit(0);
    }

    /* Parent */
    close(pipe1[0]);
    return (pipe1[1]);
}

io_closeserver(int fd)
{
    close(fd);
}

/***** host starts tasks to serve fts-resident read/write ******/

#define RSF_MAXSERVER 10
typedef struct rsf_server {
    int s_count;
    int s_fd;
} t_server;
t_server rsf_servers[RSF_MAXSERVER];

rsf_openserver(serverno)
{
    int th1, th2, th3;
    char *errorstring;
    h_getthrottle(&th1, &th2, &th3);
    if (th1 < 0)
        error("throttle1 too small"), th1 = 0;
    if (th2 < 0)
        error("throttle2 too small"), th2 = 0;
    if (th3 < 0)
        error("throttle3 too small"), th3 = 0;
    if ((rsf_servers[serverno - 1].s_fd =
         io_openserver(h_getwherebooted(), serverno, th1, th2, th3,
                       &errorstring)) < 0)
        ouch("couldn't open server for readsf~/writesf~; %s", errorstring);
}

rsf__restart()
{
    int i;
    for (i = 0; i < RSF_MAXSERVER; i++)
    {
        if (rsf_servers[i].s_count)
        {
            io_closeserver(rsf_servers[i].s_fd);
            rsf_openserver(i + 1);
        }
    }
}
#endif

#ifdef __unix
#define RSF_MAXSERVER 1
#endif

#ifdef FTSSOUNDDISK
/******* "afd": asychronous I/O handling; not a MAX class itself *******/


#ifdef I860
#define AFDSTORAGECLASS volatile
char *uncachedMalloc(unsigned int n);
#endif

#if defined (__unix)
    /* on SGI don't need to declare shared memory volatile */
#define AFDSTORAGECLASS
#include <stdio.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif
#ifdef __sgi
#include <sys/schedctl.h>
#endif

afd_open(AFDSTORAGECLASS t_asyncfd *x, int mode, long onset, int format)
{
    x->afd_h.h_closerequest = 0;
    x->afd_onset = onset;
    x->afd_cponset = 0;
    x->afd_hostonset = 0;
    x->afd_onsetformat = format;
    if (mode)
        x->afd_h.h_openrequest = AFDREQ_OPENTOWRITE;
    else
        x->afd_h.h_openrequest = AFDREQ_OPENTOREAD;
}

afd_close(AFDSTORAGECLASS t_asyncfd *x)
{
    x->afd_h.h_openrequest = 0;
    x->afd_h.h_closerequest = 1;
}

void afd_free(AFDSTORAGECLASS t_asyncfd *x)     /* mark it unused */
{
    int fd = x->afd_fd;
    /* prevent any servers from taking an interest */
    x->afd_h.h_serverno = 0;
#ifdef I860
    uncachedFree(x->afd_buf);
    uncachedFree(x->afd_filename);
#endif
#ifdef __unix
    /* detach the shared memory ( the child disposes of it.) */
    if (shmdt((char *) x) < 0)
        perror("shmdt");
    /* send a character down the pipe.  The server, when it gets this,
     * will know to exit  (it isn't enough just to close our end
     * of the pipe because another forked process might still have
     * that open,  so the server might not really get an EOF.
     */
    write(fd, "\n", 1);
    close(fd);
#endif
}

t_symbol *cp_getdir();

/* On ISPW, afd_new gets one from host/CP common structure .
 * On SGI, it makes a shared memory region and forks a server.
 */

t_asyncfd *afd_new(int nchans, int who, int bidirectional, int bufsize)
{
    int i;
    t_asyncfd *a;

#ifdef I860
    char *soundbuf, *namebuf;
    if (!(sndbuf = uncachedMalloc(bufsize)))
    {
        error("read/writesf~: out of memory");
        return (0);
    }
    if (!(namebuf = uncachedMalloc(AFD_NAMESIZE)))
    {
        error("read/writesf~: out of memory");
        uncachedFree(sndbuf);
        return (0);
    }
    for (i = 0, a = PROTO->p_asyncfd; i < MAXAFD; i++, a++)
    {
        if (!a->afd_h.h_serverno)
        {
            a->afd_buf = sndbuf;
            a->afd_filename = namebuf;
            goto gotone;
        }
    }
    error("read/writesf~: only %d allowed", MAXAFD);
    return 0;
#endif
#ifdef __unix
    char *sharedmem, idbuf[20], execbuf[1024];
    int shmemid, spoon, pipe1[2];
    struct stat statbuf;

    if ((shmemid = shmget(IPC_PRIVATE, SOUNDONSET + bufsize,
                          SHM_R | SHM_W)) == -1)
    {
        sysouch("read/writesf: shmget");
        return 0;
    }

    /* Point `shmaddr' at the memory */
    if ((sharedmem = shmat(shmemid, 0, 0)) == (void *) -1)
    {
        sysouch("read/writesf: shmat");
        return 0;
    }
    a = (t_asyncfd *) sharedmem;
    a->afd_filename = sharedmem + NAMEONSET;
    a->afd_buf = sharedmem + SOUNDONSET;
    a->afd_self = a;
#endif

  gotone:
    a->afd_size = bufsize;
    a->afd_h.h_openrequest = 0;
    a->afd_h.h_closerequest = 0;
    a->afd_h.h_eof = 0;
    a->afd_nchans = nchans;
    a->afd_bidirectional = bidirectional;
    a->afd_h.h_serverno = who;

#ifdef __unix

    /* check that the file exists, belongs to root, and is setuid */
    strcpy(execbuf, cp_getdir()->s_name);
    strcat(execbuf, "/max-sounddisk");
    sprintf(idbuf, "%d", shmemid);

    if (stat(execbuf, &statbuf) < 0)
    {
        sysouch(execbuf);
        return (0);
    }
    /* open a pipe to child */
    if (pipe(pipe1) < 0)
    {
        sysouch("read/writesf: pipe");
        return (0);
    }
    if ((spoon = fork()) < 0)
    {
        sysouch("read/writesf: fork");
        return (0);
    }

    /* Child */
    if (!spoon)
    {
        close(pipe1[1]);
#ifdef __sgi
        if (sgi_gotrealtime())
        {
            if (schedctl(NDPRI, 0, 0) < 0)
                fprintf(stderr,
                        "max-sounddisk: warning: couldn't unset priority\n");
        }
#endif
        if (pipe1[0] != 0)
        {
            if (dup2(pipe1[0], 0) < 0)
            {
                fprintf(stderr, "read/writesf: dup failed\n");
                _exit(1);
            }
            close(pipe1[0]);
        }

        execlp(execbuf, "max-sounddisk", idbuf, 0);
        fprintf(stderr, "%s: exec failed\n", execbuf);
        _exit(1);
    }

    /* Parent */
    close(pipe1[0]);
    a->afd_fd = pipe1[1];

#endif
    return (a);
}
#endif


#define RSF_MAXCHANS MAXIN
static int RSF_BUFSIZE = 0x40000;       /* should be a DEFINE but it breaks gcc */

#define S_ACTIVE  0
#define S_OPENED  1
#define S_INACTIVE  2

typedef struct rsfctl {
    float *c_inout[RSF_MAXCHANS];
    int c_nchans;
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *c_afd;
#endif                          /* FTSSOUNDDISK */
#ifdef NEXTSOUNDHACK
    FILE *c_fd;
    float c_sr;
    t_symbol *c_vol;
#endif                          /* NEXTSOUNDHACK */
    char *c_buf;
    int c_state;
    int c_bufsize;
} t_rsfctl;


/************** WRITESF ***************/

typedef struct sigwritesf {
    t_head w_h;
    t_sig *w_io[U_IN(1) + U_OUT(RSF_MAXCHANS)];
    t_rsfctl w_cspace;
    t_rsfctl *w_ctl;
    t_symbol *w_vol;
#ifdef GRAPHICS
    int w_serverno;
#else                           /* GRAPHICS */
    int w_bufsize;
#endif                          /* GRAPHICS */
} t_sigwritesf;

soundtofile(float *fp, short int *sp, int n, int skip)  /* rewrite this later */
{
    while (n--)
    {
        double f = *fp++;
        if (f > 1.)
            f = 1.;
        else if (f < -1.)
            f = -1.;
#ifdef I860
        {
            short x = 32767. * f;
            char *z1 = (char *) &x;
            char *z2 = (char *) sp;
            z2[0] = z1[1];
            z2[1] = z1[0];
        }
#else                           /* I860 */
        *sp = 32767. * f;
#endif                          /* I860 */
        sp += skip;
    }
}

void cu_writesf(t_rsfctl *c, long int n)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    int i, nchans = c->c_nchans;
    long head = a->afd_cponset, tail = a->afd_hostonset, bytesavail,
        transfersize = sizeof(t_filesamp) * nchans * n,
        bufsize = c->c_bufsize;
    short *qbuf = (short *) (c->c_buf + head);
    bytesavail = tail - head;
    if (c->c_state != S_ACTIVE)
        return;
    if (a->afd_h.h_eof)
    {
        error("writesf: %s", strerror((int) a->afd_h.h_eof));
        c->c_state = S_INACTIVE;
        return;
    }
    if (bytesavail <= 0)
        bytesavail += bufsize;
    if (bytesavail <= transfersize)
    {
        static int nextwarn = 0;
        if (logicaltime >= nextwarn)
            post("writesf~: waiting for host... ");
        while (bytesavail <= transfersize)
        {
#ifdef __unix
            /* On the I860 we could just spin; here we sleep */
            sys_trytoflush();
            microsleep(5000);
#endif
            tail = a->afd_hostonset;
            bytesavail = tail - head;
            if (bytesavail < 0)
                bytesavail += bufsize;
        }
        if (logicaltime >= nextwarn)
        {
            post("done.");
            nextwarn = logicaltime + 2000;
        }
    }
    for (i = 0; i < nchans; i++)
        soundtofile(c->c_inout[i], qbuf++, n, nchans);
    head += transfersize;
    if (head == bufsize)
        head = 0;
    a->afd_cponset = head;
#endif
#if defined(NEXTSOUNDHACK)
    int i, nchans = c->c_nchans;
    short buf[RSF_MAXCHANS * MAXVS];
    if (c->c_state == S_ACTIVE)
    {
        for (i = 0; i < nchans; i++)
            soundtofile(c->c_inout[i], buf + i, n, nchans);
        if (fwrite((char *) buf, sizeof(t_filesamp), nchans * n, c->c_fd)
            < nchans * n)
        {
            error("writesf: %s", strerror(errno));
            fclose(c->c_fd);
            c->c_state = S_INACTIVE;
        }
    }
#endif                          /* NEXTSOUNDHACK */
}

void sigwritesf_put(t_sigwritesf *x, long int whether)
{
    int i;
    if (whether)
    {
        u_stdout((t_ugen *) x);
        for (i = 0; i < x->w_cspace.c_nchans; i++)
            x->w_cspace.c_inout[i] = x->w_io[i]->s_shit;
        x->w_ctl = dspchain_putstate(u_chain, &x->w_cspace,
                                     sizeof(x->w_cspace), sizeof(float));
#ifdef FTSSOUNDDISK
        x->w_ctl->c_afd->afd_sr = x->w_io[0]->s_sr;
#endif
#ifdef NEXTSOUNDHACK
        x->w_ctl->c_sr = x->w_io[0]->s_sr;
#endif
        dspchain_putc(u_chain, cu_writesf, 2, x->w_ctl, x->w_io[0]->s_n);
        if (u_debug)
            post("writesf %lx: %lx(...)->", x, x->w_io[0]->s_shit);
    }
    else
    {
        bcopy(x->w_ctl, &x->w_cspace, sizeof(x->w_cspace));
        x->w_ctl = &x->w_cspace;
    }
}

void sigwritesf_print(t_sigwritesf *x)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->w_ctl->c_afd;
    if (x->w_ctl->c_state == S_OPENED)
        post("opened");
    else if (x->w_ctl->c_state == S_ACTIVE)
        post("active");
    else if (x->w_ctl->c_state == S_INACTIVE)
        post("inactive");
    else
        post("state?");
    post("host onset %ld; cp onset %ld", a->afd_hostonset, a->afd_cponset);
#endif                          /* FTSSOUNDDISK */
}

void sigwritesf_open(t_sigwritesf *x, t_symbol *s)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->w_ctl->c_afd;
    if (a->afd_h.h_openrequest)
    {
        error("sigwritesf_open: an open request is already pending");
        return;
    }
    if (x->w_ctl->c_state == S_ACTIVE)
    {
        error("sigwritesf_open: already active");
        return;
    }
    if (a->afd_h.h_closerequest)
    {
        post("sigwritesf_open: waiting for host... ");
        while (a->afd_h.h_closerequest)
        {
#ifdef __unix
            sys_trytoflush();
            microsleep(5000);
#else
            int count;
            for (count = 0; count < 1000; count++);
#endif
        }
        poststring("done");
    }
    if (*s->s_name == '/')
    {
        strncpy(a->afd_filename, s->s_name, AFD_NAMESIZE);
        a->afd_filename[AFD_NAMESIZE - 1] = 0;
    }
    else
        t_composename(a->afd_filename, AFD_NAMESIZE,
                      x->w_vol->s_name, s->s_name);
    afd_open(a, 1, 0, 0);
    x->w_ctl->c_state = S_OPENED;
#endif                          /* FTSSOUNDDISK */
#if defined(NEXTSOUNDHACK)
    int ufd;
    if (x->w_ctl->c_state == S_OPENED || x->w_ctl->c_state == S_ACTIVE)
        fclose(x->w_ctl->c_fd);
    if (createfile(s->s_name, x->w_ctl->c_vol, 0, &ufd))
    {
        x->w_ctl->c_fd = fdopen(ufd, "w");
        x->w_ctl->c_state = S_OPENED;
    }
    else
    {
        error("writesf: %s", strerror(errno));
        x->w_ctl->c_state = S_INACTIVE;
    }
    io_soundheader.samplingRate = x->w_ctl->c_sr;
    io_soundheader.channelCount = x->w_ctl->c_nchans;

    fwrite(&io_soundheader, sizeof(io_soundheader), 1, x->w_ctl->c_fd);
#endif                          /* NEXTSOUNDHACK */
}

void sigwritesf_int(t_sigwritesf *x, long int n)
{
#ifdef FTSSOUNDDISK
    t_rsfctl *c = x->w_ctl;
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    if (n)
    {
        if (c->c_state != S_OPENED)
        {
            error
                ("sigwritesf_int: start requested without preceeding 'open'");
            return;
        }
        c->c_state = S_ACTIVE;
    }
    else
    {
        if (a->afd_h.h_openrequest)
        {
            /* this is not good; it should be possible to have a simultaneous
               open and close request in the "write" case. */
            post("sigwritesf_close: waiting for open to complete... ");
            while (a->afd_h.h_openrequest)
            {
#ifdef __unix
                sys_trytoflush();
                microsleep(5000);
#else
                int count;
                for (count = 0; count < 1000; count++);
#endif
            }
            poststring("done");
        }
        if (!a->afd_h.h_closerequest)
            afd_close(a);
        else
            post("close request pending?");
        c->c_state = S_INACTIVE;
    }
#endif                          /* FTSSOUNDDISK */
#if defined(NEXTSOUNDHACK)
    t_rsfctl *c = x->w_ctl;
    if (n)
    {
        if (c->c_state != S_OPENED)
        {
            error
                ("sigwritesf_int: start requested without preceeding 'open'");
            return;
        }
        c->c_state = S_ACTIVE;
    }
    else
    {
        if (c->c_state == S_OPENED || c->c_state == S_ACTIVE)
            fclose(c->c_fd);
        c->c_state = S_INACTIVE;
    }
#endif
}

void sigwritesf_play(t_sigwritesf *x, t_symbol *s)
{
#if defined(NEXTSOUNDHACK)
    char buf[1024];
    if (*s->s_name == '/')
        sprintf(buf, "sndplay %s\n", s->s_name);
    else
        sprintf(buf, "sndplay %s/%s\n", x->w_ctl->c_vol->s_name,
                s->s_name);
    fprintf(stderr, "%s", buf);
    system(buf);
#endif
}

void sigwritesf_ff(t_sigwritesf *x)
{
#ifdef HOSTSTARTSERVER
    t_server *foo = &rsf_servers[x->w_serverno - 1];
    if ((!--foo->s_count) && foo->s_fd >= 0)
        io_closeserver(foo->s_fd);
#endif
    u_clean((t_ugen *) x);
#ifdef FTSSOUNDDISK
    afd_free(x->w_ctl->c_afd);
#endif
}

void *sigwritesf_class;
void sigwritesf_setup()
{
    int i;
    c_new(&s__sigwritesf, sigwritesf_ff, sizeof(t_sigwritesf), 0, 0);
    c_addmess(sigwritesf_put, &s_put, A_CANT, 0);
    c_addmess(sigwritesf_open, &s_open, A_DEFSYM, 0);
    c_addint(sigwritesf_int);
    c_addmess(sigwritesf_print, &s_print, 0);
    c_addmess(sigwritesf_play, gensym("play"), A_SYM, 0);
    for (i = 0; i < RSF_MAXCHANS; i++)
        u_inletmethod(i);
}

void *sigwritesf_new(long int nchans, long int serverno, long int bufsize)
{
    t_sigwritesf *x;
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a;
#endif
    if (nchans < 1)
        nchans = 1;
    else if (nchans > RSF_MAXCHANS)
        nchans = RSF_MAXCHANS;
    if (serverno > RSF_MAXSERVER)
        serverno = RSF_MAXSERVER;
    else if (serverno < 1)
        serverno = 1;
#ifdef FTSSOUNDDISK
    if (bufsize < RSF_BUFSIZE)
        bufsize = RSF_BUFSIZE;
    bufsize &= ~0xfff;
    bufsize *= nchans;
    if (!(a = afd_new((int) nchans, (int) serverno, 0, bufsize)))
    {
        error("writesf~: error opening file handler");
        return (0);
    }
    post("write: fd = %d", a->afd_fd);
#else
#if defined(NEXTSOUNDHACK)
#else
    if (!h_ishost)
        post("writesf~: standalone only");
#endif                          /* NEXTSOUNDHACK */
#endif
    x = (t_sigwritesf *) obj_new((t_messlist **) & sigwritesf_class,
                                 sigwritesf_setup);
    u_setup((t_ugen *) x, nchans, 0);
#ifdef HOSTSTARTSERVER
    if (whichFts() == standaloneFts)
    {
        if (!rsf_servers[serverno - 1].s_count)
            rsf_openserver(serverno);
        rsf_servers[serverno - 1].s_count++;
        x->w_serverno = serverno;
    }
#endif

#ifdef FTSSOUNDDISK
    x->w_cspace.c_afd = a;
    x->w_cspace.c_buf = a->afd_buf;
    x->w_cspace.c_bufsize = bufsize;
#endif                          /* FTSSOUNDDISK */
#ifdef NEXTSOUNDHACK
    x->w_cspace.c_sr = dsp_sr;
    x->w_cspace.c_fd = 0;
    x->w_cspace.c_vol = defvolume;
#endif                          /* NEXTSOUNDHACK */

    x->w_cspace.c_nchans = nchans;
    x->w_cspace.c_state = S_INACTIVE;
    x->w_ctl = &x->w_cspace;
    x->w_vol = defvolume;
    return (x);
}

/************** readSF ***************/

typedef struct sigreadsf {
    t_head r_h;
    t_sig *r_io[U_IN(1) + U_OUT(RSF_MAXCHANS)];
    t_rsfctl r_cspace;
    t_rsfctl *r_ctl;
    t_symbol *r_vol;
#ifdef HOSTSTARTSERVER
    int r_serverno;
#endif
} t_sigreadsf;

#ifndef I860
    /* untested c version -- I860 uses assembler */
soundfromfile(short int *sp, float *fp, int n, int skip)
{
    while (n--)
    {
        *fp++ = *sp * (1. / 32768.);
        sp += skip;
    }
}
#endif

void cu_readsf(t_rsfctl *c, long int n)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    int i, nchans = c->c_nchans;
    long tail, head, bytesavail,
        transfersize = sizeof(t_filesamp) * nchans * n,
        bufsize = c->c_bufsize;
    short *qbuf;
    if (c->c_state != S_ACTIVE)
        goto zeroit;
    tail = a->afd_cponset;
    head = a->afd_hostonset;
    qbuf = (short *) (c->c_buf + tail);
    bytesavail = head - tail;
    if (bytesavail < 0)
        bytesavail += bufsize;
    if (bytesavail < transfersize)
    {
        static int nextwarn = 0;
        if (a->afd_h.h_eof)
        {
            if (a->afd_h.h_eof != '\377')
                error("readsf: %s", strerror((int) a->afd_h.h_eof));
            c->c_state = S_INACTIVE;
            goto zeroit;
        }
        if (logicaltime >= nextwarn)
        {
            error("readsf~: data is late");
            nextwarn = logicaltime + 2000;
        }
        while (bytesavail < transfersize)
        {
#ifdef __unix
            sys_trytoflush();
            microsleep(5000);
#endif
            head = a->afd_hostonset;
            bytesavail = head - tail;
            if (bytesavail < 0)
                bytesavail += bufsize;
        }
    }
    for (i = 0; i < nchans; i++)
        soundfromfile(qbuf++, c->c_inout[i], n, nchans);
    tail += transfersize;
    if (tail == bufsize)
        tail = 0;
    a->afd_cponset = tail;
    return;
  zeroit:
    for (i = 0; i < nchans; i++)
        cu_zero(c->c_inout[i], n);
#endif
}

void sigreadsf_put(t_sigreadsf *x, long int whether)
{
#ifdef FTSSOUNDDISK
    int i;
    if (whether)
    {
        u_stdout((t_ugen *) x);
        for (i = 0; i < x->r_cspace.c_nchans; i++)
            x->r_cspace.c_inout[i] = x->r_io[i + 1]->s_shit;
        x->r_ctl = dspchain_putstate(u_chain, &x->r_cspace,
                                     sizeof(x->r_cspace), sizeof(float));
        dspchain_putc(u_chain, cu_readsf, 2, x->r_ctl, x->r_io[1]->s_n);
        if (u_debug)
            post("readsf %lx: %lx->%lx...",
                 x, x->r_io[0]->s_shit, x->r_io[1]->s_shit);
    }
    else
    {
        bcopy(x->r_ctl, &x->r_cspace, sizeof(x->r_cspace));
        x->r_ctl = &x->r_cspace;
    }
#endif
}

void sigreadsf_open(t_sigreadsf *x, t_symbol *s, long onset,
                    t_symbol *format)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->r_ctl->c_afd;
    if (a->afd_h.h_openrequest)
    {
        error("sigreadsf_open: an open request is already pending");
        return;
    }
    if (*s->s_name == '/')
    {
        strncpy(a->afd_filename, s->s_name, AFD_NAMESIZE);
        a->afd_filename[AFD_NAMESIZE - 1] = 0;
    }
    else
        t_composename(a->afd_filename, AFD_NAMESIZE,
                      x->r_vol->s_name, s->s_name);
    afd_open(a, 0, onset, (*format->s_name == 'b' ? F_BYTE : F_MSEC));
    x->r_ctl->c_state = S_OPENED;
#endif                          /* FTSSOUNDDISK */
}

void sigreadsf_print(t_sigreadsf *x)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->r_ctl->c_afd;
    if (x->r_ctl->c_state == S_OPENED)
        post("opened");
    else if (x->r_ctl->c_state == S_ACTIVE)
        post("active");
    else if (x->r_ctl->c_state == S_INACTIVE)
        post("inactive");
    else
        post("state?");
    post("head %ld; tail %ld", a->afd_hostonset, a->afd_cponset);
#endif                          /* FTSSOUNDDISK */
}

void sigreadsf_int(t_sigreadsf *x, long int n)
{
#ifdef FTSSOUNDDISK
    t_rsfctl *c = x->r_ctl;
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    if (n)
    {
        if (c->c_state != S_OPENED)
        {
            error
                ("sigreadsf_int: start requested without preceeding 'open'");
            return;
        }
        if (a->afd_h.h_openrequest)
        {
            post("sigreadsf_int: waiting for host... ");
            while (a->afd_h.h_openrequest)
            {
#ifdef __unix
                sys_trytoflush();
                microsleep(5000);
#else
                int count;
                for (count = 0; count < 1000; count++);
#endif

            }
            poststring("done.");
        }
        if (!a->afd_hostonset && a->afd_h.h_eof)
        {
            post("%s: open failed\n", a->afd_filename);
            sigreadsf_int(x, 0L);
        }
        else
            c->c_state = S_ACTIVE;
    }
    else
    {
        if (!a->afd_h.h_closerequest)
            afd_close(a);
        c->c_state = S_INACTIVE;
    }
#endif                          /* FTSSOUNDDISK */
}

void sigreadsf_ff(t_sigreadsf *x)
{
#ifdef HOSTSTARTSERVER
    t_server *foo = &rsf_servers[x->r_serverno - 1];
    if ((!--foo->s_count) && foo->s_fd >= 0)
        io_closeserver(foo->s_fd);
#endif
    u_clean((t_ugen *) x);
#ifdef FTSSOUNDDISK
    afd_free(x->r_ctl->c_afd);
#endif
}

void *sigreadsf_class;
void sigreadsf_setup()
{
    c_new(&s__sigreadsf, sigreadsf_ff, sizeof(t_sigreadsf), 0, 0);
    c_addmess(sigreadsf_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    c_addmess(sigreadsf_open, &s_open, A_DEFSYM, A_DEFLONG, A_DEFSYM, 0);
    c_addint(sigreadsf_int);
    c_addmess(sigreadsf_print, &s_print, 0);
}

void *sigreadsf_new(long int nchans, long int serverno, long int bufsize)
{
    t_sigreadsf *x;
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a;
#endif
    if (nchans < 1)
        nchans = 1;
    else if (nchans > RSF_MAXCHANS)
        nchans = RSF_MAXCHANS;
    if (serverno > RSF_MAXSERVER)
        serverno = RSF_MAXSERVER;
    else if (serverno < 1)
        serverno = 1;
#ifdef FTSSOUNDDISK
    if (bufsize < RSF_BUFSIZE)
        bufsize = RSF_BUFSIZE;
    bufsize &= ~0xfff;
    bufsize *= nchans;
    if (!(a = afd_new((int) nchans, (int) serverno, 0, bufsize)))
    {
        error("readsf~: error opening file handler");
        return (0);
    }
#else
    if (!h_ishost)
        post("readsf~: standalone only");
#endif                          /* FTSSOUNDDISK */
    x = (t_sigreadsf *) obj_new((t_messlist **) & sigreadsf_class,
                                sigreadsf_setup);
    u_setup((t_ugen *) x, IN1, nchans);
#ifdef HOSTSTARTSERVER
    if (whichFts() != standaloneFts)
        post("readsf~: standalone only");
    else if (!rsf_servers[serverno - 1].s_count)
        rsf_openserver(serverno);
    rsf_servers[serverno - 1].s_count++;
    x->r_serverno = serverno;
#endif
#ifdef FTSSOUNDDISK
    x->r_cspace.c_afd = a;
    x->r_cspace.c_nchans = nchans;
    x->r_cspace.c_buf = a->afd_buf;
    x->r_cspace.c_state = S_INACTIVE;
    x->r_cspace.c_bufsize = bufsize;
    x->r_ctl = &x->r_cspace;
    x->r_vol = defvolume;
#endif
    return (x);
}

/*************** vreadsf -- variable speed soundfile reader *********/

#ifdef FTSSOUNDDISK
typedef struct vrsfctl          /* variable readsf control structure */
{
    float *c_inout[RSF_MAXCHANS];
    int c_nchans;
    AFDSTORAGECLASS t_asyncfd *c_afd;
    char *c_buf;
    int c_state;
    int c_bufsize;
    t_sampbuf *c_sampbuf[RSF_MAXCHANS];
    double c_phase;
    double c_nwraps;
    double c_phaseinc;
    float c_samprate;
    float c_convertfactor;
    long c_iphase;
    int c_highwritten;
    int c_sampsize;
    int c_bidirectional;
} t_vrsfctl;
#endif

typedef struct sigvreadsf {
    t_head r_h;
    t_sig *r_io[U_IN(1)];
    void *r_outlet[5];
#ifdef FTSSOUNDDISK
    t_vrsfctl r_cspace;
    t_vrsfctl *r_ctl;
    t_symbol *r_vol;
    t_symbol *r_sampname;
#endif
#ifdef HOSTSERVER
    int r_serverno;
#endif
} t_sigvreadsf;

#define XFERCHUNK DEFAULTVS     /* number of samples to transfer in a chunk */
#define MAXXFER (2*XFERCHUNK)
void sigvreadsf_stop(t_sigvreadsf * x);
static int rsf_printnext;

#ifdef FTSSOUNDDISK
void cu_vreadsf(t_vrsfctl *c)
{
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    int i, nchans = c->c_nchans;
    long tail = a->afd_cponset, head = a->afd_hostonset,
        bufsize = c->c_bufsize, bytesavail, backhead, backtail;
    if (c->c_state == S_ACTIVE ||
        (c->c_state == S_OPENED && !a->afd_h.h_openrequest))
    {
        int wantphase =
            c->c_sampsize * (c->c_phase * .5) + (c->c_sampsize / 4);
        int gotphase = c->c_highwritten;
        int transfersize = wantphase - gotphase;
        int backwards = 0;
        if (transfersize < 0)
            transfersize += c->c_sampsize / 2;
        if (transfersize >= c->c_sampsize / 2)
            transfersize -= c->c_sampsize / 2;
        if (transfersize < 0)
            bug("cu_vreadsf 1");
        if (c->c_bidirectional && transfersize > c->c_sampsize / 4)
        {
            backwards = 1;
            transfersize = c->c_sampsize / 2 - transfersize - XFERCHUNK;
            if (transfersize < 0)
                transfersize = 0;
            backhead = a->afd_hostbackonset;
            backtail =
                tail - (c->c_sampsize / 4) * nchans * sizeof(t_filesamp);
            if (backtail < 0)
                backtail += bufsize;
            bytesavail = backtail - backhead;
            if (bytesavail < 0)
                bytesavail += bufsize;
        }
        else
        {
            bytesavail = head - tail;
            if (bytesavail < 0)
                bytesavail += bufsize;
        }
        transfersize &= (~(XFERCHUNK - 1));
        if (transfersize > MAXXFER)
            transfersize = MAXXFER;
        if (rsf_printnext)
        {
            post("avail %ld, xfer %ld", bytesavail, transfersize);
            post("wantphase %ld, gotphase %ld", wantphase, gotphase);
            rsf_printnext--;
        }
        if (bytesavail < transfersize * sizeof(t_filesamp) * nchans)
        {
            if (c->c_state == S_ACTIVE)
            {
                if (a->afd_h.h_eof)     /* fix this to do zero-padding */
                {
                    if (a->afd_h.h_eof != '\377')
                        error("vreadsf: %s",
                              strerror((int) a->afd_h.h_eof));
                    else
                        post("vreadsf: EOF");
                    c->c_state = S_INACTIVE;
                    return;
                }
                post("vreadsf~: waiting for host ... ");
                while (bytesavail <
                       transfersize * sizeof(t_filesamp) * nchans)
                {
#ifdef __unix
                    sys_trytoflush();
                    microsleep(5000);
#endif
                    if (backwards)
                    {
                        backhead = a->afd_hostbackonset;
                        bytesavail = backtail - backhead;
                    }
                    else
                    {
                        head = a->afd_hostonset;
                        bytesavail = head - tail;
                    }
                    if (bytesavail < 0)
                        bytesavail += bufsize;
                }
                poststring("done.");
            }
            else
                return;
        }
        if (c->c_state == S_OPENED && a->afd_h.h_eof)
        {
            if (a->afd_h.h_eof != '\377')
                error("vreadsf: %s", strerror((int) a->afd_h.h_eof));
            else
                post("vreadsf: EOF");
            c->c_state = S_INACTIVE;
            return;
        }
        while (transfersize)
        {
            short *qbuf;
            if (backwards)
            {
                tail -= (XFERCHUNK * sizeof(t_filesamp)) * nchans;
                if (tail < 0)
                    tail += bufsize;
                backtail -= (XFERCHUNK * sizeof(t_filesamp)) * nchans;
                if (backtail < 0)
                    backtail += bufsize;
                qbuf = (short *) (c->c_buf + backtail);
            }
            else
            {
                qbuf = (short *) (c->c_buf + tail);
                tail += (XFERCHUNK * sizeof(t_filesamp)) * nchans;
                if (tail == bufsize)
                    tail = 0;
            }
            for (i = 0; i < nchans; i++)
            {
                if (backwards)
                {
                    int putphase =
                        gotphase - c->c_sampsize / 4 - XFERCHUNK;
                    if (putphase < 0)
                        putphase += c->c_sampsize / 2;
                    soundfromfile(qbuf,
                                  c->c_sampbuf[i]->b_shit + putphase,
                                  XFERCHUNK, nchans);
                    soundfromfile(qbuf,
                                  c->c_sampbuf[i]->b_shit + putphase +
                                  (c->c_sampsize / 2), XFERCHUNK, nchans);
                }
                else
                {
                    soundfromfile(qbuf,
                                  c->c_sampbuf[i]->b_shit + gotphase,
                                  XFERCHUNK, nchans);
                    soundfromfile(qbuf,
                                  c->c_sampbuf[i]->b_shit + gotphase +
                                  (c->c_sampsize / 2), XFERCHUNK, nchans);
                }
                qbuf++;
            }
            if (backwards)
            {
                gotphase -= XFERCHUNK;
                if (gotphase < 0)
                {
                    gotphase += c->c_sampsize / 2;
                }
            }
            else
            {
                gotphase += XFERCHUNK;
                if (gotphase >= (c->c_sampsize / 2))
                {
                    gotphase = 0;
                }
            }
            transfersize -= XFERCHUNK;
            c->c_highwritten = gotphase;
        }
        a->afd_cponset = tail;
        if (c->c_state == S_OPENED && tail >= c->c_sampsize / 4)
            c->c_state = S_ACTIVE;
        c->c_phase += c->c_phaseinc;
        if (c->c_phase >= 1.)
        {
            c->c_phase -= 1.;
            c->c_nwraps += 1.;
        }
        else if (c->c_phase < 0.)
        {
            c->c_phase += 1.;
            c->c_nwraps -= 1.;
        }
        if (c->c_phase < 0. || c->c_phase > 1.)
            bug("cu_vreadsf 2");
    }
}
#endif

extern void *sigtable_class;

void sigvreadsf_set(t_sigvreadsf *x, t_symbol *s)
{
#ifdef FTSSOUNDDISK
    t_sigtable *wt;
    int i, size, bad = 0;
    if (strlen(s->s_name) > 500)
        s = gensym("(name too long)");
    x->r_sampname = s;
    for (i = 0; i < x->r_ctl->c_nchans; i++)
    {
        char strbuf[512];
        t_symbol *s2;
        t_sampbuf *sampbuf;
        gensampname(strbuf, s->s_name, i + 1);
        s2 = gensym(strbuf);
        if ((wt = (t_sigtable *) (s2->s_thing)) &&
            ob_messlist(wt) == sigtable_class)
        {
            sampbuf = &wt->x_buf;
            if (wt->x_next)
                post("sampread(%s): warning: multiply defined",
                     s2->s_name);
        }
        else
        {
            error("vreadsf: %s: no such sample", s2->s_name);
            bad = 1;
            break;
        }
        if (!i)
            size = sampbuf->b_nsamps;
        else if (size != sampbuf->b_nsamps)
        {
            error("vreadsf: %s: samples of different lengths", s->s_name);
            bad = 1;
            break;
        }
        x->r_ctl->c_sampbuf[i] = sampbuf;
    }
    if (!(x->r_ctl->c_sampsize = (size & (~(XFERCHUNK * 2 - 1)))))
    {
        error("vreadsf: %s: too small", s->s_name);
        bad = 1;
    }
    if (bad)
        for (i = 0; i < x->r_ctl->c_nchans; i++)
            x->r_ctl->c_sampbuf[i] = 0;
#endif
}

void sigvreadsf_put(t_sigvreadsf *x, long int whether)
{
#ifdef FTSSOUNDDISK
    if (whether)
    {
        u_stdout((t_ugen *) x);
        sigvreadsf_stop(x);
        sigvreadsf_set(x, x->r_sampname);
        /* x->r_io[0]->s_sr is sample rate; should set a coefficient here */
        dspchain_putc(u_chain, cu_vreadsf, 1, x->r_ctl);
    }
#endif
}

void sigvreadsf_open(t_sigvreadsf *x, t_symbol *s, long onset,
                     t_symbol *format)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->r_ctl->c_afd;
    if (a->afd_h.h_openrequest)
    {
        error("sigvreadsf_open: an open request is already pending");
        return;
    }
    if (*s->s_name == '/')
    {
        strncpy(a->afd_filename, s->s_name, AFD_NAMESIZE);
        a->afd_filename[AFD_NAMESIZE - 1] = 0;
    }
    else
        t_composename(a->afd_filename, AFD_NAMESIZE,
                      x->r_vol->s_name, s->s_name);
    afd_open(a, 0, onset, (*format->s_name == 'b' ? F_BYTE : F_MSEC));
    x->r_ctl->c_state = S_OPENED;
    x->r_ctl->c_highwritten = 0;
    x->r_ctl->c_phase = 0;
    x->r_ctl->c_nwraps = 0;
    x->r_ctl->c_phaseinc = 0;

#endif
}

void sigvreadsf_float(t_sigvreadsf *x, double f)
{
#ifdef FTSSOUNDDISK
    t_vrsfctl *c = x->r_ctl;
    AFDSTORAGECLASS t_asyncfd *a = c->c_afd;
    if (!c->c_sampbuf[0])
    {
        post("sigvreadsf_float: no sample");
        return;
    }
    if (c->c_state == S_OPENED)
    {
        post("sigvreadsf_float: waiting for host... ");
        while (c->c_state == S_OPENED)
        {
#ifdef __unix
            sys_trytoflush();
            microsleep(5000);
#else
            int count;
            for (count = 0; count < 1000; count++);
#endif
            cu_vreadsf(c);
        }
        poststring("done.");
    }
    if (c->c_state == S_ACTIVE)
    {
        /* fix these constants to be variable */
        double incpersample = f / (c->c_sampsize / 2);
        double reportphase =
            .5 * (c->c_phase < .25 ? c->c_phase + 1 : c->c_phase);
        double sampinmsec = c->c_sampsize * (1000. / 44100.);
        c->c_phaseinc = incpersample * DEFAULTVS;
        outlet_float(x->r_outlet[3],
                     sampinmsec * (c->c_phase + c->c_nwraps));
        outlet_float(x->r_outlet[2], sampinmsec);
        outlet_float(x->r_outlet[1],
                     (double) incpersample * (.5 * 44100.));
        outlet_float(x->r_outlet[0], reportphase);
    }
    else
        outlet_bang(x->r_outlet[4]);
#endif
}

void sigvreadsf_int(t_sigvreadsf *x, long n)
{
#ifdef FTSSOUNDDISK
    sigvreadsf_float(x, (double) n);
#endif
}

void sigvreadsf_stop(t_sigvreadsf *x)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->r_ctl->c_afd;
    if (!a->afd_h.h_closerequest)
        afd_close(a);
    x->r_ctl->c_state = S_INACTIVE;
#endif
}

void sigvreadsf_print(t_sigvreadsf *x)
{
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a = x->r_ctl->c_afd;
    if (x->r_ctl->c_state == S_OPENED)
        post("opened");
    else if (x->r_ctl->c_state == S_ACTIVE)
        post("active");
    else if (x->r_ctl->c_state == S_INACTIVE)
        post("inactive");
    else
        post("state?");
    post("head %ld; tail %ld", a->afd_hostonset, a->afd_cponset);
    rsf_printnext = 1;
#endif
}

void sigvreadsf_ff(t_sigvreadsf *x)
{
#ifdef HOSTSTARTSERVER
    t_server *foo = &rsf_servers[x->r_serverno - 1];
    if ((!--foo->s_count) && foo->s_fd >= 0)
        io_closeserver(foo->s_fd);
#endif
    u_clean((t_ugen *) x);
#ifdef FTSSOUNDDISK
    afd_free(x->r_ctl->c_afd);
#endif
}

void *sigvreadsf_class;
void sigvreadsf_setup()
{
    c_new(&s__sigvreadsf, sigvreadsf_ff, sizeof(t_sigvreadsf), 0, 0);
    c_addmess(sigvreadsf_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    c_addmess(sigvreadsf_open, &s_open, A_DEFSYM, A_DEFLONG, A_DEFSYM, 0);
    c_addfloat(sigvreadsf_float);
    c_addint(sigvreadsf_int);
    c_addmess(sigvreadsf_print, &s_print, 0);
    c_addmess(sigvreadsf_stop, gensym("stop"), 0);
}

void *sigvreadsf_new(t_symbol *sampname, long int nchans,
                     long int serverno, long int bufsize,
                     long int bidirectional)
{
    t_sigvreadsf *x;
#ifdef FTSSOUNDDISK
    AFDSTORAGECLASS t_asyncfd *a;
#endif
    long minbufsize = (bidirectional ? 2 * RSF_BUFSIZE : RSF_BUFSIZE);
    if (nchans < 1)
        nchans = 1;
    else if (nchans > RSF_MAXCHANS)
        nchans = RSF_MAXCHANS;
    if (serverno > RSF_MAXSERVER)
        serverno = RSF_MAXSERVER;
    else if (serverno < 1)
        serverno = 1;
#ifdef FTSSOUNDDISK
    if (bufsize < minbufsize)
        bufsize = minbufsize;
    bufsize &= ~0xfff;
    bufsize *= nchans;
    if (!(a = afd_new((int) nchans, (int) serverno, (int) bidirectional,
                      bufsize)))
    {
        error("vreadsf~: error opening file handler");
        return (0);
    }
#else
    if (!h_ishost)
        post("vreadsf~: standalone only");
#endif
    x = (t_sigvreadsf *) obj_new((t_messlist **) & sigvreadsf_class,
                                 sigvreadsf_setup);
    u_setup((t_ugen *) x, IN1, OUT0);
#ifdef HOSTSTARTSERVER
    if (whichFts() != standaloneFts)
        post("vreadsf~: standalone only");
    else if (!rsf_servers[serverno - 1].s_count)
        rsf_openserver(serverno);
    rsf_servers[serverno - 1].s_count++;
    x->r_serverno = serverno;
#endif

#ifdef FTSSOUNDDISK
    x->r_cspace.c_afd = a;
    x->r_sampname = sampname;
    x->r_cspace.c_nchans = nchans;
    x->r_cspace.c_buf = a->afd_buf;
    x->r_cspace.c_state = S_INACTIVE;
    x->r_cspace.c_bufsize = bufsize;
    x->r_cspace.c_bidirectional = bidirectional;
    x->r_ctl = &x->r_cspace;
    x->r_vol = defvolume;
#endif
    x->r_outlet[4] = outlet_new(x, &s_bang);
    x->r_outlet[3] = outlet_new(x, &s_float);
    x->r_outlet[2] = outlet_new(x, &s_float);
    x->r_outlet[1] = outlet_new(x, &s_float);
    x->r_outlet[0] = outlet_new(x, &s_float);
    return (x);
}
