
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.4 $ IRCAM $Date: 1994/01/04 17:55:00 $
 *
 * FTS by Miller Puckette
 * Structures for asynchronous disk I/O for real-time sound.
 * Shared between CP and host.
 * Formerly in proto56.h
 *
 */

/*
 * This file contains both the "afd" structure which supports asynchronous
 * disk I/O (which is used by d_disk.c and by the NeXTStep disk server)
 * and also the "proto" structure that controls all communication from
 * NeXT to ISPW.  This second wsa originally in ../6a and should be put
 * back there.
 */

typedef struct afdhead          /* separate struct for fast copy */
{
    char h_serverno;            /* who is serving this "afd" (1-10 for ISPW) */
    char h_openrequest;         /* request from CP, from defs below */
    char h_closerequest;        /* request from CP to close */
    char h_eof;                 /* EOF or error */
} t_afdhead;

#define F_MSEC 0
#define F_BYTE 1

typedef struct asyncfd {
    t_afdhead afd_h;            /* above */
    char *afd_buf;              /* buffer in CP mem */
    long afd_size;              /* buffer size in bytes */
    long afd_hostonset;         /* index of next char host will write or read */
    long afd_cponset;           /* index of next char CP will write or read */
    long afd_hostbackonset;     /* backward onset for bidirectional reading */
    char *afd_filename;         /* filename for open request */
    long afd_sr;                /* sample rate of file (should be float really) */
    long afd_nchans;            /* number of channels */
    long afd_openresult;        /* result of open comand */
    long afd_onset;             /* onset into file in msec or bytes */
    char afd_bidirectional;     /* true if allowed to go forward and backward */
    char afd_onsetformat;       /* from F_MSEC or F_BYTE */
    char afd_pad[2];
#ifdef __unix
    int afd_fd;                 /* pipe to server; used only to close at end */
    struct asyncfd *afd_self;   /* pointer to beginning */
#endif
} t_asyncfd;

#define AFD_NAMESIZE 512        /* max stringlength of filename */
#define AFDREQ_DONE 0           /* defs for afd_request above */
#define AFDREQ_OPENTOREAD 1     /* open a soundfile for reading */
#define AFDREQ_OPENTOWRITE  2   /* open a soundfile for writing */

#if defined (NEXTSTEP) || defined (I860)
/*
 * this is the part that should live in ../6a/proto.h!
 */
#define MAXSLOTNO 8
#define MAXAFD  8               /* max. number of asyncfd */

/* controls for fifo from another CP */
typedef struct ftscpu {
    long cpu_latency;           /* 0 if not present, otherwise latency of other CPU */
    long cpu_head;              /* head offset */
    long cpu_tail;              /* tail offset */
} t_ftscpu;

typedef struct proto {
    char p_running;             /* program sets this at start */
    char pad1[3];
    char p_spinning;            /* program sets this to tell boot to exit */
    char pad2[3];
    long p_inhead;              /* fifo head for CP input (offset into FIFO) */
    long p_intail;              /* fifo tail for CP input */
    long p_outhead;             /* and for CP output */
    long p_outtail;
    long p_checksum;            /* slot for boot PROM to put checksum */
    long p_psr;                 /* snapshots of registers -- psr */
    long p_epsr;
    long p_dirbase;
    long p_regs[32];            /* integer registers */
    long p_fir;                 /* trap address */
    long p_fsr;                 /* floating-point status register */
    long p_fregs[32];           /* floating-point registers */
    char p_iorequest;           /* request to io.c to do open, read, etc. */
    char p_pad3[3];
    long p_ioarg1;              /* args for request */
    long p_ioarg2;
    long p_ioarg3;
    long p_ioresult;            /* return value from request */
    char p_lockrequest;
    char p_locked;
    char p_pad4[2];
    long p_ftsslot;             /* 2 = slot 2 / proc a; 3 = slot 2 / proc b, etc */
    t_ftscpu p_ftscpu[MAXSLOTNO];
    long p_ftsflags;            /* see FTS_ defs above */
    long p_animal;              /* turned on if Animal has been run */
    long p_inhead_animal;       /* in and out head and tail for Animal */
    long p_intail_animal;
    long p_outhead_animal;
    long p_outtail_animal;
    long p_srate;               /* DAC sample rate */
    t_asyncfd p_asyncfd[MAXAFD];        /* soundfile i/o; defs above */
    int p_dactype;              /* dac type from converterType.h */
} t_proto;
#endif

#ifdef __unix

/*
 * on the SGI there only appears to be one FTS process and only
 * one possible AFD for it; that's because each readsf~,  etc.,  spawns
 * its own server.  THe shared memory region between FTS and max-sounddisk
 * consistss of a "t_proto" structure followed by a disk buffer of a
 * size determined by FTS.  Here are the onsets into the shared memory
 * buffer we'll use for name and sound:
 */

#define NAMEONSET (sizeof(t_asyncfd))
#define SOUNDONSET (sizeof(t_asyncfd) + AFD_NAMESIZE)

#endif
