#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:18 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/05/10 10:50:18 $
 *
 * FTS by Miller Puckette
 *
 */


/* definitions of fft~, ifft~.  */

#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__

char *getpage(int size, int cacheable);

extern t_symbol s_put, s__sigfft, s__sigifft;
extern int dsp_ticks;

/******************* fft *************************/

typedef struct fftctl {
    int c_vecsize;              /* input/output vector size */
    int c_countdown;            /* number of vectors to wait before collecting input */
    float *c_fillbuf;           /* next element of buf to fill from input */
    float *c_emptybuf;          /* next element of buf to output */
    float *c_endbuf;            /* end of buffer */
    struct sigfft *c_backptr;   /* pointer to owner */
} t_fftctl;

typedef struct sigfft {
    t_head x_h;
    t_sig *x_io[IN2 + OUT2];
    t_fftctl *x_ctl;            /* pointer to control block */
    t_fftctl x_cspace;          /* control block if not in a chain */
    float *x_buf;               /* input/output buffer */
    int x_npoints;              /* number of complex points in buffer */
    int x_phase;                /* phase of FFT duty cycle relative to dsp_ticks */
    int x_period;               /* period in samples of FFT (at least npoints) */
    int x_inverse;              /* inverse flag */
    void *x_clock;
    void *x_bangout;
} t_sigfft;

#define COEFTABPOINTS 512
#define COEFTABSIZE (COEFTABPOINTS * 2 * sizeof(float))
static float *coef, *icoef;
#define FFTMINPOINTS  16
#define FFTMAXPOINTS  1024

fft_init()
{
#ifdef I860
    int i;
    double phase = 0, phsinc;
    float *fp1, *fp2;
    if (coef)
        return (0);
    if (!(coef = (float *) getpage(COEFTABSIZE, 1)))
        return (-1);
    if (!(icoef = (float *) getpage(COEFTABSIZE, 1)))
    {
        ungetpage(coef, COEFTABSIZE);
        return (-1);
    }
    fp1 = coef;
    fp2 = icoef;
    phsinc = 3.14159265 / COEFTABPOINTS;
    for (i = COEFTABPOINTS; i--;)
    {
        fp1[0] = fp2[0] = cos(phase);
        fp1[1] = -(fp2[1] = sin(phase));
        phase += phsinc;
        fp1 += 2;
        fp2 += 2;
    }
#endif
    return (0);
}

#ifndef I860
#include "d_cfft.h"

void cu_mux(float *in, float *out, int step, int n)
{
    while (n--)
    {
        *out = *in++;
        out += step;
    }
}

#endif

fft(float *buf, int npoints, int inverse)
{
#ifdef I860
    if (!coef)
        error("fft: not initialized");
    else
        sfft(npoints, buf, inverse ? icoef : coef);
#else
    double renorm;
    float *fp, *fp2;
    int i;
    /* post("fft %x %d",  buf,  npoints); */
    renorm =
        /* sqrt((inverse ? (double)npoints : 1./(double)npoints)); */
        (inverse ? npoints : 1.);
    cfft((inverse ? INVERSE : FORWARD), npoints, RECTANGULAR,
         buf, RECT, LINEAR, buf, RECT, LINEAR, 0);
    for (i = npoints << 1, fp = buf; i--; fp++)
        *fp *= renorm;
#endif
}

ilog2(int n)
{
    int ret = -1;
    while (n)
    {
        n >>= 1;
        ret++;
    }
    return (ret);
}

void cu_fft(float *in1, float *in2, float *out1, float *out2,
            t_fftctl *ctl)
{
    float *bptr, *ebuf = ctl->c_endbuf;
    int n = ctl->c_vecsize;
    if ((bptr = ctl->c_emptybuf) != ebuf)
    {
        cu_down(bptr - 1, out1, n);
        cu_down(bptr, out2, n);
        ctl->c_emptybuf = bptr + (n << 1);
    }
    if (ctl->c_countdown)
        ctl->c_countdown--;
    else
    {
        bptr = ctl->c_fillbuf;
        cu_mux(in1, bptr, 2, n);
        cu_mux(in2, bptr + 1, 2, n);
        bptr += (n << 1);
        if (bptr == ebuf)
        {
            t_sigfft *x = ctl->c_backptr;
            fft(x->x_buf, x->x_npoints, x->x_inverse);
            ctl->c_fillbuf = ctl->c_emptybuf = x->x_buf;
            ctl->c_countdown =
                (x->x_period - x->x_npoints) / ctl->c_vecsize;
            clock_delay(x->x_clock, 0L);
        }
        else
            ctl->c_fillbuf = bptr;
    }
}

void sigfft_put(t_sigfft *x, long int whether)
{
    if (whether)
    {
        int vecsize = x->x_io[0]->s_n, phase, ticksperperiod;
        void *s1, *s2, *sig_getnew(int size);
        /* temporarily get two signals so that calculation can't
           be in-place */
        if (x->x_period == x->x_npoints)
        {
            s1 = sig_getnew(x->x_io[0]->s_n);
            s2 = sig_getnew(x->x_io[0]->s_n);
        }
        u_stdout((t_ugen *) x);
        if (x->x_period == x->x_npoints)
        {
            sig_reusable(s1);
            sig_reusable(s2);
        }
        if (vecsize != (1 << ilog2(vecsize)))
        {
            error("fft: power-of-two vector size only");
            return;
        }
        else if (vecsize > x->x_npoints)
        {
            error("fft: vector size larger than FFT size");
            return;
        }
        /* fix me -- this will break if tick rate of fft differs
           from global tick rate!!! */
        phase = (dsp_ticks * vecsize - x->x_phase) % x->x_period;
        x->x_cspace.c_vecsize = x->x_io[0]->s_n;
        x->x_cspace.c_emptybuf = (phase < x->x_npoints ?
                                  x->x_buf +
                                  (phase << 1) : x->x_cspace.c_endbuf);
        if (x->x_npoints + phase >= x->x_period)
        {
            x->x_cspace.c_countdown = 0;
            x->x_cspace.c_fillbuf = x->x_buf +
                ((x->x_npoints + phase - x->x_period) << 1);
        }
        else
        {
            x->x_cspace.c_countdown =
                (x->x_period - phase - x->x_npoints) / vecsize;
            x->x_cspace.c_fillbuf = x->x_buf;
        }
        x->x_cspace.c_vecsize = vecsize;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_fftctl), sizeof(float));
        dspchain_putc(u_chain, cu_fft, 5,
                      x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                      x->x_io[2]->s_shit, x->x_io[3]->s_shit, x->x_ctl);
        if (u_debug)
            post("sigfft %lx: (%lx,%lx)->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                 x->x_io[2]->s_shit);
        /* post("vecsize %x; empty %x",
           x->x_cspace.c_vecsize, x->x_cspace.c_emptybuf);
           post("fill %x; countdown %d",
           x->x_cspace.c_fillbuf, x->x_cspace.c_countdown); */
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_fftctl));
        x->x_ctl = &x->x_cspace;
    }
}

void fft_tick(t_sigfft *x)
{
    outlet_bang(x->x_bangout);
}

void sigfft_bang(t_sigfft *x)
{
    if (!x->x_ctl->c_vecsize)
        return;
    x->x_ctl->c_countdown =
        (x->x_period - x->x_npoints) / x->x_ctl->c_vecsize;
    x->x_ctl->c_fillbuf = x->x_buf;
}

void sigfft_ff(t_sigfft *x)
{
    ungetpage(x->x_buf, (x->x_npoints == 1024 ? 8192 : 4096));
    u_clean((t_ugen *) x);
    clock_free(x->x_clock);
}

void *sigfft_class;
void sigfft_setup()
{
    c_new(&s__sigfft, sigfft_ff, sizeof(t_sigfft), 0, 0);
    c_addmess(sigfft_put, &s_put, A_CANT, 0);
    c_addbang(sigfft_bang);
    u_inletmethod(0);
    u_inletmethod(1);
}

void *sigfft_donew(long int npoints, long int period, long int phase,
                   long int inverse)
{
    t_sigfft *x;
    float *buf;
    int bad = 0;
    if (!npoints)
        npoints = 512;
    if (!period)
        period = npoints;
    if (fft_init())
    {
        error("fft: can't init");
        return (0);
    }
    if (npoints != (1 << ilog2(npoints)) || period != (1 << ilog2(period)))
        error("fft: must use a power of two for size and period"), bad = 1;
    if (npoints < FFTMINPOINTS)
        error("fft: needs at least %d points", FFTMINPOINTS), bad = 1;
    if (npoints > FFTMAXPOINTS)
        error("fft: no more than %d points allowed", FFTMAXPOINTS), bad =
            1;
    if (npoints > period)
        error("fft: period must be at least one frame"), bad = 1;
    phase %= period;
    if (phase % MAXVS)
        error("fft: phase must be multiple of %d", MAXVS), bad = 1;
    if (npoints == FFTMAXPOINTS)
        buf = (float *) getpage(2 * sizeof(float) * FFTMAXPOINTS, 0);
    else
        buf = (float *) getpage(sizeof(float) * FFTMAXPOINTS, 0);
    if (!buf)
        error("fft: can't allocate buffer"), bad = 1;
    if (bad)
        return (0);
    x = (t_sigfft *) obj_new((t_messlist **) & sigfft_class, sigfft_setup);
    x->x_bangout = outlet_new(x, &s_bang);
    u_setup((t_ugen *) x, IN2, OUT2);
    x->x_buf = buf;
    x->x_npoints = npoints;
    x->x_period = period;
    x->x_phase = phase;
    x->x_inverse = inverse;
    x->x_cspace.c_endbuf = x->x_buf + (npoints << 1);
    x->x_cspace.c_backptr = x;
    x->x_ctl = &x->x_cspace;
    x->x_clock = clock_new(x, fft_tick);
    if (u_debug)
        post("sigfft %lx", x);
    return (x);
}

void *sigfft_new(long int npoints, long int period, long int phase)
{
    return (sigfft_donew(npoints, period, phase, 0L));
}

void *sigifft_new(long int npoints, long int period, long int phase)
{
    return (sigfft_donew(npoints, period, phase, 1L));
}
