#ifndef lint
static char rcsid[] =
    "$Revision: 2.8 $ IRCAM $Date: 1994/05/10 10:50:27 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/05/10 10:50:27 $
 *
 * FTS by Miller Puckette
 *
 */
#include  <stdio.h>

/* sample and hold; 2pole "sweep" filter; clip */

#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"

extern t_symbol s_put, s_reset, s_set,
    s__sigsamphold, s__sig2pole, s__sig2p2z, s__sigwahwah, s__sigclip,
    s_ft1, s_ft2, s_ft3, s_ft4, s_ft5, s_ft6, s_ft7;

#ifdef I860
void su_samphold(), cu_f2p(), su_wahwah(), su_clip();
#endif
extern t_sig *sig_zero;

/******************* sample/hold *************************/

/* When input 1 decreases, gate input 2.  */

typedef struct sampholdctl {
    float c_val;
    float c_last;
} t_sampholdctl;

typedef struct sigsamphold {
    t_head x_h;
    t_sig *x_io[IN2 + OUT1];
    t_sampholdctl *x_ctl;
    t_sampholdctl x_cspace;
} t_sigsamphold;

/******************* 2pole wahwah filter *************************/
typedef struct twopole {
    float c_ym1;                /* last output value of filt */
    float c_ym2;                /* second - to - last */
    float c_gain;               /* input gain */
    float c_coef1;              /* first feedback coefficient */
    float c_coef2;              /* second feedback coefficient */
} t_twopole;

typedef struct f1ctl {
    t_twopole c_2p;
    float c_fconv;              /* hertz to angular sample inc conversion */
    long c_n;                   /* vector size */
} t_f1ctl;

typedef struct sigwahwah {
    t_head x_h;
    t_sig *x_io[IN3 + OUT1];
    t_f1ctl *x_ctl;
    t_f1ctl x_cspace;
    int x_old;                  /* compatibility with old version */
} t_sigwahwah;

/******************* 2pole, 2zero raw filter *************************/
typedef struct sigf2p2z {
    t_head x_h;
    t_sig *x_io[IN2 + OUT1];
    t_ctlf2p2z *x_ctl;
    t_ctlf2p2z x_cspace;
} t_sigf2p2z;

/******************* clip *************************/
typedef struct clipctl {
    float c_min;
    float c_max;
} t_clipctl;

typedef struct sigclip {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_clipctl *x_ctl;
    t_clipctl x_cspace;
} t_sigclip;

void cu_samphold(float *in1, float *in2, float *out,
                 t_sampholdctl * samphold, long int n);
void cu_do2pole(float *in, float *out, int n, t_twopole * x);
void cu_wahwah(float *in, float *cfreq, float *q, float *out, t_f1ctl * x);
void cu_clip(float *in, float *out, t_clipctl * x, long int n);

#ifndef I860

void cu_samphold(float *in1, float *in2, float *out,
                 t_sampholdctl *samphold, long int n)
{
    float last = samphold->c_last;
    float val = samphold->c_val;
    float next;
    while (n--)
    {
        next = *(in1++);
        if (next < last)
            val = *in2;
        in2++;
        last = next;
        *out++ = val;
    }
    samphold->c_last = last;
    samphold->c_val = val;
}
#endif

void sigsamphold_put(t_sigsamphold *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_sampholdctl), sizeof(float));
#ifdef I860
        dspchain_put860(u_chain, su_samphold, 5,
                        x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                        x->x_io[2]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#else
        dspchain_putc(u_chain, cu_samphold, 5,
                      x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                      x->x_io[2]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#endif
        if (u_debug)
            post("sigsamphold %lx: (%lx,%lx)->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                 x->x_io[2]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_sampholdctl));
        x->x_ctl = &x->x_cspace;
    }
}

void sigsamphold_float(t_sigsamphold *x, double f)
{
    x->x_ctl->c_val = f;
}

void sigsamphold_int(t_sigsamphold *x, long int n)
{
    x->x_ctl->c_val = n;
}

void sigsamphold_reset(t_sigsamphold *x)
{
    x->x_ctl->c_last = 1.e30;
}

void *sigsamphold_class;
void sigsamphold_setup()
{
    c_new(&s__sigsamphold, u_clean, sizeof(t_sigsamphold), 0, 0);
    c_addmess(sigsamphold_put, &s_put, A_CANT, 0);
    c_addfloat(sigsamphold_float);
    c_addint(sigsamphold_int);
    c_addmess(sigsamphold_reset, &s_reset, 0);
    u_inletmethod(0);
    u_inletmethod(1);
}

void *sigsamphold_new()
{
    t_sigsamphold *x =
        (t_sigsamphold *) obj_new((t_messlist **) & sigsamphold_class,
                                  sigsamphold_setup);
    u_setup((t_ugen *) x, IN2, OUT1);
    x->x_cspace.c_last = 1.e30;
    x->x_cspace.c_val = 0;
    x->x_ctl = &x->x_cspace;
    if (u_debug)
        post("sigsamphold %lx", x);
    return (x);
}

#ifndef I860
void cu_do2pole(float *in, float *out, int n, t_twopole *x)
{
    float ym1 = x->c_ym1, ym2 = x->c_ym2, tmp;
    float a = x->c_coef1, b = x->c_coef2, g = x->c_gain;
    while (n--)
    {
        *out++ = g * (tmp = a * ym1 + b * ym2 + *in++ + 1E-20);
        ym2 = ym1;
        ym1 = tmp;
    }
    if (BADFLOAT(ym1))
        ym1 = 0;
    if (BADFLOAT(ym2))
        ym2 = 0;
    x->c_ym1 = ym1;
    x->c_ym2 = ym2;
}
#endif

void cu_oldwahwah(float *in, float *cfreq, float *rad, float *out, t_f1ctl *x)  /* old version */
{
    float angle = x->c_fconv * *cfreq;
    float cos_omega = 1 - .5 * angle * angle;
    float sin_omega = angle;
    float tmp = *rad, r = (tmp < 0. ? 0. : (tmp > 1. ? 1. : tmp));
    if (cos_omega < 0)
        cos_omega = 0;
    if (sin_omega < 0)
        sin_omega = 0;
    else if (sin_omega > 1)
        sin_omega = 1;
    x->c_2p.c_gain = (1. - r) * ((1. - r) + r * sin_omega);
    x->c_2p.c_coef1 = 2 * r * cos_omega;
    x->c_2p.c_coef2 = -r * r;
    cu_do2pole(in, out, x->c_n, (t_twopole *) x);
}

#ifndef I860
void cu_wahwah(float *in, float *cfreq, float *q, float *out, t_f1ctl *x)
{
    float theta = x->c_fconv * *cfreq, cos_theta, sin_theta, tmp, r;
    if (theta > 2.)
        theta = 2.;
    else if (theta < 0)
        theta = 0.;
    cos_theta = 1. - .5 * theta * theta;
    sin_theta = theta;
    if (sin_theta > 1.)
        sin_theta = 1.;
    tmp = *q;
    if (tmp <= .00001)
        r = 0.;
    else if ((r = 1. - theta / tmp) < 0)
        r = 0.;
    x->c_2p.c_gain = (1. - r) * ((1. - r) + r * sin_theta);
    x->c_2p.c_coef1 = 2. * r * cos_theta;
    x->c_2p.c_coef2 = -r * r;
    cu_do2pole(in, out, x->c_n, (t_twopole *) x);
}
#endif                          /* I860 */

void sigwahwah_put(t_sigwahwah *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_cspace.c_n = x->x_io[3]->s_n;
        x->x_cspace.c_fconv = (2. * 3.14159265) / dsp_sr;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_f1ctl), sizeof(float));
        if (x->x_old)
            dspchain_putc(u_chain, cu_oldwahwah, 5,
                          x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                          x->x_io[2]->s_shit, x->x_io[3]->s_shit,
                          x->x_ctl);
        else
            dspchain_putc(u_chain, cu_wahwah, 5,
                          x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                          x->x_io[2]->s_shit, x->x_io[3]->s_shit,
                          x->x_ctl);
        if (u_debug)
            post("sigwahwah %lx: (%lx,%lx,%lx)->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                 x->x_io[2]->s_shit, x->x_io[3]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_f1ctl));
        x->x_ctl = &x->x_cspace;
    }
}

void *sigwahwah_class;
void sigwahwah_setup()
{
    c_new(&s__sigwahwah, u_clean, sizeof(t_sigwahwah), 0, 0);
    c_addmess(sigwahwah_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    u_inletmethod(1);
    u_inletmethod(2);
}

void *sigwahwah_new(long int old)
{
    t_sigwahwah *x =
        (t_sigwahwah *) obj_new((t_messlist **) & sigwahwah_class,
                                sigwahwah_setup);
    u_setup((t_ugen *) x, IN3, OUT1);
    x->x_cspace.c_2p.c_ym1 = 0;
    x->x_cspace.c_2p.c_ym2 = 0;
    x->x_cspace.c_2p.c_gain = 0;
    x->x_cspace.c_n = DEFAULTVS;
    x->x_cspace.c_fconv = (2. * 3.14159265) / dsp_sr;
    x->x_ctl = &x->x_cspace;
    x->x_old = old;
    if (u_debug)
        post("sigwahwah %lx", x);
    return (x);
}

#ifndef I860
void cu_f2p2z(float *in, float *out, t_ctlf2p2z *x, int n)
{
    float ym1 = x->c_state1, ym2 = x->c_state2, ym0;
    float a = x->c_fbcoef1, b = x->c_fbcoef2, g = x->c_gain;
    float ff0 = x->c_ffcoef0, ff1 = x->c_ffcoef1, ff2 = x->c_ffcoef2;
    while (n--)
    {
        ym0 = a * ym1 + b * ym2 + g * *in++ + 1E-20;
        *out++ = ff0 * ym0 + ff1 * ym1 + ff2 * ym2;
        ym2 = ym1;
        ym1 = ym0;
    }
    if (BADFLOAT(ym1))
        ym1 = 0;
    if (BADFLOAT(ym2))
        ym2 = 0;
    x->c_state1 = ym1;
    x->c_state2 = ym2;
}
#endif

void sigf2p2z_put(t_sigf2p2z *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_ctlf2p2z), sizeof(float));
        dspchain_putc(u_chain, cu_f2p2z, 4,
                      x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                      x->x_ctl, x->x_io[0]->s_n);
        if (u_debug)
            post("sigf2p2z %lx: %lx,%lx->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                 x->x_io[2]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_ctlf2p2z));
        x->x_ctl = &x->x_cspace;
    }
}

void sigf2p2z_ft1(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_gain = f;
}

void sigf2p2z_ft2(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_fbcoef1 = f;
}

void sigf2p2z_ft3(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_fbcoef2 = f;
}

void sigf2p2z_ft4(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_ffcoef0 = f;
}

void sigf2p2z_ft5(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_ffcoef1 = f;
}

void sigf2p2z_ft6(t_sigf2p2z *x, double f)
{
    x->x_ctl->c_ffcoef2 = f;
}

float atom_getfloat(t_atom * at);

void sigf2p2z_set(t_sigf2p2z *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ctl->c_gain = (ac >= 1 ? atom_getfloat(av) : 0);
    x->x_ctl->c_fbcoef1 = (ac >= 2 ? atom_getfloat(av + 1) : 0);
    x->x_ctl->c_fbcoef2 = (ac >= 3 ? atom_getfloat(av + 2) : 0);
    x->x_ctl->c_ffcoef0 = (ac >= 4 ? atom_getfloat(av + 3) : 0);
    x->x_ctl->c_ffcoef1 = (ac >= 5 ? atom_getfloat(av + 4) : 0);
    x->x_ctl->c_ffcoef2 = (ac >= 6 ? atom_getfloat(av + 5) : 0);
}

void *sigf2p2z_class;
void sigf2p2z_setup()
{
    c_new(&s__sig2p2z, u_clean, sizeof(t_sigf2p2z), 0, 0);
    c_addmess(sigf2p2z_put, &s_put, A_CANT, 0);
    c_addmess(sigf2p2z_set, &s_set, A_GIMME, 0);
    u_inletmethod(0);
    u_inletmethod(1);
    c_addftx(sigf2p2z_ft1, 1);
    c_addftx(sigf2p2z_ft2, 2);
    c_addftx(sigf2p2z_ft3, 3);
    c_addftx(sigf2p2z_ft4, 4);
    c_addftx(sigf2p2z_ft5, 5);
    c_addftx(sigf2p2z_ft6, 6);
}

void *sig2p2z_new(t_symbol *s, int ac, t_atom *av)
{
    t_sigf2p2z *x =
        (t_sigf2p2z *) obj_new((t_messlist **) & sigf2p2z_class,
                               sigf2p2z_setup);
    inlet(x, x, &s_float, &s_ft6);
    inlet(x, x, &s_float, &s_ft5);
    inlet(x, x, &s_float, &s_ft4);
    inlet(x, x, &s_float, &s_ft3);
    inlet(x, x, &s_float, &s_ft2);
    inlet(x, x, &s_float, &s_ft1);
    u_setup((t_ugen *) x, IN1, OUT1);
    x->x_ctl = &x->x_cspace;
    sigf2p2z_set(x, NIL, ac, av);
    x->x_cspace.c_state1 = 0;
    x->x_cspace.c_state2 = 0;
    if (u_debug)
        post("sig2p2z %lx", x);
    return (x);
}

/*************** 2-pole filter, which just reuses the above ******/

void sigf2p_set(t_sigf2p2z *x, double f1, double f2, double f3)
{
    x->x_ctl->c_gain = f1;
    x->x_ctl->c_fbcoef1 = f2;
    x->x_ctl->c_fbcoef2 = f3;
}

void *sigf2p_class;
sigf2p_setup()
{
    c_new(&s__sig2pole, u_clean, sizeof(t_sigf2p2z), 0, 0);
    c_addmess(sigf2p2z_put, &s_put, A_CANT, 0);
    c_addmess(sigf2p_set, &s_set, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    u_inletmethod(0);
    c_addftx(sigf2p2z_ft1, 1);
    c_addftx(sigf2p2z_ft2, 2);
    c_addftx(sigf2p2z_ft3, 3);
}

void *sigf2p_new(double g, double c1, double c2)
{
    t_sigf2p2z *x =
        (t_sigf2p2z *) obj_new((t_messlist **) & sigf2p2z_class,
                               sigf2p2z_setup);
    inlet(x, x, &s_float, &s_ft3);
    inlet(x, x, &s_float, &s_ft2);
    inlet(x, x, &s_float, &s_ft1);
    u_setup((t_ugen *) x, IN1, OUT1);
    x->x_ctl = &x->x_cspace;
    sigf2p_set(x, g, c1, c2);
    x->x_cspace.c_ffcoef0 = 1.;
    x->x_cspace.c_ffcoef1 = x->x_cspace.c_ffcoef2 = 0.;
    if (u_debug)
        post("f2p %lx", x);
    return (x);
}

/* clip to the range [min, max]  */

#ifndef I860
void cu_clip(float *in, float *out, t_clipctl *x, long int n)
{
    float min = x->c_min, max = x->c_max, f;
    while (n--)
    {
        f = *(in++);
        if (f < min)
            f = min;
        else if (f > max)
            f = max;
        *out++ = f;
    }
}
#endif

void sigclip_put(t_sigclip *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_clipctl), sizeof(float));
#ifdef I860
        dspchain_put860(u_chain, su_clip, 4,
                        x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                        x->x_ctl, x->x_io[0]->s_n);
#else
        dspchain_putc(u_chain, cu_clip, 4,
                      x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                      x->x_ctl, x->x_io[0]->s_n);
#endif
        if (u_debug)
            post("sigclip %lx: (%lx)->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_clipctl));
        x->x_ctl = &x->x_cspace;
    }
}

void sigclip_ft1(t_sigclip *x, double f)
{
    x->x_ctl->c_min = f;
}

void sigclip_ft2(t_sigclip *x, double f)
{
    x->x_ctl->c_max = f;
}

void *sigclip_class;
void sigclip_setup()
{
    c_new(&s__sigclip, u_clean, sizeof(t_sigclip), 0, 0);
    c_addmess(sigclip_put, &s_put, A_CANT, 0);
    c_addftx(sigclip_ft1, 1);
    c_addftx(sigclip_ft2, 2);
    u_inletmethod(0);
}

void *sigclip_new(double min, double max)
{
    t_sigclip *x =
        (t_sigclip *) obj_new((t_messlist **) & sigclip_class,
                              sigclip_setup);
    u_setup((t_ugen *) x, IN1, OUT1);
    inlet_new(x, x, &s_float, &s_ft2);
    inlet_new(x, x, &s_float, &s_ft1);
    x->x_cspace.c_min = min;
    x->x_cspace.c_max = max;
    x->x_ctl = &x->x_cspace;
    if (u_debug)
        post("sigclip %lx", x);
    return (x);
}
