#ifndef lint
static char rcsid[] =
    "$Revision: 1.11 $ IRCAM $Date: 1994/01/29 07:47:33 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.11 $ IRCAM $Date: 1994/01/29 07:47:33 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
#include <stdio.h>
#include "m_inout.h"
#include "d_graph.h"
#include "d_ugen.h"

extern struct symbol s_sig, s_dspchain,
    s_stov, s_qoscil, s_writesf, s_tick,
    s_sig, s__sig, s_put, s_set, s_open, s_close,
    s__sigwritesf, s_clear, s_ft1, s_play;

static t_ugen *ulist;
static t_symbol *sig_syms[MAXIN];
int u_doneinit = 0, u_lock = 0;
void *u_chain;
int u_debug;
int u_abortflag;

static void *u_defaultchain;
void *sig_getnew(int size), *dspchain_new();
t_sig *sig_zero;

void u_init1();

int u_unsetall();

void u_init1()
{
    int i;
    char xx[4];
    u_doneinit = 1;
    sig_syms[0] = &s_sig;
    for (i = 1; i < MAXIN; i++)
    {
        sprintf(xx, "s%d", i);
        sig_syms[i] = gensym(xx);
    }
}

void u_setup(void *v, int nin, int nout)
{
    t_ugen *x = (t_ugen *) v;
    int i;
    t_sig **vp = x->u_io;
    x->u_h.h_next = ulist;
    ulist = x;
    x->u_h.h_count = 0;
    for (i = nin; i--;)
    {
        *(vp++) = 0;
        if (i)
            inlet(x, x, &s_sig, sig_syms[i]);
    }
    for (i = 0; i < nout; i++)
    {
        *(vp++) = 0;
        /* outlet(x, &s_sig); -- problem with m_inout.c here */
        outlet(x, NIL);
    }
    x->u_h.h_nin = nin;
    x->u_h.h_nout = nout;
    x->u_h.h_bazzed = 0;
    x->u_h.h_recall = 0;
    x->u_h.h_forcer = 0;

}

void u_setforcer(void *v)
{
    t_ugen *x = (t_ugen *) v;
    x->u_h.h_forcer = 1;
}

extern int cp_this;

void u_clean(void *v)
{
    t_ugen *x = (t_ugen *) v;
    t_ugen stub, *sp, *sp2, *u;
    stub.u_h.h_next = ulist;
    for (sp = &stub; sp2 = sp->u_h.h_next; sp = sp2)
        if (sp2 == x)
        {
            sp->u_h.h_next = sp2->u_h.h_next;
            break;
        }
    ulist = stub.u_h.h_next;
    /* clear the orchestra */
    u_unsetall();
}

#define COUNTINPUTS 1
#define COUNTOUTPUTS 2
#define DOIT 3

void u_abort()
{
    u_abortflag = 1;
}

/* find signals to hold the output of the module.  If no vector
   size is given, guess one from the inputs; if they are all zero,
   use the default.  Also checks that all input vector sizes are equal
   (except that zero can be used at any vector size).
   later this needs to be made more general since some ugen might
   actually want to deal with this itself.  The error this
   generates aborts the chain building by setting an abort
   flag. */

void u_genout(void *v, int vs)
{
    int i;
    int invs = -1;
    t_sig **iop;
    t_ugen *x = (t_ugen *) v;
    /* post("%s: ", ob_name(x));
       for (i = x->u_h.h_nin, iop = x->u_io; i--; iop++)
       {
       post("sig %x", *iop);
       if (*iop == sig_zero) post("zero");
       else post("%x: %d", *iop, (*iop)->s_n);
       } */
    for (i = x->u_h.h_nin, iop = x->u_io; i--; iop++)
        if (*iop != sig_zero)
        {
            if (invs < 0)
                invs = (*iop)->s_n;
            else if (invs != (*iop)->s_n)
            {
                error("%s: inputs don't match\n", ob_name(x));
                u_abort();
            }
        }
    if (invs < 0)
        invs = DEFAULTVS;
    if (vs < 0)
        vs = invs;
    for (i = x->u_h.h_nout, iop = x->u_io + x->u_h.h_nin; i--; iop++)
        *iop = sig_getnew(vs);
    for (i = x->u_h.h_nout, iop = x->u_io + x->u_h.h_nin; i--; iop++)
        if (*iop == sig_zero)
        {
            bug("sig_zero");
            *(char *) 0 = 0;
        }
    /* cheat and set zero signal to vs */
    sig_zero->s_n = invs;
    sig_zero->s_sr = dsp_sr / (double) (DEFAULTVS / vs);
}

void u_stdout(void *v)
{
    t_ugen *x = (t_ugen *) v;
    u_genout(x, -1);
}

/*
  int u_debugcount;
  
  u_dodeb()
  {
  -- put your favorite sanity test here --
  u_debugcount++;
  }
  */

void u_baz(t_ugen *x)
{
    int i;
    struct outlet *op;
    t_sig **iop;
    sig_zero->s_n = DEFAULTVS;
    for (i = x->u_h.h_nin, iop = x->u_io; i--; iop++)
        if (!*iop)
            *iop = sig_zero;
    mess1(x, &s_put, U_PUTON);
    x->u_h.h_bazzed = 1;
    for (i = x->u_h.h_nout, op = x->u_h.h_ob.o_outlet,
         iop = x->u_io + x->u_h.h_nin; i--; iop++, op = op->o_next)
    {
        vfun(op, outlet_anything, &s_sig,
             "ll", (long) COUNTOUTPUTS, (long) (*iop));
        if (!(*iop)->s_count)
            sig_reusable(*iop);
    }
    sig_reclaim();
    for (i = x->u_h.h_nout, op = x->u_h.h_ob.o_outlet,
         iop = x->u_io + x->u_h.h_nin; i--; iop++, op = op->o_next)
        vfun(op, outlet_anything, &s_sig,
             "ll", (long) DOIT, (long) (*iop));
    if (x->u_h.h_recall)
        mess1(x, &s_put, U_DONEPUTTINGON);
}

/* zcheck(n)  consistency check ...
   {
   if (!sig_zero) return;
   if (sig_zero->s_shit[2] != 0.)
   {
   error("zcheck %d", n);
   *(char *)0 = 0;
   }
   } */

void u_nsig(t_ugen *x, long int n, int ac, t_atom *av)
{
    int i;
    long phase;
    t_sig *sig;
    t_sig **iop;
    if (!u_lock)
        return;                 /*  protect from unauthorized calling */
    phase = av[0].a_w.w_long;
    sig = (t_sig *) (av[1].a_w.w_long);
    switch (phase)
    {
    case COUNTINPUTS:
        x->u_h.h_count++;
        break;
    case COUNTOUTPUTS:
        sig->s_count++;
        break;
    case DOIT:
        /* We really should add multiple inputs, as long as the input
           is not order-forced. Here we just release the old one and replace it */
        if (x->u_io[n])
        {
            if (!--(x->u_io[n]->s_count))
                sig_reusable(x->u_io[n]);
            if (!x->u_h.h_forcer)
                post("extra signal connected to %s input: ignored",
                     ob_name(x));
        }
        x->u_io[n] = sig;
        if (!--(x->u_h.h_count))
        {
            for (i = x->u_h.h_nin, iop = x->u_io; i--; iop++)
                if (*iop)
                {
                    if (!--((*iop)->s_count))
                        sig_reusable(*iop);
                }
            u_baz(x);
        }
        break;
    }
}


void u_sig(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 0L, ac, av);
}

void u_sg1(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 1L, ac, av);
}

void u_sg2(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 2L, ac, av);
}

void u_sg3(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 3L, ac, av);
}

void u_sg4(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 4L, ac, av);
}

void u_sg5(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 5L, ac, av);
}

void u_sg6(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 6L, ac, av);
}

void u_sg7(t_ugen *x, t_symbol *s, int ac, t_atom *av)
{
    u_nsig(x, 7L, ac, av);
}


void u_siginlet(void *x, int n)
{
    if (!u_doneinit)
        u_init1();
    if (n <= 0 || n >= MAXIN)
        bug("siginlet out of range");
    inlet(x, x, &s_sig, sig_syms[n]);
}

static t_method u_meths[] =
    { u_sig, u_sg1, u_sg2, u_sg3, u_sg4, u_sg5, u_sg6, u_sg7 };

void u_inletmethod(int n)
{
    if (!u_doneinit)
        u_init1();
    if (n < 0 || n >= MAXIN)
        bug("inlet method out of range");
    c_addmess(u_meths[n], sig_syms[n], A_GIMME, 0);
}

void u_unset()
{
    u_chain = 0;
    if (sig_zero)
        sig_reusable(sig_zero), sig_zero = 0, sig_reclaim();
}

void u_set(long int vecsize)
{
    int i, loop = 0;
    t_ugen *u;
    t_sig **iop;

    /* 1. free the current chain and all the signals, and create
       a new dspchain.  All the signals should have already been marked
       "reusable" by the last call to this routine, except the zero
       signal which is marked reusable in u_unset. */
    u_abortflag = 0;
    u_unset();
    sig_setup(vecsize);
    sig_zero = sig_getnew(DEFAULTVS);
    cu_zero(sig_zero->s_shit, DEFAULTVS);
    if (!u_defaultchain)
        u_defaultchain = dspchain_new();
    u_chain = u_defaultchain;
    /* tell ugens they can take the (secure) "sig" messages */
    u_lock = 1;
    /* 2. clear all signal inputs */
    for (u = ulist; u; u = u->u_h.h_next)
    {
        for (i = u->u_h.h_nin, iop = u->u_io; i--; iop++)
            *iop = 0;
        u->u_h.h_count = 0;
    }
    /* 3. send "sig 0" to all outlets of all ugens.  Also warn the
       ugens that the dspchain is to be cleared, so that they must
       retrieve any state they maintain in the dspchain. */
    for (u = ulist; u; u = u->u_h.h_next)
    {
        struct outlet *op;
        for (i = u->u_h.h_nout, op = u->u_h.h_ob.o_outlet;
             i--; op = op->o_next)
            vfun(op, outlet_anything, &s_sig,
                 "ll", (long) COUNTINPUTS, 0L);
        if (u->u_h.h_bazzed)
        {
            mess1(u, &s_put, U_TAKEOFF);
            u->u_h.h_bazzed = 0;
        }
    }
    /* Now you can blow away the old dspchain call list */
    dspchain_begin(u_defaultchain);
    /* 4. baz all ugens with no sig inlets, or whose sig inlets
       are all disconnected, hence ready */
    for (u = ulist; u; u = u->u_h.h_next)
        if (!u->u_h.h_bazzed && !u->u_h.h_count)
            u_baz(u);
    /* 5. check that all rendezvous are satisfied (and clear them);
       be careful that u_io is really zeroed out beforehand */
    for (u = ulist; u; u = u->u_h.h_next)
        if (u->u_h.h_count)
        {
            loop = 1;
            u->u_h.h_count = 0;
        }
    dspchain_end(u_defaultchain);
    u_lock = 0;
    if (u_abortflag)
        u_unset();
    else if (loop)
        error("signal processing loop detected");
    sig_checkreusable();
}

u_check(int n, int loud)
{
    t_ugen *u;

    for (u = ulist; u; u = u->u_h.h_next)
    {
        if (u->u_h.h_ob.o_magic != MAGIC)
        {
            post("check %d: bad one at %x", n, u);
            *(char *) 0 = 0;
        }
    }
    if (loud)
        post("check %d: ok", n);
}

u_unsetall()
{
#ifndef GRAPHICS
    rtfifo_broadcast(u_unset, 0);
#endif
}

u_setall(void *zzz, int vecsize)
/* unused first arg so this can be a method of dac, etc */
{
#ifndef GRAPHICS
    if (!vecsize)
        vecsize = DEFAULTVS;
    rtfifo_broadcast((t_method) u_set, (long) vecsize);
    {
        /* a hook for the user to be informed that DSP has been turned on. */
        void *x = gensym("dsp-is-starting")->s_thing;
        if (x)
            obj_bang(x);
    }
#endif
}
