#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1993/11/19 16:22:04 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1993/11/19 16:22:04 $
 *
 * FTS by Miller Puckette
 *
 */


/* definitions of dac~, adc~, print~.  */
extern struct symbol
 s__sigadc, s__sigdac, s__sigprint,
    s_tick, s_put, s_start, s_stop, s_debug, s_;

#include <stdio.h>
#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__
#define MAXCHAN 8

void u_setall(void *zzz, int vecsize), u_unsetall();
/************** DAC ***************/
typedef struct sigdac {
    t_head x_h;
    t_sig *x_io[U_IN(MAXCHAN) + U_OUT(0)];
    int x_nchans;
    char x_assign[MAXCHAN];
} t_sigdac;

#ifndef I860
float soundout[MAXCHAN * MAXVS], soundin[MAXCHAN * MAXVS];
#else
extern float soundout[], soundin[];
#endif

void su_vadd();
void cu_vadd(float *in1, float *in2, float *out, long int n);

void sigdac_put(t_sigdac *x, long int whether)
{
    if (whether)
    {
        int i;
        u_stdout((t_ugen *) x);
        for (i = 0; i < x->x_nchans; i++)
        {
            float *buf = soundout + MAXVS * x->x_assign[i];
#ifdef I860
            dspchain_put860(u_chain, su_vadd, 4,
                            x->x_io[i]->s_shit, buf, buf, x->x_io[i]->s_n);
#else
            dspchain_putc(u_chain, cu_vadd, 4,
                          x->x_io[i]->s_shit, buf, buf, x->x_io[i]->s_n);
#endif                          /* I860 */
        }
        if (u_debug)
            post("sigdac %lx: (%lx,%lx)",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit);
    }
}

void sigdac_debug(t_sigdac *x, long int n)
{
    u_debug = n;
}

long atom_getlong(t_atom * at);

sigdac_parse(t_sigdac *x, int ac, t_atom *av)
{
    if (!ac)
    {
        x->x_nchans = 2;
        x->x_assign[0] = 0;
        x->x_assign[1] = 1;
        return;
    }
    if (ac > 8)
        error("dac/adc: 8 channels maximum (ignoring rest)"), ac = 8;
    for (x->x_nchans = 0; x->x_nchans < ac; x->x_nchans++)
    {
        int n;
        if ((n = atom_getlong(av + x->x_nchans) - 1) < 0 || n >= MAXCHAN)
        {
            error
                ("dac/adc: channel number out of range 1-8 (setting to 1)");
            n = 0;
        }
        x->x_assign[x->x_nchans] = n;
    }
}

void *sigdac_class;
void sigdac_setup()
{
    c_new(&s__sigdac, u_clean, sizeof(t_sigdac), 0, 0);
    c_addmess(sigdac_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    u_inletmethod(1);
    u_inletmethod(2);
    u_inletmethod(3);
    u_inletmethod(4);
    u_inletmethod(5);
    u_inletmethod(6);
    u_inletmethod(7);
    c_addmess(u_setall, &s_start, A_DEFLONG, 0);
    c_addmess(u_unsetall, &s_stop, 0);
    c_addmess(sigdac_debug, &s_debug, A_DEFLONG, 0);
}

void *sigdac_new(t_symbol *s, int ac, t_atom *av)
{
    t_sigdac *x =
        (t_sigdac *) obj_new((t_messlist **) & sigdac_class, sigdac_setup);
    sigdac_parse(x, ac, av);
    u_setup((t_ugen *) x, x->x_nchans, OUT0);
    if (u_debug)
        post("sigdac %lx", x);
    return (x);
}

/************** ADC ***************/

void copy_in(char *in, char *out, int bytes);

void sigadc_put(t_sigdac *x, long int whether)
{
    if (whether)
    {
        int i;
        u_stdout((t_ugen *) x);
        for (i = 0; i < x->x_nchans; i++)
        {
            float *buf = soundin + MAXVS * x->x_assign[i];
            dspchain_putc(u_chain, copy_in, 3,
                          buf, x->x_io[i]->s_shit,
                          sizeof(long) * x->x_io[0]->s_n);
        }
        if (u_debug)
            post("sigadc %lx: (->%lx,%lx)",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit);
    }
}

void *sigadc_class;
void sigadc_setup()
{
    c_new(&s__sigadc, u_clean, sizeof(t_sigdac), 0, 0);
    c_addmess(sigadc_put, &s_put, A_CANT, 0);
    c_addmess(u_setall, &s_start, A_DEFLONG, 0);
    c_addmess(u_unsetall, &s_stop, 0);
    c_addmess(sigdac_debug, &s_debug, A_DEFLONG, 0);
}

void *sigadc_new(t_symbol *s, int ac, t_atom *av)
{
    t_sigdac *x =
        (t_sigdac *) obj_new((t_messlist **) & sigadc_class, sigadc_setup);
    sigdac_parse(x, ac, av);
    u_setup((t_ugen *) x, IN0, x->x_nchans);
    if (u_debug)
        post("sigadc %lx", x);
    return (x);
}

/******************* PRINT *********************/

typedef struct sigprint {
    t_head x_h;
    t_sig *x_io[IN1 + OUT0];
    long x_count;
    t_symbol *x_sym;
} t_sigprint;

void cu_print(float *fp, t_sigprint *x, long int n)
{
    if (x->x_count)
    {
        post("%s:", x->x_sym->s_name);
        while (n)
        {
            char buf[160];
            sprintf(buf, "%12f %12f %12f %12f", fp[0], fp[1], fp[2],
                    fp[3]);
            post("%s", buf);
            /* post("%12f %12f %12f %12f", fp[0], fp[1], fp[2], fp[3]); */
            fp += 4;
            n -= 4;
        }
        x->x_count--;
    }
}

void sigprint_bang(t_sigprint *x)
{
    x->x_count = 1;
}

void sigprint_int(t_sigprint *x, long int n)
{
    if (n < 0 || n > 10)
        n = 10;
    x->x_count = n;
}

void sigprint_put(t_sigprint *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        dspchain_putc(u_chain, cu_print, 3,
                      x->x_io[0]->s_shit, x, x->x_io[0]->s_n);
        if (u_debug)
            post("sigprint %lx: (%lx)", x, x->x_io[0]->s_shit);
    }
}

void *sigprint_class;
void sigprint_setup()
{
    c_new(&s__sigprint, u_clean, sizeof(t_sigprint), 0, 0);
    c_addmess(sigprint_put, &s_put, A_CANT, 0);
    c_addint(sigprint_int);
    c_addbang(sigprint_bang);
    u_inletmethod(0);
}

void *sigprint_new(t_symbol *s)
{
    t_sigprint *x =
        (t_sigprint *) obj_new((t_messlist **) & sigprint_class,
                               sigprint_setup);
    u_setup((t_ugen *) x, IN1, OUT0);
    if (u_debug)
        post("sigprint %lx", x);
    x->x_count = 0;
    if (s == &s_)
        s = &s__sigprint;
    x->x_sym = s;
    return (x);
}
