#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:55:22 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:55:22 $
 *
 * FTS by Miller Puckette
 *
 */


/* definitions of phasor~, osc1~, tab1~.  */
#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__

    /* bit-twiddling constants */
#define SIN_BITS 9
#define SIN_NPOINTS (1 << SIN_BITS)
#define SIN_FRACBITS 8
#define SIN_FRAC (1 << SIN_FRACBITS)
#define UNITBIT32 1572864.      /* 3*2^19 -- bit 32 has value 1 */
#define FRACBITS (32-SIN_BITS)
#define FRACRANGE (1L << FRACBITS)

    /* the value to normalize differences into slopes in the table */
#define SLOPEMULTIPLIER (1. / (float) (FRACRANGE-1))

char *getpage(int size, int cacheable), *uncachedMalloc(unsigned int n);

extern t_symbol s_, s_put, s_set,
    s__sigphasor, s__sigosc1, s__sigtab1, s_print;

/************** PHASOR ***************/
typedef struct phctl {
    double c_phase;
    float c_fconv;
} t_phctl;

typedef struct phasor {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_phctl *x_ctl;
    t_phctl x_cspace;
} t_phasor;

#ifndef I860
long *ugen_phasor(long *arg)
{
    float *freq = (float *) (arg[1]);
    float *out = (float *) (arg[2]);
    t_phctl *x = (t_phctl *) (arg[3]);
    long n = (long) (arg[4]);
    int phase = 0x7fffffffL * x->c_phase;
    float fconv = x->c_fconv * 0x7fffffffL;
    while (n--)
    {
        phase = (phase + (int) (fconv * *freq++)) & 0x7fffffffL;
        *out++ = (1. / 0x7fffffffL) * phase;
    }
    x->c_phase = (1. / 0x7fffffffL) * phase;
    return (arg + 5);
}
#endif

void phasor_print(t_phasor *x)
{
    post("phasor: %lf", x->x_ctl->c_phase);
}

void phasor_float(t_phasor *x, double f)
{
    x->x_ctl->c_phase = f;
}

void phasor_int(t_phasor *x, long int n)
{
    phasor_float(x, (double) n);
}

void cu_phasor(float *freq, float *out, t_phctl * x, long int n),
su_phasor();
void phasor_put(t_phasor *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_cspace.c_fconv = 1. / x->x_io[0]->s_sr;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_phctl), sizeof(double));
#ifdef I860
        dspchain_put860(u_chain, su_phasor, 4,
                        x->x_io[0]->s_shit, x->x_io[1]->s_shit, x->x_ctl,
                        x->x_io[0]->s_n);
#else
        dspchain_addugen(ugen_phasor, 4,
                         x->x_io[0]->s_shit, x->x_io[1]->s_shit, x->x_ctl,
                         x->x_io[0]->s_n);
#endif
        if (u_debug)
            post("phasor %lx: %lx->%lx", x->x_ctl,
                 x->x_io[0]->s_shit, x->x_io[1]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_phctl));
        x->x_ctl = &x->x_cspace;
    }
}

void *phasor_class;
void phasor_setup()
{
    c_new(&s__sigphasor, u_clean, sizeof(t_phasor), 0, 0);
    c_addfloat(phasor_float);
    c_addint(phasor_int);
    c_addmess(phasor_put, &s_put, A_CANT, 0);
    c_addmess(phasor_print, &s_print, 0);
    u_inletmethod(0);
}

void *phasor_new()
{
    t_phasor *x =
        (t_phasor *) obj_new((t_messlist **) & phasor_class, phasor_setup);
    u_setup((t_ugen *) x, IN1, OUT1);
    x->x_ctl = &x->x_cspace;
    x->x_cspace.c_phase = 0.;
    x->x_cspace.c_fconv = 0.;
    if (u_debug)
        post("phasor %lx", x);
    return (x);
}

/************** tab1 ***************/
typedef struct tcommon {
    t_object t_obj;
    t_sinsamp *t_samps;
    t_symbol *t_sym;
    t_symbol *t_vol;            /* directory you're in */
    int t_refcount;
} t_tcommon;

#define HEADERSIZE 28
#define NINTERP 2
#define TABLESIZE 0x1000

#define NEEDBYTES (HEADERSIZE + TABLESIZE*sizeof(t_filesamp)/sizeof(t_sinsamp))

    /* return 1 if machine is big-endian, like Motorola, unlike i86. */
int ambigendian()
{
    unsigned short s = 1;
    unsigned char c = *(char *) (&s);
    return (c == 0);
}

void tcommon_bang(t_tcommon *x) /* read table in from NeXT soundfile */
{
    char *name = x->t_sym->s_name;
    int fd = open_via_path(name, x->t_vol, 0), nowrap;
    char *basename, *s2;
    char *tempbuf, *rats;
    long gimme, n;
    t_sinsamp *sptr;
    /* soundfile is big endian; if we aren't, swap bytes. */
    int swap = !ambigendian();
    if (fd < 0)
    {
        error("%s: can't open", name);
        return;
    }
    if (!(tempbuf = uncachedMalloc(NEEDBYTES)))
    {
        error("tab1: can't get buffer space for read");
        close(fd);
        return;
    }
    n = read(fd, tempbuf, NEEDBYTES);
    if (n < NEEDBYTES)
    {
        if (n < 0)
            error("%s: read failed", name);
        else
            error("%s: too short", name);
        close(fd);
        uncachedFree(tempbuf);
        return;
    }
    for (n = 0, rats = tempbuf + HEADERSIZE, sptr = x->t_samps;
         n < SIN_NPOINTS; rats += sizeof(t_filesamp), sptr++, n++)
    {
        if (swap)
        {
            char shit[2];
            shit[0] = rats[1];
            shit[1] = rats[0];
            sptr->value = (1. / 32767.) * (*(short *) shit);
        }
        else
            sptr->value = (1. / 32767.) * *(short *) rats;
        if (n)
            (sptr - 1)->slope = SLOPEMULTIPLIER *
                (sptr->value - (sptr - 1)->value);
    }
    for (s2 = basename = name; *s2; s2++)
        if (*s2 == '/')
            basename = s2 + 1;
    if (!strncmp(basename, "nowrap", 6))
    {
        /* post("tab1: reading %s (non-wraparound)", name); */
        x->t_samps[SIN_NPOINTS - 1].slope =
            x->t_samps[SIN_NPOINTS - 2].slope;
    }
    else
    {
        /* post("tab1: reading %s", name);  */
        x->t_samps[SIN_NPOINTS - 1].slope = SLOPEMULTIPLIER *
            (x->t_samps[0].value - x->t_samps[SIN_NPOINTS - 1].value);
    }
    close(fd);
    uncachedFree(tempbuf);
}

void tcommon_ff(t_tcommon *x)
{
    u_unset();                  /* stop the music (some oscillator might be using our table) */
    if (x->t_sym->s_thing == (t_object *) x)
        x->t_sym->s_thing = 0;
    ungetpage(x->t_samps, SINTABSIZE);
}

void *tcommon_class;
void tcommon_setup()
{
    c_new(&s__sigtab1, tcommon_ff, sizeof(t_tcommon), 0, 0);
}

void *tcommon_new(t_symbol *s)
{
    t_sinsamp *samps = (t_sinsamp *) getpage(SINTABSIZE, 1);
    if (!samps)
        return (0);
    else
    {
        t_tcommon *x =
            (t_tcommon *) obj_new((t_messlist **) & tcommon_class,
                                  tcommon_setup);
        x->t_sym = s;
        x->t_refcount = 0;
        x->t_samps = samps;
        x->t_vol = defvolume;
        s->s_thing = (t_object *) x;
        tcommon_bang(x);
        if (u_debug)
            post("tcommon %lx", x);
        return (x);
    }
}

typedef struct sigtab1 {
    t_object x_ob;
    t_tcommon *x_com;
} t_sigtab1;

void tab1_bang(t_sigtab1 *x)
{
    if (x->x_com)
        tcommon_bang(x->x_com);
}

void tab1_ff(t_sigtab1 *x)
{
    t_tcommon *y = x->x_com;
    if (y && !--y->t_refcount)
        freeobject((t_object *) y);
}

void *tab1_class;
void tab1_setup()
{
    c_new(&s__sigtab1, tab1_ff, sizeof(t_sigtab1), 0, 0);
    c_addbang(tab1_bang);
}

void *sigtab1_new(t_symbol *s)
{
    t_sigtab1 *x =
        (t_sigtab1 *) obj_new((t_messlist **) & tab1_class, tab1_setup);
    void *tc;
    if (!(*s->s_name))
    {
        x->x_com = 0;
        return (x);
    }
    if (!s->s_thing)
    {
        t_tcommon *y = tcommon_new(s);
        if (!y)
            error("tab1: out of space");
        y->t_refcount++;
        x->x_com = y;
    }
    else
    {
        if (ob_messlist(s->s_thing) == tcommon_class)
        {
            ((t_tcommon *) (s->s_thing))->t_refcount++;
            x->x_com = (t_tcommon *) (s->s_thing);
        }
        else
        {
            error("tab1~: %s: not a signal table", s->s_name);
            x->x_com = 0;
        }
    }
    if (u_debug)
        post("tab1 %lx", x);
    return (x);
}

/************** OSC1 ***************/
typedef struct oscctl {
    double c_phase;
    float c_fconv;
    t_sinsamp *c_samps;
} t_oscctl;

typedef struct sigosc1 {
    t_head x_h;
    t_sig *x_io[IN2 + OUT1];
    t_symbol *x_sym;
    t_oscctl *x_ctl;
    t_oscctl x_cspace;
} t_sigosc1;

t_sinsamp *sin_tab;
t_sinsamp *zero_tab;

    /* machine-dependent definitions.  These ifdefs really
       should have been by CPU type and not by operating system! */
#ifdef __sgi
    /* big-endian.  Most significant byte is at low address in memory */
#define HIOFFSET 0              /* word offset to find MSB */
#define LOWOFFSET 1             /* word offset to find LSB */
#define int32 long              /* a data type that has 32 bits */
#else

#ifdef  __linux__

#include <endian.h>

#if !defined(__BYTE_ORDER) || !defined(__LITTLE_ENDIAN)
#error No byte order defined
#endif

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define HIOFFSET 1
#define LOWOFFSET 0
#else
#define HIOFFSET 0              /* word offset to find MSB */
#define LOWOFFSET 1             /* word offset to find LSB */
#endif                          /* __BYTE_ORDER */

#include <sys/types.h>
#define int32 int32_t

#endif                          /* __linux__ */
#endif                          /* SGI */

long *ugen_osc1(long *arg)
{
    float *freq = (float *) arg[1];
    float *phaseshift = (float *) arg[2];
    float *out = (float *) arg[3];
    t_oscctl *x = (t_oscctl *) arg[4];
    long n = (long) arg[5];
    float fconv = x->c_fconv, f1;
    double oscphase = UNITBIT32 + x->c_phase + fconv * *freq++;
    t_sinsamp *tab = x->c_samps, *s;
    double phase = oscphase + *phaseshift++, phase2;
    int32 foo = ((int32 *) (&phase))[LOWOFFSET], foo2, x1;
    int count;
    for (count = 0; count < n - 1; count++)
    {
        oscphase += fconv * *freq;
        s = tab + ((foo >> (32 - SIN_BITS)) & (SIN_NPOINTS - 1));
        f1 = s->slope * (foo & (FRACRANGE - 1));
        phase = oscphase + *phaseshift++;
        foo = ((int32 *) (&phase))[LOWOFFSET];
        *out++ = s->value + f1;
        freq++;
    }
    {
        s = tab + ((foo >> (32 - SIN_BITS)) & (SIN_NPOINTS - 1));
        *out++ = s->value + s->slope * (foo & (FRACRANGE - 1));
        phase = oscphase;
        foo2 = ((int32 *) (&phase))[LOWOFFSET];
        x->c_phase = foo2 * (1. / 4294967296.);
    }
    return (arg + 6);
}

#if 0                           /* here is a slower, machine-independent version: */
long *ugen_osc1(long *arg)
{
    float *freq = (float *) arg[1];
    float *phaseshift = (float *) arg[2];
    float *out = (float *) arg[3];
    t_oscctl *x = (t_oscctl *) arg[4];
    long n = (long) arg[5];
    int ophase = 0x7fffffffL * x->c_phase;
    float fconv = x->c_fconv * 0x7fffffffL;
    t_sinsamp *tab = x->c_samps;
    while (n--)
    {
        long x1;
        t_sinsamp *s;
        ophase = (ophase + (int) (fconv * *freq++)) & 0x7fffffffL;
        x1 = *phaseshift++ * (SIN_NPOINTS * SIN_FRAC) +
            (ophase >> (31 - SIN_BITS - SIN_FRACBITS));
        s = tab + ((x1 & (SIN_NPOINTS * SIN_FRAC - 1)) >> SIN_FRACBITS);
        *out++ = s->value + s->slope *
            ((x1 & (SIN_FRAC - 1)) * (1. / (SIN_NPOINTS * SIN_FRAC)));
    }
    x->c_phase = (1. / 0x7fffffffL) * ophase;
    return (arg + 6);
}
#endif

void sigosc1_float(t_sigosc1 *x, double f)
{
    x->x_ctl->c_phase = f;
}

void sigosc1_int(t_sigosc1 *x, long int n)
{
    sigosc1_float(x, (double) n);
}

void sigosc1_set(t_sigosc1 *x, t_symbol *s)
{
    t_tcommon *wt;
    if (*s->s_name)
    {
        x->x_sym = s;
        if ((wt = (t_tcommon *) (x->x_sym->s_thing)) &&
            ob_messlist(wt) == tcommon_class)
            x->x_ctl->c_samps = wt->t_samps;
        else
            error("oscil~: %s: no such table", s->s_name);
    }
    else
    {
        x->x_sym = 0;
        x->x_ctl->c_samps = sin_tab;
    }
}

void cu_osc1(float *freq, float *phase, float *out, t_oscctl * x,
             long int n), su_osc1();
void sigosc1_put(t_sigosc1 *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        x->x_cspace.c_fconv = 1. / x->x_io[0]->s_sr;
        if (x->x_sym)
        {
            t_tcommon *y = (t_tcommon *) x->x_sym->s_thing;
            if (!y)
            {
                error("osc1~: %s: no such table", x->x_sym->s_name);
                x->x_cspace.c_samps = zero_tab;
            }
            else if (ob_messlist(y) != tcommon_class)
            {
                error("osc1~: %s: not a signal table", x->x_sym->s_name);
                x->x_cspace.c_samps = zero_tab;
            }
            else
                x->x_cspace.c_samps = y->t_samps;
        }
        else
            x->x_cspace.c_samps = sin_tab;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_oscctl), sizeof(double));
#ifdef I860
        dspchain_put860(u_chain, su_osc1, 5,
                        x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                        x->x_io[2]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#else
        dspchain_addugen(ugen_osc1, 5,
                         x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                         x->x_io[2]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#endif
        if (u_debug)
            post("sigosc1 %lx: (%lx,%lx)->%lx", x->x_ctl,
                 x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                 x->x_io[2]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_oscctl));
        x->x_ctl = &x->x_cspace;
    }
}

double cos(double);

int make_sin_tab()
{
    int i;
    if (SINTABSIZE != 0x1000)
        bug("make_sin_tab");
    if (!(zero_tab = (t_sinsamp *) getpage(SINTABSIZE, 1)))
        return (0);
    if (!(sin_tab = (t_sinsamp *) getpage(SINTABSIZE, 1)))
        return (0);
    for (i = 0; i < SIN_NPOINTS; i++)
        sin_tab[i].value = cos(i * (2 * 3.141593 / SIN_NPOINTS));
    for (i = 0; i < SIN_NPOINTS; i++)
        sin_tab[i].slope = SLOPEMULTIPLIER *
            (sin_tab[(i + 1) & (SIN_NPOINTS - 1)].value -
             sin_tab[i].value);
    for (i = 0; i < SIN_NPOINTS; i++)
        zero_tab[i].value = zero_tab[i].slope = 0;
    return (1);
}

void *sigosc1_class;
void sigosc1_setup()
{
    c_new(&s__sigosc1, u_clean, sizeof(t_sigosc1), 0, 0);
    c_addfloat(sigosc1_float);
    c_addint(sigosc1_int);
    c_addmess(sigosc1_set, &s_set, A_DEFSYM, 0);
    c_addmess(sigosc1_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
    u_inletmethod(1);
}

void *sigosc1_new(t_symbol *s)
{
    t_sigosc1 *x =
        (t_sigosc1 *) obj_new((t_messlist **) & sigosc1_class,
                              sigosc1_setup);
    if (!*s->s_name)
        s = 0;
    if (!sin_tab && !make_sin_tab())
        error("osc1: out of memory");
    x->x_ctl = &x->x_cspace;
    x->x_sym = s;
    x->x_cspace.c_fconv = 0;
    x->x_cspace.c_phase = 0;
    x->x_cspace.c_samps = sin_tab;
    u_setup((t_ugen *) x, IN2, OUT1);
    if (u_debug)
        post("sigosc1 %lx", x);
    return (x);
}
