#ifndef lint
static char rcsid[] =
    "$Revision: 2.8 $ IRCAM $Date: 1994/01/04 17:55:30 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/01/04 17:55:30 $
 *
 * FTS by Miller Puckette
 *
 */
#include <stdio.h>              /* get rid of this */

/* table, sampread, sampwrite, samppeek */

/*
  "Table" stores a floating-point array of samples, which the
  other ugens read in different ways.
  Table's creation argument gives the initial size of the sample in seconds.
  
  The table's size may be changed:
  size <seconds>
  
  and cleared:
  clear
  
  Table can be passed a message to get a soundfile from disk:
  read <filename> [starting byte]
  
  (LATER this could be extended to:
  read [flags] <filename> [starting byte] [ending byte]
  where [flags] are:
  -f  floating-point
  -b  byte-swapped
  -n  normalize amplitude to [-1, 1])
  
  The table is not automatically resized to fit the file.  The onset is in
  bytes to allow you to get rid of a header.
  
  Table can save its contents to disk:
  write <filename>
  
  (later this can take normalize, byte-swap, and floating-point options)
  
  "Sampread" takes an input signal as a number of milliseconds, and reads
  the table at that delay, interpolating over four points, and putting
  the result on its signal output.  If the
  incoming signal changes violently during a frame, the output is set
  to zero for the frame.  By default, the limit is a 4:1 variation,
  for a two-octave transposition; this is changed by the message,
  limit <new transposition limit>
  
  Sampread, sampwrite, and samppeak can be directed to use
  a different sample:
  set <table-name>
  
  "Sampwrite" records a sample from a signal input on "bang".
  
  Later , two floating- point inlets could optionally restrict the range of
  capture, in milliseconds.
  
  "Samppeek" takes a floating-point index and outputs a floating-point
  sample as a message.
  
  */

#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"
#if (defined(NEXTSTEP) || defined(I860))
#include "/usr/include/sound/soundstruct.h"     /* for writing NeXT soundfiles */
#elif defined(__unix)
#include "cp_sound.h"
#endif
#ifdef I860
#include "../6a/6a.h"
#endif

char *uncachedMalloc(unsigned int n);

#define GUARDPTS 4

extern t_symbol s_put, s_set, s_read, s_write, s_size, s_jump, s_in1,
    s__sigsampread, s__sigsampwrite, s__sigsamppeek, s__sigtable;

#ifdef I860
void su_sampread(), su_sampwrite();
#endif                          /* I860 */
static t_sample nowhere[2 * GUARDPTS];
t_sampfilt *sampfilt_tab;

/******************** TABLE ***********************/

#define TEMPBUFSIZE 16384
#define TEMPBUFSAMPS (TEMPBUFSIZE/sizeof(t_filesamp))
char *uncachedMalloc(unsigned int n);
long lseek();

void sigtable_read(t_sigtable *x, t_symbol *s, long int n)
{
    int fd;
    long bufsize, sampsleft, sampstoread;
    char tempbuf[TEMPBUFSIZE + 320];
    /* char *tempbuf; */
    t_sample *outptr;
    int swap = !ambigendian();
    /* static int dc = 0; */
    /* if (!(tempbuf = uncachedMalloc(TEMPBUFSIZE)))
       {
       error("table: can't get buffer space for read");
       return;
       } */
    if ((fd = open_via_path(s->s_name, x->x_vol, 0)) < 0)
    {
        error("table: %s: can't open", s->s_name);
        /* uncachedFree(tempbuf); */
        return;
    }
    if (lseek(fd, n, 0) < 0)
    {
        error("table: %s: can't seek to beginning", s->s_name);
        close(fd);
        /* uncachedFree(tempbuf); */
        return;
    }
    /* filesize -= n; */
    sampstoread = x->x_buf.b_nsamps + GUARDPTS;
    for (sampsleft = sampstoread, outptr = x->x_buf.b_shit; sampsleft > 0;)
    {
        long gimme = sampsleft, n, bytesread;
        char *rats;
        if (gimme > TEMPBUFSAMPS)
            gimme = TEMPBUFSAMPS;
        bytesread = read(fd, tempbuf, gimme * sizeof(t_filesamp));
        if (bytesread & 1)
        {
            post("warning: dropping odd byte");
            bytesread &= ~1;
        }
        if (bytesread < 0)
            error("%s: read error", s->s_name);
        if (bytesread <= 0)
            break;
        for (n = bytesread / sizeof(t_filesamp), rats = tempbuf;
             n--; rats += sizeof(t_filesamp))
        {
            if (swap)
            {
                char shit[2];
                shit[0] = rats[1];
                shit[1] = rats[0];
                *(outptr++) = (1. / 32767.) * *(t_filesamp *) shit;
            }
            else
                *(outptr++) = (1. / 32767.) * *(t_filesamp *) rats;
            /* if (dc++ < 10) post("%d %d -> %d",
               (long)(rats[0]), (long)(rats[1]), outptr[-1]); */
        }
        sampsleft -= bytesread / sizeof(t_filesamp);
    }
    post("%s: read %d samples (out of %d in table)",
         s->s_name, sampstoread - sampsleft, x->x_buf.b_nsamps + GUARDPTS);
    while (sampsleft-- > 0)
        *outptr++ = 0.;
    close(fd);
    /* uncachedFree(tempbuf); */
}

void swaplong(void *vp)
{
    char *lp = (char *) vp;
    char temp;
    temp = lp[0];
    lp[0] = lp[3];
    lp[3] = temp;
    temp = lp[1];
    lp[1] = lp[2];
    lp[2] = temp;
}

void sigtable_write(t_sigtable *x, t_symbol *s)
{
    int fd;
    long filesize, bufsize, sampsleft, sampstowrite;
    char tempbuf[TEMPBUFSIZE];
    SNDSoundStruct header;
    t_sample *outptr;
    int bufno = 0;
    int swap = !ambigendian();
    if ((fd = creat(s->s_name, 0666)) < 0)
    {
        error("table: %s: can't create", s->s_name);
        return;
    }
    header.magic = SND_MAGIC;
    header.dataLocation = sizeof(header);
    header.dataSize = 0x10000000;
    header.dataFormat = SND_FORMAT_LINEAR_16;
    header.samplingRate = dsp_sr;
    header.channelCount = 1;
    strcpy(header.info, "Max");
    if (swap)
    {
        swaplong(&header.magic);
        swaplong(&header.dataLocation);
        swaplong(&header.dataSize);
        swaplong(&header.dataFormat);
        swaplong(&header.samplingRate);
        swaplong(&header.channelCount);
    }
    if (write(fd, &header, sizeof(header)) < sizeof(header))
    {
        error("%s: write error in header", s->s_name);
        close(fd);
        return;
    }
    for (outptr = x->x_buf.b_shit,
         sampstowrite = x->x_buf.b_nsamps + GUARDPTS; sampstowrite > 0;)
    {
        int writehere = (sampstowrite > TEMPBUFSAMPS ?
                         TEMPBUFSAMPS : sampstowrite), n;
        char *rats;
        for (n = writehere, rats = tempbuf; n--;
             rats += sizeof(t_filesamp))
        {
            int printme = (bufno == 24 && n == writehere - 97);
            float foobar;
            int fixsamp = 32767. * (foobar = *(outptr++));
            short fixshort;
            char *shit = (char *) (&fixshort);
#ifdef I860
            if (printme)
                setLEDS(LED0, 16);
#endif                          /* I860 */
            if (fixsamp > 32767)
                fixsamp = 32767;
            else if (fixsamp < -32767)
                fixsamp = -32767;
            fixshort = fixsamp;
            if (swap)
            {
                rats[0] = shit[1];
                rats[1] = shit[0];
            }
            else
                *(t_filesamp *) rats = fixshort;

            /* if (bufno == 24 && n == writehere - 97) */
            if (printme)
            {
                short f2;
                char *shit = (char *) (&f2);
                shit[0] = rats[1];
                shit[1] = rats[0];
            }
        }
        if (write(fd, tempbuf, writehere * sizeof(t_filesamp)) <
            writehere * sizeof(t_filesamp))
        {
            error("%s: write error", s->s_name);
            break;
        }
        sampstowrite -= writehere;
        bufno++;
    }
    close(fd);
    post("wrote %s", s->s_name);
}

/* problem: you don't really know the sample rate here ...  */

void sigtable_size(t_sigtable *x, double f)
{
    long samps;
    int i;
    t_sample *sp;

    if (f < 0)
        f = 0;
    if (x->x_buf.b_shit != nowhere)
        uncachedFree(x->x_buf.b_shit);
    samps = GUARDPTS + .001 * dsp_sr * f;
    if (samps < GUARDPTS)
        samps = GUARDPTS;
    if (!(x->x_buf.b_shit =
          (t_sample *) uncachedMalloc(samps * sizeof(t_sample))))
    {
        error("sample: not enough memory");
        x->x_buf.b_shit = nowhere;
        samps = GUARDPTS;
    }
    x->x_buf.b_nsamps = samps - GUARDPTS;
    for (i = samps, sp = x->x_buf.b_shit; i--;)
        *(sp++) = 0;
}

void sigtable_set(t_sigtable *x, long int addr, long int nsamps)        /* obsolete */
{
    post("warning: 'set' message to table~ is obsolete -- see help window");
    sigtable_size(x, 1000. * nsamps / dsp_sr);
}

void *sigtable_class;
void sigtable_ff(t_sigtable *x)
{
    u_unset();                  /* because there may be pointers to x in the call chain */
    if (x->x_sym->s_thing
        && ob_messlist(x->x_sym->s_thing) == sigtable_class)
    {
        t_sigtable stub, *y, *z;
        stub.x_next = (t_sigtable *) (x->x_sym->s_thing);
        for (y = &stub; z = y->x_next; y = z)
        {
            if (z == x)
            {
                y->x_next = z->x_next;
                break;
            }
        }
        x->x_sym->s_thing = (t_object *) (stub.x_next);
    }
    if (x->x_buf.b_shit != nowhere)
        uncachedFree(x->x_buf.b_shit);
}

void sigtable_setup()
{
    c_new(&s__sigtable, sigtable_ff, sizeof(t_sigtable), 0, 0);
    c_addmess(sigtable_read, &s_read, A_SYM, A_DEFLONG, 0);
    c_addmess(sigtable_write, &s_write, A_SYM, 0);
    c_addmess(sigtable_set, &s_set, A_LONG, A_LONG, 0);
}

extern int h_ishost;
void *sigtable_new(t_symbol *s, double maxdel)
{
    t_sigtable *x =
        (t_sigtable *) obj_new((t_messlist **) & sigtable_class,
                               sigtable_setup);
    if (h_ishost)
        maxdel = 0;
    x->x_buf.b_shit = nowhere;
    sigtable_size(x, maxdel);
    x->x_buf.b_conv = dsp_sr * .001;
    if (u_debug)
        post("table %lx", x);
    x->x_sym = s;
    x->x_vol = defvolume;
    if (s->s_thing)
    {
        if (ob_messlist(s->s_thing) != sigtable_class)
        {
            error("%s: already exists and is not a sample table",
                  s->s_name);
        }
        else
        {
            x->x_next = ((t_sigtable *) (s->s_thing));
            s->s_thing = (t_object *) x;
        }
    }
    else
    {
        s->s_thing = (t_object *) x;
        x->x_next = 0;
    }
    return (x);
}

/******************** SAMPREAD ***********************/

typedef struct srctl {
    float c_1overn;             /* 1 divided by vector length */
    float c_lastx;              /* previous value of input */
    float c_maxextent;          /* maximum span between lastx and new input */
    t_sampbuf *c_buf;
} t_srctl;

typedef struct sampread {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_symbol *x_sym;            /* symbol bound to table we'll use */
    float x_maxshift;           /* maximum up-shift of pitch */
    long x_vecsize;             /* vector size (for recalculating maxextent) */
    t_srctl *x_ctl;
    t_srctl x_cspace;
} t_sampread;

#ifndef I860
void cu_sampread(float *in1, float *out, t_srctl *x, long int n)
{
    static int didit;
    long phase;
    long addr1, addr2, n2;
    float f1, f2, f3, f4, extent, frac, *fp;
    t_sampbuf *b = x->c_buf;
    t_sample *sp;
    float tbuf[512];
    if (!b)
        goto zero;
    f1 = x->c_lastx;
    f2 = in1[0] * b->b_conv;
    x->c_lastx = f2;
    if (f2 >= f1)               /* reading forward */
    {
        if (f1 < 0 || f2 >= b->b_nsamps - 6 ||
            (extent = f2 - f1) > x->c_maxextent)
            goto zero;
        n2 = f1;
        sp = b->b_shit + n2;
        addr1 = (f1 - n2) * 65536.;
        addr2 = 65536. * extent * x->c_1overn;
    }
    else                        /* backward */
    {
        if (f2 < 1 || f1 >= b->b_nsamps - 6 ||
            (extent = f1 - f2) > x->c_maxextent)
            goto zero;
        n2 = f2 - .5;
        sp = b->b_shit + n2;
        addr1 = (f1 - n2) * 65536.;
        addr2 = -(long) (65536. * extent * x->c_1overn);
        /* are you sure this won't go past the bottom? */
    }
    n2 = ((long) extent) + 5;

    for (fp = tbuf; n2--;)
        *fp++ = *sp++;

    while (n--)
    {
        t_sampfilt *t = sampfilt_tab + ((addr1 >> 8) & 255);
        fp = tbuf + (addr1 >> 16);
        *out++ = t->f1 * fp[0] + t->f2 * fp[1] +
            t->f3 * fp[2] + t->f4 * fp[3];
        addr1 += addr2;
    }
    didit = 1;
    return;
  zero:
    while (n--)
        *out++ = 0;
}
#endif                          /* I860 */

gensampname(char *buf, char *base, long int n)
{
    char b2[30];
    char *s = b2 + 29;
    *s = 0;
    if (n < 0)
        n = 0;
    while (*buf = *base++)
        buf++;
    if (!n)
        *(--s) = '0';
    else
        while (n)
        {
            *(--s) = '0' + (n % 10);
            n /= 10;
        }
    while (*buf++ = *s++);
}

void sampread_set(t_sampread *x, t_symbol *s)
{
    t_sigtable *wt;
    x->x_sym = s;
    if ((wt = (t_sigtable *) (s->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
    {
        x->x_ctl->c_buf = &wt->x_buf;
        if (wt->x_next)
            post("sampread(%s): warning: multiply defined", s->s_name);
    }
    else
    {
        error("sampread: %s: no such sample", s->s_name);
        x->x_ctl->c_buf = 0;
    }
}

void sampread_int(t_sampread *x, long int n)
{
    char buf[30];
    gensampname(buf, "sample", n);
    sampread_set(x, gensym(buf));
}

void sampread_jump(t_sampread *x, double f)     /* jump to a new point */
{
    char buf[30];
    x->x_ctl->c_lastx = f * x->x_ctl->c_buf->b_conv;
}

void sampread_put(t_sampread *x, long int whether)
{
    if (whether)
    {
        t_sigtable *wt;
        u_stdout((t_ugen *) x);
        x->x_cspace.c_maxextent = x->x_maxshift *
            (x->x_vecsize = x->x_io[0]->s_n);
        x->x_cspace.c_1overn = 1. / x->x_vecsize;
        sampread_set(x, x->x_sym);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_srctl), sizeof(float));
#ifdef I860
        dspchain_put860(u_chain, su_sampread, 4,
                        x->x_io[0]->s_shit + (x->x_io[0]->s_n - 1),
                        x->x_io[1]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#else
        dspchain_putc(u_chain, cu_sampread, 4,
                      x->x_io[0]->s_shit + (x->x_io[0]->s_n - 1),
                      x->x_io[1]->s_shit, x->x_ctl, x->x_io[0]->s_n);
#endif                          /* I860 */
        if (u_debug)
            post("sampread %lx: %lx->%lx",
                 x, x->x_io[0]->s_shit, x->x_io[1]->s_shit);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_srctl));
        x->x_ctl = &x->x_cspace;
    }
}

void sampread_setup()
{
    c_new(&s__sigsampread, u_clean, sizeof(t_sampread), 0, 0);
    c_addmess(sampread_set, &s_set, A_SYM, 0);
    c_addmess(sampread_jump, &s_jump, A_DEFFLOAT, 0);
    c_addint(sampread_int);
    c_addmess(sampread_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
}

void *sampread_class;
void *sigsampread_new(t_symbol *s, double maxshift)
{
    t_sampread *x;
    if (!sampfilt_tab && !make_samp_tab())
    {
        error("sampread: out of memory");
        return (0);
    }
    x = (t_sampread *) obj_new((t_messlist **) & sampread_class,
                               sampread_setup);
    u_setup((t_ugen *) x, IN1, OUT1);
    x->x_sym = s;
    x->x_maxshift = (maxshift > 1. ? maxshift : 4.);
    x->x_vecsize = 4;
    x->x_cspace.c_lastx = 0;
    x->x_cspace.c_maxextent = x->x_maxshift * (x->x_vecsize = DEFAULTVS);
    x->x_cspace.c_1overn = 1. / x->x_vecsize;
    x->x_ctl = &x->x_cspace;
    if (u_debug)
        post("sampread %lx", x);
    return (x);
}

/******************** SAMPWRITE ***********************/
typedef struct wctl {
    long c_onset;
    t_sampbuf *c_buf;
} t_wctl;

typedef struct sampwrite {
    t_head w_h;
    t_sig *w_io[IN1 + OUT0];
    t_symbol *w_sym;
    t_wctl w_state;
    t_wctl *w_ctl;
} t_sampwrite;

#ifndef I860
void cu_sampwrite(float *in, t_wctl *ctl, long int n)
{
    long onset = ctl->c_onset;
    t_sampbuf *b = ctl->c_buf;
    t_sample *xp = b->b_shit + onset;
    long empty = b->b_nsamps + GUARDPTS - onset;
    if (empty <= 0)
        return;
    if (empty > n)
        empty = n;
    ctl->c_onset = onset + empty;
    while (empty--)
        *xp++ = *in++;
}
#endif                          /* I860 */

void sampwrite_bang(t_sampwrite *x)
{
    x->w_ctl->c_onset = 0;
}

void sampwrite_set(t_sampwrite *x, t_symbol *s)
{
    t_sigtable *wt;
    x->w_sym = s;
    if ((wt = (t_sigtable *) (x->w_sym->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
        x->w_ctl->c_buf = &wt->x_buf;
    else
    {
        error("sampwrite: %s: no such sample", s->s_name);
        x->w_ctl->c_buf = 0;
    }
    x->w_ctl->c_onset = 0x7fffffff;
}

void sampwrite_int(t_sampwrite *x, long int n)
{
    char buf[30];
    gensampname(buf, "sample", n);
    sampwrite_set(x, gensym(buf));
}

void sampwrite_put(t_sampwrite *x, long int whether)
{
    if (whether)
    {
        t_sigtable *wt;
        u_stdout((t_ugen *) x);
        sampwrite_set(x, x->w_sym);
        x->w_ctl = dspchain_putstate(u_chain, &x->w_state,
                                     sizeof(x->w_state), sizeof(long));
#ifdef I860
        dspchain_put860(u_chain, su_sampwrite, 3,
                        x->w_io[0]->s_shit, x->w_ctl, x->w_io[0]->s_n);
#else                           /* I860 */
        dspchain_putc(u_chain, cu_sampwrite, 3,
                      x->w_io[0]->s_shit, x->w_ctl, x->w_io[0]->s_n);
#endif                          /* I860 */
        if (u_debug)
            post("sampwrite %lx: %lx->%lx",
                 x, x->w_io[0]->s_shit, x->w_io[1]->s_shit);
    }
    else
    {
        x->w_state.c_onset = x->w_ctl->c_onset;
        x->w_state.c_buf = x->w_ctl->c_buf;
        x->w_ctl = &x->w_state;
    }
}

void sampwrite_setup()
{
    c_new(&s__sigsampwrite, u_clean, sizeof(t_sampwrite), 0, 0);
    c_addbang(sampwrite_bang);
    c_addmess(sampwrite_put, &s_put, A_CANT, 0);
    c_addmess(sampwrite_set, &s_set, A_SYM, 0);
    c_addint(sampwrite_int);
    u_inletmethod(0);
}

void *sampwrite_class;
void *sigsampwrite_new(t_symbol *s)
{
    t_sampwrite *x =
        (t_sampwrite *) obj_new((t_messlist **) & sampwrite_class,
                                sampwrite_setup);
    u_setup((t_ugen *) x, IN1, OUT0);
    x->w_sym = s;
    x->w_state.c_onset = 0x7fffffff;
    x->w_ctl = &x->w_state;
    if (u_debug)
        post("sampwrite %lx", x);
    return (x);
}

/**** samppeek -- get values from a sample as messages ****/

typedef struct sigsamppeek {
    t_object p_ob;
    t_symbol *p_sym;
} t_sigsamppeek;

void sigsamppeek_float(t_sigsamppeek *x, double f)
{
    t_sigtable *wt;
    if ((wt = (t_sigtable *) (x->p_sym->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
    {
        t_sampbuf *b = &wt->x_buf;
        float onset;
        int rats, frac;
        onset = b->b_conv * f - 1.;
        if (onset < 0)
        {
            outlet_float(x->p_ob.o_outlet, 0.);
            return;
        }
        rats = onset;
        frac = SAMP_NPOINTS * (onset - rats);
        if (rats >= b->b_nsamps)
            outlet_float(x->p_ob.o_outlet, 0.);
        else
        {
            t_sampfilt *t = sampfilt_tab + frac;
            float *fp = b->b_shit + rats;
            if (frac < 0 || frac >= SAMP_NPOINTS)
            {
                post("onset %f, frac %d", onset, frac);
                bug("sigsamppeek_float");
            }
            outlet_float(x->p_ob.o_outlet,
                         t->f1 * fp[0] + t->f2 * fp[1] +
                         t->f3 * fp[2] + t->f4 * fp[3]);
        }
    }
}

void sigsamppeek_int(t_sigsamppeek *x, long int n)
{
    sigsamppeek_float(x, (double) n);
}

void sigsamppeek_set(t_sigsamppeek *x, t_symbol *s)
{
    x->p_sym = s;
}

void sigsamppeek_in1(t_sigsamppeek *x, long int n)
{
    char buf[30];
    gensampname(buf, "sample", n);
    sigsamppeek_set(x, gensym(buf));
}

void sigsamppeek_setup()
{
    c_new(&s__sigsamppeek, NIL, sizeof(t_sigsamppeek), 0, 0);
    c_addfloat(sigsamppeek_float);
    c_addinx(sigsamppeek_in1, 1);
    c_addmess(sigsamppeek_set, &s_set, A_SYM, 0);
    c_addint(sigsamppeek_int);
}

void *sigsamppeek_class;

void *sigsamppeek_new(t_symbol *s)
{
    t_sigsamppeek *x;
    if (!sampfilt_tab && !make_samp_tab())
    {
        error("sampread: out of memory");
        return (0);
    }
    x = (t_sigsamppeek *) obj_new((t_messlist **) & sigsamppeek_class,
                                  sigsamppeek_setup);
    outlet_new(x, &s_float);
    inlet_new(x, x, &s_int, &s_in1);
    x->p_sym = s;
    return (x);
}
