#ifndef lint
static char rcsid[] =
    "$Revision: 2.2 $ IRCAM $Date: 1993/10/14 11:58:01 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.2 $ IRCAM $Date: 1993/10/14 11:58:01 $
 *
 * FTS by Miller Puckette
 *
 */


/* the signal object.  Signals are packed into pages; they are freed all
   at once by calling sig_setup.  If a signal being freed is located at
   the beginning of a page, the whole page is freed.  */

#include "m_mess.h"
#define __STRICT_BSD__
#include "math.h"
#undef __STRICT_BSD__
#include "d_graph.h"
#include "d_ugen.h"

extern struct symbol s_sig, s_dspchain, s__sig, s_tick, s_open, s_close;

void *getbytes(int n);
static t_sig *sig_freed = 0;
static int sig_count = 0;       /* for printing signal usage statistic */
extern void *u_chain;

/* signals in "limbo" do not own the space they point to;
   they are kept around so that ugens can have them as inputs,
   even while their space is being reused as an output, possibly
   at a different sampling rate. */

static t_sig *sig_limbo = 0;
static int sig_defaultvs = DEFAULTVS;   /* unused, untested, ... */
static char *sig_space, *sig_espace;
void *dspchain_getpage();

void *sigx_new()
{
    t_sig *x;
    long nbytes = MAXVS * sizeof(float);
    if (!sig_space || sig_space + nbytes > sig_espace)
    {
        sig_space = (char *) dspchain_getpage();
        sig_espace = sig_space + UCALL_PAGE_SIZE;
    }
    if (!sig_space)
    {
        error("out of memory");
        return (0);
    }
    x = (t_sig *) getbytes(sizeof(t_sig));
    sig_count++;
    x->s_alloc = (float *) sig_space;
    sig_space += nbytes;
    return (x);
}

void *sig_getnew(int size)
{
    t_sig *x;
    if (sig_freed)
    {
        x = sig_freed;
        sig_freed = x->s_next;
    }
    else
        x = (t_sig *) sigx_new();
    x->s_count = 0;
    x->s_n = size;
    x->s_shit = x->s_alloc + (MAXVS - size);
    /* fix me -- this will break for variable duty cycles */
    x->s_sr = dsp_sr / (double) (DEFAULTVS / size);
    return (x);
}

/* put a signal's vector on the free list.  Put the signal structure
   itself in limbo.  Allocate another signal structure to hold the
   freed space. */

void sig_reusable(t_sig *x)
{
    t_sig *y = (t_sig *) getbytes(sizeof(t_sig));
    y->s_next = sig_freed;
    sig_freed = y;
    y->s_alloc = x->s_alloc;
    x->s_next = sig_limbo;
    sig_limbo = x;
}

/* free all signals in limbo.  The vector isn't freed since it was
   already given to another signal. */

void sig_reclaim()
{
    while (sig_limbo)
    {
        freebytes(sig_limbo, sizeof(*sig_limbo));
        sig_limbo = sig_limbo->s_next;
    }
}

static void sig_setdefaultvs(int n)
{
    n &= -MINVS;
    if (n < 16)
        n = DEFAULTVS;
    /* 16 because cp_real.c can handle only integral multiple of dac
       buffer size */
    else if (n >= MAXVS)
        n = MAXVS;
    sig_defaultvs = n;
}

sig_getcount()
{
    return (sig_count);
}

/* clear out the free list and put all the memory back.  Called when you
   want to build a new chain.  It's better to
   reallocate everything from scratch because there might be fewer signals
   this time and you'd like to cram them into fewer physical pages. */

void sig_setup(int vecsize)
{
    t_sig *x;
    sig_setdefaultvs(vecsize);
    for (x = sig_freed; x; x = x->s_next)
    {
        if (!((int) (x->s_alloc) & (UCALL_PAGE_SIZE - 1)))
            dspchain_ungetpage(x->s_alloc);
        freebytes(x, sizeof(t_sig));
    }
    sig_freed = 0;
    sig_space = 0;
    sig_count = 0;
}

/* check that all but one of the signals (the zero signal)
   has been freed */

void sig_checkreusable()
{
    t_sig *x;
    int count = 0;
    for (x = sig_freed; x; x = x->s_next)
        count++;
    if (count != sig_count - 1)
        bug("signal count %d; freed %d", sig_count, count);
}
