#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/01/04 17:55:41 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/01/04 17:55:41 $
 *
 * FTS by Miller Puckette
 *
 */


/* threshold detector */

#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"

extern t_symbol s_put, s_set, s__sigthreshold;

typedef struct thresholdctl {
    int c_wait;                 /* ticks to wait before becoming active */
    int c_on;                   /* true if waiting for the low threshold */
    float c_hithresh;
    float c_lothresh;
    int c_hideadtime;
    int c_lodeadtime;
    void *c_clock;
} t_thresholdctl;

typedef struct sigthreshold {
    t_head x_h;
    t_sig *x_io[IN1 + OUT0];    /* 1 input; no outputs */
    void *x_outlet1;            /* "bang" out when reaches hithresh */
    void *x_outlet2;            /* "bang" out when reaches lothresh */
    t_thresholdctl *x_ctl;      /* pointer to state */
    t_thresholdctl x_cspace;    /* state when not in a chain */
    double x_hideadmsec;        /* hi dead time in msec */
    double x_lodeadmsec;        /* lo dead time in msec */
    double x_tickcvt;           /* msec to DSP chain tick conversion */
} t_sigthreshold;

void cu_threshold(float *in1, t_thresholdctl * x);
void sigthreshold_set(t_sigthreshold * x, double d1, double t1, double d2,
                      double t2);

/*
 * cu_threshold()
 *
 * Purpose:  the actiual DSP routine -- called for every n samples of input 
 *
 * Parameters:
 *  in1 -- input vector -- n samples
 *  x   -- control block
 *
 * Returns:  
 * Side-Effects:  
 * Caveats:  
 */

void cu_threshold(float *in1, t_thresholdctl *x)
{
    if (x->c_wait)
        x->c_wait--;
    else if (x->c_on)
    {
        if (*in1 < x->c_lothresh)       /* if sample < lo threshold ... */
        {
            /* set the clock to go off as soon as possible --
               i.e., at the next control period */
            clock_delay(x->c_clock, 0L);
            /* set dead time and go into "off" state */
            x->c_on = 0;
            x->c_wait = x->c_lodeadtime;
        }
    }
    else
    {
        if (*in1 >= x->c_hithresh)      /* if sample >= hi threshold ... */
        {
            /* set the clock */
            clock_delay(x->c_clock, 0L);
            /* set dead time and go into "on" state */
            x->c_on = 1;
            x->c_wait = x->c_hideadtime;
        }
    }
}

/* routine which FTS calls to put you on the DSP chain or take you off.
   This is never called explicitly; see the setup routine's use of A_CANT */

void sigthreshold_put(t_sigthreshold *x, long int whether)
{
    if (whether)                /* to put on the chain: */
    {
        u_stdout((t_ugen *) x); /* generate signals for outputs */
        x->x_tickcvt = x->x_io[0]->s_sr / (1000. * x->x_io[0]->s_n);
        sigthreshold_set(x, x->x_cspace.c_hithresh, x->x_hideadmsec,
                         x->x_cspace.c_lothresh, x->x_lodeadmsec);
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_thresholdctl),
                                     sizeof(float));
        dspchain_putc(u_chain, cu_threshold, 2,
                      x->x_io[0]->s_shit + (x->x_io[0]->s_n - 1),
                      x->x_ctl);
    }
    else                        /* to take it off the chain, copy state back into the garage */
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_thresholdctl));
        x->x_ctl = &x->x_cspace;
    }
}

void sigthreshold_set(t_sigthreshold *x, double d1, double t1, double d2, double t2)    /* sets arguments */
{
    int ticks;
    t_thresholdctl *c = x->x_ctl;
    c->c_hithresh = d1;
    x->x_hideadmsec = t1;
    ticks = (int) (x->x_tickcvt * t1) - 1;
    c->c_hideadtime = (ticks > 0 ? ticks : 0);
    if (d2 > d1)
        d2 = d1;
    c->c_lothresh = d2;
    x->x_lodeadmsec = t2;
    ticks = (int) (x->x_tickcvt * t2) - 1;
    c->c_lodeadtime = (ticks > 0 ? ticks : 0);
}

void sigthreshold_int(t_sigthreshold *x, long int n)    /* set state */
{
    x->x_ctl->c_on = n;
    x->x_ctl->c_wait = 0;
}

void sigthreshold_float(t_sigthreshold *x, double f)
{
    sigthreshold_int(x, (long) f);
}

/*
  void sigthreshold_bang(x, f)
  t_sigthreshold *x;
  double f;
  {
  post("hi thresh %f, dead %d",
  x->x_ctl->c_hithresh, x->x_ctl->c_hideadtime);
  post("lo thresh %f, dead %d",
  x->x_ctl->c_lothresh, x->x_ctl->c_lodeadtime);
  post("on %d, wait %d", x->x_ctl->c_on, x->x_ctl->c_wait);
  }
  */

void sigthreshold_tick(t_sigthreshold *x)       /* callback function for the clock */
{
    if (x->x_ctl->c_on)
        outlet_bang(x->x_outlet1);
    else
        outlet_bang(x->x_outlet2);
}

void sigthreshold_ff(t_sigthreshold *x) /* cleanup on free */
{
    clock_free(x->x_ctl->c_clock);
    u_clean((t_ugen *) x);      /* generic cleanup for tilde objects */
}

void *sigthreshold_class;
void sigthreshold_setup()
{
    c_new(&s__sigthreshold, sigthreshold_ff, sizeof(t_sigthreshold), 0, 0);
    c_addmess(sigthreshold_put, &s_put, A_CANT, 0);
    c_addfloat(sigthreshold_float);
    c_addint(sigthreshold_int);
    /* c_addbang(sigthreshold_bang); */
    c_addmess(sigthreshold_set, &s_set,
              A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, 0);
    u_inletmethod(0);
}

/* hi thresh, hi deadtime, lo thresh, lo deadtime */
void *sigthreshold_new(double f1, double f2, double f3, double f4)
{
    t_sigthreshold *x = (t_sigthreshold *)
        obj_new((t_messlist **) & sigthreshold_class, sigthreshold_setup);
    u_setup((t_ugen *) x, IN1, OUT0);
    x->x_ctl = &x->x_cspace;
    x->x_cspace.c_on = 0;       /* state 0 (low) at startup */
    x->x_cspace.c_wait = 0;     /* no dead time */
    x->x_cspace.c_clock = clock_new(x, sigthreshold_tick);
    x->x_outlet2 = outlet_new(x, &s_bang);
    x->x_outlet1 = outlet_new(x, &s_bang);
    x->x_tickcvt = 0.;          /* fake; it's reset in "put" routine */
    sigthreshold_set(x, f1, f2, f3, f4);
    if (u_debug)
        post("sigthreshold %lx", x);
    return (x);
}
