
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/01/29 07:47:39 $
 *
 * FTS by Miller Puckette
 *
 */

extern float dsp_sr;
extern int dsp_vs;
extern void *u_chain;           /* "current" dspchain */

#define MINVS 8                 /* vectors must be a multiple of MINVS points */
#define DEFAULTVS 64
#define MAXVS 64
#define SIGALIGN 16             /* alignment requirement for unit generators */

typedef struct sig {
    int s_count;
    struct sig *s_next;
    int s_n;
    float s_sr;                 /* sampling rate */
    float *s_shit;              /* logical beginning of signal (may depend on s_n) */
    float *s_alloc;             /* signal's allocated block, MAXVS samples long */
} t_sig;

#define SIN_BITS 9
#define SIN_NPOINTS (1 << SIN_BITS)
#define SIN_FRACBITS 8
#define SIN_FRAC (1 << SIN_FRACBITS)
typedef struct sinsamp {
    float value;
    float slope;
} t_sinsamp;
#define SINTABSIZE (SIN_NPOINTS * sizeof(t_sinsamp))

#define SAMP_NBITS 8
#define SAMP_NPOINTS (1 << SAMP_NBITS)

typedef struct sampfilt {
    float f1, f2, f3, f4;
} t_sampfilt;
extern t_sampfilt *sampfilt_tab;
#define SAMPFILTSIZE (SAMP_NPOINTS * sizeof(t_sampfilt))

typedef struct _dspchain {
    struct tinyobject d_ob;
    long *d_firstpage;          /* first page to "do" on tick */
    long *d_currentpage;        /* the page we are stuffing ucalls into */
    long *d_nextucall;          /* place to put the next ucall */
    long *d_ceiling;            /* end of free space in the current page */
    float d_srate;
} t_dspchain;

/* a ugen call is a function which you pass a (long *) and which
 * feeds you back the same long* advanced by the number of items that
 * got eaten.
 */
typedef long *(*t_ucall)(long *ptr);

/* these structures are here so that any object can go and look in a sample
   bugger, i mean buffer. */

typedef float t_sample;         /* samples are floating-point */

typedef struct sampbuf {
    t_sample *b_shit;           /* a floating-point vector of samples */
    float b_conv;               /* sample rate, in points per millisecond, of sample */
    long b_nsamps;              /* # of samples (excluding GUARDPTS zeros at end) */
} t_sampbuf;

typedef struct sigtable         /* object that takes care of a sampbuf */
{
    t_object x_ob;
    t_symbol *x_sym;            /* its name */
    t_sampbuf x_buf;            /* pointer to the sampbuf */
    t_symbol *x_vol;            /* volume (i.e. directory) */
    struct sigtable *x_next;    /* next sigtable hanging off this symbol */
} t_sigtable;

typedef short t_filesamp;       /* type of a sample in a soundfile */

typedef struct ctlf2p2z         /* control structure for 2p2z~ */
{
    float c_state1;             /* previous output */
    float c_state2;             /* previous previous output */
    float c_gain;               /* these three are c0, c1, c2 in help window */
    float c_fbcoef1;
    float c_fbcoef2;
    float c_ffcoef0;            /* these are d0, d1, d2 in help window */
    float c_ffcoef1;
    float c_ffcoef2;
} t_ctlf2p2z;

void cu_f2p2z(float *in, float *out, t_ctlf2p2z * x, int n);
void u_setup(void *x, int siginputs, int sigoutputs);
void u_stdout(void *x);
void u_clean(void *x);
void u_inletmethod(int inletnumber);
void u_genout(void *x, int vs);

void *dspchain_addstate(void *state, int size, int alignment);
void *dspchain_putstate(t_dspchain * x, void *shit, int size,
                        int alignment);

void dspchain_putc(t_dspchain * x, t_method f, int n, ...);
void dspchain_addc(t_method method, int nargs, ...);
void dspchain_add860(t_method method, int nargs, ...);
void dspchain_addugen(t_ucall f, int nargs, ...);
