#ifndef lint
static char rcsid[] =
    "$Revision: 2.2 $ IRCAM $Date: 1993/10/14 11:58:05 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.2 $ IRCAM $Date: 1993/10/14 11:58:05 $
 *
 * FTS by Miller Puckette
 *
 */

/* This file MUST be an ANSI-C pure code,
   because it must be able to run also on CPOS.
   
   See the nameValue.doc file for documentation.
   
   */

/* Default and internal configuration defines  */

#define MAX_TOKEN 256           /* The max length of an accepted token */
#define MAXPATHLEN  1024
/* Unix include */

#include <sys/types.h>          /* necessary with GNU CC CPOS */
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Local include */

#include "getNameValue.h"

/* Local MACROS */

/* Defines */

#define private
#define public

/* Configuration */

#ifndef I860
#ifdef NEXTSTEP
#define USE_NEXT_DEFAULT
#endif
#endif

/* BOOL Definition */

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef BOOL
#define BOOL int
#endif


/* Extern declarations */

extern int atoi(const char *);

/* Private  Type  declarations */

typedef struct _nameEnv {
    const char *key;
    const char *value;
    struct _nameEnv *next;
    int mode;
} nameEnvRec, *nameEnv;



/*
  
  Enable debug printing
  
  */

private int name_debug = 0;

public void enableErrorPrint()
{
    name_debug = 1;
}

/*********************************************************************************/
/*                                                                               */
/*           Environment Structure                                               */
/*                                                                               */
/*********************************************************************************/

/* The environment table is not built with 
   an hash table Class because this code must be ANSI C
   */

/* Constant for status */

#define EMPTY 0
#define REFERENCE 1
#define COPY_KEY  2
#define COPY_VALUE 3
#define COPY_BOTH  4


/*
  The following functions are intended to be user functions to build up
  dinamically a text environment.
  nameEnvNew()
  nameEnvAdd(env,key,value,mode)  mode declare wich string to copy, if any; the subsystem
  keep track of the allocation, and when its freed free only the copied structure.
  nameEnvGetValue(key)
  nameEnvFree() free the nameEnv.
  */

private nameEnv nameEnvNew()
{
    nameEnv env;

    env = (nameEnv) malloc(sizeof(nameEnvRec));
    env->key = (char *) 0;
    env->value = (char *) 0;
    env->next = (nameEnv) 0;
    env->mode = EMPTY;
    return env;
}


/* Changed to add values on top; in this way
   the first element of an env is always unused; 
   we should change the code ....
   */

private void
nameEnvAdd(nameEnv env, const char *key, const char *value, int mode)
{
    nameEnv lenv;


    /* Create and insert a new element */

    lenv = nameEnvNew();

    if (env->next == (nameEnv) 0)
        env->next = lenv;
    else
    {
        lenv->next = env->next;
        env->next = lenv;
    }

    /* Set the key */

    if ((mode == COPY_KEY) || (mode == COPY_BOTH))
    {
        char *p;

        p = (char *) malloc(strlen(key) + 1);
        strcpy(p, key);
        lenv->key = (const char *) p;
    }
    else
        lenv->key = key;

    /* Set the value */

    if ((mode == COPY_VALUE) || (mode == COPY_BOTH))
    {
        char *p;

        p = (char *) malloc(strlen(value) + 1);
        strcpy(p, value);
        lenv->value = (const char *) p;
    }
    else
        lenv->value = value;

    /* Set the mode */

    lenv->mode = mode;
}

/* Set the value: i.e. if the value already exist in the env overwrite the
   old value */

private void
nameEnvSet(nameEnv env, const char *key, const char *value, int mode)
{
    nameEnv lenv;

    /* First look for the value, if any */

    for (lenv = env; lenv; lenv = lenv->next)
        if ((lenv->mode != EMPTY) && (!strcmp(lenv->key, key)))
            break;

    /* If not found, create and insert a new element */

    if (!lenv)
    {
        lenv = nameEnvNew();
        if (env->next)
        {
            lenv->next = env->next;
            env->next = lenv;
        }
        else
            env->next = lenv;
    }

    /* Free the old value if not empty and if the value was copied */

    if ((env->mode == COPY_VALUE) || (env->mode == COPY_BOTH))
    {
        free((void *) (env->value));
        env->value = (char *) 0;
    }

    /* Set the key if needed */

    if (env->mode == EMPTY)
    {
        if ((mode == COPY_KEY) || (mode == COPY_BOTH))
        {
            char *p;

            p = (char *) malloc(strlen(key) + 1);
            strcpy(p, key);
            env->key = (const char *) p;
        }
        else
            env->key = key;
    }


    /* Set the new value */

    if ((mode == COPY_VALUE) || (mode == COPY_BOTH))
    {
        char *p;

        p = (char *) malloc(strlen(value) + 1);
        strcpy(p, value);
        env->value = (const char *) p;
    }
    else
        env->value = value;

    /* Set the mode */

    env->mode = mode;
}



private const char *nameEnvGetValue(nameEnv env, const char *key,
                                    nameEnv *nextEnv)
{


    /* First look for the value, if any */

    for (; env; env = env->next)
        if ((env->mode != EMPTY) && (!strcmp(env->key, key)))
        {
            if (nextEnv)
                *nextEnv = env->next;
            return env->value;
        }

    if (nextEnv)
        *nextEnv = (nameEnv) 0;
    return (const char *) 0;
}


private void nameEnvFree(nameEnv env)
{
    /* Free the tail */

    if (env->next)
    {
        nameEnvFree(env->next);
        env->next = 0;
    }

    /* Free the value (if needed)  */

    if ((env->mode == COPY_VALUE) || (env->mode == COPY_BOTH))
        free((void *) env->value);

    /* Free the key (if needed) */

    if ((env->mode == COPY_KEY) || (env->mode == COPY_BOTH))
        free((void *) env->value);

    /* Free the env itself */

    free((void *) env);
}

/******************************************************************************/
/*                                                                            */
/*           Argument line parsing                                            */
/*                                                                            */
/******************************************************************************/

/* The argument line parsing is done in two step: first,
   in the initialization phase, the options are inserted in the environment
   as STRINGS, and then in the access phase they are parsed in the 
   correct values; 
   The arguments are not copied; arguments without values (followed by a
   -<something>) are stored with value "TRUE".
   Options ("-<something>") corresponding to short names declared in the 
   argument table are converted to the long name.
   
   The function ignore any value that follow the first after an option.
   */

/*forward declaration: this function is defined together with the 
  arg table handling functions 
  */


private const char *expandArgName(argDesc * argTable, const char *name);

private const char *true = "YES";

private void
checkinArgs(nameEnv env, argDesc *argTable, int argc, char **argv)
{
    int i;

    /* To be compatible with the next, ignore NULL argv */

    i = 0;
    while (i < argc)
    {
        if (argv[i] && *argv[i] == '-')
        {
            /* Get a key, looking for the value */

            if ((i == argc - 1) || (*argv[i + 1] == '-'))
            {
                /* NO value, setting to TRUE (skip the '-' in the key) */
                nameEnvAdd(env, expandArgName(argTable, argv[i] + 1), true,
                           REFERENCE);
            }
            else
            {
                /* Value found, (skip the '-' in the key) */

                nameEnvAdd(env, expandArgName(argTable, argv[i] + 1),
                           argv[i + 1], REFERENCE);
                i++;
            }
        }
        i++;
    }
}


/*********************************************************************************/
/*                                                                               */
/*           Owner List handling                                                 */
/*                                                                               */
/*********************************************************************************/

/* This implementation allow a maximum number of owners, defined here MAX_OWNERS.
   The maximum length of the owner is 256; everything will screw up if the
   owner are longer than this.
   
   The "application owner" is the first owner specifed, and can be retrieved and set
   with special functions.
   */

#define MAX_OWNERS 16

private char *ownerTable[MAX_OWNERS];
private int ownerCount = 0;

private void setApplicationOwner(const char *owner)
{
    char *p;

    p = (char *) malloc(strlen(owner) + 1);
    strcpy(p, owner);
    ownerCount = 0;             /* reinit everything */
    ownerTable[ownerCount] = p;
    ownerCount++;
}

private void addOwner(const char *owner)
{
    char *p;

    if (ownerCount >= MAX_OWNERS)
        return;

    p = (char *) malloc(strlen(owner) + 1);
    strcpy(p, owner);
    ownerTable[ownerCount] = p;
    ownerCount++;
}

private const char *getApplicationOwner()
{
    return ownerTable[0];
}

private BOOL checkOwner(const char *owner)
{
    int i;

    for (i = 0; i < ownerCount; i++)
        if (!strcmp(owner, ownerTable[i]))
            return TRUE;

    return FALSE;
}

private char *makeOtherOwnersFiles()
{
    const char *fmt = "${ISPW}/%s.app/defs.%s:";
    char *p, buf[512];
    int i, len, base;


    /* Compute the needed lenght */

    len = 0;
    base = strlen(fmt) - 4;
    for (i = 1; i < ownerCount; i++)
        len = len + base + 2 * strlen(ownerTable[i]);
    len = len + 1;              /* last NUL */

    /* allocating and computing */

    p = (char *) malloc(len);
    *p = '\0';                  /* for the strcat */
    for (i = 1; i < ownerCount; i++)
    {
        sprintf(buf, fmt, ownerTable[i], ownerTable[i]);
        strcat(p, buf);
    }
    return p;
}

/* Parser for an column separated  list of owner. */

private void parseOwnerList(const char *ownerList)
{
    char buf[256];
    const char *p;
    char *q;

    if (!ownerList)
        return;

    q = buf;
    p = ownerList;

    while (1)
        if (*p == '\0')
        {
            *q = '\0';
            if (q != buf)
                addOwner(buf);
            return;
        }
        else if (*p == ':')
        {
            *q = '\0';
            if (q != buf)
                addOwner(buf);
            q = buf;
            p++;
        }
        else
            *(q++) = *(p++);
}



/******************************************************************************/
/*                                                                            */
/*           NEXT Default search                                              */
/*                                                                            */
/******************************************************************************/

#ifdef USE_NEXT_DEFAULT

extern char *NXGetDefaultValue(const char *owner, const char *property);

const char *getNextDefaultFor(const char *name)
{
    const char *p;
    int i;

    for (i = 0; i < ownerCount; i++)
    {
        p = NXGetDefaultValue(ownerTable[i], name);
        if (p)
            return p;
    }

    return (const char *) 0;
}

#endif

/*********************************************************************************/
/*                                                                               */
/*           Configuration file handling                                         */
/*                                                                               */
/*********************************************************************************/

/* This is the configuration file parser; it is, of course, implemented
   as finite state automata; it is called by the init function */

/* Modified to consider a '\n' as a single character token.
   The newline token is represented as a -127 (NLT0 followed by a NUL */

#define NLT   -127

#define WAITING    1
#define IN_COMMENT 2
#define IN_STRING  3
#define STRING_BACKSLASH 4
#define IN_TOKEN   5
#define BACKSLASH  6
#define FINAL_COMMENT 7
#define END        8

/* Return TRUE if a toekn as been read, FALSE if end of file  */

private BOOL readToken(FILE *f, char *buf, int max)
{
#define store(c)  buf[count++]=c
    int status, count;
    char c;


    status = WAITING;
    count = 0;

    while (status != END)
    {
        c = getc(f);
        if (feof(f))
            return FALSE;

        switch (status)
        {
        case WAITING:
            switch (c)
            {
            case '#':
                status = IN_COMMENT;
                break;
            case '"':
                status = IN_STRING;
                break;
            case '\\':
                status = BACKSLASH;
                break;
            case ' ':
            case '\t':
                status = WAITING;
                break;
            case '\n':
                store(NLT);
                store('\0');
                status = END;
                break;
            default:
                status = IN_TOKEN;
                store(c);
                break;
            }
            break;

        case IN_COMMENT:
            if (c == '\n')
            {
                store(NLT);
                store('\0');
                status = END;
            }
            else
                status = IN_COMMENT;
            break;

        case IN_STRING:
            switch (c)
            {
            case '"':
                status = END;
                store('\0');
                break;
            case '\n':
                ungetc(c, f);
                status = END;
                store('\0');
                break;
            default:
                status = IN_STRING;
                store(c);
                break;
            }
            break;

        case STRING_BACKSLASH:
            store(c);
            break;

        case IN_TOKEN:
            switch (c)
            {
            case ' ':
            case '\t':
                status = END;
                store('\0');
                break;
            case '\n':
                ungetc(c, f);
                status = END;
                store('\0');
                break;
            case '#':
                status = FINAL_COMMENT;
                break;
            default:
                status = IN_TOKEN;
                store(c);
                break;
            }
            break;

        case BACKSLASH:
            switch (c)
            {
            case 'n':
                status = IN_TOKEN;
                store('\n');
                break;
            case 't':
                status = IN_TOKEN;
                store('\t');
                break;
            default:
                status = IN_TOKEN;
                store(c);
                break;
            }
            break;

        case FINAL_COMMENT:
            if (c == '\n')
            {
                ungetc(c, f);
                status = END;
                store('\0');
            }
            else
                status = FINAL_COMMENT;
            break;
        }

        /* Check for the end of the buffer */

        if (count >= max)
        {
            status = END;
            buf[count - 1] = '\0';
        }
    }
    return TRUE;
}


/* Parser; modified to accept global names (i.e. without names),
   and to be more roubust with respect to syntax errors.
   report the syntax errors in the standard error.
   Implemented now as a finite state automata.
   */


/* status */

#define BEGIN_OF_LINE   1
#define FIRST_TOKEN     2
#define SECOND_TOKEN    3
#define THIRD_TOKEN     4
#define DISCARD         5
#define PARSER_END       6

#define newLineToken(t)  (((t)[0] == NLT) && ((t)[1] == '\0'))
#define printError(file, line)   if (name_debug) fprintf(stderr,"%s:%d: Error in configuration file.\n", file, line)

private BOOL readConfigFile(const char *filename, nameEnv env)
{
    FILE *f;
    int status;
    char v1[MAX_TOKEN];
    char v2[MAX_TOKEN];
    char v3[MAX_TOKEN];
    char currentToken[MAX_TOKEN];
    int lineCounter;
    BOOL r;

    f = fopen(filename, "r");
    if (!f)
        return FALSE;

    status = BEGIN_OF_LINE;
    lineCounter = 1;

    while (status != PARSER_END)
    {

        r = readToken(f, currentToken, MAX_TOKEN);

        switch (status)
        {
        case BEGIN_OF_LINE:
            if (!r)
                status = PARSER_END;
            else if (newLineToken(currentToken))
            {
                status = BEGIN_OF_LINE;
                lineCounter++;
            }
            else
            {
                status = FIRST_TOKEN;
                strcpy(v1, currentToken);
            }
            break;
        case FIRST_TOKEN:
            if (!r)
            {
                status = PARSER_END;
                printError(filename, lineCounter);
            }
            else if (newLineToken(currentToken))
            {
                status = BEGIN_OF_LINE;
                printError(filename, lineCounter);
                lineCounter++;
            }
            else
            {
                status = SECOND_TOKEN;
                strcpy(v2, currentToken);
            }
            break;

        case SECOND_TOKEN:
            if (!r)
            {
                status = PARSER_END;
                printError(filename, lineCounter);
            }
            else if (newLineToken(currentToken))
            {
                status = BEGIN_OF_LINE;
                nameEnvAdd(env, v1, v2, COPY_BOTH);
                lineCounter++;
            }
            else
            {
                status = THIRD_TOKEN;
                strcpy(v3, currentToken);
            }
            break;

        case THIRD_TOKEN:
            if (!r)
            {
                status = PARSER_END;
                printError(filename, lineCounter);
            }
            else if (newLineToken(currentToken))
            {
                status = BEGIN_OF_LINE;
                if (checkOwner(v1))
                    nameEnvAdd(env, v2, v3, COPY_BOTH);
                lineCounter++;
            }
            else
            {
                status = DISCARD;
                printError(filename, lineCounter);
            }
            break;

        case DISCARD:
            if (!r)
            {
                status = PARSER_END;
            }
            else if (newLineToken(currentToken))
            {
                status = BEGIN_OF_LINE;
                lineCounter++;
            }
            else
            {
                status = DISCARD;
            }
            break;
        }
    }

    fclose(f);

    return TRUE;
}



/*********************************************************************************/
/*                                                                               */
/*           Value Parsing                                                       */
/*                                                                               */
/*********************************************************************************/

private struct {
    const char *string;
    int value;
} boolValueTable[] = {
    { "YES", TRUE },
    { "Yes", TRUE },
    { "yes", TRUE },
    { "y", TRUE },
    { "Y", TRUE },
    { "OK", TRUE },
    { "Ok", TRUE },
    { "ok", TRUE },
    { "NO", FALSE },
    { "No", FALSE },
    { "no", FALSE },
    { "N", FALSE },
    { "n", FALSE },
    { "TRUE", TRUE },
    { "True", TRUE },
    { "true", TRUE },
    { "T", TRUE },
    { "FALSE", FALSE },
    { "False", FALSE },
    { "false", FALSE },
    { "F", FALSE },
    { (char *) 0, 0 }
};

private int i_ret;
private float f_ret;

void *parseValue(const char *value, int type)
{
    const char *p;

    switch (type)
    {
    case STRING_VALUE:
    case PATH_VALUE:
    case EX_PATH:
    case AB_PATH:
    case AB_EX_PATH:
        return (void *) value;

    case INTEGER_VALUE:
        /* First check the syntax */

        for (p = value; *p != '\0'; p++)
            if (((*p < '0') || (*p > '9')) && (*p != '-') && (*p != '+'))
                return (void *) 0;

        /* Then translate */

        i_ret = atoi(value);
        return (void *) &i_ret;

    case HEX_VALUE:
        /* Should check the syntax */
        i_ret = atoi(value);
        return (void *) &i_ret;

    case FLOAT_VALUE:
        /* Should check the syntax */
        f_ret = atof(value);
        return (void *) &f_ret;

    case BOOL_VALUE:
        {
            int i;

            for (i = 0; boolValueTable[i].string != (char *) 0; i++)
                if (!strcmp(boolValueTable[i].string, value))
                    return (void *) &(boolValueTable[i].value);
        }
        return (void *) 0;

    default:
        return (void *) 0;
    }
}


private char unparseBuffer[MAXPATHLEN];

public const char *unparseValue(void *value, int type)
{
    switch (type)
    {
    case STRING_VALUE:
    case PATH_VALUE:
    case EX_PATH:
    case AB_PATH:
    case AB_EX_PATH:
        return (const char *) value;

    case INTEGER_VALUE:
        sprintf(unparseBuffer, "%d", *((int *) value));
        return unparseBuffer;

    case HEX_VALUE:
        sprintf(unparseBuffer, "%x", *((int *) value));
        return unparseBuffer;

    case FLOAT_VALUE:
        sprintf(unparseBuffer, "%f", *((float *) value));
        return unparseBuffer;

    case BOOL_VALUE:
        return (*((int *) value) ? "TRUE" : "FALSE");

    default:
        return "";
    }
}


private BOOL fileCheck(const char *path, BOOL createFlag)
{
    struct stat s;
    int ret;

    ret = stat(path, &s);
    return (ret != -1);
}


/* TYpe checking */

private BOOL CheckValue(const char *value, int type)
{
    if (value == (char *) 0)
        return FALSE;

    switch (type)
    {
    case STRING_VALUE:
    case PATH_VALUE:
        return TRUE;

    case EX_PATH:
        return fileCheck(value, FALSE);

    case AB_PATH:
    case AB_EX_PATH:
        return ((*value == '/') && fileCheck(value, FALSE));

    case INTEGER_VALUE:
        for (; *value != '\0'; value++)
            if ((*value < '0') || (*value > '9'))
                return FALSE;
        return TRUE;

    case HEX_VALUE:
        for (; *value != '\0'; value++)
            if (((*value < '0') || (*value > '9'))
                && (*value != 'x') && (*value != 'a') && (*value != 'b')
                && (*value != 'c') && (*value != 'd') && (*value != 'e'))
                return FALSE;
        return TRUE;

    case FLOAT_VALUE:
        return TRUE;            /* don't check floats, at the moment */

    case BOOL_VALUE:
        {
            int i;

            for (i = 0; boolValueTable[i].string != (char *) 0; i++)
                if (!strcmp(boolValueTable[i].string, value))
                    return TRUE;
            return FALSE;
        }

    default:
        return FALSE;
    }

    return FALSE;
}



/*********************************************************************************/
/*                                                                               */
/*           Public functions                                                   */
/*                                                                               */
/*********************************************************************************/


/* This maximium should be largely enough */

#define MAX_ENV 64

#define UNIX_ENV  1
#define NEXT_DEFAULT  2
#define ENV    3
#define VOID_ENV    4
#define RUNTIME_ENV    5

private struct _envEntry {
    int type;                   /* can be: UNIX_ENV, NEXT_DEFAULT, ENV, VOID_ENV (ignore), RUNTIME_ENV */
    nameEnv env;                /* not NULL only if type == ENV */
} envTable[MAX_ENV];

private int envCount = 0;       /* The next env Entry to be filled */

/* Value access function */

/* This to variable store the current state of the search,
   to be used with getNextValue
   */

private int nextEnvIndex;       /* if < 0, no more env to search in */
private nameEnv nextEnvPointer; /* if NULL last env has been completely searched */
private char lastName[MAXPATHLEN];

public void *getNameValue(const char *name, void *default_value, int type)
{
    const char *value;
    int i;
    extern char *getenv(const char *name);
    char nameBuf[MAXPATHLEN];

    /* Expand the variable name, first */

    ExpandFileName(name, nameBuf);

    /* First get a Value */

    value = (char *) 0;
    for (i = (envCount - 1); (!value) && (i >= 0); i--)
    {
        /* Look for value in env i */

        /*
           printf("z3: i = %d\n", i);
           printf("z3z: %s\n", nameBuf);
           printf("type = %d\n", envTable[i].type);
           printf("z3a\n", nameBuf);
         */
        switch (envTable[i].type)
        {
        case UNIX_ENV:
            value = getenv(nameBuf);
            nextEnvPointer = (nameEnv) 0;
            break;

        case NEXT_DEFAULT:
#ifdef USE_NEXT_DEFAULT
            value = getNextDefaultFor(nameBuf);
#endif
            nextEnvPointer = (nameEnv) 0;
            break;

        case RUNTIME_ENV:
        case ENV:
            value =
                nameEnvGetValue(envTable[i].env, nameBuf, &nextEnvPointer);
            break;

        case VOID_ENV:
            break;
        }
    }

    /* Save the name, after all the recursive calls */

    strcpy(lastName, nameBuf);
    nextEnvIndex = i;

    /* Then parse it */

    if (!value)
        return default_value;
    else
    {
        /* Type checking */
        if (CheckValue(value, type))
            return parseValue(value, type);
        else
            return (void *) 0;
    }
}


/* 
  Found the next values for the last name given in getNameValue.
  Its result is undefined if called before a getNameValue.
  
  This function return default_value when there are no more values,
  and NULL if there is a value with a syntax error.
  So if you want to check errors, and to loop for all the values
  you have to use a not NULL default-value.
  */


public void *getNextValue(void *default_value, int type)
{
    const char *value;
    int i;
    extern char *getenv(const char *name);

    /* First look in the current environment, if any */

    if (nextEnvPointer)
        value = nameEnvGetValue(nextEnvPointer, lastName, &nextEnvPointer);
    else
        value = (char *) 0;

    /* Second, if the value has not been found, continue the loop in the env table */

    for (i = nextEnvIndex; (!value) && (i >= 0); i--)
    {
        /* Look for value in env i */

        switch (envTable[i].type)
        {
        case UNIX_ENV:
            value = getenv(lastName);
            nextEnvPointer = (nameEnv) 0;
            break;

        case NEXT_DEFAULT:
#ifdef USE_NEXT_DEFAULT
            value = getNextDefaultFor(lastName);
#endif
            nextEnvPointer = (nameEnv) 0;
            break;

        case RUNTIME_ENV:
        case ENV:
            value =
                nameEnvGetValue(envTable[i].env, lastName,
                                &nextEnvPointer);
            break;

        case VOID_ENV:
            break;
        }
    }

    nextEnvIndex = i;

    /* Then parse it */

    if (!value)
        return default_value;
    else
    {
        /* Type checking */
        if (CheckValue(value, type))
            return parseValue(value, type);
        else
            return (void *) 0;
    }
}


private void pushEnvironment(nameEnv env, int type)
{
    if (envCount <= MAX_ENV)
    {
        envTable[envCount].type = type;
        envTable[envCount].env = env;
        envCount++;
    }
}


public void setNameValue(const char *name, void *value, int type)
{
    /* First look if the run time env exist, otherwise create it */


    if ((envCount > 0) && (envTable[envCount - 1].type != RUNTIME_ENV))
        pushEnvironment(nameEnvNew(), RUNTIME_ENV);

    nameEnvSet(envTable[envCount - 1].env, name, unparseValue(value, type),
               COPY_VALUE);
}


/* Utilities */



/* Set to void all the envirnment of the given type */

private void voidEnvironment(int type)
{
    int i;

    for (i = 0; i < envCount; i++)
        if (envTable[i].type == type)
            envTable[i].type = VOID_ENV;
}

/* This function create a minimal default environment, containing
   the following variable defined:
   
   APPLICATION = the name of the application, i.e. the owner value after parsing.
   
   APPDIR      = the pathname of the application, either got as argument to initnameenv,
   or setted equal to ${ISPW}/${APPLICATION}.app
   
   ISPW        = set to /usr/local/ispw
   
   ISPW_VALUEPATH = ${HOME}/.defs:${APPDIR}/defs.application:${ISPW}/defs.machine:${ISPW}/defs.site
   */


private void makeBootEnv(const char *appdir)
{
    nameEnv env;
    char *p;
    int i;

    env = nameEnvNew();
    nameEnvAdd(env, "APPLICATION", getApplicationOwner(), COPY_VALUE);
    nameEnvAdd(env, "APPDIR",
               (appdir ? appdir : "${ISPW}/Apps/${APPLICATION}.app"),
               COPY_VALUE);
    nameEnvAdd(env, "ISPW", "/usr/local/ispw", REFERENCE);

    /* Add all the owners as values for OWNER, to be possibly read with 
       getNameValue/ getNextValue */

    for (i = 0; i < ownerCount; i++)
        nameEnvAdd(env, "OWNER", ownerTable[i], REFERENCE);

    p = makeOtherOwnersFiles();
    nameEnvAdd(env, "OWNERS_FILES", p, REFERENCE);

    nameEnvAdd(env, "ISPW_VALUEPATH",
               "$(HOME)/.defs:$(APPDIR)/defs.${APPLICATION}:${OWNERS_FILES}$(ISPW)/defs.machine:$(ISPW)/defs.site",
               REFERENCE);
    pushEnvironment(env, ENV);
}

/* The init function create and push a default environment, taking
   arguments as described in makeBootEnv,
   and then push the args, the Unix Env, the Next Dafault, and then
   the filename contained in the ISPW_VALUEPATH value at this
   point (after FILE_EXPANSION); the file name expansion is done
   file by file, *after* loading the previous file.
   
   If the owner start with '/' is considered the pathname of the application
   and the application name is taken from the filename, with any extension stripped.
   Otherwise the owner is considered the application name.
   
   If the user specify a not null argTable a check is perfomed on the arguments,
   before the initialization; if the check report an error FALSE is returned,
   otherwise it return TRUE.
   */



/* utility functions to analize a owner name */

private BOOL isOwnerList(const char *p)
{
    for (; *p != '\0'; p++)
        if (*p == ':')
            return TRUE;
    return FALSE;
}

private BOOL isPathName(const char *p)
{
    for (; *p != '\0'; p++)
        if ((*p == '.') || (*p == '/'))
            return TRUE;
    return FALSE;
}

private BOOL isAbsolutePathName(const char *p)
{
    return (*p == '/');
}


private BOOL
internalInitNameValue(const char *application, const char *appdir,
                      const char *ownerList, argDesc *argTable, int argc,
                      char **argv)
{
    char fileName[MAXPATHLEN];
    char expandedPath[16 * MAXPATHLEN];
    nameEnv env, argEnv;
    const char *p;
    char *q;

    if (application)
        setApplicationOwner(application);

    if (ownerList)
        parseOwnerList(ownerList);


    /* Build the bootstrap environment */

    makeBootEnv(appdir);        /* appdir can be NULL */

    /* Check the arguments, and return if there is a mistake */

    if (argTable)
        if (!checkArgs(getApplicationOwner(), argTable, argc, argv))
            return FALSE;

    /* Checkin the arguments */

    argEnv = nameEnvNew();
    checkinArgs(argEnv, argTable, argc, argv);
    pushEnvironment(argEnv, ENV);


    /* Pushing the NEXT Default env, if not specified otherwise
       (also if we are in CPOS, will be just ignored) */

    if (getBoolOption("NextDefault", TRUE))
    {
        pushEnvironment((nameEnv) 0, NEXT_DEFAULT);
    }
    /* Pushing the UNIX env, if not specified otherwise */

    if (getBoolOption("UnixEnv", TRUE))
        pushEnvironment((nameEnv) 0, UNIX_ENV);

    /* Read the ISPW_VALUEPATH variable and scan all the files */

    ExpandFileName(getStringOption("ISPW_VALUEPATH", (char *) 0),
                   expandedPath);
    p = expandedPath;
    while (*p != '\0')
        p++;

    /* Scan all the ISPW_VALEUPATH files */

    while (p != expandedPath)
    {
        BOOL b;

        q = fileName + MAXPATHLEN - 1;

        *q = '\0';
        q--;

        /* get the next pathname */

        while ((p != expandedPath) && (*p != ':'))
            *(q--) = *(p--);

        if (p == expandedPath)
            *q = *p;
        else
            q++;

        if (*p == ':')
            p--;

        /* Make a new env */

        env = nameEnvNew();
        if (name_debug)
            fprintf(stderr, "Reading file %s for owner %s.\n", q,
                    getApplicationOwner());

        /* Parse the file */

        b = readConfigFile(q, env);
        if (!b && name_debug)
            fprintf(stderr, "\tFile not found.\n");

        /* Push the env */

        pushEnvironment(env, ENV);
    }

    /* We repush the next, unix, and arg env in order to have them on top */

    /* Pushing the NEXT Default env, if not specified otherwise
       (also if we are in CPOS, will be just ignored) */

    if (getBoolOption("NextDefault", TRUE))
    {
        voidEnvironment(NEXT_DEFAULT);
        pushEnvironment((nameEnv) 0, NEXT_DEFAULT);
    }

    /* Pushing the UNIX env, if not specified otherwise */

    if (getBoolOption("UnixEnv", TRUE))
    {
        voidEnvironment(UNIX_ENV);
        pushEnvironment((nameEnv) 0, UNIX_ENV);
    }

    /* sorry - can't void the argEnv so easily. */
    pushEnvironment(argEnv, ENV);
    return TRUE;
}


/* PUbblic function: 
   initNameValue take a single argument, that can be a absolute path, 
   a relative path, the application name OR a owner List.
   
   initApplicationNameValue take a application argument, that should be 
   either an absolute path, a relative path or the application Name
   and a ownerList argument.
   */

public BOOL
initNameValue(const char *owner, argDesc *argTable, int argc, char **argv)
{
    if (isOwnerList(owner))
    {
        return internalInitNameValue((char *) 0, (char *) 0, owner,
                                     argTable, argc, argv);
    }
    else if (isAbsolutePathName(owner))
    {
        const char *p, *r;
        char *q;
        char applicationName[MAXPATHLEN];
        char appDir[MAXPATHLEN];

        p = owner;
        for (; *p != '\0'; p++);        /* Find the end of the string */
        for (; *p != '/'; p--); /* Find the last '/' */
        p++;

        q = applicationName;
        while ((*p != '\0') && (*p != '.'))
            *(q++) = *(p++);    /* Copy the filename without the extension */
        *q = '\0';

        q = appDir;
        r = owner;
        for (; *p != '/'; p--); /* Find the last '/' again */
        while (r != p)
            *(q++) = *(r++);    /* Copy the application directory */
        *q = '\0';

        return internalInitNameValue(applicationName, appDir, (char *) 0,
                                     argTable, argc, argv);
    }
    else if (isPathName(owner))
    {
        char applicationName[MAXPATHLEN];
        const char *p;
        char *q;

        /* A relative pathname, with or without '.' */

        p = owner;
        for (; *p != '\0'; p++);        /* Find the end of the string */
        for (; (*p != '/') && (p != owner); p--);       /* Find the last '/', or the beginning of the string */
        if (*p == '/')
            p++;

        q = applicationName;
        while ((*p != '\0') && (*p != '.'))
            *(q++) = *(p++);    /* Copy the filename without the extension */
        *q = '\0';

        return internalInitNameValue(applicationName, (char *) 0,
                                     (char *) 0, argTable, argc, argv);
    }
    else
    {
        return internalInitNameValue(owner, (char *) 0, (char *) 0,
                                     argTable, argc, argv);
    }
}


public BOOL
initApplicationNameValue(const char *owner, const char *ownerList,
                         argDesc *argTable, int argc, char **argv)
{
    if (isAbsolutePathName(owner))
    {
        const char *p, *r;
        char *q;
        char applicationName[MAXPATHLEN];
        char appDir[MAXPATHLEN];


        p = owner;
        for (; *p != '\0'; p++);        /* Find the end of the string */
        for (; *p != '/'; p--); /* Find the last '/' */
        p++;

        q = applicationName;
        while ((*p != '\0') && (*p != '.'))
            *(q++) = *(p++);    /* Copy the filename without the extension */
        *q = '\0';

        q = appDir;
        r = owner;
        for (; *p != '/'; p--); /* Find the last '/' again */
        while (r != p)
            *(q++) = *(r++);    /* Copy the application directory */
        *q = '\0';

        return internalInitNameValue(applicationName, appDir, ownerList,
                                     argTable, argc, argv);
    }
    else if (isPathName(owner))
    {
        char applicationName[MAXPATHLEN];
        const char *p;
        char *q;

        /* A relative pathname, with or without '.' */

        p = owner;
        for (; *p != '\0'; p++);        /* Find the end of the string */
        for (; (*p != '/') && (p != owner); p--);       /* Find the last '/', or the beginning of the string */
        if (*p == '/')
            p++;

        q = applicationName;
        while ((*p != '\0') && (*p != '.'))
            *(q++) = *(p++);    /* Copy the filename without the extension */
        *q = '\0';

        return internalInitNameValue(applicationName, (char *) 0,
                                     ownerList, argTable, argc, argv);
    }
    else
    {
        return internalInitNameValue(owner, (char *) 0, ownerList,
                                     argTable, argc, argv);
    }
}



/* UTILITY FUNCTION: shortcuts with fixed types */


public BOOL getBoolOption(char *name, BOOL default_value)
{
    BOOL *p;
    p = (BOOL *) getNameValue(name, (void *) &default_value, BOOL_VALUE);

    if (p)
        return *p;
    else
        return default_value;
}


public void setBoolOption(char *name, BOOL value)
{
    setNameValue(name, (void *) &value, BOOL_VALUE);
}


public int getIntOption(char *name, int default_value)
{
    int *p;

    p = (int *) getNameValue(name, (void *) &default_value, INTEGER_VALUE);

    if (p)
        return *p;
    else
        return default_value;

}

public int getNextIntOption(int default_value)
{
    int *p;

    p = (int *) getNextValue((void *) &default_value, INTEGER_VALUE);

    if (p)
        return *p;
    else
        return getNextIntOption(default_value); /* this recursive call is to skip syntax errors */
}

public void setIntOption(char *name, int value)
{
    setNameValue(name, (void *) &value, INTEGER_VALUE);
}


public float getFloatOption(char *name, float default_value)
{
    float *p;

    p = (float *) getNameValue(name, (void *) &default_value, FLOAT_VALUE);

    if (p)
        return *p;
    else
        return default_value;
}

public float getNextFloatOption(float default_value)
{
    float *p;

    p = (float *) getNextValue((void *) &default_value, FLOAT_VALUE);

    if (p)
        return *p;
    else
        return getNextFloatOption(default_value);
}

public void setFloatOption(char *name, float value)
{
    setNameValue(name, (void *) &value, FLOAT_VALUE);
}


public const char *getStringOption(char *name, const char *default_value)
{
    char *p;

    p = (char *) getNameValue(name, (void *) default_value, STRING_VALUE);

    if (p)
        return p;
    else
        return default_value;
}

/* Doesn't work if default_value is NULL */

public const char *getNextStringOption(const char *default_value)
{
    char *p;

    if (!default_value)
        return default_value;

    p = (char *) getNextValue((void *) default_value, STRING_VALUE);

    if (p)
        return p;
    else
        return getNextStringOption(default_value);
}

public void setStringOption(char *name, const char *value)
{
    setNameValue(name, (void *) value, STRING_VALUE);
}

public const char *getPathOption(char *name, const char *default_value)
{
    char *p;

    p = (char *) getNameValue(name, (void *) default_value, AB_EX_PATH);

    if (p)
        return p;
    else
        return default_value;
}

/* Doesn't work if default_value is NULL */

public const char *getNextPathOption(const char *default_value)
{
    char *p;

    if (!default_value)
        return default_value;

    p = (char *) getNextValue((void *) default_value, AB_EX_PATH);

    if (p)
        return p;
    else
        return getNextPathOption(default_value);
}


static char exBuf[MAXPATHLEN];

/* This two work as the previous one, but they expand the value; the expanded value
   is put in a static buffer !! copy it !!*/

public const char *getExStringOption(char *name, const char *default_value)
{
    char *p;

    p = (char *) getNameValue(name, (void *) default_value, STRING_VALUE);

    if (p)
    {
        ExpandFileName(p, exBuf);
        return exBuf;
    }
    else
        return default_value;
}


/* Doesn't work if default_value is NULL */

public const char *getNextExStringOption(const char *default_value)
{
    char *p;

    if (!default_value)
        return default_value;

    p = (char *) getNextValue((void *) default_value, STRING_VALUE);


    if (p)
    {
        ExpandFileName(p, exBuf);
        return exBuf;
    }
    else
        return getNextExStringOption(default_value);
}

public const char *getExPathOption(char *name, const char *default_value)
{
    char *p;

    p = (char *) getNameValue(name, (void *) default_value, AB_EX_PATH);

    if (p)
    {
        ExpandFileName(p, exBuf);
        return exBuf;
    }
    else
        return default_value;
}

/* Doesn't work if default_value is NULL */

public const char *getNextExPathOption(const char *default_value)
{
    char *p;

    if (!default_value)
        return default_value;

    p = (char *) getNextValue((void *) default_value, AB_EX_PATH);

    if (p)
    {
        ExpandFileName(p, exBuf);
        return exBuf;
    }
    else
        return getNextExPathOption(default_value);
}




/****************************************************/
/*                                                  */
/*     String espansion                             */
/*                                                  */
/****************************************************/

/****************************************************/
/*                                                  */
/*     PROTOCOL : - FILE NAME EXPANSION             */
/*                                                  */
/****************************************************/


/* 
  This function substitute in a string the substring ${name} with
  the value of "name".
  It is called ExpandFileName because it is mostly used for filename,
  but can really be used for any string.
  IT IS RECURSIVE: it continue to substitute variable names value,
  until possible.
  It return a static buffer, that must be copied.
  */


/* The function is implemented as a finite state automata */
/* A ${} (null name) substring is just stripped away,
   a not terminating ${.... sequence is stripped away,
   a stand-alone '$' is left in the string */

#define COPYING 1
#define DOLLAR 2
#define OPEN_P 3
#define IN_VAR 4
#define EX_END 5


private int ExpandFileNameStep(const char *filename, char *output)
{
    int expanded;
    int status;
    char varname[MAXPATHLEN];
    const char *pin;
    char *pout, *pvar;

    status = COPYING;
    pin = filename;
    pout = output;
    pvar = (char *) 0;          /* useless initialization */
    expanded = 0;

    while (status != EX_END)
    {
        switch (status)
        {
        case COPYING:
            if (*pin == '\0')
            {
                *(pout++) = *pin;
                status = EX_END;
            }
            else if (*pin == '$')
            {
                status = DOLLAR;
            }
            else
            {
                *(pout++) = *pin;
                status = COPYING;
            }
            break;
        case DOLLAR:
            if (*pin == '\0')
            {
                *(pout++) = '$';
                *(pout++) = *pin;
                status = EX_END;
            }
            else if ((*pin == '{') || (*pin == '('))
            {
                status = OPEN_P;
            }
            else
            {
                *(pout++) = '$';
                *(pout++) = *pin;
                status = COPYING;
            }
            break;
        case OPEN_P:
            if (*pin == '\0')
            {
                *(pout++) = *pin;
                status = EX_END;
            }
            else if ((*pin == '}') || (*pin == ')'))
            {
                status = COPYING;
            }
            else
            {
                pvar = varname;
                *(pvar++) = *pin;
                status = IN_VAR;
            }
            break;
        case IN_VAR:
            if (*pin == '\0')
            {
                *(pout++) = *pin;
                status = EX_END;
            }
            else if ((*pin == '}') || (*pin == ')'))
            {
                *(pvar++) = '\0';
                strcpy(pout, getNameValue(varname, "", STRING_VALUE));
                while (*pout != '\0')
                    pout++;
                expanded = 1;
                status = COPYING;
            }
            else
            {
                *(pvar++) = *pin;
                status = IN_VAR;
            }
            break;
        default:
            break;
        }
        pin++;
    }
    return expanded;
}


private int isExpandable(const char *p)
{
    for (; *p != '\0'; p++)
        if (*p == '$')
            return 1;
    return 0;
}

public void ExpandFileName(const char *filename, char *result)
{
    int ret;
    char *input, *output, *p;
    char buf[MAXPATHLEN];

    ret = TRUE;
    output = result;
    input = buf;                /* used ony for recursive steps */
    ret = ExpandFileNameStep(filename, output);

    while (isExpandable(filename) && ret)
    {
        p = input;
        input = output;
        output = p;

        ret = ExpandFileNameStep(input, output);
    }

    if (output != result)
        strcpy(result, output);
}



/****************************************************/
/*                                                  */
/*     ARGUMENT CHECKING                            */
/*                                                  */
/****************************************************/

/*
  Well, to use this function, the user must provide a
  pretty argument description table, that describe the arguments.
  All the checking is done here.
  A sample table can be:
  
  argDesc testArgTable[] =
  {
  {"Foo",0,0,MANDATORY,BOOL_VALUE, "to fool around the program"},
  {"Bar",0,"name",OPTIONAL, STRING_VALUE, "to barfify a name"},
  {"Zap",0,"name",INTERNAL, INTEGER_VALUE, "to zappify a name"},
  {"Help",0,0,HELP,0,0}, 
  {"Hidden",0,0,COMPLETE_HELP,0,0}, 
  { 0, 0,0,0,0,0}
  };
  
  Look at the documentation for more details.
  
  */

/* Forward declarations */

private void printUsage(const char *name, argDesc * argTable);

private argDesc *getArgDesc(const char *name, argDesc *argTable)
{
    argDesc *p;

    for (p = argTable; p->name != (char *) 0; p++)
        if ((p->name && (!strcmp(name, p->name))) ||
            (p->short_name && (!strcmp(name, p->short_name))))
            return p;

    return (argDesc *) 0;
}



private const char *describeType(int type)
{
    switch (type)
    {
    case STRING_VALUE:
        return "a string";

    case EX_PATH:
    case PATH_VALUE:
        return "a pathname";

    case AB_PATH:
    case AB_EX_PATH:
        return "an absolute pathname";

    case INTEGER_VALUE:
        return "an integer";
    case HEX_VALUE:
        return "a hex integer";
    case FLOAT_VALUE:
        return "a float";
    case BOOL_VALUE:
        return "a boolean";
    default:
        return "an unknow type";
    }
}


private void printHelp(const char *name, argDesc *argTable)
{
    argDesc *p;

    printUsage(name, argTable);

    printf("\nWhere :\n");
    p = argTable;

    while (p->name != (char *) 0)
    {
        const char *arg_name;

        if (p->value_type == BOOL_VALUE)
            arg_name = "[yes|no]";
        else
            arg_name = p->arg_name;


        switch (p->arg_type)
        {
        case MANDATORY:
            if (p->short_name)
                printf("\t-%s (or -%s) %s:\t%s (mandatory)\n", p->name,
                       p->short_name, arg_name, p->help);
            else
                printf("\t-%s %s:\t%s (mandatory)\n", p->name, arg_name,
                       p->help);
            break;
        case OPTIONAL:
            if (p->short_name)
                printf("\t-%s (or -%s) %s:\t%s\n", p->name, p->short_name,
                       arg_name, p->help);
            else
                printf("\t-%s %s:\t%s\n", p->name, arg_name, p->help);
            break;
        case HELP:
            if (p->short_name)
                printf("\t-%s (or -%s):\tto get this help text\n", p->name,
                       p->short_name);
            else
                printf("\t-%s:\tto get this help text\n", p->name);
        default:
            break;
        }
        p++;
    }
}

private void printCompleteHelp(const char *name, argDesc *argTable)
{
    argDesc *p;

    printUsage(name, argTable);

    printf("\nWhere :\n");
    p = argTable;

    while (p->name != (char *) 0)
    {
        const char *arg_name;

        if (p->value_type == BOOL_VALUE)
            arg_name = "[yes|no]";
        else
            arg_name = p->arg_name;

        switch (p->arg_type)
        {
        case MANDATORY:
            if (p->short_name)
                printf("\t-%s (or -%s) %s:\t%s (mandatory)\n", p->name,
                       p->short_name, arg_name, p->help);
            else
                printf("\t-%s %s:\t%s (mandatory)\n", p->name, arg_name,
                       p->help);
            break;
        case OPTIONAL:
            if (p->short_name)
                printf("\t-%s (or -%s) %s:\t%s\n", p->name, p->short_name,
                       arg_name, p->help);
            else
                printf("\t-%s %s:\t%s\n", p->name, arg_name, p->help);
            break;
        case INTERNAL:
            if (p->short_name)
                printf("\t-%s (or -%s) %s:\t%s (internal)\n", p->name,
                       p->short_name, arg_name, p->help);
            else
                printf("\t-%s %s:\t%s (internal)\n", p->name, arg_name,
                       p->help);
            break;
        case HELP:
            if (p->short_name)
                printf("\t-%s (or -%s):\tto get an help text\n", p->name,
                       p->short_name);
            else
                printf("\t-%s:\tto get an help text\n", p->name);
            break;
        case COMPLETE_HELP:
            if (p->short_name)
                printf("\t-%s (or -%s):\tto get this help text\n", p->name,
                       p->short_name);
            else
                printf("\t-%s:\tto get this help text\n", p->name);
            break;
        default:
            break;
        }
        p++;
    }
}

/* Dovrebbe tenere conto dello short name */
private void printUsage(const char *name, argDesc *argTable)
{
    argDesc *p, *helpArg;

    helpArg = (argDesc *) 0;

    printf("Usage: %s ", name);

    for (p = argTable; p->name != (char *) 0; p++)
    {
        const char *arg_name;

        if (p->value_type == BOOL_VALUE)
            arg_name = "[yes|no]";
        else
            arg_name = p->arg_name;

        switch (p->arg_type)
        {
        case HELP:
            helpArg = p;
            break;
        case OPTIONAL:
        case MANDATORY:
            if (p->arg_type == OPTIONAL)
                putchar('[');

            printf("-%s %s", (p->short_name ? p->short_name : p->name),
                   arg_name);

            if (p->arg_type == OPTIONAL)
                putchar(']');
            putchar(' ');
            break;
        default:
            break;
        }
    }
    putchar('\n');

    if (helpArg)
        printf("       %s -%s\n", name, helpArg->name);
}


/* Check arguments:
   return an int code reporting the result of the analisys.
   The result is non zero if args ar OK, and 0 if the
   application should not continue.
   
   Can be called as many time you want.
   */



public BOOL checkArgs(const char *appName, argDesc *argTable, int argc,
                      char **argv)
{
    int err, help, complete_help, i;
    argDesc *p;

    err = 0;
    help = 0;
    complete_help = 0;

    i = 0;

    while (i < argc)
    {
        p = getArgDesc(argv[i] + 1, argTable);  /* skip the '-' */

        if (p)
        {
            /* if it is mandatory, mark as satisfied */

            if (p->arg_type == MANDATORY)
                p->arg_type = SATISFIED;

            if (p->arg_type == HELP)
            {
                help = 1;
                if (i == 0)
                    break;      /* skip error analisys if -Help is the first option */
            }

            if (p->arg_type == COMPLETE_HELP)
            {
                complete_help = 1;
                if (i == 0)
                    break;      /* skip error analisys if -Help is the first option */
            }

            /* Check the value existance and types */

            if ((i == argc - 1) || (*argv[i + 1] == '-'))
            {
                /* No value for the option */

                if ((p->value_type == BOOL_VALUE) || (p->arg_type == HELP)
                    || (p->arg_type == COMPLETE_HELP))
                {
                    /* Boolean without arg, ok */
                    i++;
                }
                else
                {
                    /* Non boolean without args: error  */
                    fprintf(stderr,
                            "%s: option %s should have %s value.\n",
                            appName, argv[i], describeType(p->value_type));
                    err = 1;
                    i++;
                }
            }
            else
            {
                /* Option with value */
                void *v;

                v = parseValue(argv[i + 1], p->value_type);
                if (!v)
                {
                    fprintf(stderr,
                            "%s: incorrect value %s for option %s, should be %s.\n",
                            appName, argv[i + 1], argv[i],
                            describeType(p->value_type));
                    err = 1;
                }

                i += 2;
            }
        }
        else
        {
            /* Unknown parameter */

            fprintf(stderr, "%s: unknown option %s.\n", appName, argv[i]);
            err = 1;

            /* skip the option argument if any */

            if ((i != argc - 1) && (*argv[i + 1] != '-'))
                i++;

            /* next one */

            i++;
        }
    }

    /* Check fo unsatisfied mandatory */

    if (!(help || complete_help))
        for (p = argTable; p->name != (char *) 0; p++)
            if (p->arg_type == MANDATORY)
            {
                if (p->short_name)
                    fprintf(stderr,
                            "%s: should specify the mandatory -%s (or -%s) option.\n",
                            appName, p->name, p->short_name);
                else
                    fprintf(stderr,
                            "%s: should specify the mandatory -%s option.\n",
                            appName, p->name);
                err = 1;
            }

    /* Reset the satisfied to mandatory */

    for (p = argTable; p->name != (char *) 0; p++)
        if (p->arg_type == SATISFIED)
            p->arg_type = MANDATORY;

    /* print usage and/or help */

    if (complete_help)
        printCompleteHelp(appName, argTable);
    else if (help)
        printHelp(appName, argTable);
    else if (err)
        printUsage(appName, argTable);

    /* return the result */

    return (!(help || complete_help || err));

}



/* Look is a string corrspond to a short argument name
   in the arg table; if it does, return the corresponding long
   name, if not, return the name argument 
   */


private const char *expandArgName(argDesc *argTable, const char *name)
{
    argDesc *p;

    if (argTable)
        for (p = argTable; p->name != (char *) 0; p++)
            if (p->short_name && (!strcmp(name, p->short_name)))
                return p->name;

    return name;

}
