
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.2 $ IRCAM $Date: 1993/10/14 11:58:07 $
 *
 * FTS by Miller Puckette
 *
 */

/* Include this file if you want to use the getNameValue library */


/* Value type constants */

#define   STRING_VALUE         1
#define   PATH_VALUE       2
#define   EX_PATH     3
#define   AB_PATH     6
#define   AB_EX_PATH  7
#define   INTEGER_VALUE    10
#define   HEX_VALUE        11
#define   FLOAT_VALUE      12
#define   BOOL_VALUE       13

/*arg type constants */

#define MANDATORY 1
#define OPTIONAL  2
#define INTERNAL  3
#define HELP      4
#define COMPLETE_HELP 5

/* This value is used internally, and should be considered,
   by all the code apart checkArgs, as equivalent to MANDATORY
 */

#define SATISFIED 10

/* The argDesc structure, to buil argument description tables */

typedef struct _argDesc {
    const char *name;
    const char *short_name;
    const char *arg_name;
    int arg_type;
    int value_type;             /* same types as for getNameValue */
    const char *help;
} argDesc;


/* debug enable */

extern void enableErrorPrint();

/* Init function: only one of the two should be called */

extern int initNameValue(const char *owner, argDesc * argTable, int argc,
                         char **argv);
extern int initApplicationNameValue(const char *owner,
                                    const char *ownerList,
                                    argDesc * argTable, int argc,
                                    char **argv);

/* Value access function */

extern void *getNameValue(const char *name, void *default_value, int type);
extern void *getNextValue(void *default_value, int type);
extern void setNameValue(const char *name, void *value, int type);

/* Value parsing function */

extern void *parseValue(const char *value, int type);
extern const char *unparseValue(void *value, int type);

/* Argument check functions */

extern int checkArgs(const char *appName, argDesc * argTable, int argc,
                     char **argv);

/* Value access convenience functions */

extern int getBoolOption(char *name, int default_value);
extern void setBoolOption(char *name, int value);
extern int getIntOption(char *name, int default_value);
extern int getNextIntOption(int default_value);
extern void setIntOption(char *name, int value);
extern float getFloatOption(char *name, float default_value);
extern float getNextFloatOption(float default_value);
extern void setFloatOption(char *name, float value);
extern const char *getStringOption(char *name, const char *default_value);
extern const char *getNextStringOption(const char *default_value);
extern void setStringOption(char *name, const char *value);

/* To set a path variable, use the setStringOption anyway; 
   no check is perfomed in setting
 */

extern const char *getPathOption(char *name, const char *default_value);
extern const char *getNextPathOption(const char *default_value);
extern const char *getExStringOption(char *name,
                                     const char *default_value);
extern const char *getNextExStringOption(const char *default_value);
extern const char *getExPathOption(char *name, const char *default_value);
extern const char *getNextExPathOption(const char *default_value);
extern void ExpandFileName(const char *filename, char *result);
