
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/02/07 06:45:03 $
 *
 * FTS by Miller Puckette
 *
 */

typedef struct ftsTypeTable {
    /* Main setup */
    void (*h_init)(void);
    void (*h_getpost)(void);    /* This must be implemented also as ! animal */
    void (*h_eof)(void);
    int (*h_start)(void);
    void (*h_poll)(int secs);
    void (*h_dosend)(char *buf, int bytes);

    /* Run time miscellaneus */
    void (*h_ping)(void);
    void (*h_dosetthrottle)(int t1, int t2, int t3);
} *ftsType;

extern ftsType soxFts, standaloneFts, whichFts();
extern int h_nproc;             /* number of CPs */

void h_setconfig(char *s);
void h_eof();
int h_restart();
#ifdef ISPW
void h_getpost();
const char *h_getwherebooted();
void h_getispw(int nslot, int *dacType, int *in, int *out, int *nChans,
               int *EBU1, int *EBU2, const char **clockSource,
               int *backplane);
void h_setispw(int nslot, int dacType, int in, int out, int nChans,
               int EBU1, int EBU2, const char *clockSource, int backplane);
#endif
void h_setfts(ftsType fts);
char *h_getftsdir();
void h_setftsdir(const char *dir);
void h_setmax();

void h_vnew(void *hostid, int cpno, t_symbol * s, char *fmt, ...);
void *h_vnewunique(int cpno, t_symbol * s, char *fmt, ...);
void h_vmess(void *hostid, t_symbol * s, char *fmt, ...);

void h_new(void *hostid, int cpno, t_symbol * s, int ac, t_atom * av);
void *h_newunique(int cpno, t_symbol * s, int ac, t_atom * av);
void h_mess(void *hostid, t_symbol * s, int ac, t_atom * av);
void h_free(void *hostid);
void h_connect(void *id1, int n1, void *id2, int n2);
void h_disconnect(void *id1, int n1, void *id2, int n2);
int h_getstr(char **in, char *str, int maxlen);
int h_getint(char *str, int *vec, int *count);
void h_setthrottlestring(const char *str);
void h_setadvancesstring(const char *str);
void h_setverbose(int n);
int h_getverbose();
void h_restartdacs(int on);
float h_getrate();
void h_setrate(float rate);
void h_setchans(const int *vec);
void h_getchans(int *vec);
void h_setthrottle(int th1, int th2, int th3);
void h_getthrottle(int *th1, int *th2, int *th3);
void h_setadvances(int *vec, int n);
void h_getadvances(int *vec);
void h_setbootdir(const char *dir);
char *h_getbootdir();
void h_init();
char *t_config();

extern int h_dead;

enum h_platform {
    H_ISPW,                     /* first three are in order they appear in dialog window */
    H_UNIX,
    H_NONE
};

void h_sethardware(int which);  /* specify "hardware platform" */
int h_gethardware();

/* setup info for standalone case */
void h_setstd(int slot2, int slot4, int slot6, int version2);
void h_getstd(int *slot2, int *slot4, int *slot6, int *version2);

/* setup info for unix case */
void h_setunix(const char *command, int ncp);
void h_getunix(char **command, int *ncp);

void syserror(char *fmt, ...);
