#ifndef lint
static char rcsid[] =
    "$Revision: 2.11 $ IRCAM $Date: 1994/05/10 10:50:36 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.11 $ IRCAM $Date: 1994/05/10 10:50:36 $
 *
 * FTS by Miller Puckette
 *
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sgtty.h>
#include <fcntl.h>
#include <signal.h>

#include "m_mess.h"
#include "cp_mess.h"
#include "h_cp.h"

#define INITREADSIZE 8192
#define READSIZE 4096

extern int logfd;
extern int h_animal;
extern int h_dead;
extern int h_makinglogfile;
extern int h_useariel;
extern char *progname;

int infile;
int outfile;

static int readbytes, readalloc;
static char *readahead;

void h_setconfig(char *s)
{
}

void h_init_X()
{
    utils_init();

    if (h_makinglogfile)
    {
        if ((logfd = creat("/tmp/logfile", 0666)) < 0)
            fprintf(stderr, "warning: can't create log file\n");
    }
    else
        logfd = -1;

    post("- running on the cp -");

    if (!(readahead = (char *) malloc(INITREADSIZE)))
        hate("no space for readahead buffer\n");
    readalloc = INITREADSIZE;
    readbytes = 0;
}

void h_getpost_X()
{
    h_animal = 0;
}

int h_start_X()
{
    h_dead = 0;
    /* placeholder */
    return 0;
}

#define BUCKETSIZE 1024

void h_poll_X(int c)
{
    char buf[BUCKETSIZE];
    int x;
    int gotbytes;

    if (h_dead)
        return;

    /* service data that was read ahead while trying to write */

    if (readbytes)
    {
        fromsock(readahead, readbytes);
        readbytes = 0;
    }
    if ((gotbytes = read(infile, buf, BUCKETSIZE)) <= 0)
    {
        if (errno && (errno != EAGAIN))
        {
            perror("read");
            h_died();
        }
    }
    else
    {
        fromsock(buf, gotbytes);
    }
}

static int h_getfromcp(char *buf, int size, int secs)
{
    int x;
    int gotbytes;
    fd_set readset, writeset, exceptset;
    struct timeval timout;
    timout.tv_sec = secs;
    timout.tv_usec = 0;

    if (h_dead)
        return 0;
    FD_SET(infile, &readset);
    FD_SET(infile, &exceptset);
    x = select(infile + 1, &readset, &writeset, &exceptset, &timout);
    if (x < 0)
    {
        syserror("select failed\n");
        h_eof();
        return 0;
    }
    if (secs && !x)
    {
        post("timout");
        return 0;
    }
    if (FD_ISSET(0, &exceptset))
    {
        post("stdin exception!");
    }
    if (FD_ISSET(infile, &exceptset))
    {
        post("socket exception!");
    }
    if (FD_ISSET(infile, &readset))
    {
        gotbytes = read(infile, buf, size);
        if ((gotbytes <= 0) && (errno != EAGAIN))
        {
            perror("read");
            h_died();
            return 0;
        }
        else
        {
            if (gotbytes <= 0)
                return 0;
            else
                return gotbytes;
        }
    }
    else
        return 0;
}

static int sendtopipe(int fd, char *buf, int len)
{
    fd_set readset, writeset, exceptset;
    int sent = 0;
    while (1)
    {
        struct timeval tv;
        int didsomething = 0;
        FD_ZERO(&readset);
        FD_ZERO(&writeset);
        FD_ZERO(&exceptset);
        FD_SET(outfile, &writeset);
        FD_SET(infile, &readset);
        tv.tv_sec = 0;
        tv.tv_usec = 0;
        if (select((outfile > infile ? outfile : infile) + 1,
                   &readset, &writeset, &exceptset, &tv) < 0)
            perror("select");
        if (sent >= len)
            break;
        if (FD_ISSET(infile, &readset))
        {
            int nbytes;
            /* post("reading input pipe...\n"); */
            if (readbytes + READSIZE > readalloc)
            {
                readalloc += INITREADSIZE;
                if (!(readahead = (char *) realloc(readahead, readalloc)))
                    hate("no space for readahead buffer");
            }
            nbytes = read(infile, readahead + readbytes, READSIZE);
            if (nbytes <= 0)
            {
                if (nbytes < 0 && errno != EAGAIN)
                    return (-1);
            }
            else
            {
                readbytes += nbytes;
                didsomething = 1;
            }
        }
        if (FD_ISSET(outfile, &writeset))
        {
            int newsent = write(outfile, buf + sent, len - sent);
            if (newsent <= 0)
            {
                if (newsent < 0 && errno != EAGAIN)
                    return (-1);
            }
            else
            {
                sent += newsent;
                didsomething = 1;
            }
        }

        tv.tv_sec = 1;
        if (!didsomething)
        {
            microsleep(1000);
        }
    }
    return (0);
}

void h_dosend(char *buf, int bytes)
{
    if (sendtopipe(outfile, buf, bytes) < 0)
        perror("cp write"), h_died();
}

extern int pongwait;
extern void h_pong();

void h_ping_X()
{
    pongwait = 1;
    h_vmess((void *) 0, gensym("ping"), "l", h_pong);
    fromsock(readahead, readbytes);
    readbytes = 0;
    while (pongwait)
    {
        char buf[1];
        int ch = h_getfromcp(buf, 1, 0);
        if (h_dead)
            break;
        else if (ch > 0)
            doit(buf[0] & 0xff);
    }
}

void h_eof_X()
{
}

void cpos_setthrottle(int t1, int t2, int t3)
{
}

struct ftsTypeTable XFtsTable = {
    h_init_X,
    h_getpost_X,
    h_eof_X,
    h_start_X,
    h_poll_X,
    h_dosend,
    h_ping_X,
    cpos_setthrottle
};

ftsType XFts = &XFtsTable;

void h_restartdacs(int on)
{
    h_vmess(0, gensym("#_restartdacs_"), "l", (long) on);
}

void h_setchans(const int *vec)
{
}

void h_getchans(int *vec)
{
}

const char *h_getwherebooted()
{
    return ".";
}
