#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:56:48 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:56:48 $
 *
 * This file contains support for Animal, by implementing a class, "animal",
 * on the CP, which acts as an object with a dynamically-changeable class.
 * The comments here assume that you have read the fts documentation,
 * in particular the section on Animal.
 */

#include "m_mess.h"
#include "l_animal.h"
#include "l_animal2.h"
#include "cp_mess.h"

extern t_symbol
    s_graphic, s_rm, s_a_animal, s_, s_tick, s_midi, s_new, s_free,
    s_add, s_bind, s_unbind, s_send_float, s_rename,
    s_set, s_init, s_symbol, s_any_animal, s_send;


extern void *h_objlist;
void *animal_new(), *animal_copy(), animal_update(t_animal * x);
extern char *progname;
extern void *a_class_messlist;
t_maxclass *animal_messlist;
t_animal *animal_self;          /* "current instance" for animal methods in MAX */
extern void *newest;

/************************ animal **************************/

void animal_graphic(t_animal *x, long int fn, long int freefn,
                    long int ptr)
{
    hq_connect(x->a_hqgfx, fn, freefn, ptr);
}

/* utility routine to find an object, if any, bound to a name made
   by concatenating the two strings onto the animal's instance or
   class name. */

void *animal_target(t_animal *x, char *s2, char *s3)
{
    void *ret;
    char name[1024];
    if (x->a_instancename)
    {
        strcpy(name, x->a_instancename->s_name);
        strcat(name, s2);
        strcat(name, s3);
        if (ret = gensym(name)->s_thing)
            return (ret);
    }
    strcpy(name, x->a_class->c_name->s_name);
    strcat(name, s2);
    strcat(name, s3);
    return (gensym(name)->s_thing);
}

/* update a named field; called from "anything" if the unmatched message
   starts with "update" */

void animal_update_x(t_animal *x, t_symbol *s, int ac, t_atom *av)
{
    t_field *fp = x->a_class->c_fields;
    int count = x->a_class->c_nitems;
    void **vp = x->a_bag->x_x;
    t_symbol *fs;
    if (ac < 1)
    {
        error("animal_update_%s: no argument", fs->s_name);
        return;
    }
    fs = gensym(s->s_name + 7); /* field name after "update_" prefix */
    for (; count--; fp++, vp++)
    {
        if (fp->f_name == fs)
        {
            if (fp->f_type == &s_float)
            {
                if (av->a_type == A_FLOAT)
#ifdef __alpha
                    *(double *) vp = av->a_w.w_float;
#else
                    *(float *) vp = av->a_w.w_float;
#endif
                else if (av->a_type == A_LONG)
                    *(float *) vp = av->a_w.w_long;
                else
                    error("animal_update_%s: bad argument type",
                          fs->s_name);
            }
            else if (fp->f_type == &s_int)
            {
                if (av->a_type == A_FLOAT)
                    *(long *) vp = av->a_w.w_float;
                else if (av->a_type == A_LONG)
                    *(long *) vp = av->a_w.w_long;
                else
                    error("animal_update_%s: bad argument type",
                          fs->s_name);
            }
            else if (fp->f_type == &s_symbol)
            {
                if (av->a_type == A_SYM)
                    *(t_symbol **) vp = av->a_w.w_sym;
            }
            else
                error("animal_update_%s: bad argument  type", fs->s_name);
            animal_update(x);
            return;
        }
    }
    error("%s: animal_update_%s: no such field",
          x->a_class->c_name->s_name, fs->s_name);
}

/* set a slot of an animal */

void animal_set_or_init(t_animal *x, t_symbol *s, int ac, t_atom *av)
{
    t_field *tp;
    void *vp;
    long onset;
    t_symbol *type;
    if (ac < 2 || av[0].a_type != A_LONG)
    {
        error("animal_set: bag arguments\n");
        return;
    }
    onset = av[0].a_w.w_long;
    if (onset < 0 || onset >= x->a_class->c_nitems)
        error("animal_set: bad offset");
    tp = &x->a_class->c_fields[onset];
    type = tp->f_type;
    if (type == &s_float)
    {
        if (av[1].a_type == A_LONG)
            *((float *) (x->a_bag->x_x + onset)) = av[1].a_w.w_long;
        else if (av[1].a_type == A_FLOAT)
#ifdef __alpha
            *((double *) (x->a_bag->x_x + onset)) = av[1].a_w.w_float;
#else
            *((float *) (x->a_bag->x_x + onset)) = av[1].a_w.w_float;
#endif
        else
            error("animal_set: type doesn't match field");
    }
    else if (type == &s_int)
    {
        if (av[1].a_type == A_LONG)
            *((long *) (x->a_bag->x_x + onset)) = av[1].a_w.w_long;
        else if (av[1].a_type == A_FLOAT)
            *((long *) (x->a_bag->x_x + onset)) = av[1].a_w.w_float;
        else
            error("animal_set: type doesn't match field");
    }
    else if (type == &s_any_animal)
    {
        if (av[1].a_type == A_OBJ)
            *((void **) (x->a_bag->x_x + onset)) = av[1].a_w.w_obj;
        else
            error("animal_set: type doesn't match field");
    }
    else if (type == &s_symbol)
    {
        if (av[1].a_type == A_SYM)
            *((t_symbol **) (x->a_bag->x_x + onset)) = av[1].a_w.w_sym;
        else
            error("animal_set: type doesn't match field");
    }
    else
        error("animal_set: unknown type");
    if (vp = animal_target(x, (s == &s_set ? "_set_" : "_init_"),
                           tp->f_name->s_name))
    {
        t_animal *selfwas = animal_self;
        animal_self = x;
        if (type == &s_int)
            mess1(vp, &s_int, av[1].a_w.w_long);
        else if (type == &s_float)
            mess1(vp, &s_float, av[1].a_w.w_float);
        else if (type == &s_symbol)
            mess1(vp, &s_symbol, av[1].a_w.w_sym);
        animal_self = selfwas;
    }
}

void animal_bind(t_animal *x, t_symbol *s)      /* set an animal's name (bind it to a symbol). */
{
    s->s_thing = (t_object *) x;
    x->a_instancename = s;
}

void animal_unbind(t_animal *x, t_symbol *s)    /* unbind it */
{
    if (s->s_thing == (t_object *) x)
        s->s_thing = 0;
    x->a_instancename = 0;
}

/* send a message to an instance variable (which must be "any_animal") */

void animal_send(t_animal *x, t_symbol *s, int ac, t_atom *av)
{
    t_field *fp = x->a_class->c_fields;
    int count = x->a_class->c_nitems;
    void **vp = x->a_bag->x_x;
    t_symbol *fs;
    if (ac < 1 || av[0].a_type != A_SYM)
    {
        error("animal_send: bad arguments");
        return;
    }
    fs = av[0].a_w.w_sym;
    ac--;
    av++;
    for (; count--; fp++, vp++)
    {
        if (fp->f_name == fs)
        {
            if (fp->f_type == &s_any_animal)
            {
                if (ac == 0)
                    mess0(*(void **) vp, &s_bang);
                else if (av->a_type == A_LONG)
                {
                    if (ac == 1)
                        mess1(*(void **) vp, &s_int, av->a_w.w_long);
                    else
                        typedmess(*(void **) vp, &s_list, ac, av);
                }
                else if (av->a_type == A_FLOAT)
                {
                    if (ac == 1)
                        mess1(*(void **) vp, &s_float, av->a_w.w_float);
                    else
                        typedmess(*(void **) vp, &s_list, ac, av);
                }
                else if (av->a_type == A_SYM)
                    typedmess(*(void **) vp, av->a_w.w_sym, ac - 1,
                              av + 1);
            }
            else
                error("self_send: %s: sorry, animals only", fs->s_name);
            return;
        }
    }
    error("animal_send: %s: no field named %s",
          x->a_class->c_name->s_name, fs->s_name);
}

void animal_link_class(t_animal *x, t_a_class *cl)      /* put an animal in its class's inst list */
{
    x->a_class = cl;
    x->a_next = cl->c_animal;
    if (x->a_next)
        x->a_next->a_prev = x;
    x->a_prev = 0;
    cl->c_animal = x;
}

void animal_unlink_class(t_animal *x)   /* remove an animal from its class's inst list */
{
    t_a_class *c = x->a_class;
    x->a_class = 0;
    if (c->c_animal == x)
        c->c_animal = x->a_next;
    if (x->a_next)
        x->a_next->a_prev = x->a_prev;
    if (x->a_prev)
        x->a_prev->a_next = x->a_next;
}

void animal_hqfn(t_animal *x, void *hq) /* copy into hq, swapping bytes */
{
    int i, nitems = x->a_class->c_nitems;
    void **ap, **bp;
    void *buf[HQ_MAX];
    for (i = 0, ap = x->a_bag->x_x, bp = buf; i < nitems; i++, ap++, bp++)
        *bp = *ap, cvt_long(bp);
    hq_doit(hq, buf, nitems * sizeof(void *));
}

void animal_update(t_animal *x)
{
    hq_set(x->a_hqgfx);
}

/************ I/O for animals: attach to midi or DAC frame clock *****/

animal_dotick(t_animal *x, t_method fun)
{
    (*fun) (x->a_bag);
}

/* byte renamed in byte_in, because byte is a type in CPOS */

animal_domidi(t_animal *x, t_method fun, int byte_in)
{
    (*fun) (x->a_bag, byte_in);
}

animal_installio(t_animal *x)
{
    t_method fun;
    if ((fun = zgetfn(x->a_bag, &s_midi)))
        sched_installmidi(animal_domidi, x, fun);
}

animal_uninstallio(t_animal *x)
{
    t_method fun;
    if ((fun = zgetfn(x->a_bag, &s_midi)))
        sched_uninstallmidi(x);
}

/*
  Any other message is being sent from Max.  It can be defined by
  Animal (in C) but then no arguments are passed; or it can be defined
  back in Max, in any of 3 (count 'em) ways:
  first, search for an object named instance-name_selector
  second, for class-name_selector;
  if either is found send the rest of the message to that;
  third, see if the message is "update_field-name" and call animal_update_x.
  */

void animal_anything(t_animal *x, t_symbol *s, int ac, t_atom *av)
{
    t_method fn = zgetfn(x->a_bag, s);
    void *handler;
    if (fn)
        (*fn) (x->a_bag);
    else if (handler = animal_target(x, "_", s->s_name))
    {
        t_animal *selfwas = animal_self;
        animal_self = x;
        if (!ac)
            mess0(handler, &s_bang);
        else if (av[0].a_type == A_LONG)
        {
            if (ac == 1)
                mess1(handler, &s_int, av[0].a_w.w_long);
            else
                typedmess(handler, &s_list, ac, av);
        }
        else if (av[0].a_type == A_FLOAT)
        {
            if (ac == 1)
                mess1(handler, &s_float, av[0].a_w.w_float);
            else
                typedmess(handler, &s_list, ac, av);
        }
        else if (av[0].a_type == A_SYM)
            typedmess(handler, av[0].a_w.w_sym, ac - 1, av + 1);
        else
            bug("animal_anything");
        animal_self = selfwas;
    }
    else if (!strncmp(s->s_name, "update_", 7))
        animal_update_x(x, s, ac, av);
    else
        error("%s: no method to %s", x->a_class->c_name->s_name,
              s->s_name);
}

void animal_ff(t_animal *x)
{
    int i, nitems = x->a_class->c_nitems;
    t_symbol **tp;
    void **item;
    t_animal *a2;
    t_method fn;

    animal_uninstallio(x);
    if (fn = zgetfn(x->a_bag, &s_free))
        (*fn) (x->a_bag);
    freebytes(x->a_bag,
              sizeof(t_animal_header) +
              x->a_class->c_nitems * sizeof(void *));
    hq_free(x->a_hqgfx);
    animal_unlink_class(x);
}

void c_animal()
{
    c_new(&s_a_animal, animal_ff, sizeof(t_animal), 1, 0);

    c_addmess(animal_set_or_init, &s_set, A_GIMME, 0);
    c_addmess(animal_set_or_init, &s_init, A_GIMME, 0);
    c_addmess(animal_graphic, &s_graphic, A_LONG, A_LONG, A_LONG, 0);
    c_addmess(animal_bind, &s_bind, A_SYM, 0);
    c_addmess(animal_unbind, &s_unbind, A_SYM, 0);
    c_addmess(animal_send, &s_send, A_GIMME, 0);
    c_addmess(animal_anything, &s_anything, A_GIMME, 0);
}

void *a_animal_new(t_a_class *cl)
{
    t_animal *x;
    int nitems, i;
    t_method fn;
    if (ob_messlist(cl) != ((t_messlist *) a_class_messlist))
    {
        error("%s: not a class", ob_name(cl));
        return (0);
    }
    x = (t_animal *) obj_new((t_messlist **) & animal_messlist, c_animal);
    x->a_keygfx = 0;
    animal_link_class(x, cl);
    nitems = cl->c_nitems;
    x->a_bag = (t_ax *) getbytes(sizeof(t_animal_header) +
                                 nitems * sizeof(void *));
    c_setclass(x->a_bag, cl->c_messlist);
    for (i = 0; i < nitems; i++)
    {
        if (cl->c_fields[i].f_type == &s_symbol)
            x->a_bag->x_x[i] = &s_;
        else
            x->a_bag->x_x[i] = 0;
    }
    x->a_bag->x_header.h_backptr = x;
    x->a_parent = 0;

    x->a_hqgfx = hq_new(0L, 0L, 0L, x,
                        animal_hqfn, sizeof(void *) * nitems, currentPort);
    if (fn = zgetfn(x->a_bag, &s_new))
        (*fn) (x->a_bag);
    animal_installio(x);
    return (x);
}
