#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1994/01/29 07:48:19 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1994/01/29 07:48:19 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
#include "cp_mess.h"

extern t_symbol s__portal, s_;

/* "portal".  Non-patchable object which talks to 3rd-party apps */
#define MAXPORTALSIZE 10
#define HQARGS 5

typedef struct portal {
    t_tinyobject p_ob;
    void *p_hq;
    t_symbol *p_name;
    t_symbol *p_callback;
    int p_argc;
    t_atom p_argv[MAXPORTALSIZE];
} t_portal;


void portal_int(t_portal *x, long int n)
{
    if (!x->p_hq)
        return;
    if (x->p_argv[0].a_type == A_LONG)
        x->p_argv[0].a_w.w_long = n;
    else
        x->p_argv[0].a_w.w_float = n;
    hq_set(x->p_hq);
}

void portal_float(t_portal *x, double f)
{
    if (!x->p_hq)
        return;
    if (x->p_argv[0].a_type == A_LONG)
        x->p_argv[0].a_w.w_long = f;
    else
        x->p_argv[0].a_w.w_float = f;
    hq_set(x->p_hq);
}

void portal_list(t_portal *x, t_symbol *s, int ac, t_atom *av)
{
    int i;
    t_atom *ap;
    if (!x->p_hq)
        return;
    if (ac > x->p_argc)
        ac = x->p_argc;
    for (i = 0, ap = x->p_argv; i < ac; i++, ap++, av++)
    {
        if (ap->a_type == A_LONG)
            ap->a_w.w_long = (av->a_type == A_LONG ? av->a_w.w_long :
                              (av->a_type ==
                               A_FLOAT ? (long) (av->a_w.w_float) : 0));
        else
            ap->a_w.w_float = (av->a_type == A_LONG ?
                               (float) (av->a_w.w_long) :
                               (av->a_type ==
                                A_FLOAT ? av->a_w.w_float : 0.));

    }
    hq_set(x->p_hq);
}

void portal_anything(t_portal *x, t_symbol *s, int ac, t_atom *av)
{
    void *y = s->s_thing;
    if (y)
    {
        if (!ac)
            obj_bang(y);
        else
        {
            int type = av[0].a_type;
            if (type == A_SYM)
                typedmess(y, av[0].a_w.w_sym, ac - 1, av + 1);
            else if (ac > 1)
                typedmess(y, &s_list, ac, av);
            else if (type == A_FLOAT)
                obj_float(y, av[0].a_w.w_float);
            else
                obj_int(y, av[0].a_w.w_long);
        }
    }
    else
        error("%s: no such object", s->s_name);
}

void portal_hqfn(t_portal *x, void *hq) /* copy into hq buffer, swapping bytes */
{
    int i, nitems = x->p_argc;
    long buf[MAXPORTALSIZE], *bp;
    t_atom *ap;
    for (i = 0, ap = x->p_argv, bp = buf; i < nitems; i++, ap++, bp++)
        *bp = ap->a_w.w_long, cvt_long(bp);
    hq_doit(hq, buf, nitems * sizeof(long));
    if (x->p_callback && x->p_callback->s_thing)
        obj_bang(x->p_callback->s_thing);
}

void portal_ff(t_portal *x)
{
    if (x->p_hq)
    {
        hq_free(x->p_hq);
        if (x->p_name->s_thing == (t_object *) x)
            x->p_name->s_thing = 0;
    }
}

t_maxclass *portal_class;
void portal_setup()
{
    c_new(&s__portal, portal_ff, sizeof(t_portal), 1, 0);
    c_addint(portal_int);
    c_addfloat(portal_float);
    c_addmess(portal_list, &s_list, A_GIMME, 0);
    c_addmess(portal_anything, &s_anything, A_GIMME, 0);
}

void *portal_new(t_symbol *s, int ac, t_atom *av)
{
    t_portal *x;
    if (ac)
    {
        if (ac < HQARGS || av[0].a_type != A_SYM || av[1].a_type != A_SYM
            || av[2].a_type != A_LONG || av[3].a_type != A_LONG
            || av[4].a_type != A_LONG)
        {
            error("portal: bad arguments for return path");
            return (0);
        }
        if (ac > HQARGS + MAXPORTALSIZE)
        {
            error("portal: max number of elements is %d", MAXPORTALSIZE);
            return (0);
        }
    }
    x = (t_portal *) obj_new((t_messlist **) & portal_class, portal_setup);
    if (ac)
    {
        int i;
        t_atom *ap;
#ifdef ISPW
        x->p_hq = hq_new(av[2].a_w.w_long, av[3].a_w.w_long,
                         av[4].a_w.w_long, x,
                         portal_hqfn, sizeof(long) * (ac - HQARGS),
                         (ClientConnection) 1);
#else
        x->p_hq = hq_new(av[2].a_w.w_long, av[3].a_w.w_long,
                         av[4].a_w.w_long, x,
                         portal_hqfn, sizeof(long) * (ac - HQARGS),
                         currentPort);
#endif
        x->p_name = av[0].a_w.w_sym;
        if (x->p_name->s_thing)
            error("portal: %s: already exists", x->p_name);
        else
            x->p_name->s_thing = (t_object *) x;
        if ((x->p_callback = av[1].a_w.w_sym) == &s_)
            x->p_callback = 0;
        ac -= HQARGS;
        av += HQARGS;
        x->p_argc = ac;
        for (ap = x->p_argv; ac--; ap++, av++)
        {
            if (av->a_type != A_LONG && av->a_type != A_FLOAT)
            {
                error("portal: numbers only");
                ap->a_type = A_LONG;
            }
            else
                ap->a_type = av->a_type;
            ap->a_w.w_long = 0;
        }
    }
    else
        x->p_hq = 0;
    x->p_name->s_thing == (t_object *) x;
    return (x);
}
