#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:57:39 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1994/01/04 17:57:39 $
 *
 * The remote-send ("rsend") and remote-receive ("rreceive") objects.
 * 
 * The usage is as follows.  The host creates an object to receive and
 * one or more to send; the receiving object will own an "rreceive" and
 * the senders will own "rsend"s.  The following order must be followed:
 * 
 * 1. host creates the receiver.
 * 2. host creates an "rreceive" and sends its ID to the receiver.
 * 3. host tells the sender(s) about the "rreceive"'s host ID so that
 * the senders can tell their "rsends" about it.  (The host may wait
 * until now to create the sender and send it the rreceive's ID as a
 * creation argument, if desired.)
 * 
 * To add a sender to an existing receiver, the following must be done in
 * order:
 * 
 * 1.  host sends the receiver a message causing it to call rreceive_addone()
 * 2.  host tells the sender about the receiver.
 * 
 * To get rid of a connection, the following may occur in any order:
 * 
 * a.  The senders free their "rsends" with rsend_free;
 * b.  The receiver frees its "rreceive" with rreceive_bag (which delays the
 * free until all the sends send "break" messages to the rreceive.)
 * 
 */

#include "m_mess.h"
#include "cp_mess.h"
#include "d_graph.h"
#include "d_ugen.h"

extern t_symbol s_rreceive, s_put, s_;
extern void *h_objlist;

void *sigio_new(float **where, int wflag, int cpno);

typedef struct rreceive {
    t_head r_head;
    t_sig *r_io[IN0 + OUT1];
    int r_count;
    void *r_target;
    void *r_sigio;
    float *r_rbuf;
    int r_sig;
} t_rreceive;


void rreceive_bag(t_rreceive *x)
{
    if (x->r_target && x->r_sig)
        outlet_rm(x->r_head.h_ob.o_outlet, x->r_target);
    x->r_target = 0;
    if (!x->r_count)
        freeobject((t_object *) x);
}

void rreceive_anything(t_rreceive *x, t_symbol *s, int ac, t_atom *av)
{
    if (x->r_target)
        typedmess(x->r_target, s, ac, av);
}

rreceive_to(t_rreceive *x, void *target)
{
    if (x->r_target && x->r_sig)
        outlet_rm(x->r_head.h_ob.o_outlet, x->r_target);
    x->r_target = target;
    if (target && x->r_sig)
        outlet_add(x->r_head.h_ob.o_outlet, target);
}

/* In this file the variable id has been renamed to
   id_o, because id is a predefined cpos type */

rreceive_break(long int id_o)
{
    t_rreceive *x = (t_rreceive *) alist_get(h_objlist, id_o);
    if (!CP_LOCAL(x))
        bug("rreceive_break");
    else
    {
        if (x->r_sigio)
        {
            sigio_free(x->r_sigio);
            x->r_sigio = 0;
        }
        x->r_count--;
        if (!x->r_count && !x->r_target)
        {
            freeobject((t_object *) x);
        }
    }
}

rreceive_signal(long int id_o)
{
    t_rreceive *x = (t_rreceive *) alist_get(h_objlist, id_o);
    if (!CP_LOCAL(x))
        bug("rreceive_break");
    else
        x->r_sigio = sigio_new(&x->r_rbuf, 0, 0);
}

void cu_rreceive(float **inp, float *out, long int n)
{
#ifdef ZAPPO
    /* This version is to cope with the copy_in accessing in read one byte more
       than the requested */

    float *in = *inp;
    if (in)
    {
        n--;
        copy_in(in, out, n << 2);
        out[n] = in[n];
    }

    /* while (n--) *out++ = *in++; */
#else
    float *in = *inp;
    if (in)
        copy_in(in, out, n << 2);
    /* while (n--) *out++ = *in++; */
#endif
}

void rreceive_put(t_rreceive *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        dspchain_putc(u_chain, cu_rreceive, 3,
                      &x->r_rbuf, x->r_io[0]->s_shit, x->r_io[0]->s_n);
        if (u_debug)
            post("rreceive %lx: (%lx)", x, x->r_io[0]->s_shit);
    }
}

void rreceive_addone(t_rreceive *x)
{
    x->r_count++;
}

void rreceive_ff(t_rreceive *x)
{
    if (x->r_sig)
    {
        u_clean((t_ugen *) x);
        rreceive_to(x, 0);
    }
}

t_maxclass *rreceive_class;
void rreceive_setup()
{
    c_new(&s_rreceive, rreceive_ff, sizeof(t_rreceive), 1, 0);
    c_addmess(rreceive_anything, &s_anything, A_GIMME, 0);
    c_addmess(rreceive_put, &s_put, A_CANT, 0);
}

void *rreceive_new(long int n, long int sig)
/* number of sends that will connect */
/* whether to do a signal */
{
    t_rreceive *x =
        (t_rreceive *) obj_new((t_messlist **) & rreceive_class,
                               rreceive_setup);
    x->r_target = 0;
    x->r_count = n;
    x->r_sig = sig;
    x->r_sigio = 0;
    x->r_rbuf = 0;
    if (sig)
    {
        u_setup((t_ugen *) x, IN0, OUT1);
    }
    return (x);
}

typedef struct rsend {
    t_head s_head;
    t_sig *s_io[IN1 + OUT0];
    long s_whoto;
    int s_cpno;
    float *s_rbuf;
    void *s_sigio;
} t_rsend;

void rsend_bang(t_rsend *x)
{
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, &s_bang, 0, NIL);
}

void rsend_int(t_rsend *x, long int n)
{
    t_atom a;
    a.a_type = A_LONG;
    a.a_w.w_long = n;
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, &s_int, 1, &a);
}

void rsend_float(t_rsend *x, double f)
{
    t_atom a;
    a.a_type = A_FLOAT;
    a.a_w.w_float = f;
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, &s_float, 1, &a);
}

void rsend_symbol(t_rsend *x, t_symbol *s)
{
    t_atom a;
    a.a_type = A_SYM;
    a.a_w.w_sym = s;
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, &s_symbol, 1, &a);
}

void rsend_list(t_rsend *x, t_symbol *s, int ac, t_atom *av)
{
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, &s_list, ac, av);
}

void rsend_anything(t_rsend *x, t_symbol *s, int ac, t_atom *av)
{
    rtfifo_sendmess(x->s_cpno, CP_MESS, x->s_whoto, s, ac, av);
}

void cu_rsend(float *in, float **outp, long int n)
{
    float *out = *outp;
    if (out)
        copy_in(in, out, n << 2);
    /* while (n--) *out++ = *in++; */
}

void rsend_put(t_rsend *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        dspchain_putc(u_chain, cu_rsend, 3,
                      x->s_io[0]->s_shit, &x->s_rbuf, x->s_io[0]->s_n);
        if (u_debug)
            post("rsend %lx: (%lx)", x, x->s_io[0]->s_shit);
    }
}

int rsend_getcp(t_rsend *x)
{
    return (x->s_cpno);
}

void rsend_free(t_rsend *x)
{
    if (x->s_sigio)
    {
        u_clean((t_ugen *) x);
        sigio_free(x->s_sigio);
    }
    rtfifo_fcall(x->s_cpno, rreceive_break, x->s_whoto);
    freeobject((t_object *) x);
}

t_maxclass *rsend_class;
void rsend_setup()
{
    c_new(&s_, NIL, sizeof(t_rsend), 1, 0);
    c_addbang(rsend_bang);
    c_addint(rsend_int);
    c_addfloat(rsend_float);
    c_addmess(rsend_list, &s_list, A_GIMME, 0);
    c_addmess(rsend_anything, &s_anything, A_GIMME, 0);
    c_addsymbol(rsend_symbol);
    u_inletmethod(0);
    c_addmess(rsend_put, &s_put, A_CANT, 0);
}

void *rsend_new(int cpno, long int id_o, long int sig)
{
    t_rsend *x =
        (t_rsend *) obj_new((t_messlist **) & rsend_class, rsend_setup);
    if (sig)
    {
        u_setup((t_ugen *) x, IN1, OUT0);
        x->s_sigio = sigio_new(&x->s_rbuf, 1, cpno);
        rtfifo_fcall(cpno, rreceive_signal, id_o);
    }
    else
        x->s_sigio = 0;
    x->s_whoto = id_o;
    x->s_cpno = cpno;
    return (x);
}
