#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/01/04 17:57:42 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/01/04 17:57:42 $
 *
 * This complicated set of classes (through, send, receive, value)
 * maintains a set of inter-CP links to connect a "send" on one CP to
 * a "receive" on another (nothing is currently done for "value" which
 * is per-CP).  "send" and "receive" get "#connect" messages from the
 * host to set the host ID of the shared "through" object.  (all sends and
 * receives on a CP for a given symbol share one "through" object which keeps
 * reference counts.)
 */

#define MATRESET "MATRESET"     /* debug */
#include "m_mess.h"
#include "l_list.h"
void *rsend_new(int cpno, long int id_o, long int sig);
extern void *h_objlist;
extern t_symbol s_send, s_through, s_receive, s_value, s__connect;

void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);
void *dummyfn();

/* "through".  inlet chained straight to outlet.
   The Through is created on the CP when the first "send" or "receive" or
   "value" for a given symbol is created.  The host later must connect to
   it (i,e, suggest a host ID for it) and send it the ID of an rreceive to
   use; this is done by sending "connect" to a send or receive.  ("value"
   does nothing remotely and so doesn't matter here.)
   */

typedef struct through {
    struct object t_ob;
    struct symbol *t_sym;
    struct atom t_atom;
    t_sll *t_rsendlist;
    void *t_rreceive;
} t_through;

typedef struct send {
    struct object s_ob;
    struct through *s_through;
} t_send;

typedef struct receive {
    struct object r_ob;
    struct through *r_through;
} t_receive;

extern t_maxclass *send_class, *receive_class;


void through_bang(t_through *x)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_bang(z->l_item);
    outlet_bang(x->t_ob.o_outlet);
}

void through_int(t_through *x, long int n)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_int(z->l_item, n);
    outlet_int(x->t_ob.o_outlet, n);
}

void through_float(t_through *x, double f)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_float(z->l_item, f);
    outlet_float(x->t_ob.o_outlet, f);
}

void through_symbol(t_through *x, t_symbol *s)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_symbol(z->l_item, s);
    outlet_symbol(x->t_ob.o_outlet, s);
}

void through_list(t_through *x, t_symbol *s, int ac, t_atom *av)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_list(z->l_item, s, ac, av);
    outlet_list(x->t_ob.o_outlet, s, ac, av);
}

void through_anything(t_through *x, t_symbol *s, int ac, t_atom *av)
{
    t_sll *z;
    for (z = x->t_rsendlist; z; z = z->l_next)
        rsend_anything(z->l_item, s, ac, av);
    outlet_anything(x->t_ob.o_outlet, s, ac, av);
}

void through_connect(t_through *x, long int remote, long int n)
/* ID of remote object to connect to (0 if break or receive) */
/* CP it's on (or -1 to receive if remote == 0) */
{
    if (remote)                 /* make a new connection for sending */
    {
        void *s = rsend_new((int) n, remote, 0L);
        x->t_rsendlist = sll_insert(x->t_rsendlist, s);
    }
    else if (n >= 0)            /* break a connection for sending */
    {
        t_sll stub, *s2, *s3;
        stub.l_next = x->t_rsendlist;
        for (s2 = &stub; s3 = s2->l_next; s2 = s3)
        {
            if (rsend_getcp(s3->l_item) == n)
            {
                s2->l_next = s3->l_next;
                rsend_free(s3->l_item);
                sll_free(s3);
                goto success;
            }
        }
        bug("through_connect 1");
      success:
        x->t_rsendlist = stub.l_next;
    }
    else                        /* tell receiver to expect a new sender */
    {
        if (x->t_rreceive)
            rreceive_addone(x->t_rreceive);
        else
            bug("through_connect 3: %s", x->t_sym->s_name);
    }
}

void through_ff(t_through *x)
{
    t_sll *l;
    if (x->t_rreceive)
        rreceive_bag(x->t_rreceive);
    x->t_sym->s_thing = 0;
    for (l = x->t_rsendlist; l; l = l->l_next)
    {
        rsend_free(l->l_item);
        sll_free(l);
    }
}

t_maxclass *through_class;
void through_setup()
{
    c_new(&s_through, through_ff, sizeof(t_through), 0, 0);
    c_addbang(through_bang);
    c_addint(through_int);
    c_addfloat(through_float);
    c_addmess(through_list, &s_list, A_GIMME, 0);
    c_addmess(through_anything, &s_anything, A_GIMME, 0);
    c_addsymbol(through_symbol);
    c_addmess(through_connect, &s__connect, A_LONG, A_LONG, 0);
}

/* A "through" object.  Binds itself to the name given as argument.  Sets
   up for receiving messages from other CPs.  */

void *through_new(t_symbol *s, void *w)
/* an rreceive object to use */
{
    t_through *x =
        (t_through *) obj_new((t_messlist **) & through_class,
                              through_setup);
    outlet(x, NIL);
    x->t_sym = s;
    x->t_atom.a_type = A_LONG;
    x->t_atom.a_w.w_long = 0;
    x->t_rsendlist = 0;
    s->s_thing = (t_object *) x;
    x->t_rreceive = w;
    rreceive_to(w, x->t_ob.o_outlet);
    return (x);
}

/* "send". Send messages commonly. */

void send_bang(t_send *x)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_bang(z->l_item);
    outlet_bang(x->s_through->t_ob.o_outlet);
}

void send_int(t_send *x, long int n)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_int(z->l_item, n);
    outlet_int(x->s_through->t_ob.o_outlet, n);
}

void send_float(t_send *x, double f)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_float(z->l_item, f);
    outlet_float(x->s_through->t_ob.o_outlet, f);
}

void send_symbol(t_send *x, t_symbol *s)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_symbol(z->l_item, s);
    outlet_symbol(x->s_through->t_ob.o_outlet, s);
}

void send_list(t_send *x, t_symbol *s, int ac, t_atom *av)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_list(z->l_item, s, ac, av);
    outlet_list(x->s_through->t_ob.o_outlet, s, ac, av);
}

void send_anything(t_send *x, t_symbol *s, int ac, t_atom *av)
{
    t_sll *z;
    for (z = x->s_through->t_rsendlist; z; z = z->l_next)
        rsend_anything(z->l_item, s, ac, av);
    outlet_anything(x->s_through->t_ob.o_outlet, s, ac, av);
}

void send_connect(t_send *x, t_through *t)
{
    if (ob_messlist(t) != through_class)
        bug("send_connect");
    x->s_through = t;
}

void send_ff(t_send *x)
{
}

t_maxclass *send_class;
void send_setup()
{
    c_new(&s_send, send_ff, sizeof(t_send), 0, 0);
    c_addbang(send_bang);
    c_addint(send_int);
    c_addfloat(send_float);
    c_addmess(send_list, &s_list, A_GIMME, 0);
    c_addmess(send_anything, &s_anything, A_GIMME, 0);
    c_addsymbol(send_symbol);
    c_addmess(send_connect, &s__connect, A_OBJ, 0);
}

void *send_new(t_symbol *s)
{
    t_through *t;
    t_send *x =
        (t_send *) obj_new((t_messlist **) & send_class, send_setup);
    x->s_through = 0;
    return (x);
}

/* "receive".  Receive messages commonly. */

void receive_connect(t_receive *x, t_through *t)
{
    if (ob_messlist(t) != through_class)
        bug("receive_connect");
    x->r_through = t;
    outlet_add(t->t_ob.o_outlet, x->r_ob.o_outlet);
}

void receive_ff(t_receive *x)
{
    t_through *t = x->r_through;
    outlet_rm(t->t_ob.o_outlet, x->r_ob.o_outlet);
}

t_maxclass *receive_class;
void receive_setup()
{
    c_new(&s_receive, receive_ff, sizeof(t_receive), 0, 1);
    c_addmess(receive_connect, &s__connect, A_OBJ, 0);
}

void *receive_new(t_symbol *s)
{
    t_receive *x =
        (t_receive *) obj_new((t_messlist **) & receive_class,
                              receive_setup);
    x->r_through = 0;
    outlet(x, NIL);
    return (x);
}

/* "value". remember a value. */

typedef struct value {
    struct object v_ob;
    struct through *v_through;
} t_value;

void value_bang(t_value *x)
{
    t_through *t = x->v_through;
    if (t->t_atom.a_type == A_LONG)
        outlet_int(x->v_ob.o_outlet, t->t_atom.a_w.w_long);
    else if (t->t_atom.a_type == A_FLOAT)
        outlet_float(x->v_ob.o_outlet, t->t_atom.a_w.w_float);
    else
        outlet_symbol(x->v_ob.o_outlet, t->t_atom.a_w.w_sym);
}

void value_int(t_value *x, long int n)
{
    t_through *t = x->v_through;
    t->t_atom.a_type = A_LONG;
    t->t_atom.a_w.w_long = n;
}

void value_float(t_value *x, double f)
{
    t_through *t = x->v_through;
    t->t_atom.a_type = A_FLOAT;
    t->t_atom.a_w.w_float = f;
}

void value_symbol(t_value *x, t_symbol *s)
{
    t_through *t = x->v_through;
    t->t_atom.a_type = A_SYM;
    t->t_atom.a_w.w_sym = s;
}

void value_connect(t_value *x, t_through *t)
{
    if (ob_messlist(t) != through_class)
        bug("value_connect");
    x->v_through = t;
}

void value_ff(t_value *x)
{
}

t_maxclass *value_class;
void value_setup()
{
    c_new(&s_value, value_ff, sizeof(t_value), 0, 0);
    c_addbang(value_bang);
    c_addint(value_int);
    c_addfloat(value_float);
    c_addsymbol(value_symbol);
    c_addmess(value_connect, &s__connect, A_OBJ, 0);
}

void *value_new(t_symbol *s)
{
    t_through *t;
    t_value *x =
        (t_value *) obj_new((t_messlist **) & value_class, value_setup);
    outlet(x, &s_int);
    return (x);
}
