#ifndef lint
static char rcsid[] =
    "$Revision: 2.12 $ IRCAM $Date: 1994/05/10 10:51:01 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.12 $ IRCAM $Date: 1994/05/10 10:51:01 $
 *
 * FTS by Miller Puckette
 *
 */

#include <errno.h>

/* <a> removed unused functions; made locking version of binbuf_eval */
#include <stdio.h>
#include <stdarg.h>

#include "m_mess.h"

extern struct symbol s_print, s_new, s__Pat, s_, s_table, s__New, s_patcher,    /* <a> s_patcher added */
 s_int, s_float, s_v2, s_max;

#define FILEMAGIC 1             /* binbuf file magic number */
int binbuf_compat;              /* compatibility with old-style int and float objects */
int binbuf_error;
extern char lockout;
union ptr                       /* the real one's in m_mess.c */
{
    char *p_charstar;
    long *p_longstar;
#ifdef __alpha
    double *p_floatstar;
#else
    float *p_floatstar;
#endif
    struct symbol **p_sym;
    struct object **p_obj;
};

/* the "binbuf" structure.  A binary encoding of an array of atoms. For
   space efficiency (at least on "regular" CPUs) the types are all stored
   together followed by all the (union word sized) values.  The type
   array is zero-terminated.  For antique reasons, the types and the data
   are all stored in the same "handle".  "b_argc" gives the number of types,
   "b_onset" the onset in the array of the first datum (must be at least
   b_argc+1) and b_size is the total size of the array. */

typedef struct binbuf {
    char **b_shit;
    long b_onset;
    long b_argc;
    long b_size;
} t_binbuf;

void *evalbuf(char **buf, long int onset, int ac, struct atom *av,
              struct object *messto);

enum readerrortype { R_NONE = 0, R_UNIX, R_MEM, R_NOTMAX, R_OTHER };
static enum readerrortype whicherror;
char *readerrstring = "";
extern int errno;

int binbuf_insq(struct binbuf *x, int argc, struct atom *argv,
                int morespace);
int binbuf_toatoms(struct binbuf *x, t_atom * ap);
int prbuf(char **buf, long int onset);

void binbuf_ouch(char *s, ...)
{
    va_list args;
    char whatouch[BIGSTRING], *line2;
    va_start(args, s);
    vsnprintf(whatouch, s, args);
    line2 = whatouch + strlen(whatouch);
    switch (whicherror)
    {
    case R_NONE:
        sprintf(line2, "");
        break;
    case R_UNIX:
        sprintf(line2, "\n(%s)\n", strerror(errno));
        break;
    case R_MEM:
        sprintf(line2, "\n(out of memory)\n");
        break;
    case R_OTHER:
        sprintf(line2, "\n%s\n", readerrstring);
        break;
    default:
        bug("binbuf_ouch 2");
    }
    ouch(whatouch);
}

binbuf_printerror(char *s, int a, int b, int c, int d)
{
    char whatouch[1024];
    error(whatouch, s, a, b, c, d);
    switch (whicherror)
    {
    case R_NONE:
        sprintf(whatouch, "");
        break;
    case R_UNIX:
        sprintf(whatouch, "\n(%s)\n", strerror(errno));
        break;
    case R_MEM:
        sprintf(whatouch, "\n(out of memory)\n");
        break;
    case R_OTHER:
        sprintf(whatouch, "\n%s\n", readerrstring);
        break;
    default:
        bug("binbuf_ouch 2");
    }
    post(whatouch);
}

binbuf_setouch(char *s)
{
    readerrstring = s;
}

t_symbol *t_getlastopenvolume();

readtohandle(char *name, t_symbol *volume, char ***h, long int *sizep)
/* c string */
{
    int fd;
    long size;
    char **space;
    whicherror = R_NONE;
    if ((fd = open_via_path(name, volume, 0)) < 0)
    {
        whicherror = R_UNIX;
        return (0);
    }
    defvolume = t_getlastopenvolume();
    getfilelength(fd, &size);
    if (!(space = (char **) t_newhandle(size)))
    {
        whicherror = R_MEM;
        return (0);
    }
    HLock(space);
    if (FSRead((short) fd, &size, *space) != 0)
    {
        whicherror = R_UNIX;
        DisposHandle(space);
        FSClose(fd);
        return (0);
    }
    HUnlock(space);
    *h = space;
    *sizep = size;
    FSClose((short) fd);
    return (1);
}

/* if the string t contains special characters, backslash them into string s */
slash(char *s, char *t)
{
    char c, *t2 = t;
    while (c = *(t2++))
    {
        if (c == ';' || c == ',' || c == '$' || c == '\\')
        {
            while (c = *(t++))
            {
                if (c == ';' || c == ',' || c == '$' || c == '\\')
                    *(s++) = '\\';
                *(s++) = c;
            }
            *s = '\0';
            return (1);
        }
    }
    return (0);
}

/* strip backslashes from a string (but \\ goes to \.) */
unslash(struct symbol **spp)
{
    char xx[256];
    char *s, *t, c;
    t = (*spp)->s_name;
    while (c = *(t++))
        if (c == '\\')
        {
            int slash = 0;
            t = (*spp)->s_name;
            s = xx;
            while (c = *(t++))
            {
                if (slash)
                    *(s++) = c, slash = 0;
                else if (c == '\\')
                    slash = 1;
                else
                    *(s++) = c;
            }
            *s = '\0';
            *spp = gensym(xx);
            return;
        }
}

void *binbuf_bang(struct binbuf *x)
{
    return (evalbuf(x->b_shit, x->b_onset, 0, NIL, 0L));
}

void *binbuf_int(struct binbuf *x, long int n)
{
    struct atom av;
    av.a_type = A_LONG;
    av.a_w.w_long = n;
    return (evalbuf(x->b_shit, x->b_onset, 1, &av, 0L));
}

void *binbuf_float(struct binbuf *x, double f)
{
    struct atom av;
    av.a_type = A_FLOAT;
    av.a_w.w_float = f;
    return (evalbuf(x->b_shit, x->b_onset, 1, &av, 0L));
}

void *binbuf_eval(void *xx, int ac, struct atom *av, void *to)
{
    t_binbuf *x = xx;
    return (evalbuf(x->b_shit, x->b_onset, ac, av, to));
}

void *binbuf_make(void *xx, int ac, struct atom *av, void *to)
{
    t_binbuf *x = xx;
    return (evalbuf(x->b_shit, x->b_onset, ac, av, to));
}

binbuf_chkload(struct binbuf *x)
{
    char *shit = *(x->b_shit);
    struct symbol *s;
    if (*shit != A_SYM)
        return (0);
    s = *((struct symbol **) (shit + x->b_onset));
    return (s == &s_max || s == &s__New || s == &s_new || s == &s_table);
}

/* convert a binbuf to a handle full of longs.  If
   it succeeds, sets *hp to a new handle and *np to the number
   of items in it.  S is the symbol the binbuf should start with,
   like "table" or "funbuff".
   */
binbuf_gethandle(struct binbuf *x, char ***hp, long int *np,
                 struct symbol *s)
{
    char *shit = *(x->b_shit);
    if (*shit == A_SYM && *((struct symbol **) (shit + x->b_onset)) == s)
    {
        long n = 0, n2, *lp1, *lp2;
        char **contents, *cp;
        while (*(++shit) == A_LONG)
            n++;
        if (*shit)
            warn("extra baggage after table");
        if (!(contents = (char **) t_newhandle(n * sizeof(long))))
            hate("no space for table");
        lp1 =
            (long *) (*(x->b_shit) +
                      (x->b_onset + sizeof(struct symbol *)));
        lp2 = (long *) (*contents);
        for (n2 = n; n2--;)
            *(lp2++) = *(lp1++);
        *np = n;
        *hp = contents;
        return (0);
    }
    return (1);
}

extern void *newest;
void *tabfromhandle();

/* special version for loading which checks for a table */
void binbuf_loadeval(struct binbuf *x, int ac, struct atom *av, void *to)
{
    char **handle;
    long n;
    /* binbuf_print(x); */
    if (!binbuf_gethandle(x, &handle, &n, &s_table))
        newest = tabfromhandle(handle, n);
    else
        evalbuf(x->b_shit, x->b_onset, ac, av, to);
}

#define FUDIT 256               /* chunk size for alloc growing */

binbuf_text(struct binbuf *x, char **shit, long int n)
{
    struct atom abuf[FUDIT], *ap = abuf;        /* arg tmp space */
    int tc = 0;                 /* # items in temp */
    char outstr[1024];          /* stuff for read1 */
    long eaten = 0;
    binbuf_clear(x);
    binbuf_error = 0;
    while (1)
    {
        int gotone = read1(outstr, shit, &eaten, n, ap++, outstr + 1020);
        if (gotone)
            tc++;
        if (tc >= FUDIT || (!gotone && tc))
        {
            binbuf_insq(x, tc, abuf, (n > 50));
            if (binbuf_error)
            {
                binbuf_clear(x);
                return (1);
            }
            tc = 0;
            ap = abuf;
        }
        if (!gotone)
            break;
    }
    return (0);
}

binbuf_totext(struct binbuf *x, char **shit, long int *sizep)
{
    long size = 0, newsize;
    long tp, dp;
    int n, space, newspace = 0;
    char xx[256], *s1, *s2, *str;
    int handlesize = 0;
    for (tp = 0, dp = x->b_onset; tp < x->b_argc; tp++)
    {
        union ptr p;
        int rats, rats2;
        p.p_charstar = (*x->b_shit) + dp;
        space = newspace;
        newspace = 1;
        switch ((*x->b_shit)[tp])
        {
            char c;
        case A_LONG:
            sprintf(xx, "%ld", *p.p_longstar);
            dp += sizeof(long);
            break;
        case A_FLOAT:
#ifdef __alpha
            sprintf(xx, "%lf", *p.p_floatstar);
            dp += sizeof(double);
#elif defined(I860)             /* hack -- avoid sprintf which is broken for floats */
            t_ftoa((double) (*p.p_floatstar), xx);
            dp += sizeof(float);
#else
            sprintf(xx, "%f", *p.p_floatstar);
            dp += sizeof(float);
#endif
            /* eat trailing zeros */
            for (rats = 0, s2 = xx; *s2; s2++)
                if (*s2 == '.')
                    rats = 1;
            if (rats)
                for (s2--; *s2 == '0'; s2--)
                    *s2 = '\0';
            break;
        case A_SYM:
            str = (*p.p_sym)->s_name;
            if (!slash(xx, str))
                strcpy(xx, str);
            dp += sizeof(char *);
            break;
        case A_OBJ:
            sprintf(xx, "(%s)", ob_name(*p.p_obj));
            dp += sizeof(char *);
            break;
        case A_SEMI:
            strcpy(xx, ";\n");
            newspace = space = 0;
            break;
        case A_COMMA:
            strcpy(xx, ",");
            space = 0;
            break;
        case A_DOLLAR:
            sprintf(xx, "$%ld", *p.p_longstar);
            dp += sizeof(long);
            break;
        case A_DOLLSYM:
            str = (*p.p_sym)->s_name;
            /* xx[0] = '\\'; */
            xx[0] = '$';
            if (!slash(xx + 1, str))
                strcpy(xx + 1, str);
            dp += sizeof(char *);
            break;
        default:
            sprintf(xx, "???");
            break;
        }
        n = strlen(xx);
        newsize = size + n + space;
        /* if (t_growhandle((void **)shit, newsize))
           {
           t_growhandle((void **)shit, 1L);
           *sizep = 0;
           return (1);
           } */
        if (newsize > handlesize)
        {
            handlesize = newsize + (newsize >> 2);
            if (t_growhandle((void **) shit, handlesize))
            {
                t_growhandle((void **) shit, 1L);
                *sizep = 0;
                return (1);
            }
        }
        s2 = *shit + size;
        if (space)
            *(s2++) = ' ';
        for (s1 = xx; n--;)
            *(s2++) = *(s1++);
        size = newsize;
    }
    t_growhandle((void **) shit, size);
    *sizep = size;
    return (0);
}

void binbuf_clear(struct binbuf *x)
{
    x->b_argc = 0;
    x->b_onset = sizeof(long);
    t_growhandle((void **) x->b_shit, x->b_size = sizeof(long));
    **(x->b_shit) = 0;
}

binbuf_insq(struct binbuf *x, int argc, struct atom *argv, int morespace)
{
    static int memct = 100;
    int tc;                     /* # items in temp */
    long ts = 0;                /* temp arg size */
    long nonset = x->b_argc + argc + 1; /* new value for x->b_onset */
    long shiftsize = x->b_size - x->b_onset, newsize;
    struct atom *ap;
    int shift = 0, i;
    char *p1, *p2;
    for (tc = argc, ap = argv; tc--; ap++)
    {
        switch (ap->a_type)
        {
        case A_LONG:
        case A_DOLLAR:
            ts += sizeof(long);
            break;
        case A_FLOAT:
#ifdef __alpha
            ts += sizeof(double);
#else
            ts += sizeof(float);
#endif
            break;
        case A_SYM:
        case A_OBJ:
        case A_DOLLSYM:
            ts += sizeof(char *);
            break;
        }
    }
    newsize = x->b_size + ts;
    nonset += (-nonset) & (sizeof(long) - 1);   /* machine dependent */
    if (nonset > x->b_onset)
    {
        /* if (morespace) nonset += ((nonset >> 2) & (- sizeof(long))); */
        if (morespace)
            nonset <<= 1;
        shift = nonset - x->b_onset;
        if (shift & 3)
            bug("binbuf_text");
        newsize += shift;
    }
    else
        nonset = x->b_onset;
    if (t_growhandle((void **) x->b_shit, newsize))
    {
        binbuf_error = 1;
        return;
    }
    if (!memct--)
        calcfree(), memct = 100;
    x->b_onset = nonset;
    if (shift > 0)
    {
        char *l1 = *x->b_shit + x->b_size, *l2 = l1 + shift;
        for (; shiftsize--;)
            *(--l2) = *(--l1);
    }
    for (ap = argv, p1 = *x->b_shit + x->b_argc, p2 =
         *x->b_shit + (newsize - ts), i = argc; i--; ap++)
    {
        *(p1++) = ap->a_type;
        switch (ap->a_type)
        {
        case A_LONG:
            *((long *) p2) = ap->a_w.w_long;
            p2 += sizeof(long);
            break;
        case A_FLOAT:
#ifdef __alpha
            *((double *) p2) = ap->a_w.w_float;
            p2 += sizeof(double);
#else
            *((float *) p2) = ap->a_w.w_float;
            p2 += sizeof(float);
#endif
            break;
        case A_OBJ:
            *((struct object **) p2) = ap->a_w.w_obj;
            p2 += sizeof(char *);
            break;
        case A_SYM:
            *((struct symbol **) p2) = ap->a_w.w_sym;
            p2 += sizeof(char *);
            break;
        case A_COMMA:
        case A_SEMI:
            break;
        case A_DOLLAR:
            *((long *) p2) = ap->a_w.w_long;
            p2 += sizeof(long);
            break;
        case A_DOLLSYM:
            *((struct symbol **) p2) = ap->a_w.w_sym;
            p2 += sizeof(char *);
            break;
        }
    }
    x->b_argc += argc;
    x->b_size = newsize;
    (*x->b_shit)[x->b_argc] = 0;
    return (0);
}

/* insert a message into a binbuf, terminating it with a semicolon.
   Just for the hell of it, convert symbols #1-#9 into the atoms $1-$9. */
void binbuf_insert(struct binbuf *x, struct symbol *s, int argc,
                   struct atom *argv)
{
    struct atom at[2], *ap;
    int n = 0;
    /* protect the reader against long messages */
    if (argc > FUDIT)
    {
        warn("message too long");
        argc = FUDIT;
    }
    if (argc < 1)
        return;
    for (ap = argv, n = argc; n--; ap++)
    {
        char *s, c, c2;
        if (ap->a_type == A_SYM
            && *(s = ap->a_w.w_sym->s_name) == '#'
            && (c = *(++s)) >= '0' && c <= '9')
        {
            if ((c2 = s[1]) == '\0')
                SETDOLLAR(ap, (int) (c - '0'));
            else if (c2 < '1' || c2 > '9')
                SETDOLLSYM(ap, gensym(s));
        }
    }
    binbuf_insq(x, argc, argv, 1);
    at[0].a_type = A_SEMI;
    binbuf_insq(x, 1, at, 1);

    /* post("insert:");
       binbuf_print(x);
       post("ho ho"); */
}

#ifdef BINBUF_ADDMESS_USEMELATER
  /* I don't remember what this was intended for -- msp 2/94 */


void binbuf_addmess(va_alist)
va_dcl
{
    void *x;
    char *fmt;
    struct atom atoms[100];
    struct atom *atomp = atoms;
    char c;
    int count = 0;
    va_list ap;
    va_start(ap);
    x = va_arg(ap, void *);
    fmt = va_arg(ap, char *);
    while (c = *(fmt++))
    {
        switch (c)
        {
        case 'l':
            atomp->a_type = A_LONG;
            atomp->a_w.w_long = va_arg(ap, long);
            atomp++;
            count++;
            break;
        case 'f':
            atomp->a_type = A_FLOAT;
            atomp->a_w.w_float = va_arg(ap, double);
            atomp++;
            count++;
            break;
        case 's':
            atomp->a_type = A_SYM;
            atomp->a_w.w_sym = va_arg(ap, t_symbol *);
            atomp++;
            count++;
            break;
        }
    }
    va_end(ap);
    binbuf_insert(x, NIL, count, atoms);
}
#endif                          /* BINBUF_ADDMESS_USEMELATER */

void binbuf_set(struct binbuf *x, struct symbol *s, int argc,
                struct atom *argv)
{
    struct atom *ap;
    int ac;
    for (ap = argv, ac = argc; ac--; ap++)
        if (ap->a_type == A_SYM)
        {
            char *s = ap->a_w.w_sym->s_name, c = *s, c2;
            if (c == ';')
                SETSEMI(ap);
            else if (c == ',')
                SETCOMMA(ap);
            /* else if (c == '$' && (c = s[1]) >= '1' && c <= '9' && !s[2])
               SETDOLLAR(ap, c - '0'); */
            else if (c == '$' && (c = s[1]) >= '1' && c <= '9')
            {
                c2 = s[2];
                if (!c2)
                    SETDOLLAR(ap, c - '0');
                else if (c2 < '1' || c2 > '9')
                {
                    SETDOLLSYM(ap, gensym(s + 1));
                    unslash(&ap->a_w.w_sym);
                }
                else
                    unslash(&ap->a_w.w_sym);
            }

            else
                unslash(&ap->a_w.w_sym);
        }
    binbuf_clear(x);
    binbuf_insq(x, argc, argv, 0);
}

void binbuf_append(struct binbuf *x, struct symbol *s, int argc,
                   struct atom *argv)
{
    struct atom *ap;
    int ac;
    for (ap = argv, ac = argc; ac--; ap++)
        if (ap->a_type == A_SYM)
        {
            char *s = ap->a_w.w_sym->s_name, c = *s, c2;
            if (c == ';')
                SETSEMI(ap);
            else if (c == ',')
                SETCOMMA(ap);
            else if (c == '$' && (c = s[1]) >= '1' && c <= '9')
            {
                if (!(c2 = s[2]))
                    SETDOLLAR(ap, c - '0');
                else if (c2 < '1' || c2 > '9')
                {
                    SETDOLLSYM(ap, gensym(s + 1));
                    unslash(&ap->a_w.w_sym);
                }
            }
        }
    binbuf_insq(x, argc, argv, 1);
}

#define INSCHUNK 128
void binbuf_inshandle(struct binbuf *x, long int **handle, long int n)
{
    struct atom a[INSCHUNK];
    long where = 0;
    while (n)
    {
        int howmany = (n < INSCHUNK ? n : INSCHUNK), i;
        long *lp = *handle + where;
        struct atom *ap = a;
        where += howmany;
        n -= howmany;
        for (i = howmany; i--; ap++)
            SETLONG(ap, *(lp++));
        binbuf_insq(x, howmany, a, 0);
    }
}

binbuf_savebox(struct binbuf *x, void *w, struct symbol *what, int d1,
               int d2, int d3, int d4)
/* binbuf owned by a box */
/* binbuf to save into */
/* class of box */
/* dimensions of box */
{
    struct atom at[FUDIT + 8], *ap = at;
    int argc = x->b_argc, ac = argc;
    SETSYM(ap, &s__Pat);
    ap++;
    SETSYM(ap, what);
    ap++;
    SETLONG(ap, d1);
    ap++;
    SETLONG(ap, d2);
    ap++;
    SETLONG(ap, d3);
    ap++;
    SETLONG(ap, d4);
    ap++;
    binbuf_toatoms(x, ap);
    binbuf_insert(w, NIL, argc + 6, at);
}

binbuf_toatoms(struct binbuf *x, t_atom *ap)
{
    int ac = x->b_argc, ac2;
    char *tp = *x->b_shit;
    union ptr p;
    p.p_charstar = tp + x->b_onset;
    if (ac > FUDIT)
        warn("binbuf_dump: buffer truncated"), ac = FUDIT;
    ac2 = ac;
    for (; ac--; tp++, ap++)    /* copy out args */
    {
        char xx[256], *s, c;
        struct symbol *sp;
        switch (*tp)
        {
        case A_LONG:
            SETLONG(ap, *((p.p_longstar)++));
            break;
        case A_FLOAT:
            SETFLOAT(ap, *((p.p_floatstar)++));
            break;
        case A_SYM:
            sp = *((p.p_sym)++);
            if (slash(xx, sp->s_name))
                SETSYM(ap, gensym(xx));
            else
                SETSYM(ap, sp);
            break;
        case A_DOLLAR:
            sprintf(xx, "$%ld", *((p.p_longstar)++));
            SETSYM(ap, gensym(xx));
            break;
        case A_DOLLSYM:
            sp = *((p.p_sym)++);
            xx[0] = '\\';
            xx[1] = '$';
            if (slash(xx + 2, sp->s_name))
                strcpy(xx + 2, sp->s_name);
            SETDOLLSYM(ap, gensym(xx));
            break;
        case A_COMMA:
            SETSYM(ap, gensym(","));
            break;
        case A_SEMI:
            SETSYM(ap, gensym(";"));
            break;
        default:
            bug("binbuf_dump");
        }
      didit:;
    }
    return (ac2);
}

#define HEADER (3 * sizeof(long))

binbuf_read(struct binbuf *x, char *name, t_symbol *volume, int binary)
  /* returns 0 if succeeded */
{
    binbuf_clear(x);
    if (binary)
    {
        long size, n, argc, datasize, onset, nsym = 0, wheresym, *lp, sn;
        char **space, *s, *s2;
        long *xp;
        struct symbol ***symtab, **stp;
        int shift;
        char *p;
        if (!readtohandle(name, volume, &space, &size))
            return (1);
        lp = (long *) (*space);
        if (size < HEADER || *lp != FILEMAGIC)
        {
            whicherror = R_NOTMAX;
            DisposHandle(space);
            return (1);
        }
        argc = lp[1];
        datasize = lp[2];
        wheresym = argc + datasize + HEADER;
        if (argc < 0 || datasize < 0 || size < wheresym ||
            (size > wheresym && (*space)[size - 1]))
        {
            whicherror = R_NOTMAX;
            DisposHandle(space);
            return (1);
        }
        onset = argc + 1;
        onset += (-onset) & (sizeof(long) - 1);
        for (s = (*space) + wheresym, n = size - wheresym; n--; s++)
            if (!*s)
                nsym++;
        if (!(symtab = (struct symbol ***)
              t_newhandle(nsym * sizeof(struct symbol *))))
        {
            DisposHandle(space);
            whicherror = R_MEM;
            return (1);
        }
        for (s = (*space) + wheresym, n = nsym, stp = *symtab; n--; stp++)
            *stp = gensym(s), s += strlen(s) + 1;
        shift = HEADER - (onset - argc);
        for (s = *space, s2 = s + HEADER, n = argc; n--;)
            *(s++) = *(s2++);
        for (s = *space + onset, s2 = s + shift, n = datasize; n--;)
            *(s++) = *(s2++);
        p = (char *) (*space + onset);
        for (s = *space, xp = (long *) p, n = argc; n--; s++)
        {
            switch (*s)
            {
                struct symbol *sp;
            case A_LONG:
            case A_DOLLAR:
                bcopy(p, xp, 4);
                p += sizeof(long);
                xp++;
                break;
            case A_FLOAT:
                {
                    double ldexp(double, int), foo;
                    int exponent = (p[1] & 0xff) + ((p[0] & 0x7f) << 8);
                    int mantissa =
                        ((p[2] & 0xff) << 23) +
                        ((p[3] & 0xff) << 15) +
                        ((p[4] & 0xff) << 7) + ((p[5] & 0xff) >> 1);
                    /* dbug("%d %d", exponent, mantissa); */
                    if (p[0] & 0x80)
                        mantissa = -mantissa;
                    if (mantissa)
                        foo = ldexp((double) mantissa, exponent - 16413);
                    else
                        foo = 0.;
                    *(float *) xp = foo;
                }
                p += 10;
                xp++;
                break;
            case A_SYM:
            case A_DOLLSYM:
                bcopy(p, xp, 4);
                if ((sn = *xp) >= nsym || sn < 0)
                    *(t_symbol **) xp = gensym("???");
                else
                    *(t_symbol **) xp = (*symtab)[sn];
                p += sizeof(t_symbol **);
                xp++;
                break;
            case A_COMMA:
            case A_SEMI:
                break;
            default:
                warn("warning: bad datum in file");
                *s = A_SEMI;
            }
        }
        *s = 0;
        datasize = (char *) xp - (char *) (*space + onset);
        size = onset + datasize;
        DisposHandle(x->b_shit);
        DisposHandle(symtab);
        /* no need to check the following -- always shrinks it */
        t_growhandle((void **) space, size);
        x->b_shit = space;
        x->b_argc = argc;
        x->b_onset = onset;
        x->b_size = size;
        /* binbuf_print(x); */
    }
    else
    {
        int fd;
        long size, magic;
        char **space;
        void *ret;
        if ((fd = open_via_path(name, volume, 0)) < 0)
        {
            whicherror = R_UNIX;
            return (1);
        }
        getfilelength(fd, &size);
        if (!(space = (char **) t_newhandle(size)))
        {
            FSClose((short) fd);
            whicherror = R_MEM;
            return (1);
        }
        HLock(space);
        if (FSRead((short) fd, &size, *space) != 0)
        {
            whicherror = R_UNIX;
            DisposHandle(space);
            FSClose((short) fd);
            return (1);
        }
        HUnlock(space);
        FSClose((short) fd);
        if ((magic = **(long **) space) == FILEMAGIC)
        {
            post("reading Macintosh binary file...");
            return (binbuf_read(x, name, volume, 1));
        }
        else if (binbuf_text(x, space, size))
        {
            DisposHandle(space);
            whicherror = R_MEM;
            return (1);
        }
        DisposHandle(space);
    }
    return (0);
}

#define SYMCHUNK 50

binbuf_write(struct binbuf *x, char *fn, t_symbol *vol, int binary)
{
    int rtnval = 0;
    int fd;
    binary = 0;                 /* no Binary files on NeXT */
    if (createfile(fn, vol, 1, &fd))
    {
        char **shit = (char **) t_newhandle(0L);
        long size;
        if (binbuf_totext(x, shit, &size))
        {
            DisposHandle(shit);
            FSClose(fd);
            FlushVol(0L, vol);
            memouch();
            return (1);
        }
        HLock(shit);
        if (FSWrite(fd, &size, *shit))
            ouch("%s: write failed", fn), rtnval = 1;
        SetEOF(fd, size);
        HUnlock(shit);
        DisposHandle(shit);
        FSClose(fd);
        FlushVol(0L, vol);
    }
    else
        ouch("%s: can't create", fn), rtnval = 1;
    return (rtnval);
}

binbuf_print(struct binbuf *x)
{
    prbuf(x->b_shit, x->b_onset);
}

/* get next atom in binbuf; p1 is onset into types; p2 is onset into the stuff;
   /* *p1 = *p2 = 0 to init; they are automatically stepped. */
void binbuf_getatom(struct binbuf *x, long int *p1, long int *p2,
                    struct atom *ap)
{
    union ptr p;
    p.p_charstar = *x->b_shit + x->b_onset + *p2;
    switch (ap->a_type = (*x->b_shit)[*p1])
    {
    case A_LONG:
    case A_DOLLAR:
        ap->a_w.w_long = *(p.p_longstar++);
        break;
    case A_DOLLSYM:
        ap->a_w.w_sym = *(p.p_sym++);
        break;
    case A_FLOAT:
        ap->a_w.w_float = *(p.p_floatstar++);
        break;
    case A_SYM:
        ap->a_w.w_sym = *(p.p_sym++);
        break;
    case A_OBJ:
        ap->a_w.w_obj = *(p.p_obj++);
        break;
    }
    if (ap->a_type)
        (*p1)++;
    *(p2) = (p.p_charstar - *x->b_shit) - x->b_onset;
}

/* get a long item from a binbuf (returns 1 if fails.) */
int binbuf_getlong(struct binbuf *x, long int *p1, long int *p2,
                   long int *lp)
{
    struct atom a;
    binbuf_getatom(x, p1, p2, &a);
    if (a.a_type != A_LONG)
        return (1);
    *lp = a.a_w.w_long;
    return (0);
}

int binbuf_getsym(struct binbuf *x, long int *p1, long int *p2,
                  struct symbol **sp)
{
    struct atom a;
    binbuf_getatom(x, p1, p2, &a);
    if (a.a_type != A_SYM)
        return (1);
    *sp = a.a_w.w_sym;
    return (0);
}

binbuf_getsemi(struct binbuf *x, long int *p1, long int *p2)
{
    struct atom a;
    binbuf_getatom(x, p1, p2, &a);
    return (a.a_type != A_SEMI);
}

#define ISPATCH 1
#define GOTSEMIS 2
binbuf_misc(struct binbuf *x)   /* <a> routine added to do a bit of checking for newobj() */
{
    int r;
    char *s = *x->b_shit, c;
    union ptr p;
    p.p_charstar = s + x->b_onset;
    r = (*s == A_SYM && *p.p_sym == &s_patcher);
    while (c = *(s++))
    {
        if (c == A_SEMI || c == A_COMMA)
            r |= GOTSEMIS;
    }
    return (r);
}

binbuf_subset(struct binbuf *b1, struct binbuf *b2)     /* <a> return 1 if b2 is a substring of b1 */
{
    char *s1 = *b1->b_shit, c;
    union ptr p1, p2, p3;
    p1.p_charstar = s1 + b1->b_onset;
    while (*s1)
    {
        char *s2 = *b2->b_shit, *s3 = s1;
        p2.p_charstar = s2 + b2->b_onset;
        p3.p_charstar = p1.p_charstar;
        while (c = *s2)
        {
            if (c != *s3)
                goto miss;
            switch (c)
            {
            case A_SYM:
                if (*p2.p_sym++ != *p3.p_sym++)
                    goto miss;
                break;
            case A_LONG:
                if (*p2.p_longstar++ != *p3.p_longstar++)
                    goto miss;
                break;
            case A_FLOAT:
                if (*p2.p_floatstar++ != *p3.p_floatstar++)
                    goto miss;
                break;
            }
            s2++;
            s3++;
        }
        return (1);
      miss:
        switch (c = *s1)
        {
        case A_SYM:
            *p1.p_sym++;
            break;
        case A_LONG:
            *p1.p_longstar++;
            break;
        case A_FLOAT:
            *p1.p_floatstar++;
            break;
        }
        s1++;
    }
    return (0);
}

void binbuf_free(struct binbuf *x)
{
    DisposHandle(x->b_shit);
    freebytes(x, sizeof(*x));
}

void *binbuf_new()
{
    struct binbuf *x = (struct binbuf *) getbytes(sizeof(t_binbuf));
    x->b_shit = (char **) t_newhandle(x->b_size = sizeof(long));
    x->b_argc = 0;
    x->b_onset = sizeof(long);
    **x->b_shit = 0;
    return x;
}

char *attostr(char c, union ptr p);

prbuf(char **buf, long int onset)
/* handle for message buffer */
/* onset into buf of first arg */
{
    char *tp = *buf;
    union ptr p;
    int n;
    p.p_charstar = *buf + onset;
    post("prbuf:");
    for (n = 0; *tp; tp++, n++)
    {
        if (!(n & 7))
            post("");
        poststring(" ");
        poststring(attostr(*tp, p));
        if (*tp == A_LONG)
            (p.p_longstar)++;
        else if (*tp == A_FLOAT)
            (p.p_floatstar)++;
        else if (*tp == A_SYM)
            (p.p_sym)++;
        else if (*tp == A_DOLLAR)
            (p.p_longstar)++;
        else if (*tp == A_DOLLSYM)
            (p.p_sym)++;
    }
}
