
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.12 $ IRCAM $Date: 1994/01/29 07:48:40 $
 *
 * FTS by Miller Puckette
 *
 * mess.h -- define a symbol table and message-passing system.
 */

#ifndef _M_MESS_H
#define _M_MESS_H

#define CAREFUL
#define EXPERIMENT
#define MAGIC 1758379419L       /* random (but odd) */
#define NOGOOD(x) (((struct object *)x)->o_magic != MAGIC)
#define MAXARG 11

/*
 * method --
 */
typedef void (*t_method)();

/*
 * symbol --
 */
typedef struct symbol {
    char *s_name;               /* name */
    struct object *s_thing;     /* possible binding to an object */
    struct symbol *s_nextintab; /* next in hash table for gensym */
} t_symbol;

extern t_symbol *defvolume;
extern t_symbol s_bang, s_int, s_float, s_symbol, s_list, s_anything, s_;

/* 
 * messlist -- list of symbols and their corresponding t_methods,
 * complete with typechecking information 
 */
typedef struct messlist {
    struct symbol *t_symbol;
    t_method m_fun;
    char m_type[MAXARG + 1];    /* wish this were indep of MAXARG */
} t_messlist;

/* 
 * tinyobject -- this structure sits at the head of any object 
 * to which you may pass messages (and which you may feed 
 * to freeobject((t_object *))). 
 */
typedef struct tinyobject {
    struct messlist *t_messlist;        /* list of messages and t_methods */
    /* (also used as freelist link) */
#ifdef CAREFUL
    long t_magic;               /* magic number */
#endif
} t_tinyobject;

/* 
 * object -- any object which wants to have inlets or outlets
 * needs a header with more fields: 
 */
typedef struct object {
    struct messlist *o_messlist;        /* list of messages and t_methods */
    /* (also used as freelist link) */
#ifdef CAREFUL
    long o_magic;               /* magic number */
#endif
    struct inlet *o_inlet;      /* list of inlets */
    struct outlet *o_outlet;    /* list of outlets */
} t_object;

/* 
 * class -- the -1 entry of the message list of an object contains a pointer
 * to its class entry, defined as follows:
 */
typedef struct class {
    struct symbol *c_sym;       /* symbol giving name of class  */
    t_method c_freefun;         /* function to call when freeing */
    t_method c_bangfun;         /* bang method */
    t_method c_intfun;          /* int method */
    t_method c_floatfun;        /* float method */
    t_method c_listfun;         /* list method */
    t_method c_anyfun;          /* anything method */
    t_method c_symfun;          /* "symbol" method */
    int c_size;                 /* size of an instance */
    char c_tiny;                /* flag for tiny header */
    char c_noinlet;             /* flag for no first inlet in patcher */
} t_class;

/* 
 * The list of officially recognized types, including pseudotypes for
 * commas and semicolons.  Used in two places: 1. the reader, when it reads
 * a string, returns long, float, sym, comma, semi, or dollar;
 * and 2. each object t_method comes with an array of them saying what types
 * it needs, from among long, float, sym, obj, gimme, and cant.
 */
#define A_NOTHING 0
#define A_LONG    1
#define A_FLOAT   2
#define A_SYM   3
#define A_OBJ   4               /* for argtype lists; passes the value of sym */
#define A_DEFLONG 5             /* long but defaults to zero */
#define A_DEFFLOAT  6           /* float, same default */
#define A_DEFSYM  7             /* symbol, defaults to "" */
#define A_GIMME   8             /* request that args be passed as an array; */
        /* the routine will check the types itself. */
#define A_CANT    9             /* cannot typecheck args */
#define A_SEMI    10
#define A_COMMA   11
#define A_DOLLAR  12
#define A_DOLLSYM 13

typedef union word {            /* union for packing any above datum */
    long w_long;
#ifdef __alpha
    double w_float;
#else
    float w_float;
#endif
    struct symbol *w_sym;
    struct object *w_obj;
} t_word;

typedef struct atom {           /* and an atom which is a typed datum */
    int a_type;                 /* from the above defs */
    union word a_w;
} t_atom;

#define BAG(s) {error(s); return (0);}
#define BAG2(s,t) {error(s,t); return (0);}
#define ob_messlist(x) (((struct object *)(x))->o_messlist - 1)
#define ob_class(x) ((struct class *) \
         (((struct object *)(x))->o_messlist[-1].t_symbol))
#define ob_name(x) (ob_class(x)->c_sym->s_name)
t_method getfn(void *op, struct symbol *msg);
t_method egetfn(void *op, struct symbol *msg);
t_method zgetfn(void *op, struct symbol *msg);

#define mess0(x, y) (*(getfn(x, y)))(x)
#define mess1(x, y, z1) (*(getfn(x, y)))(x, z1)
#define mess2(x, y, z1,z2) (*(getfn(x, y)))(x, z1,z2)
#define mess3(x, y, z1,z2,z3) (*(getfn(x, y)))(x, z1,z2,z3)
#define mess4(x, y, z1,z2,z3,z4) (*(getfn(x, y)))(x, z1,z2,z3,z4)
#define mess5(x, y, z1,z2,z3,z4,z5) (*(getfn(x, y)))(x, z1,z2,z3,z4,z5)
#define NIL ((void *)0)
#define SETCOMMA(ap) ((ap)->a_type = A_COMMA)
#define SETSEMI(ap) ((ap)->a_type = A_SEMI)
#define SETSYM(ap, x) ((ap)->a_type = A_SYM, (ap)->a_w.w_sym = (x))
#define SETOBJ(ap, x) ((ap)->a_type = A_OBJ, (ap)->a_w.w_obj = (x))
#define SETLONG(ap, x) ((ap)->a_type = A_LONG, (ap)->a_w.w_long = (x))
#define SETFLOAT(ap, x) ((ap)->a_type = A_FLOAT, (ap)->a_w.w_float = (x))
#define SETDOLLAR(ap, x) ((ap)->a_type = A_DOLLAR, (ap)->a_w.w_long = (x))
#define SETDOLLSYM(ap,x) ((ap)->a_type = A_DOLLSYM, (ap)->a_w.w_sym = (x))

struct binbuf;
struct inlet;
struct outlet;
struct patcher;
struct box;
struct patchline;
struct clock;
struct qelem;
struct toolfile;
struct ugen;

/*
 * function prototypes --
 */

typedef t_messlist t_maxclass;
typedef void *(*t_creator)();

void *toolfile_new(char *name, t_symbol * vol, long int type);
int toolfile_fwrite(struct toolfile *x, void *what, int size);
void typedmess(void *object, t_symbol * selector, int argcount,
               t_atom * argval);

t_symbol *gensym(const char *s);

void *getbytes(int size);
void freebytes(void *pointer, int size);

void *newobject(struct messlist *ml);
void freeobject(t_object * x);

void **t_newhandle(long size);
int t_growhandle(void **handle, long newsize);
void t_disposehandle(void **handle);

void c_addmess(t_method method, t_symbol * selector, ...);
void c_extern(t_maxclass ** class, t_creator m_new, t_method m_free,
              t_symbol * sym, int size, int noinlet, ...);
void c_new(t_symbol * s, t_method freefun, int size, int tiny,
           int noinlet);
void c_addbang(t_method method);
void c_addint(t_method method);
void c_addfloat(t_method method);
void c_addsymbol(t_method method);
void c_addlist(t_method method);
void c_addanything(t_method method);
void c_addinx(t_method method, int inletnumber);
void c_addftx(t_method method, int inletnumber);

void *obj_new(t_maxclass ** class, t_method setupfunction);
t_symbol *obj_getclassname(void *obj);
void obj_free(void *obj);

void *outlet_new(void *owner, t_symbol * selector);
void *outlet(void *op, struct symbol *s2);
void outlet_bang(void *outlet);
void outlet_int(void *outlet, long n);
void outlet_float(void *outlet, double f);
void outlet_symbol(void *outlet, t_symbol * s);
void outlet_list(void *outlet, t_symbol * selector,
                 int argcount, t_atom * argval);
void outlet_anything(void *outlet, t_symbol * selector,
                     int argcount, t_atom * argval);
void *inlet_new(void *owner, void *target,
                t_symbol * insym, t_symbol * outsym);
void *inlet(void *op, void *who, struct symbol *s1, struct symbol *s2);

void *clock_new(void *owner, t_method method);
void clock_set(struct clock *x, long timeval);
void clock_delay(struct clock *x, long timeinterval);
void clock_unset(struct clock *x);
void clock_free(struct clock *x);

void *dspchain_addstate(void *state, int size, int alignment);
void dspchain_addc(t_method method, int nargs, ...);
void dspchain_add860(t_method method, int nargs, ...);

void stdlist(struct object *x, struct symbol *s, int ac, struct atom *av);
void stdfloat(struct object *x, double f);

void *binbuf_new();
void binbuf_free(struct binbuf *x);
void *binbuf_eval(void *x, int ac, struct atom *av, void *to);
void binbuf_clear(struct binbuf *x);
int binbuf_read(struct binbuf *x, char *name, t_symbol * volume,
                int binary);
int binbuf_write(struct binbuf *x, char *name, t_symbol * vol, int binary);
void binbuf_set(struct binbuf *x, struct symbol *s, int argc,
                struct atom *argv);
void binbuf_append(struct binbuf *x, struct symbol *s, int argc,
                   struct atom *argv);
void binbuf_getatom(struct binbuf *x, long int *p1, long int *p2,
                    struct atom *ap);


void poststring(char *string);
void postlong(long value);
void postfloat(double value);
void post(char *format, ...);
void warn(char *format, ...);
void hate(char *format, ...);
void error(char *format, ...);
void ouch(char *format, ...);
void bug(char *format, ...);

extern long logicaltime;
long gettime();
double sys_getrealtime(void);

void nullfn();
void nofn();

#define BADFLOAT(f) ((((*(unsigned int*)&(f))&0x7f800000)==0) || \
    (((*(unsigned int*)&(f))&0x7f800000)==0x7f800000))

#define BIGSTRING 4096
#endif                          /* not _M_MESS_H */
