#ifndef lint
static char rcsid[] =
    "$Revision: 1.10 $ IRCAM $Date: 1994/02/07 11:00:50 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.10 $ IRCAM $Date: 1994/02/07 11:00:50 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
#ifdef I860
#include "../6a/6a.h"
#endif

extern struct symbol s_;
long logicaltime;
long sched_next = 0x7fffffffL;
long midi_aread(), midi_bread(), midi_getclktime();
char *midibuf;
static int modemchan, printerchan;
static int dropmodem, dropprinter;
int outport[2] = { 0, 1 }, newoutport[2];

static long faketime;

extern int max_debug;
extern char lockout;
static long timeoffset = 0;

void *midi_noteout, *midi_nc[32],
    *midi_polyout, *midi_yc[32],
    *midi_ctlout, *midi_cc[32], *midi_pgmout, *midi_pc[32],
    *midi_touchout, *midi_tc[32], *midi_bendout, *midi_bc[32],
    *midi_abyteout, *midi_bbyteout, *midi_aexout, *midi_bexout,
    *midi_artout, *midi_brtout;
long midi_chan, midi_arg1, midi_arg2;
#define NOTEOFF 0
#define NOTEON 1
#define POLY 2
#define CTL 3
#define PGM 4
#define TCH 5
#define BEND 6

int midiPorts;

sched_init()                    /* this is called once at the beginning */
{
}

sched_suspend()
{
}

sched_resume()
{
}

sched_close()
{
}

/* the clock object */
typedef struct clock {
    struct clock *c_next;
    void *c_arg;
    t_method c_fn;
    long c_when;
} t_clock;

static struct clock *sched_clkset;

void clock_unset(struct clock *x)
{
    struct clock *c, *c2;
    struct clock stub;
    if (x->c_when < 0)
        return;
    stub.c_next = sched_clkset;
    for (c = &stub; c2 = c->c_next; c = c2)
    {
        if (c2 == x)
        {
            c->c_next = c2->c_next;
            break;
        }
    }
    x->c_when = -1;
    sched_clkset = stub.c_next;
}

void clock_set(struct clock *x, long int when)
{
    struct clock *c, *c2;
    struct clock stub;
    if (x->c_when >= 0)
        clock_unset(x);
    stub.c_next = sched_clkset;
    for (c = &stub; (c2 = c->c_next) && c2->c_when <= when; c = c2);
    c->c_next = x;
    x->c_next = c2;
    x->c_when = when;
    sched_clkset = stub.c_next;
}

void clock_delay(struct clock *x, long int when)
{
    clock_set(x, when + logicaltime);
}

void clock_free(struct clock *x)
{
    clock_unset(x);
    freebytes(x, sizeof(*x));
}

void *clock_new(void *obj, t_method fn)
{
    struct clock *x = (struct clock *) getbytes(sizeof(t_clock));
    x->c_arg = obj;
    x->c_fn = fn;
    x->c_when = -1;
    return (x);
}

extern void *u_chain;
extern int dsp_ticks;

extern long realtime();

#ifndef GRAPHICS
#define RTFIFO
#endif

sched_poll()                    /* poll this at hi pri; add calls to domidi() for midi input */
{
    long miditime;
    long nextbyte;

    faketime = realtime();

    /* while a clock is ready, but no more
       than "throttle" times, call the first clock's function. */
    while (logicaltime < faketime)
    {
        if (sched_clkset && sched_clkset->c_when <= logicaltime)
        {
            struct clock *c = sched_clkset;
            c->c_when = -1;
            sched_clkset = c->c_next;
            (*(c->c_fn)) (c->c_arg);
        }
        else
        {
            if (logicaltime == faketime - 1)
            {
#ifdef RTFIFO
                if (rtfifo_tickin())
                {
                    cp_canit();
                    return;
                }
#endif
                logicaltime++;
#ifdef RTFIFO
                rtfifo_setupsigs();
#endif
                if (u_chain)
                {
                    dspchain_tick(u_chain);
#ifdef I860
                    dsp_ticks++;
#endif
                }
#ifdef RTFIFO
                rtfifo_donesigs();
#endif
            }
            else
                logicaltime++;
        }
    }
}

static char sched_canslip = 0;  /* make functions to set this */


/* the qelem() object -- deprioritized function call for graphics.
   One element is pulled off the queue each time the clock and MIDI queues
   are emptied. */

struct qelem {
    struct qelem *q_next;
    void *q_arg;
    void *(*q_fn)();
    char q_set;
    char q_bad;
    char q_pad[2];
};

static struct qelem sched_qstub, *sched_lastq = &sched_qstub;

void *qelem_set(struct qelem *x)
{
    char savelock = lockout;    /* <a> added lockout */
    lockout = 1;
    if (x->q_bad)
        bug("bad qelem %lx\n", x);
    if (!x->q_set)
    {
        x->q_set = 1;
        sched_lastq->q_next = x;
        sched_lastq = x;
        x->q_next = 0;
    }
    lockout = savelock;
}

void *qelem_unset(struct qelem *x)
{
    char savelock = lockout;    /* <a> added lockout */
    struct qelem *q, *q2;
    lockout = 1;
    q = &sched_qstub;
    if (x->q_set)
    {
        x->q_set = 0;
        for (q = &sched_qstub; q2 = q->q_next; q = q2)
        {
            if (q2 == x)
            {
                q->q_next = q2->q_next;
                if (sched_lastq == q2)
                    sched_lastq = q;
                break;
            }
        }
    }
    lockout = savelock;
}

qelem_free(struct qelem *x)
{
    if (x->q_set)
        qelem_unset(x);
    x->q_bad = 1;
    freebytes(x, sizeof(struct qelem));
}

void *qelem_new(void *obj, void *(*fn)( /* ??? */ ))
{
    struct qelem *x =(struct qelem *) getbytes(sizeof(struct qelem));
    x->q_arg = obj;
    x->q_fn = fn;
    x->q_set = 0;
    x->q_bad = 0;
    return (x);
}

sched_dequeue()                 /* poll this at low pri */
{
    struct qelem *q;
    char savelock;

    savelock = lockout;
    lockout = 1;                /* necessary to put around "if" here? */
    if (q = sched_qstub.q_next)
    {
        q->q_set = 0;
        if (sched_lastq == q)
            sched_lastq = &sched_qstub;
        sched_qstub.q_next = q->q_next;
        lockout = savelock;
        (*(q->q_fn)) (q->q_arg);
        return (1);
    }
    lockout = savelock;
    return (0);
}

typedef struct mp {
    long m_chan;                /* incoming virtual channel */
    long m_arg1;
    long m_cmd;
    long m_basechan;
    char m_sysex;
    char m_argc;
} t_mp;

t_mp sched_midiport;

void maxmidi(struct mp *mp, void *unused, long int nextbyte)    /* set logicaltime first */
{
    /* <a> unused declaration removed */
    void *x;
    if (nextbyte >= 0xf8)
    {
        if (midi_artout)
            outlet_int(midi_artout, nextbyte);
        return;
    }
    if (midi_abyteout)
        outlet_int(midi_abyteout, nextbyte);
    if (nextbyte & 128)
    {
        if (nextbyte == 0xf0 && midi_aexout)
            mp->m_sysex = 1;
        mp->m_cmd = (nextbyte >> 4) & 7;
        mp->m_chan = (nextbyte & 0xf) + mp->m_basechan;
        mp->m_argc = 0;
    }
    else
        switch (mp->m_cmd)
        {
        case NOTEON:
            if (mp->m_argc == 0)
                mp->m_arg1 = nextbyte, mp->m_argc = 1;
            else if (mp->m_argc == 1)
            {
                midi_chan = mp->m_chan + 1;
                midi_arg1 = mp->m_arg1;
                midi_arg2 = nextbyte;
                if (midi_noteout)
                    outlet_bang(midi_noteout);
                if (x = midi_nc[mp->m_chan])
                    outlet_bang(x);
                mp->m_argc = 0;
            }
            break;
        case POLY:              /* polyphonic aftertouch */
            if (mp->m_argc == 0)
            {
                mp->m_arg1 = nextbyte;
                mp->m_argc = 1;
            }
            else if (mp->m_argc == 1)
            {
                midi_chan = mp->m_chan + 1;
                midi_arg1 = mp->m_arg1;
                midi_arg2 = nextbyte;
                if (midi_polyout)
                    outlet_bang(midi_polyout);
                if (x = midi_yc[mp->m_chan])
                    outlet_bang(x);
                mp->m_argc = 0;
            }
            break;
        case CTL:               /* controller change */
            if (mp->m_argc == 0)
                mp->m_arg1 = nextbyte, mp->m_argc = 1;
            else if (mp->m_argc == 1)
            {
                midi_chan = mp->m_chan + 1;
                midi_arg1 = mp->m_arg1;
                midi_arg2 = nextbyte;
                if (midi_ctlout)
                    outlet_bang(midi_ctlout);
                if (x = midi_cc[mp->m_chan])
                    outlet_bang(x);
                mp->m_argc = 0;
            }
            break;
        case BEND:              /* pitchbend */
            if (mp->m_argc == 0)
                mp->m_arg1 = nextbyte, mp->m_argc = 1;
            else if (mp->m_argc == 1)
            {
                midi_chan = mp->m_chan + 1;
                midi_arg1 = mp->m_arg1;
                midi_arg2 = nextbyte;
                if (midi_bendout)
                    outlet_bang(midi_bendout);
                if (x = midi_bc[mp->m_chan])
                    outlet_bang(x);
                mp->m_argc = 0;
            }
            break;
        case TCH:               /* channel aftertouch */
            midi_chan = mp->m_chan + 1;
            midi_arg1 = nextbyte;
            if (midi_touchout)
                outlet_bang(midi_touchout);
            if (x = midi_tc[mp->m_chan])
                outlet_bang(x);
            mp->m_argc = 0;
            break;
        case PGM:               /* program change */
            midi_chan = mp->m_chan + 1;
            midi_arg1 = nextbyte;
            if (midi_pgmout)
                outlet_bang(midi_pgmout);
            if (x = midi_pc[mp->m_chan])
                outlet_bang(x);
            mp->m_argc = 0;
            break;
        case NOTEOFF:
            if (mp->m_argc == 0)
                mp->m_arg1 = nextbyte, mp->m_argc = 1;
            else if (mp->m_argc == 1)
            {
                midi_chan = mp->m_chan + 1;
                midi_arg1 = mp->m_arg1;
                midi_arg2 = 0;
                if (midi_noteout)
                    outlet_bang(midi_noteout);
                if (x = midi_nc[mp->m_chan])
                    outlet_bang(x);
                mp->m_argc = 0;
            }
            break;
        }
    if (mp->m_sysex)
    {
        if (midi_aexout)
            outlet_int(midi_aexout, nextbyte);
        if (nextbyte == 0xf7)
            mp->m_sysex = 0;
    }
}

static t_method sched_midifun = maxmidi;
static void *sched_midiarg1 = &sched_midiport, *sched_midiarg2;

sched_installmidi(t_method fun, void *arg1, void *arg2)
{
    sched_midifun = fun;
    sched_midiarg1 = arg1;
    sched_midiarg2 = arg2;
}

sched_uninstallmidi(void *arg1)
{
    if (sched_midiarg1 == arg1)
    {
        sched_midiarg1 = sched_midiarg2 = 0;
        sched_midifun = 0;
    }
}

sched_domidi(int byte)
{
    if (sched_midiarg1)
        (*sched_midifun) (sched_midiarg1, sched_midiarg2, byte);
}

/* domidi(&sched_midiport, byte); */

long gettime()
{
    return (logicaltime);
}



#ifdef __unix
#include <sys/time.h>

/* time measurement utility routines */

void getrealtime(struct timeval *tv)
{
    gettimeofday(tv, 0);
}

int usecsince(struct timeval *tv)
{
    struct timeval tv2;
    gettimeofday(&tv2, 0);
    return (1000000 * (tv2.tv_sec - tv->tv_sec) +
            (tv2.tv_usec - tv->tv_usec));
}

int msecsince(struct timeval *tv)
{
    struct timeval tv2;
    gettimeofday(&tv2, 0);
    return (1000 * (tv2.tv_sec - tv->tv_sec - 1) +
            (1000000 + tv2.tv_usec - tv->tv_usec) / 1000);
}
#endif
