#ifndef lint
static char rcsid[] =
    "$Revision: 1.13 $ IRCAM $Date: 1994/02/07 11:00:53 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.13 $ IRCAM $Date: 1994/02/07 11:00:53 $
 *
 * FTS by Miller Puckette
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "m_mess.h"
#ifndef I860
#ifndef __osf
#include <sys/types.h>
#include <sys/time.h>
#endif
#include "m_mess.h"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdarg.h>

static char *max_space;         /* currently used new-memory buffer */
static char *max_espace;        /* end of space in current allocated chunk */
static char *max_warnspace;     /* location at which to "do something" */
static char *max_next;          /* another, pre-allocated buffer */
int memislow;                   /* warning that there is less than LOMEM left */
extern int max_debug;
#define CHUNK 32768L            /* chunk size for allocation */
static int FIRSTCHUNK = 0x10000L;
#define LOMEM 4096L             /* lo-memory warning point */
extern int inISR;               /* flag saying that we are in hard interrupt */
int lockout;                    /* locking interrupts out */
char *t_newptr(int n);
char **max_freelist;            /* free-space linked lists indexed by size in LONGs */
#if defined(I860) || defined(__alpha) || defined(__unix)
#define GRAIN 16
#define LOGGRAIN 4
#define NHUNKS 64
#else
#define GRAIN 4                 /* allocation size grain -- round up to multiples */
#define LOGGRAIN 2              /* log of above */
#define NHUNKS 128              /* number of free-space pointers */
#endif

#define MAXSIZEMASK (- (NHUNKS * GRAIN))

extern struct symbol s_max, s_;
#define STRSIZE 2048            /* chunk size for symbol string space */
#define SYMCHUNK 256            /* chunk size for symbols */
static char *stringspace, *estr, *mstr;
static struct symbol *symspace, *esp, *msp;
#define SYMTABSIZE 256
t_symbol **symtab;

int calcfree();
int DisposPtr(char *s);

utils_init()                    /* must call at LO */
{
    static int initted;
    t_symbol *sp, **spp;
    int i;
    int lwas = lockout;

    if (initted)
        return;
    initted = 1;

    lockout = 1;
    if (!(max_space = t_newptr((long) FIRSTCHUNK)))
        hate("not enough space");
    max_espace = max_space + FIRSTCHUNK;
    max_warnspace = max_espace - LOMEM;
    max_next = 0;
    lockout = lwas;
    if (!
        (max_freelist =
         (char **) t_newptr((long) (NHUNKS * sizeof(char *)))))
        hate("not enough space");
    for (i = 0; i < NHUNKS; i++)
        max_freelist[i] = 0;

    if (!(stringspace = t_newptr((long) STRSIZE)))
        hate("no space for symtab");
    if (!(symspace = (struct symbol *)
          t_newptr((long) (SYMCHUNK * sizeof(struct symbol)))))
        hate("no space for symtab 2");
    if (!(symtab = (t_symbol **)
          t_newptr((long) (SYMTABSIZE * sizeof(t_symbol *)))))
        hate("no space for symtab 3");
    estr = stringspace;
    mstr = stringspace + STRSIZE;
    esp = symspace;
    msp = esp + SYMCHUNK;
    for (spp = symtab, i = SYMTABSIZE; i--; spp++)
        *spp = 0;
}

/* the next routine checks whether we are actually at interrupt and if not,
   reports free memory and tries to get expansion space if needed. When you
   are getting low, this will really start to burn CPU time.  I can't find
   any other simple way to feed interrupt level routines with memory, though.
   (later, polling this from the main loop could throttle itself.) */
utils_poll()
{
    if (!inISR)
    {
        if (!max_next)
            max_next = t_newptr((long) CHUNK);
        calcfree();             /* changes posted free-mem value */
    }
}

/* allocate n bytes; return 0 on failure */
void *getbytes(int n)
{
    if (n <= 0)
        n = 1;
    n += ((-n) & (GRAIN - 1));
    if (n & MAXSIZEMASK)
        return (t_newptr(n));
    else
    {
        char **freelist = max_freelist + (n >> LOGGRAIN);
        char *it = *freelist;
        if (!it)
            goto try2;
        *freelist = *((char **) it);
        return (it);
    }
  try2:
    {
        char *was = max_space, *newpt = was + n;
        /* <b> get more memory if low (moved from inside
           the next bracket) */
        if (!max_next)
            utils_poll();
        if (!max_next && newpt > max_warnspace)
        {
            if (!memislow)
                post("out of memory");
            memislow = 1;
        }
        if (newpt > max_espace)
        {
            if (!(was = max_next))
            {
#ifdef I860
                sa_printf("\nout of memory (fatal)\n");
#else
                post("\nout of memory (fatal)\n");
#endif
                exit(1);
            }
            max_next = 0;
            utils_poll();
            max_warnspace = was + (CHUNK - LOMEM);
            max_espace = was + CHUNK;
            newpt = was + n;
        }
        max_space = newpt;
        return (was);
    }
}

char *resizebytes(char *x, int s1, int s2)
{
    /* if before and after will both be malloc'd, just use
       realloc */
    if ((s1 & MAXSIZEMASK) && (s2 & MAXSIZEMASK))
        return (realloc(x, s2));
    else
    {
        int s3 = (s1 > s2 ? s2 : s1);
        char *y = getbytes(s2), *z = x, *w = y;
        if (!y)
            return (0);
        for (; s3--; z++, w++)
            *w = *z;
        freebytes(x, s1);
        return (y);
    }
}

void freebytes(void *x, int n)
{
    /* if ((((int)x) & 0xf0000000) != 0xa0000000) error("freebytes %x", x); */
    if (n <= 0)
        n = 1;
    n += ((-n) & (GRAIN - 1));
    if (n & MAXSIZEMASK)
        DisposPtr(x);
    else
    {
        char **freelist = max_freelist + (n >> LOGGRAIN);
        *(char **) x = *freelist;
        *freelist = x;
    }
}

void freeraw(struct object *op, struct class *cp)       /* get rid of this */
{
    void (*fn)();
    /* post("freeraw: %s at %lx", cp->c_sym->s_name, op); */
    if (fn = cp->c_freefun)
        (*fn) (op);
    freebytes((char *) op, cp->c_size);
}

/******************* symbols *******************/

int sym__hash(const char *s)
{
    int hval = 0, hval2 = 0;
    int ch;
    while (ch = *(s++))
        hval += ch, hval2 += hval;
    return (hval & (SYMTABSIZE - 1));
}

sym_put(struct symbol *s)
{
    t_symbol **spp = symtab + sym__hash(s->s_name);
    s->s_nextintab = *spp;
    *spp = s;
}

struct symbol *sym__new(const char *s)
{
    int n = strlen(s) + 1;
    if (esp == msp)
    {
        if (!
            (esp =
             (struct symbol *)
             t_newptr((long) (SYMCHUNK * sizeof(struct symbol)))))
            hate("no space for new symbol");
        msp = esp + SYMCHUNK;
    }
    if (estr + n > mstr)
    {
        if (!(estr = t_newptr((long) STRSIZE)))
            hate("no space for new string");
        mstr = estr + STRSIZE;
    }
    strcpy(estr, s);
    esp->s_name = estr;
    esp->s_thing = 0;
    estr += n;
    sym_put(esp);
    return (esp++);
}

struct symbol *gensym(const char *s)
{
    struct symbol *sp = symtab[sym__hash(s)];
    for (; sp; sp = sp->s_nextintab)
    {
        const char *s1 = s;
        char *s2 = sp->s_name, c;
        while ((c = *(s1++)) == *(s2++))
            if (!c)
                return (sp);
    }
    return (sym__new(s));
}

int inISR;

calcfree()
{
}

char **handles;
#define HCH 256

char *t_newptr(int n)
{
    char *ret = malloc(n);
#ifdef I860                     /* fix me */
    if (ret)
        *(int *) ret = 0;
#endif
    return (ret);
}

DisposPtr(char *s)
{
    /* if ((((int)s) & 0xf0000000) != 0xa0000000) error("disposPtr %x", s); */
    free(s);
}

t_disposeptr(char *s)
{
    DisposPtr(s);
}

#define hchk(x)

/*
  hchk(h)
  char **h;
  {
  char *space = (char *)((int *)(*h) - 2);
  if (((int *)space)[1] != 123 ||
  space[((int *)space)[0]+2*sizeof(int)] != 123) 
  bug("handles");
  }
  */

#define HANDLE_ONSET 1
#define HANDLE_EXTRA 1

void **t_newhandle(long int n)
{
    char **ret, *space;
    if (n < 0)
        return ((void **) 0);
    if (!(space = getbytes(n + sizeof(void *))))
    {
#ifdef I860
        sa_printf("memory alloc failed\n");
#endif
        return ((void **) 0);
    }
    if (!handles)
    {
        char **newones = (char **) t_newptr(sizeof(char *) * HCH), **x;
        int i;
        if (!newones)
            hate("out of memory");
        for (x = newones, i = HCH; i--; x++)
        {
            *x = (char *) handles;
            handles = x;
        }
    }
    ret = handles;
    handles = (char **) (*ret);
    ((int *) space)[0] = n;
    *ret = space + sizeof(void *);
    return ((void **) ret);
}

int t_growhandle(void **h, long n)
{
    char *space = (char *) (*h) - sizeof(void *);
    int sizewas = *((int *) space);
    hchk(h);
    /* if ((((int)space) & 0xf0000000) != 0xa0000000)
       error("grow %x", space); */
    if (!(space = resizebytes(space,
                              sizewas + sizeof(void *),
                              n + sizeof(void *))))
    {
#ifdef I860
        sa_printf("out of memory\n");
#endif
        return (1);
    }
    ((int *) space)[0] = n;
    *h = space + sizeof(void *);
    return (0);
}

void DisposHandle(void **h)
{
    char *space = (char *) (*h) - sizeof(void *);
    int sizewas = *((int *) space);
    hchk(h);
    freebytes(space, sizewas + sizeof(void *));
    *h = (char *) handles;
    handles = (char **) h;
}

void t_disposehandle(void **h)
{
    DisposHandle(h);
}

HLock(char **h)
{
}

HUnlock(char **h)
{
}

int t_killer;

t_setkiller()
{
    t_killer = 1;
}

void bug(char *fmt, ...)
{
    va_list args;
    char msg[BIGSTRING];
    va_start(args, fmt);
    vsnprintf(msg, BIGSTRING-1, fmt, args);
    msg[BIGSTRING-1] = 0;
    va_end(args);
    post("consistency check failed: ");
    poststring(msg);
}

void error(char *fmt, ...)
{
    va_list args;
    char msg[BIGSTRING];
    va_start(args, fmt);
    vsnprintf(msg, BIGSTRING-1, fmt, args);
    msg[BIGSTRING-1] = 0;
    va_end(args);
    post("error: ");
    poststring(msg);
}

void warn(char *fmt, ...)
{
    va_list args;
    char msg[BIGSTRING];
    va_start(args, fmt);
    vsnprintf(msg, BIGSTRING-1, fmt, args);
    msg[BIGSTRING-1] = 0;
    va_end(args);
    post("warning: ");
    poststring(msg);
}

#ifdef GRAPHICS
char *h_getftsdir();
char *t_config()
{
    static char foo[80];
    char buf[4096];
    int fd, nchar;
    strcpy(buf, h_getftsdir());
    strcat(buf, "/config");
    fd = open(buf, 0);
    if (fd < 0)
        fd = open("/LocalApps/Max/config", 0);
    if (fd < 0)
        fd = open("/usr/local/ispw/config", 0);
    if (fd < 0)
        return (0);
    else if ((nchar = read(fd, foo, 79)) < 0)
        return (0);
    foo[nchar] = 0;
    return (foo);
}
#endif                          /* GRAPHICS */


#ifdef NO_LONGER_USED

void mkacv(int *ac, t_atom *av, char *fmt,      /* format string */
           char *argp)          /* argument list */
{
    struct atom *atomp = av;
    char c;
    int count = 0;
    while (c = *(fmt++))
    {
        switch (c)
        {
        case 'l':
            atomp->a_type = A_LONG;
            atomp->a_w.w_long = *(long *) argp;
            argp += sizeof(long);
            atomp++;
            count++;
            break;
        case 'f':
            atomp->a_type = A_FLOAT;
            atomp->a_w.w_float = *(double *) argp;
            argp += sizeof(double);
            atomp++;
            count++;
            break;
        case 's':
            atomp->a_type = A_SYM;
            atomp->a_w.w_sym = *(struct symbol **) argp;
            argp += sizeof(t_symbol *);
            atomp++;
            count++;
            break;
        case 'o':
            atomp->a_type = A_OBJ;
            atomp->a_w.w_long = *(long *) argp;
            argp += sizeof(long);
            atomp++;
            count++;
            break;
        }
    }
    *ac = count;
}

#endif                          /* NO_LONGER_USED */

float atom_getfloat(t_atom *at)
{
    float f;
    if (at->a_type == A_LONG)
        f = at->a_w.w_long;
    else if (at->a_type == A_FLOAT)
        f = at->a_w.w_float;
    else if (sscanf(at->a_w.w_sym->s_name, "%f", &f) < 1)
    {
        warn("%s: bad number", at->a_w.w_sym->s_name);
        f = 0;
    }
    return f;
}

long atom_getlong(t_atom *at)
{
    if (at->a_type == A_LONG)
        return (at->a_w.w_long);
    else if (at->a_type == A_FLOAT)
        return (at->a_w.w_float);
    else
    {
        long n;
        if (sscanf(at->a_w.w_sym->s_name, "%ld", &n) < 1)
        {
            warn("%s: bad number", at->a_w.w_sym->s_name);
            return (0);
        }
        else
            return (n);
    }
}

t_symbol *atom_getsym(t_atom *at)
{
    char buf[30];
    if (at->a_type == A_SYM)
        return (at->a_w.w_sym);
    else if (at->a_type == A_LONG)
        sprintf(buf, "%ld", at->a_w.w_long);
    else
#ifdef __alpha
        sprintf(buf, "%lf", at->a_w.w_float);
#else
        sprintf(buf, "%f", at->a_w.w_float);
#endif
    return (gensym(buf));
}

void microsleep(int usec)       /* cf. minisleep */
{
    struct timeval timout;
    timout.tv_sec = 0;
    timout.tv_usec = usec;
    select(0, 0, 0, 0, &timout);
}
#endif
