#ifndef lint
static char rcsid[] =
    "$Revision: 1.12 $ IRCAM $Date: 1994/02/07 06:45:45 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.12 $ IRCAM $Date: 1994/02/07 06:45:45 $
 *
 * FTS by Miller Puckette
 *
 */

#include "m_mess.h"
#include "t_mac.h"
#include "p_patcher.h"

extern struct symbol s_click, s_update, s_button, s_psave,
    s_toggle, s_slider, s_new, s_size, s_set, s_setcp,
    s_vbutton, s_vtoggle, s_vslider, s_free, s__active;

void *qelem_new();
GrafPort *patcher_setport(struct patcher *p);

extern int memislow;
extern char lockout;

void box_defaultfree(struct box *b);
void *vbutton_class;
struct binbuf;
extern void binbuf_insert(struct binbuf *x, struct symbol *s, int argc,
                          struct atom *argv);

/* parse an arglist and call the requested function */
void *patchbox(void *p, void *(*fn)( /* ??? */ ), int ac, struct atom *av,
               int defwidth)
{
    int x, y, w, f, cpno = patcher_getcp(p);
    void *ret;
    if (ac < 2 || av[0].a_type != A_LONG || av[1].a_type != A_LONG)
        x = y = 50;
    else
        x = av[0].a_w.w_long, y = av[1].a_w.w_long;
    if (ac < 3 || av[2].a_type != A_LONG)
        w = defwidth;
    else
        w = av[2].a_w.w_long;
    if (ac < 4 || av[3].a_type != A_LONG)
        f = 0;
    else
        f = av[3].a_w.w_long;
    ret = (*fn) (p, x, y, w, f);
    if (ret && cpno >= 0)
        mess2(ret, &s_setcp, cpno, CP_PUT);
    return (ret);
}

/**************** button *************************/

typedef struct vbutton {
    struct box p_box;
    void *p_clock;
    void *p_qelem;
    void *p_cpr;
    char p_pleaseinvert;
    char p_inverted;
} t_vbutton;


void vbutton_update(struct vbutton *x)
{
    Rect *r = &x->p_box.b_rect, r2;
    t_frameoval(r->left + 2, r->top + 2, r->right - 2, r->bottom - 2);
}

vbuttonblack(struct vbutton *x)
{
    x->p_pleaseinvert = 1;
    qelem_set(x->p_qelem);
}

vbuttonwhite(struct vbutton *x)
{
    x->p_pleaseinvert = 0;
    qelem_set(x->p_qelem);
}

vbutton_qfn(struct vbutton *x)
{
    Rect r;
    GrafPort *gp;
    if (x->p_pleaseinvert == x->p_inverted)
        return;
    if (gp = patcher_setport(x->p_box.b_patcher))
    {
        inbox(x, &r);
        if (x->p_pleaseinvert)
        {
            t_paintrect(r.left, r.top, r.right, r.bottom);
            x->p_inverted = 1;
        }
        else
        {
            t_eraserect(r.left, r.top, r.right, r.bottom);
            vbutton_update(x);
            x->p_inverted = 0;
        }
        t_resetport(gp);
    }
}

vbuttondragpoint(struct vbutton *x, Point pt, int but)
{
    if (!but)
    {
        if (patcher_getcp(x->p_box.b_patcher) >= 0)
            h_vmess(x, &s_vbutton, "l", 0L);
        else
            vbuttonwhite(x);
    }
}

void vbutton_tick(struct vbutton *x)
{
    vbuttonwhite(x);
}

void vbutton_click(struct vbutton *x)
{
    Rect *r = &x->p_box.b_rect;
    Point pt;
    if (patcher_getcp(x->p_box.b_patcher) >= 0)
        h_vmess(x, &s_vbutton, "l", 1L);
    box_drag(x, vbuttondragpoint, pt);
}

void vbutton_bang(struct vbutton *x)
{
    if (!x->p_inverted)
        vbuttonblack(x);
    clock_set(x->p_clock, logicaltime + 150);
    outlet_bang(x->p_box.b_ob.o_outlet);
}

void vbutton_psave(struct vbutton *x, void *w)
{
    Rect *r = &x->p_box.b_rect;
    if (r->right == r->left + SIZE1)
        vfun(w, binbuf_insert, NIL, "ssll", &s__Pat,
             &s_button, (long) (r->left), (long) (r->top));
    else
        vfun(w, binbuf_insert, NIL, "sslll", &s__Pat,
             &s_button, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left));
}

/*************** CP support *************/
void vbutton_cpr(struct vbutton *x, long *buf, int n)
{
    x->p_pleaseinvert = *buf;
    qelem_unset(x->p_qelem);
    vbutton_qfn(x);
}

/* become active or inactive */
void *cpr_new();
void vbutton_active(struct vbutton *x, int n)
{
    if (n)
    {
        if (x->p_cpr)
            bug("vbutton_active");
        x->p_cpr = cpr_new(x, vbutton_cpr);
    }
    else
    {
        if (!x->p_cpr)
            bug("vbutton_active 2");
        cpr_free(x->p_cpr);
        x->p_cpr = 0;
    }
}

/* add or remove a copy on a remote CPU */
void vbutton_setcp(struct vbutton *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->p_cpr)
            vbutton_active(x, 0);
        h_free(x);
    }
    if (flags & CP_PUT)
        h_vnew(x, cpno, &s_vbutton, "");
}

void vbutton_ff(struct vbutton *x)
{
    qelem_free(x->p_qelem);
    clock_free(x->p_clock);
}

void vbutton_setup()
{
    c_new(&s_vbutton, vbutton_ff, sizeof(t_vbutton), 0, 0);
    c_addbang(vbutton_bang);
    c_addint(vbutton_bang);
    c_addfloat(vbutton_bang);
    c_addmess(vbutton_bang, &s_list, A_GIMME, 0);
    c_addmess(vbutton_psave, &s_psave, A_CANT, 0);
    c_addmess(vbutton_bang, &s_anything, A_GIMME, 0);
    c_addmess(vbutton_update, &s_update, A_CANT, 0);
    c_addmess(vbutton_click, &s_click, A_CANT, 0);
    c_addmess(vbutton_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vbutton_active, &s__active, A_CANT, 0);
}

void *rbutton(struct patcher *patcher, int x, int y, int w)
{
    struct vbutton *p;
    if (memislow)
        return (0);
    p = (struct vbutton *) obj_new((t_messlist **) & vbutton_class,
                                   vbutton_setup);
    p->p_clock = clock_new(p, vbutton_tick);
    p->p_qelem = qelem_new(p, vbutton_qfn);
    p->p_inverted = 0;
    p->p_cpr = 0;
    box(p, patcher, F_DRAWFIRSTIN | F_GROWY, x, y, x + w, y + w);
    outlet(p, &s_bang);
    p->p_box.b_firstin = (struct inlet *) p;
    box_ready(p);
    return (p);
}

void *mbutton(void *patcher, int x, int y)
{
    return (rbutton(patcher, x, y, SIZE1));
}

void *patcher_button(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, rbutton, ac, av, SIZE1));
}

/***************** vtoggle ******************/

typedef struct vtoggle {
    struct box v_box;
    void *v_qelem;
    void *v_cpr;
    long v_state;
} t_vtoggle;


void vtoggle_bang(struct vtoggle *x),
vtoggle_int(struct vtoggle *x, long int n);
void vtoggle_click(struct vtoggle *x)
{
    char savelock = lockout;    /* <b> lockout added */
    lockout = 1;
    if (patcher_getcp(x->v_box.b_patcher) >= 0)
        h_vmess(x, &s_bang, "");
#ifdef LOCAL
    vtoggle_bang(x);
#endif
    lockout = savelock;
}

void vtoggle_bang(struct vtoggle *x)
{
    vtoggle_int(x, (long) (!x->v_state));
}

void vtoggle_int(struct vtoggle *x, long int n)
{
    outlet_int(x->v_box.b_ob.o_outlet, x->v_state = n);
    qelem_set(x->v_qelem);
}

void vtoggle_set(struct vtoggle *x, long int n)
{
    x->v_state = n;
    qelem_set(x->v_qelem);
}

void vtoggle_float(struct vtoggle *x, double f)
{
    vtoggle_int(x, (long) f);
}

void vtoggle_update(struct vtoggle *x)
{
    Rect *r = &x->v_box.b_rect;
    if (x->v_state)
    {
        int top = r->top + 1, left = r->left,
            bottom = r->bottom - 3, right = r->right - 2;
        t_line(left, top, right, bottom);
        t_line(right, top, left, bottom);
    }
    else
        t_eraserect(r->left + 1, r->top + 2, r->right - 1, r->bottom - 2);
}

vtoggle_qfn(struct vtoggle *x)
{
    GrafPort *gp = patcher_setport(x->v_box.b_patcher);
    if (gp)
    {
        vtoggle_update(x);
        t_resetport(gp);
    }
}

void vtoggle_save(struct vtoggle *v, void *x)
{
    Rect *r = &v->v_box.b_rect;
    if (r->right == r->left + SIZE1)
        vfun(x, binbuf_insert, NIL, "ssll", &s__Pat,
             &s_toggle, (long) (r->left), (long) (r->top));
    else
        vfun(x, binbuf_insert, NIL, "sslll", &s__Pat,
             &s_toggle, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left));
}

/*************** CP support *************/
void vtoggle_cpr(struct vtoggle *x, long *buf, int n)
{
    vtoggle_set(x, *buf);
}

/* become active or inactive */
void vtoggle_active(struct vtoggle *x, int n)
{
    if (n)
    {
        if (x->v_cpr)
            bug("vtoggle_active");
        x->v_cpr = cpr_new(x, vtoggle_cpr);
    }
    else
    {
        if (!x->v_cpr)
            bug("vtoggle_active 2");
        cpr_free(x->v_cpr);
        x->v_cpr = 0;
    }
}

/* add or remove a copy on a remote CPU */
void vtoggle_setcp(struct vtoggle *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->v_cpr)
            vtoggle_active(x, 0);
        h_free(x);
    }
    if (flags & CP_PUT)
        h_vnew(x, cpno, &s_vtoggle, "");
}

void vtoggle_ff(struct vtoggle *x)
{
    qelem_free(x->v_qelem);
}

void *vtoggle_class;
void vtoggle_setup()
{
    c_new(&s_vtoggle, vtoggle_ff, sizeof(t_vtoggle), 0, 0);
    c_addbang(vtoggle_bang);
    c_addint(vtoggle_int);
    c_addfloat(vtoggle_float);
    c_addmess(vtoggle_set, &s_set, A_LONG, 0);
    c_addmess(vtoggle_save, &s_psave, A_CANT, 0);
    c_addmess(vtoggle_click, &s_click, A_CANT, 0);
    c_addmess(vtoggle_update, &s_update, A_CANT, 0);
    c_addmess(vtoggle_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vtoggle_active, &s__active, A_CANT, 0);
}

void *rtoggle(struct patcher *patcher, int x, int y, int w)
{
    struct vtoggle *p;
    if (memislow)
        return (0);
    p = (struct vtoggle *) obj_new((t_messlist **) & vtoggle_class,
                                   vtoggle_setup);
    outlet(p, &s_int);
    box(p, patcher, F_DRAWFIRSTIN | F_GROWY, x, y, x + w, y + w);
    p->v_box.b_firstin = (struct inlet *) p;
    p->v_state = 0;
    p->v_qelem = qelem_new(p, vtoggle_qfn);
    p->v_cpr = 0;
    box_ready(p);
    return (p);
}

void *mtoggle(void *patcher, int x, int y)
{
    return (rtoggle(patcher, (int) x, (int) y, SIZE1));
}

void *patcher_toggle(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, rtoggle, ac, av, SIZE1));
}

typedef struct vslider {
    struct box s_box;
    void *s_qelem;
    void *s_cpr;
    long s_state;
    long s_drawn;
    long s_size;
} t_vslider;

#define DEFSIZE 128
#define MARGIN 4
#define BOTTOM(x) (x->s_size + 1)

void vslider_int(struct vslider *x, long int n);
long boxdragref;

vsliderdragpoint(struct vslider *x, Point pt, int but)
{
    char savelock = lockout;
    lockout = 1;
    if (but)
    {
        long n = boxdragref - pt.v;
        if (n < 0)
            n = 0;
        else if (n >= x->s_size)
            n = x->s_size - 1;
        if (n != x->s_state)
        {
            if (patcher_getcp(x->s_box.b_patcher) >= 0)
                h_vmess(x, &s_int, "l", n);
#ifdef LOCAL
            else
                vslider_int(x, n);
#endif
        }
    }
    lockout = savelock;
}

void vslider_click(struct vslider *x, Point pt)
{
    boxdragref = x->s_state + pt.v;
    box_drag(x, vsliderdragpoint, pt);
}

void vslider_bang(struct vslider *x)
{
    outlet_int(x->s_box.b_ob.o_outlet, x->s_state);
}

void vslider_int(struct vslider *x, long int n)
{
    x->s_state = n;
    outlet_int(x->s_box.b_ob.o_outlet, x->s_state);
    qelem_set(x->s_qelem);
}

void vslider_set(struct vslider *x, long int n)
{
    x->s_state = n;
    qelem_set(x->s_qelem);
}

void vslider_size(struct vslider *x, long int n)
{
    Rect *r = &x->s_box.b_rect;
    if (n < 1)
        n = 1;
    else if (n > 1024)
        n = 1024;
    x->s_size = n;
    if (x->s_state >= n)
        x->s_state = n - 1;
    box_size(x, r->right - r->left, (int) (n + MARGIN));
}

vslider_getsize(struct vslider *x)
{
    return (x->s_size);
}

void vslider_float(struct vslider *x, double f)
{
    vslider_int(x, (long) f);
}

void vslider_update(struct vslider *x)
{
    Rect *r = &x->s_box.b_rect;
    int bottom = r->top + BOTTOM(x), y, v = x->s_state;
    if (x->s_drawn >= 0 && x->s_drawn < x->s_size)
    {
        y = bottom - x->s_drawn;
        t_paintrect(r->left + 1, y, r->right - 1, y + 1);
    }
}

vslider_qfn(struct vslider *x)
{
    GrafPort *gp = patcher_setport(x->s_box.b_patcher);
    int v = x->s_state;
    if (v < 0)
        v = 0;
    else if (v >= x->s_size)
        v = x->s_size - 1;
    if (gp)
    {
        Rect *r = &x->s_box.b_rect;
        int bottom = r->top + BOTTOM(x), y;
        if (x->s_drawn >= 0 && x->s_drawn < x->s_size)
        {
            y = bottom - x->s_drawn;
            t_eraserect(r->left + 1, y, r->right - 1, y + 1);
        }
        y = bottom - v;
        t_paintrect(r->left + 1, y, r->right - 1, y + 1);
        x->s_drawn = v;
        t_resetport(gp);
    }
    else
        x->s_drawn = v;
}

void vslider_save(struct vslider *x, void *w)
{
    Rect *r = &x->s_box.b_rect;
    vfun(w, binbuf_insert, NIL, "ssllll", &s__Pat,
         &s_slider, (long) (r->left), (long) (r->top),
         (long) (r->right - r->left), (long) (x->s_size));
}

/*************** CP support *************/
void vslider_cpr(struct vslider *x, long *buf, int n)
{
    vslider_set(x, *buf);
}

/* become active or inactive */
void vslider_active(struct vslider *x, int n)
{
    if (n)
    {
        if (x->s_cpr)
            bug("vslider_active");
        x->s_cpr = cpr_new(x, vslider_cpr);
    }
    else
    {
        if (!x->s_cpr)
            bug("vslider_active 2");
        cpr_free(x->s_cpr);
        x->s_cpr = 0;
    }
}

/* add or remove a copy on a remote CPU */
void vslider_setcp(struct vslider *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->s_cpr)
            vslider_active(x, 0);
        h_free(x);
    }
    if (flags & CP_PUT)
        h_vnew(x, cpno, &s_vslider, "");
}

void vslider_ff(struct vslider *x)
{
    qelem_free(x->s_qelem);
}

void *vslider_class;
void vslider_setup()
{
    c_new(&s_vslider, vslider_ff, sizeof(t_vslider), 0, 0);
    c_addbang(vslider_bang);
    c_addint(vslider_int);
    c_addfloat(vslider_float);
    c_addmess(vslider_set, &s_set, A_LONG, 0);
    c_addmess(vslider_save, &s_psave, A_CANT, 0);
    c_addmess(vslider_click, &s_click, A_CANT, 0);
    c_addmess(vslider_size, &s_size, A_LONG, 0);
    c_addmess(vslider_update, &s_update, A_CANT, 0);
    c_addmess(vslider_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vslider_active, &s__active, A_CANT, 0);
}

void *rslider(struct patcher *patcher, int x, int y, int w, int s)
{
    struct vslider *p;
    if (memislow)
        return (0);
    p = (struct vslider *) obj_new((t_messlist **) & vslider_class,
                                   vslider_setup);
    if (s < 1)
        s = 1;
    outlet(p, &s_int);
    box(p, patcher, F_DRAWFIRSTIN, x, y, x + w, y + s + MARGIN);
    p->s_box.b_firstin = (struct inlet *) p;
    p->s_state = 0;
    p->s_size = s;
    p->s_drawn = 0;
    p->s_qelem = qelem_new(p, vslider_qfn);
    p->s_cpr = 0;
    box_ready(p);
    return (p);
}

void *mslider(void *patcher, int x, int y)
{
    return (rslider(patcher, x, y, SIZE1, 128));
}

void *patcher_slider(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, rslider, ac, av, SIZE1));
}
