#ifndef lint
static char rcsid[] =
    "$Revision: 1.14 $ IRCAM $Date: 1994/05/10 10:52:05 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.14 $ IRCAM $Date: 1994/05/10 10:52:05 $
 *
 * FTS by Miller Puckette
 *
 */


#include <stdio.h>
#include "m_mess.h"
#include "t_mac.h"
#include "p_patcher.h"
#include "h_cp.h"

extern struct symbol s_inlet, s_outlet, s_click, s_update, s_psave, s_new,
    s__Pat, s_number, s_flonum, s_key, s_set, s_free, s_dblclick,
    s_vnumber, s_vflonum, s_vinlet, s_voutlet, s_setcp, s_setparentcp,
    s_connect, s_disconnect, s_rreceive, s_inchild, s_outchild, s__active;
void *patchbox(void *p, void *(*fn)( /* ??? */ ), int ac, struct atom *av,
               int defwidth);
void *qelem_new();
void box_defaultfree(struct box *b);
long box_inindex(struct box *b), box_outindex(struct box *b);
GrafPort *patcher_setport(struct patcher *p);

struct binbuf;
extern void binbuf_insert(struct binbuf *x, struct symbol *s, int argc,
                          struct atom *argv);

extern int memislow;
extern char lockout;

#define MARGIN 6
#define VNWIDTH 35
static char gotminus, gotkeys, gotpoint, exp;

typedef struct vnumber {
    struct box n_box;
    void *n_qelem;
    int n_down;
    int n_fsize;
    struct atom n_val;
    void *n_cpr;
} t_vnumber;


void vnumber_fsize(struct vnumber *x, int n1, int n2)
{
    x->n_fsize = n1;
    x->n_down = n2;
}

void vnumber_draw(struct vnumber *x)
{
    Rect *r = &x->n_box.b_rect;
    char xx[256], *s2;
    int rats, w, avail = r->right - r->left - (MARGIN + 1), shit = 0;
    t_textsize(x->n_fsize);
    sprintf(xx, "%ld", x->n_val.a_w.w_long);
    if (t_stringwidth(xx) > avail)
    {
        xx[0] = (x->n_val.a_w.w_long >= 0 ? '+' : '-');
        xx[1] = 0;
        if (t_charwidth(xx[0]) > avail)
            xx[0] = 0;
    }
    t_drawstring(r->left + MARGIN, r->top + x->n_down, xx);
}

void vnumber_sign(struct vnumber *x)
{
    Rect *r = &x->n_box.b_rect;
    int top = r->top, left = r->left;
    t_line(left + 1, top + 3, left + 4, top + 6);
    t_line(left + 4, top + 7, left + 1, top + 10);
}

void vnumber_fill(struct vnumber *x)
{
    Rect *r = &x->n_box.b_rect;
    int top = r->top, left = r->left;
    patcher_nonum(x->n_box.b_patcher);
    left++;
    t_line(left, top + 3, left, top + 10);
    left++;
    t_line(left, top + 4, left, top + 9);
    left++;
    t_line(left, top + 5, left, top + 8);
    left++;
    t_line(left, top + 6, left, top + 7);
}

void vnumber_update(struct vnumber *x)
{
    vnumber_draw(x);
    vnumber_sign(x);
}

vnumber_erase(struct vnumber *x)
{
    Rect *r = &x->n_box.b_rect;
    t_eraserect(r->left + MARGIN, r->top + 2, r->right - 1, r->bottom - 2);
}

vnumber_qfn(struct vnumber *x)  /* queued function */
{
    GrafPort *gp = patcher_setport(x->n_box.b_patcher);
    if (gp)
    {
        vnumber_erase(x);
        vnumber_draw(x);
        t_resetport(gp);
    }
}

void vnumber_int(struct vnumber *x, long int n)
{
    outlet_int(x->n_box.b_ob.o_outlet, x->n_val.a_w.w_long = n);
    qelem_set(x->n_qelem);
}

void vnumber_set(struct vnumber *x, long int n)
{
    x->n_val.a_w.w_long = n;
    qelem_set(x->n_qelem);
}

void vnumber_float(struct vnumber *x, double f)
{
    vnumber_int(x, (long) f);
}

void vnumber_bang(struct vnumber *x)
{
    outlet_int(x->n_box.b_ob.o_outlet, x->n_val.a_w.w_long);
}

numdragpoint(struct vnumber *x, Point pt, int but)
{
    char savelock = lockout;
    lockout = 1;
    if (but)
    {
        long n = boxdragref - pt.v;
        if (n != x->n_val.a_w.w_long)
        {
            if (patcher_getcp(x->n_box.b_patcher) >= 0)
                h_vmess(x, &s_int, "l", n);
#ifdef LOCAL
            else
                vnumber_int(x, n);
#endif
        }
        patcher_numdrag(x->n_box.b_patcher, x);
    }
    lockout = savelock;
}

void vnumber_click(struct vnumber *x, Point pt)
{
    boxdragref = x->n_val.a_w.w_long + pt.v;
    box_drag(x, numdragpoint, pt);
    gotminus = gotkeys = 0;
    vnumber_fill(x);
    patcher_numdrag(x->n_box.b_patcher, x);
}

void vnumber_key(struct vnumber *x, int what)
{
    char k = what;
    box_drag(x, NIL, NIL);
    if (k == '-')
    {
        if (!gotkeys)
            gotminus = 1;
    }
    else if (k >= '0' && k <= '9')
    {
        long it = k - '0';
        if (gotminus)
            it = -it;
        if (gotkeys)
            it += 10 * x->n_val.a_w.w_long;
        if (patcher_getcp(x->n_box.b_patcher) >= 0)
        {
            h_vmess(x, &s_int, "l", it);
            x->n_val.a_w.w_long = it;
        }
#ifdef LOCAL
        else
            vnumber_int(x, it);
#endif
        gotkeys = 1;
    }
}

void vnumber_psave(struct vnumber *x, void *w)
{
    Rect *r = &x->n_box.b_rect;
    struct symbol *cp =
        (x->n_val.a_type == A_LONG ? &s_number : &s_flonum);
    vfun(w, binbuf_insert, NIL, "ssllll", &s__Pat, cp, (long) (r->left),
         (long) (r->top), (long) (r->right - r->left),
         (long) (x->n_box.b_fsize));
}

/*************** CP support *************/

void vnumber_cpr(struct vnumber *x, long int *buf, int n)
{
    vnumber_set(x, *buf);
}

void *cpr_new();
/* become active or inactive */
void vnumber_active(struct vnumber *x, int n)
{
    if (n)
    {
        if (x->n_cpr)
            bug("vnumber_active");
        x->n_cpr = cpr_new(x, vnumber_cpr);
    }
    else
    {
        if (!x->n_cpr)
            bug("vnumber_active 2");
        cpr_free(x->n_cpr);
        x->n_cpr = 0;
    }
}

/* add or remove a copy on a remote CPU */
void vnumber_setcp(struct vnumber *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->n_cpr)
            vnumber_active(x, 0);
        h_free(x);
    }
    if (flags & CP_PUT)
        h_vnew(x, cpno, &s_vnumber, "");
}

void vnumber_ff(struct vnumber *x)
{
    if (x->n_cpr)
        bug("free before disconnect");
    qelem_free(x->n_qelem);
}

void *vnumber_class;
void vnumber_setup()
{
    c_new(&s_vnumber, vnumber_ff, sizeof(t_vnumber), 0, 0);
    c_addint(vnumber_int);
    c_addfloat(vnumber_float);
    c_addbang(vnumber_bang);
    c_addmess(vnumber_set, &s_set, A_LONG, 0);
    c_addmess(vnumber_click, &s_click, A_CANT, 0);
    c_addmess(vnumber_key, &s_key, A_CANT, 0);
    c_addmess(vnumber_update, &s_update, A_CANT, 0);
    c_addmess(vnumber_psave, &s_psave, A_CANT, 0);
    c_addmess(vnumber_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vnumber_active, &s__active, A_CANT, 0);
}

void *xnumber(struct patcher *patcher, int x, int y, int w, int f)
{
    struct vnumber *p;
    if (memislow)
        return (0);
    p = (struct vnumber *) obj_new((t_messlist **) & vnumber_class,
                                   vnumber_setup);
    box(p, patcher, F_DRAWFIRSTIN, x, y, x + w, y + SIZE1 - 1);
    outlet(p, &s_int);
    p->n_box.b_firstin = (struct inlet *) p;
    p->n_val.a_type = A_LONG;
    p->n_val.a_w.w_long = 0;
    p->n_qelem = qelem_new(p, vnumber_qfn);
    p->n_cpr = 0;
    box_num(p, f);
    box_ready(p);
    return (p);
}

void *mnumber(void *patcher, int x, int y, int f)
{
    return (xnumber(patcher, x, y, VNWIDTH, f));
}

void *patcher_number(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, xnumber, ac, av, VNWIDTH));
}

void vflonum_draw(struct vnumber *x)
{
    Rect *r = &x->n_box.b_rect;
    char xx[256], *s2;
    int rats, w, avail = r->right - r->left - (MARGIN + 1), shit = 0;
    t_textsize(x->n_fsize);
    sprintf(xx, "%lf", x->n_val.a_w.w_float);
    /* eat trailing zeros */
    for (rats = 0, s2 = xx; *s2; s2++)
        if (*s2 == '.')
            rats = 1;
    if (rats)
        for (s2--; *s2 == '0'; s2--)
            *s2 = '\0';
    w = t_stringwidth(xx);
    while (w > avail)
    {
        if (*s2 > '9' || *s2 < '0' || s2 == xx)
        {
            xx[0] = (x->n_val.a_w.w_float >= 0 ? '+' : '-');
            if (t_charwidth(xx[0]) > avail)
                xx[0] = 0;
            xx[1] = 0;
            break;
        }
        w -= t_charwidth(*s2);
        *(s2--) = '\0';
    }
    t_drawstring(r->left + MARGIN, r->top + x->n_down, xx);
}

void vflonum_update(struct vnumber *x)
{
    vflonum_draw(x);
    vnumber_sign(x);
}

void vflonum_float(struct vnumber *x, double f)
{
    outlet_float(x->n_box.b_ob.o_outlet, x->n_val.a_w.w_float = f);
    qelem_set(x->n_qelem);
}

vflonum_qfn(struct vnumber *x)  /* queued function */
{
    GrafPort *gp = patcher_setport(x->n_box.b_patcher);
    if (gp)
    {
        vnumber_erase(x);
        vflonum_draw(x);
        t_resetport(gp);
    }
}

void vflonum_int(struct vnumber *x, long int n)
{
    vflonum_float(x, (double) n);
}

void vflonum_bang(struct vnumber *x)
{
    outlet_float(x->n_box.b_ob.o_outlet, x->n_val.a_w.w_float);
}

void vflonum_set(struct vnumber *x, double f)
{
    x->n_val.a_w.w_float = f;
    qelem_set(x->n_qelem);
}

static float clickwas, placeval, ref;

fndragpoint(struct vnumber *x, Point pt, int but)
{
    char savelock = lockout;
    lockout = 1;
    if (but)
    {
        float newval;
        long n = boxdragref - pt.v;
        long abs = (n > 0 ? n : -n), clip = (abs > 128 ? 128 : abs);
        long grow = ((64 + (clip & 63)) << (clip >> 6)) - 63;
        newval =
            (n >
             0 ? clickwas + grow * ref : (n <
                                          0 ? clickwas -
                                          grow * ref : clickwas));
        if (newval != x->n_val.a_w.w_float)
        {
            if (patcher_getcp(x->n_box.b_patcher) >= 0)
                h_vmess(x, &s_float, "f", (double) newval);
#ifdef LOCAL
            else
                vflonum_float(x, (double) newval);
#endif
        }
        patcher_numdrag(x->n_box.b_patcher, x);
    }
    lockout = savelock;
}

void vflonum_click(struct vnumber *x, Point pt)
{
    clickwas = x->n_val.a_w.w_float;
    ref = .001 * (clickwas > 0 ? clickwas : -clickwas);
    if (ref < .001)
        ref = .001;
    boxdragref = pt.v;
    gotminus = gotkeys = gotpoint = 0;
    box_drag(x, fndragpoint, pt);
    vnumber_fill(x);
    patcher_numdrag(x->n_box.b_patcher, x);
}

void vflonum_key(struct vnumber *x, int what)
{
    char k = what;
    box_drag(x, NIL, NIL);
    if (k == '-')
    {
        if (!gotkeys && !gotpoint)
            gotminus = 1;
    }
    else if (k == '.')
    {
        if (!gotkeys)
        {
            if (patcher_getcp(x->n_box.b_patcher) >= 0)
            {
                h_vmess(x, &s_float, "f", (double) 0.);
                x->n_val.a_w.w_float = 0.;
            }
#ifdef LOCAL
            else
                vflonum_float(x, (double) 0.);
#endif
        }
        gotpoint = 1;
        placeval = .1;
    }
    else if (k >= '0' && k <= '9')
    {
        float it = k - '0';
        if (gotminus)
            it = -it;
        if (gotpoint)
        {
            it = x->n_val.a_w.w_float + it * placeval;
            placeval *= .1;
        }
        else if (gotkeys)
            it += 10 * x->n_val.a_w.w_float;
        if (patcher_getcp(x->n_box.b_patcher) >= 0)
        {
            h_vmess(x, &s_float, "f", (double) it);
            x->n_val.a_w.w_float = it;
        }
#ifdef LOCAL
        else
            vflonum_float(x, (double) it);
#endif
        gotkeys = 1;
    }
}

/*************** CP support *************/

void vflonum_cpr(struct vnumber *x, float *buf, int n)
{
    vflonum_set(x, (double) (*buf));
}

/* become active or inactive */
void vflonum_active(struct vnumber *x, int n)
{
    if (n)
    {
        if (x->n_cpr)
            bug("vflonum_active");
        x->n_cpr = cpr_new(x, vflonum_cpr);
    }
    else
    {
        if (!x->n_cpr)
            bug("vflonum_active 2");
        cpr_free(x->n_cpr);
        x->n_cpr = 0;
    }
}

/* add or remove a copy on a remote CPU */
void vflonum_setcp(struct vnumber *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->n_cpr)
            vflonum_active(x, 0);
        h_free(x);
    }
    if (flags & CP_PUT)
        h_vnew(x, cpno, &s_vflonum, "");
}

void vflonum_ff(struct vnumber *x)
{
    if (x->n_cpr)
        bug("free before disconnect");
    qelem_free(x->n_qelem);
}

void *vflonum_class;
void vflonum_setup()
{
    c_new(&s_vnumber, vflonum_ff, sizeof(t_vnumber), 0, 0);
    c_addint(vflonum_int);
    c_addfloat(vflonum_float);
    c_addbang(vflonum_bang);
    c_addmess(vflonum_set, &s_set, A_FLOAT, 0);
    c_addmess(vflonum_click, &s_click, A_CANT, 0);
    c_addmess(vflonum_key, &s_key, A_CANT, 0);
    c_addmess(vflonum_update, &s_update, A_CANT, 0);
    c_addmess(vnumber_psave, &s_psave, A_CANT, 0);
    c_addmess(vflonum_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vflonum_active, &s__active, A_CANT, 0);
}

void *xflonum(struct patcher *patcher, int x, int y, int w, int f)
{
    struct vnumber *p;
    if (memislow)
        return (0);
    p = (struct vnumber *) obj_new((t_messlist **) & vflonum_class,
                                   vflonum_setup);
    box(p, patcher, F_DRAWFIRSTIN, x, y, x + w, y + SIZE1 - 1);
    outlet(p, &s_float);
    p->n_box.b_firstin = p;
    p->n_val.a_type = A_FLOAT;
    p->n_val.a_w.w_float = 0;
    p->n_qelem = qelem_new(p, vflonum_qfn);
    p->n_cpr = 0;
    box_num(p, f);
    box_ready(p);
    return (p);
}

void *mflonum(void *patcher, int x, int y, int f)
{
    return (xflonum(patcher, x, y, VNWIDTH, f));
}

void *patcher_flonum(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, xflonum, ac, av, VNWIDTH));
}

typedef struct vinlet {
    struct box i_box;
    void *i_inlet;
    int i_cp;
    int i_parentcp;
    int i_sig;
} t_vinlet;

void vinlet_update(struct vinlet *i)
{
    Rect *r = &i->i_box.b_rect;
    int x = r->left + 1, y = r->top + 2, w = (r->right - r->left - 3) >> 1;
    t_line(x, y, x + w, y + w);
    t_line(x + w, y + w, x + w + w, y);
    if (i->i_sig)
        t_drawstring(x + w - 2, y + w - 2, "~");
}

void vinlet_psave(struct vinlet *i, void *w)
{
    Rect *r = &i->i_box.b_rect;
    if (i->i_sig)
        vfun(w, binbuf_insert, NIL, "ssllll", &s__Pat,
             &s_inlet, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left), 1L);
    else if (r->right == r->left + SIZE2)
        vfun(w, binbuf_insert, NIL, "ssll", &s__Pat,
             &s_inlet, (long) (r->left), (long) (r->top));
    else
        vfun(w, binbuf_insert, NIL, "sslll", &s__Pat,
             &s_inlet, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left));
}

void vinlet_put(struct vinlet *x)
{
    if (x->i_cp != x->i_parentcp)
    {
        void *connect = h_vnewunique(x->i_cp, &s_rreceive, "ll",
                                     1L, (long) (x->i_sig));
        h_vmess(x, &s_connect, "o", connect);
        h_vmess(&x->i_inlet, &s_connect, "ll", connect, (long) (x->i_cp));
    }
    else
        h_vmess(&x->i_inlet, &s_connect, "ll", x, (long) (x->i_cp));
}

void vinlet_unput(struct vinlet *x)
{
    h_vmess(&x->i_inlet, &s_disconnect, "");
    h_vmess(x, &s_disconnect, "");
}

void vinlet_setcp(struct vinlet *x, int cpno, int flags)
{
    if (flags & CP_RESTART)
        x->i_cp = x->i_parentcp = -1;
    if (flags & CP_UNPUT)
    {
        x->i_cp = -1;
        if (x->i_parentcp >= 0)
            vinlet_unput(x);
        h_free(x);
    }
    if (flags & CP_PUT)
    {
        x->i_cp = cpno;
        h_vnew(x, cpno, &s_inchild, "l", (long) (x->i_sig));
        if (x->i_parentcp >= 0)
            vinlet_put(x);
    }
}

void vinlet_setparentcp(struct vinlet *x, int cpno)
{
    if (cpno >= 0)
    {
        h_vnew(&x->i_inlet, cpno, &s_vinlet, "oll",
               x->i_box.b_patcher, (long) (x->i_box.b_rect.left),
               (long) (x->i_sig));
        x->i_parentcp = cpno;
        if (x->i_cp >= 0)
            vinlet_put(x);
    }
    else
    {
        if (x->i_cp >= 0)
            vinlet_unput(x);
        h_free(&x->i_inlet);
        x->i_parentcp = -1;
    }
}

/* update x-location on CP if necessary */
void vinlet_int(struct vinlet *x, long int n)
{
    if (x->i_parentcp >= 0)
        h_vmess(&x->i_inlet, &s_int, "l", n);
}

void vinlet_dblclick(struct vinlet *x)
{
    patcher_visparent(x->i_box.b_patcher);
}

void vinlet_ff(struct vinlet *x)
{
    if (x->i_parentcp >= 0)
        h_free(&x->i_inlet);
}

void *vinlet_class;
void vinlet_setup()
{
    c_new(&s_vinlet, vinlet_ff, sizeof(t_vinlet), 0, 0);
    c_addmess(vinlet_setcp, &s_setcp, A_CANT, 0);
    c_addmess(vinlet_setparentcp, &s_setparentcp, A_CANT, 0);
    c_addint(vinlet_int);
    c_addmess(vinlet_update, &s_update, A_CANT, 0);
    c_addmess(vinlet_psave, &s_psave, A_CANT, 0);
    c_addmess(vinlet_dblclick, &s_dblclick, A_CANT, 0);
}

void *rinlet(struct patcher *patcher, int x, int y, int w, int sig)
{
    struct vinlet *p;
    void *v;
    if (memislow)
        return (0);
    p = (struct vinlet *) obj_new((t_messlist **) & vinlet_class,
                                  vinlet_setup);
    box(p, patcher, F_GROWY, x, y, x + w, y + w);
    p->i_cp = p->i_parentcp = -1;
    p->i_sig = sig;
    v = outlet(p, NIL);
    p->i_inlet = inlet(patcher, v, NIL, NIL);
    inlet_chain(p->i_inlet, v);
    box_vinlet(p, p->i_inlet);
    p->i_box.b_firstin = (struct inlet *) p;
    box_ready(p);
    return (p);
}

void *minlet(void *patcher, int x, int y)
{
    return (rinlet(patcher, x, y, SIZE1, 0));
}

void *msiginlet(void *patcher, int x, int y)
{
    return (rinlet(patcher, x, y, SIZE1, 1));
}

void *patcher_inlet(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, rinlet, ac, av, SIZE1));
}

typedef struct voutlet {
    struct box o_box;
    void *o_outlet;
    int o_cp;
    int o_parentcp;
    int o_sig;
} t_voutlet;

void voutlet_update(struct voutlet *o)
{
    Rect *r = &o->o_box.b_rect, r2;
    int x = r->left + 1, y = r->bottom - 3, w =
        (r->right - r->left - 3) >> 1;
    t_line(x, y, x + w, y - w);
    t_line(x + w, y - w, x + w + w, y);
    if (o->o_sig)
        t_drawstring(x + w - 2, y + w - 2, "~");
}

void voutlet_psave(struct voutlet *o, void *w)
{
    Rect *r = &o->o_box.b_rect;
    if (o->o_sig)
        vfun(w, binbuf_insert, NIL, "ssllll", &s__Pat,
             &s_outlet, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left), 1L);
    else if (r->right == r->left + SIZE2)
        vfun(w, binbuf_insert, NIL, "ssll", &s__Pat,
             &s_outlet, (long) (r->left), (long) (r->top));
    else
        vfun(w, binbuf_insert, NIL, "sslll", &s__Pat,
             &s_outlet, (long) (r->left), (long) (r->top),
             (long) (r->right - r->left));
}

void voutlet_put(struct voutlet *x)
{
    if (x->o_cp != x->o_parentcp)
    {
        void *connect = h_vnewunique(x->o_parentcp, &s_rreceive, "ll",
                                     1L, x->o_sig);
        h_vmess(&x->o_outlet, &s_connect, "o", connect);
        h_vmess(x, &s_connect, "ll", connect, (long) (x->o_parentcp));
    }
    else
        h_vmess(x, &s_connect, "ll", &x->o_outlet, (long) (x->o_cp));
}

void voutlet_unput(struct voutlet *x)
{
    h_vmess(x, &s_disconnect, "");
    h_vmess(&x->o_outlet, &s_disconnect, "");
}

void voutlet_setcp(struct voutlet *x, int cpno, int flags)
{
    if (flags & CP_UNPUT)
    {
        x->o_cp = -1;
        if (x->o_parentcp >= 0)
            voutlet_unput(x);
        h_free(x);
    }
    if (flags & CP_PUT)
    {
        x->o_cp = cpno;
        h_vnew(x, cpno, &s_outchild, "l", x->o_sig);
        if (x->o_parentcp >= 0)
            voutlet_put(x);
    }
}

void voutlet_setparentcp(struct voutlet *x, int cpno)
{
    if (cpno >= 0)
    {
        h_vnew(&x->o_outlet, cpno, &s_voutlet, "oll",
               x->o_box.b_patcher, (long) (x->o_box.b_rect.left),
               x->o_sig);
        x->o_parentcp = cpno;
        if (x->o_cp >= 0)
            voutlet_put(x);
    }
    else
    {
        if (x->o_cp >= 0)
            voutlet_unput(x);
        h_free(&x->o_outlet);
        x->o_parentcp = -1;
    }
}

/* update x-location on CP if necessary */
void voutlet_int(struct voutlet *x, long int n)
{
    if (x->o_parentcp >= 0)
        h_vmess(&x->o_outlet, &s_int, "l", n);
}

void voutlet_dblclick(struct voutlet *x)
{
    patcher_visparent(x->o_box.b_patcher);
}

void voutlet_ff(struct voutlet *x)
{
    if (x->o_parentcp >= 0)
        h_free(&x->o_outlet);
}

void *voutlet_class;
void voutlet_setup()
{
    c_new(&s_voutlet, voutlet_ff, sizeof(t_voutlet), 0, 0);
    c_addmess(voutlet_setcp, &s_setcp, A_CANT, 0);
    c_addmess(voutlet_setparentcp, &s_setparentcp, A_CANT, 0);
    c_addmess(voutlet_setcp, &s_setcp, A_CANT, 0);
    c_addint(voutlet_int);
    c_addmess(voutlet_update, &s_update, A_CANT, 0);
    c_addmess(voutlet_psave, &s_psave, A_CANT, 0);
    c_addmess(voutlet_dblclick, &s_dblclick, A_CANT, 0);
}

void *routlet(struct patcher *patcher, int x, int y, int w, int sig)
{
    struct voutlet *p;
    void *v;
    if (memislow)
        return (0);
    p = (struct voutlet *) obj_new((t_messlist **) & voutlet_class,
                                   voutlet_setup);
    box(p, patcher, F_GROWY, x, y, x + w, y + w);
    p->o_cp = p->o_parentcp = -1;
    p->o_sig = sig;
    p->o_outlet = outlet(patcher, NIL); /* LATER could type outlet if sig */
    v = inlet(p, p->o_outlet, NIL, NIL);
    inlet_chain(v, p->o_outlet);
    box_voutlet(p, p->o_outlet);
    p->o_box.b_firstin = (struct inlet *) p;
    box_ready(p);
    return (p);
}

void *moutlet(void *patcher, int x, int y)
{
    return (routlet(patcher, x, y, SIZE1, 0));
}

void *msigoutlet(void *patcher, int x, int y)
{
    return (routlet(patcher, x, y, SIZE1, 1));
}

void *patcher_outlet(void *p, struct symbol *s, int ac, struct atom *av)
{
    return (patchbox(p, routlet, ac, av, SIZE1));
}
