#ifndef lint
static char rcsid[] =
    "$Revision: 1.10 $ IRCAM $Date: 1994/01/04 18:02:50 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.10 $ IRCAM $Date: 1994/01/04 18:02:50 $
 *
 * FTS by Miller Puckette
 *
 */


/* menu handling + utility routines */

#include "m_mess.h"
#include "t_mac.h"
#include "t_wind.h"
#include "p_patcher.h"
#include "m_inout.h"
#include "p_patcher2.h"
#include "t_menu.h"
extern int h_nproc;

extern struct symbol s_patcher, s_psave, s_connect, s__active;
extern int binbuf_error;
extern int max_lock;
extern int memislow;
int unact;
void *unwind;
extern char lockout;
extern void *patcher_class, *vnewobj_class;
t_symbol *t_getpath();

void binbuf_insert(struct binbuf *x, struct symbol *s, int argc,
                   struct atom *argv);
GrafPort *patcher_setport(struct patcher *p);
int patcher_clear(struct patcher *p);
int patcher_setunmove(struct patcher *p, int new);
int patcher_setuntype(struct patcher *p);
int patcher_docopy(struct patcher *p, int toundo);
int patcher_doclear(struct patcher *p);
int patcher_savelines(struct patcher *p);
int noundo(void *x);
int patcher_cpto(struct patcher *p, void *buf);
int patcher_pfrom(struct patcher *p, void *buf);
int patcher_findstring(t_patcher * p, char *s, int ouchonfail);
int patcher_findagain(int ouchonfail);
int sc1d(int x1, int x2, int x3, int x4);

patcher_key(struct patcher *p, int k)
{
    struct box *b;
    if (!max_lock && (k == BACKSPACE))
    {
        TEHandle teh;
        if (!(b = p->p_seltext) ||
            (!te_selstart(teh = b->b_un.b_teh) &&
             te_selend(teh) == te_length(teh)))
        {
            patcher_clear(p);
            return;
        }
    }
    else if (k >= LEFTARROW && k <= DOWNARROW && !p->p_seltext)
    {
        if (!p->p_nselect)
            return;
        patcher_setunmove(p, 0);
        for (b = p->p_box; b; b = b->b_next)
        {
            if (b->b_selected)
            {
                int x = b->b_rect.left, y = b->b_rect.top;
                switch (k)
                {
                case LEFTARROW:
                    x--;
                    break;
                case RIGHTARROW:
                    x++;
                    break;
                case UPARROW:
                    y--;
                    break;
                case DOWNARROW:
                    y++;
                    break;
                }
                box_move(b, x, y);
            }
        }
    }
    if (b = p->p_seltext)
    {
        TEHandle teh = b->b_un.b_teh;
        if (!(b->b_textdirty))
        {
            patcher_setuntype(p);
            b->b_textdirty = 1;
        }
        te_key(teh, k);
        box_checktextheight(b);
        patcher_dirty(p, 1);
    }
    else if (b = p->p_typeat)
    {
        if (logicaltime > p->p_notypetime)
            p->p_typeat = 0;
        else
        {
            mess1(b, &s_key, k);
            p->p_notypetime = logicaltime + TYPEINTER;
        }
    }
}

patcher_close(struct patcher *p)
{
    freeobject((t_object *) p);
}

struct patcher *patcher_saving;

void *patcher_saveto(struct patcher *p, char *fn, t_symbol *vol)
{
    struct binbuf *x = binbuf_new();
    int ret = 0;
    int ac, i;
    t_atom at[256], *ap;
    t_symbol **declist;
    binbuf_error = 0;
    p->p_vol = vol;
    SETSYM(at, gensym("max"));
    SETSYM(at + 1, gensym("v2"));
    declare__getforpatcher(p, &ac, &declist);
    if (ac > 256)
        post("warning: declarations truncated to 256"), ac = 256;
    for (ap = at + 2, i = ac; i--; ap++, declist++)
        SETSYM(ap, *declist);
    binbuf_insert(x, NIL, ac + 2, at);
    patcher_flip(p);
    patcher_saving = p;
#if protoerror
    patcher_save(p, x, 1L);
#else
    patcher_save(p, x);
#endif
    patcher_saving = 0;
    patcher_flip(p);
    /* post("saved... err = %d", binbuf_error); */
    if (!binbuf_error)
        ret = binbuf_write(x, fn, vol, p->p_wind->w_bin), binbuf_free(x);
    else
        binbuf_free(x), memouch(), ret = 1;
    return ((void *) ret);
}

static int vertical;

void *patcher_lineup(struct patcher *p)
{
    long i, to = 1234567, xx;
    struct box *b;
    long xmin = 0x40000000, xmax = -0x40000000;
    long ymin = 0x40000000, ymax = -0x40000000;
    int n = 0;
    for (b = p->p_box; b; b = b->b_next)
    {
        b->b_temp = 0;
        if (b->b_selected)
        {
            if (b->b_rect.left < xmin)
                xmin = b->b_rect.left;
            if (b->b_rect.left > xmax)
                xmax = b->b_rect.left;
            if (b->b_rect.top < ymin)
                ymin = b->b_rect.top;
            if (b->b_rect.top > ymax)
                ymax = b->b_rect.top;
            n++;
        }
    }
    if (!n)
        return;
    vertical = (xmax - xmin > ymax - ymin);
    unact = UNLINEUP;
    unwind = p->p_wind;
    for (b = p->p_box; b; b = b->b_next)
    {
        if (b->b_selected)
        {
            if (vertical)
            {
                if (b->b_rect.left == xmin)
                    to = b->b_rect.top;
            }
            else
            {
                if (b->b_rect.top == ymin)
                    to = b->b_rect.left;
            }
        }
    }
    if (to == 1234567)
    {
        bug("lineup");
        return;
    }
    for (b = p->p_box; b; b = b->b_next)
        if (b->b_selected)
        {
            b->b_temp = (vertical ? b->b_rect.top : b->b_rect.left);
            if (vertical)
                box_move(b, b->b_rect.left, (int) to);
            else
                box_move(b, (int) to, b->b_rect.top);
        }
}

void *patcher_fixwidth(struct patcher *p)
{
    struct box *b;
    unact = UNSTRETCH;
    unwind = p->p_wind;
    for (b = p->p_box; b; b = b->b_next)
        if (b->b_selected && b->b_binbuf)
        {
            int newwidth = te_wantwidth(b->b_un.b_teh);
            if (newwidth + 7 != b->b_rect.right - b->b_rect.left)
            {
                b->b_temp = b->b_rect.right - b->b_rect.left;
                box_size(b, newwidth + 7,
                         b->b_rect.bottom - b->b_rect.top);
            }
        }
        else
            b->b_temp = 0;
}

patcher_clear(struct patcher *p)
{
    if (!patcher_docopy(p, 1))
    {
        patcher_doclear(p);
        unact = UNCLEAR;
        unwind = p->p_wind;
    }
}

patcher_cut(struct patcher *p)
{
    if (!patcher_docopy(p, 0))
    {
        patcher_doclear(p);
        unact = UNCUT;
        unwind = p->p_wind;
    }
}

void *patcher_copy(struct patcher *p)
{
    patcher_docopy(p, 0);
}

patcher_doclear(struct patcher *p)
{
    struct box *b;
    struct patchline *l;
    TEHandle teh;
    if ((b = p->p_seltext) &&
        (te_selstart(teh = b->b_un.b_teh) ||
         te_selend(teh) < te_length(teh)))
    {
        te_delete(teh);
        box_checktextheight(b);
        if (!(b->b_textdirty))
        {
            patcher_setuntype(p);
            b->b_textdirty = 1;
        }
    }
    else
    {
        char savelock = lockout;
        lockout = 1;
        patcher_savelines(p);
        for (l = p->p_line; l; l = l->l_next)
            if (l->l_selected)
                freeobject((t_object *) l);
        for (b = p->p_box; b; b = b->b_next)
            if (b->b_selected)
                box_free(b);
        lockout = savelock;
    }
    patcher_dirty(p, 1);
    patcher_span(p);
}

static void *undobuf = 0;
void *copybuf = 0;              /* this is used in patcher.c */

patcher_docopy(struct patcher *p, int toundo)
{
    struct box *b;
    void **buf = (toundo ? &undobuf : &copybuf);
    if (!toundo && (b = p->p_seltext))
    {
        if (te_selstart(b->b_un.b_teh) < te_selend(b->b_un.b_teh))
            t_cpytext(), te_copy(p->p_seltext->b_un.b_teh);
        if (b->b_textdirty)
            binbuf_text(b->b_binbuf, te_text(b->b_un.b_teh),
                        te_length(b->b_un.b_teh));
    }
    noundo(p->p_wind);
    if (copybuf)
        binbuf_free(copybuf), copybuf = 0;
    *buf = binbuf_new();
    patcher_cpto(p, *buf);
    patcher_span(p);
    return (binbuf_error);
}

extern PicHandle t_pictbuf;

patcher_cpto(struct patcher *p, void *buf)
{
    /* first copy to a patcher-save buffer */
    struct box *b;
    long n;
    struct patchline *l;
    binbuf_error = 0;
    for (n = 1, b = p->p_box; b && !binbuf_error; b = b->b_next)
    {
        if (b->b_selected)
        {
            mess3(b, &s_psave, buf, n, &s_patcher);
            b->b_temp = p->p_nselect - n++;
        }
    }
    for (l = p->p_line; l && !binbuf_error; l = l->l_next)
    {
        long pono = 0, pino = 0;
        struct outlet *po;
        struct inlet *pi;
        if (!(l->l_box1->b_selected && l->l_box2->b_selected))
            continue;
        for (po = l->l_box1->b_ob.o_outlet; po && po != l->l_outlet;
             po = po->o_next, pono++);
        b = l->l_box2;
        if (!(b->b_drawfirstin) ||
            l->l_inlet != (struct inlet *) (b->b_firstin))
        {
            if ((b->b_drawfirstin))
                pino++;
            for (pi = b->b_ob.o_inlet; pi && pi != l->l_inlet;
                 pi = pi->i_next, pino++);
        }
        vfun(buf, binbuf_insert, NIL,
             "ssllll", &s_patcher, &s_connect,
             (long) (l->l_box1->b_temp), pono,
             (long) (l->l_box2->b_temp), pino);
    }
    if (binbuf_error)
        binbuf_clear(buf), memouch();


    /* Copy graphics.  */
    if (buf == copybuf)
    {
        Rect r;
        char pastewas = p->p_topasteboard;      /* should be 0 */
        p->p_topasteboard = 1;
        r.top = r.left = 32000;
        r.bottom = r.right = -32000;
        for (b = p->p_box; b; b = b->b_next)
            if (b->b_selected)
            {
                if (b->b_rect.left < r.left)
                    r.left = b->b_rect.left;
                if (b->b_rect.top < r.top)
                    r.top = b->b_rect.top;
                if (b->b_rect.right > r.right)
                    r.right = b->b_rect.right;
                if (b->b_rect.bottom > r.bottom)
                    r.bottom = b->b_rect.bottom;
            }
        for (l = p->p_line; l; l = l->l_next)
            if (l->l_selected)
            {
                if (l->l_x1 < r.left)
                    r.left = l->l_x1;
                if (l->l_x1 > r.right)
                    r.right = l->l_x1;
                if (l->l_x2 < r.left)
                    r.left = l->l_x2;
                if (l->l_x2 > r.right)
                    r.right = l->l_x2;
                if (l->l_y1 < r.top)
                    r.top = l->l_y1;
                if (l->l_y1 > r.bottom)
                    r.bottom = l->l_y1;
                if (l->l_y2 < r.top)
                    r.top = l->l_y2;
                if (l->l_y2 > r.bottom)
                    r.bottom = l->l_y2;
            }
        if (r.left == 32000)
            goto refuse;        /* this shouldn't be possible */
        r.top--;                /* add one pixel to top and bottom */
        r.bottom++;

        wind_copytorect(p->p_wind, r.left, r.top, r.bottom, r.right);
      refuse:
        p->p_topasteboard = pastewas;
    }
}

extern int box_offset;          /* <a> */

patcher_paste(struct patcher *p)
{
    t_symbol *was = defvolume;
    defvolume = p->p_vol;
    if (p->p_seltext)
    {
        struct box *b = p->p_seltext;
        TEHandle teh = b->b_un.b_teh;
        te_paste(teh);
        box_checktextheight(b);
        if (!(b->b_textdirty))
        {
            patcher_setuntype(p);
            b->b_textdirty = 1;
        }
    }
    else
    {
        box_offset = 7;         /* <a> offset by 7 */
        if (copybuf)
            patcher_pfrom(p, copybuf);
        box_offset = 0;         /* <a> reset afterward */
    }
    defvolume = was;
    patcher_dirty(p, 1);
    patcher_span(p);
}

patcher_pfrom(struct patcher *p, void *buf)
{
    /* <a> long nhidden;
       void *hidden; */
    t_symbol *vwas = defvolume, *pwas = t_getpath();
    struct atom at[10];
    struct box *b;

    defvolume = p->p_vol;
    crosshatch(at);
    if (p->p_lock)
        patcher_lock(p, 0);
    else
        patcher_noselect(p);
    s_patcher.s_thing = (struct object *) p;
    patcher_becomecurrent(p);
    memislow = 0;
    binbuf_make(buf, 9, at, NIL);
    if (memislow)
        memouch();
    patcher_becomecurrent(NIL);
    p->p_pasted = 1;
    if (p->p_cpno >= 0 && p->p_cpno < h_nproc)
        for (b = p->p_box; b; b = b->b_next)
            if (b->b_selected)
                (*egetfn(b, &s__active)) (b, 1);
    if (p->p_seltext)
        patcher_deselectbox(p, p->p_seltext);
    defvolume = vwas;
    t_setpath(pwas);
}

patcher_dup(struct patcher *p)
{
    patcher_copy(p);
    patcher_noselect(p);
    patcher_paste(p);
    patcher_dirty(p, 1);
    patcher_span(p);
}

static int un_xwas, un_ywas;
static struct box *un_boxwas;

patcher_setunmove(struct patcher *p, int new)
{
    struct box *b;
    unact = 0;
    if (new)
        return;
    for (b = p->p_box; b; b = b->b_next)
    {
        if ((b->b_temp = b->b_selected) && !unact)
        {
            un_boxwas = b;
            un_xwas = b->b_rect.left;
            un_ywas = b->b_rect.top;
            unact = UNMOVE;
            unwind = p->p_wind;
        }
    }
}

patcher_setunstretch(struct patcher *p)
{
    struct box *b;
    unact = UNSTRETCH;
    unwind = p->p_wind;
    for (b = p->p_box; b; b = b->b_next)
        b->b_temp = (b->b_selected ?
                     (b->b_rect.right - b->b_rect.left) : 0);
}

#define NLWAS 100
static struct linewas {
    int x_box1;
    int x_box2;
    char x_index1;
    char x_index2;
} **linebuf = 0;
static int nbaglines;

patcher_setuntype(struct patcher *p)
{
    struct box *b;
    if (undobuf)
        binbuf_free(undobuf);
    undobuf = binbuf_new();
    patcher_cpto(p, undobuf);
    for (b = p->p_box; b; b = b->b_next)
        b->b_temp = b->b_selected;
    unact = UNTYPE;
    unwind = p->p_wind;
    nbaglines = 1;
}

patcher_enum(struct patcher *p) /* put the index of each box in its b_temp field */
{
    struct box *b;
    int x = 0;
    for (b = p->p_box; b; b = b->b_next)
        b->b_temp = x++;
}

patcher_enumtocut(struct patcher *p)    /* same, but assumes you will cut and repaste */
/* (putting all selected boxes in front.) */
{
    struct box *b;
    int x = 0;
    for (b = p->p_box; b; b = b->b_next)
        if (b->b_selected)
            b->b_temp = p->p_nselect - 1 - (x++);
    for (b = p->p_box; b; b = b->b_next)
        if (!b->b_selected)
            b->b_temp = x++;
}

/* remember all the lines connected to a box (and possibly zap them).  Also 
   set the temp field to remember which box this was for undoing. */
patcher_baglines(struct patcher *p, struct box *b, int zap)
{
    struct box *b2;
    struct patchline *l, *l2;
    struct linewas *lp;
    if (!linebuf)
        linebuf =
            (struct linewas **)
            t_newhandle((long) (NLWAS * sizeof(struct linewas)));
    patcher_enum(p);
    lp = *linebuf;
    nbaglines = 0;
    for (l = p->p_line; l;)
    {
        l2 = l->l_next;
        if (l->l_box1 == b || l->l_box2 == b)
        {
            if (nbaglines < NLWAS)
            {
                lp->x_box1 = l->l_box1->b_temp;
                lp->x_box2 = l->l_box2->b_temp;
                lp->x_index1 = patchline_index1(l);
                lp->x_index2 = patchline_index2(l);
                nbaglines++;
                lp++;
            }
            if (zap)
                freeobject((t_object *) l);
        }
        l = l2;
    }
    for (b2 = p->p_box; b2; b2 = b2->b_next)
        b2->b_temp = (b == b2);
}

patcher_savelines(struct patcher *p)
/* remember lines for undoing a cut or clear */
/* only bother with ones that the paste mechanism */
/* won't put back by itself. */
{
    struct box *b2;
    struct patchline *l, *l2;
    struct linewas *lp;
    /* <a> bug fix: cast following arg to long */
    if (!linebuf)
        linebuf =
            (struct linewas **)
            t_newhandle((long) (NLWAS * sizeof(struct linewas)));
    patcher_enumtocut(p);
    lp = *linebuf;
    nbaglines = 0;
    for (l = p->p_line; l; l = l->l_next)
    {
        /* save selected lines that don't connect two selected boxes,
           and also lines between selected and unselected boxes. */
        if ((l->l_selected
             && (!l->l_box1->b_selected || !l->l_box2->b_selected))
            || l->l_box1->b_selected != l->l_box2->b_selected)
        {
            if (nbaglines < NLWAS)
            {
                lp->x_box1 = l->l_box1->b_temp;
                lp->x_box2 = l->l_box2->b_temp;
                lp->x_index1 = patchline_index1(l);
                lp->x_index2 = patchline_index2(l);
                nbaglines++;
                lp++;
            }
        }
    }
}

patcher_restorelines(struct patcher *p)
{
    int i;
    if (!linebuf)
        return;
    /* try to reconnect lines */
    for (i = 0; i < nbaglines; i++)
    {
        struct linewas *lp = (*linebuf) + i;
        patcher_connect(p, (long) (lp->x_box1), (long) (lp->x_index1),
                        (long) (lp->x_box2), (long) (lp->x_index2));
    }
}

noundo(void *x)
{
    if (x == unwind)
    {
        if (undobuf)
            binbuf_free(undobuf), undobuf = 0;
        if (linebuf)
            DisposHandle(linebuf), linebuf = 0;
        unact = 0;
    }
}

whatundo(void *wind)
{
    return (wind == unwind ? unact : 0);
}

patcher_undo(struct patcher *p)
{
    switch (unact)
    {
    case UNMOVE:
        {
            int dx, dy;
            struct box *b = un_boxwas;
            if (NOGOOD(b))
                bug("undo move");
            dx = un_xwas - b->b_rect.left;
            dy = un_ywas - b->b_rect.top;
            for (b = p->p_box; b; b = b->b_next)
            {
                if (b->b_temp)
                    box_move(b, b->b_rect.left + dx, b->b_rect.top + dy);
            }
        }
        break;
    case UNSTRETCH:
        {
            struct box *b;
            for (b = p->p_box; b; b = b->b_next)
            {
                if (b->b_temp)
                {
                    Rect *r = &b->b_rect;
                    int w = r->right - r->left;
                    int oldw = b->b_temp;
                    int h = r->bottom - r->top;
                    if (w != oldw)
                        box_size(b, oldw, (b->b_growy ? oldw : h));
                }
            }
        }
        break;
    case UNTYPE:
        {
            struct box *b;
            if (!undobuf)
                bug("undo");
            for (b = p->p_box; b; b = b->b_next)
                if (b->b_temp)
                    box_free(b);
            patcher_pfrom(p, undobuf);
#if 0
            if (p->p_seltext)   /* why was this ever here? */
                patcher_restorelines(p, p->p_seltext);
#else
            patcher_restorelines(p);
#endif
        }
        break;
    case UNCUT:
    case UNCLEAR:
        {
            int i;
            void *buf = (unact == UNCLEAR ? undobuf : copybuf);
            if (!buf)
                bug("undo cut");
            patcher_pfrom(p, buf);
            if (!linebuf)
                bug("lines");
            patcher_restorelines(p);
        }
        break;
    case UNLINEUP:
        {
            struct box *b;
            for (b = p->p_box; b; b = b->b_next)
                if (b->b_temp)
                {
                    if (vertical)
                        box_move(b, b->b_rect.left, b->b_temp);
                    else
                        box_move(b, b->b_temp, b->b_rect.top);
                }
        }
        break;
    default:
        post("sorry: can't undo your last edit");
    }
    noundo(p->p_wind);
}

void *patcher_getmain(t_patcher *p)
{
    while (p->p_vnewobj)
        p = ((struct box *) (p->p_vnewobj))->b_patcher;
    return (p);
}


static void *findbuf;
struct patcher *findpat;
int findindex;
t_box *patcher_lastboxfound;

#define OKITEM 1
#define CANCELITEM 2
#define TEXTITEM 3

void *wind_front();
patcher_find()
{
#if (0)
    void *front;
    DialogRecord dog;
    short hit = 2001;
    char doit = 0, touched = 0;
    char buf[256];
    int foo = 0;
    long nsize;
    char **tbuf = t_newhandle(4L);
    if (!findbuf)
        findbuf = binbuf_new();
    binbuf_totext(findbuf, tbuf, &nsize);
    if (nsize > 255)
        nsize = 255;
    strncpy(buf, *tbuf, (int) nsize);
    buf[nsize] = 0;
    DisposHandle(tbuf);
    GetNewDialog(1005, &dog, -1L);
    settextitem(&dog, TEXTITEM, buf);
    SelIt_text(&dog, TEXTITEM, (short) 0, (short) 32767);
    while (hit != OKITEM)
    {
        ModalDialog(NIL, &hit);
        if (hit == TEXTITEM)
            touched = 1;
        else if (hit == CANCELITEM)
        {
            CloseDialog(&dog);
            return;
        }
    }
    if (touched)
    {
        char *x = buf;
        gettextitem(&dog, TEXTITEM, buf);
        binbuf_text(findbuf, &x, (long) strlen(buf));
    }
    CloseDialog(&dog);
    if (!(front = wind_front()) || ob_messlist(front) != patcher_class)
    {
        error("search only works in patchers");
        return;
    }
    findpat = (t_patcher *) patcher_getmain(front);
    findindex = 0;
#endif
}

void patcher__findstringouch(char *s)
{
    t_patcher *p;
    if (!(p = wind_front()) || ob_messlist(p) != patcher_class)
        ouch("sorry, patchers only (so far!)");
    else if (!patcher_findstring(p, s, 0))
        ouch("%s: not found", s);
}

int patcher_findstring(t_patcher *p, char *s, int ouchonfail)
{
    if (!findbuf)
        findbuf = binbuf_new();
    binbuf_text(findbuf, &s, (long) strlen(s));
    findpat = (t_patcher *) patcher_getmain(p);
    findindex = 0;
    return (patcher_findagain(ouchonfail));
}

int patcher_findagain(int ouchonfail)
{
    struct box *b, *b2;
    int count = 0;
    if (!findpat)
    {
        ouch("No search in progress");
        return (0);
    }
    for (b = findpat->p_box, count = 0; b && count < findindex;
         b = b->b_next, count++);
    while (1)
    {
        struct patcher *p2;
        while (!b)
        {
            if (!(b = findpat->p_vnewobj))
                goto nodice;
            findpat = b->b_patcher;
            b = b->b_next;
        }
        if (!b->b_binbuf)
        {
            b = b->b_next;
            continue;
        }
        if (binbuf_subset(b->b_binbuf, findbuf))
        {
            GrafPort *gp;
            patcher_front(findpat);
            gp = patcher_setport(findpat);
            if (findpat->p_lock)
                patcher_lock(findpat, 0);
            patcher_noselect(findpat);
            patcher_selectbox(findpat, b);
            patcher_lastboxfound = b;
            for (b2 = findpat->p_box, findindex = 1; b2 && b2 != b;
                 b2 = b2->b_next, findindex++);
            t_resetport(gp);
            return (1);
        }
        if (ob_messlist(b) == vnewobj_class && (p2 = b->b_firstin)
            && ob_messlist(p2) == patcher_class)
        {
            findpat = p2;
            b = findpat->p_box;
        }
        else
            b = b->b_next;
    }
  nodice:
    findpat = 0;
    patcher_lastboxfound = 0;
    if (ouchonfail)
        ouch("Not found");
    return (0);
}

void *patcher_lastfound()
{
    if (!patcher_lastboxfound)
        return (0);
    else
        return (patcher_lastboxfound->b_firstin);
}

/* return 1 if segment from (x1, y1) to (x2, y2) hits the segment
   /* from (x3, y3) to (x4, y3)
    */
segscross(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4)
{
    long dx21 = x2 - x1, dx31 = x3 - x1, dx41 = x4 - x1;
    long dy21 = y2 - y1, dy31 = y3 - y1, dy41 = y4 - y1;
    long dx32 = x3 - x2, dx34 = x3 - x4;
    long dy32 = y3 - y2, dy34 = y3 - y4;
    long det1 = dx31 * dy21 - dy31 * dx21;
    long det2 = dx41 * dy21 - dy41 * dx21;
    long det3 = dx34 * dy31 - dy34 * dx31;
    long det4 = dx34 * dy32 - dy34 * dx32;
    if (!det1 && !det2)         /* if collinear */
    {
        if (y1 == y2 && y1 == y3 && y1 == y4)
            return (sc1d(x1, x2, x3, x4));
        else
            return (sc1d(y1, y2, y3, y4));
    }
    else
        return (((det1 <= 0 && det2 >= 0) || (det1 >= 0 && det2 <= 0))
                && ((det3 <= 0 && det4 >= 0) || (det3 >= 0 && det4 <= 0)));
}

/* same but second seg is horizontal */
segxh(int x1, int y1, int x2, int y2, int x3, int y3, int x4)
{
    long dx21 = x2 - x1, dx31 = x3 - x1, dx41 = x4 - x1;
    long dy21 = y2 - y1, dy31 = y3 - y1;
    long tmp = dy31 * dx21;
    long det1 = dx31 * dy21 - tmp;
    long det2 = dx41 * dy21 - tmp;
    /* if collinear check in 1d */
    if (y1 == y2)
    {
        if (y1 == y3)
            return (sc1d(x1, x2, x3, x4));
        else
            return (0);
    }
    else
        return (((det1 < 0 && det2 > 0) || (det1 > 0 && det2 < 0))
                && ((y1 < y3 && y2 > y3) || (y1 > y3 && y2 < y3)));
}

/* test whether the segments (x1, x2) and (x3, x4) intersect */
sc1d(int x1, int x2, int x3, int x4)
{
    int foo;
    if (x2 < x1)
        foo = x2, x2 = x1, x1 = foo;
    if (x4 < x3)
        foo = x4, x4 = x3, x3 = foo;
    return (x2 > x3 && x4 > x1);
}

/* test whether the line from (x1, y1) to (x2, y2) hits the
   rectangle (which contains its left and top sides, not its
   right and bottom). The line contains all points the toolbox
   would draw; i.e. all points within 1/2 either in x or y of
   the (real) line.  So we grow the rectangle by 1/2 point on all
   sides.  I think you can then test for intersection of the closed line
   with the open rectangle. */

seghitsrect(int x1, int y1, int x2, int y2, Rect *r)
{
    int left = (r->left << 1) - 1, right = (r->right << 1) + 1,
        top = (r->top << 1) - 1, bottom = (r->bottom << 1) + 1;
    x1 <<= 1;
    y1 <<= 1;
    x2 <<= 1;
    y2 <<= 1;
    return (segxh(x1, y1, x2, y2, left, top, right) ||
            segxh(x1, y1, x2, y2, left, bottom, right) ||
            segxh(y1, x1, y2, x2, top, right, bottom) ||
            segxh(y1, x1, y2, x2, top, left, bottom) ||
            (x1 >= left && x1 <= right && y1 >= top && y1 <= bottom));
}

interrect(Rect *r1, Rect *r2)   /* return 1 if 2 rectangles intersect */
{
    return (r1->right > r2->left && r2->right > r1->left &&
            r1->bottom > r2->top && r2->bottom > r1->top);
}

ptnearseg(int a1, int b1, int x1, int y1, int x2, int y2)
{
    int dx = x2 - x1, dy = y2 - y1;
    if (dx < 0)
        dx = -dx;
    if (dy < 0)
        dy = -dy;
    if (dx > dy)
        return (segscross(a1, b1 - 3, a1, b1 + 3, x1, y1, x2, y2));
    else
        return (segscross(a1 - 3, b1, a1 + 3, b1, x1, y1, x2, y2));
}
