
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.4 $ IRCAM $Date: 1993/09/08 09:00:38 $
 *
 * FTS by Miller Puckette
 *
 */

#define DEFSIZE 128             /* default size 128 elements */
#define DEFRANGE 128            /* range 0-127 */
#define ELEM long               /* table element type */

typedef struct tabwind {
    struct object w_ob;
    struct symbol *w_sym;
    ELEM **w_tab;               /* relocatable vector for table contents */
    long w_size;                /* size of table */
#ifdef GRAPHICS
    t_wind *w_wind;
    long w_range;               /* editable range (y values) */
    long w_xoff;                /* x-value of first point displayed */
    long w_yoff;                /* y-value of top line */
    long w_xshift;              /* x-shift for displaying */
    long w_yshift;              /* y-shift */
    long w_xzoom;               /* x zoom */
    long w_yzoom;               /* y zoom */
    long w_cliprange;           /* range, anded to be 0 mod 1 << w_xzoom */
    long w_dpoints;             /* display points (size >> xshift rounded up) */
    long w_dragx;               /* previous x-value changed by dragging */
    long w_selstart;            /* starting sample of selected region (inclusive) */
    long w_selend;              /* ending sample of selected region (exclusive) */
    long w_saveno;              /* serial number of last save */
    int w_xdisp;                /* number of points displayed */
    int w_ydisp;                /* number of y-values displayed */
    int w_cpno;                 /* CP number, -1 if none */
    void *w_cpr;                /* CP responder */
    char w_signed;              /* flag for signed display */
    char w_selmode;             /* flag for select mode */
    char w_embed;               /* save with containing patcher */
#else
    int w_refcount;             /* reference count */
    long w_lodirty;             /* index of first element to send to host */
    long w_hidirty;             /* 1 + index of last element to send */
    long w_nextup;              /* index of next element to send up; <0 if clean */
    void *w_hq;                 /* host queue element */
#endif
} t_tabwind;

#ifdef GRAPHICS
#define VIS 1
#define SIGNED 2
#define SCRATCH 4
#define EMBED 16

#define LMARGIN 5
#define RMARGIN 48
#define TMARGIN 5
#define BMARGIN 45
#define BAR 40
#define STDWIDTH (LMARGIN + RMARGIN + DEFSIZE)
#define STDHEIGHT (TMARGIN + BMARGIN + DEFRANGE)

#define SHWIDTH 25
#define NWIDTH 35
#define NHEIGHT 13
#define ASCENT 10
#define XSHLEFT LMARGIN
#define XSHRIGHT (XSHLEFT + SHWIDTH)
#define XSHTOP (15 + TMARGIN)
#define XSHBOTTOM (XSHTOP + NHEIGHT)
#define XLOCX (XSHRIGHT + 3)
#define XLOCY (XSHTOP + ASCENT)
#define YSHLEFT (100)
#define YSHRIGHT (YSHLEFT + SHWIDTH)
#define YSHTOP (XSHTOP)
#define YSHBOTTOM (YSHTOP + NHEIGHT)
#define YLOCX (YSHRIGHT + 3)
#define YLOCY XLOCY
#define MINWIDTH (XSHLEFT + NWIDTH + 2 * SHWIDTH + 5)
#define MINDISP 64
#define HOP -10
#define DROP 3
#define MODEX (7 + LMARGIN)
#define MODEY 19
#define MODEW 21
#define MODEH 21
#define MAXSHIFT 6
#define MINSHIFT (-3)
#endif

void *stabwind(struct symbol *s), *gettabwind(struct symbol *s,
                                              long int size,
                                              long int range, int x1,
                                              int y1, int x2, int y2,
                                              int xshift, int yshift,
                                              int flags);

#define UPCHUNK 10
typedef struct updbuf           /* buffer for cp-to-host update */
{
    long u_x;
    ELEM u_shit[UPCHUNK];
} t_updbuf;
