#ifndef lint
static char rcsid[] =
    "$Revision: 2.3 $ IRCAM $Date: 1993/11/17 14:47:16 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.3 $ IRCAM $Date: 1993/11/17 14:47:16 $
 *
 * FTS by Miller Puckette
 *
 */


#include <stdio.h>
#include "m_mess.h"
#include "t_mac.h"
#include "t_wind.h"
#include "t_menu.h"
#include "p_table.h"
#include "p_pic.h"
#include "m_ctype.h"

t_wind *wind_new(void *assoc, int a, int b, int c, int d, int flags);
void *outlet(void *op, struct symbol *s2), binbuf_set(struct binbuf *x,
                                                      struct symbol *s,
                                                      int argc,
                                                      struct atom *argv),
binbuf_insert(struct binbuf *x, struct symbol *s, int argc,
              struct atom *argv);

extern struct symbol s_table, s_tabwind, s__event, s_outlet,
    s_int1, s_int2, s_randi, s_inv, s_save, s_vtable,
    s_, s_dblclick, s_filename, s_front, s_size, s_float, s_in1,
    s_set, s_new, s_vtab, s__t, s_invtab, s_obj,
    s_quantile, s_sum, s_clear, s_const, s__New, s__Tab, s__set;

void *tabwind_erase(struct tabwind *x);
void *tabwind_chkscroll(struct tabwind *x);
void tabwind_list(struct tabwind *x, struct symbol *s, int ac,
                  struct atom *av);
void tabwind_set(struct tabwind *x, struct symbol *s, int ac,
                 struct atom *av);
void tabwind_const(struct tabwind *x, long int n);
void tabwind_clear(struct tabwind *x);
extern void *tabwind_class;

/* xvals to/from pixels into table */
#define PTOX(p) (((long)(x->w_wind->w_xoffset + (p)) \
      << x->w_xshift) >> x->w_xzoom)
#define XTOP(n) ((((n) << x->w_xzoom) >> x->w_xshift) - x->w_wind->w_xoffset)

#define PTOY(p) (x->w_cliprange - \
     ((((long)(x->w_wind->w_yoffset + (p))) << x->w_yshift) >> x->w_yzoom))

#define YTOP(n) (((((x->w_cliprange - (n)) << x->w_yzoom) >> x->w_yshift))\
     - x->w_wind->w_yoffset)

int tabwind_cliprange(struct tabwind *x);
int tabwind_invsel(struct tabwind *x, long int start, long int end);
int tabwind_clrsel(struct tabwind *x, long int start, long int end);

drawrect(int x, int y, int w, int h, int color)
{
    int i;
    Rect r1, r2;
    if (color)
        t_paintrect(x, y, x + w, y + h);
    else
        t_eraserect(x, y, x + w, y + h);
}

tabwind_drawpoint(struct tabwind *x, long int index, long int val,
                  int color)
{
    int pictx = XTOP(index), picty;
    long foo;
    if (pictx < 0 || pictx >= x->w_xdisp)
        return;
    picty = YTOP(val);
    foo = 1 << x->w_yzoom;
    if (picty < 0)
        picty = 0;
    else if (picty >= x->w_ydisp)
        picty = x->w_ydisp - foo;
    drawrect(LMARGIN + pictx, (TMARGIN + 1) + picty,
             1 << (int) (x->w_xzoom), (int) foo, color);
}

tabwind_movepoint(struct tabwind *x, long int n1, long int old,
                  long int new)
{
    Rect r;
    int xval, y, pictx;
    if ((pictx = XTOP(n1)) >= 0 &&
        pictx < x->w_xdisp && !(n1 & ((1 << x->w_xshift) - 1)))
    {
        void *port = wind_setport(x->w_wind);
        tabwind_drawpoint(x, n1, old, 0);
        tabwind_drawpoint(x, n1, new, 1);
        t_resetport(port);
    }
}

drawzoom(struct tabwind *x, long int val, int horiz)
{
    int height = x->w_wind->w_y2 - x->w_wind->w_y1;
    int xp, yp;
    char buf[80];
    long up = 1, down = 1;
    if (horiz)
        xp = XSHLEFT, yp = x->w_ydisp + XSHTOP;
    else
        xp = YSHLEFT, yp = x->w_ydisp + YSHTOP;
    if (val >= 0)
        up <<= val;
    else
        down <<= -val;
    sprintf(buf, "%ld:%ld", down, up);
    t_eraserect(xp + 1, yp + 1, xp + (SHWIDTH - 1), yp + (NHEIGHT - 1));
    t_drawstring(xp + 2, yp + ASCENT, buf);
}

getshift(int val, long int *pp, long int *np)
{
    if (val > 0)
        *pp = val, *np = 0;
    else
        *pp = 0, *np = -val;
}

void *tabwind_update(struct tabwind *x, int margin);
void *tabwind_erase(struct tabwind *x)
{
    struct windpub *w = x->w_wind;
    GrafPort *gp = wind_setport(w);
    t_eraserect(0, 0, w->w_x2 - w->w_x1, w->w_y2 - w->w_y1);
#if protoerror
    tabwind_update(x);
#else
    tabwind_update(x, 0);
#endif
    t_resetport(gp);
}

static int zooming = 0;

void *tabwind_idle(struct tabwind *x, Point pt)
{
    int xpx, ypx;
    long xv, yv;
    int height = x->w_wind->w_y2 - x->w_wind->w_y1;
    static struct tabwind *tabwas;
    static long xwas, ywas;
    static int alt;
    if (zooming)
        return;
    if (pt.h > x->w_xdisp + 2 * LMARGIN || pt.h < 0 ||
        pt.v > x->w_ydisp + 2 * LMARGIN || pt.v < 0)
    {
        if (xwas != 123456 && ywas != 123456)
        {
            t_eraserect(XLOCX - 1, x->w_ydisp + (HOP + XLOCY),
                        XLOCX + NWIDTH, x->w_ydisp + (DROP + XLOCY));
            t_eraserect(YLOCX - 1, x->w_ydisp + (HOP + YLOCY),
                        YLOCX + NWIDTH, x->w_ydisp + (DROP + YLOCY));
            xwas = ywas = 123456;
        }
        arrowcursor();
        return;
    }
    else if ((alt++) & 1)
    {
        xpx = pt.h - LMARGIN;
        if (xpx >= x->w_xdisp)
            xpx = x->w_xdisp - 1;
        else if (xpx < 0)
            xpx = 0;
        xv = PTOX(xpx);
        if (x != tabwas || xv != xwas)
        {
            t_eraserect(XLOCX - 1, x->w_ydisp + (HOP + XLOCY),
                        XLOCX + NWIDTH, x->w_ydisp + (DROP + XLOCY));
            t_drawlong(XLOCX, x->w_ydisp + XLOCY, xv);
        }
        tabwas = x;
        xwas = xv;
    }
    else
    {
        ypx = pt.v - (TMARGIN + 1);
        if (ypx < 0)
            ypx = 0;
        else if (ypx >= x->w_ydisp)
            ypx = x->w_ydisp - 1;
        yv = PTOY(ypx);
        if (x != tabwas || yv != ywas)
        {
            t_eraserect((YLOCX - 1), x->w_ydisp + (HOP + YLOCY),
                        (YLOCX + NWIDTH), x->w_ydisp + (DROP + YLOCY));
            t_drawlong(YLOCX, x->w_ydisp + YLOCY, yv);
        }
        tabwas = x;
        ywas = yv;
    }
    if (x->w_selmode)
        editcursor();
    else
        arrowcursor();
}

static long dragtime;
#define SH 12

void tabledrag(struct tabwind *x, Point pt)
{
    int xpix = pt.h - LMARGIN, y = pt.v - (TMARGIN + 1), i;
    long xval, xwas = x->w_dragx;
    Rect r;
    long yval, xx;
    /* if (logicaltime < dragtime) return;
       dragtime = logicaltime + 30; */
    if (xwas >= x->w_size)
        xwas = x->w_size - 1;
    else if (xwas < 0)
        xwas = 0;
    x->w_dragx = xwas;
    if (xpix >= x->w_xdisp)
        xpix = x->w_xdisp - 1;
    else if (xpix < 0)
        xpix = 0;
    xval = PTOX(xpix);
    if (xval < 0 || xval >= x->w_size)
        bug("table");
    if (xval < xwas)
        xwas--;
    else if (xval > xwas)
        xwas++;
    if (y < 0)
        y = 0;
    else if (y >= x->w_ydisp)
        y = x->w_ydisp - 1;
    yval = PTOY(y);
    if (xval == xwas)
    {
        long *lp = (*x->w_tab) + xval, ywas = *lp;
        *lp = yval;
        tabwind_drawpoint(x, xval, ywas, 0);
        tabwind_drawpoint(x, xval, yval, 1);
        if (x->w_cpno >= 0)
            h_vmess(x, &s__set, "ll", xval, yval);
    }
    else
    {
        int backwards = xval < xwas, inc = (backwards ? 1 : -1);
        long xlim = xwas + inc;
        long slope =
            ((yval - (*x->w_tab)[x->w_dragx]) << SH) / (xlim - xval);
        long cum = yval << SH;
        if (backwards)
            slope = -slope;
        for (xx = xval; xx != xlim; xx += inc)
        {
            long *lp = (*x->w_tab) + xx, ywas = *lp;
            *lp = yval = ((cum + (1 << (SH - 1))) >> SH);
            tabwind_drawpoint(x, xx, ywas, 0);
            tabwind_drawpoint(x, xx, yval, 1);
            if (x->w_cpno >= 0)
                h_vmess(x, &s__set, "ll", xx, yval);
            cum += slope;
        }
    }
    x->w_dragx = xval;
}

static int horiz, yref, curshift;

void zoomdrag(struct tabwind *x, Point pt, int but)
{
    if (but)
    {
        int newsh = (yref + pt.v) >> 3;
        if (newsh > MAXSHIFT)
            newsh = MAXSHIFT;
        else if (newsh < MINSHIFT)
            newsh = MINSHIFT;
        if (newsh == curshift)
            return;
        curshift = newsh;
        drawzoom(x, (long) newsh, horiz);
        if (horiz)
        {
            if (newsh > 0)
                x->w_xshift = newsh, x->w_xzoom = 0;
            else
                x->w_xshift = 0, x->w_xzoom = -newsh;
        }
        else
        {
            long lowy = PTOY(x->w_ydisp - 1);
            if (newsh > 0)
                x->w_yshift = newsh, x->w_yzoom = 0;
            else
                x->w_yshift = 0, x->w_yzoom = -newsh;
            tabwind_cliprange(x);
#if OLD
            /* reset scrollbar so that bottom y val is unchanged */
            x->w_wind->w_yoffset = (((x->w_range - lowy)
                                     >> x->w_yshift) << x->w_yzoom) -
                (x->w_ydisp - 1);
#endif
        }
        tabwind_chkscroll(x);
    }
    else
        zooming = 0;
}

/* update x->w_cliprange, used for fast y <--> pixel conversion */
tabwind_cliprange(struct tabwind *x)
{
    long foo = -(1 << x->w_yshift);
    x->w_cliprange = (x->w_range & foo) + foo;
}

void tseldrag(struct tabwind *x, Point pt, int but)
{
    int into = pt.h - LMARGIN, clipx = (into < 0 ? 0 :
                                        (into >
                                         x->w_xdisp ? x->w_xdisp : into));
    long xval = PTOX(clipx);
    if (xval > x->w_dragx)
        tabwind_invsel(x, x->w_dragx, xval);
    else
        tabwind_invsel(x, xval, x->w_dragx);
    x->w_dragx = xval;
    if (!but)
    {
        if (xval < x->w_selstart)
            x->w_selend = x->w_selstart, x->w_selstart = xval;
        else
            x->w_selend = xval;
    }
}

void tabwind_selectall(struct tabwind *x)
{
    x->w_selmode = 1;
    x->w_selstart = 0;
    x->w_selend = x->w_size;
    tabwind_erase(x);
}

void *tabwind_click(struct tabwind *x, Point pt, int dbl, int modifiers)
{
    int height = x->w_wind->w_y2 - x->w_wind->w_y1;
    int x2 = x->w_xdisp + MODEX, y2 =
        (x->w_selmode ? MODEY : (2 * MODEY - 1));
    if (pt.h >= 0 && pt.h < x->w_xdisp + LMARGIN + 10 && pt.v >= 0
        && pt.v < x->w_ydisp + TMARGIN + 10)
    {
        int xpix = pt.h - LMARGIN;
        if (xpix < 0)
            xpix = 0;
        else if (xpix >= x->w_xdisp)
            xpix = x->w_xdisp - 1;
        dragtime = logicaltime;
        if (x->w_selmode)
        {
            long xval = PTOX(xpix);
            x->w_dragx = xval;
            if (modifiers & shiftKey)
            {
                /* do immediate change in select region */
                long mid = (x->w_selstart + x->w_selend) >> 1;
                if (xval >= mid)
                {
                    if (xval > x->w_selend)
                        tabwind_invsel(x, x->w_selend, xval);
                    else
                        tabwind_invsel(x, xval, x->w_selend);
                    x->w_selend = x->w_selstart;
                }
                else
                {
                    if (xval > x->w_selstart)
                        tabwind_invsel(x, x->w_selstart, xval);
                    else
                        tabwind_invsel(x, xval, x->w_selstart);
                    x->w_selstart = x->w_selend;
                }
            }
            else
            {
                tabwind_invsel(x, x->w_selstart, x->w_selend);
                x->w_selstart = x->w_selend = xval;
            }
            tseldrag(x, pt, 1);
            wind_drag(x->w_wind, tseldrag, x);
        }
        else
        {
            if (!(modifiers & shiftKey))
                x->w_dragx = PTOX(xpix);
            x->w_wind->w_dirty = 1;
#if protoerror
            tabledrag(x, pt, 1);
#else
            tabledrag(x, pt);
#endif
            wind_drag(x->w_wind, tabledrag, x);
        }
    }
    else if (pt.h >= x2 && pt.h <= x2 + MODEW
             && pt.v >= y2 && pt.v <= y2 + MODEH)
    {
        x->w_selmode = !x->w_selmode;
        if (x->w_selmode)
        {
            x->w_selstart = 0;
            x->w_selend = x->w_size;
        }
        tabwind_erase(x);
    }
    else if (pt.v >= x->w_ydisp + XSHTOP && pt.v < x->w_ydisp + XSHBOTTOM
             && pt.h >= XSHLEFT && pt.h < XSHRIGHT)
    {
        horiz = 1;
        yref = ((curshift = x->w_xshift - x->w_xzoom) << 3) + 4 - pt.v;
        zooming = 1;
        wind_drag(x->w_wind, zoomdrag, x);
    }
    else if (pt.v >= x->w_ydisp + YSHTOP && pt.v < x->w_ydisp + YSHBOTTOM
             && pt.h >= YSHLEFT && pt.h < YSHRIGHT)
    {
        horiz = 0;
        yref = ((curshift = x->w_yshift - x->w_yzoom) << 3) + 4 - pt.v;
        zooming = 1;
        wind_drag(x->w_wind, zoomdrag, x);
    }
}

/* draw in margin */
void *tabwind_update(struct tabwind *x, int margin)
{
    int i, bottom = x->w_ydisp + TMARGIN + 1, right = x->w_xdisp + LMARGIN,
        max = x->w_ydisp, height = x->w_wind->w_y2 - x->w_wind->w_y1;
    char buf[80];
    if (margin)
        return;                 /* nothing in margin (yet) */

    /* low and hi bounds for x and y */
    t_drawlong(LMARGIN - 4, bottom + 9, PTOX(0));
    t_drawlong(right > 25 ? right - 15 : 10, bottom + 9,
               PTOX(x->w_xdisp - 1));
    t_drawlong(right + 3, TMARGIN + 8, PTOY(0));
    t_drawlong(right + 3, bottom, PTOY(x->w_ydisp - 1));
    /* zoom */
    t_framerect(XSHLEFT, x->w_ydisp + XSHTOP,
                XSHRIGHT, x->w_ydisp + XSHBOTTOM);
    drawzoom(x, x->w_xshift - x->w_xzoom, 1);
    t_framerect(YSHLEFT, x->w_ydisp + YSHTOP,
                YSHRIGHT, x->w_ydisp + YSHBOTTOM);
    drawzoom(x, x->w_yshift - x->w_yzoom, 0);
    /* edit/select buttons */
    if (x->w_ydisp >= MINDISP)
    {
        int x2 = x->w_xdisp + MODEX, y2 = MODEY + 1;
        t_drawbits(PICTABLE, 0, 0, x2, y2, x2 + MODEH,
                   y2 + (2 * MODEH - 1));
        if (x->w_selmode)
            y2 += (MODEH - 1);
        t_invertrect(x2, y2, x2 + (MODEW - 2), y2 + (MODEH - 2));
    }
    /* erase selected region */
    if (x->w_selmode)
        tabwind_clrsel(x, x->w_selstart, x->w_selend);
    /* contents */
    for (i = 0; i < x->w_xdisp; i++)
    {
        long index = PTOX(i);
        if (index >= x->w_size)
            index = x->w_size - 1;
        tabwind_drawpoint(x, index, (*x->w_tab)[index], 1);
    }
    /* rectangle for contents */
    t_framerect(LMARGIN - 1, TMARGIN, right + 1, bottom + 1);
    if (x->w_selmode)
        tabwind_invsel(x, x->w_selstart, x->w_selend);
}

tabwind_invsel(struct tabwind *x, long int start, long int end)
{
    int sraw = XTOP(start), eraw = XTOP(end);
    int pixmin = (sraw > 0 ? sraw : 0);
    int pixmax = (eraw < x->w_xdisp ? eraw : x->w_xdisp);
    t_invertrect(pixmin + LMARGIN, TMARGIN + 1,
                 pixmax + LMARGIN, x->w_ydisp + TMARGIN + 1);
}

tabwind_clrsel(struct tabwind *x, long int start, long int end)
{
    int sraw = XTOP(start), eraw = XTOP(end);
    int pixmin = (sraw > 0 ? sraw : 0);
    int pixmax = (eraw < x->w_xdisp ? eraw : x->w_xdisp);
    t_eraserect(pixmin + LMARGIN, TMARGIN + 1,
                pixmax + LMARGIN, x->w_ydisp + TMARGIN + 1);
}

void *tabwind_filename(struct tabwind *x, char *fn, t_symbol *vol, int bin)
{
    struct symbol *s;
    if (ob_messlist(x) != tabwind_class)
        bug("1");
    s = gensym(fn);
    if (s->s_thing)
        wind_settitle(x->w_wind, fn);
    else
    {
        x->w_sym = s;
        s->s_thing = (struct object *) x;
        wind_filename(x->w_wind, fn, vol, bin);
    }
}

void *tabwind_front(struct tabwind *x)
{
    wind_vis(x->w_wind);
}

void *binbuf_new();
void *tabwind_saveto(struct tabwind *x, char *fn, t_symbol *vol)
{
    int ret;
    struct binbuf *b = binbuf_new();
    vfun((struct object *) b, binbuf_set, NIL, "s", &s_table);
    binbuf_inshandle(b, x->w_tab, x->w_size);
    ret = binbuf_write(b, fn, vol, x->w_wind->w_bin);
    binbuf_free(b);
    return ((void *) ret);
}

extern long max_serialno;
#define SCHUNK 128              /* save table in chunks of 128 */

void *tabwind_savit(struct tabwind *x, struct symbol *name, void *w, long int n)        /* called from table and sigtab */
{
    long flags = ((x->w_wind->w_vis ? VIS : 0) | (x->w_signed ? SIGNED : 0)
                  | (x->w_wind->w_scratch ? SCRATCH : 0) | (x->
                                                            w_embed ? EMBED
                                                            : 0)) +
        100L * x->w_yshift + 10000L * x->w_xshift;
    if (x->w_sym)
        vfun(w, binbuf_insert, NIL, "ssllllllls", &s__New, name,
             x->w_size, (long) (x->w_wind->w_x1), (long) (x->w_wind->w_y1),
             (long) (x->w_wind->w_x2), (long) (x->w_wind->w_y2), flags,
             x->w_range, x->w_sym);
    else
        vfun(w, binbuf_insert, NIL, "sslllllll", &s__New, name,
             x->w_size, (long) (x->w_wind->w_x1), (long) (x->w_wind->w_y1),
             (long) (x->w_wind->w_x2), (long) (x->w_wind->w_y2), flags,
             x->w_range);
    if (x->w_embed && x->w_saveno != max_serialno)
    {
        long index, i, thismax;
        struct atom mess[SCHUNK + 4];
        for (index = 0; index < x->w_size; index += SCHUNK)
        {
            int num = (index + SCHUNK <= x->w_size ?
                       SCHUNK : x->w_size - index);
            SETSYM(mess, &s__Tab);
            SETSYM(mess + 1, &s_set);
            SETLONG(mess + 2, index);
            for (i = 0; i < num; i++)
                SETLONG(mess + (3 + i), (*x->w_tab)[index + i]);
            SETSEMI(mess + (3 + num));
            binbuf_insq(w, num + 4, mess, 1);
        }
        x->w_wind->w_dirty = 0; /* mark table clean after saving it */
    }
    x->w_saveno = max_serialno; /* mark table as saved, thank you. */
}

void *tabwind_wsize(struct tabwind *x, int a, int b)
{
    tabwind_chkscroll(x);
    /* ValidRect(&thePort->portRect); */
}

/* call this anytime size, display size, x or y shift are changed */
void *tabwind_chkscroll(struct tabwind *x)
{
    /* int width = x->w_wind->w_x2 - x->w_wind->w_x1;
       int height = x->w_wind->w_y2 - x->w_wind->w_y1; */
    long dpoints = ((x->w_size >> x->w_xshift) << x->w_xzoom);
    long drange = ((x->w_range >> x->w_yshift) << x->w_yzoom);
    if (x->w_signed)
        drange <<= 1;
    x->w_xdisp = dpoints;
    x->w_ydisp = drange;
    x->w_wind->w_xoffset = 0;
    x->w_wind->w_yoffset = 0;
    /* if (x->w_wind->w_xoffset > dpoints - x->w_xdisp)
       x->w_wind->w_xoffset = dpoints - x->w_xdisp;
       if (x->w_wind->w_yoffset > drange - x->w_ydisp)
       x->w_wind->w_yoffset = drange - x->w_ydisp; */
    /* wind_setsmax(x->w_wind, (int)(dpoints + 1 - x->w_xdisp),
       (int)(drange + 1 - x->w_ydisp));
       wind_setsval(x->w_wind,
       x->w_wind->w_xoffset, x->w_wind->w_yoffset); */
    wind_setscrollrect(x->w_wind, 0, 0, (int) (dpoints + 30),
                       (int) (drange + 30));
    /* post("range %ld disp %d", drange, x->w_ydisp); */
    x->w_yoff = (((long) (x->w_signed ?
                          ((x->w_ydisp - 1) >> 1) : x->w_ydisp - 1))
                 << x->w_yshift) >> x->w_yzoom;
    if (x->w_wind->w_vis)
    {
        /* zapgrowicon(x->w_wind); */
        tabwind_erase(x);
    }
}

/*
  #define OKITEM 1
  #define CANCELITEM 2
  #define TITLEITEM 3
  #define SIZEITEM 4
  #define RANGEITEM 5
  #define SIGNEDITEM 6
  #define SCRATCHITEM 7
  
  void *tabwind_dialog(x)
  struct tabwind *x;
  {
  DialogRecord dog;
  short hit = 2001, shit;
  char stouched = 0, rtouched = 0;
  int scheck = x->w_signed, scratch = x->w_wind->w_scratch;
  char buf[50];
  Rect r;
  GetNewDialog(1001, &dog, -1L);
  if (x->w_sym) settextitem(&dog, TITLEITEM, x->w_sym->s_name);
  sprintf(buf, "%ld", x->w_size);
  settextitem(&dog, SIZEITEM, buf);
  sprintf(buf, "%ld", x->w_range);
  settextitem(&dog, RANGEITEM, buf);
  setctlval(&dog, SIGNEDITEM, scheck);
  setctlval(&dog, SCRATCHITEM, scratch);
  SelIText(&dog, SIZEITEM, (short)0, (short)32767);
  while (hit != OKITEM)
  {
  ModalDialog(NIL, &hit);
  switch (hit)
  {
  case SIZEITEM: stouched = 1; break;
  case RANGEITEM: rtouched = 1; break;
  case SIGNEDITEM:
  scheck = !scheck;
  setctlval(&dog, SIGNEDITEM, scheck);
  break;
  case SCRATCHITEM:
  scratch = !scratch;
  setctlval(&dog, SCRATCHITEM, scratch);
  break;
  case CANCELITEM: goto closeit;
  }
  }
  if (stouched)
  {
  long nsize;
  gettextitem(&dog, SIZEITEM, buf);
  if (sscanf(buf, "%ld", &nsize) < 1)
  {
  ouch("bad size: ", buf);
  goto closeit;
  }
  if (nsize < 1) nsize = 1;
  tabwind_size(x, nsize);
  }
  if (rtouched)
  {
  long nx;
  gettextitem(&dog, RANGEITEM, buf);
  if (sscanf(buf, "%ld", &nx) < 1)
  {
  ouch("bad range: ", buf);
  goto closeit;
  }
  if (nx < 2) nx = 2;
  x->w_range = nx;
  tabwind_cliprange(x);
  }
  x->w_signed = scheck;
  x->w_wind->w_scratch = scratch;
  tabwind_chkscroll(x);
  closeit:
  CloseDialog(&dog); 
  }
  */

void *tabwind_chkmenu(struct tabwind *x, struct menuinfo *m)
{
    if (x->w_selmode)
    {
        /* m->i_paste = (TEScrpLength != 0); */
        m->i_cut = x->w_selend > x->w_selstart;
    }
    m->i_save = m->i_range = 1;
}

void *tabwind_clr(struct tabwind *x)
{
    long *lp1, *lp2, *lp3, newsize;
    if (x->w_selend > x->w_size)
        x->w_selend = x->w_size;
    if (x->w_selstart > x->w_selend)
        x->w_selstart = x->w_selend;
    lp1 = (*x->w_tab) + x->w_selstart;
    lp2 = (*x->w_tab) + x->w_selend;
    lp3 = (*x->w_tab) + x->w_size;
    while (lp2 < lp3)
        *(lp1++) = *(lp2++);
    newsize = x->w_size - (x->w_selend - x->w_selstart);
    if (newsize < 1)
    {
        newsize = 1;
        (*x->w_tab)[0] = 0;
    }
    tabwind_size(x, newsize);
    x->w_wind->w_dirty = 1;
}

long atol(char *s);

/* void *tabwind_paste(x)
   struct tabwind *x;
   {
   long n = 0, n2, asize, newsize, bytesleft = TEScrpLength, putwhere;
   ELEM **newh, *lp1, *lp2;
   char *sc, buf[80], *ebuf = buf+79, *bp;
   newh = (ELEM **)t_newhandle((asize = 100) * sizeof(ELEM));
   if (!newh)
   {
   memouch();
   return;
   }
   sc = *TEScrpHandle;
   if (x->w_selend > x->w_size) x->w_selend = x->w_size;
   if (x->w_selend <= x->w_selstart) x->w_selend = 0;
   while (bytesleft)
   {
   bp = buf;
   while (bytesleft && isspace(*sc)) sc++, bytesleft--;
   if (!bytesleft) break;
   while (bytesleft && !isspace(*sc) && bp < ebuf)
   *(bp++) = *(sc++), bytesleft--;
   *bp = 0;
   if (n >= asize)
   {
   asize += 100;
   if (t_growhandle(newh, asize * sizeof(ELEM)))
   {
   DisposHandle(newh);
   memouch();
   return;
   }
   sc = *TEScrpHandle + (TEScrpLength - bytesleft);
   }
   (*newh)[n] = atol(buf);
   n++;
   }
   newsize = x->w_size + n;
   if (t_growhandle(x->w_tab, newsize * sizeof(ELEM)))
   {
   DisposHandle(newh);
   memouch();
   return;
   }
   if (x->w_selend > x->w_selstart) putwhere = x->w_selstart;
   else putwhere = 0;
   for (lp1 = *(x->w_tab) + newsize, lp2 = lp1 - n,
   n2 = x->w_size - putwhere; n2--;) *(--lp1) = *(--lp2);
   for (lp1 = *(x->w_tab) + putwhere, lp2 = *newh, n2 = n; n2--;)
   *(lp1++) = *(lp2++);
   DisposHandle(newh);
   x->w_size = newsize;
   x->w_selstart += n;
   x->w_selend += n;
   x->w_wind->w_dirty = 1;
   tabwind_chkscroll(x);
   } */

/* had had --> x->w_selstart = putwhere;
   x->w_selend = putwhere + n; */


#define MAXTEXT 32700
/* void *tabwind_copy(x)
   struct tabwind *x;
   {
   long n, asize = TEScrpLength, index = 0, tsize;
   if (x->w_selstart >= x->w_selend || x->w_selend > x->w_size) return;
   for (n = x->w_selstart; n < x->w_selend; n++)
   {
   char buf[80], *bp = buf;
   int len;
   sprintf(buf, "%ld%c", (*x->w_tab)[n], (n%10 == 9 ? '\r' : ' '));
   if ((len = strlen(buf)) + index + 1 > asize)
   {
   if (asize >= MAXTEXT - 20)
   {
   ouch("sorry: truncated to %ld items", n);
   break;
   }
   asize += 1000;
   if (asize > MAXTEXT) asize = MAXTEXT;
   t_growhandle(TEScrpHandle, asize);
   }
   strcpy(*TEScrpHandle + index, buf);
   index += len;
   }
   t_growhandle(TEScrpHandle, index);
   TEScrpLength = index;
   } */

/* void *tabwind_cut(x)
   struct tabwind *x;
   {
   tabwind_copy(x);
   tabwind_clr(x);
   } */
