#ifndef lint
static char rcsid[] =
    "$Revision: 1.9 $ IRCAM $Date: 1994/02/07 06:45:59 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.9 $ IRCAM $Date: 1994/02/07 06:45:59 $
 *
 * FTS by Miller Puckette
 *
 */


typedef void *t_wind;

#include "m_mess.h"
#include "p_table.h"
#include "t_mac.h"              /* needed by patcher.h */
#include "p_patcher.h"          /* for CP_PUT, CP_UNPUT defs */
#include "h_cp.h"

void *outlet(void *op, struct symbol *s2);
extern struct symbol s_table,
    s_inv, s_save, s_vtable,
    s_, s_dblclick, s_size, s_float, s_in1, s_tabwind,
    s_set, s__t, s_quantile, s_sum, s_clear, s_const, s__New, s__Tab,
    s_setcp, s__set;
extern int memislow;

typedef struct table {
    struct object t_ob;
    long t_yval;
    struct tabwind *t_tabwind;
    void *t_outlet;
    ELEM **t_tab;
    char t_in1;
} t_table;

void table_quantile(struct table *x, long int n);

void table_bang(struct table *x)
{
    long r = rand() & 0x7fff;
    table_quantile(x, r);
}

void table_int(struct table *x, long int n)
{
    if (n < 0)
        n = 0;
    else if (n >= x->t_tabwind->w_size)
        n = x->t_tabwind->w_size - 1;
    if (x->t_in1)
    {
        tabwind_fixfix(x->t_tabwind, n, x->t_yval);
        x->t_in1 = 0;
    }
    else
        outlet_int(x->t_outlet, (long) ((*x->t_tab)[n]));
}

void table_float(struct table *x, double f)
{
    long n = f;
    if (n < 0)
        n = 0;
    else if (n >= x->t_tabwind->w_size)
        n = x->t_tabwind->w_size - 1;
    if (x->t_yval >= 0)
    {
        tabwind_fixfix(x->t_tabwind, n, x->t_yval);
        x->t_yval = -1;
    }
    else
        outlet_int(x->t_outlet, (long) ((*x->t_tab)[n]));
}

void table_in1(struct table *x, long int n)
{
    x->t_in1 = 1;
    x->t_yval = n;
}

#ifdef OBSOLETE
void *table_fixfix(x, n1, n2)
struct table *x;
long n1, n2;
{
    tabwind_fixfix(x->t_tabwind, n1, n2);
}
#endif

void table_set(struct table *x, struct symbol *s, int ac, struct atom *av)
{
    tabwind_set(x->t_tabwind, s, ac, av);
}

#ifndef GRAPHICS
void table__set(x, s, ac, av)   /* version without graphics update */
struct table *x;
struct symbol *s;
int ac;
struct atom *av;
{
    tabwind__set(x->t_tabwind, s, ac, av);
}
#endif

void table_inv(struct table *x, long int n)
{
    long xval = 0;
    ELEM *tp = *x->t_tab;
    ELEM *xp = tp + xval, *ep = tp + x->t_tabwind->w_size;
    if (xp >= ep)
        xp = tp, xval = 0;
    if (xval && n < *xp)
        xp = tp, xval = 0;
    while (xp < ep && *xp <= n)
        xp++, xval++;
    if (xval)
        xval--;
    outlet_int(x->t_outlet, xval);
}

void table_quantile(struct table *x, long int n)
{
    long v = 0, index;
    long i;
    ELEM *ip;
    long size = x->t_tabwind->w_size;
    for (i = size, ip = (*x->t_tab); i--;)
        v += *(ip++);
    if (!v)
    {
        outlet_int(x->t_outlet, 0L);
        return;
    }
    index = ((n * v) >> 15) + 1;
    for (i = 0, ip = (*x->t_tab); i < size; i++)
    {
        index -= *(ip++);
        if (index <= 0)
            break;
    }
    if (i >= size)
        i--;
    outlet_int(x->t_outlet, i);
}

void table_clear(struct table *x)
{
    tabwind_clear(x->t_tabwind);
}

void table_const(struct table *x, long int n)
{
    tabwind_const(x->t_tabwind, n);
}

void table_size(struct table *x, long int n)
{
    tabwind_size(x->t_tabwind, n);
}

void table_read(struct table *x, t_symbol *f)
{
    post("unimplemented\n");
    /* tabwind_read(x->t_tabwind, f); */
}

void table_sum(struct table *x, long int n)
{
    long i;
    ELEM *ip, v = 0;
    long size = x->t_tabwind->w_size;
    for (i = size, ip = (*x->t_tab); i--;)
        v += *(ip++);
    outlet_int(x->t_outlet, (long) v);
}

#ifdef GRAPHICS
void table_dblclick(struct table *x)
{
    wind_vis(x->t_tabwind->w_wind);
}

void table_save(struct table *x, void *w, long int n)
{
    tabwind_savit(x->t_tabwind, &s_vtable, w, n);
}

#define CPCHUNK 16

void *cpr_new(void *w, void (*fn)( /* ??? */ )), tabwind_cpr(t_tabwind * x,
                                                             t_updbuf *
                                                             buf,
                                                             int size);

void table_setcp(t_table *x, int n, int flags)
{
    if (flags & CP_UNPUT)
    {
        if (x->t_tabwind->w_cpno >= 0)
        {
            cpr_free(x->t_tabwind->w_cpr);
            x->t_tabwind->w_cpr = 0;
            h_cutoff(x->t_tabwind);
            x->t_tabwind->w_cpno = -1;
        }
    }
    if (flags & CP_PUT)
    {
        if (x->t_tabwind->w_cpno >= 0 && x->t_tabwind->w_cpno != n)
            error("%s: tables on more than one CP not yet allowed",
                  x->t_tabwind->w_sym->s_name);
        else if (x->t_tabwind->w_cpno < 0)
        {
            int i;
            t_atom at[CPCHUNK + 4], *ap;
            x->t_tabwind->w_cpno = n;
            h_vnew(x->t_tabwind, n, &s_tabwind, "o", x);
            if (x->t_tabwind->w_size != DEFSIZE)
                h_vmess(x, &s_size, "l", x->t_tabwind->w_size);
            for (i = 0; i < x->t_tabwind->w_size;)
            {
                int ct = (x->t_tabwind->w_size - i < CPCHUNK ?
                          x->t_tabwind->w_size - i : CPCHUNK);
                ELEM *tp;
                int nonzero = 0, ct2;
                tp = (*x->t_tabwind->w_tab) + i;
                SETLONG(at, (long) i);
                for (ct2 = 0, ap = at + 1; ct2 < ct; ct2++, ap++)
                {
                    ELEM n = *tp++;
                    SETLONG(ap, n);
                    if (n)
                        nonzero = 1;
                }
                if (nonzero)
                    h_mess(x, &s__set, ct + 1, at);
                i += ct;
            }
            x->t_tabwind->w_cpr = cpr_new(x->t_tabwind, tabwind_cpr);
        }
    }
}
#endif                          /* GRAPHICS */


void table_ff(struct table *x)
{
    untabwind(x->t_tabwind);
}

void *table_class;
void table_setup()
{
    c_new(&s_table, table_ff, sizeof(t_table), 0, 0);
    c_addint(table_int);
    c_addbang(table_bang);
    c_addfloat(table_float);
    c_addinx(table_in1, 1);
    c_addmess(table_inv, &s_inv, A_LONG, 0);
    c_addmess(table_quantile, &s_quantile, A_LONG, 0);
    c_addmess(table_sum, &s_sum, 0);
    c_addmess(table_set, &s_set, A_GIMME, 0);
    c_addmess(table_size, &s_size, A_LONG, 0);
    c_addmess(table_read, gensym("read"), A_SYM, 0);
    c_addmess(table_clear, &s_clear, 0);
    c_addmess(table_const, &s_const, A_LONG, 0);
#ifdef GRAPHICS
    c_addmess(table_save, &s_save, A_CANT, 0);
    c_addmess(table_dblclick, &s_dblclick, 0);
    c_addmess(table_setcp, &s_setcp, A_CANT, 0);
#else
    c_addmess(table__set, gensym("_set"), A_GIMME, 0);
#endif
}

long get1long(struct atom *ap)
{
    return (ap->a_type == A_LONG ? ap->a_w.w_long : 0);
}

#ifdef GRAPHICS
void *argtotab(int argc, struct atom *argv)
{
    struct symbol *s;
    int x1 = 0, x2 = 0, y1 = 32000, y2 = 32000, xshift = 0, yshift =
        0, flags = 0;
    long flagarg, size, range = 128;
    void *answer;
    if (argc > 5)
    {
        size = get1long(argv++);
        x1 = get1long(argv++);
        y1 = get1long(argv++);
        x2 = get1long(argv++);
        y2 = get1long(argv++);
        flagarg = get1long(argv++);
        argc -= 6;
        if (size < 1)
            size = 1;
        if (x1 + 50 > x2 || y1 + 50 > y2 || x2 < 0 || y2 < 0)
            x1 = y1 = 0, x2 = y2 = 32000;
        flags = flagarg % 100;
        flagarg /= 100;
        yshift = flagarg % 100;
        if (yshift >= 16)
            yshift -= 32;
        flagarg /= 100;
        xshift = flagarg % 100;
        if (xshift >= 16)
            xshift -= 32;
        if (xshift > MAXSHIFT)
            xshift = MAXSHIFT;
        else if (xshift < MINSHIFT)
            xshift = MINSHIFT;
        if (yshift > MAXSHIFT)
            yshift = MAXSHIFT;
        else if (yshift < MINSHIFT)
            yshift = MINSHIFT;
    }
    if (argc && argv->a_type == A_LONG)
    {
        if ((range = argv->a_w.w_long) < 1)
            range = 1;
        argc--;
        argv++;
    }
    if (argc && argv->a_type == A_SYM)
        s = argv->a_w.w_sym;
    else
        s = 0;
    answer = gettabwind(s, size, range, x1, y1, x2, y2,
                        xshift, yshift, flags);
    s__Tab.s_thing = answer;
    return (answer);
}

void *vtable_new(struct symbol *sel, int argc, struct atom *argv)
{
    struct table *x;
    if (memislow)
        return (0);
    x = (struct table *) obj_new((t_messlist **) & table_class,
                                 table_setup);
    x->t_outlet = outlet(x, NIL);
    inlet(x, x, &s_int, &s_in1);
    x->t_tabwind = (struct tabwind *) argtotab(argc, argv);
    x->t_tab = x->t_tabwind->w_tab;
    x->t_yval = 0;
    x->t_in1 = 0;
    return (x);
}
#endif                          /* GRAPHICS */

void *table_new(struct symbol *s)
{
    struct table *x =
        (struct table *) obj_new((t_messlist **) & table_class,
                                 table_setup);
    if (s == &s_)
        s = 0;
#ifdef GRAPHICS
    x->t_tabwind = gettabwind(s, (long) DEFSIZE, (long) DEFRANGE,
                              50, 50, 50 + STDWIDTH, 50 + STDHEIGHT, 0, 0,
                              EMBED | VIS);
#else
    x->t_tabwind = gettabwind(s, (long) DEFSIZE, (long) DEFRANGE,
                              0, 0, 0, 0, 0, 0, 0);
#endif
    x->t_tab = x->t_tabwind->w_tab;
    x->t_yval = 0;
    x->t_in1 = 0;
    x->t_outlet = outlet(x, NIL);
    inlet(x, x, &s_int, &s_in1);
    return (x);
}

#ifdef GRAPHICS
void *menutable()
{
    struct tabwind *x = stabwind(NIL);
    wind_vis(x->w_wind);
}
#endif

#ifndef GRAPHICS
void *tabwind_fakenew(t)
struct table *t;
{
    return (t->t_tabwind);
}
#endif
