#ifndef lint
static char rcsid[] =
    "$Revision: 2.4 $ IRCAM $Date: 1993/11/17 18:32:40 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.4 $ IRCAM $Date: 1993/11/17 18:32:40 $
 *
 * FTS by Miller Puckette
 *
 */


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

extern char *progname;

void *shmaddr;                  /* Shared memory address */
int shmid;                      /* Shared memory ID */

/*
 * mmap_open()
 *
 * Return a pointer to a chunk of shared memory.
 *
 * In the server this function is called with a non-zero `shm_size'
 * which results in a new shared memory ID being created (via IPC_PRIVATE)
 * The global `shmaddr' points to the memory allocated.
 *
 * Later the client is passed this ID via an exec() parameter.  When the
 * client calls mmap_open() with a zero `shm_size' the `shmaddr' pointer
 * is made point to the shared memory identified by the ID.
 */

void *mmap_open(int shm_size)
{
    if (shm_size)
    {
        /* Being called from server: get an identifier */
        if ((shmid = shmget(IPC_PRIVATE, shm_size, SHM_R | SHM_W)) == -1)
        {
            fprintf(stderr, "shared memory error in %s\n", progname);
            perror("unable to shmget() in mmap_open()");
            return 0;
        }
    }

    /* Point `shmaddr' at the memory */
    if ((shmaddr = shmat(shmid, 0, 0)) == (void *) -1)
    {
        fprintf(stderr, "shared memory error in %s\n", progname);
        perror("unable to shmat() in mmap_open()");
        return 0;
    }

    return shmaddr;
}

/*
 * mmap_close()
 *
 * Detach and free the shared memory segment indexed by `shmid'
 */

void mmap_close()
{
    /* detach the shared memory */
    if (shmdt(shmaddr) < 0)
    {
        perror("unable to shmdt() in mmap_close()");
    }

    /* remove the ID */
    if (shmctl(shmid, IPC_RMID, 0) < 0)
    {
        perror("error removing shared memory id");
    }
}
