
/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.8 $ IRCAM $Date: 1993/11/17 14:47:41 $
 *
 * FTS by Miller Puckette
 * the window object.
 * Part of the structure is public for dirty purposes.
 *
 */


#ifndef   _OBJC_
#ifdef XMAX
#include <X11/Intrinsic.h>
#else
#endif
#endif
typedef void *t_dummy;

/* LATER, hide this struct in X_wind.c */
typedef struct windpub {
    t_object w_ob;
    int w_x1;                   /* location of window */
    int w_x2;
    int w_y1;
    int w_y2;
    int w_xoffset;              /* scroll offsets */
    int w_yoffset;
    int w_scrollgrain;          /* scroll grain in pixels, 0 if none */
    int w_refcount;             /* reference count */
    int w_topmargin;            /* margins in pixels outside scrollview */
    int w_bottommargin;
    int w_rightmargin;
    int w_leftmargin;
    t_dummy w_dummy1;           /* NextStep window */
    t_dummy w_dummy2;           /* scroller doc view (content view if none) */
    t_dummy w_dummy3;           /* content view */
    void *w_assoc;
    t_symbol *w_vol;
    int w_xmin;                 /* scroll region size */
    int w_xmax;
    int w_ymin;
    int w_ymax;
    char w_name[256];
    char w_vis;                 /* visible */
    char w_scroll;              /* has a scroll region */
    char w_titled;              /* has a title */
    char w_grow;                /* has a grow region */
    char w_close;               /* has a close region */
    char w_scrollx;             /* has an x scroll region */
    char w_scrolly;             /* has a y scroll region */
    char w_dirty;               /* dirty flag (can save) */
    char w_scratch;             /* scratch flag -- don't complain on close */
    char w_bin;                 /* binary save */
    char w_font;                /* text font */
    char w_fsize;               /* font size */
    char w_fontindex;           /* font index (0 = 9 points, 7 = 24 points) */
    char w_willclose;           /* NeXT will close the (NeXT) window object */
#if NOT_USED
    void *w_qelem;              /* queue element for forcing first update */
#endif
    char w_keeptitle;
    char w_printing;
    /* forward & backward pointers */
    struct windpub *w_previous;
    struct windpub *w_next;
    Widget w_palette;           /* not yet used */
    Widget w_maincanvas;        /* drawing canvas */
    Widget w_margincanvas;      /* canvas outside scroll rgn */
    Widget w_canvasshell;       /* "canvas shell" */
    Widget w_closedialog;
    struct windpub *w_viacontent;       /* see below */
    struct windpub *w_viamargin;        /* ditto */
} t_wind;

/*
 * The w_viacontent and w_viamargin slots contain pointers back to the
 * structure.  They in turn are pointed to by the current "grafport"
 * which describes a context.  If grafport points to "viacontent"
 * drawing is done into the content portion of the window, otherwise
 * on the margin.  Meanwhile we can still chase the grafport back to
 * the wind structure if needed.
 */

/* flags for window object */
#define WVIS    01              /* visible */
#define WGROW   02              /* grow region */
#define WSCROLLX  04            /* x scrollbar */
#define WSCROLLY  010           /* y scrollbar */
#define WCLOSE    020           /* close region */
#define WKEEPT    040           /* keep window title from saveas */
#define WSGROW    0100          /* silent grow region */
#define WCOLOR    0200          /* color window */

void *wind();
#define W_WIDTH(w) (w->w_x2 - w->w_x1)
#define W_HEIGHT(w) (w->w_y2 - w->w_y1)
#define SBARWIDTH 0

extern int w_sizes[];

/* Events reported by window object */
extern t_symbol s_click, s_invis, s_key, s_help, s_activate, s_update,
    s_scroll, s_move, s_wsize, s_dialog, s_close, s_saveto,
    s_cut, s_copy, s_paste,
    s_clr, s_eval, s_idle, s_chkmenu, s_dup, s_undo, s_font,
    s_edit, s_lineup, s_fixwidth, s_front;

void wind_filename(t_wind * x, char *name, t_symbol * vol, int bin);
void wind_settitle(t_wind * x, char *name);
void wind_unfilename(t_wind * x);
void wind_invis(t_wind * x);
void wind_vis(t_wind * x);
GrafPort *wind_setport(t_wind * x);
GrafPort *wind_marginport(t_wind * x);
void wind_issubpatch(t_wind * x);
void wind_setsval(t_wind * x, int a, int b);
void wind_setsmax(t_wind * x, int a, int b);
int wind_defaultscroll(t_wind * x, Point pt);
void wind_free();
t_wind *wind_new(void *assoc, int a, int b, int c, int d, int flags);
void t_resetport(GrafPort * gp);
void wind_drag(t_wind * x, t_method fun, void *arg);
void wind_nodrag();
t_method wind_getdrag(void);

#define BACKSPACE 127
#define LEFTARROW 172
#define RIGHTARROW 174
#define UPARROW 173
#define DOWNARROW 175
