#ifndef lint
static char rcsid[] =
    "$Revision: 1.7 $ IRCAM $Date: 1993/11/19 10:51:45 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.7 $ IRCAM $Date: 1993/11/19 10:51:45 $
 *
 * FTS by Miller Puckette
 *
 */


/* This file defines the "self" MAX object which acts as the MAX/Animal
   interface.  The object receives messages to set and get values of the
   current Animal instance, if any. */

#include "m_mess.h"
#include "l_animal.h"
#include "l_animal2.h"
extern void *animal_messlist, *a_set_class;

extern t_symbol s_animal, s_set, s_get, s_send, s_any_animal, s_symbol,
    s_my, s_;

typedef struct myentry {
    t_symbol *e_sym;
    int e_onset;
} t_myentry;

typedef struct my {
    t_object m_ob;
    int m_ac;
    t_myentry *m_entry;
    int m_serial;
} t_my;


#ifndef GRAPHICS
extern int animal_serial;

t_field *my_search(t_my *x, void ***vpp, t_animal **ap)
{
    int depth = x->m_ac;
    t_myentry *entry = x->m_entry;
    int count;
    t_field *fp;
    void **vp;
    t_animal *a = animal_self;
    if (!a)
    {
        error("my_search: no current animal");
        return (0);
    }
    if (x->m_serial == animal_serial)
    {
        for (; depth--; entry++)
        {
            vp = a->a_bag->x_x + entry->e_onset;
            if (depth)
            {
                if (!(a = (t_animal *) (*vp)))
                {
                    error("my_search: %s: empty", entry->e_sym->s_name);
                    return (0);
                }
            }
        }
        /* post("get onset %d, addr %x", entry->e_onset, entry); */
        fp = a->a_class->c_fields + (entry - 1)->e_onset;
    }
    else
    {
        for (; depth--; entry++)
        {
            t_symbol *s = entry->e_sym;
            fp = a->a_class->c_fields;
            for (count = a->a_class->c_nitems, vp = a->a_bag->x_x;
                 count--; fp++, vp++)
            {
                /* post("compare %s with %s",
                   s->s_name, fp->f_name->s_name); */
                if (fp->f_name == s)
                {
                    entry->e_onset = a->a_class->c_nitems - count - 1;
                    /* post("put onset %d, addr %x", entry->e_onset, entry); */
                    if (depth)
                    {
                        if (fp->f_type != &s_any_animal)
                        {
                            error("my_search: %s: not an animal",
                                  s->s_name);
                            return (0);
                        }
                        else if (!(a = (t_animal *) (*vp)))
                        {
                            error("my_search: %s: empty", s->s_name);
                            return (0);
                        }
                    }
                    goto foundit;
                }
            }
            error("my_search: %s: field not found", s->s_name);
            return (0);
          foundit:;
        }
        x->m_serial = animal_serial;
    }
    *ap = a;
    *vpp = vp;
    return (fp);
}
#endif                          /* GRAPHICS */

void my_bang(t_my *x)
{
#ifndef GRAPHICS
    t_animal *a;
    void **vp;
    t_field *fp = my_search(x, &vp, &a);
    if (fp)
    {
        /* post("animal %x", a);
           post("check %s", ob_name(a));
           post("class %x", a->a_class);
           post("field %x, first field %x", fp, a->a_class->c_fields);
           post("index %d", fp - a->a_class->c_fields);
           post("field name %s, type %d", fp->f_name->s_name, fp->f_type); */
        if (fp->f_type == &s_float)
            outlet_float(x->m_ob.o_outlet, *(float *) vp);
        else if (fp->f_type == &s_int)
            outlet_int(x->m_ob.o_outlet, *(long *) vp);
        else if (fp->f_type == &s_symbol)
            vfun(x->m_ob.o_outlet, outlet_list, NIL,
                 "s", *(t_symbol **) vp);
        else
            error("my_bang: sorry, numbers and symbols only");
    }
#endif
}

void my_int(t_my *x, long int n)
{
#ifndef GRAPHICS
    t_animal *a;
    void **vp;
    t_field *fp = my_search(x, &vp, &a);
    if (fp)
    {
        if (fp->f_type == &s_float)
            *(float *) vp = n;
        else if (fp->f_type == &s_int)
            *(long *) vp = n;
        else
            error("my_int: field is not a number");
        animal_update(a);
    }
#endif
}

void my_float(t_my *x, double f)
{
#ifndef GRAPHICS
    t_animal *a;
    void **vp;
    t_field *fp = my_search(x, &vp, &a);
    if (fp)
    {
        if (fp->f_type == &s_float)
            *(float *) vp = f;
        else if (fp->f_type == &s_int)
            *(long *) vp = f;
        else
            error("my_float: field is not a number");
        animal_update(a);
    }
#endif
}

void my_anything(t_my *x, t_symbol *s, int ac, t_atom *av)
{
#ifndef GRAPHICS
    t_animal *a;
    void **vp;
    t_field *fp = my_search(x, &vp, &a);
    if (fp)
    {
        if (fp->f_type == &s_any_animal)
            typedmess(*(void **) vp, s, ac, av);
        else
            error("my_send: field is not a number");
    }
#endif
}

void my_ff(t_my *x)
{
    if (x->m_ac)
        freebytes(x->m_entry, x->m_ac * sizeof(*x->m_entry));
}

void *my_class;
void my_setup()
{
    c_new(&s_my, my_ff, sizeof(t_my), 0, 0);
    c_addbang(my_bang);
    c_addint(my_int);
    c_addfloat(my_float);
    c_addmess(my_anything, &s_send, A_GIMME, 0);
}

void *my_new(t_symbol *s, int ac, t_atom *av)
{
    t_my *x = (t_my *) obj_new((t_messlist **) & my_class, my_setup);
    t_myentry *entry;
    x->m_ac = ac;
    if (x->m_ac)
    {
        x->m_entry = (t_myentry *) getbytes(x->m_ac * sizeof(*x->m_entry));
        for (entry = x->m_entry; ac--; av++, entry++)
        {
            if (av->a_type == A_SYM)
                entry->e_sym = av->a_w.w_sym;
            else
            {
                error("my: symbols only as arguments");
                entry->e_sym = &s_;
            }
        }
    }
    else
        x->m_entry = 0;
    outlet(x, NIL);
    x->m_serial = -1;
    return (x);
}

/*********** old version called "animal" **********/
typedef struct self {
    t_object s_ob;
    t_animal *s_animal;
    t_symbol *s_sym;
} t_self;

#ifndef GRAPHICS
void self_get(t_self *x, t_symbol *s)
{
    if (animal_self)
    {
        t_field *fp = animal_self->a_class->c_fields;
        int count = animal_self->a_class->c_nitems;
        void **vp = animal_self->a_bag->x_x;
        for (; count--; fp++, vp++)
        {
            if (fp->f_name == s)
            {
                if (fp->f_type == &s_float)
                    outlet_float(x->s_ob.o_outlet, *(float *) vp);
                else if (fp->f_type == &s_int)
                    outlet_int(x->s_ob.o_outlet, *(long *) vp);
                else if (fp->f_type == &s_symbol)
                    vfun(x->s_ob.o_outlet, outlet_list, NIL,
                         "s", *(t_symbol **) vp);
                else
                    error("self_get: %s: sorry, numbers and symbols only",
                          s->s_name);
                return;
            }
        }
        error("self_get: %s: no field named %s",
              animal_self->a_class->c_name->s_name, s->s_name);
    }
    else
        error("self_get: no current animal");
}

void self_set(t_self *x, t_symbol *s, int ac, t_atom *av)
{
    if (!ac)
    {
        x->s_animal = animal_self;
        return;
    }
    if (animal_self)
    {
        t_field *fp = animal_self->a_class->c_fields;
        int count = animal_self->a_class->c_nitems;
        void **vp = animal_self->a_bag->x_x;
        t_symbol *fs;
        if (ac < 2 || av[0].a_type != A_SYM)
        {
            error("self_set: bad arguments");
            return;
        }
        fs = av[0].a_w.w_sym;
        for (; count--; fp++, vp++)
        {
            if (fp->f_name == fs)
            {
                if (fp->f_type == &s_float)
                {
                    if (av[1].a_type == A_FLOAT)
                        *(float *) vp = av[1].a_w.w_float;
                    else if (av[1].a_type == A_LONG)
                        *(float *) vp = av[1].a_w.w_long;
                    else
                        error("self_set: bad arguments");
                }
                else if (fp->f_type == &s_int)
                {
                    if (av[1].a_type == A_FLOAT)
                        *(long *) vp = av[1].a_w.w_float;
                    else if (av[1].a_type == A_LONG)
                        *(long *) vp = av[1].a_w.w_long;
                    else
                        error("self_set: bad arguments");
                }
                else if (fp->f_type == &s_symbol)
                {
                    if (av[1].a_type == A_SYM)
                        *(t_symbol **) vp = av[1].a_w.w_sym;
                }
                else
                    error("self_set: %s: sorry, numbers only", fs->s_name);
                animal_update(animal_self);
                return;
            }
        }
        error("self_set: %s: no field named %s",
              animal_self->a_class->c_name->s_name, fs->s_name);
    }
    else
        error("self_set: no current animal");
}

void self_send(t_self *x, t_symbol *s, int ac, t_atom *av)
{
    if (animal_self)
        animal_send(animal_self, s, ac, av);
}
#endif

void self_bang(t_self *x)
{
#ifndef GRAPHICS
    if (x->s_sym)
    {
        t_animal *animal_was = animal_self;
        void *a = x->s_sym->s_thing;
        animal_self = a;
        if (!a)
            error("%s: undefined", x->s_sym->s_name);
        else if (ob_messlist(a) == (t_messlist *) animal_messlist)
            outlet_bang(x->s_ob.o_outlet);
        else if (ob_messlist(a) == (t_messlist *) a_set_class)
            a_set_uzi(a, x->s_ob.o_outlet);
        else
            error("%s: not an animal", x->s_sym->s_name);
        animal_self = animal_was;
    }
    else
    {
        if (!x->s_animal)
            error("animal_bang: unset");
        else
        {
            t_animal *animal_was = animal_self;
            animal_self = x->s_animal;
            outlet_bang(x->s_ob.o_outlet);
            animal_self = animal_was;
        }
    }
#endif
}

void *self_class;
void self_setup()
{
    c_new(&s_animal, NIL, sizeof(t_self), 0, 0);
    c_addbang(self_bang);
#ifndef GRAPHICS
    c_addmess(self_set, &s_set, A_GIMME, 0);
    c_addmess(self_get, &s_get, A_SYM, 0);
    c_addmess(self_send, &s_send, A_GIMME, 0);
#endif
}

void *maxanimal_new(t_symbol *s)
{
    t_self *x =
        (t_self *) obj_new((t_messlist **) & self_class, self_setup);
    outlet(x, NIL);
    x->s_animal = 0;
    if (s != &s_)
        x->s_sym = s;
    else
        x->s_sym = 0;
    return (x);
}
