#ifndef lint
static char rcsid[] =
    "$Revision: 1.5 $ IRCAM $Date: 1993/11/17 14:48:19 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.5 $ IRCAM $Date: 1993/11/17 14:48:19 $
 *
 * FTS by Miller Puckette
 * x_exp2.c -- score following, playback, MIDI import/export
 *
 */


#include "m_mess.h"
#include "t_mac.h"
#include "t_wind.h"
#include "x_explode.h"

void explode_clear(t_explode * x), explode_next(t_explode * x),
explode_hangon(t_explode * x, t_evt * e);
extern t_symbol s_in1, s_in2, s_in3, s_in4;
extern void *explode_class;
long explode_nextserial;

int explode_sendevt(t_explode * x, t_evt * e, int noteon);

void *explode_stop(t_explode *x)
{
    t_hang *h;
    t_skip *s;
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        if (x->x_mode == MRECORD)
        {
            explode_markdirty(x);
        }
#ifdef GRAPHICS
        explode_remax(x);
        if (x->x_wind)
        {
            x->x_qflags[QERASEMODE] = 1;
            x->x_qflags[QDRAWPANEL] = 1;
            qelem_set(x->x_qelem);
        }
#endif
        x->x_mode = x->x_savemode;
        x->x_set1 = x->x_set2 = x->x_set3 = 0;
        for (h = x->x_hang; h; h = h->h_next)
            freebytes(h, sizeof(t_hang));
        for (s = x->x_skip; s; s = s->s_next)
            freebytes(s, sizeof(t_skip));
        x->x_skip = 0;
        x->x_hang = 0;
        x->x_nhang = 0;
        x->x_serial = ++explode_nextserial;
    }
    RestoreA4();
    return (x);
}

void explode_doappend(t_explode *x, long int time, long int pit,
                      long int vel, long int dur, long int chan)
{
    t_evt *e, *cur = x->x_current;
    if (x->x_evt && (!cur || cur->e_next))
        post("oops! lost my marbles");
    else
    {
        t_evt *e = (t_evt *) getbytes(sizeof(t_evt));
        if (time < 0)
            time = 0;
        e->e_time = x->x_rectime = time + x->x_rectime;
        e->e_pit = pit;
        e->e_vel = vel;
        e->e_dur = dur;
        e->e_chan = chan;
        e->e_next = 0;
        e->e_selected = 0;
        e->e_dirty = 1;
        if (cur)
            cur->e_next = e;
        else
            x->x_evt = e;
        x->x_current = e;
#ifdef GRAPHICS
        if (x->x_wind)
        {
            x->x_qflags[QREMAX] = 1;
            x->x_qflags[QDRAWEVTS] = 1;
            qelem_set(x->x_qelem);
        }
#endif
    }
}

void explode_dofollow(t_explode *x, long int n)
{
    long tquit = 0x7fffffff;
    int count = 0, gotem = 0, nskip = 0, i;
    t_skip newskip, *sp, *sp2;
    int eat = 0;
    t_evt *evtwas = x->x_current, *e;
    newskip.s_next = x->x_skip;
    /* check against skip array */
    for (sp = &newskip; sp2 = sp->s_next;)
    {
        int drop = 0;
        long pit = sp2->s_evt->e_pit;
        /* if a skip is too old (score time skipped
           >= FTIME) then kill it */
        if (x->x_matchscoretime - sp2->s_evt->e_time >= x->x_ftime)
            drop = 1;
        /* here's the payoff for the skip array:
           if the pitch is in the array eat it. */
        if (pit == n || (x->x_oct && (pit == n + 12 || pit == n - 12)))
        {
            if (!eat)           /* if you haven't already sent one */
            {
                /* send the match */
                explode_sendevt(x, sp2->s_evt, 1);
                /* and add to hang list */
                explode_hangon(x, sp2->s_evt);
            }
            drop = eat = 1;
        }
        if (drop)
        {
            sp->s_next = sp2->s_next;
            freebytes(sp2, sizeof(t_skip));
        }
        else
            sp = sp2;
    }
    x->x_skip = newskip.s_next;
    if (eat || !evtwas)
        return;
    /* search forward in list for matching note */
    for (;; x->x_current = e->e_next)
    {
        e = x->x_current;
        /* if at end of score no match */
        if (!e)
            goto restore;
        /* interested only in channel 1 */
        if (e->e_chan != 1)
            continue;
        if (gotem)              /* if already ate all the anytime-skippable notes */
        {
            /* and if, past that, you ate ftime into the future */
            if (e->e_time > tquit)
                /* then there is no match. */
                goto restore;
        }
        /* break if got the good note */
        if (e->e_pit == n || (x->x_oct &&
                              (e->e_pit == n + 12 || e->e_pit == n - 12)))
            break;
        /* if you got NFWD notes already */
        if (!gotem && ++count > x->x_nfwd)
        {
            gotem = 1;
            tquit = e->e_time + x->x_ftime;
        }
    }
    /* If you break out here, it means you just matched x->x_current. */
    x->x_matchscoretime = x->x_current->e_time;
    /* put note in note-hang list for matching note-off later */
    explode_hangon(x, x->x_current);
    /* put skipped notes into skip list; output any notes in
       other channels we
       have jumped get output */
    for (e = evtwas; e != x->x_current; e = e->e_next)
        if (e->e_chan == 1)
        {
            sp = (t_skip *) getbytes(sizeof(t_skip));
            sp->s_evt = e;
            sp->s_next = x->x_skip;
            x->x_skip = sp;
        }
        else
            explode_sendevt(x, e, 1);
    /* output the winning note */
    explode_sendevt(x, x->x_current, 1);
    /* and set pointer to the note after that */
    x->x_current = x->x_current->e_next;
    return;

  restore:                      /* if no match, restore state to what it was */
    x->x_current = evtwas;
}

void explode_hangon(t_explode *x, t_evt *e)
{
    if (x->x_nhang < MAXFOLLOWHANG)
    {

        t_hang *h = (t_hang *) getbytes(sizeof(t_hang));
        x->x_nhang++;
        h->h_next = x->x_hang;
        h->h_evt = e;
        x->x_hang = h;
    }
}

/* note-off handler for score following */
void explode_fnoteoff(t_explode *x, long int n)
{
    t_hang *h, *h2;
    t_hang stub;
    stub.h_next = x->x_hang;
    for (h = &stub; h2 = h->h_next; h = h2)
    {
        if (h2->h_evt->e_pit == n)
        {
            t_evt *e = h2->h_evt;
            explode_sendevt(x, e, 0L);
            h->h_next = h2->h_next;
            freebytes(h2, sizeof(t_hang));
            x->x_nhang--;
            break;
        }
    }
    x->x_hang = stub.h_next;
}

explode_sendevt(t_explode *x, t_evt *e, int noteon)     /* send out an event matched by follower */
{
    outlet_int(x->x_out5, e->e_chan);
    outlet_int(x->x_out4, e->e_dur);
    outlet_int(x->x_out3, noteon ? e->e_vel : 0L);
    outlet_int(x->x_out2, e->e_pit);
}

void explode_start(t_explode *x)
{
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        explode_stop(x);
        x->x_set1 = x->x_set2 = x->x_set3 = 0;
        if (x->x_current = x->x_evt)
        {
            outlet_int(x->x_ob.o_outlet, x->x_current->e_time);
            x->x_mode = MPLAY;
        }
        else
            post("no sequence");
    }
    RestoreA4();
}

void explode_record(t_explode *x)
{
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        explode_clear(x);
        x->x_mode = MRECORD;
        x->x_rectime = 0;
        x->x_serial = 0;
    }
    RestoreA4();
}

void explode_follow(t_explode *x, long int n)
{
    t_evt *e;

    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        explode_stop(x);
        x->x_matchscoretime = 0;
        for (e = x->x_evt; e && n--; e = e->e_next);
        if (x->x_current = e)
            x->x_mode = MFOLLOW;
        else
            post("follow: out of range");
    }
    RestoreA4();
}

void explode_at(t_explode *x, long int n1, long int n2, long int n3)
{
    t_evt *e;
    for (e = x->x_evt; e; e = e->e_next)
        if (e->e_pit == n1 && e->e_vel == n2 && e->e_chan == n3)
            break;
    if (!(x->x_current = e))
        post("explode: note not found");
}

void explode_followat(t_explode *x, long int n1, long int n2, long int n3)
{
    t_evt *e;
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        explode_stop(x);
        x->x_matchscoretime = 0;
        explode_at(x, n1, n2, n3);
        if (e = x->x_current)
            x->x_current = e->e_next;
        if (x->x_current)
            x->x_mode = MFOLLOW;
    }
    RestoreA4();
}

void explode_startat(t_explode *x, long int n1, long int n2, long int n3)
{
    t_evt *e;
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        explode_stop(x);
        explode_at(x, n1, n2, n3);
        if ((e = x->x_current) && e->e_next)
        {
            x->x_mode = MPLAY;
            outlet_int(x->x_ob.o_outlet, e->e_next->e_time - e->e_time);
            x->x_current = e->e_next;
        }
    }
    RestoreA4();
}

void explode_nth(t_explode *x, long int n)
{
    t_evt *e;

    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        for (e = x->x_evt; e && n--; e = e->e_next);
        if (e)
            explode_sendevt(x, e, 1),
                outlet_int(x->x_ob.o_outlet, e->e_time);
    }
    RestoreA4();
}

void explode_params(t_explode *x, long int n1, long int n2, long int n3)
{
    SetUpA4();
    x->x_nfwd = n1;
    x->x_ftime = n2;
    x->x_oct = n3;
    RestoreA4();
}

void explode_clear(t_explode *x)
{
    t_evt *e;
    SetUpA4();
    if (x->x_mode != MUPDATE)
    {
        x->x_serial = 0;
        explode_stop(x);
        for (e = x->x_evt; e; e = e->e_next)
            freebytes(e, sizeof(t_evt));
        x->x_evt = x->x_current = 0;
#if (0)
        x->x_scrollmax = 0;
        if (x->x_wind)
        {
            x->x_qflags[QDRAWBG] = 1;
            qelem_set(x->x_qelem);
        }
#endif
    }
    RestoreA4();
}

void explode_next(t_explode *x)
{
    t_evt *e;

    SetUpA4();
    if (x->x_mode == MPLAY && (e = x->x_current))
    {
        outlet_int(x->x_out5, e->e_chan);
        outlet_int(x->x_out4, e->e_dur);
        outlet_int(x->x_out3, e->e_vel);
        outlet_int(x->x_out2, e->e_pit);
        if (x->x_current = e->e_next)
            outlet_int(x->x_ob.o_outlet, x->x_current->e_time - e->e_time);
        else
            explode_stop(x);
    }
    else
        post("not playing");
    RestoreA4();
}

void explode_int(t_explode *x, long int n)
{
    t_evt *e;

    SetUpA4();
    if (n < 0)
        n = 0;
    if (x->x_mode == MRECORD)
    {
        if (x->x_set3)
            explode_doappend(x, n, x->x_n1, x->x_n2, x->x_n3, x->x_n4);
        else
        {
            if (x->x_n2)        /* if nonzero velocity */
            {
                t_hang *h = (t_hang *) getbytes(sizeof(t_hang));
                /* add short note */
                explode_doappend(x, n, x->x_n1, x->x_n2, 10L, x->x_n4);
                /* mark it hanging */
                h->h_next = x->x_hang;
                h->h_evt = x->x_current;
                x->x_hang = h;
            }
            else
            {
                t_hang *h, *h2;
                t_hang stub;
                stub.h_next = x->x_hang;
                x->x_rectime += n;
                for (h = &stub; h2 = h->h_next; h = h2)
                {
                    if (h2->h_evt->e_pit == x->x_n1)
                    {
                        t_evt *e = h2->h_evt;
                        if ((e->e_dur = x->x_rectime - e->e_time) < 0)
                            e->e_dur = 0;
#ifdef GRAPHICS
                        if (x->x_wind)
                        {
                            x->x_qflags[QREMAX] = 1;
                            qelem_set(x->x_qelem);
                        }
#endif
                        h->h_next = h2->h_next;
                        freebytes(h2, sizeof(t_hang));
                        break;
                    }
                }
                x->x_hang = stub.h_next;
            }
        }
    }
    else if (x->x_mode == MFOLLOW)
    {
        if (x->x_set2)
        {
            if (x->x_n2)
                explode_dofollow(x, x->x_n1);
            else
                explode_fnoteoff(x, x->x_n1);
        }
        else if (x->x_n2)
            explode_dofollow(x, x->x_n1);
    }
    RestoreA4();
}

void explode_float(t_explode *x, double f)
{
    SetUpA4();
    explode_int(x, (long) f);
    RestoreA4();
}

void explode_in1(t_explode *x, long int n)
{
    SetUpA4();
    x->x_n1 = n;
    x->x_set1 = 1;
    RestoreA4();
}

void explode_in2(t_explode *x, long int n)
{
    SetUpA4();
    x->x_n2 = n;
    x->x_set2 = 1;
    RestoreA4();
}

void explode_in3(t_explode *x, long int n)
{
    SetUpA4();
    x->x_n3 = n;
    x->x_set3 = 1;
    RestoreA4();
}

void explode_in4(t_explode *x, long int n)
{
    SetUpA4();
    x->x_n4 = n;
    RestoreA4();
}

void explode_list(t_explode *x, struct symbol *s, int ac, struct atom *av)
{
    if (ac >= 5 && av[4].a_type == A_LONG)
        explode_in4(x, av[4].a_w.w_long);
    if (ac >= 4 && av[3].a_type == A_LONG)
        explode_in3(x, av[3].a_w.w_long);
    if (ac >= 3 && av[2].a_type == A_LONG)
        explode_in2(x, av[2].a_w.w_long);
    if (ac >= 2 && av[1].a_type == A_LONG)
        explode_in1(x, av[1].a_w.w_long);
    if (ac >= 1 && av[0].a_type == A_LONG)
        explode_int(x, av[0].a_w.w_long);
}

explode_init1(t_explode *x, t_symbol *s)
{
    x->x_out5 = outlet(x, &s_int);
    x->x_out4 = outlet(x, &s_int);
    x->x_out3 = outlet(x, &s_int);
    x->x_out2 = outlet(x, &s_int);
    outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in4);
    inlet(x, x, &s_int, &s_in3);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    x->x_n1 = x->x_n2 = x->x_n3 = 0;
    x->x_n4 = 1;
    x->x_set1 = x->x_set2 = x->x_set3 = 0;
    x->x_savemode = x->x_mode = MVERT;
    x->x_oct = 0;
    x->x_ftime = FTIME;
    x->x_nfwd = NFWD;
    x->x_skip = 0;
    x->x_hang = 0;
    x->x_serial = ++explode_nextserial;
    x->x_evt = 0;
    x->x_sym = s;
    if (s->s_thing)
    {
        if (ob_messlist(s->s_thing) != explode_class)
        {
            error("%s: already exists and is not an explode", s->s_name);
        }
        else
        {
            x->x_next = ((t_explode *) (s->s_thing));
            s->s_thing = (t_object *) x;
        }
    }
    else
    {
        s->s_thing = (t_object *) x;
        x->x_next = 0;
    }
}

#ifdef MIDIFILES

#define BS 1024
int t_fd;
int t_write;
int t_eof;
int t_vol;
char *t_next;
char *t_readend;
char *t_writeend;
char *t_buf;

t_getc()
{
    if (t_eof)
        return (-1);
    if (t_next >= t_readend)
    {
        long size = BS;
        int err = FSRead(t_fd, &size, t_buf);
        if (size <= 0)
        {
            t_eof = 1;
            return (-1);
        }
        t_next = t_buf;
        t_readend = t_buf + size;
    }
    return (0xff & *(t_next++));
}

long t_fread(buf, n)
char *buf;
long n;
{
    long c;
    for (c = 0; c < n; c++)
    {
        int b = t_getc();
        if (b < 0)
            return (c);
        *(buf++) = b;
    }
    return (n);
}

t_putc(c)
char c;
{
    if (t_next >= t_writeend)
    {
        long size = BS;
        int err = FSWrite(t_fd, &size, t_buf);
        t_next = t_buf;
        if (err != noErr)
        {
            t_eof = 1;
            return (-1);
        }
    }
    *(t_next++) = c;
    return (0);
}

long t_fwrite(buf, n)
char *buf;
long n;
{
    long c;
    for (c = 0; c < n; c++)
    {
        if (t_putc(*(buf++)))
            return (c);
    }
    return (n);
}

void *t_flush()
{
    long size = t_next - t_buf;
    if (size)
    {
        FSWrite(t_fd, &size, t_buf);
        t_next = t_buf;
    }
}

t_seek(n)
long n;
{
    t_flush();
    SetFPos(t_fd, 1, n);
}

t_close()
{
    if (t_write)
    {
        t_flush();
        FSClose(t_fd);
        FlushVol(0L, t_vol);
    }
    else
        FSClose(t_fd);
    DisposPtr(t_buf);
}

t_open(name, vol, type)
char *name;
int vol;
long type;
{
    int fd;
    if (!(t_buf = NewPtr((long) BS)))
        return (0);
    if (type)
    {
        int io;
        FInfo shit;
        io = GetFInfo(name, vol, &shit);
        if (io == noErr)
        {
            shit.fdCreator = 'max2';
            shit.fdType = type;
            if (SetFInfo(name, vol, &shit) != noErr)
                return (0);
        }
        else if (Create(name, vol, 'max2', type) != noErr)
            return (0);
        if (FSOpen(name, vol, &fd) != noErr)
            return (0);
        SetEOF(fd, 0L);
    }
    else
    {
        int io = FSOpen(name, vol, &fd);
        if (io != noErr)
            return (0);
    }
    t_write = (type != 0);
    t_vol = vol;
    t_fd = fd;
    t_eof = 0;
    t_readend = t_next = t_buf;
    t_writeend = t_buf + BS;
    return (1);
}

long readvar()
{
    long n = 0;
    int c;
    do
    {
        c = t_getc();
        if (c < 0)
            return (-1);
        n = (n << 7) | (c & 0x7f);
    }
    while (c & 0x80);
    return (n);
}

#define DEFTEMPO 120
#define TIMEBASE 96
#define SYSEX 0xf0
#define ENDEX 0xf7
#define NOTEON 0x90
#define NOTEOFF 0x80
#define THREEBYTE(x) ((x & 0xe0) != 0xc0)

void explode_import(x, flag)
t_explode *x;
long flag;
{
    long type = 'Midi';
    char name[256];
    Point where;
    SFReply reply;
    int i;
    long lx, cum;
    char *bag, buf[8];
    short doo[3];
    long jitter;
    long timebase;
    int tempo;
    long ticpersec;

    SetUpA4();
    where.h = 90;
    where.v = 82;
    name[0] = 0;
    SFGetFile(where, name, NIL, 1, &type, NIL, &reply);
    if (!reply.good)
        goto shit;
    if (!t_open(reply.fName, reply.vRefNum, 0L))
        goto shit;

    bag = "MThd";
    for (i = 4; i--; bag++)
        if (*bag != t_getc())
        {
            post("bad midi file");
            goto freeit;
        }
    if (t_fread(&lx, (long) sizeof(long)) < sizeof(long)
        || lx != 6 ||
        t_fread(doo, (long) sizeof(doo)) < sizeof(doo) ||
        t_fread(buf, (long) sizeof(buf)) < sizeof(buf))
    {
        post("bad midi file");
        goto freeit;
    }
    if (doo[0])
    {
        post("sorry, I take single-track files only");
        goto freeit;
    }
    if (doo[2] & 0x8000)        /* SMPTE-like timebase */
    {
        int frames = -(doo[2] >> 8);
        int perframe = doo[2] & 0xff;
        timebase = frames * perframe;
    }
    else
        timebase = doo[2];      /* normal timebase */

    if (timebase <= 0)
        timebase = 1;
    tempo = DEFTEMPO;
    ticpersec = (timebase * tempo) / 60;
    cum = 0;
    jitter = 0;
    explode_clear(x);
    while (1)
    {
        long nextbyte, mtype;
        long df = readvar();
        long dthi = 0, dt;
        if (!flag)
        {
            dthi = (df < 1024 ? (df * 1024000L) / ticpersec :
                    1024 * ((df * 1000L) / ticpersec)) + jitter;
            dt = dthi >> 10;
            jitter = dthi & 1023;
            if (dthi < 0)
                break;
        }
        else
            dt = df;
        if (df < 0)
            break;
        if ((nextbyte = t_getc()) < 0)
            break;
        cum += dt;
        if ((nextbyte & 0xf0) == 0xf0)
        {
            long len;
            if (nextbyte == SYSEX || nextbyte == ENDEX)
            {
                len = readvar();
                while (len--)
                {
                    if ((nextbyte = t_getc()) < 0
                        || (nextbyte & 0x80) && (len || nextbyte != ENDEX))
                    {
                        post("bad MIDI file");
                        goto freeit;
                    }
                }
            }
            else
            {
                t_getc();
                len = readvar();
                while (len--)
                    t_getc();
            }
        }
        else
        {
            if (nextbyte & 0x80)
            {
                mtype = nextbyte;
                if ((nextbyte = t_getc()) < 0)
                    break;
            }
            if (THREEBYTE(mtype))
            {
                long byte3 = t_getc();
                int foo = mtype & 0xf0;
                if (foo == NOTEON || foo == NOTEOFF)
                {
                    if (byte3 && foo == NOTEON)
                    {
                        t_hang *h = (t_hang *) getbytes(sizeof(t_hang));
                        /* add short note */
                        x->x_rectime = 0;
                        explode_doappend(x, cum, nextbyte, byte3, 10L, 1L);
                        /* mark it hanging */
                        h->h_next = x->x_hang;
                        h->h_evt = x->x_current;
                        x->x_hang = h;
                    }
                    else
                    {
                        t_hang *h, *h2;
                        t_hang stub;
                        stub.h_next = x->x_hang;
                        for (h = &stub; h2 = h->h_next; h = h2)
                        {
                            if (h2->h_evt->e_pit == nextbyte)
                            {
                                t_evt *e = h2->h_evt;
                                if ((e->e_dur = cum - e->e_time) < 0)
                                    e->e_dur = 0;
                                x->x_qflags[QREMAX] = 1;
                                qelem_set(x->x_qelem);
                                h->h_next = h2->h_next;
                                freebytes(h2, sizeof(t_hang));
                                break;
                            }
                        }
                        x->x_hang = stub.h_next;
                    }
                }
            }
        }
    }
  freeit:t_close();
    explode_stop(x);
  shit:
    RestoreA4();
}

static long wbytes;

writevar(n)
long n;
{
    char buf[4], *bp = buf + 4, *ep = bp;
    n &= 0xfffffff;             /* force it to 28 bits maximum */
    *(--bp) = n & 0x7f;
    while (n >>= 7)
        *(--bp) = (n & 0x7f) | 0x80;
    while (bp < ep)
        wbytes++, t_putc(*(bp++));
}

static char hdr26[] = { 77, 84, 104, 100, 0, 0, 0, 6, 0, 0, 0, 1, 0, 96,
    77, 84, 114, 107, 0, 0, 0, 0
};

#define HSIZE 22
#define SIZEONSET 18

void explode_export(x, flag)
t_explode *x;
long flag;
{
    char name[256];
    Point where;
    SFReply reply;
    long df, jitter, lasttime;
    int i, on;
    t_evt *e;
    long type;

    SetUpA4();
    where.h = 90;
    where.v = 82;
    name[0] = 0;
    SFPutFile(where, "\p", name, NIL, &reply);
    if (!reply.good)
        goto shit;
    if (!t_open(reply.fName, reply.vRefNum, 'Midi'))
        goto shit;

    df = jitter = lasttime = 0;
    explode_stop(x);
    e = x->x_evt;
    wbytes = 0;
    post("writing: %s", name);
    t_fwrite(hdr26, (long) HSIZE);

    while (e || x->x_hang)
    {
        long nexttime = 0x7fffffff, shit, big, petit;
        int noteon = 0;
        t_hang *h, *htake = 0;
        int pit, vel;
        if (e)
        {
            noteon = 1;
            nexttime = e->e_time;
        }
        for (h = x->x_hang; h; h = h->h_next)
        {
            if ((shit = h->h_evt->e_time + h->h_evt->e_dur) < nexttime)
            {
                nexttime = shit;
                htake = h;
                noteon = 0;
            }
        }
        if (noteon)
        {
            h = getbytes(sizeof(t_hang));
            h->h_evt = e;
            h->h_next = x->x_hang;
            x->x_hang = h;
            pit = e->e_pit;
            vel = e->e_vel;
            e = e->e_next;
        }
        else if (htake)
        {
            t_hang *h2, stub;
            pit = htake->h_evt->e_pit;
            vel = 0;
            stub.h_next = x->x_hang;
            for (h = &stub; h2 = h->h_next; h = h2)
                if (h2 == htake)
                {
                    h->h_next = h2->h_next;
                    freebytes(h2, sizeof(t_hang));
                }
            x->x_hang = stub.h_next;
        }
        else
            post("export bug");

        shit = nexttime - lasttime;

        if (flag)
            df = shit;
        else
        {
            big = (shit > 60000 ?
                   ((shit * 1024L) / 125) * 24 :
                   (shit * (24 * 1024L)) / 125) + jitter;
            df = big >> 10;
            jitter = big & 1023;
        }

        writevar(df);

        wbytes += 3;
        t_putc(144);
        t_putc(pit);
        t_putc(vel);
        lasttime = nexttime;
    }
    writevar(0L);               /* write out end of track */
    wbytes += 3;
    t_putc(0xff);
    t_putc(0x2f);
    t_putc(0x00);

    t_seek((long) SIZEONSET);
    t_fwrite(&wbytes, (long) sizeof(wbytes));
    t_close();
    explode_stop(x);
  shit:
    RestoreA4();
}
#endif
