#ifndef lint
static char rcsid[] =
    "$Revision: 2.7 $ IRCAM $Date: 1994/01/04 18:03:44 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.7 $ IRCAM $Date: 1994/01/04 18:03:44 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"

/* <a> totally rewritten to add voice-stealing */

extern struct symbol s_note, s_in1, s_in2, s_print, s_poly,
    s_clear, s_, s_pipe, s_flush, s_sustain, s_poly,
    s_stop, s_stripnote, s_makenote, s_bag, s_cut;

void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);

typedef struct voice {
    long v_state;
    long v_pitch;
} t_voice;

typedef struct loco {
    struct object l_ob;
    long l_in1;
    void *l_out1, *l_out2;
    long l_nvoice;
    long l_count;
    struct voice **l_v;
    int l_strat;
} t_loco;


/* pitch or other unique index */
void loco_int(struct loco *x, long int n)
{
    long i;
    struct voice *v;
    if (x->l_in1)               /* if velo nonzero */
    {
        if (x->l_strat)
        {
            struct voice *von = 0, *voff = 0;
            long ion, ioff;
            for (i = 0, v = *x->l_v; i < x->l_nvoice; i++, v++)
            {
                if (v->v_state & 1)
                {
                    if (!von || v->v_state < von->v_state)
                        von = v, ion = i;
                }
                else
                {
                    if (!voff || v->v_state < voff->v_state)
                        voff = v, ioff = i;
                }
            }
            if (!voff)          /* if all notes are on steal one */
            {
                long pwas = von->v_pitch;
                if (!von)
                {
                    bug("poly");
                    return;
                }
                von->v_state = ((x->l_count++) << 1) + 1;
                von->v_pitch = n;
                ioff = ion;
                outlet_int(x->l_out2, 0L);
                outlet_int(x->l_out1, pwas);
                outlet_int(x->l_ob.o_outlet, ion + 1);
            }
            else
            {
                voff->v_state = ((x->l_count++) << 1) + 1;
                voff->v_pitch = n;
            }
            outlet_int(x->l_out2, x->l_in1);
            outlet_int(x->l_out1, n);
            outlet_int(x->l_ob.o_outlet, ioff + 1);
        }
        else                    /* strategy zero: no voice stealing */
        {
            for (i = 0, v = *x->l_v; i < x->l_nvoice; i++, v++)
            {
                if (!(v->v_state & 1))
                {
                    v->v_state = 1;
                    v->v_pitch = n;
                    outlet_int(x->l_out2, x->l_in1);
                    outlet_int(x->l_out1, n);
                    outlet_int(x->l_ob.o_outlet, i + 1);
                    return;
                }
            }
        }
    }
    else                        /* note off */
    {
        for (i = 0, v = *x->l_v; i < x->l_nvoice; i++, v++)
        {
            if (v->v_pitch == n && (v->v_state & 1))
            {
                v->v_state = (x->l_count++) << 1;
                outlet_int(x->l_out2, 0L);
                outlet_int(x->l_out1, n);
                outlet_int(x->l_ob.o_outlet, i + 1);
                return;
            }
        }
    }
}

void loco_in1(struct loco *x, long int n)
{
    x->l_in1 = (n > 0 ? n : 0);
}

void loco_stop(struct loco *x)
{
    long i;
    struct voice *v;
    HLock(x->l_v);
    for (i = 0, v = *x->l_v; i < x->l_nvoice; i++, v++)
        if (v->v_state & 1)
        {
            outlet_int(x->l_out2, 0L);
            outlet_int(x->l_out1, v->v_pitch);
            outlet_int(x->l_ob.o_outlet, i + 1);
            v->v_state = (x->l_count++) << 1;
        }
    HUnlock(x->l_v);
}

void loco_clear(struct loco *x)
{
    long i;
    struct voice *v;
    for (v = *x->l_v, i = x->l_nvoice; i--; v++)
        v->v_pitch = v->v_state = 0;
}

void loco_set(struct loco *x, long int pit, long int vel, long int which)
{
    struct voice *v = (*x->l_v) + (which - 1);
    if (which > x->l_nvoice || which < 1)
    {
        error("poly: 'set' voice %ld is out of range", which);
        return;
    }
    v->v_pitch = pit;
    if (vel)
        v->v_state = ((x->l_count++) << 1) + 1;
    else
        v->v_state = (x->l_count++) << 1;
}

void loco_ff(struct loco *x)
{
    DisposHandle(x->l_v);
}

void *loco_class;
void loco_setup()
{
    c_new(&s_poly, loco_ff, sizeof(t_loco), 0, 0);
    c_addint(loco_int);
    c_addfloat(stdfloat);
    c_addinx(loco_in1, 1);
    c_addmess(loco_stop, &s_stop, 0);
    c_addmess(loco_clear, gensym("clear"), 0);
    c_addmess(loco_set, gensym("set"), A_LONG, A_LONG, A_LONG, 0);
}

void *loco_new(long int n, long int strat)
{
    long i;
    struct voice *v;
    struct loco *x =
        (struct loco *) obj_new((t_messlist **) & loco_class, loco_setup);
    if (n < 1)
        n = 16;
    x->l_nvoice = n;
    x->l_v = (struct voice **) t_newhandle(n * sizeof(struct voice));
    for (v = *x->l_v, i = n; i--; v++)
        v->v_pitch = v->v_state = 0;
    x->l_in1 = 0;
    x->l_strat = strat;
    inlet(x, x, &s_int, &s_in1);
    x->l_out2 = outlet(x, &s_int);
    x->l_out1 = outlet(x, &s_int);
    outlet(x, &s_int);
    return (x);
}
