#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1993/11/22 14:35:15 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1993/11/22 14:35:15 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
extern void *midi_noteout, *midi_nc[32],
    *midi_ctlout, *midi_cc[32],
    *midi_polyout, *midi_yc[32],
    *midi_pgmout, *midi_pc[32],
    *midi_touchout, *midi_tc[32], *midi_bendout, *midi_bc[32],
    *midi_abyteout, *midi_bbyteout, *midi_aexout, *midi_bexout,
    *midi_artout, *midi_brtout;
extern long midi_chan, midi_arg1, midi_arg2;
extern int outport[];
void *outlet(void *, struct symbol *), *inlet(void *, void *,
                                              struct symbol *,
                                              struct symbol *);
extern struct symbol s_print, s_pgmin, s_pgmout, s_touchin, s_touchout,
    s_bendin, s_bendout, s_notein, s_noteout, s_ctlin, s_ctlout, s_in1,
    s_in2, s_midiin, s_midiout, s_sysexin, s_rtin, s_tempoin, s_tick,
    s_start, s_stop, s_polyin, s_polyout;
#define SNOTE 0x90
#define SPOLY 0xa0
#define SCTL 0xb0
#define SPGM 0xc0
#define STCH 0xd0
#define SBEND 0xe0

#ifdef GRAPHICS
#define Xmit(a,b)
#endif

/* 1. program change */
typedef struct pgmout {
    struct object p_ob;
    long p_status;              /* status byte to send */
    int p_port;                 /* go to other port */
} t_pgmout;

void pgmout_int(struct pgmout *x, long int n)
{
    int what = outport[x->p_port];
    if ((--n) & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->p_status);
        Xmit(0L, n);
    }
    else if (what == 1)
    {
        Xmit(1L, x->p_status);
        Xmit(1L, n);
    }
}

void pgmout_float(struct pgmout *x, double f)
{
    pgmout_int(x, (long) f);
}

void pgmout_in1(struct pgmout *x, long int n)
{
    x->p_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->p_port = 1, n -= 16;
    x->p_status = (SPGM + n - 1);
}

void *pgmout_class;
void pgmout_setup()
{
    c_new(&s_pgmout, NIL, sizeof(t_pgmout), 0, 0);
    c_addint(pgmout_int);
    c_addfloat(stdfloat);
    c_addinx(pgmout_in1, 1);
}

void *pgmout_new(long int n)
{
    struct pgmout *x =
        (struct pgmout *) obj_new((t_messlist **) & pgmout_class,
                                  pgmout_setup);
    inlet(x, x, &s_int, &s_in1);
    pgmout_in1(x, n);
    return (x);
}

typedef struct pgmin {
    struct object p_ob;
    void *p_outlet;             /* outlet for receive */
    void *p_chanout;            /* outlet for channel if any */
    void *p_sysout;             /* outlet that that one is plugged into */
} t_pgmin;

void pgmin_bang(struct pgmin *x)
{
    if (x->p_chanout)
        outlet_int(x->p_chanout, midi_chan);
    outlet_int(x->p_outlet, midi_arg1 + 1);
}

void pgmin_ff(struct pgmin *x)
{
    outlet_rm(x->p_sysout, x);
}

void *pgmin_class;
void pgmin_setup()
{
    c_new(&s_pgmin, pgmin_ff, sizeof(t_pgmin), 0, 1);
    c_addbang(pgmin_bang);
}

void *pgmin_new(long int n)
{
    struct pgmin *x =
        (struct pgmin *) obj_new((t_messlist **) & pgmin_class,
                                 pgmin_setup);
    void **outp;
    if (n > 0 && n <= 32)
        outp = &midi_pc[n - 1], x->p_chanout = 0;
    else
        outp = &midi_pgmout, x->p_chanout = outlet(x, &s_int);
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->p_outlet = outlet(x, &s_int);
    outlet_add(x->p_sysout = *outp, x);
    return (x);
}

/* 2. pitch bend */
typedef struct bendout {
    struct object b_ob;
    long b_status;              /* status byte to send */
    int b_port;                 /* go to other port */
} t_bendout;

void bendout_int(struct bendout *x, long int n)
{
    int what = outport[x->b_port];
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->b_status);
        Xmit(0L, n);
        Xmit(0L, n);
    }
    else if (what == 1)
    {
        Xmit(1L, x->b_status);
        Xmit(1L, n);
        Xmit(1L, n);
    }
}

void bendout_float(struct bendout *x, double f)
{
    bendout_int(x, (long) f);
}

void bendout_in1(struct bendout *x, long int n)
{
    x->b_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->b_port = 1, n -= 16;
    x->b_status = (SBEND + n - 1);
}

void *bendout_class;
void bendout_setup()
{
    c_new(&s_bendout, NIL, sizeof(t_bendout), 0, 0);
    c_addint(bendout_int);
    c_addfloat(stdfloat);
    c_addinx(bendout_in1, 1);
}

void *bendout_new(long int n)
{
    struct bendout *x =
        (struct bendout *) obj_new((t_messlist **) & bendout_class,
                                   bendout_setup);
    inlet(x, x, &s_int, &s_in1);
    bendout_in1(x, n);
    return (x);
}

typedef struct bendin {
    struct object b_ob;
    void *b_outlet;             /* outlet for receive */
    void *b_chanout;            /* outlet for channel if any */
    void *b_sysout;             /* outlet that that one is plugged into */
} t_bendin;

void bendin_bang(struct bendin *x)
{
    if (x->b_chanout)
        outlet_int(x->b_chanout, midi_chan);
    outlet_int(x->b_outlet, midi_arg2);
}

void bendin_ff(struct bendin *x)
{
    outlet_rm(x->b_sysout, x);
}

void *bendin_class;
void bendin_setup()
{
    c_new(&s_bendin, bendin_ff, sizeof(t_bendin), 0, 1);
    c_addbang(bendin_bang);
}

void *bendin_new(long int n)
{
    struct bendin *x =
        (struct bendin *) obj_new((t_messlist **) & bendin_class,
                                  bendin_setup);
    void **outp;
    if (n > 0 && n <= 32)
        outp = &midi_bc[n - 1], x->b_chanout = 0;
    else
        outp = &midi_bendout, x->b_chanout = outlet(x, &s_int);
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->b_outlet = outlet(x, &s_int);
    outlet_add(x->b_sysout = *outp, x);
    return (x);
}

/* 3. channel aftertouch */
typedef struct touchout {
    struct object t_ob;
    long t_status;              /* status byte to send */
    int t_port;                 /* go to other port */
} t_touchout;

void touchout_int(struct touchout *x, long int n)
{
    int what = outport[x->t_port];
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->t_status);
        Xmit(0L, n);
    }
    else if (what == 1)
    {
        Xmit(1L, x->t_status);
        Xmit(1L, n);
    }
}

void touchout_float(struct touchout *x, double f)
{
    touchout_int(x, (long) f);
}

void touchout_in1(struct touchout *x, long int n)
{
    x->t_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->t_port = 1, n -= 16;
    x->t_status = (STCH + n - 1);
}

void *touchout_class;
void touchout_setup()
{
    c_new(&s_touchout, NIL, sizeof(t_touchout), 0, 0);
    c_addint(touchout_int);
    c_addfloat(stdfloat);
    c_addinx(touchout_in1, 1);
}

void *touchout_new(long int n)
{
    struct touchout *x =
        (struct touchout *) obj_new((t_messlist **) & touchout_class,
                                    touchout_setup);
    inlet(x, x, &s_int, &s_in1);
    touchout_in1(x, n);
    return (x);
}

typedef struct touchin {
    struct object t_ob;
    void *t_outlet;             /* outlet for receive */
    void *t_chanout;            /* outlet for channel if any */
    void *t_sysout;             /* outlet that that one is plugged into */
} t_touchin;

void touchin_bang(struct touchin *x)
{
    if (x->t_chanout)
        outlet_int(x->t_chanout, midi_chan);
    outlet_int(x->t_outlet, midi_arg1);
}

void touchin_ff(struct touchin *x)
{
    outlet_rm(x->t_sysout, x);
}

void *touchin_class;
void touchin_setup()
{
    c_new(&s_touchin, touchin_ff, sizeof(t_touchin), 0, 1);
    c_addbang(touchin_bang);
}

void *touchin_new(long int n)
{
    struct touchin *x =
        (struct touchin *) obj_new((t_messlist **) & touchin_class,
                                   touchin_setup);
    void **outp;
    if (n > 0 && n <= 32)
        outp = &midi_tc[n - 1], x->t_chanout = 0;
    else
        outp = &midi_touchout, x->t_chanout = outlet(x, &s_int);
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->t_outlet = outlet(x, &s_int);
    outlet_add(x->t_sysout = *outp, x);
    return (x);
}

/* 4. midi control */
typedef struct ctlout {
    struct object c_ob;
    long c_status;              /* status byte to send */
    long c_ctlno;               /* control number */
    int c_port;                 /* go to other port */
} t_ctlout;

void ctlout_int(struct ctlout *x, long int n)
{
    int what = outport[x->c_port];
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->c_status);
        Xmit(0L, x->c_ctlno);
        Xmit(0L, n);
    }
    else if (what == 1)
    {
        Xmit(1L, x->c_status);
        Xmit(1L, x->c_ctlno);
        Xmit(1L, n);
    }
}

void ctlout_float(struct ctlout *x, double f)
{
    ctlout_int(x, (long) f);
}

void ctlout_in1(struct ctlout *x, long int n)
{
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    x->c_ctlno = n;
}

void ctlout_in2(struct ctlout *x, long int n)
{
    x->c_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->c_port = 1, n -= 16;
    x->c_status = (SCTL + n - 1);
}

void *ctlout_class;
void ctlout_setup()
{
    c_new(&s_ctlout, NIL, sizeof(t_ctlout), 0, 0);
    c_addint(ctlout_int);
    c_addfloat(ctlout_float);
    c_addinx(ctlout_in1, 1);
    c_addinx(ctlout_in2, 2);
}

void *ctlout_new(long int n1, long int n2)
              /* control number and channel */
{
    struct ctlout *x =
        (struct ctlout *) obj_new((t_messlist **) & ctlout_class,
                                  ctlout_setup);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    ctlout_in1(x, n1);
    ctlout_in2(x, n2);
    return (x);
}

typedef struct ctlin {
    struct object c_ob;
    void *c_outlet;             /* outlet for receive */
    void *c_numout;             /* ctl number out */
    void *c_chanout;            /* outlet for channel if any */
    void *c_sysout;             /* outlet that that one is plugged into */
    long c_ctlno;
} t_ctlin;

void ctlin_bang(struct ctlin *x)
{
    if (x->c_numout)
    {
        if (x->c_chanout)
            outlet_int(x->c_chanout, midi_chan);
        outlet_int(x->c_numout, midi_arg1);
        outlet_int(x->c_outlet, midi_arg2);
    }
    else if (x->c_ctlno == midi_arg1)
    {
        if (x->c_chanout)
            outlet_int(x->c_chanout, midi_chan);
        outlet_int(x->c_outlet, midi_arg2);
    }
}

void ctlin_ff(struct ctlin *x)
{
    outlet_rm(x->c_sysout, x);
}

void *ctlin_class;
void ctlin_setup()
{
    c_new(&s_ctlin, ctlin_ff, sizeof(t_ctlin), 0, 1);
    c_addbang(ctlin_bang);
}

void *ctlin_new(long int n1, long int n2)
              /* ctl number and channel */
{
    struct ctlin *x =
        (struct ctlin *) obj_new((t_messlist **) & ctlin_class,
                                 ctlin_setup);
    void **outp;
    if (n1 == 128)
        n1 = 0;
    else if (!n1 || n1 & -128)
        n1 = -1;
    x->c_ctlno = n1;
    if (n2 > 0 && n2 <= 32)
        outp = &midi_cc[n2 - 1], x->c_chanout = 0;
    else
        outp = &midi_ctlout, x->c_chanout = outlet(x, &s_int);
    if (n1 == -1)
        x->c_numout = outlet(x, &s_int);
    else
        x->c_numout = 0;
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->c_outlet = outlet(x, &s_int);
    outlet_add(x->c_sysout = *outp, x);
    return (x);
}

/* 5. notes */
typedef struct noteout {
    struct object n_ob;
    long n_status;              /* status byte to send */
    long n_vel;                 /* velocity */
    int n_port;                 /* go to other port */
} t_noteout;

void noteout_int(struct noteout *x, long int n)
{
    int what = outport[x->n_port];
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->n_status);
        Xmit(0L, n);
        Xmit(0L, x->n_vel);
    }
    else if (what == 1)
    {
        Xmit(1L, x->n_status);
        Xmit(1L, n);
        Xmit(1L, x->n_vel);
    }
}

void noteout_float(struct noteout *x, double f)
{
    noteout_int(x, (long) f);
}

void noteout_in1(struct noteout *x, long int n)
{
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    x->n_vel = n;
}

void noteout_in2(struct noteout *x, long int n)
{
    x->n_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->n_port = 1, n -= 16;
    x->n_status = (SNOTE + n - 1);
}

void *noteout_class;
void noteout_setup()
{
    c_new(&s_noteout, NIL, sizeof(t_noteout), 0, 0);
    c_addint(noteout_int);
    c_addfloat(noteout_float);
    c_addinx(noteout_in1, 1);
    c_addinx(noteout_in2, 2);
}

void *noteout_new(long int n)
{
    struct noteout *x =
        (struct noteout *) obj_new((t_messlist **) & noteout_class,
                                   noteout_setup);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    noteout_in2(x, n);
    x->n_vel = 0;
    return (x);
}

typedef struct notein {
    struct object n_ob;
    void *n_outlet;             /* outlet for receive */
    void *n_velout;             /* velocity outlet */
    void *n_chanout;            /* outlet for channel if any */
    void *n_sysout;             /* outlet that that one is plugged into */
} t_notein;

void notein_bang(struct notein *x)
{
    if (x->n_chanout)
        outlet_int(x->n_chanout, midi_chan);
    outlet_int(x->n_velout, midi_arg2);
    outlet_int(x->n_outlet, midi_arg1);
}

void notein_ff(struct notein *x)
{
    outlet_rm(x->n_sysout, x);
}

void *notein_class;
void notein_setup()
{
    c_new(&s_notein, notein_ff, sizeof(t_notein), 0, 1);
    c_addbang(notein_bang);
}

void *notein_new(long int n)
{
    struct notein *x =
        (struct notein *) obj_new((t_messlist **) & notein_class,
                                  notein_setup);
    void **outp;
    if (n > 0 && n <= 32)
        outp = &midi_nc[n - 1], x->n_chanout = 0;
    else
        outp = &midi_noteout, x->n_chanout = outlet(x, &s_int);
    x->n_velout = outlet(x, &s_int);
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->n_outlet = outlet(x, &s_int);
    outlet_add(x->n_sysout = *outp, x);
    return (x);
}

/* DDZ 7. polyphonic keypressure */

typedef struct polyout {
    struct object p_ob;
    long p_status;              /* status byte to send */
    long p_keyno;               /* pressure value */
    int p_port;                 /* go to other port */
} t_polyout;

void polyout_int(struct polyout *x, long int n)
{
    int what = outport[x->p_port];
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    if (!what)
    {
        Xmit(0L, x->p_status);
        Xmit(0L, x->p_keyno);
        Xmit(0L, n);
    }
    else if (what == 1)
    {
        Xmit(1L, x->p_status);
        Xmit(1L, x->p_keyno);
        Xmit(1L, n);
    }
}

void polyout_float(struct polyout *x, double f)
{
    polyout_int(x, (long) f);
}

void polyout_in1(struct polyout *x, long int n)
{
    if (n & -128)
        n = (n > 0 ? 127 : 0);
    x->p_keyno = n;
}

void polyout_in2(struct polyout *x, long int n)
{
    x->p_port = 0;
    if (n < 1)
        n = 1;
    else if (n > 32)
        n = 32;
    if (n > 16)
        x->p_port = 1, n -= 16;
    x->p_status = (SPOLY + n - 1);
}

void *polyout_class;
void polyout_setup()
{
    c_new(&s_polyout, NIL, sizeof(t_polyout), 0, 0);
    c_addint(polyout_int);
    c_addfloat(polyout_float);
    c_addinx(polyout_in1, 1);
    c_addinx(polyout_in2, 2);
}

void *polyout_new(long int n1, long int n2)
              /* key number and pressure value */
{
    struct polyout *x =
        (struct polyout *) obj_new((t_messlist **) & polyout_class,
                                   polyout_setup);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    polyout_in1(x, n1);
    polyout_in2(x, n2);
    return (x);
}

/* polyphonic key pressure in */

typedef struct polyin {
    struct object p_ob;
    void *p_outlet;             /* outlet for pressure value */
    void *p_keyout;             /* key number out */
    void *p_chanout;            /* outlet for channel if any */
    void *p_sysout;             /* outlet that that one is plugged into */
    long p_keyno;
} t_polyin;

void polyin_bang(struct polyin *x)
{
    if (x->p_keyout)
    {
        if (x->p_chanout)
            outlet_int(x->p_chanout, midi_chan);
        outlet_int(x->p_keyout, midi_arg1);
        outlet_int(x->p_outlet, midi_arg2);
    }
    else if (x->p_keyno == midi_arg1)
    {
        if (x->p_chanout)
            outlet_int(x->p_chanout, midi_chan);
        outlet_int(x->p_outlet, midi_arg2);
    }
}

void polyin_ff(struct polyin *x)
{
    outlet_rm(x->p_sysout, x);
}

void *polyin_class;
void polyin_setup()
{
    c_new(&s_polyin, polyin_ff, sizeof(t_polyin), 0, 1);
    c_addbang(polyin_bang);
}

void *polyin_new(long int n1, long int n2)
              /* key number and channel */
{
    struct polyin *x =
        (struct polyin *) obj_new((t_messlist **) & polyin_class,
                                  polyin_setup);
    void **outp;
    if (n1 == 128)
        n1 = 0;
    else if (!n1 || n1 & -128)
        n1 = -1;
    x->p_keyno = n1;
    if (n2 > 0 && n2 <= 32)
    {
        outp = &midi_yc[n2 - 1];
        x->p_chanout = 0;
    }
    else
    {
        outp = &midi_polyout;
        x->p_chanout = outlet(x, &s_int);
    }
    if (n1 == -1)
        x->p_keyout = outlet(x, &s_int);
    else
        x->p_keyout = 0;
    if (!(*outp))
        *outp = outlet(NIL, &s_bang);
    x->p_outlet = outlet(x, &s_int);
    outlet_add(x->p_sysout = *outp, x);
    return (x);
}


/* midi exclusive in, bytewise in and out */
typedef struct midiout {
    struct object m_ob;
    long m_which;
} t_midiout;

void midiout_int(struct midiout *x, long int n)
{
    if (x->m_which)
        Xmit(1L, n);
    else
        Xmit(0L, n);
}

void *midiout_class;
void midiout_setup()
{
    c_new(&s_midiout, NIL, sizeof(t_midiout), 0, 0);
    c_addint(midiout_int);
    c_addfloat(stdfloat);
}

void *midiout_new(long int n)
{
    struct midiout *x =
        (struct midiout *) obj_new((t_messlist **) & midiout_class,
                                   midiout_setup);
    x->m_which = n;
    return (x);
}

typedef struct midiin {
    struct object m_ob;
    void *m_outlet;
    void *m_sysout;
} t_midiin;

void midiin_ff(struct midiin *x)
{
    outlet_rm(x->m_sysout, x->m_outlet);
}

void *midiin_class;
void midiin_setup()
{
    c_new(&s_midiin, midiin_ff, sizeof(t_midiin), 0, 1);
}

void *midiin_new(long int n)
{
    struct midiin *x =
        (struct midiin *) obj_new((t_messlist **) & midiin_class,
                                  midiin_setup);
    void **outp = (n ? &midi_bbyteout : &midi_abyteout);
    if (!(*outp))
        *outp = outlet(NIL, &s_int);
    x->m_outlet = outlet(x, &s_int);
    outlet_add(x->m_sysout = *outp, x->m_outlet);
    return (x);
}

void *sysexin_class;
void sysexin_setup()
{
    c_new(&s_sysexin, midiin_ff, sizeof(t_midiin), 0, 1);
}

void *sysexin_new(long int n)
{
    struct midiin *x =
        (struct midiin *) obj_new((t_messlist **) & sysexin_class,
                                  sysexin_setup);
    void **outp = (n ? &midi_bexout : &midi_aexout);
    if (!(*outp))
        *outp = outlet(NIL, &s_int);
    x->m_outlet = outlet(x, &s_int);
    outlet_add(x->m_sysout = *outp, x->m_outlet);
    return (x);
}

void *rtin_class;
void rtin_setup()
{
    c_new(&s_rtin, midiin_ff, sizeof(t_midiin), 0, 1);
}

void *rtin_new(long int n)
{
    struct midiin *x =
        (struct midiin *) obj_new((t_messlist **) & rtin_class,
                                  rtin_setup);
    void **outp = (n ? &midi_brtout : &midi_artout);
    if (!(*outp))
        *outp = outlet(NIL, &s_int);
    x->m_outlet = outlet(x, &s_int);
    outlet_add(x->m_sysout = *outp, x->m_outlet);
    return (x);
}
