#ifndef lint
static char rcsid[] =
    "$Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:03:50 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:03:50 $
 *
 * FTS by Miller Puckette
 *
 */

#include "m_mess.h"

extern struct symbol s_note, s_in1, s_in2,
    s_clear, s_, s_pipe, s_flush, s_sustain,
    s_stop, s_stripnote, s_makenote, s_bag, s_cut;

typedef struct linknote {
    long l_pit;
    void *l_clock;
    void *l_who;
    void (*l_fn)();
    struct linknote *l_next;
    struct linknote *l_prev;
} t_linknote;


void linknote_bang(struct linknote *x)
{
    (*(x->l_fn)) (x->l_who, x->l_pit);
    x->l_next->l_prev = x->l_prev;
    x->l_prev->l_next = x->l_next;
    clock_free(x->l_clock);
    freebytes(x, sizeof(t_linknote));
}

void linknote_whimper(struct linknote *x)
{
    x->l_next->l_prev = x->l_prev;
    x->l_prev->l_next = x->l_next;
    clock_free(x->l_clock);
    freebytes(x, sizeof(t_linknote));
}

void linknote_new(struct linknote *next, void *who, void (*fn)(),
                  long int pit, long int dur)
{
    struct linknote *x =(struct linknote *) getbytes(sizeof(t_linknote));
    x->l_next = next;
    (x->l_prev = next->l_prev)->l_next = x;
    next->l_prev = x;
    x->l_who = who;
    x->l_pit = pit;
    x->l_fn = fn;
    x->l_clock = clock_new(x, linknote_bang);
    clock_set(x->l_clock, logicaltime + dur);
}

typedef struct makenote {
    struct object m_ob;
    void *m_outlet;
    long m_vel;
    long m_dur;
    struct linknote m_linknote;
} t_makenote;

void makenote_in1(struct makenote *x, long int n)
{
    x->m_vel = n;
}

void makenote_in2(struct makenote *x, long int n)
{
    x->m_dur = n;
}

void domakenote(struct makenote *x, long int val)
{
    outlet_int(x->m_outlet, 0L);
    outlet_int(x->m_ob.o_outlet, val);
}

void makenote_int(struct makenote *x, long int n)
{
    outlet_int(x->m_outlet, x->m_vel);
    outlet_int(x->m_ob.o_outlet, n);
    linknote_new(&x->m_linknote, x, domakenote, n, x->m_dur);
}

void makenote_float(struct makenote *x, double f)
{
    makenote_int(x, (long) f);
}

void makenote_stop(struct makenote *x)
{
    struct linknote *ln = &x->m_linknote, *ln2;
    while ((ln2 = ln->l_next) != ln)
        linknote_bang(ln2);
}

void makenote_clear(struct makenote *x)
{
    struct linknote *ln = &x->m_linknote, *ln2;
    while ((ln2 = ln->l_next) != ln)
        linknote_whimper(ln2);
}

void makenote_ff(struct makenote *x)
{
    makenote_stop(x);
}

void *makenote_class;

void makenote_setup()
{
    c_new(&s_makenote, makenote_ff, sizeof(t_makenote), 0, 0);
    c_addint(makenote_int);
    c_addmess(makenote_clear, &s_clear, 0);
    c_addmess(makenote_stop, &s_stop, 0);
    c_addfloat(makenote_float);
    c_addinx(makenote_in1, 1);
    c_addinx(makenote_in2, 2);
}

void *makenote_new(long int n1, long int n2)
{
    struct makenote *x =
        (struct makenote *) obj_new((t_messlist **) & makenote_class,
                                    makenote_setup);
    x->m_linknote.l_next = x->m_linknote.l_prev = &x->m_linknote;
    x->m_outlet = outlet(x, &s_int);
    outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    x->m_vel = n1;
    x->m_dur = n2;
    return (x);
}

typedef struct stripnote {
    struct object n_ob;
    long n_n;
    void *n_out2;
} t_stripnote;

void stripnote_int(struct stripnote *x, long int n)
{
    if (x->n_n)
    {
        outlet_int(x->n_out2, x->n_n);
        outlet_int(x->n_ob.o_outlet, n);
    }
}

void stripnote_float(struct stripnote *x, double f)
{
    if (x->n_n)
        outlet_int(x->n_ob.o_outlet, (long) f);
}

void stripnote_in1(struct stripnote *x, long int n)
{
    x->n_n = n;
}

void *stripnote_class;
void stripnote_setup()
{
    c_new(&s_stripnote, NIL, sizeof(t_stripnote), 0, 0);
    c_addint(stripnote_int);
    c_addfloat(stripnote_float);
    c_addinx(stripnote_in1, 1);
}

void *stripnote_new()
{
    struct stripnote *x =
        (struct stripnote *) obj_new((t_messlist **) & stripnote_class,
                                     stripnote_setup);
    x->n_out2 = outlet(x, &s_int);
    outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in1);
    x->n_n = 0;
    return (x);
}

typedef struct slink {
    long l_pit;
    struct slink *l_next;
} t_slink;

slink_free(t_slink *x)
{
    freebytes(x, sizeof(t_slink));
}

struct slink *slink_new(struct slink *next, long int pit)
{
    struct slink *x = (struct slink *) getbytes(sizeof(t_slink));
    x->l_next = next;
    x->l_pit = pit;
    return (x);
}

typedef struct sustain {
    struct object s_ob;
    long s_vel;
    long s_sust;
    void *s_out2;
    struct slink *s_slink;
} t_sustain;

void sustain_int(struct sustain *x, long int n)
{
    if (x->s_vel || !x->s_sust)
        outlet_int(x->s_out2, x->s_vel), outlet_int(x->s_ob.o_outlet, n);
    else
        x->s_slink = slink_new(x->s_slink, n);
}

void sustain_in1(struct sustain *x, long int n)
{
    x->s_vel = n;
}

void sustain_in2(struct sustain *x, long int n)
{
    struct slink *s, *s2;
    if (!(x->s_sust = n))
    {
        for (s = x->s_slink; s; s = s2)
        {
            s2 = s->l_next;
            outlet_int(x->s_out2, 0L),
                outlet_int(x->s_ob.o_outlet, s->l_pit);
            slink_free(s);
        }
        x->s_slink = 0;
    }
}

void sustain_clear(struct sustain *x)
{
    struct slink *s, *s2;
    for (s = x->s_slink; s; s = s2)
    {
        s2 = s->l_next;
        slink_free(s);
    }
    x->s_slink = 0;
}

void sustain_ff(struct sustain *x)
{
    sustain_clear(x);
}

void *sustain_class;
void sustain_setup()
{
    c_new(&s_sustain, sustain_ff, sizeof(t_sustain), 0, 0);
    c_addint(sustain_int);
    c_addmess(sustain_clear, &s_clear, 0);
    c_addfloat(stdfloat);
    c_addinx(sustain_in1, 1);
    c_addinx(sustain_in2, 2);
}

void *sustain_new(long int vel, long int sust)
{
    struct sustain *x =
        (struct sustain *) obj_new((t_messlist **) & sustain_class,
                                   sustain_setup);
    x->s_slink = 0;
    x->s_out2 = outlet(x, &s_int);
    outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    x->s_vel = vel;
    x->s_sust = sust;
    return (x);
}

typedef struct blink {
    long l_shit;
    struct blink *l_next;
    struct blink *l_prev;
} t_blink;

blink_free(t_blink *x)
{
    freebytes(x, sizeof(t_blink));
}

t_blink *blink_new()
{
    return ((t_blink *) getbytes(sizeof(t_blink)));
}

typedef struct bag {
    struct object b_ob;
    long b_in;
    struct blink *b_blink;
    int b_multi;
} t_bag;

void bag_int(struct bag *x, long int n)
{
    struct blink *b;
    if (x->b_in)
    {
        if (!x->b_multi)
        {
            for (b = x->b_blink; b; b = b->l_next)
                if (b->l_shit == n)
                    return;
        }
        b = blink_new();
        b->l_shit = n;
        if (b->l_next = x->b_blink)
            b->l_next->l_prev = b;
        b->l_prev = 0;
        x->b_blink = b;
    }
    else
    {
        for (b = x->b_blink; b; b = b->l_next)
            if (b->l_shit == n)
            {
                if (x->b_blink == b)
                    x->b_blink = b->l_next;
                if (b->l_next)
                    b->l_next->l_prev = b->l_prev;
                if (b->l_prev)
                    b->l_prev->l_next = b->l_next;
                blink_free(b);
                return;
            }
    }
}

void bag_cut(struct bag *x)
{
    struct blink *b;
    long shit;

    if (!x->b_blink)
        return;

    for (b = x->b_blink; b->l_next; b = b->l_next);

    shit = b->l_shit;
    if (x->b_blink == b)
        x->b_blink = b->l_next;
    if (b->l_next)
        b->l_next->l_prev = b->l_prev;
    if (b->l_prev)
        b->l_prev->l_next = b->l_next;
    blink_free(b);
    outlet_int(x->b_ob.o_outlet, shit);
    return;
}

void bag_in1(struct bag *x, long int n)
{
    x->b_in = n;
}

void bag_bang(struct bag *x)
{
    struct blink *b;
    for (b = x->b_blink; b; b = b->l_next)
        outlet_int(x->b_ob.o_outlet, b->l_shit);
}

void bag_clear(struct bag *x)
{
    struct blink *b;
    for (b = x->b_blink; b; b = b->l_next)
        blink_free(b);
    x->b_blink = 0;
}

void bag_ff(struct bag *x)
{
    bag_clear(x);
}

void *bag_class;
void bag_setup()
{
    c_new(&s_bag, bag_ff, sizeof(t_bag), 0, 0);
    c_addint(bag_int);
    c_addmess(bag_clear, &s_clear, 0);
    c_addmess(bag_bang, &s_bang, 0);
    c_addmess(bag_cut, &s_cut, 0);
    c_addfloat(stdfloat);
    c_addinx(bag_in1, 1);
}

void *bag_new(struct symbol *s)
{
    struct bag *x =
        (struct bag *) obj_new((t_messlist **) & bag_class, bag_setup);
    x->b_blink = 0;
    outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in1);
    x->b_multi = (s != &s_);
    x->b_in = 0;
    return (x);
}
