#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:04:09 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:04:09 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
extern struct symbol s__plus, s__minus, s__times, s__div,
    s__gt, s__ge, s__lt, s__le, s__ee, s__ne, s__ba, s__bo, s__la, s__lo,
    s__pc, s__ls, s__rs, s_binop1, s_binop2, s_logic, s_shift, s_set,
    s_in1, s_ft1, s_change;
void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);

/* the following objects do numeric low-level things. */

/* "change" (i.e. register).  Sends an output when its input changes. */
typedef struct change {
    struct object r_ob;
    long r_state;
    void *r_outlet;
} t_change;


void change_int(t_change *x, long int n)
{
    if (n != x->r_state)
    {
        outlet_int(x->r_outlet, n);
        x->r_state = n;
    }
}

void change_set(t_change *x, long int n)
{
    x->r_state = n;
}

void c_change()
{
    c_new(&s_change, NIL, sizeof(t_change), 0, 0);
    c_addint(change_int);
    c_addmess(change_set, &s_set, A_LONG, 0);
}

void *change_messlist;

void *change_new(long int n)
{
    t_change *x =
        (t_change *) obj_new((t_messlist **) & change_messlist, c_change);
    x->r_outlet = outlet(x, &s_int);
    x->r_state = n;
    return (x);
}

struct messlist *plus_messlist, *minus_messlist, *times_messlist,
    *div_messlist, *gt_messlist, *ge_messlist, *lt_messlist, *le_messlist,
    *ee_messlist, *ne_messlist, *la_messlist, *lo_messlist, *ba_messlist,
    *bo_messlist, *pc_messlist, *ls_messlist, *rs_messlist;
struct messlist *fplus_messlist, *fminus_messlist, *ftimes_messlist,
    *fdiv_messlist, *fgt_messlist, *fge_messlist, *flt_messlist,
    *fle_messlist, *fee_messlist, *fne_messlist;

typedef struct ibinop           /* fixed-point binop */
{
    struct object i_ob;
    long i_n1;
    long i_n2;
} t_ibinop;

void ibinop_set(struct ibinop *x, long int n)
{
    x->i_n1 = n;
}

void ibinop_in1(struct ibinop *x, long int n)
{
    x->i_n2 = n;
}

typedef struct fbinop           /* floating-point */
{
    struct object f_ob;
    float f_x1;
    float f_x2;
} t_fbinop;

void fbinop_set(struct fbinop *x, double f)
{
    x->f_x1 = f;
}

void fbinop_ft1(struct fbinop *x, double f)
{
    x->f_x2 = f;
}

  /* return a binary operation (new object.) Checks for
     floating-point arguments and if given them, returns a
     floating-point version.  Needs the messlists for the fixed
     and floating-point versions in fixmess and floatmess 
     (0 if none).  If floatout is set, the outlet follows the
     type; otherwise it's always fixed-point (as in ==.) */
void *binop(struct symbol *s, int argc, struct atom *argv, void *fixmess,
            t_method fixcl, void *floatmess, t_method floatcl,
            int floatout)
{
    struct atom a;
    if (argc)
        a = *argv;
    else
        a.a_type = 0;
    if (a.a_type == A_FLOAT)
    {
        /* it's a floatint-point operation */
        struct fbinop *x;
        if (!floatmess)
        {
            error("%s: fixed-point only", s->s_name);
            return (0);
        }
        x = (struct fbinop *) obj_new((t_messlist **) floatmess, floatcl);
        x->f_x1 = 0;
        x->f_x2 = (a.a_type == A_FLOAT ? a.a_w.w_float :
                   (a.a_type == A_LONG ? a.a_w.w_long : 0));
        outlet(x, floatout ? &s_float : &s_int);
        inlet(x, x, &s_float, &s_ft1);
        return (x);
    }
    else
    {
        struct ibinop *x =
            (struct ibinop *) obj_new((t_messlist **) fixmess, fixcl);
        x->i_n1 = 0;
        x->i_n2 = (a.a_type == A_LONG ? a.a_w.w_long : 0);
        outlet(x, &s_int);
        inlet(x, x, &s_int, &s_in1);
        return (x);
    }
}

ibinop_cl(t_method fbang, t_method fint, t_method ffloat)
{
    c_new(&s_binop1, NIL, sizeof(t_ibinop), 0, 0);
    c_addint(fint);
    c_addfloat(ffloat);
    c_addbang(fbang);
    c_addinx(ibinop_in1, 1);
    c_addmess(ibinop_set, &s_set, A_LONG, 0);
}

fbinop_cl(t_method fbang, t_method fint, t_method ffloat)
{
    c_new(&s_binop1, NIL, sizeof(t_fbinop), 0, 0);
    c_addint(fint);
    c_addfloat(ffloat);
    c_addbang(fbang);
    c_addftx(fbinop_ft1, 1);
    c_addmess(fbinop_set, &s_set, A_FLOAT, 0);
}

/* "-." subtraction. */
#define BINOP1(bif, iif, fif, cif, iml, bff, iff, fff, cff, fml, fn, op) \
void *iml, *fml; \
void bif(x) struct ibinop *x; \
  {outlet_int(x->i_ob.o_outlet, x->i_n1 op x->i_n2);} \
void iif(x, n) struct ibinop *x; long n; \
  {outlet_int(x->i_ob.o_outlet, (x->i_n1 = n) op x->i_n2);} \
void fif(x, f) struct ibinop *x; double f; \
  {outlet_int(x->i_ob.o_outlet, (x->i_n1 = f) op x->i_n2);} \
void cif() {ibinop_cl(bif, iif, fif); } \
void bff(x) struct fbinop *x; \
  {outlet_float(x->f_ob.o_outlet, x->f_x1 op x->f_x2);} \
void iff(x, n) struct fbinop *x; long n; \
  {outlet_float(x->f_ob.o_outlet, (x->f_x1 = n) op x->f_x2);} \
void fff(x, f) struct fbinop *x; double f; \
{outlet_float(x->f_ob.o_outlet, (x->f_x1 = f) op x->f_x2);} \
void cff() {fbinop_cl(bff, iff, fff); } \
void *fn(s, argc, argv) struct symbol *s; struct atom *argv; \
{return (binop(s, argc, argv, &iml, cif, &fml, cff, 1));}

BINOP1(plus_bang, plus_int, plus_float, plus_cl, plus_ml,
       fplus_bang, fplus_int, fplus_float, fplus_cl, fplus_ml, plus_new,
       +)BINOP1(minus_bang, minus_int, minus_float, minus_cl, minus_ml,
                fminus_bang, fminus_int, fminus_float, fminus_cl,
                fminus_ml, minus_new, -)BINOP1(times_bang, times_int,
                                               times_float, times_cl,
                                               times_ml, ftimes_bang,
                                               ftimes_int, ftimes_float,
                                               ftimes_cl, ftimes_ml,
                                               times_new,
                                               *)BINOP1(div_bang, div_int,
                                                        div_float, div_cl,
                                                        div_ml, fdiv_bang,
                                                        fdiv_int,
                                                        fdiv_float,
                                                        fdiv_cl, fdiv_ml,
                                                        div_new, /)
/* The relational binops can be fix or float 
  but their outlets are always fixed-point.*/
#define BINOP2(bif, iif, fif, cif, iml, bff, iff, fff, cff, fml, fn, op) \
void *iml, *fml; \
void bif(x) struct ibinop *x; \
  {outlet_int(x->i_ob.o_outlet, (long)(x->i_n1 op x->i_n2));} \
void iif(x, n) struct ibinop *x; long n; \
  {outlet_int(x->i_ob.o_outlet, (long)((x->i_n1 = n) op x->i_n2));} \
void fif(x, f) struct ibinop *x; double f; \
  {outlet_int(x->i_ob.o_outlet, (long)((x->i_n1 = f) op x->i_n2));} \
void cif() {ibinop_cl(bif, iif, fif); } \
void bff(x) struct fbinop *x; \
  {outlet_float(x->f_ob.o_outlet, (long)(x->f_x1 op x->f_x2));} \
void iff(x, n) struct fbinop *x; long n; \
  {outlet_float(x->f_ob.o_outlet, (long)((x->f_x1 = n) op x->f_x2));} \
void fff(x, f) struct fbinop *x; double f; \
{outlet_int(x->f_ob.o_outlet, (long)((x->f_x1 = f) op x->f_x2));} \
void cff() {fbinop_cl(bff, iff, fff); } \
void *fn(s, argc, argv) struct symbol *s; struct atom *argv; \
{return (binop(s, argc, argv, &iml, cif, &fml, cff, 0));}
    BINOP2(gt_bang, gt_int, gt_float, gt_cl, gt_ml,
       fgt_bang, fgt_int, fgt_float, fgt_cl, fgt_ml, gt_new,
       >)BINOP2(ge_bang, ge_int, ge_float, ge_cl, ge_ml, fge_bang, fge_int,
                fge_float, fge_cl, fge_ml, ge_new, >=)BINOP2(lt_bang,
                                                             lt_int,
                                                             lt_float,
                                                             lt_cl, lt_ml,
                                                             flt_bang,
                                                             flt_int,
                                                             flt_float,
                                                             flt_cl,
                                                             flt_ml,
                                                             lt_new,
                                                             <)BINOP2
    (le_bang, le_int, le_float, le_cl, le_ml, fle_bang, fle_int, fle_float,
     fle_cl, fle_ml, le_new, <=)BINOP2(ee_bang, ee_int, ee_float, ee_cl,
                                       ee_ml, fee_bang, fee_int, fee_float,
                                       fee_cl, fee_ml, ee_new,
                                       ==) BINOP2(ne_bang, ne_int,
                                                  ne_float, ne_cl, ne_ml,
                                                  fne_bang, fne_int,
                                                  fne_float, fne_cl,
                                                  fne_ml, ne_new, !=)
/* logical and bitwise operators.  Fixed-point in and out only. */
#define BINOP3(bif, iif, fif, cif, iml, fn, op) \
void *iml; \
void bif(x) struct ibinop *x; \
  {outlet_int(x->i_ob.o_outlet, (long)(x->i_n1 op x->i_n2));} \
void iif(x, n) struct ibinop *x; long n; \
  {outlet_int(x->i_ob.o_outlet, (long)((x->i_n1 = n) op x->i_n2));} \
void fif(x, f) struct ibinop *x; double f; \
  {outlet_int(x->i_ob.o_outlet, (long)((x->i_n1 = f) op x->i_n2));} \
void cif() {ibinop_cl(bif, iif, fif); } \
void *fn(s, argc, argv) struct symbol *s; struct atom *argv; \
{return (binop(s, argc, argv, &iml, cif, NIL, NIL, 0));}
    BINOP3(ba_bang, ba_int, ba_float, ba_cl, ba_ml, ba_new,
       &)BINOP3(bo_bang, bo_int, bo_float, bo_cl, bo_ml, bo_new,
                |)BINOP3(la_bang, la_int, la_float, la_cl, la_ml, la_new,
                         &&)BINOP3(lo_bang, lo_int, lo_float, lo_cl, lo_ml,
                                   lo_new, ||)BINOP3(pc_bang, pc_int,
                                                     pc_float, pc_cl,
                                                     pc_ml, pc_new,
                                                     %)BINOP3(ls_bang,
                                                              ls_int,
                                                              ls_float,
                                                              ls_cl, ls_ml,
                                                              ls_new,
                                                              <<)BINOP3
    (rs_bang, rs_int, rs_float, rs_cl, rs_ml, rs_new, >>)
