#ifndef lint
static char rcsid[] =
    "$Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:04:16 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:04:16 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
extern t_symbol s_metro, s_speedlim, s_gate, s_del,
    s_stop, s_in1, s_random;


/* "metro". Act like a metronome. */
typedef struct metro {
    t_object m_ob;
    long m_n;
    long m_run;
    void *m_clock;
} t_metro;

void metro_tick(t_metro *x)
{
    outlet_bang(x->m_ob.o_outlet);
    if (x->m_run)
        clock_set(x->m_clock, logicaltime + x->m_n);
}

void metro_bang(t_metro *x)
{
    x->m_run = 1;
    clock_set(x->m_clock, logicaltime);
}

void metro_stop(t_metro *x)
{
    x->m_run = 0;
    clock_unset(x->m_clock);
}

void metro_int(t_metro *x, long int n)
{
    if (x->m_run = n)
        clock_set(x->m_clock, logicaltime);
    else
        clock_unset(x->m_clock);
}

void metro_in1(t_metro *x, long int n)
{
    if (n <= 0)
        n = 5;
    x->m_n = n;
}

void metro_ff(t_metro *x)
{
    clock_free(x->m_clock);
}

void *metro_class;
void metro_setup()
{
    c_new(&s_metro, metro_ff, sizeof(t_metro), 0, 0);
    c_addbang(metro_bang);
    c_addint(metro_int);
    c_addfloat(stdfloat);
    c_addinx(metro_in1, 1);
    c_addmess(metro_stop, &s_stop, 0);
}

void *metro_new(long int n)
{
    t_metro *x =
        (t_metro *) obj_new((t_messlist **) & metro_class, metro_setup);
    outlet_new(x, &s_bang);
    inlet_new(x, x, &s_int, &s_in1);
    x->m_clock = clock_new(x, metro_tick);
    if (n <= 0)
        n = 5;
    x->m_n = n;
    return (x);
}

/* "speedlim". limit speed of a number */
typedef struct speedlim {
    t_object s_ob;
    long s_n;
    long s_time;
    long s_set;
    long s_lim;
    void *s_outlet;
    void *s_clock;
} t_speedlim;

void speedlim_int(t_speedlim *x, long int n)
{
    x->s_n = n;
    if (logicaltime >= x->s_time)
    {
        outlet_int(x->s_outlet, n);
        x->s_time = logicaltime + x->s_lim;
    }
    else
    {
        if (!x->s_set)
        {
            clock_set(x->s_clock, x->s_time);
            x->s_set = 1;
        }
    }
}

void speedlim_float(t_speedlim *x, double f)
{
    speedlim_int(x, (long) f);
}

void speedlim_in1(t_speedlim *x, long int n)
{
    x->s_lim = n;
}

void speedlim_tick(t_speedlim *x)
{
    outlet_int(x->s_outlet, x->s_n);
    x->s_time = logicaltime + x->s_lim;
    x->s_set = 0;
}

void speedlim_ff(t_speedlim *x)
{
    clock_free(x->s_clock);
}

void speedlim_setup()
{
    c_new(&s_speedlim, speedlim_ff, sizeof(t_speedlim), 0, 0);
    c_addint(speedlim_int);
    c_addfloat(speedlim_float);
    c_addinx(speedlim_in1, 1);
}

void *speedlim_class;

void *speedlim_new(long int n)
{
    t_speedlim *x =
        (t_speedlim *) obj_new((t_messlist **) & speedlim_class,
                               speedlim_setup);
    x->s_outlet = outlet_new(x, &s_int);
    x->s_clock = clock_new(x, speedlim_tick);
    x->s_lim = n;
    x->s_set = 0;
    x->s_time = 0;
    inlet_new(x, x, &s_int, &s_in1);
    return (x);
}

/* "gate". */
typedef struct gate {
    t_object g_ob;
    void *g_outlet;
    void *g_inlet;
} t_gate;

void gate_int(t_gate *x, long int n)
{
    inlet_to(x->g_inlet, n ? x->g_outlet : 0);
}

void gate_setup()
{
    c_new(&s_gate, NIL, sizeof(t_gate), 0, 0);
    c_addint(gate_int);
    c_addfloat(stdfloat);
}

void *gate_class;

void *gate_new()
{
    t_gate *x =
        (t_gate *) obj_new((t_messlist **) & gate_class, gate_setup);
    x->g_outlet = outlet_new(x, NIL);
    x->g_inlet = inlet_new(x, NIL, NIL, NIL);
    return (x);
}

/* "del". */

typedef struct del {
    t_object d_ob;
    void *d_outlet;
    void *d_clock;
    long d_del;
} t_del;

void del_in1(t_del *x, long int n)
{
    if (n < 0)
        n = 0;
    x->d_del = n;
}

void del_bang(t_del *x)
{
    clock_set(x->d_clock, logicaltime + x->d_del);
}

void del_stop(t_del *x)
{
    clock_unset(x->d_clock);
}

void del_tick(t_del *x)
{
    outlet_bang(x->d_outlet);
}

void del_ff(t_del *x)
{
    clock_free(x->d_clock);
}

void del_setup()
{
    c_new(&s_del, del_ff, sizeof(t_del), 0, 0);
    c_addbang(del_bang);
    c_addinx(del_in1, 1);
    c_addmess(del_stop, &s_stop, 0);
}

void *del_class;

void *del_new(long int n)
{
    t_del *x = (t_del *) obj_new((t_messlist **) & del_class, del_setup);
    x->d_del = (n > 0 ? n : 0);
    x->d_clock = clock_new(x, del_tick);
    x->d_outlet = outlet_new(x, &s_bang);
    inlet_new(x, x, &s_int, &s_in1);
    return (x);
}

typedef struct random {
    t_object r_ob;
    long r_n;
} t_random;

void random_bang(t_random *x)
{
    outlet_int(x->r_ob.o_outlet, (x->r_n * (rand() & 0x7fffL)) >> 15);
}

void random_in1(t_random *x, long int n)
{
    if (n <= 1)
        n = 1;
    x->r_n = n;
}

void random_setup()
{
    c_new(&s_random, NIL, sizeof(t_random), 0, 0);
    c_addbang(random_bang);
    c_addinx(random_in1, 1);
}

void *random_class;

void *random_new(long int n)
{
    t_random *x =
        (t_random *) obj_new((t_messlist **) & random_class, random_setup);
    outlet_new(x, &s_int);
    inlet_new(x, x, &s_int, &s_in1);
    if (n < 1)
        n = 1;
    x->r_n = n;
    return (x);
}
