#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1993/11/19 10:51:58 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1993/11/19 10:51:58 $
 *
 * FTS by Miller Puckette
 *
 */


/* objects by Cort Lippe */

#include "m_mess.h"
extern t_symbol s_line, s_timer, s_stop, s_in1, s_in2, s_set;

/* "line". Act like some sort of breakpoint luni-type  guy. */
typedef struct line {
    t_object l_ob;
    void *l_outlet;
    void *l_clock;
    long l_target;
    long l_grain;
    long l_cur;
    long l_steps;
    long l_inc;
    long l_inval;               /* last value sent to inlet */
} t_line;

void line_stop(t_line *x)       /* STOP clock */
{
    x->l_steps = 0;
    x->l_target = x->l_cur >> 10;
    clock_unset(x->l_clock);
}

void line_tick(t_line *x)
{
    if (x->l_steps)
        x->l_steps = x->l_steps - 1;
    if (x->l_steps)
    {
        x->l_cur = x->l_cur + x->l_inc;
        clock_set(x->l_clock, logicaltime + x->l_grain);
        outlet_int(x->l_outlet, x->l_cur >> 10);
    }
    else
        outlet_int(x->l_outlet, x->l_target);
}

void line_fixfix(t_line *x, long int target, long int time)     /* START clock */
/* start the calculation up here */
{
    long old_target, distance;
    if (x->l_steps)
        old_target = x->l_cur >> 10;
    else
        old_target = x->l_target;
    distance = target - old_target;
    if (time < 1)
        time = 1;
    x->l_steps = ((time - 1) / x->l_grain) + 1;
    x->l_inc = (distance * 1024) / x->l_steps;
    x->l_cur = old_target * 1024;
    x->l_target = target;
    x->l_inval = 0;
    line_tick(x);
}

void line_int(t_line *x, long int target)       /* jump to given target val and STOP clock */
{
    /* if some dweezul has talked to middle inlet, do fixfix */
    if (x->l_inval)
        line_fixfix(x, target, x->l_inval);
    else
    {
        x->l_target = target;
        x->l_steps = 0;
        clock_unset(x->l_clock);
        outlet_int(x->l_outlet, x->l_target);
    }
}

void line_set(t_line *x, long int target)       /* jump to given target val and STOP clock */
{
    x->l_target = target;
    x->l_steps = 0;
    clock_unset(x->l_clock);
}

void line_float(t_line *x, double f)
{
    line_int(x, (long) f);
}

void line_in2(t_line *x, long int grain)        /* fix message into rite corner */
{
    if (grain < 1)
        grain = 20;
    x->l_grain = grain;
}

void line_in1(t_line *x, long int n)    /* fix message into middle corner */
{
    if (n < 0)
        n = 0;
    x->l_inval = n;
}

void line_ff(t_line *x)
{
    clock_free(x->l_clock);
}

void *line_class;
void line_setup()
{
    c_new(&s_line, line_ff, sizeof(t_line), 0, 0);
    c_addint(line_int);
    c_addmess(line_stop, &s_stop, 0);
    c_addmess(line_set, &s_set, A_LONG, 0);
    c_addfloat(line_float);
    c_addinx(line_in1, 1);
    c_addinx(line_in2, 2);
}

void *line_new(long int target, long int grain) /* init vals in struc */
{
    t_line *x =
        (t_line *) obj_new((t_messlist **) & line_class, line_setup);
    x->l_outlet = outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    x->l_clock = clock_new(x, line_tick);
    x->l_target = target;
    x->l_steps = x->l_inval = 0;
    if (grain < 1)
        grain = 20;
    x->l_grain = grain;
    return (x);
}

/***************************************************************************/

/* "timer"  bang to save logicaltime &
   bang rite corner to send elapsed time message */

typedef struct timer {
    t_object t_ob;
    long t_oldtime;
    void *t_outlet;
    void *t_inlet;
} t_timer;

void timer_bang(t_timer *x)     /* store time-now on default bang message */
{
    x->t_oldtime = logicaltime;
}

void timer_in1(t_timer *x)      /* bang message into rite corner sends fix out  */
{
    outlet_int(x->t_outlet, logicaltime - x->t_oldtime);
}

void *timer_class;
void timer_setup()
{
    c_new(&s_timer, NIL, sizeof(t_timer), 0, 0);
    c_addbang(timer_bang);
    c_addmess(timer_in1, &s_in1, 0);
}

void *timer_new()               /* init vals in struc */
{
    t_timer *x =
        (t_timer *) obj_new((t_messlist **) & timer_class, timer_setup);
    x->t_outlet = outlet(x, &s_int);
    x->t_inlet = inlet(x, x, &s_bang, &s_in1);
    x->t_oldtime = logicaltime;
    return (x);
}
