#ifndef lint
static char rcsid[] =
    "$Revision: 1.9 $ IRCAM $Date: 1994/01/04 18:04:24 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.9 $ IRCAM $Date: 1994/01/04 18:04:24 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
#include "cp_mess.h"
#include "t_mac.h"
#include "p_patcher.h"

void *cpr_new();
extern t_symbol s_openpanel, s_savepanel;

typedef struct _panel {
    struct object p_ob;
    int p_write;                /* true for write; false for read */
#ifdef GRAPHICS
    void *p_cpr;                /* CP responder */
#else
    void *p_hq;
#endif
} t_panel;

    /* code to flush a panel up from CP to host */
void panel_bang(x)
t_panel *x;
{
#ifndef GRAPHICS
    if (!x->p_hq)
        bug("panel_bang");
    else
        hq_set(x->p_hq);
#endif
}

#ifndef GRAPHICS
void panel_hqfn(x, hq)
t_panel *x;
void *hq;
{
    char c = 0;
    hq_doit(hq, &c, 1);
}

void panel_connect(x, ptr, fun, freefun)
t_panel *x;
long ptr, fun, freefun;
{
    if (ptr)
    {
        if (x->p_hq)
            bug("panel_connect 1");
        x->p_hq = hq_new(fun, freefun, ptr, x, panel_hqfn, 1, currentPort);
    }
    else
    {
        if (!x->p_hq)
            bug("panel_connect 2");
        else
            hq_free(x->p_hq);
        x->p_hq = 0;
    }
}

void panel_symbol(t_panel *x, t_symbol *s)
{
    outlet_symbol(x->p_ob.o_outlet, s);
}

#endif                          /* GRAPHICS */

/* CP->host update, host side */
#ifdef GRAPHICS
#ifdef NEXTSTEP
t_symbol *t_openpanel(), *t_savepanel();

void panel_cpr(x, buf, size)
t_panel *x;
char *buf;
{
    t_symbol *s = (x->p_write ? t_savepanel() : t_openpanel());
    if (s)
        h_vmess(x, &s_symbol, "s", s);
}
#endif                          /* NEXTSTEP */

#ifdef __unix
void panel_open(t_method f, void *arg);
void panel_save(t_method f, void *arg);
void panel_cancel(void);

static void panel_gotit(t_panel *x, char *s)
{
    h_vmess(x, &s_symbol, "s", gensym(s));
}

void panel_cpr(x, buf, size)
t_panel *x;
char *buf;
{
    if (x->p_write)
        panel_save(panel_gotit, x);
    else
        panel_open(panel_gotit, x);
}
#endif                          /* __unix */

void panel_setcp(x, n, flags)
t_panel *x;
{
    if (flags & CP_UNPUT)
    {
        if (x->p_cpr)
        {
            cpr_free(x->p_cpr);
            x->p_cpr = 0;
        }
    }
    if (flags & CP_PUT)
    {
        x->p_cpr = cpr_new(x, panel_cpr);
    }
}

#endif                          /* GRAPHICS */

void panel_ff(t_panel *x)
{
#ifndef GRAPHICS
    if (x->p_hq)
        hq_free(x->p_hq);
#endif
}

void *openpanel_class;
void openpanel_setup()
{
    c_new(&s_openpanel, panel_ff, sizeof(t_panel), 0, 0);
#ifdef GRAPHICS
    c_addmess(panel_setcp, gensym("setcp"), A_CANT, 0);
#else
    c_addsymbol(panel_symbol);
    c_addmess(panel_connect, gensym("connect"), A_LONG, A_LONG, A_LONG, 0);
#endif
    c_addbang(panel_bang);
}

void *openpanel_new()
{
    t_panel *x = (t_panel *) obj_new((t_messlist **) & openpanel_class,
                                     openpanel_setup);
#ifdef GRAPHICS
    x->p_cpr = 0;
#else
    x->p_hq = 0;
#endif
    x->p_write = 0;
    outlet_new(x, &s_symbol);
    return (x);
}

void *savepanel_class;
void savepanel_setup()
{
    c_new(&s_savepanel, panel_ff, sizeof(t_panel), 0, 0);
#ifdef GRAPHICS
    c_addmess(panel_setcp, gensym("setcp"), A_CANT, 0);
#else
    c_addsymbol(panel_symbol);
    c_addmess(panel_connect, gensym("connect"), A_LONG, A_LONG, A_LONG, 0);
#endif
    c_addbang(panel_bang);
}

void *savepanel_new()
{
    t_panel *x = (t_panel *) obj_new((t_messlist **) & savepanel_class,
                                     savepanel_setup);
#ifdef GRAPHICS
    x->p_cpr = 0;
#else
    x->p_hq = 0;
#endif
    x->p_write = 1;
    outlet_new(x, &s_symbol);
    return (x);
}
