#ifndef lint
static char rcsid[] =
    "$Revision: 2.9 $ IRCAM $Date: 1994/01/04 18:04:29 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.9 $ IRCAM $Date: 1994/01/04 18:04:29 $
 *
 * FTS by Miller Puckette
 *
 */


/*
  these objects are in support of Manoury's pieces until we can make them
  external */

#include "m_mess.h"
#define __STRICT_BSD__
#include <math.h>
#undef __STRICT_BSD__


extern t_symbol s_pafctl, s_in1, s_ft2, s_ft3, s_ft4, s_ft5;

/* "pafctl".  inlet chained straight to outlet */

typedef struct pafctl {
    t_object p_ob;
    void *p_ampout;             /* amplitude */
    void *p_ampout2;            /* amplitude correction */
    void *p_freqout;            /* freq */
    void *p_cfout;              /* center freq */
    void *p_bwout;              /* bandwidth */
    void *p_vibout;             /* vibrato depth */
    float p_amp;
    float p_freq;
    float p_cf;
    float p_bw;
    float p_vib;
    long p_amptime;
} t_pafctl;


void pafctl__put(void *z, double f, long int n)
{
    t_atom at[2];
    at[0].a_type = A_FLOAT;
    at[0].a_w.w_float = f;
    at[1].a_type = A_LONG;
    at[1].a_w.w_long = n;
    outlet_list(z, &s_list, 2, at);
}

void pafctl_float(t_pafctl *x, double f)
{
    x->p_amp = (f > 1 ? .05 * exp(.06623 * (f - 127.)) : 0.);
    pafctl__put(x->p_ampout, x->p_amp, x->p_amptime);
    /* outlet_float(x->p_ampout, x->p_amp); */
}

void pafctl_int(t_pafctl *x, long int n)
{
    pafctl_float(x, (double) n);
}

void pafctl_bwf(t_pafctl *x)
{
    float quot = x->p_bw / x->p_freq, sq;
    if (quot <= .01)
        quot = .01;
    sq = sqrt(quot);
    outlet_float(x->p_bwout, quot);
    pafctl__put(x->p_ampout2, 1. + (sq > 16 ? 16 : sq), 30);
}

void pafctl_in1(t_pafctl *x, int n)     /* amp ramp time */
{
    x->p_amptime = n;
}

void pafctl_ft2(t_pafctl *x, double f)  /* base freq */
{
    x->p_freq = 440. * exp(.00057762265 * (f - 6900.));
    if (x->p_freq < .1)
        x->p_freq = .1;
    pafctl__put(x->p_cfout, x->p_cf / x->p_freq, 30);
    pafctl_bwf(x);
    outlet_float(x->p_vibout, (.5 / 256.) * x->p_vib * x->p_freq);
    outlet_float(x->p_freqout, .5 * x->p_freq);
}

void pafctl_ft3(t_pafctl *x, double f)  /* bandwidth */
{
    x->p_bw = (f >= 0 ? (440. * exp(.057762265 * (f - 69.))) : 0.);
    pafctl_bwf(x);
}

void pafctl_ft4(t_pafctl *x, double f)  /* center freq */
{
    x->p_cf = (f >= 0 ? (440. * exp(.057762265 * (f - 69.))) : 0.);
    pafctl__put(x->p_cfout, x->p_cf / x->p_freq, 30);
}

void pafctl_ft5(t_pafctl *x, double f)  /* vibrato depth */
{
    x->p_vib = (f * (1. / 127.));
    outlet_float(x->p_vibout, x->p_vib * x->p_freq);
}

void *pafctl_class;
void pafctl_setup()
{
    c_new(&s_pafctl, NIL, sizeof(t_pafctl), 0, 0);
    c_addint(pafctl_int);
    c_addfloat(pafctl_float);
    c_addinx(pafctl_in1, 1);
    c_addftx(pafctl_ft2, 2);
    c_addftx(pafctl_ft3, 3);
    c_addftx(pafctl_ft4, 4);
    c_addftx(pafctl_ft5, 5);
}

void *pafctl_new()
{
    t_pafctl *x =
        (t_pafctl *) obj_new((t_messlist **) & pafctl_class, pafctl_setup);
    x->p_vibout = outlet(x, &s_float);
    x->p_bwout = outlet(x, NIL);
    x->p_cfout = outlet(x, NIL);
    x->p_freqout = outlet(x, &s_float);
    x->p_ampout2 = outlet(x, NIL);
    x->p_ampout = outlet(x, NIL);
    inlet(x, x, &s_float, &s_ft5);
    inlet(x, x, &s_float, &s_ft4);
    inlet(x, x, &s_float, &s_ft3);
    inlet(x, x, &s_float, &s_ft2);
    inlet(x, x, &s_int, &s_in1);
    x->p_freq = 1.;
    x->p_cf = 1.;
    x->p_bw = 1.;
    x->p_amptime = 30;
    return (x);
}

/*************** mark2 **************************/
extern t_symbol s_mark2, s_in1;

#define DIM 88
#define GOO 8
#define COLS (DIM + GOO)
#define FIRST 21
#define LAST (FIRST + DIM)

struct trans {
    unsigned char t_vel;
    unsigned char t_length;
    unsigned short t_inter;
};
struct hang {
    struct trans *h_trans;
    long h_time;
};
typedef struct mark2 {
    struct object m_ob;
    struct trans **m_matrix;
    int m_recpit;
    int m_restart;
    long m_rectime;
    struct hang m_hang[DIM];
    char m_goophase[DIM];
    struct trans m_firsttrans;
    int m_firstpit;
    int m_playpit;
    struct trans *m_playtrans;

    void *m_clock;
    void *m_velout;
    void *m_durout;
    long m_val;
    long m_maxdiff;
    long m_record;
    long m_inval;
    long m_repeat;
} t_mark2;

long state;
randi()
{
    state = state * 210427233L + 1489352071L;
    return ((state >> 16) & 0x7fff);
}

void mark2_int(struct mark2 *m, long int x1)
{
    long x2;
    int lastrec;
    long timediff;

    x2 = m->m_inval;
    lastrec = m->m_recpit;
    if (!m->m_record)
    {
        return;
    }
    x1 -= FIRST;
    if (!x2 && x1 >= 0 && x1 < DIM)
    {
        struct hang *h = &m->m_hang[x1];
        if (h->h_trans)
        {
            long del = (logicaltime - h->h_time) >> 3;
            if (del > 255)
                del = 255;
            else if (del < 1)
                del = 1;
            h->h_trans->t_length = del;
            h->h_trans = 0;
        }
        return;
    }
    if (x2 && m->m_firstpit < 0)
    {
        m->m_firstpit = x1;
        m->m_firsttrans.t_vel = x2;
        m->m_firsttrans.t_length = 64;
        m->m_hang[x1].h_time = logicaltime;
        m->m_hang[x1].h_trans = &m->m_firsttrans;
    }
    else if (x2 && x1 >= 0 && x1 < DIM && lastrec >= 0 && lastrec < DIM
             && (timediff = logicaltime - m->m_rectime) < m->m_val)
    {
        struct trans *t = &(*m->m_matrix)[lastrec * COLS + x1];
        /* save up to GOO old transitions for repeated notes */
        if (x1 == lastrec)
        {
            int phase = m->m_goophase[lastrec];
            struct trans *t2 = &(*m->m_matrix)[lastrec * COLS +
                                               DIM + phase];
            *t2 = *t;
            m->m_goophase[lastrec] = (phase + 1) & (GOO - 1);
        }
        t->t_vel = x2;
        t->t_length = 64;
        if (timediff < m->m_maxdiff)
            timediff = 0;
        t->t_inter = timediff;
        m->m_hang[x1].h_time = logicaltime;
        m->m_hang[x1].h_trans = t;
    }
    if (x2)
    {
        m->m_recpit = x1;
        m->m_rectime = logicaltime;
    }
}

void mark2_print(struct mark2 *m)
{
    post("record %ld time %ld", (long) m->m_recpit, m->m_rectime);
    post("val %ld", m->m_val);
    post("maxdiff %ld", m->m_maxdiff);
}

void mark2_clear(struct mark2 *m)
{
    int i, j;
    struct trans *tp;

    m->m_record = 0;
    for (i = 0; i < DIM; i++)
        for (j = COLS, tp = (*m->m_matrix) + i * COLS; j--; tp++)
            tp->t_vel = 0;
    for (i = 0; i < DIM; i++)
        m->m_hang[i].h_trans = 0, m->m_goophase[i] = 0;
    m->m_playtrans = 0;
    m->m_playpit = 0;
    m->m_recpit = -1;
    m->m_firstpit = -1;
    clock_unset(m->m_clock);
}

void mark2_record(struct mark2 *m)
{
    m->m_record = 1;
}

void mark2_stoprec(struct mark2 *m)
{
    m->m_record = 0;
}

void mark2_bang(struct mark2 *m)
{
    if (m->m_firstpit < 0)
    {
        post("mark2: no sequence");
        return;
    }
    m->m_playpit = m->m_firstpit;
    m->m_playtrans = &m->m_firsttrans;
    clock_set(m->m_clock, logicaltime);
}

void mark2_tick(struct mark2 *m)
{
    long x, r, index;
    int i, row, lastplay, ncolumn;
    struct trans *tp;

    x = 0;
    r = randi();
    ncolumn = (m->m_repeat ? COLS : DIM);
    outlet_int(m->m_durout, (m->m_playtrans->t_length & 0xffL) << 3);
    outlet_int(m->m_velout, m->m_playtrans->t_vel & 0xffL);
    outlet_int(m->m_ob.o_outlet, (long) m->m_playpit + FIRST);
    row = m->m_playpit;
    for (tp = (*m->m_matrix) + COLS * row, i = ncolumn; i--; tp++)
        x += tp->t_vel;
    /* post("%ld->%ld", (long)row, x); */
    if (!x)                     /* if nowhere to go stop */
    {
        if (m->m_restart)
        {
            if (m->m_firstpit < 0)
            {
                post("mark2: no sequence");
                return;
            }
            m->m_playpit = m->m_firstpit;
            m->m_playtrans = &m->m_firsttrans;
            clock_set(m->m_clock, logicaltime + 500L);
        }
        return;
    }
    index = (r * x) >> 15;
    if (!index)
        index++;
    for (tp = (*m->m_matrix) + COLS * row, i = 0; i < ncolumn; tp++, i++)
    {
        index -= tp->t_vel;
        if (index <= 0)
            break;
    }
    if (i >= ncolumn)
        post("fuck"), i--;
    m->m_playtrans = (*m->m_matrix) + COLS * row + i;
    m->m_playpit = (i >= DIM ? row : i);
    clock_set(m->m_clock, logicaltime + (long) (m->m_playtrans->t_inter));
}

void mark2_stop(struct mark2 *m)
{
    clock_unset(m->m_clock);
}

void mark2_set(struct mark2 *m, long int n)
{
    m->m_val = n;
}

void mark2_in1(struct mark2 *m, long int n)
{
    m->m_inval = n;
}

void mark2_restart(struct mark2 *m, long int n)
{
    m->m_restart = n;
}

void mark2_r(struct mark2 *m, long int n)
{
    m->m_repeat = n;
}

void mark2_maxdiff(struct mark2 *m, long int n)
{
    m->m_maxdiff = n;
}

void mark2_ff(struct mark2 *m)
{
    clock_free(m->m_clock);
    DisposHandle(m->m_matrix);
}

void *mark2_class;

void mark2_setup()
{
    c_new(&s_mark2, mark2_ff, sizeof(t_mark2), 0, 0);
    c_addbang(mark2_bang);
    c_addint(mark2_int);
    c_addinx(mark2_in1, 1);
    c_addmess(mark2_print, gensym("print"), 0);
    c_addmess(mark2_clear, gensym("clear"), 0);
    c_addmess(mark2_record, gensym("record"), 0);
    c_addmess(mark2_stoprec, gensym("stoprec"), 0);
    c_addmess(mark2_stop, gensym("stop"), 0);
    c_addmess(mark2_set, gensym("set"), A_LONG, 0);
    c_addmess(mark2_maxdiff, gensym("maxdiff"), A_LONG, 0);
    c_addmess(mark2_restart, gensym("restart"), A_LONG, 0);
    c_addmess(mark2_r, gensym("r"), A_LONG, 0);
}

void *mark2_new()
{
    struct mark2 *m;
    int i, j;
    struct trans *tp;

    m = (struct mark2 *) obj_new((t_messlist **) & mark2_class,
                                 mark2_setup);
    if (!
        (m->m_matrix =
         (struct trans **)
         t_newhandle((sizeof(struct trans) * (long) DIM * (long) COLS))))
    {
        post("mark2: not enough memory");
        return (0);
    }
    m->m_clock = clock_new(m, mark2_tick);
    inlet(m, m, &s_int, &s_in1);
    m->m_durout = outlet(m, &s_int);
    m->m_velout = outlet(m, &s_int);
    outlet(m, &s_int);
    m->m_restart = 0;
    mark2_clear(m);
    m->m_val = 2000;
    m->m_maxdiff = 50;
    m->m_inval = 0;
    m->m_repeat = 0;
    return (m);
}

/* "gumbo". send out large numbers of controls. */
extern t_symbol s_gumbo;

#define NCTL 48
#define ADDPITCH 1
#define RMPITCH 2
#define ADDVIB 3
#define RMVIB 4
#define SKEW 11
#define GLISS 12
#define GLISS2 13
#define GLISS3 14
#define INT1 15
#define DECAL 32

typedef struct gumbo {
    struct object g_ob;
    void *g_out2;               /* outlet for frequencies */
    void *g_out1;
    void *g_clock;
    long g_mode;                /* what's happening */
    long g_starttime;           /* arpegio start time */
    long g_msec;                /* arpegio time interval */
    int g_nosc;                 /* number of oscilators */
    int g_count;                /* arpegio count */
    long g_shit[NCTL];          /* pitches */
    long g_newshit[NCTL];       /* new pitches (to gliss to) */
} t_gumbo;

sendit(struct gumbo *x, long int n1, long int n2)
{
    outlet_int(x->g_out2, n2);
    outlet_int(x->g_out1, n1);
}

void gumbo_bang(struct gumbo *x)
{
    long *lp, count;
    int nosc;
    nosc = x->g_nosc;
    for (lp = x->g_shit, count = 0; count < nosc; count++)
    {
        long n = *(lp++);
        if (n)
            sendit(x, count, n);
        else
            sendit(x, count, 0L);
    }
}

void gumbo_tick(struct gumbo *x)
{
    int i;
    int nosc;
    nosc = x->g_nosc;
    if (x->g_mode == SKEW)
    {
        long count = x->g_count;
        long *lp = &x->g_shit[count], n;
        for (; count < nosc; count++, lp++)
        {
            sendit(x, count, 0L);
            if (n = *lp)
            {
                sendit(x, count, n);
                x->g_count = count + 1;
                clock_delay(x->g_clock, x->g_msec);
                break;
            }
        }
    }
    else if (x->g_mode == GLISS)
    {
        long count = x->g_count, t = logicaltime - x->g_starttime;
        long msec = x->g_msec, f1, f2;
        while ((count < nosc) && (t > (count + 1) * msec))
        {
            if (!x->g_shit[count])
                goto nomore;
            sendit(x, count, x->g_newshit[count]);
            x->g_count = (++count);
        }
        if ((count >= nosc) || (!x->g_shit[count]))
            goto nomore;
        while ((count < nosc) && (f1 = x->g_shit[count]))
        {
            if (f2 = x->g_newshit[count])
            {
                long totdel = (count + 1) * msec;
                long freq = f1 + ((f2 - f1) * t) / totdel;
                sendit(x, count, freq);
            }
            else
                sendit(x, count, 0L);
            count++;
        }
        clock_delay(x->g_clock, 25L);
    }
    else if (x->g_mode == GLISS2)
    {
        long count = x->g_count, t = logicaltime - x->g_starttime;
        long msec = x->g_msec, f1, f2;
        while ((count < nosc) && (t > (count + 1) * msec))
        {
            if (!x->g_shit[count])
                goto nomore;
            sendit(x, count, x->g_newshit[count]);
            x->g_count = (++count);
        }
        if ((count >= nosc) || (!x->g_shit[count]))
        {
            for (i = 0; i < x->g_nosc; i++)
                x->g_shit[i] = x->g_newshit[i];
            goto nomore;
        }
        f1 = x->g_shit[count];
        if (f2 = x->g_newshit[count])
        {
            long begdel = count * msec;
            long freq = f1 + ((f2 - f1) * (t - begdel)) / msec;
            sendit(x, count, freq);
        }
        else
            sendit(x, count, 0L);
        clock_delay(x->g_clock, 15L);
    }
    else if (x->g_mode == GLISS3)
    {
        long count, t = logicaltime - x->g_starttime;
        long msec = x->g_msec;
        if (t > msec)
            t = msec;
        for (count = 0; count < nosc; count++)
        {
            long was = x->g_shit[count], is = x->g_newshit[count];
            if (is && was)
                sendit(x, count,
                       was + (t * (x->g_newshit[count] - was)) / msec);
        }
        if (t == msec)
        {
            for (i = 0; i < x->g_nosc; i++)
                x->g_shit[i] = x->g_newshit[i];
            goto nomore;
        }
        clock_delay(x->g_clock, 25L);
    }
  nomore:;
}

void gumbo_skew(struct gumbo *x, long int n)
{
    x->g_msec = n;
    x->g_count = 0;
    x->g_mode = SKEW;
    gumbo_tick(x);
}

/* gumbo_getem(x, x2)
   struct gumbo *x, *x2;
   {
   long n;
   for (n = 0; n < x->g_nosc; n++) x->g_newshit[n] = x2->g_shit[n];
   } */


void gumbo_slide(struct gumbo *x, struct symbol *s, int ac,
                 struct atom *av)
{
    int n;
    n = 0;
    if (ac < 3 || av[0].a_type != A_LONG || av[1].a_type != A_LONG)
    {
        post("gumbo_gliss: usage: gliss [123] <msec> <pitch...>");
        return;
    }
    if (av[0].a_w.w_long == 2)
        x->g_mode = GLISS2;
    else if (av[0].a_w.w_long == 3)
        x->g_mode = GLISS3;
    else
        x->g_mode = GLISS;
    if ((x->g_msec = av[1].a_w.w_long) < 10)
        x->g_msec = 10;
    x->g_count = 0;
    x->g_starttime = logicaltime;
    ac -= 2;
    av += 2;
    while (n < x->g_nosc && ac-- && av->a_type == A_LONG)
        x->g_newshit[n++] = (av++)->a_w.w_long;
    while (n < x->g_nosc)
        x->g_newshit[n++] = 0;
    gumbo_tick(x);
}

void gumbo_stop(struct gumbo *x)
{
    clock_unset(x->g_clock);
    x->g_mode = 0;
}

void gumbo_sort(struct gumbo *x)
{
    int nosc;
    long *lp, *lp2, *ep;
    int count, count2, n;
    nosc = x->g_nosc;
    ep = x->g_shit + nosc;
    n = 0;
    for (lp = lp2 = x->g_shit, count = nosc; count--; lp2++)
        if (*lp = *lp2)
            lp++, n++;
    for (lp2 = x->g_shit + nosc; lp < lp2; lp++)
        *lp = 0;
    for (count = n, lp = x->g_shit; count--; lp++)
    {
        long min = 0x7fffffff, tmp, *mp = 0;
        for (lp2 = lp, count2 = count + 1; count2--; lp2++)
            if (*lp2 < min)
                min = *(mp = lp2);
        if (!mp)
            post("sort bug");
        tmp = *lp;
        *lp = min;
        *mp = tmp;
    }
    for (count = 0; count < nosc; count += 4)
    {
        if (!(count & 15))
            post("");
        post("%d %d %d %d", x->g_shit[count], x->g_shit[count + 1],
             x->g_shit[count + 2], x->g_shit[count + 3]);
    }
}

void gumbo_mode(struct gumbo *x, long int n)
{
    x->g_mode = (n > 10 ? 0 : n);
}

void gumbo_print(struct gumbo *x)
{
    long *lp, count, i, n;

    lp = x->g_shit;
    count = 0;
    switch (x->g_mode)
    {
    case 0:
        post("gumbo: stopped");
        break;
    case ADDPITCH:
        post("gumbo: add pitches");
        break;
    case RMPITCH:
        post("gumbo: clear pitches");
        break;
    case ADDVIB:
        post("gumbo: add vibrato");
        break;
    case RMVIB:
        post("gumbo: clear vibrato");
        break;
    default:
        post("gumbo: other");
        break;
    }
    for (i = x->g_nosc; i--;)
    {
        if (n = *(lp++))
        {
            post("%ld", n);
        }
    }
}

void gumbo_dump(struct gumbo *x)
{
    struct atom at[NCTL + 4], *ap;
    int count, ac;
    long *lp, n;

    ac = 0;
    for (count = x->g_nosc, ap = at, lp = x->g_shit; count--;)
    {
        if (n = *(lp++))
        {
            ap->a_type = A_LONG;
            ap->a_w.w_long = n;
            ap++;
            ac++;
        }
    }
    ap->a_type = A_NOTHING;
    outlet_anything(x->g_out1, gensym("set"), ac, at);
}

void gumbo_set(struct gumbo *x, struct symbol *s, int ac, struct atom *av)
{
    int nosc;
    long *lp;
    int count;

    nosc = x->g_nosc;
    lp = x->g_shit;
    if (ac > nosc)
        ac = nosc;
    for (count = 0; count < ac; count++, av++)
    {
        if (av->a_type != A_LONG)
            post("gumbo: bad input"), *(lp++) = 0;
        else
            *(lp++) = av->a_w.w_long;
    }
    for (; count < nosc; count++)
        *(lp++) = 0;
}

void gumbo_int(struct gumbo *x, long int n)
{
    int nosc;
    long *lp, lt;
    int count, hole;

    hole = nosc = x->g_nosc;
    switch (x->g_mode)
    {
    case ADDPITCH:
        for (count = nosc, lp = x->g_shit + nosc; count--;)
        {
            if ((lt = *(--lp)) == n)
                goto nodice;
            if (!lt)
                hole = count;
        }
        if (hole == nosc)
            post("gumbo: full");
        else
        {
            x->g_shit[hole] = n;
            sendit(x, (long) hole, n);
        }
        break;
    case RMPITCH:
        for (count = nosc, lp = x->g_shit; count--; lp++)
            if (*lp == n)
            {
                *lp = 0;
                sendit(x, (long) (nosc - 1 - count), 0L);
            }
        break;
    case INT1:                  /* unused.  did something like pick out harmonics of
                                   one gumbo that were close to pitches in the other. */
        for (count = 0, lp = x->g_shit; count < nosc; lp++, count++)
        {
            long pit = *lp, *lp2, c2;
            if (!pit)
                continue;
            for (c2 = x->g_nosc, lp2 = x->g_newshit; c2--; lp2++)
            {
                long inter = *lp2, tryit = inter + n;
                if (!lp2)
                    continue;
                if (tryit > pit - 50 && tryit < pit + 50)
                {
                    sendit(x, (long) (DECAL + count), pit);
                    goto didit;
                }
            }
            sendit(x, (long) (DECAL + count), 0L);
          didit:;
        }
        break;
    }
  nodice:;
}

void gumbo_ff(struct gumbo *x)
{
    clock_free(x->g_clock);
}

void *gumbo_class;

void gumbo_setup()
{
    c_new(&s_gumbo, gumbo_ff, sizeof(struct gumbo), 0, 0);
    c_addint(gumbo_int);
    c_addbang(gumbo_bang);
    c_addmess(gumbo_set, gensym("set"), A_GIMME, 0);
    c_addmess(gumbo_mode, gensym("mode"), A_LONG, 0);
    c_addmess(gumbo_slide, gensym("slide"), A_GIMME, 0);
    c_addmess(gumbo_stop, gensym("stop"), 0);
    c_addmess(gumbo_skew, gensym("skew"), A_LONG, 0);
    c_addmess(gumbo_dump, gensym("dump"), 0);
    c_addmess(gumbo_print, gensym("print"), 0);
    c_addmess(gumbo_sort, gensym("sort"), 0);
}

void *gumbo_new(long int n)
{
    int i;
    struct gumbo *x;

    x = (struct gumbo *) obj_new((t_messlist **) & gumbo_class,
                                 gumbo_setup);
    if (n <= 0)
        n = NCTL;
    x->g_out2 = outlet(x, &s_int);
    x->g_out1 = outlet(x, &s_int);
    x->g_clock = clock_new(x, gumbo_tick);
    x->g_mode = 0;
    for (i = 0; i < NCTL; i++)
        x->g_shit[i] = 0;
    if (n < 0)
        n = 0;
    else if (n > NCTL)
        n = NCTL;
    x->g_nosc = n;
    return (x);
}

/***************** freak.c -- obsolete!!! ***********/

#define SBEND 0xe0L             /* pitch bend "status byte" */
typedef struct freak {
    t_object f_ob;
    long f_in1;
    long f_off;
} t_freak;

extern t_symbol s_freak, s_in1;

void freak_int(t_freak *x, long int n)
{
    long poobah;
    poobah = ((n + x->f_off) << 12) + (((x->f_in1 << 3) / 25) & 0xfff);
    outlet_int(x->f_ob.o_outlet, SBEND + ((poobah >> 14) & 0xf));
    outlet_int(x->f_ob.o_outlet, ((poobah >> 7) & 0x7f));
    outlet_int(x->f_ob.o_outlet, ((poobah) & 0x7f));
}

void freak_in1(t_freak *x, long int n)
{
    x->f_in1 = n;
}

void *freak_class;
void freak_setup()
{
    c_new(&s_freak, NIL, sizeof(t_freak), 0, 0);
    c_addint(freak_int);
    c_addinx(freak_in1, 1);
}


void *freak_new(int n)
{
    t_freak *x =
        (t_freak *) obj_new((t_messlist **) & freak_class, freak_setup);

    x->f_off = (n & 63);
    outlet_new(x, &s_int);
    inlet(x, x, &s_int, &s_in1);
    return (x);
}

/***************** frog.c -- obsolete!!! ***********/

typedef struct frog {
    t_object f_ob;
    long f_in1;
    long f_offset;
} t_frog;

extern t_symbol s_frog, s_in1;

void frog_int(t_frog *x, long int n)
{
    long vel, n2;
    long poobah;

    n2 = x->f_in1;
    if (n2 & -128)
        n2 = (n2 > 0 ? 127 : 0);
    vel = n2 >> 4;
    if (n2 && !vel)
        vel = 1;
    poobah = ((vel + x->f_offset) << 12) + (((n << 3) / 25) & 0xfff);
    outlet_int(x->f_ob.o_outlet, SBEND + ((poobah >> 14) & 0xf));
    outlet_int(x->f_ob.o_outlet, ((poobah >> 7) & 0x7f));
    outlet_int(x->f_ob.o_outlet, ((poobah) & 0x7f));
}

void frog_in1(t_frog *x, long int n)
{
    x->f_in1 = n;
}

void *frog_class;
void frog_setup()
{
    c_new(&s_frog, NIL, sizeof(t_frog), 0, 0);
    c_addint(frog_int);
    c_addinx(frog_in1, 1);
}

void *frog_new(int n)
{
    t_frog *x =
        (t_frog *) obj_new((t_messlist **) & frog_class, frog_setup);

    x->f_offset = n;
    outlet_new(x, &s_int);
    inlet(x, x, &s_int, &s_in1);
    return (x);
}

/************* vdelay by Zack while on acid or something **************/

extern t_symbol s_vdel, s_ft1, s_ft2;
#define FOREVER 0x7fffffff

struct vdel {
    struct object v_ob;
    void *v_outlet2;
    void *v_clock;
    float v_tscaler;
    float v_del;
    double v_duedate;
    long v_startdate;
    float v_microsecs;
    int dbflag;
};

void vdel_bang(struct vdel *x)
{
    int zerror;

    x->v_duedate = (double) ((x->v_startdate = logicaltime) +
                             (x->v_del * x->v_tscaler));
    if (x->dbflag)
        post("setAt= %ld  DueDate=%ld", logicaltime,
             (x->v_duedate + 1 >=
              FOREVER) ? FOREVER : (long) x->v_duedate);
    if (x->v_duedate + 1 >= FOREVER)
        clock_unset(x->v_clock);        /* don't even set a clock */
    else
    {
        if ((long) x->v_microsecs)      /*  if accumulated zerror is >= 1 */
            clock_set(x->v_clock, (long) (x->v_duedate + 1));
        else
            clock_set(x->v_clock, (long) x->v_duedate);
    }
}

void vdel_db(struct vdel *x, long int n)
{
    x->dbflag = n;
    post("logicaltime = %ld", logicaltime);
}

void vdel_in1(struct vdel *x, double f) /*  delay time <float>   */
{
    x->v_del = ((f < 0.) ? 0. : f);
}

void vdel_in2(struct vdel *x, double f) /*  current delay scaler <float>   */
{
    f = ((!(f > .0001)) ? .0001 : f);   /*  no 0's allowd */
    if (x->v_duedate)           /* if pending delay  */
    {
        double remainingDelay;
        long noww;
        remainingDelay = x->v_duedate - (noww = logicaltime);
        remainingDelay /= x->v_tscaler;
        x->v_duedate = noww + (remainingDelay * (x->v_tscaler = f));
        if (x->dbflag)
            post("TChange:noww=%ld newDueDate=%ld", noww,
                 (long) x->v_duedate);
        if (x->v_duedate + 1 >= FOREVER)
            clock_unset(x->v_clock);
        else
        {
            if ((long) x->v_microsecs)  /*  if the accumulated zerror is >= 1 */
                clock_set(x->v_clock, (long) (x->v_duedate + 1));
            else
                clock_set(x->v_clock, (long) x->v_duedate);
        }
    }
    else
        x->v_tscaler = f;
}

void vdel_time(struct vdel *x)  /*  output elapsed time of pending delay   */
{
    if (x->v_duedate)
        outlet_float(x->v_outlet2,
                     ((logicaltime - x->v_startdate) / x->v_tscaler));
    else
        outlet_float(x->v_outlet2, 0.);
}

void vdel_stop(struct vdel *x)
{
    clock_unset(x->v_clock);
    x->v_microsecs = 0.;
    x->v_duedate = 0.;
}

void vdel_tick(struct vdel *x)
{
    /*  1ms zerror has been eliminated */
    if ((long) x->v_microsecs)
        x->v_microsecs -= 1;
    x->v_microsecs += (x->v_duedate - (long) x->v_duedate);
    if (x->dbflag)
        post("expiredAt = %ld", logicaltime);
    x->v_duedate = 0.;
    outlet_bang(x->v_ob.o_outlet);
}

void vdel_dispose(struct vdel *x)
{
    clock_free(x->v_clock);
}

void vdel_fixit(struct vdel *x, long int n)
{
    post("delay set to %ld", n);
    clock_set(x->v_clock, n);
    post("now= %ld", logicaltime);
}

void *vdel_class;
void vdel_setup()               /* set up the class */
{
    c_new(&s_vdel, vdel_dispose, sizeof(struct vdel), 0, 0);
    c_addbang(vdel_bang);       /* "bang" method */
    c_addftx(vdel_in1, 1);      /* "float" in 1st inlet */
    c_addftx(vdel_in2, 2);      /* "float" in 2nd inlet */
    c_addmess(vdel_stop, gensym("stop"), 0);    /* "stop" method; no arg */
    c_addmess(vdel_time, gensym("time"), 0);    /* "time" method; no arg */
    c_addmess(vdel_fixit, gensym("fixit"), A_DEFLONG, 0);       /* "fixit" opt arg */
    c_addmess(vdel_db, gensym("debug"), A_DEFLONG, 0);  /* "db" opt arg */
    post("vdelayDEBUG version 1.1 NeXT");
}

void *vdel_new(double f)
{
    struct vdel *x;
    x = (struct vdel *) obj_new((t_messlist **) & vdel_class, vdel_setup);
    x->v_outlet2 = outlet_new(x, &s_float);
    outlet_new(x, &s_bang);
    inlet_new(x, x, &s_ft2, &s_float);
    inlet_new(x, x, &s_ft1, &s_float);
    x->v_clock = clock_new(x, vdel_tick);
    x->v_duedate = 0.;
    x->v_tscaler = 1.;
    x->v_del = (f > 0. ? f : 0.);
    x->v_microsecs = 0.;
    x->dbflag = 0;
}
