#ifndef lint
static char rcsid[] =
    "$Revision: 1.9 $ IRCAM $Date: 1994/01/04 18:05:10 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 1.9 $ IRCAM $Date: 1994/01/04 18:05:10 $
 *
 * FTS by Miller Puckette
 *
 */


#include <stdio.h>
#include "m_mess.h"
#include "t_mac.h"              /* needed by patcher.h */
#include "p_patcher.h"          /* for CP_PUT, CP_UNPUT defs */
#include "h_cp.h"

extern t_symbol s_send, s_through, s_receive, s_value, s_setcp, s__connect,
    s_rreceive, s_dblclick;

void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);
void *dummyfn();
extern int h_nproc;

/* "through".  inlet chained straight to outlet */

typedef struct cpent            /* CP entry -- each "through" keeps one per CP */
{
    int c_receivecount;         /* number of "receive"s on this CP */
    void *c_rreceive;           /* remote receiver object if any */
} t_cpent;

typedef struct through {
    struct object t_ob;
    struct symbol *t_sym;
    long t_count;
    struct atom t_atom;
    t_cpent *t_cpvec;
    int t_ok;
} t_through;


static void through_dummy()
{
/* methods given as "through_dummy" are really defined in the CP.  It doesn't
   work to reuse the CP versions here as they make explicit reference
   to rsends and rreceives. 
*/
}

void through_ff(t_through *x)
{
    t_cpent *cpent;
    int i;
    for (i = 0, cpent = x->t_cpvec; i < h_nproc; i++, cpent++)
        h_free(cpent);
    freebytes(x->t_cpvec, h_nproc * sizeof(t_cpent));
    x->t_sym->s_thing = 0;
}

void through_puton(t_through *x)
{
    int i;
    t_cpent *cpent;
    for (i = 0, cpent = x->t_cpvec; i < h_nproc; i++, cpent++)
    {
        cpent->c_receivecount = 0;
        cpent->c_rreceive = h_vnewunique(i, &s_rreceive, "ll", 0L, 0L);
        h_vnew(cpent, i, &s_through, "so", x->t_sym, cpent->c_rreceive);
    }
    x->t_ok = 1;
}

void *through_class;
void through_setup()
{
    c_new(&s_through, through_ff, sizeof(t_through), 0, 0);
    c_addbang(through_dummy);
    c_addint(through_dummy);
    c_addfloat(through_dummy);
    c_addsymbol(through_dummy);
    c_addmess(through_dummy, &s_list, A_GIMME, 0);
    c_addmess(through_dummy, &s_anything, A_GIMME, 0);
}

void *through(t_symbol *s)
{
    t_through *x =
        (t_through *) obj_new((t_messlist **) & through_class,
                              through_setup);
    outlet(x, NIL);
    x->t_count = 0;
    x->t_sym = s;
    x->t_atom.a_type = A_LONG;
    x->t_atom.a_w.w_long = 0;
    x->t_cpvec = (t_cpent *) getbytes(h_nproc * sizeof(t_cpent));
    s->s_thing = (t_object *) x;
    through_puton(x);
    return (x);
}

/* "send". Send messages commonly. */

typedef struct send {
    struct object s_ob;
    struct through *s_through;
    int s_cpno;
} t_send;

static int findingsend;

int send__newfind(t_symbol *s)
{
    char buf[256];
    void *wind_front(), *wack = wind_front();
    if (!wack || !ispatcher(wack))
    {
        bug("send_dblclick");
        return (0);
    }
    sprintf(buf, "%c %.250s", (findingsend ? 's' : 'r'), s->s_name);
    return (patcher_findstring(wack, buf, 0));
}

void send__dodblclick(void *x, t_symbol *s, int issend)
{
    void *patcher_lastfound();
    if (x == patcher_lastfound())
    {
        if (patcher_findagain(0))
            return;
        findingsend = !findingsend;
    }
    else
        findingsend = !issend;
    if (!send__newfind(s))
    {
        findingsend = !findingsend;
        if (!send__newfind(s))
            ouch("not found");
    }
}

void send_dblclick(t_send *x)
{
    send__dodblclick(x, x->s_through->t_sym, 1);
}

void send_setcp(t_send *x, int n, int flags)
{
    if (flags & CP_RESTART)
        x->s_through->t_ok = 0;
    if (flags & CP_PUT)
    {
        if (!x->s_through->t_ok)
            through_puton(x->s_through);
        h_vmess(x, &s__connect, "o", &x->s_through->t_cpvec[n]);
    }
}

void send_ff(t_send *x)
{
    t_through *t = x->s_through;
    if (!--(t->t_count))
        freeobject((t_object *) t);
}

void *send_class;
void send_setup()
{
    c_new(&s_send, send_ff, sizeof(t_send), 0, 0);
    c_addmess(send_setcp, &s_setcp, A_CANT, 0);
    c_addbang(through_dummy);
    c_addint(through_dummy);
    c_addfloat(through_dummy);
    c_addsymbol(through_dummy);
    c_addmess(through_dummy, &s_list, A_GIMME, 0);
    c_addmess(through_dummy, &s_anything, A_GIMME, 0);
    c_addmess(send_dblclick, &s_dblclick, A_CANT, 0);
}

void *send_new(t_symbol *s)
{
    t_through *t;
    t_send *x;
    if (t = (t_through *) (s->s_thing))
    {
        if (ob_messlist(t) != through_class)
        {
            error("send: %s: already exists", s->s_name);
            return (0);
        }
    }
    else
        t = through(s);
    x = (t_send *) obj_new((t_messlist **) & send_class, send_setup);
    x->s_through = t;
    x->s_cpno = -1;
    t->t_count++;
    return (x);
}

/* "receive". receive messages commonly. */

typedef struct receive {
    struct object r_ob;
    struct through *r_through;
    int r_cpno;
} t_receive;

/* tell the receive object what CP it's on.
   
   If you put a receive on a CP which does not have one yet, all the "through"
   objects (one per CP) chould be informed that they will now send messages to
   the new CP.
   
   If you take out the last receive on a CP you have to tell
   sends on all other CPs to quit sending to it. */

void receive_setcp(t_receive *x, int n, int flags)
{
    t_cpent *cpent = x->r_through->t_cpvec;
    if (flags & CP_RESTART)
        x->r_through->t_ok = 0;
    else if (!x->r_through->t_ok)
        through_puton(x->r_through);
    if (flags & CP_UNPUT)
    {
        t_cpent *e2 = &cpent[x->r_cpno], *e3;
        if (x->r_cpno < 0)
            bug("receive_setcp 1");
        if (--(e2->c_receivecount) < 0)
            bug("receive_setcp 2");
        if (!e2->c_receivecount)
        {
            long cp;
            for (cp = 0, e3 = cpent; cp < h_nproc; e3++, cp++)
                if (cp != x->r_cpno)
                {
                    /* tell sender that its receiver is gone */
                    h_vmess(e3, &s__connect, "ll", 0L, (long) (x->r_cpno));
                }
        }
        x->r_cpno = -1;
    }
    if (flags & CP_PUT)
    {
        t_cpent *e2 = &cpent[x->r_cpno = n], *e3;
        h_vmess(x, &s__connect, "o", e2);
        if (!(e2->c_receivecount)++)
        {
            long cp;
            for (cp = 0, e3 = cpent; cp < h_nproc; e3++, cp++)
                if (cp != n)
                {
                    /* tell receiver that a new sender will appear */
                    h_vmess(e2, &s__connect, "ll", 0L, -1L);
                    /* tell the sender to include the receiver */
                    h_vmess(e3, &s__connect, "ll", e2->c_rreceive, n);
                }
        }
    }
}

void receive_dblclick(t_receive *x)
{
    send__dodblclick(x, x->r_through->t_sym, 0);
}

void receive_ff(t_receive *x)
{
    t_through *t = x->r_through;
    outlet_rm(t->t_ob.o_outlet, x->r_ob.o_outlet);
    if (!--(t->t_count))
        freeobject((t_object *) t);
}

void *receive_class;
void receive_setup()
{
    c_new(&s_receive, receive_ff, sizeof(t_receive), 0, 1);
#ifdef GRAPHICS
    c_addmess(receive_setcp, &s_setcp, A_CANT, 0);
    c_addmess(receive_dblclick, &s_dblclick, A_CANT, 0);
#endif
}

void *receive_new(t_symbol *s)
{
    t_through *t;
    t_receive *x;
    if (t = (t_through *) (s->s_thing))
    {
        if (ob_messlist(t) != through_class)
        {
            error("receive: %s: already exists", s->s_name);
            return (0);
        }
    }
    else
        t = through(s);
    x = (t_receive *) obj_new((t_messlist **) & receive_class,
                              receive_setup);
    outlet(x, NIL);
    x->r_through = t;
    x->r_cpno = -1;
    outlet_add(t->t_ob.o_outlet, x->r_ob.o_outlet);
    t->t_count++;
    return (x);
}

/* "value". remember a value. */

typedef struct value {
    struct object v_ob;
    struct through *v_through;
} t_value;

void value_setcp(t_value *x, int n, int flags)
{
    if (flags & CP_RESTART)
        x->v_through->t_ok = 0;
    if (flags & CP_PUT)
    {
        if (!x->v_through->t_ok)
            through_puton(x->v_through);
        h_vmess(x, &s__connect, "o", &x->v_through->t_cpvec[n]);
    }
}

void value_ff(t_value *x)
{
    t_through *t = x->v_through;
    if (!--(t->t_count))
        freeobject((t_object *) t);
}

void *value_class;
void value_setup()
{
    c_new(&s_value, value_ff, sizeof(t_value), 0, 0);
    c_addmess(value_setcp, &s_setcp, A_CANT, 0);
    c_addbang(through_dummy);
    c_addint(through_dummy);
    c_addfloat(through_dummy);
    c_addsymbol(through_dummy);
}

void *value_new(t_symbol *s)
{
    t_through *t;
    t_value *x;
    if (t = (t_through *) (s->s_thing))
    {
        if (ob_messlist(t) != through_class)
        {
            error("value: %s: already exists", s->s_name);
            return (0);
        }
    }
    else
        t = through(s);
    x = (t_value *) obj_new((t_messlist **) & value_class, value_setup);
    x->v_through = t;
    t->t_count++;
    outlet(x, &s_int);
    return (x);
}
