#ifndef lint
static char rcsid[] =
    "$Revision: 2.1 $ IRCAM $Date: 1993/09/08 09:01:28 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.1 $ IRCAM $Date: 1993/09/08 09:01:28 $
 *
 * vexp_func.c -- this file include all the functions for vexp.
 *      the first two arguments to the function are the number
 *      of argument and an array of arguments (argc, argv)
 *      the last argument is a pointer to a struct ex_ex for
 *      the result.  Up do this point, the content of the
 *      struct ex_ex that these functions receive are either
 *      ET_INT (long), ET_FLT (float), or ET_SYM (char **, it is
 *      char ** and not char * since NewHandle of Mac returns
 *      a char ** for relocatability.)  The common practice in
 *      these functions is that they figure out the type of their
 *      result according to the type of the arguments. In general
 *      the ET_SYM is used an ET_INT when we expect a value.
 *      It is the users responsibility not to pass strings to the
 *      function.
 */

#ifndef SUN
#define MAXMAX
#endif
#define __STRICT_BSD__
#include <math.h>
#undef __STRICT_BSD__
#ifdef MAXMAX
#include "m_mess.h"
#endif
#include "x_vexp.h"

extern long ex_min(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_max(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_toint(long int argc, struct ex_ex *argv,
                     struct ex_ex *optr);
extern long ex_tofloat(long int argc, struct ex_ex *argv,
                       struct ex_ex *optr);
extern long ex_pow(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_exp(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_log(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_ln(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_sin(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_cos(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_asin(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_acos(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_tan(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_atan(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_sinh(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_cosh(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_asinh();
extern long ex_acosh();
extern long ex_tanh(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_atanh();
extern long ex_atan2(long int argc, struct ex_ex *argv,
                     struct ex_ex *optr);
extern long ex_size(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_sum(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_Sum(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_sqrt(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_avg(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_Avg(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_std();
extern long ex_fact(long int argc, struct ex_ex *argv, struct ex_ex *optr);
extern long ex_store(long int argc, struct ex_ex *argv,
                     struct ex_ex *optr);
extern long ex_random(long int argc, struct ex_ex *argv,
                      struct ex_ex *optr);

t_ex_func ex_funcs[] = {
    { "min", ex_min, 2 },       /* min */
    { "max", ex_max, 2 },       /* max */
    { "int", ex_toint, 1 },     /* convert to integer */
    { "float", ex_tofloat, 1 }, /* convert to float */
    { "pow", ex_pow, 2 },       /* power */
    { "sqrt", ex_sqrt, 1 },     /* sqrt */
    { "exp", ex_exp, 1 },       /* power of e to x */
    { "log10", ex_log, 1 },     /* log */
    { "ln", ex_ln, 1 },         /* nat. log */
    { "log", ex_ln, 1 },        /* nat. log */
    { "sin", ex_sin, 1 },       /* sinus */
    { "cos", ex_cos, 1 },       /* cosin */
    { "tan", ex_tan, 1 },       /* tan */
    { "asin", ex_asin, 1 },     /* asin */
    { "acos", ex_acos, 1 },     /* acos */
    { "atan", ex_atan, 1 },     /* atan */
    { "atan2", ex_atan2, 2 },   /* atan2 */
    { "sinh", ex_sinh, 1 },     /* hyperbolic sinus */
    { "cosh", ex_cosh, 1 },     /* hyperbolic cosin */
    { "tanh", ex_tanh, 1 },     /* hyperbolic tan */
    { "fact", ex_fact, 1 },     /* factorial */
    { "random", ex_random, 2 }, /* random number */
#ifndef MAXMAX
    { "asinh", ex_asinh, 1 },   /* hyperbolic asin */
    { "acosh", ex_acosh, 1 },   /* hyperbolic acos */
    { "atanh", ex_atanh, 1 },   /* hyperbolic atan */
#else
    { "size", ex_size, 1 },     /* size of table */
    { "sum", ex_sum, 1 },       /* sum of contect of a table */
    { "Sum", ex_Sum, 3 },       /* sum with boundries */
    { "avg", ex_avg, 1 },       /* avarage of contect of table */
    { "Avg", ex_Avg, 3 },       /* avarage with boundries */
    { "store", ex_store, 3 },   /* store in the table */
#endif                          /* MAXMAX */
    { 0, 0, 0 }
};

/*
 * ex_min -- the type of the result is the type of the smaller value
 */
/* ARGSUSED */
long ex_min(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    struct ex_ex *left, *right;

    left = argv++;
    right = argv;
    if (left->ex_type == ET_FLT)
    {
        if (right->ex_type == ET_FLT)
        {
            if (left->ex_flt < right->ex_flt)
                *optr = *left;
            else
                *optr = *right;
        }
        else
        {
            if (left->ex_flt < (float) right->ex_int)
                *optr = *left;
            else
                *optr = *right;
        }
    }
    else
    {
        if (right->ex_type == ET_FLT)
        {
            if ((float) left->ex_int < right->ex_flt)
                *optr = *left;
            else
                *optr = *right;
        }
        else
        {
            if (left->ex_int < right->ex_int)
                *optr = *left;
            else
                *optr = *right;
        }
    }
}

/*
 *ex_max -- the type of the result is the type of the larger value
 */
/* ARGSUSED */
long ex_max(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    struct ex_ex *left, *right;

    left = argv++;
    right = argv;
    if (left->ex_type == ET_FLT)
    {
        if (right->ex_type == ET_FLT)
        {
            if (left->ex_flt > right->ex_flt)
                *optr = *left;
            else
                *optr = *right;
        }
        else
        {
            if (left->ex_flt > (float) right->ex_int)
                *optr = *left;
            else
                *optr = *right;
        }
    }
    else
    {
        if (right->ex_type == ET_FLT)
        {
            if ((float) left->ex_int > right->ex_flt)
                *optr = *left;
            else
                *optr = *right;
        }
        else
        {
            if (left->ex_int > right->ex_int)
                *optr = *left;
            else
                *optr = *right;
        }
    }
}

/*
 * ex_toint -- convert to integer
 */
long ex_toint(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    if (argv->ex_type == ET_INT)
        *optr = *argv;
    else if (argv->ex_type == ET_FLT)
    {
        optr->ex_type = ET_INT;
        optr->ex_int = (int) argv->ex_flt;
    }
    else
    {
        optr->ex_type = ET_INT;
        optr->ex_int = (int) argv->ex_ptr;
    }
}

/*
 * ex_tofloat -- convert to float
 */
long ex_tofloat(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    if (argv->ex_type == ET_FLT)
        *optr = *argv;
    else
    {
        optr->ex_type = ET_FLT;
        optr->ex_flt = (float) argv->ex_int;
    }
}


/*
 * ex_pow -- the power of
 */
/* ARGSUSED */
long ex_pow(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    struct ex_ex *left, *right;

    left = argv++;
    right = argv;
    optr->ex_type = ET_FLT;
    if (left->ex_type == ET_FLT)
    {
        if (right->ex_type == ET_FLT)
        {
            optr->ex_flt = (float)
                pow((double) left->ex_flt, (double) right->ex_flt);
        }
        else
        {
            optr->ex_flt = (float)
                pow((double) left->ex_flt, (double) right->ex_int);
        }
    }
    else
    {
        if (right->ex_type == ET_FLT)
        {
            optr->ex_flt = (float)
                pow((double) left->ex_int, (double) right->ex_flt);
        }
        else
        {
            optr->ex_type = ET_INT;
            optr->ex_int = (int)
                pow((double) left->ex_int, (double) right->ex_int);
        }
    }
}

/*
 * ex_sqrt -- square root
 */
/* ARGSUSED */
long ex_sqrt(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) sqrt((double) argv->ex_flt);
    else
        optr->ex_flt = (float) sqrt((double) argv->ex_int);
}

/*
 * ex_exp -- e to the power of
 */
/* ARGSUSED */
long ex_exp(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) exp((double) argv->ex_flt);
    else
        optr->ex_flt = (float) exp((double) argv->ex_int);
}

/*
 * ex_log -- 10 based logarithm
 */
/* ARGSUSED */
long ex_log(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) log10((double) argv->ex_flt);
    else
        optr->ex_flt = (float) log10((double) argv->ex_int);
}

/*
 * ex_ln -- natural log
 */
/* ARGSUSED */
long ex_ln(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) log((double) argv->ex_flt);
    else
        optr->ex_flt = (float) log((double) argv->ex_int);
}

/* ARGSUSED */
long ex_sin(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) sin((double) argv->ex_flt);
    else
        optr->ex_flt = (float) sin((double) argv->ex_int);
}

/* ARGSUSED */
long ex_cos(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) cos((double) argv->ex_flt);
    else
        optr->ex_flt = (float) cos((double) argv->ex_int);
}

/* ARGSUSED */
long ex_tan(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) tan((double) argv->ex_flt);
    else
        optr->ex_flt = (float) tan((double) argv->ex_int);
}

/* ARGSUSED */
long ex_asin(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) asin((double) argv->ex_flt);
    else
        optr->ex_flt = (float) asin((double) argv->ex_int);
}

/* ARGSUSED */
long ex_acos(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) acos((double) argv->ex_flt);
    else
        optr->ex_flt = (float) acos((double) argv->ex_int);
}

/* ARGSUSED */
long ex_atan(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) atan((double) argv->ex_flt);
    else
        optr->ex_flt = (float) atan((double) argv->ex_int);
}

/*
 *ex_atan2 --
 */
/* ARGSUSED */
long ex_atan2(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    struct ex_ex *left, *right;

    left = argv++;
    right = argv;
    optr->ex_type = ET_FLT;
    if (left->ex_type == ET_FLT)
    {
        if (right->ex_type == ET_FLT)
        {
            optr->ex_flt = (float) atan2((double) left->ex_flt,
                                         (double) right->ex_flt);
        }
        else
        {
            optr->ex_flt = (float) atan2((double) left->ex_flt,
                                         (double) right->ex_int);
        }
    }
    else
    {
        if (right->ex_type == ET_FLT)
        {
            optr->ex_flt = (float) atan2((double) left->ex_int,
                                         (double) right->ex_flt);
        }
        else
        {
            optr->ex_flt = (float) atan2((double) left->ex_int,
                                         (double) right->ex_int);
        }
    }
}

/* ARGSUSED */
long ex_sinh(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) sinh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) sinh((double) argv->ex_int);
}

/* ARGSUSED */
long ex_cosh(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) cosh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) cosh((double) argv->ex_int);
}

/* ARGSUSED */
long ex_tanh(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) tanh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) tanh((double) argv->ex_int);
}

#ifndef MAXMAX
long ex_asinh(argc, argv, optr)
long argc;
struct ex_ex *argv, *optr;
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) asinh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) asinh((double) argv->ex_int);
}

long ex_acosh(argc, argv, optr)
long argc;
struct ex_ex *argv, *optr;
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) acosh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) acosh((double) argv->ex_int);
}

long ex_atanh(argc, argv, optr)
long argc;
struct ex_ex *argv, *optr;
{
    optr->ex_type = ET_FLT;
    if (argv->ex_type == ET_FLT)
        optr->ex_flt = (float) atanh((double) argv->ex_flt);
    else
        optr->ex_flt = (float) atanh((double) argv->ex_int);
}
#endif                          /* not MAXMAX */

long ex_fact(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    long i;

    if (argv->ex_type == ET_FLT)
        i = (long) argv->ex_flt;
    else
        i = argv->ex_int;
    optr->ex_type = ET_INT;
    if (i)
        optr->ex_int = 1;
    else
    {
        optr->ex_int = 0;
        return;
    }
    do
    {
        optr->ex_int *= i;
    }
    while (--i);
}

int i1, i2;

/*
 * ex_random -- return a random number
 */
long ex_random(long int argc, struct ex_ex *argv, struct ex_ex *optr)
{
    if (argv->ex_type == ET_FLT)
        i1 = (int) argv->ex_flt;
    else
        i1 = argv->ex_int;
    if (argv[1].ex_type == ET_FLT)
        i2 = (int) argv[1].ex_flt;
    else
        i2 = argv[1].ex_int;
    optr->ex_type = ET_INT;
    optr->ex_int = i1 + (((i2 - i1) * (rand() & 0x7fffL)) >> 15);
}
