#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:05:53 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:05:53 $
 *
 * FTS by Miller Puckette
 *
 */


#include "m_mess.h"
extern struct symbol s_set, s_in1, s_in2, s_accum;

void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);

/*   faccum object    -a sort of accumulater   */
typedef struct faccum {
    struct object s_ob;
    void *s_outlet;
    void *s_inlet1;
    void *s_inlet2;
    float s_reg;
} t_faccum;


void faccum_bang(struct faccum *x)      /*  output contents   */
{
    outlet_float(x->s_outlet, x->s_reg);
}

void faccum_in1(struct faccum *x, double n)     /*   increment contents by n  */
{
    x->s_reg += n;
}

void faccum_in2(struct faccum *x, double n)     /*   multiplies contents by n  */
{
    x->s_reg *= n;
}

void faccum_set(struct faccum *x, double n)     /*   set contents to n  */
{
    x->s_reg = n;
}

void faccum_int(struct faccum *x, long int n)
{
    x->s_reg = (float) n;
    outlet_float(x->s_outlet, x->s_reg);
}

void faccum_float(struct faccum *x, double n)
{
    x->s_reg = n;
    outlet_float(x->s_outlet, x->s_reg);
}

void *faccum_class;
void faccum_setup()
{
    c_new(&s_accum, NIL, sizeof(t_faccum), 0, 0);
    c_addint(faccum_int);
    c_addfloat(faccum_float);
    c_addbang(faccum_bang);
    c_addmess(faccum_set, &s_set, A_FLOAT, 0);
    c_addinx(faccum_in1, 1);
    c_addinx(faccum_in2, 2);
}

void *faccum(double n)
{
    struct faccum *x =
        (struct faccum *) obj_new((t_messlist **) & faccum_class,
                                  faccum_setup);
    x->s_outlet = outlet(x, &s_float);
    x->s_inlet2 = inlet(x, x, &s_float, &s_in2);
    x->s_inlet1 = inlet(x, x, &s_float, &s_in1);
    x->s_reg = n;
    return (x);
}

/* ==================================================================== */
/*   iaccum object    -a sort of iaccumulater   */
typedef struct iaccum {
    struct object s_ob;
    void *s_outlet;
    void *s_inlet1;
    void *s_inlet2;
    float s_reg;
} t_iaccum;

void iaccum_bang(struct iaccum *x)      /*  output contents   */
{
    outlet_int(x->s_outlet, (long) (x->s_reg + .5));
}

void iaccum_in1(struct iaccum *x, long int n)   /*   increment contents by n  */
{
    x->s_reg += n;
}

void iaccum_in2(struct iaccum *x, double n)     /*   multiplies contents by n  */
{
    x->s_reg *= n;
}

void iaccum_set(struct iaccum *x, long int n)   /*   set contents to n  */
{
    x->s_reg = n;
}

void iaccum_int(struct iaccum *x, long int n)
{
    x->s_reg = n;
    outlet_int(x->s_outlet, (long) (x->s_reg + .5));
}

void iaccum_float(struct faccum *x, double n)
{
    x->s_reg = n;
    outlet_int(x->s_outlet, (long) (x->s_reg + .5));
}

void *iaccum_class;
void iaccum_setup()
{
    c_new(&s_accum, NIL, sizeof(t_iaccum), 0, 0);
    c_addint(iaccum_int);
    c_addfloat(iaccum_float);
    c_addbang(iaccum_bang);
    c_addmess(iaccum_set, &s_set, A_FLOAT, 0);
    c_addinx(iaccum_in1, 1);
    c_addinx(iaccum_in2, 2);
}

void *iaccum(long int n)
{
    struct iaccum *x =
        (struct iaccum *) obj_new((t_messlist **) & iaccum_class,
                                  iaccum_setup);
    x->s_outlet = outlet(x, &s_int);
    x->s_inlet2 = inlet(x, x, &s_float, &s_in2);
    x->s_inlet1 = inlet(x, x, &s_int, &s_in1);
    x->s_reg = (float) n;
    return (x);
}

/* ==================  choice between "subclasses" =================== */

void *accumulate_new(struct symbol *sel, int argc, struct atom *argv)
{
    if (argc < 1)
        return (iaccum(0L));
    if (argv->a_type == A_LONG)
        return (iaccum(argv->a_w.w_long));
    if (argv->a_type == A_FLOAT)
        return (faccum(argv->a_w.w_float));
    error("accum: bad input arg");
    return (0);
}
