#ifndef lint
static char rcsid[] =
    "$Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:06:02 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.8 $ IRCAM $Date: 1994/01/04 18:06:02 $
 *
 * FTS by Miller Puckette
 * zack  4/88
 */

#define __STRICT_BSD__
#include <math.h>
#undef __STRICT_BSD__

typedef void *t_wind;
#include "m_mess.h"
#include "p_table.h"

#define funbuff_first(FB) (FB)->f_head.z_next

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif
#define VBAG(s) {error(s); return;}

extern struct symbol s_zllf, s_zll, s_set, s_print,
    s_funbuff, s_dump, s_find, s_cut,
    s_copy, s_paste, s_clear, s_reduce, s_undo, s_interp, s_goto, s_next,
    s_read, s_write, s_select, s_in1, s_interptab;


char f_fileprompt[256] = "SNOX.fb";

typedef struct zll {
    long z_x;
    long z_y;
    struct zll *z_next;
    struct zll *z_prev;
} t_zll;

typedef struct funbuff {
    struct object f_obj;
    void *f_outlet;
    void *f_outlet1;
    void *f_outlet2;
    struct zll f_head;          /* head of double linked list of function elts */
    struct zll *f_foot;         /* foot in the door pointer for list  */
    long f_gotoDelta;           /* used by goto and next */
    long f_selectX;             /* selected region start   */
    long f_selectW;             /* selected region width   */
    struct symbol *f_sym;       /* filename */
    long f_y;                   /* y-value from inlet */
    char f_yvalid;              /* flag that y has been set since x has */
} t_funbuff;

/* 
  void *zll_freefun(x)        NOT USED;    DOING IT BY HAND    
  struct zll *x;
  {
  x->z_next->z_prev = x->z_prev;
  x->z_prev->z_next = x->z_next;
  freeobject((t_object *)x);
  }           */

/* note:  to kill a zll element , do zll_delete(zll, zll)  then  freeraw(zll, zll_messlist)   */

zll_free(struct zll *z)
{
    freebytes(z, sizeof(struct zll));
}

void *zll_new()                 /*   create empty zll  */
{
    struct zll *x = (struct zll *) getbytes(sizeof(struct zll));
    return (x);
}

/*    ZLL LOCAL FUNCTIONS +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  */

void zll_delete(struct zll *xfirst, struct zll *xlast)
/*  deletes xfirst thru xlast and rejoines  */
{
    if (xfirst->z_prev)
        xfirst->z_prev->z_next = xlast->z_next;
    if (xlast->z_next)
        xlast->z_next->z_prev = xfirst->z_prev;
    xlast->z_next = NIL;
}

void zll_insert(struct zll *x1, struct zll *xnew)       /* xnew must contain at least 1 element and be NIL terminated */
{
    struct zll *x2 = x1->z_next;
    x1->z_next = xnew;          /*  inserts right after x1 */
    xnew->z_prev = x1;
    for (xnew; xnew->z_next; xnew = xnew->z_next);      /*  go to end of new elements */
    xnew->z_next = x2;
    if (x2)
        x2->z_prev = xnew;
}

/*  returns copy of region from x to x + w of list  or NIL  */
struct zll *zll_copy(struct zll *list, long int x1, long int w)
{
    long x2 = x1 + w;

    struct zll *head, *copy;

    for (list; list; list = list->z_next)
        if (list->z_x >= x1)
            break;
    if (!list || list->z_x > x2)
    {
        post("zll_copy: no data to copy");
        return (NIL);
    }
    head = copy = zll_new();
    head->z_prev = NIL;
    head->z_x = list->z_x;
    head->z_y = list->z_y;
    list = list->z_next;
    for (list; list; list = list->z_next)
    {
        if (list->z_x > x2)
            break;
        copy->z_next = zll_new();
        copy->z_next->z_prev = copy;
        copy = copy->z_next;
        copy->z_x = list->z_x;
        copy->z_y = list->z_y;
    }
    copy->z_next = NIL;
    return (head);
}

struct zll *zll_cut(struct zll *list, long int x1, long int w)
{
    struct zll *head, *tail;
    long x2 = x1 + w;

    for (list; list; list = list->z_next)
        if (list->z_x >= x1)
            break;
    if (!list || list->z_x > x2)
        return (NIL);
    head = list;
    for (list; list->z_x <= x2; list = list->z_next)
        if (!list->z_next)
            break;
    tail = (list->z_x <= x2) ? list : list->z_prev;
    zll_delete(head, tail);
    return (head);
}

void *zll_last(struct zll *list)
{
    for (NIL; list->z_next; list = list->z_next);
    return (list);
}

void *zll_rewind(struct zll *list)
{
    for (NIL; list->z_prev; list = list->z_prev);
    return (list);
}

void *zll_freefun(struct zll *list)
{
    for (NIL; list; list = list->z_next)
        freebytes(list, sizeof(struct zll));
}


long zll_length(struct zll *list)
{
    long length = 0;
    for (list; list; list = list->z_next)
        length++;
    return (length);
}

void *zll_DBprin(struct zll *x)
{
    post("ZLL: z_x = %ld  z_y = %ld", x->z_x, x->z_y);
    if (x->z_prev)
        post("    prev ptr = %ld", x->z_prev);
    else
        post("   prev ptr = ()");
    if (x->z_next)
        post("    next ptr = %ld", x->z_next);
    else
        post("   next ptr = ()");
    return;
}

void *zll_prinlist(struct zll *list)
{
    for (NIL; list; list = list->z_next)
        zll_DBprin(list);
}

/* accepts a vector : x1 y1 x2 y2 x3 y3 ... places it in a zll list  returns list or NIL */
struct zll *zll_vecToZll(int ac, struct atom *av)
{
    struct zll *head, *current;
    int count;
    long xn, yn;

    if (ac & 1)
    {
        post("zll_vecToZll: bad input - odd arg count= %d", ac);
        return (NIL);
    }
    if (!ac)
    {
        post("zll_vecToZllf: bad input - that is,  no input");
        return (NIL);
    }
    head = current = zll_new();
    head->z_prev = NIL;
    for (count = 0; count < 2; count++, av++)
    {                           /* pick off first element */
        if (av->a_type != A_LONG)
            post("zll_vecToZll: bad input"), xn =
                0, post(" X bad %ld", av->a_w.w_long);
        else
            xn = av->a_w.w_long;
        count++, av++;
        if (av->a_type != A_LONG)
            post("zll_vecToZll: bad input"), yn =
                0, post(" Y bad %ld", av->a_w.w_long);
        else
            yn = av->a_w.w_long;
        current->z_x = xn, current->z_y = yn;
    }
    for (count = 2; count < ac; count++, av++)
    {                           /* pick off any others */
        current->z_next = zll_new();
        current->z_next->z_prev = current;
        current = current->z_next;
        if (av->a_type != A_LONG)
            post("zll_vecToZll: bad input"), xn =
                0, post(" X bad %ld", av->a_w.w_long);
        else
            xn = av->a_w.w_long;
        count++, av++;
        if (av->a_type != A_LONG)
            post("zll_vecToZll: bad input"), yn =
                0, post(" Y bad %ld", av->a_w.w_long);
        else
            yn = av->a_w.w_long;
        current->z_x = xn, current->z_y = yn;
    }
    current->z_next = NIL;
    return (head);
}

/* ++++++++++++++   ZLLF object   (floting point zll) +++++++++++++*/
struct zllf                     /* doubly linked list */
{                               /* note:  no obj header  */
    long z_x;
    float z_y;                  /* floating pt val for Y  */
    struct zllf *z_next;
    struct zllf *z_prev;
};

void *zllf_new()                /*   create empty zllf  */
{
    struct zllf *x = (struct zllf *) getbytes(sizeof(struct zllf));
    return (x);
}

void *zllf_freefun(struct zllf *list)
{
    for (NIL; list; list = list->z_next)
        freebytes(list, sizeof(struct zllf));
}

void *zllf_DBprin(struct zllf *x)
{

    post("zllf: z_x = %ld  z_y = %f", x->z_x, x->z_y);
    if (x->z_prev)
        post("    prev ptr = %ld", x->z_prev);
    else
        post("   prev ptr = ()");
    if (x->z_next)
        post("    next ptr = %ld", x->z_next);
    else
        post("   next ptr = ()");
    return;
}

void *zllf_prinlist(struct zllf *list)
{
    for (NIL; list; list = list->z_next)
        zllf_DBprin(list);
}

/* +++++++++++++++++++++++++++ ZLL DATA  REDUCTION ++++++++++++++++  */
float zll_distance(long int x, long int y, long int x1, long int y1,
                   long int x2, long int y2)
{
    long x1_x2 = x1 - x2;
    long y2_y1 = y2 - y1;
    long x2y1_x1y2 = (x2 * y1) - (x1 * y2);
    return ((float)
            fabs((double) ((x * y2_y1) + (y * x1_x2) + x2y1_x1y2)));
}

struct zllf *zll_getslope(struct zll *function)
{
    struct zllf *slope_head, *slope;

    if (!function)
    {
        post(" zll_getslope: no data ");
        return (NIL);
    }

    slope_head = slope = zllf_new();
    slope->z_prev = NIL;
    slope->z_x = function->z_x;
    slope->z_y = 1.0;           /*  assumes a slope of 1 on entry   */

    for (NIL; function->z_next; function = function->z_next)
    {
        slope->z_next = zllf_new();
        slope->z_next->z_prev = slope;
        slope = slope->z_next;
        slope->z_x = function->z_next->z_x;
        slope->z_y =
            (function->z_next->z_y -
             function->z_y) / (function->z_next->z_x - function->z_x);
    }
    slope->z_next = NIL;
    return (slope_head);
}

int zll_relmaxmin(float slope1_left, float slope1_right, float slope2)
{
    if (slope1_left <= slope2)
    {
        if (slope1_right > slope2)
            return (TRUE);
        else
            return (FALSE);
    }
    if (slope1_right <= slope2)
        return (TRUE);
    return (FALSE);
}

int zll_seek(struct zll *left, struct zll *right, struct zllf *slope_ptr,
             float ray_slope, float threshold);

struct zll *zll_hop1(struct zll *function, float precision)
{
    struct zll *outfun, *outfun_head;
    struct zll *ray_ptr = function->z_next;
    struct zllf *killptr, *slopes;
    long x1, x2, y1, y2, x2_x1, y2_y1;
    float ray_slope, threshold;

    outfun = outfun_head = zll_new();   /* always  keep the first point  */
    outfun->z_prev = NIL;
    outfun->z_x = function->z_x;
    outfun->z_y = function->z_y;
    killptr = slopes = zll_getslope(function);

    for (NIL; ray_ptr->z_next; ray_ptr = ray_ptr->z_next)
    {
        x1 = function->z_x;
        x2 = ray_ptr->z_next->z_x;
        y1 = function->z_y;
        y2 = ray_ptr->z_next->z_y;
        x2_x1 = x2 - x1;
        y2_y1 = y2 - y1;
        ray_slope = (float) (y2_y1 / x2_x1);
        threshold =
            precision *
            (float) sqrt((double) ((x2_x1 * x2_x1) + (y2_y1 * y2_y1)));

        if (zll_seek
            (function, ray_ptr->z_next, slopes->z_next, ray_slope,
             threshold))
        {

            outfun->z_next = zll_new(); /* append newval and point to end of list  */
            outfun->z_next->z_prev = outfun;    /*   build outfun from left to right   */
            outfun = outfun->z_next;
            outfun->z_next = NIL;
            outfun->z_x = ray_ptr->z_x;
            outfun->z_y = ray_ptr->z_y;

            for (NIL; x2 > slopes->z_next->z_x; slopes = slopes->z_next);
            function = ray_ptr;
        }
    }
    /*  no ray_ptr->z_next  so return  */
    outfun->z_next = zll_new(); /* append last point to end of output fun */
    outfun->z_next->z_prev = outfun;    /*  build outfun from left to right  */
    outfun = outfun->z_next;
    outfun->z_next = NIL;
    outfun->z_x = ray_ptr->z_x;
    outfun->z_y = ray_ptr->z_y;
    zllf_freefun(killptr);
    return (outfun_head);       /*  finally return pointer to reduced data */
}

int zll_seek(struct zll *left, struct zll *right, struct zllf *slope_ptr,
             float ray_slope, float threshold)
{
    struct zll *current_point = left->z_next;
    int flag;

    for (NIL; current_point->z_x < right->z_x;
         current_point = current_point->z_next, slope_ptr =
         slope_ptr->z_next)
    {

        flag =
            zll_relmaxmin(slope_ptr->z_y, slope_ptr->z_next->z_y,
                          ray_slope);
        if (flag)
        {
            if (threshold <
                zll_distance(current_point->z_x, current_point->z_y,
                             left->z_x, left->z_y, right->z_x, right->z_y))
                return (TRUE);
        }
    }
    return (FALSE);
}

struct zll *zll_reduce(struct zll *zllist, float tolerance)
{
    struct zll *reduced_list;

    if (!zllist)
        VBAG("no data, quoi");
    if (!zllist->z_next)
        VBAG("too short to reduce, quoi");

    reduced_list = zll_hop1(zllist, tolerance);
    return (reduced_list);
}


/*    ++++++++++++++++++++++++++++++++++++++++++FUN-BUFFER OBJECT DEFINITION IN FUN.H    */

struct zll *funbuff_global_undo = NIL;  /*  GLOBAL funbuff undobuffer   */
struct zll *funbuff_global_clip = NIL;  /*  GLOBAL funbuff clipboard   */

void funbuff_print(struct funbuff *x)
{
    struct zll *list;
    list = &x->f_head;
    post("funbuff: selectX= %ld  selectW= %ld ", x->f_selectX,
         x->f_selectW);
    zll_prinlist(list);
}

/* select region x and width */
void funbuff_select(struct funbuff *x, struct symbol *s, int ac,
                    struct atom *av)
{
    if (ac < 2)
        return;
    if (av->a_type == A_LONG)
        x->f_selectX = av->a_w.w_long;
    else if (av->a_type == A_FLOAT)
        x->f_selectX = (long) (av->a_w.w_long);
    av++;
    if (av->a_type == A_LONG)
    {
        if (av->a_w.w_long < 0)
            VBAG("funbuff_select:  arg2 (width)  must be non-negative");
        x->f_selectW = av->a_w.w_long;
    }
    else if (av->a_type == A_FLOAT)
    {
        if (av->a_w.w_long < 0)
            VBAG("funbuff_select:  arg2 (width)  must be non-negative");
        x->f_selectW = (long) (av->a_w.w_long);
    }
}


/* returns next element and a difference value between its x value and the x value of the 
   event just before; UNLESS "GOTO n" has been called, in which case 
   the difference value is the x of next event - n */

void funbuff_next(struct funbuff *x)
{
    long delta_x, y;

    if (!funbuff_first(x))
        return;                 /* funbuff:  no data, quoi  */
    if (!x->f_foot)
    {
        outlet_bang(x->f_outlet2);      /*  funbuff:   at end of data   send bang out outlet1 */
        return;
    }
    if (x->f_foot == funbuff_first(x))  /* (!x->f_foot->z_prev)  */
    {
        delta_x = (x->f_gotoDelta > -1) ? x->f_gotoDelta : x->f_foot->z_x;
        y = x->f_foot->z_y;
    }
    else
    {
        delta_x =
            (x->f_gotoDelta >
             -1) ? x->f_gotoDelta : x->f_foot->z_x -
            x->f_foot->z_prev->z_x;
        y = x->f_foot->z_y;
    }
    x->f_foot = x->f_foot->z_next;
    x->f_gotoDelta = -1;
    outlet_int(x->f_outlet1, delta_x);
    outlet_int(x->f_outlet, y);
}

struct zll *funbuff_getElement(struct funbuff *x, long int index)
{
    struct zll *list = x->f_foot ? x->f_foot : funbuff_first(x);

    if (!funbuff_first(x))
    {
        /*  post("funbuff_fix:  buffer empty");  */
        return (NIL);
    }
    if (index < list->z_x)
    {
        list = list->z_prev;
        for (list; list->z_prev; list = list->z_prev)
            if (index >= list->z_x)
            {
                x->f_foot = list;
                return (list);
            }
        x->f_foot = list->z_next;
        return (list->z_next);
    }
    for (list; list->z_x <= index; list = list->z_next)
        if (!list->z_next)
        {
            x->f_foot = list;
            return (list);
        }
    x->f_foot = list->z_prev;
    return (list->z_prev);
}

/*  works in conjunction with "NEXT".  Using "GOTO n",  the next call to "NEXT" will be 
    relative to N instead of the x value of the previous event returned by NEXT"  */

void funbuff_goto(struct funbuff *x, long int index)
{
    if (!funbuff_first(x))
        return;

    funbuff_getElement(x, index);
    if (index <= x->f_foot->z_x)
        x->f_gotoDelta = x->f_foot->z_x - index;
    else
    {
        x->f_foot = x->f_foot->z_next;
        if (x->f_foot)
            x->f_gotoDelta = x->f_foot->z_x - index;
        else
            x->f_gotoDelta = 0;
    }
}

struct zll *funbuff_write(struct funbuff *x, struct zll *list)  /*  returns overwriten list  or NIL  */
/* list must be NIL terminated  */
{
    struct zll *current;
    struct zll *kill, *splice1, *splice2 = NIL, *start, *end;
    long x1, x2;


    if (!list)
    {
        post("funbuff_write: no data to write");
        return (NIL);
    }
    start = list;
    end = zll_last(list);
    x1 = start->z_x;
    x2 = end->z_x;
    if (!funbuff_first(x))
    {                           /* if funbuff is empty then attach list     */
        x->f_head.z_next = list;
        list->z_prev = &x->f_head;
        x->f_foot = end;
        return (NIL);
    }
    for (current = funbuff_getElement(x, x1); current->z_next;
         current = current->z_next)
    {
        if (current->z_x >= x1)
            break;
    }
    x->f_foot = end;
    if (x1 > current->z_x)
    {                           /*  append list  */
        current->z_next = list;
        list->z_prev = current;
        return (NIL);
    }
    splice1 = current->z_prev;  /*   insert list start   */
    splice1->z_next = list;
    list->z_prev = splice1;
    if (x1 < current->z_x && x2 < current->z_x)
    {                           /*  insert list end    no overwrite  */
        end->z_next = current;
        current->z_prev = end;
        return (NIL);
    }
    kill = current;
    for (NIL; current; current = current->z_next)
    {
        if (current->z_x > x2)
        {
            splice2 = current;
            break;
        }
    }
    if (splice2)
    {                           /*     insert list end  and NIL cap overwrite   */
        splice2->z_prev->z_next = NIL;
        end->z_next = splice2;
        splice2->z_prev = end;
    }
    return (kill);
}

/* method for inlet */
void funbuff_in1(struct funbuff *x, long int n)
{
    x->f_y = n;
    x->f_yvalid = 1;
}

/* sends nearest value (<=)  index  or lowest val in buffer  */
void funbuff_int(struct funbuff *x, long int index)
{
    if (x->f_yvalid)
    {
        struct zll *list, *element = zll_new();
        element->z_x = index;
        element->z_y = x->f_y;
        element->z_next = NIL;

        list = funbuff_write(x, element);
        zll_freefun(list);
        x->f_yvalid = 0;
    }
    else
    {
        struct zll *list;
        list = funbuff_getElement(x, index);
        if (list)
            outlet_int(x->f_outlet, list->z_y);
    }
}

void funbuff_interp(struct funbuff *x, long int n)
{
    struct zll *z = funbuff_getElement(x, n), *z2;
    if (!z)
        outlet_int(x->f_outlet, 0L);
    else if (n <= z->z_x)
        outlet_int(x->f_outlet, z->z_y);
    else
    {
        long x2, y2;
        for (; z2 = z->z_next; z = z->z_next)
            if (n <= (x2 = z2->z_x))
            {
                long dx = x2 - z->z_x, dy = (y2 = z2->z_y) - z->z_y;
                outlet_int(x->f_outlet, y2 - ((x2 - n) * dy) / dx);
                return;
            }
        outlet_int(x->f_outlet, z->z_y);
    }
}

/* uses the function described in the table named by the symbol in the message-  this method 
   checks each time the length and max/min vals of the table SO the function in the tabled
   is always relative to itself,  that is it is only a shape   */

extern void *tabwind_class;

void funbuff_interptab(struct funbuff *x, struct symbol *s, int argc,
                       struct atom *argv)
{
    struct zll *z;
    long xn, x1, x2, delta_x, tab_domain, tab_range, tab_x, tab_y, i;
    long low = 0, high = 0;
    struct tabwind *tab;
    ELEM *tp;

    if ((argc != 2) || argv->a_type != A_LONG)
        VBAG("funbuff_interptab: bad argument(s) -  args must be <integer  tablename>");
    xn = argv->a_w.w_long;
    argv++;
    if ((argv->a_type != A_SYM)
        || !(tab = (struct tabwind *) argv->a_w.w_sym->s_thing)
        || ob_messlist(tab) != tabwind_class)
        VBAG("funbuff_interptab: bad argument(s)! -  args must be <integer  tablename>");
    z = funbuff_getElement(x, xn);
    if (!z)
    {
        outlet_int(x->f_outlet, 0L);    /*  no data in funbuff output zero?????  */
        return;
    }
    else if ((xn <= z->z_x) || (!z->z_next))
    {
        outlet_int(x->f_outlet, z->z_y);        /*  requested point can't be interpolated */
        return;
    }
    tp = *(tab->w_tab);         /* the value = *tp */
    tab_domain = tab->w_size;
    for (i = 0, high = low = *tp; i < tab_domain; i++, tp++)
        if (*tp < low)
            low = *tp;
        else if (*tp > high)
            high = *tp;
    if ((tab_range = high - low) == 0)
        VBAG("funbuff_interptab: Table must be non zero");
    x1 = z->z_x, x2 = z->z_next->z_x;
    tab_x = tab_domain * ((xn - x1) / (float) (x2 - x1));
    tab_y = *(* (tab->w_tab) + tab_x) - low;
    xn = z->z_y + (z->z_next->z_y - z->z_y) * (float) tab_y / tab_range;
    outlet_int(x->f_outlet, xn);
}

void funbuff_set(struct funbuff *x, struct symbol *s, int ac,
                 struct atom *av)
{
    struct zll *list, *overwrite;

    list = zll_vecToZll(ac, av);
    if (!list)
        VBAG("funbuff_set:  no data to set");
    overwrite = funbuff_write(x, list);
    x->f_foot = funbuff_first(x);
    if (overwrite)
    {
        if (funbuff_global_undo)
            zll_freefun(funbuff_global_undo);
        funbuff_global_undo = overwrite;
    }
}

void funbuff_find(struct funbuff *x, long int yval)
{
    struct zll *list;

    if (!funbuff_first(x))
        VBAG("funbuff_dump: nothing to dump, chump");

    for (list = funbuff_first(x); list; list = list->z_next)
        if (list->z_y == yval)
            outlet_int(x->f_outlet, list->z_x);
}

void funbuff_dump(struct funbuff *x)
{
    struct zll *list;

    if (!funbuff_first(x))
        VBAG("funbuff_dump: nothing to dump, chump");

    for (list = funbuff_first(x); list; list = list->z_next)
    {
        outlet_int(x->f_outlet1, list->z_y);
        outlet_int(x->f_outlet, list->z_x);
    }
}

/*  returns *copy of selected region of list or NIL  */
void funbuff_copy(struct funbuff *x)
{
    struct zll *copy;

    copy = zll_copy(funbuff_first(x), x->f_selectX, x->f_selectW);
    if (funbuff_global_clip)
        zll_freefun(funbuff_global_clip);
    funbuff_global_clip = copy;
}

/* cuts zll in selected region returns cut zll or NIL  */
void funbuff_cut(struct funbuff *x)
{
    struct zll *head;

    x->f_foot = NIL;

    head = zll_cut(funbuff_first(x), x->f_selectX, x->f_selectW);
    if (!head)
        VBAG("funbuff_cut:  no data in selected region, quoi");
    if (funbuff_global_clip)
        zll_freefun(funbuff_global_clip);
    funbuff_global_clip = head;
}

void funbuff_clear(struct funbuff *x)   /* clears zll in selected region */
{
    struct zll *list;

    x->f_foot = NIL;
    list = zll_cut(funbuff_first(x), x->f_selectX, x->f_selectW);
    zll_freefun(list);
    return;
}

void funbuff_reduce(struct funbuff *x, double factor)   /* reduces zll in selected region */
{
    struct zll *newlist, *oldlist;

    x->f_foot = NIL;

    if (!funbuff_first(x))
        VBAG("funbuff_reduce: buffer empty, quoi");
    oldlist = zll_cut(funbuff_first(x), x->f_selectX, x->f_selectW);
    if (!oldlist)
        VBAG("funbuff_reduce: no data to reduce in selected region, quoi");

    newlist = zll_reduce(oldlist, factor);
    funbuff_write(x, newlist);
    if (funbuff_global_undo)
        zll_freefun(funbuff_global_undo);
    funbuff_global_undo = oldlist;
}

void funbuff_undo(struct funbuff *x)    /* takes contents of global undo buffer and writes it into funbuff */
{
    struct zll *relist;

    if (!funbuff_global_undo)
        return;
    relist = funbuff_global_undo;
    funbuff_global_undo = funbuff_write(x, relist);
}

/* pastes a copy corresponding to my selected region from clipboard into myself */
void funbuff_paste(struct funbuff *x)
{
    struct zll *list, *overwrite;

    if (!funbuff_global_clip)
        VBAG("funbuff: clipboard empty, quoi");
    list = zll_copy(funbuff_global_clip, x->f_selectX, x->f_selectW);
    overwrite = funbuff_write(x, list);
    if (funbuff_global_undo)
        zll_freefun(funbuff_global_undo);
    funbuff_global_undo = overwrite;
    return;
}

void funbuff_bang(struct funbuff *x)    /*    DEBUG HERE +++++++++++++++++++++++++++++++++   */
{
    post("funbuff info:    %ld elements long",
         zll_length(funbuff_first(x)));
    if (!funbuff_first(x))
        return;
    else
    {
        struct zll *list = funbuff_first(x);
        long lowX, highX, lowY, highY;

        lowX = highX = list->z_x, lowY = highY = list->z_y;
        for (list = list->z_next; list; list = list->z_next)
        {
            if (list->z_x < lowX)
                lowX = list->z_x;
            else if (list->z_x > highX)
                highX = list->z_x;
            if (list->z_y < lowY)
                lowY = list->z_y;
            else if (list->z_y > highY)
                highY = list->z_y;
        }
        post("  ->  minX= %ld maxX= %ld", lowX, highX);
        post("  ->  minY = %ld maxY = %ld", lowY, highY);
        post("  ->  domain= %ld range = %ld", highX - lowX, highY - lowY);
    }

}

/* sorry -- too lazy to port this right now --- msp 
   
   void *funbuff_readfrom(x, name, volume)
   struct funbuff *x;
   char *name;  /* c string  ---
       {
       struct zll *z;
       short fd, io;
       int bin;
       long i, size, type, *np, **handle;
       void *ret, *binbuf(), *b;
       getfiletype(name, volume, &type);
       if (type != 'TEXT' && type != 'maxb')
       {
       error ("funbuff: bad file type or file not found");
       return;
       }
       b = binbuf();
       /* set up loading environment ---
          binbuf_read(b, name, volume, (bin = (type == 'maxb')));
          if (!binbuf_gethandle(b, &handle, &size, &s_funbuff))     
          {
          if(size & 1)
          {
          post("funbuff_readfrom:  odd data count,  x val without a y val ");
          freeobject((t_object *)b);
          return;
          }
          np = *handle;
          zll_freefun(funbuff_first(x));    /*  clear existing data  ---
                    x->f_head.z_next = z = zll_new();
                    z->z_prev = &x->f_head;
                    z->z_x = *np++;
                    z->z_y = *np++;
                    for (i=0;i < size - 2;i+=2)
                    {
                    z->z_next = zll_new();
                    z->z_next->z_prev = z;
                    z = z->z_next;
                    z->z_x = *np++;
                    if (z->z_prev->z_x >= z->z_x)
                    {
                    post("funbuff_readfrom:  bad data  x values must be increasing");
                    z->z_next = NIL;
                    zll_freefun(funbuff_first(x));
                    x->f_head.z_next = NIL;
                    freeobject((t_object *)b);
                    return;
                    }
                    z->z_y = *np++;
                    }
                    z->z_next = NIL;
                    post("funbuff_read:  %s read!!", name);
                    }
                    else post("funbuff_read:  %s can't be read by funbuff, quoi", name);
                    freeobject((t_object *)b);
                    }
                    
                    void *funbuff_read(x)
                    struct funbuff *x;
                    {
                    SFTypeList types;
                    char fn[256];
                    int vol;
                    long type;
                    
                    types[0]='TEXT';
                    types[1]='maxb';
                    fn[0] = 0;
                    if (getfile(fn, &vol, &type, types, 2)) return;
                    funbuff_readfrom(x, fn, vol);
                    }
                    
                    funbuff_writeto(x, fn, vol, bin)
                    struct funbuff *x;
                    char *fn;
                    int vol;
                    {
                    void *binbuf(), *b = binbuf();
                    long **handle;
                    long n, *np;
                    struct zll *z = funbuff_first(x);
                    n = zll_length(z);
                    handle = (long **)t_newhandle(2 * sizeof(long) * n);
                    for (np = *handle; z; z = z->z_next)
                    {
                    *np++ = z->z_x;
                    *np++ = z->z_y;
                    }
                    vfun(b, binbuf_set, NIL, "s",  &s_funbuff);
                    binbuf_inshandle(b, handle, 2*n);
                    binbuf_write(b, fn, vol, bin);
                    freeobject((t_object *)b);
                    }
                    
                    void *funbuff_filewrite(x)
                    struct funbuff *x;
                    {
                    char fn[256];
                    int vol, bin=0;
                    
                    if (!funbuff_first(x)) VBAG("funbuff_write:  no data to write, quoi");
                    
                    strcpy(fn, (x->f_sym ? x->f_sym->s_name : f_fileprompt));
                    if (getnewfile(fn, &vol, &bin)) return;
                    funbuff_writeto(x, fn, vol, bin);
                    }
                    */

void *funbuff_class;
void funbuff_setup()
{
    c_new(&s_funbuff, NIL, sizeof(t_funbuff), 0, 0);
    c_addmess(funbuff_bang, &s_bang, 0);
    c_addmess(funbuff_int, &s_int, A_LONG, 0);
    /*c_addmess(funbuff_filewrite,&s_write, 0);
       c_addmess(funbuff_read,   &s_read,  0); */
    c_addmess(stdfloat, &s_float, A_FLOAT, 0);
    c_addmess(funbuff_in1, &s_in1, A_LONG, 0);
    c_addmess(funbuff_next, &s_next, 0);
    c_addmess(funbuff_goto, &s_goto, A_LONG, 0);
    c_addmess(funbuff_reduce, &s_reduce, A_FLOAT, 0);
    c_addmess(funbuff_interp, &s_interp, A_LONG, 0);
    c_addmess(funbuff_print, &s_print, 0);
    c_addmess(funbuff_set, &s_set, A_GIMME, 0);
    c_addmess(funbuff_clear, &s_clear, 0);
    c_addmess(funbuff_interptab, &s_interptab, A_GIMME, 0);
#ifdef EXPERIMENT
    c_addmess(funbuff_cut, &s_cut, 0);
    c_addmess(funbuff_copy, &s_copy, 0);
    c_addmess(funbuff_paste, &s_paste, 0);
    c_addmess(funbuff_dump, &s_dump, 0);
    c_addmess(funbuff_find, &s_find, A_LONG, 0);
    c_addmess(funbuff_undo, &s_undo, 0);
    c_addmess(funbuff_select, &s_select, A_GIMME, 0);
#endif
}

void *funbuff_new(struct symbol *s, int argc, struct atom *argv)
{
    struct funbuff *x =
        (struct funbuff *) obj_new((t_messlist **) & funbuff_class,
                                   funbuff_setup);
    x->f_outlet2 = outlet(x, &s_bang);
    x->f_outlet1 = outlet(x, &s_int);
    x->f_outlet = outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in1);
    x->f_head.z_prev = NIL;
    x->f_head.z_next = NIL;
    x->f_head.z_x = -0x0FFFFFFFF;
    x->f_head.z_y = -0x0FFFFFFFF;
    x->f_selectX = -0x040000000;
    x->f_selectW = 0x07FFFFFFF;
    x->f_foot = NIL;
    x->f_yvalid = 0;
    x->f_gotoDelta = -1L;
    if (argc && argv->a_type == A_SYM)
    {
        x->f_sym = argv->a_w.w_sym;
        /* funbuff_readfrom(x, x->f_sym->s_name, defvolume, 0); */
    }
    else
        x->f_sym = 0;
    return (x);
}
