#ifndef lint
static char rcsid[] =
    "$Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:06:07 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.5 $ IRCAM $Date: 1994/01/04 18:06:07 $
 *
 * FTS by Miller Puckette
 *
 */


/* objects by Zack Settel */
#include "m_mess.h"

extern struct symbol s_in1, s_in2, s_bangbang, s_set, s_int, s_symbol,
    s_split, s_absval, s_drunk, s_sum, s_select, s_route;

void *outlet(void *op, struct symbol *s2), *inlet(void *op, void *who,
                                                  struct symbol *s1,
                                                  struct symbol *s2);

#define MAXOUTS 10

void *outlet(void *op, struct symbol *s2);

/*  BANG    modified zack   7/88        optional args allow for upto MAXVAL outlets which fire
    from left to right */

typedef struct bang {
    struct object b_ob;
    void *b_outlets[MAXOUTS];
    int b_outletcount;
} t_bang;

void bang_bang(struct bang *x)
{
    int i;

    for (i = 0; i < x->b_outletcount; i++)
        outlet_bang(x->b_outlets[i]);
}

void *bang_class;
void bang_setup()
{
    c_new(&s_bangbang, NIL, sizeof(t_bang), 0, 0);
    c_addbang(bang_bang);
    c_addint(bang_bang);
    c_addfloat(bang_bang);
    c_addmess(bang_bang, &s_anything, A_GIMME, 0);
}

void *bangbang_new(long int n)  /* msp version -- right to left */
{
    void **baba;
    struct bang *x =
        (struct bang *) obj_new((t_messlist **) & bang_class, bang_setup);
    if (n <= 0)
        n = 2;
    else if (n > MAXOUTS)
        n = MAXOUTS;
    x->b_outletcount = n;
    for (baba = x->b_outlets; n--; baba++)
        *baba = outlet(x, &s_bang);
    return (x);
}

typedef struct split {
    struct object z_ob;
    void *z_outlet1;
    void *z_outlet2;
    long z_n1;
    long z_n2;
} t_split;

void split_int(struct split *x, long int n)
{
    if (n >= x->z_n1 && n <= x->z_n2)
        outlet_int(x->z_outlet2, n);
    else
        outlet_int(x->z_outlet1, n);
}

void split_in1(struct split *x, long int n)
{
    x->z_n1 = n;
}

void split_in2(struct split *x, long int n)
{
    x->z_n2 = n;
}

void *split_class;
void split_setup()
{
    c_new(&s_split, NIL, sizeof(t_split), 0, 0);
    c_addint(split_int);
    c_addfloat(stdfloat);
    c_addinx(split_in1, 1);
    c_addinx(split_in2, 2);
}

void *split_new(long int n1, long int n2)
{
    struct split *x =
        (struct split *) obj_new((t_messlist **) & split_class,
                                 split_setup);
    x->z_outlet1 = outlet(x, &s_int);
    x->z_outlet2 = outlet(x, &s_int);
    inlet(x, x, &s_int, &s_in2);
    inlet(x, x, &s_int, &s_in1);
    x->z_n1 = (n1 < n2) ? n1 : n2;
    x->z_n2 = (n2 > n1) ? n2 : n1;
    return (x);
}

/*  select object    zack   4/88  replaces old-fashioned "sel"  */

typedef struct sel {
    struct object t_ob;
    void *t_outlet;
    void *t_outlets[MAXOUTS];
    int t_triglist[MAXOUTS];
    int t_outletcount;
} t_sel;

void sel_int(struct sel *x, long int n)
{
    int i;

    for (i = 0; i < x->t_outletcount; i++)
    {
        if (n == x->t_triglist[i])
        {
            outlet_bang(x->t_outlets[i]);
            return;
        }
    }
    outlet_int(x->t_outlet, n);
}

void sel_float(struct sel *x, double f)
{
    sel_int(x, (long) f);
}

void *sel_class;
void sel_setup()
{
    c_new(&s_select, NIL, sizeof(t_sel), 0, 0);
    c_addint(sel_int);
    c_addfloat(sel_float);
}

void *sel1_new(struct symbol *sel, int argc, struct atom *argv);
void *zkselect_new(struct symbol *sel, int argc, struct atom *argv)
{
    int count;
    struct sel *x;
    if (argc < 2)
        return (sel1_new(NIL, argc, argv));
    if (argc > MAXOUTS)
    {
        error("select: only %d arguments allowed", MAXOUTS);
        return (0);
    }
    x = (struct sel *) obj_new((t_messlist **) & sel_class, sel_setup);
    x->t_outlet = outlet(x, &s_int);    /*   standard outlet    */
    for (count = 0; count < argc; count++, argv++)
    {
        if (argv->a_type != A_LONG)
        {
            warn("select: bad argument replaced with 0");
            x->t_triglist[count] = 0;
        }
        else
            x->t_triglist[count] = (int) argv->a_w.w_long;
        x->t_outlets[argc - 1 - count] = outlet(x, &s_bang);
    }
    x->t_outletcount = argc;
    return (x);
}

typedef struct sel1 {
    struct object s_ob;
    void *s_out2;
    long s_n;
} t_sel1;

void sel1_int(struct sel1 *x, long int n)
{
    if (n == x->s_n)
        outlet_bang(x->s_ob.o_outlet);
    else
        outlet_int(x->s_out2, n);
}

void sel1_in1(struct sel1 *x, long int n)
{
    x->s_n = n;
}

void sel1_float(struct sel1 *x, double f)
{
    long n = f;
    if (n == x->s_n)
        outlet_bang(x->s_ob.o_outlet);
    else
        outlet_int(x->s_out2, n);
}

void *sel1_class;
void sel1_setup()
{
    c_new(&s_select, NIL, sizeof(t_sel1), 0, 0);
    c_addint(sel1_int);
    c_addfloat(sel1_float);
    c_addinx(sel1_in1, 1);
}

void *sel1_new(struct symbol *sel, int argc, struct atom *argv)
{
    struct sel1 *x;
    long n;
    if (!argc)
        n = 0;
    else
    {
        if (argv->a_type == A_LONG)
            n = argv->a_w.w_long;
        else if (argv->a_type == A_FLOAT)
            n = argv->a_w.w_float;
        else
            n = 0;
    }
    x = (struct sel1 *) obj_new((t_messlist **) & sel1_class, sel1_setup);
    x->s_out2 = outlet(x, &s_int);
    outlet(x, &s_bang);
    inlet(x, x, &s_int, &s_in1);        /* inlet to change selected value */
    x->s_n = n;
    return (x);
}


/* MESFILT  -Zack may '89  Message type parser AND filter object. Input is routed by
   matching message type OR first element, to corresponding outlet.  Messages
   routed by a first element match are stripped of the matched (first) element */

#define A_SYM_MATCH 127
#define NARGS 10
#define TRUE 1
#define FALSE 0

struct boa {
    void *b_outlet;
    struct symbol *b_sym;
};

typedef struct messfilt {
    struct object m_ob;
    int m_argc;
    void *m_outlet;
    struct boa m_boa[NARGS];
} t_messfilt;

void messfilt_int(struct messfilt *x, long int n)
{
    struct boa *bp = x->m_boa + x->m_argc;
    int ac = x->m_argc;
    char matchFlag = FALSE;

    while (ac--)
        if ((--bp)->b_sym == &s_int)
            matchFlag = TRUE, outlet_int(bp->b_outlet, n);
    if (!matchFlag)
        outlet_int(x->m_outlet, n);
}

void messfilt_float(struct messfilt *x, double f)
{
    struct boa *bp = x->m_boa + x->m_argc;
    int ac = x->m_argc;
    char matchFlag = FALSE;

    while (ac--)
        if ((--bp)->b_sym == &s_float)
            matchFlag = TRUE, outlet_float(bp->b_outlet, f);
    if (!matchFlag)
        outlet_float(x->m_outlet, f);
}

/* passes the input message (stripped of selector) thru any
   outlets corresponding to the stripped selector  */

void messfilt_anything(struct messfilt *x, struct symbol *s, int argc,
                       struct atom *argv)
{
    struct boa *bp = x->m_boa + x->m_argc;
    int ac = x->m_argc;
    char matchFlag = FALSE;

    while (ac--)
        if ((--bp)->b_sym == s)
        {
            matchFlag = TRUE;
            if (!argc)
                outlet_bang(bp->b_outlet);
            else if (argv->a_type == A_SYM)
                outlet_anything(bp->b_outlet, argv->a_w.w_sym, argc - 1,
                                argv + 1);
            else if (argv->a_type == A_LONG)
            {
                if (argc == 1)
                    outlet_int(bp->b_outlet, argv->a_w.w_long);
                else
                    outlet_list(bp->b_outlet, NIL, argc, argv);
            }
            else if (argv->a_type == A_FLOAT)
            {
                if (argc == 1)
                    outlet_float(bp->b_outlet, argv->a_w.w_float);
                else
                    outlet_list(bp->b_outlet, NIL, argc, argv);
            }
        }
    if (!matchFlag)
        outlet_anything(x->m_outlet, s, argc, argv);
}

void messfilt_list(struct messfilt *x, struct symbol *s, int argc,
                   struct atom *argv)
{
    struct boa *bp = x->m_boa + x->m_argc;
    int ac = x->m_argc;
    char matchFlag = FALSE;

    while (ac--)
        if ((--bp)->b_sym == &s_list)
            matchFlag = TRUE, outlet_list(bp->b_outlet, s, argc, argv);
    if (!matchFlag)
        outlet_list(x->m_outlet, s, argc, argv);
}

void messfilt_bang(struct messfilt *x)
{
    struct boa *bp = x->m_boa + x->m_argc;
    int ac = x->m_argc;
    char matchFlag = FALSE;

    while (ac--)
        if ((--bp)->b_sym == &s_bang)
            matchFlag = TRUE, outlet_bang(bp->b_outlet);
    if (!matchFlag)
        outlet_bang(x->m_outlet);
}

void *messfilt_class;
void messfilt_setup()
{
    c_new(&s_route, NIL, sizeof(t_messfilt), 0, 0);
    c_addbang(messfilt_bang);
    c_addint(messfilt_int);
    c_addfloat(messfilt_float);
    c_addmess(messfilt_list, &s_list, A_GIMME, 0);
    c_addmess(messfilt_anything, &s_anything, A_GIMME, 0);
}

/*  arguments are all assumed to be symbols */
void *messfilt_new(struct symbol *s, int ac, struct atom *av)
{
    struct messfilt *x =
        (struct messfilt *) obj_new((t_messlist **) & messfilt_class,
                                    messfilt_setup);
    struct boa *bp;

    if (ac > NARGS)
        ac = NARGS;
    x->m_argc = ac;
    av += ac;
    bp = x->m_boa + ac;

    /*   rightmost failure outlet for unmatched message types   */
    x->m_outlet = outlet(x, NIL);
    while (ac--)
    {
        av--;
        bp--;
        bp->b_sym = av->a_w.w_sym;
        bp->b_outlet = outlet(x, NIL);  /* hooks up to all types */
    }
    return (x);
}

/* strip  -Zack may '89  list filter object. Input (lists) are routed by
   matching the list's first element to the integer argument assoiated with a given 
   outlet.  The output lists are stripped of the matched (first) element */

struct ipair {
    long b_match;
    void *b_outlet;
};

typedef struct strip {
    struct object s_ob;
    void *s_outlet;
    int s_argc;
    struct ipair s_ipair[NARGS];
} t_strip;

void strip_int(struct strip *x, long int n)
{
    struct ipair *bp = x->s_ipair + x->s_argc;
    int ac = x->s_argc;
    char matchFlag = 0;

    while (ac--)
        if ((--bp)->b_match == n)
        {
            matchFlag = 1;
            outlet_bang(bp->b_outlet);
        }
    if (!matchFlag)
        outlet_int(x->s_outlet, n);
}

void strip_list(struct strip *x, struct symbol *s, int argc,
                struct atom *argv)
{
    struct ipair *bp = x->s_ipair + x->s_argc;
    int ac = x->s_argc;
    char matchFlag = 0;
    long matchme = argv->a_w.w_long;

    if (!argc)
        return;                 /* to hell with it */
    argc--;
    argv++;
    while (ac--)
        if ((--bp)->b_match == matchme)
        {
            matchFlag = 1;
            if (!argc)
                outlet_bang(bp->b_outlet);
            else if (argv->a_type == A_SYM)
                outlet_anything(bp->b_outlet, argv->a_w.w_sym, argc - 1,
                                argv + 1);
            else if (argv->a_type == A_LONG)
            {
                if (argc == 1)
                    outlet_int(bp->b_outlet, argv->a_w.w_long);
                else
                    outlet_list(bp->b_outlet, NIL, argc, argv);
            }
            else if (argv->a_type == A_FLOAT)
            {
                if (argc == 1)
                    outlet_float(bp->b_outlet, argv->a_w.w_float);
                else
                    outlet_list(bp->b_outlet, NIL, argc, argv);
            }
        }
    if (!matchFlag)
        outlet_list(x->s_outlet, s, argc + 1, argv - 1);
}

void *strip_class;
void strip_setup()
{
    c_new(&s_route, NIL, sizeof(t_strip), 0, 0);
    c_addint(strip_int);
    c_addfloat(stdfloat);
    c_addmess(strip_list, &s_list, A_GIMME, 0);
}

/*  arguments: up to NARGS integer arguments  */
void *strip_new(struct symbol *s, int ac, struct atom *av)
{
    int i, type;
    struct ipair *bp;
    struct atom *av2;
    struct strip *x;

    /* if no value for AC return a reasonable default */
    if (!ac)
    {
        struct atom at;
        SETLONG(&at, 0);
        return (strip_new(s, 1, &at));
    }
    if ((type = av->a_type) != A_LONG && type != A_SYM)
        return (0);
    if (ac > NARGS)
    {
        error("strip: too many arguments");
        return (0);
    }
    for (i = ac, av2 = av; i--; av2++)
        if (av2->a_type != type)
        {
            error("strip: argument types don't match");
            return (0);
        }
    if (type == A_SYM)
        return (messfilt_new((t_symbol *) x, ac, av));

    x = (struct strip *) obj_new((t_messlist **) & strip_class,
                                 strip_setup);
    /* failure outlet through which passes un-matched input: */
    x->s_outlet = outlet(x, NIL);
    x->s_argc = 0;
    av += ac;
    bp = x->s_ipair;
    while (ac--)
    {
        av--;
        bp->b_match = av->a_w.w_long;
        (bp++)->b_outlet = outlet(x, &s_int);
        x->s_argc++;
    }
    return (x);
}
