#ifndef lint
static char rcsid[] =
    "$Revision: 2.6 $ IRCAM $Date: 1994/01/29 07:50:03 $";
#endif

/*
 *                      Copyright (c) 1993 by IRCAM
 *                          All rights reserved.
 *
 *  For any information regarding this and other IRCAM software, please
 *  send email to:
 *                              manager@ircam.fr
 *
 *      $Revision: 2.6 $ IRCAM $Date: 1994/01/29 07:50:03 $
 *
 * FTS by Miller Puckette
 *
 */

#include "m_mess.h"
#include "d_graph.h"
#include "d_ugen.h"

/******************** vuckterbean ***********************/

typedef struct vuckterbean {
    t_object f_ob;
    t_symbol *f_sym;
    long f_len;                 /* lentgth of FFT */
} t_vuckterbean;

#define TOOLONG 2050            /* dont deal with long shit in a ctl. period */

extern void *sigtable_class;
extern t_symbol s_set, s_in1, s_hook, s_vuckterbean, s_put, s_bang, s_jump,
    s_ft1, s__sigboner;

void vuckterbean_float(t_vuckterbean *x, double f)
{
    float onset;
    t_sigtable *wt;
    /* chase the symbol to find the table and verify it's really a sample */
    if ((wt = (t_sigtable *) (x->f_sym->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
    {
        t_sampbuf *b = &wt->x_buf;
        /* convert milliseconds to samples */
        float onset = b->b_conv * f;
        int rats = onset;
        if (onset < 0 || rats >= b->b_nsamps)
            outlet_float(x->f_ob.o_outlet, 0.);
        else
        {
            double mofo = b->b_shit[rats];
            outlet_float(x->f_ob.o_outlet, mofo);
        }
    }
}

void vuckterbean_int(t_vuckterbean *x, long int n)
{
    vuckterbean_float(x, (double) n);
}

void vuckterbean_set(t_vuckterbean *x, t_symbol *s)
{
    x->f_sym = s;
}

void vuckterbean_shit(t_vuckterbean *x) /* compute center of mass of fft in table */
{
    float onset;
    t_sigtable *wt;
    /* chase the symbol to find the table and verify it's really a sample */
    if ((wt = (t_sigtable *) (x->f_sym->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
    {
        t_sampbuf *b = &wt->x_buf;
        long z;
        double result;
        double top = 0., bottom = 0.;

        if (b->b_nsamps < x->f_len)
        {
            error("vuckterbean_shit: samptab too short");
            return;
        }
        else
        {
            for (z = 0; z < x->f_len; z++)
            {
                top += (z * b->b_shit[z]);
                bottom += b->b_shit[z];
            }
            result = (bottom) ? top / bottom : 0.;
            outlet_float(x->f_ob.o_outlet, result);
        }
    }
}

void vuckterbean_in1(t_vuckterbean *x, long int n)
{
    char buf[30];
    gensampname(buf, n);
    vuckterbean_set(x, gensym(buf));
}

void vuckterbean_setup()
{
    c_new(&s_vuckterbean, NIL, sizeof(t_vuckterbean), 0, 0);
    c_addfloat(vuckterbean_float);
    c_addinx(vuckterbean_in1, 1);
    c_addmess(vuckterbean_set, &s_set, A_SYM, 0);
    c_addmess(vuckterbean_shit, &s_hook, 0);
    c_addint(vuckterbean_int);
}

void *vuckterbean_class;

void *vuckterbean_new(t_symbol *s, long int n)
{
    t_vuckterbean *x;
    if (!sampfilt_tab && !make_samp_tab())
    {
        error("sampread: out of memory");
        return (0);
    }
    else if (n < 0 || n > 1024)
    {
        error("2nd argument out of range -1024 maxium");
        return (0);
    }
    x = (t_vuckterbean *) obj_new((t_messlist **) & vuckterbean_class,
                                  vuckterbean_setup);
    outlet_new(x, &s_float);
    inlet_new(x, x, &s_int, &s_in1);
    x->f_sym = s;
    x->f_len = n;
    return (x);
}

/******************** boner ***********************/
/* also allows for reading and writing single samples to samptab */
typedef struct srctl {
    long c_FFTlen;              /* FFT size */
    long c_offset;              /* sample offset (grows by the vectorSize) */
    long c_wrap;                /* used for index wrap */
    t_sampbuf *c_buf;           /* samp tab buffer */
} t_srctl;

typedef struct boner {
    t_head x_h;
    t_sig *x_io[IN1 + OUT1];
    t_srctl *x_ctl;
    t_srctl x_cspace;
    t_symbol *x_sym;            /* symbol bound to table we'll use */
    void *x_outlet1;            /* "float" for reading */
    float x_yval;               /* value to write ot samptab */
    int x_state;                /* inlet state - table style */
} t_boner;

void cu_boner(float *in1, float *out, t_srctl *x, long int n)
{
    long z;
    t_sampbuf *b = x->c_buf;

    if (!b)
        goto zero;

    z = x->c_offset;
    x->c_offset = (x->c_offset + n) & x->c_wrap;
    while (n--)
        *out++ = b->b_shit[z++];
    return;
  zero:
    while (n--)
        *out++ = .11111111;
    return;
}

void boner_set(t_boner *x, t_symbol *s)
{
    t_sigtable *wt;
    x->x_sym = s;
    if ((wt = (t_sigtable *) (x->x_sym->s_thing)) &&
        ob_messlist(wt) == sigtable_class)
        x->x_ctl->c_buf = &wt->x_buf;
    else
    {
        error("sampread: %s: no such sample", s->s_name);
        x->x_ctl->c_buf = 0;
        return;
    }
    x->x_ctl->c_offset = 0;
    if (x->x_ctl->c_buf->b_nsamps < x->x_ctl->c_FFTlen)
    {
        error("sampread: %s: samptab not long enough", s->s_name);
        x->x_ctl->c_buf = 0;
    }
}


void boner_bang(t_boner *x)
{
    x->x_ctl->c_offset = 0L;
}

void boner_int(t_boner *x, long int n)  /* read (or write x->x_yval) sample n of samptab */
{
    t_sigtable *wt;

    if ((wt = (t_sigtable *) (x->x_sym->s_thing)) && ob_messlist(wt) == sigtable_class) /* table~ OK ? */
    {
        t_sampbuf *b = &wt->x_buf;

        if (b->b_nsamps < n)
            n = b->b_nsamps;
        else if (n < 0)
            n = 0;

        if (x->x_state)
        {
            b->b_shit[n] = x->x_yval;
            x->x_state = 0;
        }
        else
            outlet_float(x->x_outlet1, (double) b->b_shit[n]);
    }
    else
        error("boner int: table~ %s not found", x->x_sym->s_name);
}

void boner_ft1(t_boner *x, double f)
{
    x->x_yval = (float) f;
    x->x_state = 1;
}

void boner_jump(t_boner *x, double f)   /* jump to a new point */
{
    char buf[30];
}

void boner_put(t_boner *x, long int whether)
{
    if (whether)
    {
        u_stdout((t_ugen *) x);
        boner_set(x, x->x_sym);
        x->x_cspace.c_offset = 0;
        x->x_ctl = dspchain_putstate(u_chain, &x->x_cspace,
                                     sizeof(t_srctl), sizeof(float));
        dspchain_putc(u_chain, cu_boner, 4,
                      x->x_io[0]->s_shit, x->x_io[1]->s_shit,
                      x->x_ctl, x->x_io[0]->s_n);
    }
    else
    {
        bcopy(x->x_ctl, &x->x_cspace, sizeof(t_srctl));
        x->x_ctl = &x->x_cspace;
    }
}

void boner_setup()
{
    c_new(&s__sigboner, u_clean, sizeof(t_boner), 0, 0);
    c_addmess(boner_set, &s_set, A_SYM, 0);
    c_addmess(boner_jump, &s_jump, A_DEFFLOAT, 0);
    c_addint(boner_int);
    c_addftx(boner_ft1, 1);
    c_addbang(boner_bang);
    c_addmess(boner_put, &s_put, A_CANT, 0);
    u_inletmethod(0);
}

void *boner_class;
void *sigboner_new(t_symbol *s, long int n)     /* samptab name and FFT length */
{
    int z = 64, shift = 6;

    t_boner *x;
    if (!sampfilt_tab && !make_samp_tab())
    {
        error("boner: out of memory");
        return (0);
    }
    while (z < 1024)
    {
        if (n == z)
            break;
        z <<= 1;
        shift++;
    }
    if (z > 1024)
    {
        error("arg2 must be power of 2 between 64 and 1024");
        return (0);
    }
    x = (t_boner *) obj_new((t_messlist **) & boner_class, boner_setup);
    x->x_outlet1 = outlet_new(x, &s_float);
    inlet_new(x, x, &s_float, &s_ft1);
    u_setup((t_ugen *) x, IN1, OUT1);
    x->x_sym = s;
    x->x_state = 0;
    x->x_cspace.c_wrap = (long) (1 << shift) - 1;
    x->x_cspace.c_FFTlen = n;
    x->x_cspace.c_offset = 0L;
    x->x_ctl = &x->x_cspace;
    if (u_debug)
        post("boner %lx", x);
    return (x);
}
